(* Content-type: application/vnd.wolfram.mathematica *)

(*** Wolfram Notebook File ***)
(* http://www.wolfram.com/nb *)

(* CreatedBy='Mathematica 9.0' *)

(*CacheID: 234*)
(* Internal cache information:
NotebookFileLineBreakTest
NotebookFileLineBreakTest
NotebookDataPosition[       157,          7]
NotebookDataLength[     15139,        457]
NotebookOptionsPosition[     12081,        368]
NotebookOutlinePosition[     14137,        420]
CellTagsIndexPosition[     14052,        415]
WindowTitle->CoalescentTree - Wolfram Mathematica
WindowFrame->Normal*)

(* Beginning of Notebook Content *)
Notebook[{
Cell[" ", "SymbolColorBar",
 CellMargins->{{Inherited, Inherited}, {-5, 0}}],

Cell[TextData[{
 ButtonBox["Genealogy",
  BaseStyle->{"Link", "LinkTrail"},
  ButtonData->"paclet:Genealogy/guide/Genealogy"],
 StyleBox[" > ", "LinkTrailSeparator"]
}], "LinkTrail"],

Cell[BoxData[GridBox[{
   {Cell["GENEALOGY PACLET SYMBOL", "PacletNameCell"], Cell[TextData[{
     Cell[BoxData[
      ActionMenuBox[
       FrameBox["\<\"More About \[RightGuillemet]\"\>",
        StripOnInput->False], {"\<\"Genealogy\"\>":>Documentation`HelpLookup[
       "paclet:Genealogy/guide/Genealogy"]},
       Appearance->None,
       MenuAppearance->Automatic]],
      LineSpacing->{1.4, 0}],
     "\[ThickSpace]\[ThickSpace]\[ThickSpace]\[ThickSpace]\[ThickSpace]\
\[ThickSpace]",
     Cell[BoxData[
      ActionMenuBox[
       FrameBox["\<\"URL \[RightGuillemet]\"\>",
        StripOnInput->False], {"\<\"Genealogy/ref/CoalescentTree\"\>":>
       None, "\<\"Copy Mathematica url\"\>":>
       Module[{DocumentationSearch`Private`nb$}, 
        DocumentationSearch`Private`nb$ = NotebookPut[
           Notebook[{
             Cell["Genealogy/ref/CoalescentTree"]}, Visible -> False]]; 
        SelectionMove[DocumentationSearch`Private`nb$, All, Notebook]; 
        FrontEndTokenExecute[DocumentationSearch`Private`nb$, "Copy"]; 
        NotebookClose[DocumentationSearch`Private`nb$]; Null], 
       Delimiter, "\<\"Copy web url\"\>":>
       Module[{DocumentationSearch`Private`nb$}, 
        DocumentationSearch`Private`nb$ = NotebookPut[
           Notebook[{
             Cell[
              BoxData[
               MakeBoxes[
                Hyperlink[
                "http://reference.wolfram.com/mathematica/Genealogy/ref/\
CoalescentTree.html"], StandardForm]], "Input", TextClipboardType -> 
              "PlainText"]}, Visible -> False]]; 
        SelectionMove[DocumentationSearch`Private`nb$, All, Notebook]; 
        FrontEndTokenExecute[DocumentationSearch`Private`nb$, "Copy"]; 
        NotebookClose[DocumentationSearch`Private`nb$]; 
        Null], "\<\"Go to web url\"\>":>FrontEndExecute[{
         NotebookLocate[{
           URL[
            StringJoin[
            "http://reference.wolfram.com/mathematica/", 
             "Genealogy/ref/CoalescentTree", ".html"]], None}]}]},
       Appearance->None,
       MenuAppearance->Automatic]],
      LineSpacing->{1.4, 0}]
    }], "AnchorBar"]}
  }]], "AnchorBarGrid",
 GridBoxOptions->{GridBoxItemSize->{"Columns" -> {
     Scaled[0.65], {
      Scaled[0.34]}}, "ColumnsIndexed" -> {}, "Rows" -> {{1.}}, 
   "RowsIndexed" -> {}}},
 CellID->1],

Cell["CoalescentTree", "ObjectName",
 CellID->1224892054],

Cell[BoxData[GridBox[{
   {"", Cell[TextData[{
     Cell[BoxData[
      RowBox[{"CoalescentTree", "[", 
       RowBox[{"tlist", ",", "elist"}], "]"}]], "InlineFormula"],
     " \[LineSeparator]is a data structure for a coalescent tree. Time is \
measured backward in unit of  2 Ne, where Ne is the effective population \
size. We use same lables for tree leaves and  sequences, and they are set as \
integers from 1 to n (the number of sequences).  The internal nodes are \
labled by integers from n+1 to 2 n-1, according the occurent order of \
coalescent events. ",
     StyleBox["tlist",
      FontSlant->"Italic"],
     " is a sequence of coalescent times  starting from zero at tree leaves, \
and ",
     StyleBox["elist",
      FontSlant->"Italic"],
     " is a sequence of branch pars between which coalescence occurs. "
    }]]}
  }]], "Usage",
 GridBoxOptions->{
 GridBoxBackground->{
  "Columns" -> {{None}}, "ColumnsIndexed" -> {}, "Rows" -> {{None}}, 
   "RowsIndexed" -> {}}},
 CellID->982511436],

Cell[CellGroupData[{

Cell[TextData[ButtonBox["EXAMPLES",
 BaseStyle->None,
 Appearance->{Automatic, None},
 Evaluator->None,
 Method->"Preemptive",
 ButtonFunction:>(FrontEndExecute[{
    FrontEnd`SelectionMove[
     FrontEnd`SelectedNotebook[], All, ButtonCell], 
    FrontEndToken["OpenCloseGroup"], 
    FrontEnd`SelectionMove[
     FrontEnd`SelectedNotebook[], After, 
     CellContents]}]& )]], "PrimaryExamplesSection",
 CellTags->"PrimaryExamplesSection",
 CellID->50338187],

Cell[CellGroupData[{

Cell[TextData[{
 "Basic Examples",
 "\[NonBreakingSpace]\[NonBreakingSpace]",
 Cell["(1)", "ExampleCount"]
}], "ExampleSection",
 CellID->509688211],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"<<", "Genealogy`"}]], "Input",
 CellLabel->"In[1]:=",
 CellID->166385708],

Cell[BoxData[
 RowBox[{
  StyleBox[
   RowBox[{"General", "::", "compat"}], "MessageName"], ":", 
  " ", "\<\"Combinatorica Graph and Permutations functionality has been \
superseded by preloaded functionaliy. The package now being loaded may \
conflict with this. Please see the Compatibility Guide for details.\"\>"}]], \
"Message", "MSG",
 CellLabel->"During evaluation of In[1]:=",
 CellID->284564170]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[{
 RowBox[{"tree", "=", 
  RowBox[{"RandomCoalescentTree", "[", "5", "]"}]}], "\[IndentingNewLine]", 
 RowBox[{"Head", "[", "tree", "]"}], "\[IndentingNewLine]", 
 RowBox[{"ShowLabeledGraph", "[", 
  RowBox[{
   RowBox[{"TreeToCombinatoricaGraph", "[", "tree", "]"}], ",", 
   RowBox[{"VertexColor", "\[Rule]", "Red"}], ",", 
   RowBox[{"EdgeColor", "\[Rule]", "Blue"}]}], "]"}]}], "Input",
 CellLabel->"In[2]:=",
 CellID->652827085],

Cell[BoxData[
 RowBox[{"CoalescentTree", "[", 
  RowBox[{
   RowBox[{"{", 
    RowBox[{
    "0", ",", "0", ",", "0", ",", "0", ",", "0", ",", "0.002566104540038227`",
      ",", "0.12375279767208577`", ",", "0.590820602243904`", ",", 
     "1.082198813195186`"}], "}"}], ",", 
   RowBox[{"{", 
    RowBox[{
     RowBox[{"{", 
      RowBox[{
       RowBox[{"{", 
        RowBox[{"3", ",", "6"}], "}"}], ",", 
       RowBox[{"{", 
        RowBox[{"1", ",", "6"}], "}"}]}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{
       RowBox[{"{", 
        RowBox[{"4", ",", "7"}], "}"}], ",", 
       RowBox[{"{", 
        RowBox[{"2", ",", "7"}], "}"}]}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{
       RowBox[{"{", 
        RowBox[{"5", ",", "8"}], "}"}], ",", 
       RowBox[{"{", 
        RowBox[{"7", ",", "8"}], "}"}]}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{
       RowBox[{"{", 
        RowBox[{"6", ",", "9"}], "}"}], ",", 
       RowBox[{"{", 
        RowBox[{"8", ",", "9"}], "}"}]}], "}"}]}], "}"}]}], "]"}]], "Output",
 ImageSize->{478, 33},
 ImageMargins->{{0, 0}, {0, 0}},
 ImageRegion->{{0, 1}, {0, 1}},
 CellLabel->"Out[2]=",
 CellID->560315760],

Cell[BoxData["CoalescentTree"], "Output",
 ImageSize->{102, 15},
 ImageMargins->{{0, 0}, {0, 0}},
 ImageRegion->{{0, 1}, {0, 1}},
 CellLabel->"Out[2]=",
 CellID->130778101],

Cell[BoxData[
 GraphicsBox[{{
    {RGBColor[0, 0, 1], Thickness[0.005], 
     ArrowBox[{{0.2, 0.}, {0.30000000000000004`, 0.0005132209080076454}}], 
     {GrayLevel[0]}}, 
    {RGBColor[0, 0, 1], Thickness[0.005], 
     ArrowBox[{{0.4, 0.}, {0.30000000000000004`, 0.0005132209080076454}}], 
     {GrayLevel[0]}}, 
    {RGBColor[0, 0, 1], Thickness[0.005], 
     ArrowBox[{{0.8, 0.}, {0.9, 0.024750559534417155`}}], 
     {GrayLevel[0]}}, 
    {RGBColor[0, 0, 1], Thickness[0.005], 
     ArrowBox[{{1., 0.}, {0.9, 0.024750559534417155`}}], 
     {GrayLevel[0]}}, 
    {RGBColor[0, 0, 1], Thickness[0.005], 
     ArrowBox[{{0.6000000000000001, 0.}, {0.75, 0.1181641204487808}}], 
     {GrayLevel[0]}}, 
    {RGBColor[0, 0, 1], Thickness[0.005], 
     ArrowBox[{{0.9, 0.024750559534417155`}, {0.75, 0.1181641204487808}}], 
     {GrayLevel[0]}}, 
    {RGBColor[0, 0, 1], Thickness[0.005], 
     ArrowBox[{{0.30000000000000004`, 0.0005132209080076454}, {0.525, 
      0.21643976263903722`}}], 
     {GrayLevel[0]}}, 
    {RGBColor[0, 0, 1], Thickness[0.005], 
     ArrowBox[{{0.75, 0.1181641204487808}, {0.525, 0.21643976263903722`}}], 
     {GrayLevel[0]}}}, {
    {RGBColor[1, 0, 0], 
     {PointSize[0.025], PointBox[{0.4, 0.}]}, 
     {GrayLevel[0], InsetBox["1", Scaled[{-0.02, -0.02}, {0.4, 0.}], {1, 0}]}, 
     {GrayLevel[0]}}, 
    {RGBColor[1, 0, 0], 
     {PointSize[0.025], PointBox[{1., 0.}]}, 
     {GrayLevel[0], InsetBox["2", Scaled[{-0.02, -0.02}, {1., 0.}], {1, 0}]}, 
     {GrayLevel[0]}}, 
    {RGBColor[1, 0, 0], 
     {PointSize[0.025], PointBox[{0.2, 0.}]}, 
     {GrayLevel[0], InsetBox["3", Scaled[{-0.02, -0.02}, {0.2, 0.}], {1, 0}]}, 
     {GrayLevel[0]}}, 
    {RGBColor[1, 0, 0], 
     {PointSize[0.025], PointBox[{0.8, 0.}]}, 
     {GrayLevel[0], InsetBox["4", Scaled[{-0.02, -0.02}, {0.8, 0.}], {1, 0}]}, 
     {GrayLevel[0]}}, 
    {RGBColor[1, 0, 0], 
     {PointSize[0.025], PointBox[{0.6000000000000001, 0.}]}, 
     {GrayLevel[0], 
      InsetBox["5", 
       Scaled[{-0.02, -0.02}, {0.6000000000000001, 0.}], {1, 0}]}, 
     {GrayLevel[0]}}, 
    {RGBColor[1, 0, 0], 
     {PointSize[0.025], 
      PointBox[{0.30000000000000004`, 0.0005132209080076454}]}, 
     {GrayLevel[0], 
      InsetBox["6", 
       Scaled[{-0.02, -0.02}, {0.30000000000000004`, 0.0005132209080076454}], \
{1, 0}]}, 
     {GrayLevel[0]}}, 
    {RGBColor[1, 0, 0], 
     {PointSize[0.025], PointBox[{0.9, 0.024750559534417155`}]}, 
     {GrayLevel[0], 
      InsetBox["7", 
       Scaled[{-0.02, -0.02}, {0.9, 0.024750559534417155`}], {1, 0}]}, 
     {GrayLevel[0]}}, 
    {RGBColor[1, 0, 0], 
     {PointSize[0.025], PointBox[{0.75, 0.1181641204487808}]}, 
     {GrayLevel[0], 
      InsetBox["8", 
       Scaled[{-0.02, -0.02}, {0.75, 0.1181641204487808}], {1, 0}]}, 
     {GrayLevel[0]}}, 
    {RGBColor[1, 0, 0], 
     {PointSize[0.025], PointBox[{0.525, 0.21643976263903722`}]}, 
     {GrayLevel[0], 
      InsetBox["9", 
       Scaled[{-0.02, -0.02}, {0.525, 0.21643976263903722`}], {1, 0}]}, 
     {GrayLevel[0]}}}},
  AlignmentPoint->Center,
  AspectRatio->Automatic,
  Axes->False,
  AxesLabel->None,
  AxesOrigin->Automatic,
  AxesStyle->{},
  Background->None,
  BaseStyle->{},
  BaselinePosition->Automatic,
  ColorOutput->Automatic,
  ContentSelectable->Automatic,
  CoordinatesToolOptions:>Automatic,
  DisplayFunction:>$DisplayFunction,
  Epilog->{},
  FormatType:>TraditionalForm,
  Frame->False,
  FrameLabel->None,
  FrameStyle->{},
  FrameTicks->Automatic,
  FrameTicksStyle->{},
  GridLines->None,
  GridLinesStyle->{},
  ImageMargins->0.,
  ImagePadding->All,
  ImageSize->Automatic,
  ImageSizeRaw->Automatic,
  LabelStyle->{},
  Method->Automatic,
  PlotLabel->None,
  PlotRange->{{0.15000000000000002`, 1.05}, {-0.20844678534714806`, 
   0.42488654798618525`}},
  PlotRangeClipping->False,
  PlotRangePadding->Automatic,
  PlotRegion->Automatic,
  PreserveImageOptions->Automatic,
  Prolog->{},
  RotateLabel->True,
  Ticks->Automatic,
  TicksStyle->{}]], "Output",
 ImageSize->{364, 253},
 ImageMargins->{{0, 0}, {0, 0}},
 ImageRegion->{{0, 1}, {0, 1}},
 CellLabel->"Out[2]=",
 CellID->202030714]
}, Open  ]]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell[TextData[ButtonBox["MORE ABOUT",
 BaseStyle->None,
 Appearance->{Automatic, None},
 Evaluator->None,
 Method->"Preemptive",
 ButtonFunction:>(FrontEndExecute[{
    FrontEnd`SelectionMove[
     FrontEnd`SelectedNotebook[], All, ButtonCell], 
    FrontEndToken["OpenCloseGroup"], 
    FrontEnd`SelectionMove[
     FrontEnd`SelectedNotebook[], After, 
     CellContents]}]& )]], "MoreAboutSection",
 CellID->38303248],

Cell[TextData[ButtonBox["Genealogy",
 BaseStyle->"Link",
 ButtonData->"paclet:Genealogy/guide/Genealogy"]], "MoreAbout",
 CellID->417333301]
}, Open  ]],

Cell[" ", "FooterCell"]
},
Saveable->False,
ScreenStyleEnvironment->"Working",
WindowSize->{725, 730},
WindowMargins->{{0, Automatic}, {Automatic, 0}},
WindowTitle->"CoalescentTree - Wolfram Mathematica",
TaggingRules->{
 "ModificationHighlight" -> False, 
  "Metadata" -> {
   "built" -> "{2013, 7, 27, 22, 6, 59.3687610}", "context" -> "Genealogy`", 
    "keywords" -> {}, "index" -> True, "label" -> "Genealogy Paclet Symbol", 
    "language" -> "en", "paclet" -> "Genealogy", "status" -> "None", 
    "summary" -> 
    "CoalescentTree[tlist, elist] is a data structure for a coalescent tree. \
Time is measured backward in unit of 2 Ne, where Ne is the effective \
population size. We use same lables for tree leaves and sequences, and they \
are set as integers from 1 to n (the number of sequences). The internal nodes \
are labled by integers from n+1 to 2 n-1, according the occurent order of \
coalescent events. tlist is a sequence of coalescent times starting from zero \
at tree leaves, and elist is a sequence of branch pars between which \
coalescence occurs. ", "synonyms" -> {}, "title" -> "CoalescentTree", "type" -> 
    "Symbol", "uri" -> "Genealogy/ref/CoalescentTree"}, "LinkTrails" -> "", 
  "SearchTextTranslated" -> ""},
CellContext->"Global`",
FrontEndVersion->"9.0 for Microsoft Windows (64-bit) (January 25, 2013)",
StyleDefinitions->Notebook[{
   Cell[
    StyleData[
    StyleDefinitions -> FrontEnd`FileName[{"Wolfram"}, "Reference.nb"]]], 
   Cell[
    StyleData["Input"], CellContext -> "Global`"], 
   Cell[
    StyleData["Output"], CellContext -> "Global`"]}, Visible -> False, 
  FrontEndVersion -> "9.0 for Microsoft Windows (64-bit) (January 25, 2013)", 
  StyleDefinitions -> "Default.nb"]
]
(* End of Notebook Content *)

(* Internal cache information *)
(*CellTagsOutline
CellTagsIndex->{
 "PrimaryExamplesSection"->{
  Cell[4295, 117, 460, 13, 70, "PrimaryExamplesSection",
   CellTags->"PrimaryExamplesSection",
   CellID->50338187]}
 }
*)
(*CellTagsIndex
CellTagsIndex->{
 {"PrimaryExamplesSection", 13910, 408}
 }
*)
(*NotebookFileOutline
Notebook[{
Cell[607, 21, 76, 1, 70, "SymbolColorBar"],
Cell[686, 24, 182, 5, 70, "LinkTrail"],
Cell[871, 31, 2327, 53, 70, "AnchorBarGrid",
 CellID->1],
Cell[3201, 86, 57, 1, 70, "ObjectName",
 CellID->1224892054],
Cell[3261, 89, 1009, 24, 70, "Usage",
 CellID->982511436],
Cell[CellGroupData[{
Cell[4295, 117, 460, 13, 70, "PrimaryExamplesSection",
 CellTags->"PrimaryExamplesSection",
 CellID->50338187],
Cell[CellGroupData[{
Cell[4780, 134, 148, 5, 70, "ExampleSection",
 CellID->509688211],
Cell[CellGroupData[{
Cell[4953, 143, 97, 3, 70, "Input",
 CellID->166385708],
Cell[5053, 148, 405, 9, 70, "Message",
 CellID->284564170]
}, Open  ]],
Cell[CellGroupData[{
Cell[5495, 162, 446, 10, 70, "Input",
 CellID->652827085],
Cell[5944, 174, 1168, 38, 54, "Output",
 CellID->560315760],
Cell[7115, 214, 172, 5, 36, "Output",
 CellID->130778101],
Cell[7290, 221, 4126, 118, 274, "Output",
 CellID->202030714]
}, Open  ]]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[11477, 346, 419, 12, 70, "MoreAboutSection",
 CellID->38303248],
Cell[11899, 360, 140, 3, 70, "MoreAbout",
 CellID->417333301]
}, Open  ]],
Cell[12054, 366, 23, 0, 70, "FooterCell"]
}
]
*)

(* End of internal cache information *)

