(* Content-type: application/vnd.wolfram.mathematica *)

(*** Wolfram Notebook File ***)
(* http://www.wolfram.com/nb *)

(* CreatedBy='Mathematica 9.0' *)

(*CacheID: 234*)
(* Internal cache information:
NotebookFileLineBreakTest
NotebookFileLineBreakTest
NotebookDataPosition[       157,          7]
NotebookDataLength[     13868,        429]
NotebookOptionsPosition[     11276,        347]
NotebookOutlinePosition[     12932,        394]
CellTagsIndexPosition[     12847,        389]
WindowTitle->RandomCoalescentTree - Wolfram Mathematica
WindowFrame->Normal*)

(* Beginning of Notebook Content *)
Notebook[{
Cell[" ", "SymbolColorBar",
 CellMargins->{{Inherited, Inherited}, {-5, 0}}],

Cell[TextData[{
 ButtonBox["Genealogy",
  BaseStyle->{"Link", "LinkTrail"},
  ButtonData->"paclet:Genealogy/guide/Genealogy"],
 StyleBox[" > ", "LinkTrailSeparator"]
}], "LinkTrail"],

Cell[BoxData[GridBox[{
   {Cell["GENEALOGY PACLET SYMBOL", "PacletNameCell"], Cell[TextData[{
     Cell[BoxData[
      ActionMenuBox[
       FrameBox["\<\"More About \[RightGuillemet]\"\>",
        StripOnInput->False], {"\<\"Genealogy\"\>":>Documentation`HelpLookup[
       "paclet:Genealogy/guide/Genealogy"]},
       Appearance->None,
       MenuAppearance->Automatic]],
      LineSpacing->{1.4, 0}],
     "\[ThickSpace]\[ThickSpace]\[ThickSpace]\[ThickSpace]\[ThickSpace]\
\[ThickSpace]",
     Cell[BoxData[
      ActionMenuBox[
       FrameBox["\<\"URL \[RightGuillemet]\"\>",
        StripOnInput->False], {"\<\"Genealogy/ref/RandomCoalescentTree\"\>":>
       None, "\<\"Copy Mathematica url\"\>":>
       Module[{DocumentationSearch`Private`nb$}, 
        DocumentationSearch`Private`nb$ = NotebookPut[
           Notebook[{
             Cell["Genealogy/ref/RandomCoalescentTree"]}, Visible -> False]]; 
        SelectionMove[DocumentationSearch`Private`nb$, All, Notebook]; 
        FrontEndTokenExecute[DocumentationSearch`Private`nb$, "Copy"]; 
        NotebookClose[DocumentationSearch`Private`nb$]; Null], 
       Delimiter, "\<\"Copy web url\"\>":>
       Module[{DocumentationSearch`Private`nb$}, 
        DocumentationSearch`Private`nb$ = NotebookPut[
           Notebook[{
             Cell[
              BoxData[
               MakeBoxes[
                Hyperlink[
                "http://reference.wolfram.com/mathematica/Genealogy/ref/\
RandomCoalescentTree.html"], StandardForm]], "Input", TextClipboardType -> 
              "PlainText"]}, Visible -> False]]; 
        SelectionMove[DocumentationSearch`Private`nb$, All, Notebook]; 
        FrontEndTokenExecute[DocumentationSearch`Private`nb$, "Copy"]; 
        NotebookClose[DocumentationSearch`Private`nb$]; 
        Null], "\<\"Go to web url\"\>":>FrontEndExecute[{
         NotebookLocate[{
           URL[
            StringJoin[
            "http://reference.wolfram.com/mathematica/", 
             "Genealogy/ref/RandomCoalescentTree", ".html"]], None}]}]},
       Appearance->None,
       MenuAppearance->Automatic]],
      LineSpacing->{1.4, 0}]
    }], "AnchorBar"]}
  }]], "AnchorBarGrid",
 GridBoxOptions->{GridBoxItemSize->{"Columns" -> {
     Scaled[0.65], {
      Scaled[0.34]}}, "ColumnsIndexed" -> {}, "Rows" -> {{1.}}, 
   "RowsIndexed" -> {}}},
 CellID->1],

Cell["RandomCoalescentTree", "ObjectName",
 CellID->1224892054],

Cell[BoxData[GridBox[{
   {"", Cell[TextData[{
     Cell[BoxData[
      RowBox[{"RandomCoalescentTree", "[", "n", "]"}]], "InlineFormula"],
     " \[LineSeparator]generates a coalescent tree with n leaves, based on \
the standard coalesent without recombination model."
    }]]}
  }]], "Usage",
 GridBoxOptions->{
 GridBoxBackground->{
  "Columns" -> {{None}}, "ColumnsIndexed" -> {}, "Rows" -> {{None}}, 
   "RowsIndexed" -> {}}},
 CellID->982511436],

Cell[CellGroupData[{

Cell[TextData[ButtonBox["EXAMPLES",
 BaseStyle->None,
 Appearance->{Automatic, None},
 Evaluator->None,
 Method->"Preemptive",
 ButtonFunction:>(FrontEndExecute[{
    FrontEnd`SelectionMove[
     FrontEnd`SelectedNotebook[], All, ButtonCell], 
    FrontEndToken["OpenCloseGroup"], 
    FrontEnd`SelectionMove[
     FrontEnd`SelectedNotebook[], After, 
     CellContents]}]& )]], "PrimaryExamplesSection",
 CellTags->"PrimaryExamplesSection",
 CellID->3895953],

Cell[CellGroupData[{

Cell[TextData[{
 "Basic Examples",
 "\[NonBreakingSpace]\[NonBreakingSpace]",
 Cell["(1)", "ExampleCount"]
}], "ExampleSection",
 CellID->776197355],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"<<", "Genealogy`"}]], "Input",
 CellLabel->"In[1]:=",
 CellID->3374645],

Cell[BoxData[
 RowBox[{
  StyleBox[
   RowBox[{"General", "::", "compat"}], "MessageName"], ":", 
  " ", "\<\"Combinatorica Graph and Permutations functionality has been \
superseded by preloaded functionaliy. The package now being loaded may \
conflict with this. Please see the Compatibility Guide for details.\"\>"}]], \
"Message", "MSG",
 CellLabel->"During evaluation of In[1]:=",
 CellID->34954588]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[{
 RowBox[{"RandomCoalescentTree", "[", "5", "]"}], "\[IndentingNewLine]", 
 RowBox[{"ShowLabeledGraph", "[", 
  RowBox[{
   RowBox[{"TreeToCombinatoricaGraph", "[", "%", "]"}], ",", 
   RowBox[{"VertexColor", "\[Rule]", "Red"}], ",", 
   RowBox[{"EdgeColor", "\[Rule]", "Blue"}]}], "]"}]}], "Input",
 CellLabel->"In[2]:=",
 CellID->232874246],

Cell[BoxData[
 RowBox[{"CoalescentTree", "[", 
  RowBox[{
   RowBox[{"{", 
    RowBox[{
    "0", ",", "0", ",", "0", ",", "0", ",", "0", ",", "0.056365599052572046`",
      ",", "0.18056811189719352`", ",", "0.4978172696109599`", ",", 
     "0.6214107307358409`"}], "}"}], ",", 
   RowBox[{"{", 
    RowBox[{
     RowBox[{"{", 
      RowBox[{
       RowBox[{"{", 
        RowBox[{"1", ",", "6"}], "}"}], ",", 
       RowBox[{"{", 
        RowBox[{"3", ",", "6"}], "}"}]}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{
       RowBox[{"{", 
        RowBox[{"5", ",", "7"}], "}"}], ",", 
       RowBox[{"{", 
        RowBox[{"6", ",", "7"}], "}"}]}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{
       RowBox[{"{", 
        RowBox[{"2", ",", "8"}], "}"}], ",", 
       RowBox[{"{", 
        RowBox[{"7", ",", "8"}], "}"}]}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{
       RowBox[{"{", 
        RowBox[{"8", ",", "9"}], "}"}], ",", 
       RowBox[{"{", 
        RowBox[{"4", ",", "9"}], "}"}]}], "}"}]}], "}"}]}], "]"}]], "Output",
 ImageSize->{484, 33},
 ImageMargins->{{0, 0}, {0, 0}},
 ImageRegion->{{0, 1}, {0, 1}},
 CellLabel->"Out[2]=",
 CellID->176900080],

Cell[BoxData[
 GraphicsBox[{{
    {RGBColor[0, 0, 1], Thickness[0.005], 
     ArrowBox[{{0.6000000000000001, 0.}, {0.7000000000000001, 
      0.01127311981051441}}], 
     {GrayLevel[0]}}, 
    {RGBColor[0, 0, 1], Thickness[0.005], 
     ArrowBox[{{0.8, 0.}, {0.7000000000000001, 0.01127311981051441}}], 
     {GrayLevel[0]}}, 
    {RGBColor[0, 0, 1], Thickness[0.005], 
     ArrowBox[{{0.4, 0.}, {0.55, 0.03611362237943871}}], 
     {GrayLevel[0]}}, 
    {RGBColor[0, 0, 1], Thickness[0.005], 
     ArrowBox[{{0.7000000000000001, 0.01127311981051441}, {0.55, 
      0.03611362237943871}}], 
     {GrayLevel[0]}}, 
    {RGBColor[0, 0, 1], Thickness[0.005], 
     ArrowBox[{{0.2, 0.}, {0.375, 0.09956345392219199}}], 
     {GrayLevel[0]}}, 
    {RGBColor[0, 0, 1], Thickness[0.005], 
     ArrowBox[{{0.55, 0.03611362237943871}, {0.375, 0.09956345392219199}}], 
     {GrayLevel[0]}}, 
    {RGBColor[0, 0, 1], Thickness[0.005], 
     ArrowBox[{{0.375, 0.09956345392219199}, {0.6875, 0.12428214614716818`}}], 
     {GrayLevel[0]}}, 
    {RGBColor[0, 0, 1], Thickness[0.005], 
     ArrowBox[{{1., 0.}, {0.6875, 0.12428214614716818`}}], 
     {GrayLevel[0]}}}, {
    {RGBColor[1, 0, 0], 
     {PointSize[0.025], PointBox[{0.6000000000000001, 0.}]}, 
     {GrayLevel[0], 
      InsetBox["1", 
       Scaled[{-0.02, -0.02}, {0.6000000000000001, 0.}], {1, 0}]}, 
     {GrayLevel[0]}}, 
    {RGBColor[1, 0, 0], 
     {PointSize[0.025], PointBox[{0.2, 0.}]}, 
     {GrayLevel[0], InsetBox["2", Scaled[{-0.02, -0.02}, {0.2, 0.}], {1, 0}]}, 
     {GrayLevel[0]}}, 
    {RGBColor[1, 0, 0], 
     {PointSize[0.025], PointBox[{0.8, 0.}]}, 
     {GrayLevel[0], InsetBox["3", Scaled[{-0.02, -0.02}, {0.8, 0.}], {1, 0}]}, 
     {GrayLevel[0]}}, 
    {RGBColor[1, 0, 0], 
     {PointSize[0.025], PointBox[{1., 0.}]}, 
     {GrayLevel[0], InsetBox["4", Scaled[{-0.02, -0.02}, {1., 0.}], {1, 0}]}, 
     {GrayLevel[0]}}, 
    {RGBColor[1, 0, 0], 
     {PointSize[0.025], PointBox[{0.4, 0.}]}, 
     {GrayLevel[0], InsetBox["5", Scaled[{-0.02, -0.02}, {0.4, 0.}], {1, 0}]}, 
     {GrayLevel[0]}}, 
    {RGBColor[1, 0, 0], 
     {PointSize[0.025], PointBox[{0.7000000000000001, 0.01127311981051441}]}, 
     {GrayLevel[0], 
      InsetBox["6", 
       Scaled[{-0.02, -0.02}, {0.7000000000000001, 0.01127311981051441}], \
{1, 0}]}, 
     {GrayLevel[0]}}, 
    {RGBColor[1, 0, 0], 
     {PointSize[0.025], PointBox[{0.55, 0.03611362237943871}]}, 
     {GrayLevel[0], 
      InsetBox["7", 
       Scaled[{-0.02, -0.02}, {0.55, 0.03611362237943871}], {1, 0}]}, 
     {GrayLevel[0]}}, 
    {RGBColor[1, 0, 0], 
     {PointSize[0.025], PointBox[{0.375, 0.09956345392219199}]}, 
     {GrayLevel[0], 
      InsetBox["8", 
       Scaled[{-0.02, -0.02}, {0.375, 0.09956345392219199}], {1, 0}]}, 
     {GrayLevel[0]}}, 
    {RGBColor[1, 0, 0], 
     {PointSize[0.025], PointBox[{0.6875, 0.12428214614716818`}]}, 
     {GrayLevel[0], 
      InsetBox["9", 
       Scaled[{-0.02, -0.02}, {0.6875, 0.12428214614716818`}], {1, 0}]}, 
     {GrayLevel[0]}}}},
  AlignmentPoint->Center,
  AspectRatio->Automatic,
  Axes->False,
  AxesLabel->None,
  AxesOrigin->Automatic,
  AxesStyle->{},
  Background->None,
  BaseStyle->{},
  BaselinePosition->Automatic,
  ColorOutput->Automatic,
  ContentSelectable->Automatic,
  CoordinatesToolOptions:>Automatic,
  DisplayFunction:>$DisplayFunction,
  Epilog->{},
  FormatType:>TraditionalForm,
  Frame->False,
  FrameLabel->None,
  FrameStyle->{},
  FrameTicks->Automatic,
  FrameTicksStyle->{},
  GridLines->None,
  GridLinesStyle->{},
  ImageMargins->0.,
  ImagePadding->All,
  ImageSize->Automatic,
  ImageSizeRaw->Automatic,
  LabelStyle->{},
  Method->Automatic,
  PlotLabel->None,
  PlotRange->{{0.15000000000000002`, 1.05}, {-0.2545255935930826, 
   0.3788077397402507}},
  PlotRangeClipping->False,
  PlotRangePadding->Automatic,
  PlotRegion->Automatic,
  PreserveImageOptions->Automatic,
  Prolog->{},
  RotateLabel->True,
  Ticks->Automatic,
  TicksStyle->{}]], "Output",
 ImageSize->{364, 253},
 ImageMargins->{{0, 0}, {0, 0}},
 ImageRegion->{{0, 1}, {0, 1}},
 CellLabel->"Out[2]=",
 CellID->182396069]
}, Open  ]]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell[TextData[ButtonBox["MORE ABOUT",
 BaseStyle->None,
 Appearance->{Automatic, None},
 Evaluator->None,
 Method->"Preemptive",
 ButtonFunction:>(FrontEndExecute[{
    FrontEnd`SelectionMove[
     FrontEnd`SelectedNotebook[], All, ButtonCell], 
    FrontEndToken["OpenCloseGroup"], 
    FrontEnd`SelectionMove[
     FrontEnd`SelectedNotebook[], After, 
     CellContents]}]& )]], "MoreAboutSection",
 CellID->38303248],

Cell[TextData[ButtonBox["Genealogy",
 BaseStyle->"Link",
 ButtonData->"paclet:Genealogy/guide/Genealogy"]], "MoreAbout",
 CellID->53881601]
}, Open  ]],

Cell[" ", "FooterCell"]
},
Saveable->False,
ScreenStyleEnvironment->"Working",
WindowSize->{725, 730},
WindowMargins->{{0, Automatic}, {Automatic, 0}},
WindowTitle->"RandomCoalescentTree - Wolfram Mathematica",
TaggingRules->{
 "ModificationHighlight" -> False, 
  "Metadata" -> {
   "built" -> "{2013, 7, 27, 22, 7, 8.7287774}", "context" -> "Genealogy`", 
    "keywords" -> {}, "index" -> True, "label" -> "Genealogy Paclet Symbol", 
    "language" -> "en", "paclet" -> "Genealogy", "status" -> "None", 
    "summary" -> 
    "RandomCoalescentTree[n] generates a coalescent tree with n leaves, based \
on the standard coalesent without recombination model.", "synonyms" -> {}, 
    "title" -> "RandomCoalescentTree", "type" -> "Symbol", "uri" -> 
    "Genealogy/ref/RandomCoalescentTree"}, "LinkTrails" -> "", 
  "SearchTextTranslated" -> ""},
CellContext->"Global`",
FrontEndVersion->"9.0 for Microsoft Windows (64-bit) (January 25, 2013)",
StyleDefinitions->Notebook[{
   Cell[
    StyleData[
    StyleDefinitions -> FrontEnd`FileName[{"Wolfram"}, "Reference.nb"]]], 
   Cell[
    StyleData["Input"], CellContext -> "Global`"], 
   Cell[
    StyleData["Output"], CellContext -> "Global`"]}, Visible -> False, 
  FrontEndVersion -> "9.0 for Microsoft Windows (64-bit) (January 25, 2013)", 
  StyleDefinitions -> "Default.nb"]
]
(* End of Notebook Content *)

(* Internal cache information *)
(*CellTagsOutline
CellTagsIndex->{
 "PrimaryExamplesSection"->{
  Cell[3773, 105, 459, 13, 70, "PrimaryExamplesSection",
   CellTags->"PrimaryExamplesSection",
   CellID->3895953]}
 }
*)
(*CellTagsIndex
CellTagsIndex->{
 {"PrimaryExamplesSection", 12706, 382}
 }
*)
(*NotebookFileOutline
Notebook[{
Cell[613, 21, 76, 1, 70, "SymbolColorBar"],
Cell[692, 24, 182, 5, 70, "LinkTrail"],
Cell[877, 31, 2351, 53, 70, "AnchorBarGrid",
 CellID->1],
Cell[3231, 86, 63, 1, 70, "ObjectName",
 CellID->1224892054],
Cell[3297, 89, 451, 12, 70, "Usage",
 CellID->982511436],
Cell[CellGroupData[{
Cell[3773, 105, 459, 13, 70, "PrimaryExamplesSection",
 CellTags->"PrimaryExamplesSection",
 CellID->3895953],
Cell[CellGroupData[{
Cell[4257, 122, 148, 5, 70, "ExampleSection",
 CellID->776197355],
Cell[CellGroupData[{
Cell[4430, 131, 95, 3, 70, "Input",
 CellID->3374645],
Cell[4528, 136, 404, 9, 70, "Message",
 CellID->34954588]
}, Open  ]],
Cell[CellGroupData[{
Cell[4969, 150, 356, 8, 70, "Input",
 CellID->232874246],
Cell[5328, 160, 1170, 38, 54, "Output",
 CellID->176900080],
Cell[6501, 200, 4111, 118, 274, "Output",
 CellID->182396069]
}, Open  ]]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[10673, 325, 419, 12, 70, "MoreAboutSection",
 CellID->38303248],
Cell[11095, 339, 139, 3, 70, "MoreAbout",
 CellID->53881601]
}, Open  ]],
Cell[11249, 345, 23, 0, 70, "FooterCell"]
}
]
*)

(* End of internal cache information *)

