(* Content-type: application/vnd.wolfram.mathematica *)

(*** Wolfram Notebook File ***)
(* http://www.wolfram.com/nb *)

(* CreatedBy='Mathematica 9.0' *)

(*CacheID: 234*)
(* Internal cache information:
NotebookFileLineBreakTest
NotebookFileLineBreakTest
NotebookDataPosition[       157,          7]
NotebookDataLength[     17304,        525]
NotebookOptionsPosition[     14445,        440]
NotebookOutlinePosition[     16383,        491]
CellTagsIndexPosition[     16298,        486]
WindowTitle->RandomGenealogy - Wolfram Mathematica
WindowFrame->Normal*)

(* Beginning of Notebook Content *)
Notebook[{
Cell[" ", "SymbolColorBar",
 CellMargins->{{Inherited, Inherited}, {-5, 0}}],

Cell[TextData[{
 ButtonBox["Genealogy",
  BaseStyle->{"Link", "LinkTrail"},
  ButtonData->"paclet:Genealogy/guide/Genealogy"],
 StyleBox[" > ", "LinkTrailSeparator"]
}], "LinkTrail"],

Cell[BoxData[GridBox[{
   {Cell["GENEALOGY PACKAGE PACLET SYMBOL", "PacletNameCell"], Cell[TextData[{
     Cell[BoxData[
      ActionMenuBox[
       FrameBox["\<\"More About \[RightGuillemet]\"\>",
        StripOnInput->False], {"\<\"Genealogy Package\"\>":>
       Documentation`HelpLookup["paclet:Genealogy/guide/Genealogy"]},
       Appearance->None,
       MenuAppearance->Automatic]],
      LineSpacing->{1.4, 0}],
     "\[ThickSpace]\[ThickSpace]\[ThickSpace]\[ThickSpace]\[ThickSpace]\
\[ThickSpace]",
     Cell[BoxData[
      ActionMenuBox[
       FrameBox["\<\"URL \[RightGuillemet]\"\>",
        StripOnInput->False], {"\<\"Genealogy/ref/RandomGenealogy\"\>":>
       None, "\<\"Copy Mathematica url\"\>":>
       Module[{DocumentationSearch`Private`nb$}, 
        DocumentationSearch`Private`nb$ = NotebookPut[
           Notebook[{
             Cell["Genealogy/ref/RandomGenealogy"]}, Visible -> False]]; 
        SelectionMove[DocumentationSearch`Private`nb$, All, Notebook]; 
        FrontEndTokenExecute[DocumentationSearch`Private`nb$, "Copy"]; 
        NotebookClose[DocumentationSearch`Private`nb$]; Null], 
       Delimiter, "\<\"Copy web url\"\>":>
       Module[{DocumentationSearch`Private`nb$}, 
        DocumentationSearch`Private`nb$ = NotebookPut[
           Notebook[{
             Cell[
              BoxData[
               MakeBoxes[
                Hyperlink[
                "http://reference.wolfram.com/mathematica/Genealogy/ref/\
RandomGenealogy.html"], StandardForm]], "Input", TextClipboardType -> 
              "PlainText"]}, Visible -> False]]; 
        SelectionMove[DocumentationSearch`Private`nb$, All, Notebook]; 
        FrontEndTokenExecute[DocumentationSearch`Private`nb$, "Copy"]; 
        NotebookClose[DocumentationSearch`Private`nb$]; 
        Null], "\<\"Go to web url\"\>":>FrontEndExecute[{
         NotebookLocate[{
           URL[
            StringJoin[
            "http://reference.wolfram.com/mathematica/", 
             "Genealogy/ref/RandomGenealogy", ".html"]], None}]}]},
       Appearance->None,
       MenuAppearance->Automatic]],
      LineSpacing->{1.4, 0}]
    }], "AnchorBar"]}
  }]], "AnchorBarGrid",
 GridBoxOptions->{GridBoxItemSize->{"Columns" -> {
     Scaled[0.65], {
      Scaled[0.34]}}, "ColumnsIndexed" -> {}, "Rows" -> {{1.}}, 
   "RowsIndexed" -> {}}},
 CellID->1],

Cell["RandomGenealogy", "ObjectName",
 CellID->1224892054],

Cell[BoxData[GridBox[{
   {"", Cell[TextData[{
     Cell[BoxData[
      RowBox[{"RandomGenealogy", "[", 
       StyleBox["n", "TI"], "]"}]], "InlineFormula"],
     "\[LineSeparator]generates a coalescent tree for ",
     StyleBox["n", "TI"],
     " sequences. "
    }]]},
   {"", Cell[TextData[{
     Cell[BoxData[
      RowBox[{"RandomGenealogy", "[", 
       RowBox[{
        StyleBox["n", "TI"], ",", "\[Rho]", ",", 
        StyleBox[
         RowBox[{"metho", 
          StyleBox["d", "TI"]}]]}], 
       StyleBox["]", "TI"]}]], "InlineFormula"],
     "\[LineSeparator]generates an ancestral recombination graph for ",
     StyleBox["n", "TI"],
     " sequences and the scaled recombination rate ",
     StyleBox["\[Rho]", "TI"],
     ", using the algorithm specified by the optional ",
     StyleBox["method", "TI"],
     ". The optional method takes \"HUDSON\" (default), \"SRG\", or \"SMC\". \
If method=\"SMC\", return  a series of local tress at recombination points."
    }]]}
  }]], "Usage",
 GridBoxOptions->{
 GridBoxBackground->{
  "Columns" -> {{None}}, "ColumnsIndexed" -> {}, "Rows" -> {None, {None}}, 
   "RowsIndexed" -> {}}}],

Cell[CellGroupData[{

Cell[TextData[ButtonBox["EXAMPLES",
 BaseStyle->None,
 Appearance->{Automatic, None},
 Evaluator->None,
 Method->"Preemptive",
 ButtonFunction:>(FrontEndExecute[{
    FrontEnd`SelectionMove[
     FrontEnd`SelectedNotebook[], All, ButtonCell], 
    FrontEndToken["OpenCloseGroup"], 
    FrontEnd`SelectionMove[
     FrontEnd`SelectedNotebook[], After, 
     CellContents]}]& )]], "PrimaryExamplesSection",
 CellTags->"PrimaryExamplesSection",
 CellID->64003287],

Cell[CellGroupData[{

Cell[TextData[{
 "Basic Examples",
 "\[NonBreakingSpace]\[NonBreakingSpace]",
 Cell["(1)", "ExampleCount"]
}], "ExampleSection",
 CellID->84169028],

Cell[BoxData[
 RowBox[{"Needs", "[", "\"\<Genealogy`\>\"", "]"}]], "Input",
 CellLabel->"In[1]:=",
 CellID->24234475],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"ShowLabeledGraph", "[", 
  RowBox[{
   RowBox[{"RandomGenealogy", "[", "5", "]"}], ",", 
   RowBox[{"VertexColor", "\[Rule]", "Red"}], ",", 
   RowBox[{"EdgeColor", "\[Rule]", "Blue"}]}], "]"}]], "Input",
 CellLabel->"In[2]:=",
 CellID->605413967],

Cell[BoxData[
 GraphicsBox[{{
    {RGBColor[0, 0, 1], Thickness[0.005], 
     ArrowBox[{{0.8, 0.}, {0.7000000000000001, 0.039778291552724505`}}], 
     {GrayLevel[0]}}, 
    {RGBColor[0, 0, 1], Thickness[0.005], 
     ArrowBox[{{0.6000000000000001, 0.}, {0.7000000000000001, 
      0.039778291552724505`}}], 
     {GrayLevel[0]}}, 
    {RGBColor[0, 0, 1], Thickness[0.005], 
     ArrowBox[{{0.4, 0.}, {0.30000000000000004`, 0.08103337056921733}}], 
     {GrayLevel[0]}}, 
    {RGBColor[0, 0, 1], Thickness[0.005], 
     ArrowBox[{{0.2, 0.}, {0.30000000000000004`, 0.08103337056921733}}], 
     {GrayLevel[0]}}, 
    {RGBColor[0, 0, 1], Thickness[0.005], 
     ArrowBox[{{0.30000000000000004`, 0.08103337056921733}, {0.65, 
      0.09137437591563585}}], 
     {GrayLevel[0]}}, 
    {RGBColor[0, 0, 1], Thickness[0.005], 
     ArrowBox[{{1., 0.}, {0.65, 0.09137437591563585}}], 
     {GrayLevel[0]}}, 
    {RGBColor[0, 0, 1], Thickness[0.005], 
     ArrowBox[{{0.7000000000000001, 0.039778291552724505`}, {0.675, 
      0.2996621290657539}}], 
     {GrayLevel[0]}}, 
    {RGBColor[0, 0, 1], Thickness[0.005], 
     ArrowBox[{{0.65, 0.09137437591563585}, {0.675, 0.2996621290657539}}], 
     {GrayLevel[0]}}}, {
    {RGBColor[1, 0, 0], 
     {PointSize[0.025], PointBox[{0.2, 0.}]}, 
     {GrayLevel[0], InsetBox["1", Scaled[{-0.02, -0.02}, {0.2, 0.}], {1, 0}]}, 
     {GrayLevel[0]}}, 
    {RGBColor[1, 0, 0], 
     {PointSize[0.025], PointBox[{0.4, 0.}]}, 
     {GrayLevel[0], InsetBox["2", Scaled[{-0.02, -0.02}, {0.4, 0.}], {1, 0}]}, 
     {GrayLevel[0]}}, 
    {RGBColor[1, 0, 0], 
     {PointSize[0.025], PointBox[{0.6000000000000001, 0.}]}, 
     {GrayLevel[0], 
      InsetBox["3", 
       Scaled[{-0.02, -0.02}, {0.6000000000000001, 0.}], {1, 0}]}, 
     {GrayLevel[0]}}, 
    {RGBColor[1, 0, 0], 
     {PointSize[0.025], PointBox[{0.8, 0.}]}, 
     {GrayLevel[0], InsetBox["4", Scaled[{-0.02, -0.02}, {0.8, 0.}], {1, 0}]}, 
     {GrayLevel[0]}}, 
    {RGBColor[1, 0, 0], 
     {PointSize[0.025], PointBox[{1., 0.}]}, 
     {GrayLevel[0], InsetBox["5", Scaled[{-0.02, -0.02}, {1., 0.}], {1, 0}]}, 
     {GrayLevel[0]}}, 
    {RGBColor[1, 0, 0], 
     {PointSize[0.025], 
      PointBox[{0.7000000000000001, 0.039778291552724505`}]}, 
     {GrayLevel[0], 
      InsetBox["6", 
       Scaled[{-0.02, -0.02}, {0.7000000000000001, 0.039778291552724505`}], \
{1, 0}]}, 
     {GrayLevel[0]}}, 
    {RGBColor[1, 0, 0], 
     {PointSize[0.025], 
      PointBox[{0.30000000000000004`, 0.08103337056921733}]}, 
     {GrayLevel[0], 
      InsetBox["7", 
       Scaled[{-0.02, -0.02}, {0.30000000000000004`, 0.08103337056921733}], \
{1, 0}]}, 
     {GrayLevel[0]}}, 
    {RGBColor[1, 0, 0], 
     {PointSize[0.025], PointBox[{0.65, 0.09137437591563585}]}, 
     {GrayLevel[0], 
      InsetBox["8", 
       Scaled[{-0.02, -0.02}, {0.65, 0.09137437591563585}], {1, 0}]}, 
     {GrayLevel[0]}}, 
    {RGBColor[1, 0, 0], 
     {PointSize[0.025], PointBox[{0.675, 0.2996621290657539}]}, 
     {GrayLevel[0], 
      InsetBox["9", 
       Scaled[{-0.02, -0.02}, {0.675, 0.2996621290657539}], {1, 0}]}, 
     {GrayLevel[0]}}}},
  AlignmentPoint->Center,
  AspectRatio->Automatic,
  Axes->False,
  AxesLabel->None,
  AxesOrigin->Automatic,
  AxesStyle->{},
  Background->None,
  BaseStyle->{},
  BaselinePosition->Automatic,
  ColorOutput->Automatic,
  ContentSelectable->Automatic,
  CoordinatesToolOptions:>Automatic,
  DisplayFunction:>$DisplayFunction,
  Epilog->{},
  FormatType:>TraditionalForm,
  Frame->False,
  FrameLabel->None,
  FrameStyle->{},
  FrameTicks->Automatic,
  FrameTicksStyle->{},
  GridLines->None,
  GridLinesStyle->{},
  ImageMargins->0.,
  ImagePadding->All,
  ImageSize->Automatic,
  ImageSizeRaw->Automatic,
  LabelStyle->{},
  Method->Automatic,
  PlotLabel->None,
  PlotRange->{{0.15000000000000002`, 1.05}, {-0.05, 0.3496621290657539}},
  PlotRangeClipping->False,
  PlotRangePadding->Automatic,
  PlotRegion->Automatic,
  PreserveImageOptions->Automatic,
  Prolog->{},
  RotateLabel->True,
  Ticks->Automatic,
  TicksStyle->{}]], "Output",
 ImageSize->{364, 160},
 ImageMargins->{{0, 0}, {0, 0}},
 ImageRegion->{{0, 1}, {0, 1}},
 CellLabel->"Out[2]=",
 CellID->33690642]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"ShowLabeledGraph", "[", 
  RowBox[{
   RowBox[{"RandomGenealogy", "[", 
    RowBox[{"3", ",", "2"}], "]"}], ",", 
   RowBox[{"VertexColor", "\[Rule]", "Red"}], ",", 
   RowBox[{"EdgeColor", "\[Rule]", "Blue"}]}], "]"}]], "Input",
 CellLabel->"In[3]:=",
 CellID->97854467],

Cell[BoxData[
 GraphicsBox[{{
    {RGBColor[0, 0, 1], Thickness[0.005], 
     ArrowBox[{{0.3333333333333333, 0.}, {0.3333333333333333, 
      0.1065346675940361}}], 
     {GrayLevel[0]}}, 
    {RGBColor[0, 0, 1], Thickness[0.005], 
     ArrowBox[{{0.6666666666666666, 0.}, {0.5, 0.13582778188310018`}}], 
     {GrayLevel[0]}}, 
    {RGBColor[0, 0, 1], Thickness[0.005], 
     ArrowBox[{{0.3333333333333333, 0.1065346675940361}, {0.5, 
      0.13582778188310018`}}], 
     {GrayLevel[0]}}, 
    {RGBColor[0, 0, 1], Thickness[0.005], 
     ArrowBox[{{0.5, 0.13582778188310018`}, {0.75, 0.3229496152829952}}], 
     {GrayLevel[0]}}, 
    {RGBColor[0, 0, 1], Thickness[0.005], 
     ArrowBox[{{1., 0.}, {0.75, 0.3229496152829952}}], 
     {GrayLevel[0]}}, 
    {RGBColor[0, 0, 1], Thickness[0.005], 
     ArrowBox[{{0.3333333333333333, 0.1065346675940361}, {0.5416666666666666, 
      0.48429984679773225`}}], 
     {GrayLevel[0]}}, 
    {RGBColor[0, 0, 1], Thickness[0.005], 
     ArrowBox[{{0.75, 0.3229496152829952}, {0.5416666666666666, 
      0.48429984679773225`}}], 
     {GrayLevel[0]}}}, {
    {RGBColor[1, 0, 0], 
     {PointSize[0.025], PointBox[{0.3333333333333333, 0.}]}, 
     {GrayLevel[0], 
      InsetBox["1", 
       Scaled[{-0.02, -0.02}, {0.3333333333333333, 0.}], {1, 0}]}, 
     {GrayLevel[0]}}, 
    {RGBColor[1, 0, 0], 
     {PointSize[0.025], PointBox[{0.6666666666666666, 0.}]}, 
     {GrayLevel[0], 
      InsetBox["2", 
       Scaled[{-0.02, -0.02}, {0.6666666666666666, 0.}], {1, 0}]}, 
     {GrayLevel[0]}}, 
    {RGBColor[1, 0, 0], 
     {PointSize[0.025], PointBox[{1., 0.}]}, 
     {GrayLevel[0], InsetBox["3", Scaled[{-0.02, -0.02}, {1., 0.}], {1, 0}]}, 
     {GrayLevel[0]}}, 
    {RGBColor[1, 0, 0], 
     {PointSize[0.025], PointBox[{0.3333333333333333, 0.1065346675940361}]}, 
     {GrayLevel[0], 
      InsetBox["4", 
       Scaled[{-0.02, -0.02}, {0.3333333333333333, 0.1065346675940361}], \
{1, 0}]}, 
     {GrayLevel[0]}}, 
    {RGBColor[1, 0, 0], 
     {PointSize[0.025], PointBox[{0.5, 0.13582778188310018`}]}, 
     {GrayLevel[0], 
      InsetBox["5", 
       Scaled[{-0.02, -0.02}, {0.5, 0.13582778188310018`}], {1, 0}]}, 
     {GrayLevel[0]}}, 
    {RGBColor[1, 0, 0], 
     {PointSize[0.025], PointBox[{0.75, 0.3229496152829952}]}, 
     {GrayLevel[0], 
      InsetBox["6", 
       Scaled[{-0.02, -0.02}, {0.75, 0.3229496152829952}], {1, 0}]}, 
     {GrayLevel[0]}}, 
    {RGBColor[1, 0, 0], 
     {PointSize[0.025], PointBox[{0.5416666666666666, 0.48429984679773225`}]}, 
     {GrayLevel[0], 
      InsetBox["7", 
       Scaled[{-0.02, -0.02}, {0.5416666666666666, 0.48429984679773225`}], \
{1, 0}]}, 
     {GrayLevel[0]}}}},
  AlignmentPoint->Center,
  AspectRatio->Automatic,
  Axes->False,
  AxesLabel->None,
  AxesOrigin->Automatic,
  AxesStyle->{},
  Background->None,
  BaseStyle->{},
  BaselinePosition->Automatic,
  ColorOutput->Automatic,
  ContentSelectable->Automatic,
  CoordinatesToolOptions:>Automatic,
  DisplayFunction:>$DisplayFunction,
  Epilog->{},
  FormatType:>TraditionalForm,
  Frame->False,
  FrameLabel->None,
  FrameStyle->{},
  FrameTicks->Automatic,
  FrameTicksStyle->{},
  GridLines->None,
  GridLinesStyle->{},
  ImageMargins->0.,
  ImagePadding->All,
  ImageSize->Automatic,
  ImageSizeRaw->Automatic,
  LabelStyle->{},
  Method->Automatic,
  PlotLabel->None,
  PlotRange->{{0.2833333333333333, 1.05}, {-0.05, 0.5342998467977322}},
  PlotRangeClipping->False,
  PlotRangePadding->Automatic,
  PlotRegion->Automatic,
  PreserveImageOptions->Automatic,
  Prolog->{},
  RotateLabel->True,
  Ticks->Automatic,
  TicksStyle->{}]], "Output",
 ImageSize->{364, 274},
 ImageMargins->{{0, 0}, {0, 0}},
 ImageRegion->{{0, 1}, {0, 1}},
 CellLabel->"Out[3]=",
 CellID->615152044]
}, Open  ]]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell[TextData[ButtonBox["MORE ABOUT",
 BaseStyle->None,
 Appearance->{Automatic, None},
 Evaluator->None,
 Method->"Preemptive",
 ButtonFunction:>(FrontEndExecute[{
    FrontEnd`SelectionMove[
     FrontEnd`SelectedNotebook[], All, ButtonCell], 
    FrontEndToken["OpenCloseGroup"], 
    FrontEnd`SelectionMove[
     FrontEnd`SelectedNotebook[], After, 
     CellContents]}]& )]], "MoreAboutSection",
 CellID->38303248],

Cell[TextData[ButtonBox["Genealogy Package",
 BaseStyle->"Link",
 ButtonData->"paclet:Genealogy/guide/Genealogy"]], "MoreAbout",
 CellID->1665078683]
}, Open  ]],

Cell[" ", "FooterCell"]
},
Saveable->False,
ScreenStyleEnvironment->"Working",
WindowSize->{725, 730},
WindowMargins->{{0, Automatic}, {Automatic, 0}},
WindowTitle->"RandomGenealogy - Wolfram Mathematica",
TaggingRules->{
 "ModificationHighlight" -> False, 
  "Metadata" -> {
   "built" -> "{2013, 7, 27, 22, 7, 9.6179790}", "context" -> "Genealogy`", 
    "keywords" -> {}, "index" -> True, "label" -> 
    "Genealogy Package Paclet Symbol", "language" -> "en", "paclet" -> 
    "Genealogy Package", "status" -> "None", "summary" -> 
    "RandomGenealogy[n] generates a coalescent tree for n sequences. \
RandomGenealogy[n, \\[Rho], method] generates an ancestral recombination \
graph for n sequences and the scaled recombination rate \\[Rho], using the \
algorithm specified by the optional method. The optional method takes \
\"HUDSON\" (default), \"SRG\", or \"SMC\". If method=\"SMC\", return a series \
of local tress at recombination points.", "synonyms" -> {}, "title" -> 
    "RandomGenealogy", "type" -> "Symbol", "uri" -> 
    "Genealogy/ref/RandomGenealogy"}, "LinkTrails" -> "", 
  "SearchTextTranslated" -> ""},
CellContext->"Global`",
FrontEndVersion->"9.0 for Microsoft Windows (64-bit) (January 25, 2013)",
StyleDefinitions->Notebook[{
   Cell[
    StyleData[
    StyleDefinitions -> FrontEnd`FileName[{"Wolfram"}, "Reference.nb"]]], 
   Cell[
    StyleData["Input"], CellContext -> "Global`"], 
   Cell[
    StyleData["Output"], CellContext -> "Global`"]}, Visible -> False, 
  FrontEndVersion -> "9.0 for Microsoft Windows (64-bit) (January 25, 2013)", 
  StyleDefinitions -> "Default.nb"]
]
(* End of Notebook Content *)

(* Internal cache information *)
(*CellTagsOutline
CellTagsIndex->{
 "PrimaryExamplesSection"->{
  Cell[4453, 124, 460, 13, 70, "PrimaryExamplesSection",
   CellTags->"PrimaryExamplesSection",
   CellID->64003287]}
 }
*)
(*CellTagsIndex
CellTagsIndex->{
 {"PrimaryExamplesSection", 16156, 479}
 }
*)
(*NotebookFileOutline
Notebook[{
Cell[608, 21, 76, 1, 70, "SymbolColorBar"],
Cell[687, 24, 182, 5, 70, "LinkTrail"],
Cell[872, 31, 2347, 53, 70, "AnchorBarGrid",
 CellID->1],
Cell[3222, 86, 58, 1, 70, "ObjectName",
 CellID->1224892054],
Cell[3283, 89, 1145, 31, 70, "Usage"],
Cell[CellGroupData[{
Cell[4453, 124, 460, 13, 70, "PrimaryExamplesSection",
 CellTags->"PrimaryExamplesSection",
 CellID->64003287],
Cell[CellGroupData[{
Cell[4938, 141, 147, 5, 70, "ExampleSection",
 CellID->84169028],
Cell[5088, 148, 117, 3, 70, "Input",
 CellID->24234475],
Cell[CellGroupData[{
Cell[5230, 155, 271, 7, 70, "Input",
 CellID->605413967],
Cell[5504, 164, 4188, 121, 181, "Output",
 CellID->33690642]
}, Open  ]],
Cell[CellGroupData[{
Cell[9729, 290, 295, 8, 70, "Input",
 CellID->97854467],
Cell[10027, 300, 3744, 111, 295, "Output",
 CellID->615152044]
}, Open  ]]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[13832, 418, 419, 12, 70, "MoreAboutSection",
 CellID->38303248],
Cell[14254, 432, 149, 3, 70, "MoreAbout",
 CellID->1665078683]
}, Open  ]],
Cell[14418, 438, 23, 0, 70, "FooterCell"]
}
]
*)

(* End of internal cache information *)

