(* Content-type: application/vnd.wolfram.mathematica *)

(*** Wolfram Notebook File ***)
(* http://www.wolfram.com/nb *)

(* CreatedBy='Mathematica 9.0' *)

(*CacheID: 234*)
(* Internal cache information:
NotebookFileLineBreakTest
NotebookFileLineBreakTest
NotebookDataPosition[       157,          7]
NotebookDataLength[     18826,        584]
NotebookOptionsPosition[     15710,        489]
NotebookOutlinePosition[     17652,        540]
CellTagsIndexPosition[     17567,        535]
WindowTitle->RandomSequentialTrees - Wolfram Mathematica
WindowFrame->Normal*)

(* Beginning of Notebook Content *)
Notebook[{
Cell[" ", "SymbolColorBar",
 CellMargins->{{Inherited, Inherited}, {-5, 0}}],

Cell[TextData[{
 ButtonBox["Genealogy",
  BaseStyle->{"Link", "LinkTrail"},
  ButtonData->"paclet:Genealogy/guide/Genealogy"],
 StyleBox[" > ", "LinkTrailSeparator"]
}], "LinkTrail"],

Cell[BoxData[GridBox[{
   {Cell["GENEALOGY PACLET SYMBOL", "PacletNameCell"], Cell[TextData[{
     Cell[BoxData[
      ActionMenuBox[
       FrameBox["\<\"More About \[RightGuillemet]\"\>",
        StripOnInput->False], {"\<\"Genealogy\"\>":>Documentation`HelpLookup[
       "paclet:Genealogy/guide/Genealogy"]},
       Appearance->None,
       MenuAppearance->Automatic]],
      LineSpacing->{1.4, 0}],
     "\[ThickSpace]\[ThickSpace]\[ThickSpace]\[ThickSpace]\[ThickSpace]\
\[ThickSpace]",
     Cell[BoxData[
      ActionMenuBox[
       FrameBox["\<\"URL \[RightGuillemet]\"\>",
        StripOnInput->False], {"\<\"Genealogy/ref/RandomSequentialTrees\"\>":>
       None, "\<\"Copy Mathematica url\"\>":>
       Module[{DocumentationSearch`Private`nb$}, 
        DocumentationSearch`Private`nb$ = NotebookPut[
           Notebook[{
             Cell["Genealogy/ref/RandomSequentialTrees"]}, Visible -> False]]; 
        SelectionMove[DocumentationSearch`Private`nb$, All, Notebook]; 
        FrontEndTokenExecute[DocumentationSearch`Private`nb$, "Copy"]; 
        NotebookClose[DocumentationSearch`Private`nb$]; Null], 
       Delimiter, "\<\"Copy web url\"\>":>
       Module[{DocumentationSearch`Private`nb$}, 
        DocumentationSearch`Private`nb$ = NotebookPut[
           Notebook[{
             Cell[
              BoxData[
               MakeBoxes[
                Hyperlink[
                "http://reference.wolfram.com/mathematica/Genealogy/ref/\
RandomSequentialTrees.html"], StandardForm]], "Input", TextClipboardType -> 
              "PlainText"]}, Visible -> False]]; 
        SelectionMove[DocumentationSearch`Private`nb$, All, Notebook]; 
        FrontEndTokenExecute[DocumentationSearch`Private`nb$, "Copy"]; 
        NotebookClose[DocumentationSearch`Private`nb$]; 
        Null], "\<\"Go to web url\"\>":>FrontEndExecute[{
         NotebookLocate[{
           URL[
            StringJoin[
            "http://reference.wolfram.com/mathematica/", 
             "Genealogy/ref/RandomSequentialTrees", ".html"]], None}]}]},
       Appearance->None,
       MenuAppearance->Automatic]],
      LineSpacing->{1.4, 0}]
    }], "AnchorBar"]}
  }]], "AnchorBarGrid",
 GridBoxOptions->{GridBoxItemSize->{"Columns" -> {
     Scaled[0.65], {
      Scaled[0.34]}}, "ColumnsIndexed" -> {}, "Rows" -> {{1.}}, 
   "RowsIndexed" -> {}}},
 CellID->1],

Cell["RandomSequentialTrees", "ObjectName",
 CellID->1224892054],

Cell[CellGroupData[{

Cell[BoxData[GridBox[{
   {"", Cell[TextData[{
     Cell[BoxData[
      RowBox[{"RandomSequentialTrees", "[", 
       RowBox[{"tree", ",", "\[Rho]", ",", "x0", ",", "x1"}], "]"}]], 
      "InlineFormula"],
     " \[LineSeparator]generates local trees along chromsomes based on the \
modified Sequential Markov Coalescent process with recombation rate \[Rho], \
starting from",
     StyleBox[" tree",
      FontSlant->"Italic"],
     " at x=",
     StyleBox["x0",
      FontSlant->"Italic"],
     " and continuing until x=",
     StyleBox["x1",
      FontSlant->"Italic"],
     ". Returns a list with each element consisting of change-point location, \
transition operator from the tree before recombation event, and local tree on \
and after change-point. "
    }]]}
  }]], "Usage",
 GridBoxOptions->{
 GridBoxBackground->{
  "Columns" -> {{None}}, "ColumnsIndexed" -> {}, "Rows" -> {{None}}, 
   "RowsIndexed" -> {}}},
 CellID->982511436],

Cell[CellGroupData[{

Cell[TextData[Cell[BoxData[
 ButtonBox[
  FrameBox[
   StyleBox[
    RowBox[{"MORE", " ", "INFORMATION"}], "NotesFrameText"],
   StripOnInput->False],
  Appearance->{Automatic, None},
  BaseStyle->None,
  ButtonFunction:>(FrontEndExecute[{
     FrontEnd`SelectionMove[
      FrontEnd`SelectedNotebook[], All, ButtonCell], 
     FrontEndToken["OpenCloseGroup"], 
     FrontEnd`SelectionMove[
      FrontEnd`SelectedNotebook[], After, CellContents]}]& ),
  Evaluator->None,
  Method->"Preemptive"]]]], "NotesSection",
 CellGroupingRules->{"SectionGrouping", 50},
 CellID->703655122],

Cell["\<\
There are two settings for x0 and x1. First, the sequences are represented \
continuously from 0 to 1 and thus 0\[LessEqual]x0<x1\[LessEqual]1. Second, \
the sequences are represented discretely in unit of bp, and thus both x0 and \
x1 are positive integers and x0<x1.\
\>", "Notes",
 CellID->1067943069]
}, Closed]]
}, Open  ]],

Cell[CellGroupData[{

Cell[TextData[ButtonBox["EXAMPLES",
 BaseStyle->None,
 Appearance->{Automatic, None},
 Evaluator->None,
 Method->"Preemptive",
 ButtonFunction:>(FrontEndExecute[{
    FrontEnd`SelectionMove[
     FrontEnd`SelectedNotebook[], All, ButtonCell], 
    FrontEndToken["OpenCloseGroup"], 
    FrontEnd`SelectionMove[
     FrontEnd`SelectedNotebook[], After, 
     CellContents]}]& )]], "PrimaryExamplesSection",
 CellTags->"PrimaryExamplesSection",
 CellID->53314144],

Cell[CellGroupData[{

Cell[TextData[{
 "Basic Examples",
 "\[NonBreakingSpace]\[NonBreakingSpace]",
 Cell["(1)", "ExampleCount"]
}], "ExampleSection",
 CellID->262122630],

Cell[BoxData[
 RowBox[{"<<", "Genealogy`"}]], "Input",
 CellLabel->"In[1]:=",
 CellID->487177798],

Cell[CellGroupData[{

Cell[BoxData[{
 RowBox[{
  RowBox[{"tree", "=", 
   RowBox[{"RandomCoalescentTree", "[", "5", "]"}]}], 
  ";"}], "\[IndentingNewLine]", 
 RowBox[{"RandomSequentialTrees", "[", 
  RowBox[{"tree", ",", "1", ",", "0", ",", "1"}], "]"}]}], "Input",
 CellLabel->"In[2]:=",
 CellID->53260314],

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{
   RowBox[{"{", 
    RowBox[{"0", ",", "\<\"Initial\"\>", ",", 
     RowBox[{"CoalescentTree", "[", 
      RowBox[{
       RowBox[{"{", 
        RowBox[{
        "0", ",", "0", ",", "0", ",", "0", ",", "0", ",", 
         "0.012666025474938742`", ",", "0.05326178470528031`", ",", 
         "0.11584123238242541`", ",", "0.1563755897626996`"}], "}"}], ",", 
       RowBox[{"{", 
        RowBox[{
         RowBox[{"{", 
          RowBox[{
           RowBox[{"{", 
            RowBox[{"2", ",", "6"}], "}"}], ",", 
           RowBox[{"{", 
            RowBox[{"3", ",", "6"}], "}"}]}], "}"}], ",", 
         RowBox[{"{", 
          RowBox[{
           RowBox[{"{", 
            RowBox[{"5", ",", "7"}], "}"}], ",", 
           RowBox[{"{", 
            RowBox[{"6", ",", "7"}], "}"}]}], "}"}], ",", 
         RowBox[{"{", 
          RowBox[{
           RowBox[{"{", 
            RowBox[{"4", ",", "8"}], "}"}], ",", 
           RowBox[{"{", 
            RowBox[{"7", ",", "8"}], "}"}]}], "}"}], ",", 
         RowBox[{"{", 
          RowBox[{
           RowBox[{"{", 
            RowBox[{"8", ",", "9"}], "}"}], ",", 
           RowBox[{"{", 
            RowBox[{"1", ",", "9"}], "}"}]}], "}"}]}], "}"}]}], "]"}]}], 
    "}"}], ",", 
   RowBox[{"{", 
    RowBox[{"1.5008818957755499`", ",", 
     RowBox[{"{", 
      RowBox[{
       RowBox[{"{", 
        RowBox[{"8", ",", "9"}], "}"}], ",", "1.3380120556630537`", ",", 
       RowBox[{"{", 
        RowBox[{"9", ",", 
         RowBox[{"-", "1"}]}], "}"}]}], "}"}], ",", 
     RowBox[{"CoalescentTree", "[", 
      RowBox[{
       RowBox[{"{", 
        RowBox[{
        "0", ",", "0", ",", "0", ",", "0", ",", "0", ",", 
         "0.012666025474938742`", ",", "0.05326178470528031`", ",", 
         "0.11584123238242541`", ",", "1.3380120556630537`"}], "}"}], ",", 
       RowBox[{"{", 
        RowBox[{
         RowBox[{"{", 
          RowBox[{
           RowBox[{"{", 
            RowBox[{"2", ",", "6"}], "}"}], ",", 
           RowBox[{"{", 
            RowBox[{"3", ",", "6"}], "}"}]}], "}"}], ",", 
         RowBox[{"{", 
          RowBox[{
           RowBox[{"{", 
            RowBox[{"5", ",", "7"}], "}"}], ",", 
           RowBox[{"{", 
            RowBox[{"6", ",", "7"}], "}"}]}], "}"}], ",", 
         RowBox[{"{", 
          RowBox[{
           RowBox[{"{", 
            RowBox[{"4", ",", "8"}], "}"}], ",", 
           RowBox[{"{", 
            RowBox[{"7", ",", "8"}], "}"}]}], "}"}], ",", 
         RowBox[{"{", 
          RowBox[{
           RowBox[{"{", 
            RowBox[{"8", ",", "9"}], "}"}], ",", 
           RowBox[{"{", 
            RowBox[{"1", ",", "9"}], "}"}]}], "}"}]}], "}"}]}], "]"}]}], 
    "}"}]}], "}"}]], "Output",
 ImageSize->{576, 84},
 ImageMargins->{{0, 0}, {0, 0}},
 ImageRegion->{{0, 1}, {0, 1}},
 CellLabel->"Out[2]=",
 CellID->162116426]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"RandomSequentialTrees", "[", 
  RowBox[{"tree", ",", "0.001", ",", "1", ",", "1000"}], "]"}]], "Input",
 CellLabel->"In[3]:=",
 CellID->72198143],

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{
   RowBox[{"{", 
    RowBox[{"1", ",", "\<\"Initial\"\>", ",", 
     RowBox[{"CoalescentTree", "[", 
      RowBox[{
       RowBox[{"{", 
        RowBox[{
        "0", ",", "0", ",", "0", ",", "0", ",", "0", ",", 
         "0.012666025474938742`", ",", "0.05326178470528031`", ",", 
         "0.11584123238242541`", ",", "0.1563755897626996`"}], "}"}], ",", 
       RowBox[{"{", 
        RowBox[{
         RowBox[{"{", 
          RowBox[{
           RowBox[{"{", 
            RowBox[{"2", ",", "6"}], "}"}], ",", 
           RowBox[{"{", 
            RowBox[{"3", ",", "6"}], "}"}]}], "}"}], ",", 
         RowBox[{"{", 
          RowBox[{
           RowBox[{"{", 
            RowBox[{"5", ",", "7"}], "}"}], ",", 
           RowBox[{"{", 
            RowBox[{"6", ",", "7"}], "}"}]}], "}"}], ",", 
         RowBox[{"{", 
          RowBox[{
           RowBox[{"{", 
            RowBox[{"4", ",", "8"}], "}"}], ",", 
           RowBox[{"{", 
            RowBox[{"7", ",", "8"}], "}"}]}], "}"}], ",", 
         RowBox[{"{", 
          RowBox[{
           RowBox[{"{", 
            RowBox[{"8", ",", "9"}], "}"}], ",", 
           RowBox[{"{", 
            RowBox[{"1", ",", "9"}], "}"}]}], "}"}]}], "}"}]}], "]"}]}], 
    "}"}], ",", 
   RowBox[{"{", 
    RowBox[{"339", ",", 
     RowBox[{"{", 
      RowBox[{
       RowBox[{"{", 
        RowBox[{"1", ",", "9"}], "}"}], ",", "2.178989692484272`", ",", 
       RowBox[{"{", 
        RowBox[{"9", ",", 
         RowBox[{"-", "1"}]}], "}"}]}], "}"}], ",", 
     RowBox[{"CoalescentTree", "[", 
      RowBox[{
       RowBox[{"{", 
        RowBox[{
        "0", ",", "0", ",", "0", ",", "0", ",", "0", ",", 
         "0.012666025474938742`", ",", "0.05326178470528031`", ",", 
         "0.11584123238242541`", ",", "2.178989692484272`"}], "}"}], ",", 
       RowBox[{"{", 
        RowBox[{
         RowBox[{"{", 
          RowBox[{
           RowBox[{"{", 
            RowBox[{"2", ",", "6"}], "}"}], ",", 
           RowBox[{"{", 
            RowBox[{"3", ",", "6"}], "}"}]}], "}"}], ",", 
         RowBox[{"{", 
          RowBox[{
           RowBox[{"{", 
            RowBox[{"5", ",", "7"}], "}"}], ",", 
           RowBox[{"{", 
            RowBox[{"6", ",", "7"}], "}"}]}], "}"}], ",", 
         RowBox[{"{", 
          RowBox[{
           RowBox[{"{", 
            RowBox[{"4", ",", "8"}], "}"}], ",", 
           RowBox[{"{", 
            RowBox[{"7", ",", "8"}], "}"}]}], "}"}], ",", 
         RowBox[{"{", 
          RowBox[{
           RowBox[{"{", 
            RowBox[{"8", ",", "9"}], "}"}], ",", 
           RowBox[{"{", 
            RowBox[{"1", ",", "9"}], "}"}]}], "}"}]}], "}"}]}], "]"}]}], 
    "}"}], ",", 
   RowBox[{"{", 
    RowBox[{"415", ",", 
     RowBox[{"{", 
      RowBox[{
       RowBox[{"{", 
        RowBox[{"1", ",", "9"}], "}"}], ",", "2.031466524317489`", ",", 
       RowBox[{"{", 
        RowBox[{"8", ",", "9"}], "}"}]}], "}"}], ",", 
     RowBox[{"CoalescentTree", "[", 
      RowBox[{
       RowBox[{"{", 
        RowBox[{
        "0", ",", "0", ",", "0", ",", "0", ",", "0", ",", 
         "0.012666025474938742`", ",", "0.05326178470528031`", ",", 
         "0.11584123238242541`", ",", "2.031466524317489`"}], "}"}], ",", 
       RowBox[{"{", 
        RowBox[{
         RowBox[{"{", 
          RowBox[{
           RowBox[{"{", 
            RowBox[{"2", ",", "6"}], "}"}], ",", 
           RowBox[{"{", 
            RowBox[{"3", ",", "6"}], "}"}]}], "}"}], ",", 
         RowBox[{"{", 
          RowBox[{
           RowBox[{"{", 
            RowBox[{"5", ",", "7"}], "}"}], ",", 
           RowBox[{"{", 
            RowBox[{"6", ",", "7"}], "}"}]}], "}"}], ",", 
         RowBox[{"{", 
          RowBox[{
           RowBox[{"{", 
            RowBox[{"4", ",", "8"}], "}"}], ",", 
           RowBox[{"{", 
            RowBox[{"7", ",", "8"}], "}"}]}], "}"}], ",", 
         RowBox[{"{", 
          RowBox[{
           RowBox[{"{", 
            RowBox[{"8", ",", "9"}], "}"}], ",", 
           RowBox[{"{", 
            RowBox[{"1", ",", "9"}], "}"}]}], "}"}]}], "}"}]}], "]"}]}], 
    "}"}], ",", 
   RowBox[{"{", 
    RowBox[{"1036", ",", 
     RowBox[{"{", 
      RowBox[{
       RowBox[{"{", 
        RowBox[{"1", ",", "9"}], "}"}], ",", "2.010062740663244`", ",", 
       RowBox[{"{", 
        RowBox[{"8", ",", "9"}], "}"}]}], "}"}], ",", 
     RowBox[{"CoalescentTree", "[", 
      RowBox[{
       RowBox[{"{", 
        RowBox[{
        "0", ",", "0", ",", "0", ",", "0", ",", "0", ",", 
         "0.012666025474938742`", ",", "0.05326178470528031`", ",", 
         "0.11584123238242541`", ",", "2.010062740663244`"}], "}"}], ",", 
       RowBox[{"{", 
        RowBox[{
         RowBox[{"{", 
          RowBox[{
           RowBox[{"{", 
            RowBox[{"2", ",", "6"}], "}"}], ",", 
           RowBox[{"{", 
            RowBox[{"3", ",", "6"}], "}"}]}], "}"}], ",", 
         RowBox[{"{", 
          RowBox[{
           RowBox[{"{", 
            RowBox[{"5", ",", "7"}], "}"}], ",", 
           RowBox[{"{", 
            RowBox[{"6", ",", "7"}], "}"}]}], "}"}], ",", 
         RowBox[{"{", 
          RowBox[{
           RowBox[{"{", 
            RowBox[{"4", ",", "8"}], "}"}], ",", 
           RowBox[{"{", 
            RowBox[{"7", ",", "8"}], "}"}]}], "}"}], ",", 
         RowBox[{"{", 
          RowBox[{
           RowBox[{"{", 
            RowBox[{"8", ",", "9"}], "}"}], ",", 
           RowBox[{"{", 
            RowBox[{"1", ",", "9"}], "}"}]}], "}"}]}], "}"}]}], "]"}]}], 
    "}"}]}], "}"}]], "Output",
 ImageSize->{576, 186},
 ImageMargins->{{0, 0}, {0, 0}},
 ImageRegion->{{0, 1}, {0, 1}},
 CellLabel->"Out[3]=",
 CellID->353115850]
}, Open  ]]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell[TextData[ButtonBox["MORE ABOUT",
 BaseStyle->None,
 Appearance->{Automatic, None},
 Evaluator->None,
 Method->"Preemptive",
 ButtonFunction:>(FrontEndExecute[{
    FrontEnd`SelectionMove[
     FrontEnd`SelectedNotebook[], All, ButtonCell], 
    FrontEndToken["OpenCloseGroup"], 
    FrontEnd`SelectionMove[
     FrontEnd`SelectedNotebook[], After, 
     CellContents]}]& )]], "MoreAboutSection",
 CellID->38303248],

Cell[TextData[ButtonBox["Genealogy",
 BaseStyle->"Link",
 ButtonData->"paclet:Genealogy/guide/Genealogy"]], "MoreAbout",
 CellID->204906720]
}, Open  ]],

Cell[" ", "FooterCell"]
},
Saveable->False,
ScreenStyleEnvironment->"Working",
WindowSize->{725, 730},
WindowMargins->{{0, Automatic}, {Automatic, 0}},
WindowTitle->"RandomSequentialTrees - Wolfram Mathematica",
TaggingRules->{
 "ModificationHighlight" -> False, 
  "Metadata" -> {
   "built" -> "{2013, 7, 27, 22, 7, 11.3651821}", "context" -> "Genealogy`", 
    "keywords" -> {}, "index" -> True, "label" -> "Genealogy Paclet Symbol", 
    "language" -> "en", "paclet" -> "Genealogy", "status" -> "None", 
    "summary" -> 
    "RandomSequentialTrees[tree, \\[Rho], x0, x1] generates local trees along \
chromsomes based on the modified Sequential Markov Coalescent process with \
recombation rate \\[Rho], starting from tree at x=x0 and continuing until \
x=x1. Returns a list with each element consisting of change-point location, \
transition operator from the tree before recombation event, and local tree on \
and after change-point. ", "synonyms" -> {}, "title" -> 
    "RandomSequentialTrees", "type" -> "Symbol", "uri" -> 
    "Genealogy/ref/RandomSequentialTrees"}, "LinkTrails" -> "", 
  "SearchTextTranslated" -> ""},
CellContext->"Global`",
FrontEndVersion->"9.0 for Microsoft Windows (64-bit) (January 25, 2013)",
StyleDefinitions->Notebook[{
   Cell[
    StyleData[
    StyleDefinitions -> FrontEnd`FileName[{"Wolfram"}, "Reference.nb"]]], 
   Cell[
    StyleData["Input"], CellContext -> "Global`"], 
   Cell[
    StyleData["Output"], CellContext -> "Global`"]}, Visible -> False, 
  FrontEndVersion -> "9.0 for Microsoft Windows (64-bit) (January 25, 2013)", 
  StyleDefinitions -> "Default.nb"]
]
(* End of Notebook Content *)

(* Internal cache information *)
(*CellTagsOutline
CellTagsIndex->{
 "PrimaryExamplesSection"->{
  Cell[5235, 152, 460, 13, 70, "PrimaryExamplesSection",
   CellTags->"PrimaryExamplesSection",
   CellID->53314144]}
 }
*)
(*CellTagsIndex
CellTagsIndex->{
 {"PrimaryExamplesSection", 17425, 528}
 }
*)
(*NotebookFileOutline
Notebook[{
Cell[614, 21, 76, 1, 70, "SymbolColorBar"],
Cell[693, 24, 182, 5, 70, "LinkTrail"],
Cell[878, 31, 2355, 53, 70, "AnchorBarGrid",
 CellID->1],
Cell[3236, 86, 64, 1, 70, "ObjectName",
 CellID->1224892054],
Cell[CellGroupData[{
Cell[3325, 91, 939, 26, 70, "Usage",
 CellID->982511436],
Cell[CellGroupData[{
Cell[4289, 121, 580, 17, 70, "NotesSection",
 CellGroupingRules->{"SectionGrouping", 50},
 CellID->703655122],
Cell[4872, 140, 314, 6, 70, "Notes",
 CellID->1067943069]
}, Closed]]
}, Open  ]],
Cell[CellGroupData[{
Cell[5235, 152, 460, 13, 70, "PrimaryExamplesSection",
 CellTags->"PrimaryExamplesSection",
 CellID->53314144],
Cell[CellGroupData[{
Cell[5720, 169, 148, 5, 70, "ExampleSection",
 CellID->262122630],
Cell[5871, 176, 97, 3, 70, "Input",
 CellID->487177798],
Cell[CellGroupData[{
Cell[5993, 183, 286, 8, 70, "Input",
 CellID->53260314],
Cell[6282, 193, 2872, 86, 105, "Output",
 CellID->162116426]
}, Open  ]],
Cell[CellGroupData[{
Cell[9191, 284, 169, 4, 70, "Input",
 CellID->72198143],
Cell[9363, 290, 5682, 170, 207, "Output",
 CellID->353115850]
}, Open  ]]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[15106, 467, 419, 12, 70, "MoreAboutSection",
 CellID->38303248],
Cell[15528, 481, 140, 3, 70, "MoreAbout",
 CellID->204906720]
}, Open  ]],
Cell[15683, 487, 23, 0, 70, "FooterCell"]
}
]
*)

(* End of internal cache information *)

