(* Content-type: application/vnd.wolfram.mathematica *)

(*** Wolfram Notebook File ***)
(* http://www.wolfram.com/nb *)

(* CreatedBy='Mathematica 9.0' *)

(*CacheID: 234*)
(* Internal cache information:
NotebookFileLineBreakTest
NotebookFileLineBreakTest
NotebookDataPosition[       157,          7]
NotebookDataLength[     14977,        471]
NotebookOptionsPosition[     12152,        381]
NotebookOutlinePosition[     13816,        428]
CellTagsIndexPosition[     13731,        423]
WindowTitle->TreeToAdjacencyList - Wolfram Mathematica
WindowFrame->Normal*)

(* Beginning of Notebook Content *)
Notebook[{
Cell[" ", "SymbolColorBar",
 CellMargins->{{Inherited, Inherited}, {-5, 0}}],

Cell[TextData[{
 ButtonBox["Genealogy",
  BaseStyle->{"Link", "LinkTrail"},
  ButtonData->"paclet:Genealogy/guide/Genealogy"],
 StyleBox[" > ", "LinkTrailSeparator"]
}], "LinkTrail"],

Cell[BoxData[GridBox[{
   {Cell["GENEALOGY PACLET SYMBOL", "PacletNameCell"], Cell[TextData[{
     Cell[BoxData[
      ActionMenuBox[
       FrameBox["\<\"See Also \[RightGuillemet]\"\>",
        StripOnInput->False], {"\<\"TreeNodeParent\"\>":>
       Documentation`HelpLookup["paclet:Genealogy/ref/TreeNodeParent"]},
       Appearance->None,
       MenuAppearance->Automatic]],
      LineSpacing->{1.4, 0}],
     "\[ThickSpace]\[ThickSpace]\[ThickSpace]\[ThickSpace]\[ThickSpace]\
\[ThickSpace]",
     Cell[BoxData[
      ActionMenuBox[
       FrameBox["\<\"More About \[RightGuillemet]\"\>",
        StripOnInput->False], {"\<\"Genealogy\"\>":>Documentation`HelpLookup[
       "paclet:Genealogy/guide/Genealogy"]},
       Appearance->None,
       MenuAppearance->Automatic]],
      LineSpacing->{1.4, 0}],
     "\[ThickSpace]\[ThickSpace]\[ThickSpace]\[ThickSpace]\[ThickSpace]\
\[ThickSpace]",
     Cell[BoxData[
      ActionMenuBox[
       FrameBox["\<\"URL \[RightGuillemet]\"\>",
        StripOnInput->False], {"\<\"Genealogy/ref/TreeToAdjacencyList\"\>":>
       None, "\<\"Copy Mathematica url\"\>":>
       Module[{DocumentationSearch`Private`nb$}, 
        DocumentationSearch`Private`nb$ = NotebookPut[
           Notebook[{
             Cell["Genealogy/ref/TreeToAdjacencyList"]}, Visible -> False]]; 
        SelectionMove[DocumentationSearch`Private`nb$, All, Notebook]; 
        FrontEndTokenExecute[DocumentationSearch`Private`nb$, "Copy"]; 
        NotebookClose[DocumentationSearch`Private`nb$]; Null], 
       Delimiter, "\<\"Copy web url\"\>":>
       Module[{DocumentationSearch`Private`nb$}, 
        DocumentationSearch`Private`nb$ = NotebookPut[
           Notebook[{
             Cell[
              BoxData[
               MakeBoxes[
                Hyperlink[
                "http://reference.wolfram.com/mathematica/Genealogy/ref/\
TreeToAdjacencyList.html"], StandardForm]], "Input", TextClipboardType -> 
              "PlainText"]}, Visible -> False]]; 
        SelectionMove[DocumentationSearch`Private`nb$, All, Notebook]; 
        FrontEndTokenExecute[DocumentationSearch`Private`nb$, "Copy"]; 
        NotebookClose[DocumentationSearch`Private`nb$]; 
        Null], "\<\"Go to web url\"\>":>FrontEndExecute[{
         NotebookLocate[{
           URL[
            StringJoin[
            "http://reference.wolfram.com/mathematica/", 
             "Genealogy/ref/TreeToAdjacencyList", ".html"]], None}]}]},
       Appearance->None,
       MenuAppearance->Automatic]],
      LineSpacing->{1.4, 0}]
    }], "AnchorBar"]}
  }]], "AnchorBarGrid",
 CellID->1],

Cell["TreeToAdjacencyList", "ObjectName",
 CellID->1224892054],

Cell[BoxData[GridBox[{
   {"", Cell[TextData[{
     Cell[BoxData[
      RowBox[{"TreeToAdjacencyList", "[", "tree", "]"}]], "InlineFormula"],
     " \[LineSeparator]gets the adjacency list of ",
     StyleBox["tree",
      FontSlant->"Italic"],
     "., where the ",
     StyleBox["i ",
      FontSlant->"Italic"],
     "th element of the adjacency list is the parent node of node ",
     StyleBox["i",
      FontSlant->"Italic"],
     ". "
    }]]}
  }]], "Usage",
 GridBoxOptions->{
 GridBoxBackground->{
  "Columns" -> {{None}}, "ColumnsIndexed" -> {}, "Rows" -> {{None}}, 
   "RowsIndexed" -> {}}},
 CellID->982511436],

Cell[CellGroupData[{

Cell[TextData[ButtonBox["EXAMPLES",
 BaseStyle->None,
 Appearance->{Automatic, None},
 Evaluator->None,
 Method->"Preemptive",
 ButtonFunction:>(FrontEndExecute[{
    FrontEnd`SelectionMove[
     FrontEnd`SelectedNotebook[], All, ButtonCell], 
    FrontEndToken["OpenCloseGroup"], 
    FrontEnd`SelectionMove[
     FrontEnd`SelectedNotebook[], After, 
     CellContents]}]& )]], "PrimaryExamplesSection",
 CellTags->"PrimaryExamplesSection",
 CellID->139389486],

Cell[CellGroupData[{

Cell[TextData[{
 "Basic Examples",
 "\[NonBreakingSpace]\[NonBreakingSpace]",
 Cell["(1)", "ExampleCount"]
}], "ExampleSection",
 CellID->558086636],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"<<", "Genealogy`"}]], "Input",
 CellLabel->"In[1]:=",
 CellID->90703928],

Cell[BoxData[
 RowBox[{
  StyleBox[
   RowBox[{"General", "::", "compat"}], "MessageName"], ":", 
  " ", "\<\"Combinatorica Graph and Permutations functionality has been \
superseded by preloaded functionaliy. The package now being loaded may \
conflict with this. Please see the Compatibility Guide for details.\"\>"}]], \
"Message", "MSG",
 CellLabel->"During evaluation of In[1]:=",
 CellID->18859487]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[{
 RowBox[{
  RowBox[{"n", "=", "5"}], ";"}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{"tree", "=", 
   RowBox[{"RandomCoalescentTree", "[", "n", "]"}]}], 
  ";"}], "\[IndentingNewLine]", 
 RowBox[{"ShowLabeledGraph", "[", 
  RowBox[{
   RowBox[{"TreeToCombinatoricaGraph", "[", "%", "]"}], ",", 
   RowBox[{"VertexColor", "\[Rule]", "Red"}], ",", 
   RowBox[{"EdgeColor", "\[Rule]", "Blue"}]}], "]"}], "\[IndentingNewLine]", 
 RowBox[{"TreeToAdjacencyList", "[", "tree", "]"}], "\[IndentingNewLine]", 
 RowBox[{"%", "\[Equal]", 
  RowBox[{"(", 
   RowBox[{
    RowBox[{
     RowBox[{"TreeNodeParent", "[", 
      RowBox[{"tree", ",", "#"}], "]"}], "&"}], "/@", 
    RowBox[{"Range", "[", 
     RowBox[{
      RowBox[{"2", " ", "n"}], "-", "1"}], "]"}]}], ")"}]}]}], "Input",
 CellLabel->"In[2]:=",
 CellID->54007850],

Cell[BoxData[
 GraphicsBox[{{
    {RGBColor[0, 0, 1], Thickness[0.005], 
     ArrowBox[{{0.4, 0.}, {0.5, 0.009624511227024082}}], 
     {GrayLevel[0]}}, 
    {RGBColor[0, 0, 1], Thickness[0.005], 
     ArrowBox[{{0.6000000000000001, 0.}, {0.5, 0.009624511227024082}}], 
     {GrayLevel[0]}}, 
    {RGBColor[0, 0, 1], Thickness[0.005], 
     ArrowBox[{{0.5, 0.009624511227024082}, {0.65, 0.010399265194569046`}}], 
     {GrayLevel[0]}}, 
    {RGBColor[0, 0, 1], Thickness[0.005], 
     ArrowBox[{{0.8, 0.}, {0.65, 0.010399265194569046`}}], 
     {GrayLevel[0]}}, 
    {RGBColor[0, 0, 1], Thickness[0.005], 
     ArrowBox[{{0.2, 0.}, {0.42500000000000004`, 0.04227661849329123}}], 
     {GrayLevel[0]}}, 
    {RGBColor[0, 0, 1], Thickness[0.005], 
     ArrowBox[{{0.65, 0.010399265194569046`}, {0.42500000000000004`, 
      0.04227661849329123}}], 
     {GrayLevel[0]}}, 
    {RGBColor[0, 0, 1], Thickness[0.005], 
     ArrowBox[{{0.42500000000000004`, 0.04227661849329123}, {0.7125, 
      0.08334638818879492}}], 
     {GrayLevel[0]}}, 
    {RGBColor[0, 0, 1], Thickness[0.005], 
     ArrowBox[{{1., 0.}, {0.7125, 0.08334638818879492}}], 
     {GrayLevel[0]}}}, {
    {RGBColor[1, 0, 0], 
     {PointSize[0.025], PointBox[{1., 0.}]}, 
     {GrayLevel[0], InsetBox["1", Scaled[{-0.02, -0.02}, {1., 0.}], {1, 0}]}, 
     {GrayLevel[0]}}, 
    {RGBColor[1, 0, 0], 
     {PointSize[0.025], PointBox[{0.6000000000000001, 0.}]}, 
     {GrayLevel[0], 
      InsetBox["2", 
       Scaled[{-0.02, -0.02}, {0.6000000000000001, 0.}], {1, 0}]}, 
     {GrayLevel[0]}}, 
    {RGBColor[1, 0, 0], 
     {PointSize[0.025], PointBox[{0.4, 0.}]}, 
     {GrayLevel[0], InsetBox["3", Scaled[{-0.02, -0.02}, {0.4, 0.}], {1, 0}]}, 
     {GrayLevel[0]}}, 
    {RGBColor[1, 0, 0], 
     {PointSize[0.025], PointBox[{0.8, 0.}]}, 
     {GrayLevel[0], InsetBox["4", Scaled[{-0.02, -0.02}, {0.8, 0.}], {1, 0}]}, 
     {GrayLevel[0]}}, 
    {RGBColor[1, 0, 0], 
     {PointSize[0.025], PointBox[{0.2, 0.}]}, 
     {GrayLevel[0], InsetBox["5", Scaled[{-0.02, -0.02}, {0.2, 0.}], {1, 0}]}, 
     {GrayLevel[0]}}, 
    {RGBColor[1, 0, 0], 
     {PointSize[0.025], PointBox[{0.5, 0.009624511227024082}]}, 
     {GrayLevel[0], 
      InsetBox["6", 
       Scaled[{-0.02, -0.02}, {0.5, 0.009624511227024082}], {1, 0}]}, 
     {GrayLevel[0]}}, 
    {RGBColor[1, 0, 0], 
     {PointSize[0.025], PointBox[{0.65, 0.010399265194569046`}]}, 
     {GrayLevel[0], 
      InsetBox["7", 
       Scaled[{-0.02, -0.02}, {0.65, 0.010399265194569046`}], {1, 0}]}, 
     {GrayLevel[0]}}, 
    {RGBColor[1, 0, 0], 
     {PointSize[0.025], 
      PointBox[{0.42500000000000004`, 0.04227661849329123}]}, 
     {GrayLevel[0], 
      InsetBox["8", 
       Scaled[{-0.02, -0.02}, {0.42500000000000004`, 0.04227661849329123}], \
{1, 0}]}, 
     {GrayLevel[0]}}, 
    {RGBColor[1, 0, 0], 
     {PointSize[0.025], PointBox[{0.7125, 0.08334638818879492}]}, 
     {GrayLevel[0], 
      InsetBox["9", 
       Scaled[{-0.02, -0.02}, {0.7125, 0.08334638818879492}], {1, 0}]}, 
     {GrayLevel[0]}}}},
  AlignmentPoint->Center,
  AspectRatio->Automatic,
  Axes->False,
  AxesLabel->None,
  AxesOrigin->Automatic,
  AxesStyle->{},
  Background->None,
  BaseStyle->{},
  BaselinePosition->Automatic,
  ColorOutput->Automatic,
  ContentSelectable->Automatic,
  CoordinatesToolOptions:>Automatic,
  DisplayFunction:>$DisplayFunction,
  Epilog->{},
  FormatType:>TraditionalForm,
  Frame->False,
  FrameLabel->None,
  FrameStyle->{},
  FrameTicks->Automatic,
  FrameTicksStyle->{},
  GridLines->None,
  GridLinesStyle->{},
  ImageMargins->0.,
  ImagePadding->All,
  ImageSize->Automatic,
  ImageSizeRaw->Automatic,
  LabelStyle->{},
  Method->Automatic,
  PlotLabel->None,
  PlotRange->{{0.15000000000000002`, 1.05}, {-0.2749934725722692, 
   0.3583398607610641}},
  PlotRangeClipping->False,
  PlotRangePadding->Automatic,
  PlotRegion->Automatic,
  PreserveImageOptions->Automatic,
  Prolog->{},
  RotateLabel->True,
  Ticks->Automatic,
  TicksStyle->{}]], "Output",
 ImageSize->{364, 253},
 ImageMargins->{{0, 0}, {0, 0}},
 ImageRegion->{{0, 1}, {0, 1}},
 CellLabel->"Out[2]=",
 CellID->65287473],

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{
  "9", ",", "6", ",", "6", ",", "7", ",", "8", ",", "7", ",", "8", ",", "9", 
   ",", 
   RowBox[{"-", "1"}]}], "}"}]], "Output",
 ImageSize->{177, 15},
 ImageMargins->{{0, 0}, {0, 0}},
 ImageRegion->{{0, 1}, {0, 1}},
 CellLabel->"Out[2]=",
 CellID->92465505],

Cell[BoxData["True"], "Output",
 ImageSize->{32, 15},
 ImageMargins->{{0, 0}, {0, 0}},
 ImageRegion->{{0, 1}, {0, 1}},
 CellLabel->"Out[2]=",
 CellID->246657515]
}, Open  ]]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell[TextData[ButtonBox["SEE ALSO",
 BaseStyle->None,
 Appearance->{Automatic, None},
 Evaluator->None,
 Method->"Preemptive",
 ButtonFunction:>(FrontEndExecute[{
    FrontEnd`SelectionMove[
     FrontEnd`SelectedNotebook[], All, ButtonCell], 
    FrontEndToken["OpenCloseGroup"], 
    FrontEnd`SelectionMove[
     FrontEnd`SelectedNotebook[], After, 
     CellContents]}]& )]], "SeeAlsoSection",
 CellID->1255426704],

Cell[TextData[Cell[BoxData[
 StyleBox[
  ButtonBox["TreeNodeParent",
   BaseStyle->"Link",
   ButtonData->"paclet:Genealogy/ref/TreeNodeParent"],
  FontFamily->"Verdana"]], "InlineFormula"]], "SeeAlso",
 CellID->929782353]
}, Open  ]],

Cell[CellGroupData[{

Cell[TextData[ButtonBox["MORE ABOUT",
 BaseStyle->None,
 Appearance->{Automatic, None},
 Evaluator->None,
 Method->"Preemptive",
 ButtonFunction:>(FrontEndExecute[{
    FrontEnd`SelectionMove[
     FrontEnd`SelectedNotebook[], All, ButtonCell], 
    FrontEndToken["OpenCloseGroup"], 
    FrontEnd`SelectionMove[
     FrontEnd`SelectedNotebook[], After, 
     CellContents]}]& )]], "MoreAboutSection",
 CellID->38303248],

Cell[TextData[ButtonBox["Genealogy",
 BaseStyle->"Link",
 ButtonData->"paclet:Genealogy/guide/Genealogy"]], "MoreAbout",
 CellID->37845943]
}, Open  ]],

Cell[" ", "FooterCell"]
},
Saveable->False,
ScreenStyleEnvironment->"Working",
WindowSize->{725, 730},
WindowMargins->{{0, Automatic}, {Automatic, 0}},
WindowTitle->"TreeToAdjacencyList - Wolfram Mathematica",
TaggingRules->{
 "ModificationHighlight" -> False, 
  "Metadata" -> {
   "built" -> "{2013, 7, 27, 22, 7, 23.5644035}", "context" -> "Genealogy`", 
    "keywords" -> {}, "index" -> True, "label" -> "Genealogy Paclet Symbol", 
    "language" -> "en", "paclet" -> "Genealogy", "status" -> "None", 
    "summary" -> 
    "TreeToAdjacencyList[tree] gets the adjacency list of tree., where the i \
th element of the adjacency list is the parent node of node i. ", 
    "synonyms" -> {}, "title" -> "TreeToAdjacencyList", "type" -> "Symbol", 
    "uri" -> "Genealogy/ref/TreeToAdjacencyList"}, "LinkTrails" -> "", 
  "SearchTextTranslated" -> ""},
CellContext->"Global`",
FrontEndVersion->"9.0 for Microsoft Windows (64-bit) (January 25, 2013)",
StyleDefinitions->Notebook[{
   Cell[
    StyleData[
    StyleDefinitions -> FrontEnd`FileName[{"Wolfram"}, "Reference.nb"]]], 
   Cell[
    StyleData["Input"], CellContext -> "Global`"], 
   Cell[
    StyleData["Output"], CellContext -> "Global`"]}, Visible -> False, 
  FrontEndVersion -> "9.0 for Microsoft Windows (64-bit) (January 25, 2013)", 
  StyleDefinitions -> "Default.nb"]
]
(* End of Notebook Content *)

(* Internal cache information *)
(*CellTagsOutline
CellTagsIndex->{
 "PrimaryExamplesSection"->{
  Cell[4180, 119, 461, 13, 70, "PrimaryExamplesSection",
   CellTags->"PrimaryExamplesSection",
   CellID->139389486]}
 }
*)
(*CellTagsIndex
CellTagsIndex->{
 {"PrimaryExamplesSection", 13588, 416}
 }
*)
(*NotebookFileOutline
Notebook[{
Cell[612, 21, 76, 1, 70, "SymbolColorBar"],
Cell[691, 24, 182, 5, 70, "LinkTrail"],
Cell[876, 31, 2589, 59, 70, "AnchorBarGrid",
 CellID->1],
Cell[3468, 92, 62, 1, 70, "ObjectName",
 CellID->1224892054],
Cell[3533, 95, 622, 20, 70, "Usage",
 CellID->982511436],
Cell[CellGroupData[{
Cell[4180, 119, 461, 13, 70, "PrimaryExamplesSection",
 CellTags->"PrimaryExamplesSection",
 CellID->139389486],
Cell[CellGroupData[{
Cell[4666, 136, 148, 5, 70, "ExampleSection",
 CellID->558086636],
Cell[CellGroupData[{
Cell[4839, 145, 96, 3, 70, "Input",
 CellID->90703928],
Cell[4938, 150, 404, 9, 70, "Message",
 CellID->18859487]
}, Open  ]],
Cell[CellGroupData[{
Cell[5379, 164, 833, 23, 70, "Input",
 CellID->54007850],
Cell[6215, 189, 4128, 119, 274, "Output",
 CellID->65287473],
Cell[10346, 310, 299, 10, 36, "Output",
 CellID->92465505],
Cell[10648, 322, 161, 5, 36, "Output",
 CellID->246657515]
}, Open  ]]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[10870, 334, 417, 12, 70, "SeeAlsoSection",
 CellID->1255426704],
Cell[11290, 348, 222, 6, 70, "SeeAlso",
 CellID->929782353]
}, Open  ]],
Cell[CellGroupData[{
Cell[11549, 359, 419, 12, 70, "MoreAboutSection",
 CellID->38303248],
Cell[11971, 373, 139, 3, 70, "MoreAbout",
 CellID->37845943]
}, Open  ]],
Cell[12125, 379, 23, 0, 70, "FooterCell"]
}
]
*)

(* End of internal cache information *)

