(* Content-type: application/vnd.wolfram.mathematica *)

(*** Wolfram Notebook File ***)
(* http://www.wolfram.com/nb *)

(* CreatedBy='Mathematica 8.0' *)

(*CacheID: 234*)
(* Internal cache information:
NotebookFileLineBreakTest
NotebookFileLineBreakTest
NotebookDataPosition[       157,          7]
NotebookDataLength[    250790,       5987]
NotebookOptionsPosition[    244220,       5776]
NotebookOutlinePosition[    246494,       5831]
CellTagsIndexPosition[    246407,       5826]
WindowTitle->AdaptiveProposal - Wolfram Mathematica
WindowFrame->Normal*)

(* Beginning of Notebook Content *)
Notebook[{
Cell[" ", "SymbolColorBar",
 CellMargins->{{Inherited, Inherited}, {-5, 0}}],

Cell[TextData[{
 ButtonBox["Metropolis",
  BaseStyle->{"Link", "LinkTrail"},
  ButtonData->"paclet:Metropolis/guide/Metropolis"],
 StyleBox[" > ", "LinkTrailSeparator"]
}], "LinkTrail"],

Cell[BoxData[GridBox[{
   {Cell["METROPOLIS PACLET SYMBOL", "PacletNameCell"], Cell[TextData[{
     Cell[BoxData[
      ActionMenuBox[
       FrameBox["\<\"More About \[RightGuillemet]\"\>",
        StripOnInput->False], {"\<\"Metropolis\"\>":>Documentation`HelpLookup[
       "paclet:Metropolis/guide/Metropolis"]},
       Appearance->None,
       MenuAppearance->Automatic]],
      LineSpacing->{1.4, 0}],
     "\[ThickSpace]\[ThickSpace]\[ThickSpace]\[ThickSpace]\[ThickSpace]\
\[ThickSpace]",
     Cell[BoxData[
      ActionMenuBox[
       FrameBox["\<\"URL \[RightGuillemet]\"\>",
        StripOnInput->False], {"\<\"Metropolis/ref/AdaptiveProposal\"\>":>
       None, "\<\"Copy Mathematica url\"\>":>
       Module[{DocumentationSearch`Private`nb$}, 
        DocumentationSearch`Private`nb$ = NotebookPut[
           Notebook[{
             Cell["Metropolis/ref/AdaptiveProposal"]}, Visible -> False]]; 
        SelectionMove[DocumentationSearch`Private`nb$, All, Notebook]; 
        FrontEndTokenExecute[DocumentationSearch`Private`nb$, "Copy"]; 
        NotebookClose[DocumentationSearch`Private`nb$]; Null], 
       Delimiter, "\<\"Copy web url\"\>":>
       Module[{DocumentationSearch`Private`nb$}, 
        DocumentationSearch`Private`nb$ = NotebookPut[
           Notebook[{
             Cell[
              BoxData[
               MakeBoxes[
                Hyperlink[
                "http://reference.wolfram.com/mathematica/Metropolis/ref/\
AdaptiveProposal.html"], StandardForm]], "Input", TextClipboardType -> 
              "PlainText"]}, Visible -> False]]; 
        SelectionMove[DocumentationSearch`Private`nb$, All, Notebook]; 
        FrontEndTokenExecute[DocumentationSearch`Private`nb$, "Copy"]; 
        NotebookClose[DocumentationSearch`Private`nb$]; 
        Null], "\<\"Go to web url\"\>":>FrontEndExecute[{
         NotebookLocate[{
           URL[
            StringJoin[
            "http://reference.wolfram.com/mathematica/", 
             "Metropolis/ref/AdaptiveProposal", ".html"]], None}]}]},
       Appearance->None,
       MenuAppearance->Automatic]],
      LineSpacing->{1.4, 0}]
    }], "AnchorBar"]}
  }]], "AnchorBarGrid",
 GridBoxOptions->{GridBoxItemSize->{"Columns" -> {
     Scaled[0.65], {
      Scaled[0.34]}}, "ColumnsIndexed" -> {}, "Rows" -> {{1.}}, 
   "RowsIndexed" -> {}}},
 CellID->1],

Cell[CellGroupData[{

Cell["AdaptiveProposal", "ObjectName",
 CellID->1224892054],

Cell[BoxData[GridBox[{
   {"", Cell[TextData[{
     Cell[BoxData[
      RowBox[{"AdaptiveProposal", "[", 
       RowBox[{
        StyleBox["isadp", "TI"], ",", " ", 
        StyleBox["indaccept", "TI"], ",", " ", 
        StyleBox["adprate", "TI"], ",", " ", 
        StyleBox["adpset", "TI"], ",", " ", 
        StyleBox["xls", "TI"]}], "]"}]], "InlineFormula"],
     "\[LineSeparator]gives updated ",
     StyleBox["adpset",
      FontSlant->"Italic"],
     StyleBox[". ", "TI"],
     "It calculates the recent accept rate, and the scale and the covariance \
of proposal distributions with adpative rate ",
     StyleBox["adprate", "TI"],
     " if ",
     StyleBox["isadp", "TI"],
     ". The ",
     StyleBox["xls", "TI"],
     " is a list of random sample x, and the ",
     StyleBox["indaccept", "TI"],
     " is the average accept probability of the corresponding prposal list."
    }]]},
   {"", Cell[TextData[{
     Cell[BoxData[
      RowBox[{"AdaptiveProposal", "[", 
       RowBox[{
        StyleBox["isadp", "TI"], ",", " ", 
        StyleBox["indacceptls", "TI"], ",", " ", 
        StyleBox["adprate", "TI"], ",", " ", 
        StyleBox["adpset", "TI"], ",", " ", 
        RowBox[{"{", 
         RowBox[{
          StyleBox[
           RowBox[{"xls_", "1"}], "TI"], 
          StyleBox[",", "TI"], ".."}], 
         StyleBox["}", "TI"]}]}], "]"}]], "InlineFormula"],
     "\[LineSeparator]gives updated ",
     StyleBox["adpset",
      FontSlant->"Italic"],
     StyleBox[" ", "TI"],
     "for a set of random variables. It calculates the recent list of accept \
rates, and the scale and the covariance of proposal distributions with \
adpative rate ",
     StyleBox["adprate", "TI"],
     " if ",
     StyleBox["isadp", "TI"],
     ". The ",
     StyleBox["xls_i", "TI"],
     " is a list of random sample x_i for i=1,.., and the ",
     StyleBox["indaccept", "TI"],
     " is the corresponding list of average accept probabilities. "
    }]]}
  }]], "Usage",
 GridBoxOptions->{
 GridBoxBackground->{
  "Columns" -> {{None}}, "ColumnsIndexed" -> {}, "Rows" -> {None, {None}}, 
   "RowsIndexed" -> {}}},
 CellID->982511436]
}, Open  ]],

Cell[CellGroupData[{

Cell[TextData[Cell[BoxData[
 ButtonBox[
  FrameBox[
   StyleBox[
    RowBox[{"MORE", " ", "INFORMATION"}], "NotesFrameText"],
   StripOnInput->False],
  Appearance->{Automatic, None},
  BaseStyle->None,
  ButtonFunction:>(FrontEndExecute[{
     FrontEnd`SelectionMove[
      FrontEnd`SelectedNotebook[], All, ButtonCell], 
     FrontEndToken["OpenCloseGroup"], 
     FrontEnd`SelectionMove[
      FrontEnd`SelectedNotebook[], After, CellContents]}]& ),
  Evaluator->None,
  Method->"Preemptive"]]]], "NotesSection",
 CellGroupingRules->{"SectionGrouping", 50},
 CellID->178741072],

Cell[TextData[{
 "The ",
 StyleBox["adpset",
  FontSlant->"Italic"],
 " =  {targetratio, r, q, accept, scale, mu, cov} where the ",
 StyleBox["accept", "TI"],
 " denotes the accept ratio, the ",
 StyleBox["scale", "TI"],
 " measures the width of the proposal distribution and it is used to adjust \
the accept ratio, and the ",
 StyleBox["mu", "TI"],
 " and ",
 StyleBox["cov", "TI"],
 " store the mean and covariance. "
}], "Notes",
 CellID->1067943069]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[ButtonBox["EXAMPLES",
 BaseStyle->None,
 Appearance->{Automatic, None},
 Evaluator->None,
 Method->"Preemptive",
 ButtonFunction:>(FrontEndExecute[{
    FrontEnd`SelectionMove[
     FrontEnd`SelectedNotebook[], All, ButtonCell], 
    FrontEndToken["OpenCloseGroup"], 
    FrontEnd`SelectionMove[
     FrontEnd`SelectedNotebook[], After, 
     CellContents]}]& )]], "PrimaryExamplesSection",
 CellTags->"PrimaryExamplesSection",
 CellID->184452762],

Cell[CellGroupData[{

Cell[TextData[{
 "Basic Examples",
 "\[NonBreakingSpace]\[NonBreakingSpace]",
 Cell["(1)", "ExampleCount"]
}], "ExampleSection",
 CellID->202422922],

Cell[BoxData[
 RowBox[{"Needs", "[", "\"\<MultivariateStatistics`\>\"", "]"}]], "Input",
 CellLabel->"In[1]:=",
 CellID->307212672],

Cell[BoxData[
 RowBox[{"Needs", "[", "\"\<Metropolis`\>\"", "]"}]], "Input",
 CellLabel->"In[2]:=",
 CellID->22466094],

Cell["This samples bivariate normal distribution", "ExampleText",
 CellID->13303408],

Cell[CellGroupData[{

Cell[BoxData[{
 RowBox[{
  RowBox[{"Clear", "[", 
   RowBox[{
   "rho", ",", " ", "sd1", ",", "sd2", ",", "n", ",", "x", ",", "V", ",", 
    "logpdf", ",", "adpset", ",", "scale", ",", "cov", ",", "res", ",", 
    "accept"}], "]"}], ";"}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{"rho", "=", "0.99"}], ";", 
  RowBox[{"sd1", "=", "1"}], ";", 
  RowBox[{"sd2", "=", "10"}], ";"}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{"n", "=", "5000"}], ";", " ", 
  RowBox[{"x", "=", 
   RowBox[{"{", 
    RowBox[{"5", ",", "5"}], "}"}]}], ";"}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{"adpmax", "=", "2000"}], ";"}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{"V", "=", 
   RowBox[{
    RowBox[{"DiagonalMatrix", "[", 
     RowBox[{"{", 
      RowBox[{"sd1", ",", "sd2"}], "}"}], "]"}], ".", " ", 
    RowBox[{"{", 
     RowBox[{
      RowBox[{"{", 
       RowBox[{"1", ",", "rho"}], "}"}], ",", 
      RowBox[{"{", 
       RowBox[{"rho", ",", "1"}], "}"}]}], "}"}], ".", " ", 
    RowBox[{"DiagonalMatrix", "[", 
     RowBox[{"{", 
      RowBox[{"sd1", ",", "sd2"}], "}"}], "]"}]}]}], 
  ";"}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{"logpdf", "=", 
   RowBox[{"Function", "[", 
    RowBox[{"x", ",", 
     RowBox[{"Log", "[", 
      RowBox[{"PDF", "[", 
       RowBox[{
        RowBox[{"MultinormalDistribution", "[", 
         RowBox[{
          RowBox[{"{", 
           RowBox[{"0", ",", "0"}], "}"}], ",", "V"}], "]"}], ",", "x"}], 
       "]"}], "]"}]}], "]"}]}], ";"}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{"adpset", "=", 
   RowBox[{"InitialProposal", "[", 
    RowBox[{"0.234", ",", "x"}], "]"}]}], ";"}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{
   RowBox[{"{", 
    RowBox[{"scale", ",", "cov"}], "}"}], "=", 
   RowBox[{"adpset", "[", 
    RowBox[{"[", 
     RowBox[{"{", 
      RowBox[{
       RowBox[{"-", "3"}], ",", 
       RowBox[{"-", "1"}]}], "}"}], "]"}], "]"}]}], 
  ";"}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{"res", "=", 
   RowBox[{"{", "}"}]}], ";"}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{
   RowBox[{"Do", "[", 
    RowBox[{
     RowBox[{
      RowBox[{"temp", "=", 
       RowBox[{"StudentTMetropolis", "[", 
        RowBox[{
        "logpdf", ",", "x", ",", "scale", ",", "cov", ",", " ", "2", ",", " ", 
         RowBox[{"SamplingPattern", "\[Rule]", 
          RowBox[{"{", 
           RowBox[{"\"\<Cholesky\>\"", ",", "Infinity", ",", "False"}], 
           "}"}]}], ",", 
         RowBox[{"IsReturnLast", "\[Rule]", "False"}]}], "]"}]}], ";", 
      "\[IndentingNewLine]", 
      RowBox[{"adpset", "=", 
       RowBox[{"AdaptiveProposal", "[", 
        RowBox[{
         RowBox[{"TrueQ", "[", 
          RowBox[{"it", "\[LessEqual]", "adpmax"}], "]"}], ",", " ", 
         RowBox[{"Mean", "[", 
          RowBox[{"temp", "[", 
           RowBox[{"[", 
            RowBox[{"All", ",", 
             RowBox[{"-", "1"}]}], "]"}], "]"}], "]"}], ",", " ", 
         RowBox[{"5", "/", "adpmax"}], ",", " ", "adpset", ",", 
         RowBox[{"temp", "[", 
          RowBox[{"[", 
           RowBox[{"All", ",", "1"}], "]"}], "]"}]}], "]"}]}], ";", 
      "\[IndentingNewLine]", 
      RowBox[{
       RowBox[{"{", 
        RowBox[{"accept", ",", "scale", ",", " ", "cov"}], "}"}], "=", 
       RowBox[{"adpset", "[", 
        RowBox[{"[", 
         RowBox[{"{", 
          RowBox[{
           RowBox[{"-", "4"}], ",", 
           RowBox[{"-", "3"}], ",", 
           RowBox[{"-", "1"}]}], "}"}], "]"}], "]"}]}], ";", 
      "\[IndentingNewLine]", 
      RowBox[{"x", "=", 
       RowBox[{"temp", "[", 
        RowBox[{"[", 
         RowBox[{
          RowBox[{"-", "1"}], ",", "1"}], "]"}], "]"}]}], ";", 
      "\[IndentingNewLine]", 
      RowBox[{"AppendTo", "[", 
       RowBox[{"res", ",", 
        RowBox[{"{", 
         RowBox[{"x", ",", "accept", ",", "scale"}], "}"}]}], "]"}]}], ",", 
     RowBox[{"{", 
      RowBox[{"it", ",", 
       RowBox[{"2", "  ", "n"}]}], "}"}]}], "]"}], ";"}], "//", 
  "Timing"}]}], "Input",
 CellLabel->"In[3]:=",
 CellID->231556428],

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{"11.732000000000014`", ",", "Null"}], "}"}]], "Output",
 ImageSize->{99, 15},
 ImageMargins->{{0, 0}, {0, 0}},
 ImageRegion->{{0, 1}, {0, 1}},
 CellLabel->"Out[3]=",
 CellID->202682309]
}, Open  ]],

Cell["Accept ratito:", "ExampleText",
 CellID->671059588],

Cell[CellGroupData[{

Cell[BoxData[{
 RowBox[{"N", "[", 
  RowBox[{"Mean", "[", 
   RowBox[{"res", "[", 
    RowBox[{"[", 
     RowBox[{
      RowBox[{"adpmax", ";;"}], ",", "2"}], "]"}], "]"}], "]"}], 
  "]"}], "\[IndentingNewLine]", 
 RowBox[{"ListLinePlot", "[", 
  RowBox[{
   RowBox[{"res", "[", 
    RowBox[{"[", 
     RowBox[{"All", ",", "2"}], "]"}], "]"}], ",", 
   RowBox[{"PlotRange", "\[Rule]", 
    RowBox[{"{", 
     RowBox[{"Automatic", ",", 
      RowBox[{"{", 
       RowBox[{"0.1", ",", "0.5"}], "}"}]}], "}"}]}], ",", 
   RowBox[{"AxesLabel", "\[Rule]", 
    RowBox[{"{", 
     RowBox[{"\"\<t\>\"", ",", "\"\<accept\>\""}], "}"}]}]}], "]"}]}], "Input",
 CellLabel->"In[4]:=",
 CellID->245470297],

Cell[BoxData["0.23034237330059548`"], "Output",
 ImageSize->{60, 15},
 ImageMargins->{{0, 0}, {0, 0}},
 ImageRegion->{{0, 1}, {0, 1}},
 CellLabel->"Out[4]=",
 CellID->270803938],

Cell[BoxData[
 GraphicsBox[
  TagBox[RasterBox[CompressedData["
1:eJztne2P1FQUxjfxL/HfmMSooARQs0BifIENsJgJfAAiyBcNCnxQNJBMFk1E
zG6W9wU+LC8OUdeEsLCK4g4alxDGBNCsGAMxceOHTUzq0zntnc5tz8zsTnfa
bZ9fOs3ty5yets+ce9reO32yf/ua/id6enr24XMJH7fsEEIIIWmkWiqVk/aB
LAzVUqHHpeie4XKxVi6UqlFlr1golYrWN4qURwaBNHBiMcaJl3Jwvl8uegIp
Fo0UXE0Vy2V8kcLILiYouCGhamYGyz0ewZWMqqiNjOIGgzbjRvALjvcVaiPD
SLpRKBRqNYUfIdzTrZRNEPHCR22a8iBOY2VDSB0vzFAdhBBCCGlJ4EZnI95t
UmYUecW/RxG+5iiXKIpcU79fZYnDCxq8W5Ff/LCh3KyQJyVd94qkATVu1Fco
Bmf7T1PYYCwH6PmGv7whblAVuaLxOsXTit+Uw65RqA2iQW0QDWqDaDARJRoU
BtGgNogGtUE0qA2iwVyUaFAYRIPaIBrUBtFgvkE0KAyiQW0QDWqDaFAbRIO5
aK5Q+6c4TvhvEyiMHNGsvWjEfyZQG/mhSTtztxUx40aOUfunSPtyaiPHKHHD
NDO3UxH2T8kRzfunMG7km8j+KR7UBmkbaKMyOf3w4T9JO0JSB7TxwftXzo9O
Je0ISR2SiA4N3kzaEZI6JG5QGyQMtUE0qA2iQW0QDeaiRINxg2hQG0QD2ti+
7SK1QcIw3yAajBtEQ7RxaGAiaUdI6hBtIB1N2hGSOqgNoiG56N49YzMzs0n7
Qhac6P4pfotRa77EjaefOsx0NPto7UWrVe+1no2zg9oYv3qP0SPDtHgPQrVU
Cv2fuWgDlyqb+s9VJqchD162ZJIm70+JfLuO0QbS0VdePiltRzHJFqTZo9X7
UxqbnQf6pyAdFW1U7z6iNrJJi/enWK/WcbUBSUAMGIs2MGDywYO/u+Ux6R6R
/VP8V83bgrG0cbl8R7SBcRK+kxRBbRANSxsHD4xDHpjE9WzSrpGEkZrGaANX
K6INjJN2jSSMFTdwtXJ+dIraIE5NG88v/dxoQ9pyUBvEqWkDSjDa2NR/7vCn
N6gN4gS0sbr3mCgEKQciCXtPE5OLmughN8/5WJYEhRHWxuzsf8hAkvaRJIOl
DdQs0MbG9Wc/3O+2BEP6gZlTv/zp1HSCIWl/SfewtIEBgeKdt7+UuCGXMAcP
jKN8aGDitVdPmS/ygUvmsfINo429e8aw1FQ0KG/ZPAqpPH78L8rfjP2K8q63
yjJJMklk3Pho/xXRQ1AbKGxYf+bE8QrKqHcgDEgIY7GD6ubsmZ9x/YtFWAfi
OTRwHYXxq/fkcT/Gpm6qTE6jfOO736p3HyW256QV1kWK0YY8pscYAQQhAmcT
qQjKGOOcYg7ONc4ygknvS0fXrDqO+UhOUBPh4hd10M4dX+zbOwapwBoWYf0X
VgzB/nNL3EoKa2LAIslnMCx55ghWwCTGRqKffDyBigwCg1mIEGUYn1/DEuyL
CBLj06du7drpWoO2McBJmMX8byfuY9ekbaS0VcC2EBhv/vA7NCzfRcGEypmZ
WWkmJ9Ur5mP969fuY2Ynidntqb9gAT6IG3APx3Pb1otnRn6CD90M1JHaGB66
KY9ipT0Yxjh6kAH2HecOAxyWr8PVI0e+v1X5o/lz23aOFQ5y8MjD4Hu7vzo6
/CO2JecLIoEzq3uPvrhyWG7hQqtGXStrjkkbNhkwufTZz6Tct3YEi7ALGHCo
B0rXJLhdLt+BcZjF0nWvn97x5iVsDjrEpKh65fLBvnUj2By+iDkowCykvnzZ
IMZSz8pxw8orlg1u3OBqXhZhQzCOTcB5BFUID1tEQR52nzxRuXjhNiYxSHd1
CHVV77HVq47J1sUNzMcKJ47fwjrYWbkNZYY3Np3rXAMaQW1gd+TXDc8xxo5g
EmP4g2MrtQkWYZeTbRWGrUscwEHGT8kID7oK/mYxuUDXVu3YhCemkl3fdxZh
ShI5HD2R4vatF97d/TXkLfEQGsChTlX+FsxF5Xch2oAYRKhYB2XeRc8GyvtT
vJZf4f4pwSzUaAPCkOoY6yC4URtZQGkv6nc+wOKG+WFtoAYRbZi+kPLUHtVl
d/eExEyrdub2bHlGb7JQyeUkqcB8008WyQZvii52mvRP8eaG+qdIPQIlBLUh
rTjYhzpLNI8boW5t9f4pkAeyC9EGriWlZaDcLScZQe+fAtnUZpVLje/Wkbgh
T2ChDVy3Yj5bjWaSpv1T7F6PYW3IHXJkFwga7DqdZ6CNvrUjGEQb0vMxaadI
KpC+0ogPkITcxKM2iGBe5CdtvagNYpC44dSaeFEbJIjRhjzygza2bB5N2imS
Cow2cFVSmZyWx4VJO0VSgck3BIQO/qETEUzcEKSpSYL+kPQAbciDeEIs5P0p
SXtB0gi1QTSsXJQQA+MG0aA2iAa1QTSoDaLBXJRoQBhsQE4iYdDIOuH24+2y
cNpYdJYXncNtWZaGwsX5yCPSuLbFRbdymn2b08pzshCkXCoU5iiMHpI5os90
vcMSIY1QG0SllnBQHoSQBUb5C5e5GqhbsAw2n2xNoKKM17K871LWj8+y/W7m
ziyLtXpNEF4/5qMdpMUr/1pj/9mLZbD5ZFv++ccmVst+h3F/I3FZ9pVc737c
qeXG34a1fsxHW9nwvMXhIwYsg80n27BZLPvfidOy/+MW2cXqsxFFXJbra4XX
j/doR+/JfL/eYCn4e/ENNp9sy6T1S4zDsgkaEuzi9NmJDHWdWA6eYXv9mD3X
N9yJNszN+hiVbH7bUmPGaHlOpuYcN7y6JK4omljc6DzfcKw/e4m9BjR7GKNl
PznybMdr2a8BC/EcjcTyDccLgfO/Tgn/2cuiuE7x3I7fskll6hVBJ5ati8Cu
XqcQQgghhMyX/wHwq7qr
    "], {{0, 130}, {180, 0}}, {0, 255},
    ColorFunction->RGBColor],
   BoxForm`ImageTag["Byte", ColorSpace -> "RGB", Interleaving -> True],
   Selectable->False],
  BaseStyle->"ImageGraphics",
  ImageSize->Automatic,
  ImageSizeRaw->{180, 130},
  PlotRange->{{0, 180}, {0, 130}}]], "Output",
 ImageSize->{184, 130},
 ImageMargins->{{0, 0}, {0, 0}},
 ImageRegion->{{0, 1}, {0, 1}},
 CellLabel->"Out[4]=",
 CellID->242756173]
}, Open  ]],

Cell["Proposal scale", "ExampleText",
 CellID->141799113],

Cell[CellGroupData[{

Cell[BoxData[{
 RowBox[{"N", "[", 
  RowBox[{"Mean", "[", 
   RowBox[{"res", "[", 
    RowBox[{"[", 
     RowBox[{
      RowBox[{"adpmax", ";;"}], ",", "3"}], "]"}], "]"}], "]"}], 
  "]"}], "\[IndentingNewLine]", 
 RowBox[{"ListLinePlot", "[", 
  RowBox[{
   RowBox[{"res", "[", 
    RowBox[{"[", 
     RowBox[{"All", ",", "3"}], "]"}], "]"}], ",", 
   RowBox[{"PlotRange", "\[Rule]", "All"}], ",", 
   RowBox[{"AxesLabel", "\[Rule]", 
    RowBox[{"{", 
     RowBox[{"\"\<t\>\"", ",", "\"\<scale\>\""}], "}"}]}]}], "]"}]}], "Input",
 CellLabel->"In[5]:=",
 CellID->119363098],

Cell[BoxData["2.3930350979923922`"], "Output",
 ImageSize->{53, 15},
 ImageMargins->{{0, 0}, {0, 0}},
 ImageRegion->{{0, 1}, {0, 1}},
 CellLabel->"Out[5]=",
 CellID->159232310],

Cell[BoxData[
 GraphicsBox[
  TagBox[RasterBox[CompressedData["
1:eJztnd9LHFcUxxf6l/TfWCh9KH0oLRULIfSHSaqtJFHiQ5sXfUlekhelW31o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    "], {{0, 122}, {180, 0}}, {0, 255},
    ColorFunction->RGBColor],
   BoxForm`ImageTag["Byte", ColorSpace -> "RGB", Interleaving -> True],
   Selectable->False],
  BaseStyle->"ImageGraphics",
  ImageSize->Automatic,
  ImageSizeRaw->{180, 122},
  PlotRange->{{0, 180}, {0, 122}}]], "Output",
 ImageSize->{184, 122},
 ImageMargins->{{0, 0}, {0, 0}},
 ImageRegion->{{0, 1}, {0, 1}},
 CellLabel->"Out[5]=",
 CellID->95788349]
}, Open  ]],

Cell["Time series", "ExampleText",
 CellID->303467053],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"Table", "[", 
  RowBox[{
   RowBox[{"ListLinePlot", "[", 
    RowBox[{
     RowBox[{"res", "[", 
      RowBox[{"[", 
       RowBox[{
        RowBox[{";;", ";;", "10"}], ",", "1", ",", "i"}], "]"}], "]"}], ",", 
     RowBox[{"PlotRange", "\[Rule]", "All"}], ",", 
     RowBox[{"AxesLabel", "\[Rule]", 
      RowBox[{"{", 
       RowBox[{"\"\<t\>\"", ",", 
        RowBox[{"\"\<x\>\"", "<>", 
         RowBox[{"ToString", "[", "i", "]"}]}]}], "}"}]}]}], "]"}], ",", 
   RowBox[{"{", 
    RowBox[{"i", ",", "2"}], "}"}]}], "]"}]], "Input",
 CellLabel->"In[6]:=",
 CellID->30665817],

Cell[BoxData[
 GraphicsBox[
  TagBox[RasterBox[CompressedData["
1:eJztnX+oZkd5x2/p/6Glghtosb1EI8lGkpqENLp7CtbqGn9EU/O2NdY0tosR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    "], {{0, 118}, {385, 0}}, {0, 255},
    ColorFunction->RGBColor],
   BoxForm`ImageTag["Byte", ColorSpace -> "RGB", Interleaving -> True],
   Selectable->False],
  BaseStyle->"ImageGraphics",
  ImageSize->Automatic,
  ImageSizeRaw->{385, 118},
  PlotRange->{{0, 385}, {0, 118}}]], "Output",
 ImageSize->{389, 118},
 ImageMargins->{{0, 0}, {0, 0}},
 ImageRegion->{{0, 1}, {0, 1}},
 CellLabel->"Out[6]=",
 CellID->179800696]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"ListPlot", "[", 
  RowBox[{"res", "[", 
   RowBox[{"[", 
    RowBox[{
     RowBox[{";;", ";;", "10"}], ",", "1"}], "]"}], "]"}], "]"}]], "Input",
 CellLabel->"In[7]:=",
 CellID->40892383],

Cell[BoxData[
 GraphicsBox[
  TagBox[RasterBox[CompressedData["
1:eJztnU1vE0cYxyP1zqWfoFK5NYdeIyVe20mIk5YmQQ20JBCiyhRVhVYt4oAi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    "], {{0, 119}, {180, 0}}, {0, 255},
    ColorFunction->RGBColor],
   BoxForm`ImageTag["Byte", ColorSpace -> "RGB", Interleaving -> True],
   Selectable->False],
  BaseStyle->"ImageGraphics",
  ImageSize->Automatic,
  ImageSizeRaw->{180, 119},
  PlotRange->{{0, 180}, {0, 119}}]], "Output",
 ImageSize->{184, 119},
 ImageMargins->{{0, 0}, {0, 0}},
 ImageRegion->{{0, 1}, {0, 1}},
 CellLabel->"Out[7]=",
 CellID->3429041]
}, Open  ]],

Cell["Simulated distribution vs target distribution (red line):", \
"ExampleText",
 CellID->57759205],

Cell[CellGroupData[{

Cell[BoxData[{
 RowBox[{
  RowBox[{"Clear", "[", 
   RowBox[{"f", ",", "data"}], "]"}], ";"}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{"f", "[", 
   RowBox[{"x_", ",", "i_"}], "]"}], ":=", 
  RowBox[{"PDF", "[", 
   RowBox[{
    RowBox[{"NormalDistribution", "[", 
     RowBox[{"0", ",", 
      RowBox[{"Sqrt", "[", 
       RowBox[{"V", "[", 
        RowBox[{"[", 
         RowBox[{"i", ",", "i"}], "]"}], "]"}], "]"}]}], "]"}], ",", "x"}], 
   "]"}]}], "\[IndentingNewLine]", 
 RowBox[{"Table", "[", "\[IndentingNewLine]", 
  RowBox[{
   RowBox[{
    RowBox[{"data", "=", 
     RowBox[{"res", "[", 
      RowBox[{"[", 
       RowBox[{
        RowBox[{"adpmax", ";;"}], ",", "1", ",", "i"}], "]"}], "]"}]}], ";", 
    "\[IndentingNewLine]", 
    RowBox[{"Show", "[", 
     RowBox[{
      RowBox[{"Histogram", "[", 
       RowBox[{"data", ",", "Automatic", ",", "\"\<ProbabilityDensity\>\""}], 
       "]"}], ",", 
      RowBox[{"Plot", "[", 
       RowBox[{
        RowBox[{"f", "[", 
         RowBox[{"x", ",", "i"}], "]"}], ",", 
        RowBox[{"{", 
         RowBox[{"x", ",", 
          RowBox[{"Min", "[", "data", "]"}], ",", 
          RowBox[{"Max", "[", "data", "]"}]}], "}"}], ",", 
        RowBox[{"PlotRange", "\[Rule]", "All"}], ",", 
        RowBox[{"PlotStyle", "\[Rule]", 
         RowBox[{"{", 
          RowBox[{"Red", ",", "Thick"}], "}"}]}]}], "]"}], ",", 
      RowBox[{"AxesLabel", "\[Rule]", 
       RowBox[{"{", 
        RowBox[{
         RowBox[{"\"\<x\>\"", "<>", 
          RowBox[{"ToString", "[", "i", "]"}]}], ",", "\"\<freq\>\""}], 
        "}"}]}]}], "]"}]}], ",", 
   RowBox[{"{", 
    RowBox[{"i", ",", "2"}], "}"}]}], "]"}]}], "Input",
 CellLabel->"In[8]:=",
 CellID->13266113],

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{
   GraphicsBox[{
     {RGBColor[0.798413061722744, 0.824719615472648, 0.968322270542458], 
      EdgeForm[Opacity[0.7]], {}, 
      {RGBColor[0.798413061722744, 0.824719615472648, 0.968322270542458], 
       EdgeForm[Opacity[0.7]], 
       TagBox[
        TooltipBox[
         TagBox[
          DynamicBox[{
            FEPrivate`If[
             CurrentValue["MouseOver"], 
             EdgeForm[{
               GrayLevel[0.5], 
               AbsoluteThickness[1.5], 
               Opacity[0.66]}], {}, {}], 
            
            RectangleBox[{-4., 0.}, {-3.8, 0.0012498437695288077`}, 
             "RoundingRadius" -> 0]},
           
           ImageSizeCache->{{15.5, 20.445578231292515`}, {42.889204531643735`,
             44.114928368734745`}}],
          StatusArea[#, 0.0012498437695288077`]& ,
          TagBoxNote->"0.0012498437695288077"],
         StyleBox["0.0012498437695288077`", {
           GrayLevel[0]}, StripOnInput -> False]],
        Annotation[#, 
         Style[0.0012498437695288077`, {
           GrayLevel[0]}], "Tooltip"]& ], {}, {}, 
       TagBox[
        TooltipBox[
         TagBox[
          DynamicBox[{
            FEPrivate`If[
             CurrentValue["MouseOver"], 
             EdgeForm[{
               GrayLevel[0.5], 
               AbsoluteThickness[1.5], 
               Opacity[0.66]}], {}, {}], 
            
            RectangleBox[{-3.4, 0.}, {-3.2, 0.0037495313085864324`}, 
             "RoundingRadius" -> 0]},
           
           ImageSizeCache->{{27.336734693877546`, 32.28231292517006}, {
            42.43775685746171, 44.114928368734745`}}],
          StatusArea[#, 0.0037495313085864324`]& ,
          TagBoxNote->"0.0037495313085864324"],
         StyleBox["0.0037495313085864324`", {
           GrayLevel[0]}, StripOnInput -> False]],
        Annotation[#, 
         Style[0.0037495313085864324`, {
           GrayLevel[0]}], "Tooltip"]& ], 
       TagBox[
        TooltipBox[
         TagBox[
          DynamicBox[{
            FEPrivate`If[
             CurrentValue["MouseOver"], 
             EdgeForm[{
               GrayLevel[0.5], 
               AbsoluteThickness[1.5], 
               Opacity[0.66]}], {}, {}], 
            
            RectangleBox[{-3.2, 0.}, {-3., 0.0006249218847644038}, 
             "RoundingRadius" -> 0]},
           
           ImageSizeCache->{{31.28231292517006, 36.22789115646258}, {
            43.00206645018924, 44.114928368734745`}}],
          StatusArea[#, 0.0006249218847644038]& ,
          TagBoxNote->"0.0006249218847644038"],
         StyleBox["0.0006249218847644038`", {
           GrayLevel[0]}, StripOnInput -> False]],
        Annotation[#, 
         Style[0.0006249218847644038, {
           GrayLevel[0]}], "Tooltip"]& ], 
       TagBox[
        TooltipBox[
         TagBox[
          DynamicBox[{
            FEPrivate`If[
             CurrentValue["MouseOver"], 
             EdgeForm[{
               GrayLevel[0.5], 
               AbsoluteThickness[1.5], 
               Opacity[0.66]}], {}, {}], 
            
            RectangleBox[{-3., 0.}, {-2.8, 0.008123984501937251}, 
             "RoundingRadius" -> 0]},
           
           ImageSizeCache->{{35.22789115646258, 40.173469387755105`}, {
            41.64772342764317, 44.114928368734745`}}],
          StatusArea[#, 0.008123984501937251]& ,
          TagBoxNote->"0.008123984501937251"],
         StyleBox["0.008123984501937251`", {
           GrayLevel[0]}, StripOnInput -> False]],
        Annotation[#, 
         Style[0.008123984501937251, {
           GrayLevel[0]}], "Tooltip"]& ], 
       TagBox[
        TooltipBox[
         TagBox[
          DynamicBox[{
            FEPrivate`If[
             CurrentValue["MouseOver"], 
             EdgeForm[{
               GrayLevel[0.5], 
               AbsoluteThickness[1.5], 
               Opacity[0.66]}], {}, {}], 
            
            RectangleBox[{-2.8, 0.}, {-2.6, 0.01062367204099489}, 
             "RoundingRadius" -> 0]},
           
           ImageSizeCache->{{39.173469387755105`, 44.11904761904761}, {
            41.196275753461144`, 44.114928368734745`}}],
          StatusArea[#, 0.01062367204099489]& ,
          TagBoxNote->"0.01062367204099489"],
         StyleBox["0.01062367204099489`", {
           GrayLevel[0]}, StripOnInput -> False]],
        Annotation[#, 
         Style[0.01062367204099489, {
           GrayLevel[0]}], "Tooltip"]& ], 
       TagBox[
        TooltipBox[
         TagBox[
          DynamicBox[{
            FEPrivate`If[
             CurrentValue["MouseOver"], 
             EdgeForm[{
               GrayLevel[0.5], 
               AbsoluteThickness[1.5], 
               Opacity[0.66]}], {}, {}], 
            
            RectangleBox[{-2.6, 0.}, {-2.4, 0.015623047119110098`}, 
             "RoundingRadius" -> 0]},
           
           ImageSizeCache->{{43.11904761904761, 48.064625850340136`}, {
            40.29338040509711, 44.114928368734745`}}],
          StatusArea[#, 0.015623047119110098`]& ,
          TagBoxNote->"0.015623047119110098"],
         StyleBox["0.015623047119110098`", {
           GrayLevel[0]}, StripOnInput -> False]],
        Annotation[#, 
         Style[0.015623047119110098`, {
           GrayLevel[0]}], "Tooltip"]& ], 
       TagBox[
        TooltipBox[
         TagBox[
          DynamicBox[{
            FEPrivate`If[
             CurrentValue["MouseOver"], 
             EdgeForm[{
               GrayLevel[0.5], 
               AbsoluteThickness[1.5], 
               Opacity[0.66]}], {}, {}], 
            
            RectangleBox[{-2.4, 0.}, {-2.2, 0.03062117235345586}, 
             "RoundingRadius" -> 0]},
           
           ImageSizeCache->{{47.064625850340136`, 52.010204081632644`}, {
            37.58469436000497, 44.114928368734745`}}],
          StatusArea[#, 0.03062117235345586]& ,
          TagBoxNote->"0.03062117235345586"],
         StyleBox["0.03062117235345586`", {
           GrayLevel[0]}, StripOnInput -> False]],
        Annotation[#, 
         Style[0.03062117235345586, {
           GrayLevel[0]}], "Tooltip"]& ], 
       TagBox[
        TooltipBox[
         TagBox[
          DynamicBox[{
            FEPrivate`If[
             CurrentValue["MouseOver"], 
             EdgeForm[{
               GrayLevel[0.5], 
               AbsoluteThickness[1.5], 
               Opacity[0.66]}], {}, {}], 
            
            RectangleBox[{-2.2, 0.}, {-2., 0.036245469316335426`}, 
             "RoundingRadius" -> 0]},
           
           ImageSizeCache->{{51.010204081632644`, 55.955782312925166`}, {
            36.56893709309543, 44.114928368734745`}}],
          StatusArea[#, 0.036245469316335426`]& ,
          TagBoxNote->"0.036245469316335426"],
         StyleBox["0.036245469316335426`", {
           GrayLevel[0]}, StripOnInput -> False]],
        Annotation[#, 
         Style[0.036245469316335426`, {
           GrayLevel[0]}], "Tooltip"]& ], 
       TagBox[
        TooltipBox[
         TagBox[
          DynamicBox[{
            FEPrivate`If[
             CurrentValue["MouseOver"], 
             EdgeForm[{
               GrayLevel[0.5], 
               AbsoluteThickness[1.5], 
               Opacity[0.66]}], {}, {}], 
            
            RectangleBox[{-2., 0.}, {-1.8, 0.11061117360329961`}, 
             "RoundingRadius" -> 0]},
           
           ImageSizeCache->{{54.955782312925166`, 59.90136054421768}, {
            23.138368786180255`, 44.114928368734745`}}],
          StatusArea[#, 0.11061117360329961`]& ,
          TagBoxNote->"0.11061117360329961"],
         StyleBox["0.11061117360329961`", {
           GrayLevel[0]}, StripOnInput -> False]],
        Annotation[#, 
         Style[0.11061117360329961`, {
           GrayLevel[0]}], "Tooltip"]& ], 
       TagBox[
        TooltipBox[
         TagBox[
          DynamicBox[{
            FEPrivate`If[
             CurrentValue["MouseOver"], 
             EdgeForm[{
               GrayLevel[0.5], 
               AbsoluteThickness[1.5], 
               Opacity[0.66]}], {}, {}], 
            
            RectangleBox[{-1.8, 0.}, {-1.6, 0.10311211098612676`}, 
             "RoundingRadius" -> 0]},
           
           ImageSizeCache->{{58.90136054421768, 63.846938775510196`}, {
            24.492711808726323`, 44.114928368734745`}}],
          StatusArea[#, 0.10311211098612676`]& ,
          TagBoxNote->"0.10311211098612676"],
         StyleBox["0.10311211098612676`", {
           GrayLevel[0]}, StripOnInput -> False]],
        Annotation[#, 
         Style[0.10311211098612676`, {
           GrayLevel[0]}], "Tooltip"]& ], 
       TagBox[
        TooltipBox[
         TagBox[
          DynamicBox[{
            FEPrivate`If[
             CurrentValue["MouseOver"], 
             EdgeForm[{
               GrayLevel[0.5], 
               AbsoluteThickness[1.5], 
               Opacity[0.66]}], {}, {}], 
            
            RectangleBox[{-1.6, 0.}, {-1.4, 0.14060742407199087`}, 
             "RoundingRadius" -> 0]},
           
           ImageSizeCache->{{62.846938775510196`, 67.79251700680271}, {
            17.720996695996014`, 44.114928368734745`}}],
          StatusArea[#, 0.14060742407199087`]& ,
          TagBoxNote->"0.14060742407199087"],
         StyleBox["0.14060742407199087`", {
           GrayLevel[0]}, StripOnInput -> False]],
        Annotation[#, 
         Style[0.14060742407199087`, {
           GrayLevel[0]}], "Tooltip"]& ], 
       TagBox[
        TooltipBox[
         TagBox[
          DynamicBox[{
            FEPrivate`If[
             CurrentValue["MouseOver"], 
             EdgeForm[{
               GrayLevel[0.5], 
               AbsoluteThickness[1.5], 
               Opacity[0.66]}], {}, {}], 
            
            RectangleBox[{-1.4, 0.}, {-1.2, 0.1774778152730909}, 
             "RoundingRadius" -> 0]},
           
           ImageSizeCache->{{66.79251700680271, 71.73809523809524}, {
            11.062143501811157`, 44.114928368734745`}}],
          StatusArea[#, 0.1774778152730909]& ,
          TagBoxNote->"0.1774778152730909"],
         StyleBox["0.1774778152730909`", {
           GrayLevel[0]}, StripOnInput -> False]],
        Annotation[#, 
         Style[0.1774778152730909, {
           GrayLevel[0]}], "Tooltip"]& ], 
       TagBox[
        TooltipBox[
         TagBox[
          DynamicBox[{
            FEPrivate`If[
             CurrentValue["MouseOver"], 
             EdgeForm[{
               GrayLevel[0.5], 
               AbsoluteThickness[1.5], 
               Opacity[0.66]}], {}, {}], 
            
            RectangleBox[{-1.2, 0.}, {-1., 0.19560054993125864`}, 
             "RoundingRadius" -> 0]},
           
           ImageSizeCache->{{70.73809523809524, 75.68367346938774}, {
            7.789147863991495, 44.114928368734745`}}],
          StatusArea[#, 0.19560054993125864`]& ,
          TagBoxNote->"0.19560054993125864"],
         StyleBox["0.19560054993125864`", {
           GrayLevel[0]}, StripOnInput -> False]],
        Annotation[#, 
         Style[0.19560054993125864`, {
           GrayLevel[0]}], "Tooltip"]& ], 
       TagBox[
        TooltipBox[
         TagBox[
          DynamicBox[{
            FEPrivate`If[
             CurrentValue["MouseOver"], 
             EdgeForm[{
               GrayLevel[0.5], 
               AbsoluteThickness[1.5], 
               Opacity[0.66]}], {}, {}], 
            
            RectangleBox[{-1., 0.}, {-0.8, 0.22497187851518563`}, 
             "RoundingRadius" -> 0]},
           
           ImageSizeCache->{{74.68367346938774, 79.62925170068027}, {
            2.4846376923527345`, 44.114928368734745`}}],
          StatusArea[#, 0.22497187851518563`]& ,
          TagBoxNote->"0.22497187851518563"],
         StyleBox["0.22497187851518563`", {
           GrayLevel[0]}, StripOnInput -> False]],
        Annotation[#, 
         Style[0.22497187851518563`, {
           GrayLevel[0]}], "Tooltip"]& ], 
       TagBox[
        TooltipBox[
         TagBox[
          DynamicBox[{
            FEPrivate`If[
             CurrentValue["MouseOver"], 
             EdgeForm[{
               GrayLevel[0.5], 
               AbsoluteThickness[1.5], 
               Opacity[0.66]}], {}, {}], 
            
            RectangleBox[{-0.8, 0.}, {-0.6, 0.30621172353455806`}, 
             "RoundingRadius" -> 0]},
           
           ImageSizeCache->{{78.62925170068027, 
            83.57482993197279}, {-12.187411718562963`, 44.114928368734745`}}],
          StatusArea[#, 0.30621172353455806`]& ,
          TagBoxNote->"0.30621172353455806"],
         StyleBox["0.30621172353455806`", {
           GrayLevel[0]}, StripOnInput -> False]],
        Annotation[#, 
         Style[0.30621172353455806`, {
           GrayLevel[0]}], "Tooltip"]& ], 
       TagBox[
        TooltipBox[
         TagBox[
          DynamicBox[{
            FEPrivate`If[
             CurrentValue["MouseOver"], 
             EdgeForm[{
               GrayLevel[0.5], 
               AbsoluteThickness[1.5], 
               Opacity[0.66]}], {}, {}], 
            
            RectangleBox[{-0.6, 0.}, {-0.4, 0.3330833645794276}, 
             "RoundingRadius" -> 0]},
           
           ImageSizeCache->{{82.57482993197279, 
            87.5204081632653}, {-17.04047421601973, 44.114928368734745`}}],
          StatusArea[#, 0.3330833645794276]& ,
          TagBoxNote->"0.3330833645794276"],
         StyleBox["0.3330833645794276`", {
           GrayLevel[0]}, StripOnInput -> False]],
        Annotation[#, 
         Style[0.3330833645794276, {
           GrayLevel[0]}], "Tooltip"]& ], 
       TagBox[
        TooltipBox[
         TagBox[
          DynamicBox[{
            FEPrivate`If[
             CurrentValue["MouseOver"], 
             EdgeForm[{
               GrayLevel[0.5], 
               AbsoluteThickness[1.5], 
               Opacity[0.66]}], {}, {}], 
            
            RectangleBox[{-0.4, 0.}, {-0.2, 0.44556930383702037`}, 
             "RoundingRadius" -> 0]},
           
           ImageSizeCache->{{86.5204081632653, 
            91.46598639455782}, {-37.355619554210726`, 44.114928368734745`}}],
          StatusArea[#, 0.44556930383702037`]& ,
          TagBoxNote->"0.44556930383702037"],
         StyleBox["0.44556930383702037`", {
           GrayLevel[0]}, StripOnInput -> False]],
        Annotation[#, 
         Style[0.44556930383702037`, {
           GrayLevel[0]}], "Tooltip"]& ], 
       TagBox[
        TooltipBox[
         TagBox[
          DynamicBox[{
            FEPrivate`If[
             CurrentValue["MouseOver"], 
             EdgeForm[{
               GrayLevel[0.5], 
               AbsoluteThickness[1.5], 
               Opacity[0.66]}], {}, {}], 
            
            RectangleBox[{-0.2, 0.}, {0., 0.4099487564054493}, 
             "RoundingRadius" -> 0]},
           
           ImageSizeCache->{{90.46598639455782, 
            95.41156462585033}, {-30.922490197116907`, 44.114928368734745`}}],
          StatusArea[#, 0.4099487564054493]& ,
          TagBoxNote->"0.4099487564054493"],
         StyleBox["0.4099487564054493`", {
           GrayLevel[0]}, StripOnInput -> False]],
        Annotation[#, 
         Style[0.4099487564054493, {
           GrayLevel[0]}], "Tooltip"]& ], 
       TagBox[
        TooltipBox[
         TagBox[
          DynamicBox[{
            FEPrivate`If[
             CurrentValue["MouseOver"], 
             EdgeForm[{
               GrayLevel[0.5], 
               AbsoluteThickness[1.5], 
               Opacity[0.66]}], {}, {}], 
            
            RectangleBox[{0., 0.}, {0.2, 0.3818272715910511}, 
             "RoundingRadius" -> 0]},
           
           ImageSizeCache->{{94.41156462585033, 
            99.35714285714285}, {-25.843703862569164`, 44.114928368734745`}}],
          StatusArea[#, 0.3818272715910511]& ,
          TagBoxNote->"0.3818272715910511"],
         StyleBox["0.3818272715910511`", {
           GrayLevel[0]}, StripOnInput -> False]],
        Annotation[#, 
         Style[0.3818272715910511, {
           GrayLevel[0]}], "Tooltip"]& ], 
       TagBox[
        TooltipBox[
         TagBox[
          DynamicBox[{
            FEPrivate`If[
             CurrentValue["MouseOver"], 
             EdgeForm[{
               GrayLevel[0.5], 
               AbsoluteThickness[1.5], 
               Opacity[0.66]}], {}, {}], 
            
            RectangleBox[{0.2, 0.}, {0.4, 0.3830771153605799}, 
             "RoundingRadius" -> 0]},
           
           ImageSizeCache->{{98.35714285714285, 
            103.30272108843536`}, {-26.069427699660174`, 
            44.114928368734745`}}],
          StatusArea[#, 0.3830771153605799]& ,
          TagBoxNote->"0.3830771153605799"],
         StyleBox["0.3830771153605799`", {
           GrayLevel[0]}, StripOnInput -> False]],
        Annotation[#, 
         Style[0.3830771153605799, {
           GrayLevel[0]}], "Tooltip"]& ], 
       TagBox[
        TooltipBox[
         TagBox[
          DynamicBox[{
            FEPrivate`If[
             CurrentValue["MouseOver"], 
             EdgeForm[{
               GrayLevel[0.5], 
               AbsoluteThickness[1.5], 
               Opacity[0.66]}], {}, {}], 
            
            RectangleBox[{0.4, 0.}, {0.6, 0.3574553180852394}, 
             "RoundingRadius" -> 0]},
           
           ImageSizeCache->{{102.30272108843536`, 
            107.24829931972788`}, {-21.44208903929445, 44.114928368734745`}}],
          StatusArea[#, 0.3574553180852394]& ,
          TagBoxNote->"0.3574553180852394"],
         StyleBox["0.3574553180852394`", {
           GrayLevel[0]}, StripOnInput -> False]],
        Annotation[#, 
         Style[0.3574553180852394, {
           GrayLevel[0]}], "Tooltip"]& ], 
       TagBox[
        TooltipBox[
         TagBox[
          DynamicBox[{
            FEPrivate`If[
             CurrentValue["MouseOver"], 
             EdgeForm[{
               GrayLevel[0.5], 
               AbsoluteThickness[1.5], 
               Opacity[0.66]}], {}, {}], 
            
            RectangleBox[{0.6, 0.}, {0.8, 0.3105861767279089}, 
             "RoundingRadius" -> 0]},
           
           ImageSizeCache->{{106.24829931972788`, 
            111.19387755102039`}, {-12.977445148381506`, 
            44.114928368734745`}}],
          StatusArea[#, 0.3105861767279089]& ,
          TagBoxNote->"0.3105861767279089"],
         StyleBox["0.3105861767279089`", {
           GrayLevel[0]}, StripOnInput -> False]],
        Annotation[#, 
         Style[0.3105861767279089, {
           GrayLevel[0]}], "Tooltip"]& ], 
       TagBox[
        TooltipBox[
         TagBox[
          DynamicBox[{
            FEPrivate`If[
             CurrentValue["MouseOver"], 
             EdgeForm[{
               GrayLevel[0.5], 
               AbsoluteThickness[1.5], 
               Opacity[0.66]}], {}, {}], 
            
            RectangleBox[{0.8, 0.}, {1., 0.24184476940382457`}, 
             "RoundingRadius" -> 0]},
           
           ImageSizeCache->{{110.19387755102039`, 
            115.13945578231292`}, {-0.5626341083759172, 44.114928368734745`}}],
          StatusArea[#, 0.24184476940382457`]& ,
          TagBoxNote->"0.24184476940382457"],
         StyleBox["0.24184476940382457`", {
           GrayLevel[0]}, StripOnInput -> False]],
        Annotation[#, 
         Style[0.24184476940382457`, {
           GrayLevel[0]}], "Tooltip"]& ], 
       TagBox[
        TooltipBox[
         TagBox[
          DynamicBox[{
            FEPrivate`If[
             CurrentValue["MouseOver"], 
             EdgeForm[{
               GrayLevel[0.5], 
               AbsoluteThickness[1.5], 
               Opacity[0.66]}], {}, {}], 
            
            RectangleBox[{1., 0.}, {1.2, 0.19560054993125864`}, 
             "RoundingRadius" -> 0]},
           
           ImageSizeCache->{{114.13945578231292`, 119.08503401360542`}, {
            7.789147863991495, 44.114928368734745`}}],
          StatusArea[#, 0.19560054993125864`]& ,
          TagBoxNote->"0.19560054993125864"],
         StyleBox["0.19560054993125864`", {
           GrayLevel[0]}, StripOnInput -> False]],
        Annotation[#, 
         Style[0.19560054993125864`, {
           GrayLevel[0]}], "Tooltip"]& ], 
       TagBox[
        TooltipBox[
         TagBox[
          DynamicBox[{
            FEPrivate`If[
             CurrentValue["MouseOver"], 
             EdgeForm[{
               GrayLevel[0.5], 
               AbsoluteThickness[1.5], 
               Opacity[0.66]}], {}, {}], 
            
            RectangleBox[{1.2, 0.}, {1.4, 0.20059992500937388`}, 
             "RoundingRadius" -> 0]},
           
           ImageSizeCache->{{118.08503401360542`, 123.03061224489795`}, {
            6.886252515627447, 44.114928368734745`}}],
          StatusArea[#, 0.20059992500937388`]& ,
          TagBoxNote->"0.20059992500937388"],
         StyleBox["0.20059992500937388`", {
           GrayLevel[0]}, StripOnInput -> False]],
        Annotation[#, 
         Style[0.20059992500937388`, {
           GrayLevel[0]}], "Tooltip"]& ], 
       TagBox[
        TooltipBox[
         TagBox[
          DynamicBox[{
            FEPrivate`If[
             CurrentValue["MouseOver"], 
             EdgeForm[{
               GrayLevel[0.5], 
               AbsoluteThickness[1.5], 
               Opacity[0.66]}], {}, {}], 
            RectangleBox[{1.4, 0.}, {1.6, 0.15248093988251454`}, 
             "RoundingRadius" -> 0]},
           
           ImageSizeCache->{{122.03061224489795`, 126.97619047619047`}, {
            15.57662024363141, 44.114928368734745`}}],
          StatusArea[#, 0.15248093988251454`]& ,
          TagBoxNote->"0.15248093988251454"],
         StyleBox["0.15248093988251454`", {
           GrayLevel[0]}, StripOnInput -> False]],
        Annotation[#, 
         Style[0.15248093988251454`, {
           GrayLevel[0]}], "Tooltip"]& ], 
       TagBox[
        TooltipBox[
         TagBox[
          DynamicBox[{
            FEPrivate`If[
             CurrentValue["MouseOver"], 
             EdgeForm[{
               GrayLevel[0.5], 
               AbsoluteThickness[1.5], 
               Opacity[0.66]}], {}, {}], 
            
            RectangleBox[{1.6, 0.}, {1.8, 0.07061617297837772}, 
             "RoundingRadius" -> 0]},
           
           ImageSizeCache->{{125.97619047619047`, 130.92176870748298`}, {
            30.361531573092613`, 44.114928368734745`}}],
          StatusArea[#, 0.07061617297837772]& ,
          TagBoxNote->"0.07061617297837772"],
         StyleBox["0.07061617297837772`", {
           GrayLevel[0]}, StripOnInput -> False]],
        Annotation[#, 
         Style[0.07061617297837772, {
           GrayLevel[0]}], "Tooltip"]& ], 
       TagBox[
        TooltipBox[
         TagBox[
          DynamicBox[{
            FEPrivate`If[
             CurrentValue["MouseOver"], 
             EdgeForm[{
               GrayLevel[0.5], 
               AbsoluteThickness[1.5], 
               Opacity[0.66]}], {}, {}], 
            
            RectangleBox[{1.8, 0.}, {2., 0.056867891513560816`}, 
             "RoundingRadius" -> 0]},
           
           ImageSizeCache->{{129.92176870748298`, 134.86734693877548`}, {
            32.84449378109373, 44.114928368734745`}}],
          StatusArea[#, 0.056867891513560816`]& ,
          TagBoxNote->"0.056867891513560816"],
         StyleBox["0.056867891513560816`", {
           GrayLevel[0]}, StripOnInput -> False]],
        Annotation[#, 
         Style[0.056867891513560816`, {
           GrayLevel[0]}], "Tooltip"]& ], 
       TagBox[
        TooltipBox[
         TagBox[
          DynamicBox[{
            FEPrivate`If[
             CurrentValue["MouseOver"], 
             EdgeForm[{
               GrayLevel[0.5], 
               AbsoluteThickness[1.5], 
               Opacity[0.66]}], {}, {}], 
            
            RectangleBox[{2., 0.}, {2.2, 0.03687039120109983}, 
             "RoundingRadius" -> 0]},
           
           ImageSizeCache->{{133.86734693877548`, 138.812925170068}, {
            36.45607517454992, 44.114928368734745`}}],
          StatusArea[#, 0.03687039120109983]& ,
          TagBoxNote->"0.03687039120109983"],
         StyleBox["0.03687039120109983`", {
           GrayLevel[0]}, StripOnInput -> False]],
        Annotation[#, 
         Style[0.03687039120109983, {
           GrayLevel[0]}], "Tooltip"]& ], 
       TagBox[
        TooltipBox[
         TagBox[
          DynamicBox[{
            FEPrivate`If[
             CurrentValue["MouseOver"], 
             EdgeForm[{
               GrayLevel[0.5], 
               AbsoluteThickness[1.5], 
               Opacity[0.66]}], {}, {}], 
            
            RectangleBox[{2.2, 0.}, {2.4, 0.02124734408198978}, 
             "RoundingRadius" -> 0]},
           
           ImageSizeCache->{{137.812925170068, 142.75850340136054`}, {
            39.27762313818755, 44.114928368734745`}}],
          StatusArea[#, 0.02124734408198978]& ,
          TagBoxNote->"0.02124734408198978"],
         StyleBox["0.02124734408198978`", {
           GrayLevel[0]}, StripOnInput -> False]],
        Annotation[#, 
         Style[0.02124734408198978, {
           GrayLevel[0]}], "Tooltip"]& ], 
       TagBox[
        TooltipBox[
         TagBox[
          DynamicBox[{
            FEPrivate`If[
             CurrentValue["MouseOver"], 
             EdgeForm[{
               GrayLevel[0.5], 
               AbsoluteThickness[1.5], 
               Opacity[0.66]}], {}, {}], 
            
            RectangleBox[{2.4, 0.}, {2.6, 0.018122734658167713`}, 
             "RoundingRadius" -> 0]},
           
           ImageSizeCache->{{141.75850340136054`, 146.70408163265304`}, {
            39.84193273091508, 44.114928368734745`}}],
          StatusArea[#, 0.018122734658167713`]& ,
          TagBoxNote->"0.018122734658167713"],
         StyleBox["0.018122734658167713`", {
           GrayLevel[0]}, StripOnInput -> False]],
        Annotation[#, 
         Style[0.018122734658167713`, {
           GrayLevel[0]}], "Tooltip"]& ], 
       TagBox[
        TooltipBox[
         TagBox[
          DynamicBox[{
            FEPrivate`If[
             CurrentValue["MouseOver"], 
             EdgeForm[{
               GrayLevel[0.5], 
               AbsoluteThickness[1.5], 
               Opacity[0.66]}], {}, {}], 
            
            RectangleBox[{2.6, 0.}, {2.8, 0.01062367204099489}, 
             "RoundingRadius" -> 0]},
           
           ImageSizeCache->{{145.70408163265304`, 150.64965986394554`}, {
            41.196275753461144`, 44.114928368734745`}}],
          StatusArea[#, 0.01062367204099489]& ,
          TagBoxNote->"0.01062367204099489"],
         StyleBox["0.01062367204099489`", {
           GrayLevel[0]}, StripOnInput -> False]],
        Annotation[#, 
         Style[0.01062367204099489, {
           GrayLevel[0]}], "Tooltip"]& ], 
       TagBox[
        TooltipBox[
         TagBox[
          DynamicBox[{
            FEPrivate`If[
             CurrentValue["MouseOver"], 
             EdgeForm[{
               GrayLevel[0.5], 
               AbsoluteThickness[1.5], 
               Opacity[0.66]}], {}, {}], 
            RectangleBox[{2.8, 0.}, {3., 0.004999375078115231}, 
             "RoundingRadius" -> 0]},
           
           ImageSizeCache->{{149.64965986394554`, 154.59523809523807`}, {
            42.212033020370704`, 44.114928368734745`}}],
          StatusArea[#, 0.004999375078115231]& ,
          TagBoxNote->"0.004999375078115231"],
         StyleBox["0.004999375078115231`", {
           GrayLevel[0]}, StripOnInput -> False]],
        Annotation[#, 
         Style[0.004999375078115231, {
           GrayLevel[0]}], "Tooltip"]& ], 
       TagBox[
        TooltipBox[
         TagBox[
          DynamicBox[{
            FEPrivate`If[
             CurrentValue["MouseOver"], 
             EdgeForm[{
               GrayLevel[0.5], 
               AbsoluteThickness[1.5], 
               Opacity[0.66]}], {}, {}], 
            
            RectangleBox[{3., 0.}, {3.2, 0.0031246094238220196`}, 
             "RoundingRadius" -> 0]},
           
           ImageSizeCache->{{153.59523809523807`, 158.5408163265306}, {
            42.55061877600722, 44.114928368734745`}}],
          StatusArea[#, 0.0031246094238220196`]& ,
          TagBoxNote->"0.0031246094238220196"],
         StyleBox["0.0031246094238220196`", {
           GrayLevel[0]}, StripOnInput -> False]],
        Annotation[#, 
         Style[0.0031246094238220196`, {
           GrayLevel[0]}], "Tooltip"]& ]}, {}, {}}, {{}, {}, 
      {RGBColor[1, 0, 0], Thickness[Large], LineBox[CompressedData["
1:eJwVlnc4Vv8bx/HwPDYlGYkoEZlJZvdJyWgpspVNvvbee2Vliygje4SMrM4n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        "]]}}},
    AspectRatio->NCache[GoldenRatio^(-1), 0.6180339887498948],
    Axes->{True, True},
    AxesLabel->{
      FormBox["\"x1\"", TraditionalForm], 
      FormBox["\"freq\"", TraditionalForm]},
    AxesOrigin->{-4., 0},
    FrameTicks->{{Automatic, Automatic}, {Automatic, Automatic}},
    GridLines->{None, None},
    PlotRange->NCache[{{-4, 
        Rational[16, 5]}, {All, All}}, {{-4, 3.2}, {All, All}}],
    PlotRangePadding->{{
       Scaled[0.02], 
       Scaled[0.02]}, {
       Scaled[0.02], 
       Scaled[0.1]}},
    Ticks->{Automatic, Automatic}], ",", 
   GraphicsBox[{
     {RGBColor[0.798413061722744, 0.824719615472648, 0.968322270542458], 
      EdgeForm[Opacity[0.7]], {}, 
      {RGBColor[0.798413061722744, 0.824719615472648, 0.968322270542458], 
       EdgeForm[Opacity[0.7]], 
       TagBox[
        TooltipBox[
         TagBox[
          DynamicBox[{
            FEPrivate`If[
             CurrentValue["MouseOver"], 
             EdgeForm[{
               GrayLevel[0.5], 
               AbsoluteThickness[1.5], 
               Opacity[0.66]}], {}, {}], 
            
            RectangleBox[{-38., 0.}, {-36., 0.00012498437695288088`}, 
             "RoundingRadius" -> 0]},
           
           ImageSizeCache->{{20.5, 25.30952380952381}, {40.80213463861606, 
            42.024758424985244`}}],
          StatusArea[#, 0.00012498437695288088`]& ,
          TagBoxNote->"0.00012498437695288088"],
         StyleBox["0.00012498437695288088`", {
           GrayLevel[0]}, StripOnInput -> False]],
        Annotation[#, 
         Style[0.00012498437695288088`, {
           GrayLevel[0]}], "Tooltip"]& ], {}, 
       TagBox[
        TooltipBox[
         TagBox[
          DynamicBox[{
            FEPrivate`If[
             CurrentValue["MouseOver"], 
             EdgeForm[{
               GrayLevel[0.5], 
               AbsoluteThickness[1.5], 
               Opacity[0.66]}], {}, {}], 
            
            RectangleBox[{-34., 0.}, {-32., 0.00031246094238220225`}, 
             "RoundingRadius" -> 0]},
           
           ImageSizeCache->{{28.11904761904762, 32.92857142857142}, {
            40.46819895906229, 42.024758424985244`}}],
          StatusArea[#, 0.00031246094238220225`]& ,
          TagBoxNote->"0.00031246094238220225"],
         StyleBox["0.00031246094238220225`", {
           GrayLevel[0]}, StripOnInput -> False]],
        Annotation[#, 
         Style[0.00031246094238220225`, {
           GrayLevel[0]}], "Tooltip"]& ], 
       TagBox[
        TooltipBox[
         TagBox[
          DynamicBox[{
            FEPrivate`If[
             CurrentValue["MouseOver"], 
             EdgeForm[{
               GrayLevel[0.5], 
               AbsoluteThickness[1.5], 
               Opacity[0.66]}], {}, {}], 
            
            RectangleBox[{-32., 0.}, {-30., 0.00031246094238220225`}, 
             "RoundingRadius" -> 0]},
           
           ImageSizeCache->{{31.928571428571423`, 36.738095238095234`}, {
            40.46819895906229, 42.024758424985244`}}],
          StatusArea[#, 0.00031246094238220225`]& ,
          TagBoxNote->"0.00031246094238220225"],
         StyleBox["0.00031246094238220225`", {
           GrayLevel[0]}, StripOnInput -> False]],
        Annotation[#, 
         Style[0.00031246094238220225`, {
           GrayLevel[0]}], "Tooltip"]& ], 
       TagBox[
        TooltipBox[
         TagBox[
          DynamicBox[{
            FEPrivate`If[
             CurrentValue["MouseOver"], 
             EdgeForm[{
               GrayLevel[0.5], 
               AbsoluteThickness[1.5], 
               Opacity[0.66]}], {}, {}], 
            
            RectangleBox[{-30., 0.}, {-28., 0.0005624296962879641}, 
             "RoundingRadius" -> 0]},
           
           ImageSizeCache->{{35.738095238095234`, 40.547619047619044`}, {
            40.02295138632393, 42.024758424985244`}}],
          StatusArea[#, 0.0005624296962879641]& ,
          TagBoxNote->"0.0005624296962879641"],
         StyleBox["0.0005624296962879641`", {
           GrayLevel[0]}, StripOnInput -> False]],
        Annotation[#, 
         Style[0.0005624296962879641, {
           GrayLevel[0]}], "Tooltip"]& ], 
       TagBox[
        TooltipBox[
         TagBox[
          DynamicBox[{
            FEPrivate`If[
             CurrentValue["MouseOver"], 
             EdgeForm[{
               GrayLevel[0.5], 
               AbsoluteThickness[1.5], 
               Opacity[0.66]}], {}, {}], 
            
            RectangleBox[{-28., 0.}, {-26., 0.0007499062617172854}, 
             "RoundingRadius" -> 0]},
           
           ImageSizeCache->{{39.547619047619044`, 44.357142857142854`}, {
            39.689015706770164`, 42.024758424985244`}}],
          StatusArea[#, 0.0007499062617172854]& ,
          TagBoxNote->"0.0007499062617172854"],
         StyleBox["0.0007499062617172854`", {
           GrayLevel[0]}, StripOnInput -> False]],
        Annotation[#, 
         Style[0.0007499062617172854, {
           GrayLevel[0]}], "Tooltip"]& ], 
       TagBox[
        TooltipBox[
         TagBox[
          DynamicBox[{
            FEPrivate`If[
             CurrentValue["MouseOver"], 
             EdgeForm[{
               GrayLevel[0.5], 
               AbsoluteThickness[1.5], 
               Opacity[0.66]}], {}, {}], 
            RectangleBox[{-26., 0.}, {-24., 0.002124734408198975}, 
             "RoundingRadius" -> 0]},
           
           ImageSizeCache->{{43.357142857142854`, 48.166666666666664`}, {
            37.24015405670918, 42.024758424985244`}}],
          StatusArea[#, 0.002124734408198975]& ,
          TagBoxNote->"0.002124734408198975"],
         StyleBox["0.002124734408198975`", {
           GrayLevel[0]}, StripOnInput -> False]],
        Annotation[#, 
         Style[0.002124734408198975, {
           GrayLevel[0]}], "Tooltip"]& ], 
       TagBox[
        TooltipBox[
         TagBox[
          DynamicBox[{
            FEPrivate`If[
             CurrentValue["MouseOver"], 
             EdgeForm[{
               GrayLevel[0.5], 
               AbsoluteThickness[1.5], 
               Opacity[0.66]}], {}, {}], 
            
            RectangleBox[{-24., 0.}, {-22., 0.0018747656542932134`}, 
             "RoundingRadius" -> 0]},
           
           ImageSizeCache->{{47.166666666666664`, 51.976190476190474`}, {
            37.68540162944754, 42.024758424985244`}}],
          StatusArea[#, 0.0018747656542932134`]& ,
          TagBoxNote->"0.0018747656542932134"],
         StyleBox["0.0018747656542932134`", {
           GrayLevel[0]}, StripOnInput -> False]],
        Annotation[#, 
         Style[0.0018747656542932134`, {
           GrayLevel[0]}], "Tooltip"]& ], 
       TagBox[
        TooltipBox[
         TagBox[
          DynamicBox[{
            FEPrivate`If[
             CurrentValue["MouseOver"], 
             EdgeForm[{
               GrayLevel[0.5], 
               AbsoluteThickness[1.5], 
               Opacity[0.66]}], {}, {}], 
            
            RectangleBox[{-22., 0.}, {-20., 0.005061867266591676}, 
             "RoundingRadius" -> 0]},
           
           ImageSizeCache->{{50.976190476190474`, 55.785714285714285`}, {
            32.008495077033444`, 42.024758424985244`}}],
          StatusArea[#, 0.005061867266591676]& ,
          TagBoxNote->"0.005061867266591676"],
         StyleBox["0.005061867266591676`", {
           GrayLevel[0]}, StripOnInput -> False]],
        Annotation[#, 
         Style[0.005061867266591676, {
           GrayLevel[0]}], "Tooltip"]& ], 
       TagBox[
        TooltipBox[
         TagBox[
          DynamicBox[{
            FEPrivate`If[
             CurrentValue["MouseOver"], 
             EdgeForm[{
               GrayLevel[0.5], 
               AbsoluteThickness[1.5], 
               Opacity[0.66]}], {}, {}], 
            
            RectangleBox[{-20., 0.}, {-18., 0.010436195475565554`}, 
             "RoundingRadius" -> 0]},
           
           ImageSizeCache->{{54.785714285714285`, 59.595238095238095`}, {
            22.435672263158693`, 42.024758424985244`}}],
          StatusArea[#, 0.010436195475565554`]& ,
          TagBoxNote->"0.010436195475565554"],
         StyleBox["0.010436195475565554`", {
           GrayLevel[0]}, StripOnInput -> False]],
        Annotation[#, 
         Style[0.010436195475565554`, {
           GrayLevel[0]}], "Tooltip"]& ], 
       TagBox[
        TooltipBox[
         TagBox[
          DynamicBox[{
            FEPrivate`If[
             CurrentValue["MouseOver"], 
             EdgeForm[{
               GrayLevel[0.5], 
               AbsoluteThickness[1.5], 
               Opacity[0.66]}], {}, {}], 
            
            RectangleBox[{-18., 0.}, {-16., 0.010748656417947757`}, 
             "RoundingRadius" -> 0]},
           
           ImageSizeCache->{{58.595238095238095`, 63.4047619047619}, {
            21.87911279723574, 42.024758424985244`}}],
          StatusArea[#, 0.010748656417947757`]& ,
          TagBoxNote->"0.010748656417947757"],
         StyleBox["0.010748656417947757`", {
           GrayLevel[0]}, StripOnInput -> False]],
        Annotation[#, 
         Style[0.010748656417947757`, {
           GrayLevel[0]}], "Tooltip"]& ], 
       TagBox[
        TooltipBox[
         TagBox[
          DynamicBox[{
            FEPrivate`If[
             CurrentValue["MouseOver"], 
             EdgeForm[{
               GrayLevel[0.5], 
               AbsoluteThickness[1.5], 
               Opacity[0.66]}], {}, {}], 
            
            RectangleBox[{-16., 0.}, {-14., 0.013060867391576053`}, 
             "RoundingRadius" -> 0]},
           
           ImageSizeCache->{{62.4047619047619, 67.21428571428571}, {
            17.76057274940591, 42.024758424985244`}}],
          StatusArea[#, 0.013060867391576053`]& ,
          TagBoxNote->"0.013060867391576053"],
         StyleBox["0.013060867391576053`", {
           GrayLevel[0]}, StripOnInput -> False]],
        Annotation[#, 
         Style[0.013060867391576053`, {
           GrayLevel[0]}], "Tooltip"]& ], 
       TagBox[
        TooltipBox[
         TagBox[
          DynamicBox[{
            FEPrivate`If[
             CurrentValue["MouseOver"], 
             EdgeForm[{
               GrayLevel[0.5], 
               AbsoluteThickness[1.5], 
               Opacity[0.66]}], {}, {}], 
            
            RectangleBox[{-14., 0.}, {-12., 0.01843519560054993}, 
             "RoundingRadius" -> 0]},
           
           ImageSizeCache->{{66.21428571428571, 71.02380952380952}, {
            8.187749935531158, 42.024758424985244`}}],
          StatusArea[#, 0.01843519560054993]& ,
          TagBoxNote->"0.01843519560054993"],
         StyleBox["0.01843519560054993`", {
           GrayLevel[0]}, StripOnInput -> False]],
        Annotation[#, 
         Style[0.01843519560054993, {
           GrayLevel[0]}], "Tooltip"]& ], 
       TagBox[
        TooltipBox[
         TagBox[
          DynamicBox[{
            FEPrivate`If[
             CurrentValue["MouseOver"], 
             EdgeForm[{
               GrayLevel[0.5], 
               AbsoluteThickness[1.5], 
               Opacity[0.66]}], {}, {}], 
            RectangleBox[{-12., 0.}, {-10., 0.020559930008748905`}, 
             "RoundingRadius" -> 0]},
           
           ImageSizeCache->{{70.02380952380952, 74.83333333333333}, {
            4.403145567255095, 42.024758424985244`}}],
          StatusArea[#, 0.020559930008748905`]& ,
          TagBoxNote->"0.020559930008748905"],
         StyleBox["0.020559930008748905`", {
           GrayLevel[0]}, StripOnInput -> False]],
        Annotation[#, 
         Style[0.020559930008748905`, {
           GrayLevel[0]}], "Tooltip"]& ], 
       TagBox[
        TooltipBox[
         TagBox[
          DynamicBox[{
            FEPrivate`If[
             CurrentValue["MouseOver"], 
             EdgeForm[{
               GrayLevel[0.5], 
               AbsoluteThickness[1.5], 
               Opacity[0.66]}], {}, {}], 
            
            RectangleBox[{-10., 0.}, {-8., 0.021559805024371954`}, 
             "RoundingRadius" -> 0]},
           
           ImageSizeCache->{{73.83333333333333, 78.64285714285714}, {
            2.6221552763016476`, 42.024758424985244`}}],
          StatusArea[#, 0.021559805024371954`]& ,
          TagBoxNote->"0.021559805024371954"],
         StyleBox["0.021559805024371954`", {
           GrayLevel[0]}, StripOnInput -> False]],
        Annotation[#, 
         Style[0.021559805024371954`, {
           GrayLevel[0]}], "Tooltip"]& ], 
       TagBox[
        TooltipBox[
         TagBox[
          DynamicBox[{
            FEPrivate`If[
             CurrentValue["MouseOver"], 
             EdgeForm[{
               GrayLevel[0.5], 
               AbsoluteThickness[1.5], 
               Opacity[0.66]}], {}, {}], 
            
            RectangleBox[{-8., 0.}, {-6., 0.03049618797650294}, 
             "RoundingRadius" -> 0]},
           
           ImageSizeCache->{{77.64285714285714, 
            82.45238095238095}, {-13.295445449094736`, 42.024758424985244`}}],
          StatusArea[#, 0.03049618797650294]& ,
          TagBoxNote->"0.03049618797650294"],
         StyleBox["0.03049618797650294`", {
           GrayLevel[0]}, StripOnInput -> False]],
        Annotation[#, 
         Style[0.03049618797650294, {
           GrayLevel[0]}], "Tooltip"]& ], 
       TagBox[
        TooltipBox[
         TagBox[
          DynamicBox[{
            FEPrivate`If[
             CurrentValue["MouseOver"], 
             EdgeForm[{
               GrayLevel[0.5], 
               AbsoluteThickness[1.5], 
               Opacity[0.66]}], {}, {}], 
            
            RectangleBox[{-6., 0.}, {-4., 0.033183352080989874`}, 
             "RoundingRadius" -> 0]},
           
           ImageSizeCache->{{81.45238095238095, 
            86.26190476190476}, {-18.081856856032104`, 42.024758424985244`}}],
          StatusArea[#, 0.033183352080989874`]& ,
          TagBoxNote->"0.033183352080989874"],
         StyleBox["0.033183352080989874`", {
           GrayLevel[0]}, StripOnInput -> False]],
        Annotation[#, 
         Style[0.033183352080989874`, {
           GrayLevel[0]}], "Tooltip"]& ], 
       TagBox[
        TooltipBox[
         TagBox[
          DynamicBox[{
            FEPrivate`If[
             CurrentValue["MouseOver"], 
             EdgeForm[{
               GrayLevel[0.5], 
               AbsoluteThickness[1.5], 
               Opacity[0.66]}], {}, {}], 
            
            RectangleBox[{-4., 0.}, {-2., 0.0418697662792151}, 
             "RoundingRadius" -> 0]},
           
           ImageSizeCache->{{85.26190476190476, 
            90.07142857142857}, {-33.55421000869014, 42.024758424985244`}}],
          StatusArea[#, 0.0418697662792151]& ,
          TagBoxNote->"0.0418697662792151"],
         StyleBox["0.0418697662792151`", {
           GrayLevel[0]}, StripOnInput -> False]],
        Annotation[#, 
         Style[0.0418697662792151, {
           GrayLevel[0]}], "Tooltip"]& ], 
       TagBox[
        TooltipBox[
         TagBox[
          DynamicBox[{
            FEPrivate`If[
             CurrentValue["MouseOver"], 
             EdgeForm[{
               GrayLevel[0.5], 
               AbsoluteThickness[1.5], 
               Opacity[0.66]}], {}, {}], 
            
            RectangleBox[{-2., 0.}, {0., 0.04361954755655543}, 
             "RoundingRadius" -> 0]},
           
           ImageSizeCache->{{89.07142857142857, 
            93.88095238095238}, {-36.670943017858654`, 42.024758424985244`}}],
          StatusArea[#, 0.04361954755655543]& ,
          TagBoxNote->"0.04361954755655543"],
         StyleBox["0.04361954755655543`", {
           GrayLevel[0]}, StripOnInput -> False]],
        Annotation[#, 
         Style[0.04361954755655543, {
           GrayLevel[0]}], "Tooltip"]& ], 
       TagBox[
        TooltipBox[
         TagBox[
          DynamicBox[{
            FEPrivate`If[
             CurrentValue["MouseOver"], 
             EdgeForm[{
               GrayLevel[0.5], 
               AbsoluteThickness[1.5], 
               Opacity[0.66]}], {}, {}], 
            
            RectangleBox[{0., 0.}, {2., 0.04011998500187477}, 
             "RoundingRadius" -> 0]},
           
           ImageSizeCache->{{92.88095238095238, 
            97.69047619047619}, {-30.437476999521614`, 42.024758424985244`}}],
          StatusArea[#, 0.04011998500187477]& ,
          TagBoxNote->"0.04011998500187477"],
         StyleBox["0.04011998500187477`", {
           GrayLevel[0]}, StripOnInput -> False]],
        Annotation[#, 
         Style[0.04011998500187477, {
           GrayLevel[0]}], "Tooltip"]& ], 
       TagBox[
        TooltipBox[
         TagBox[
          DynamicBox[{
            FEPrivate`If[
             CurrentValue["MouseOver"], 
             EdgeForm[{
               GrayLevel[0.5], 
               AbsoluteThickness[1.5], 
               Opacity[0.66]}], {}, {}], 
            RectangleBox[{2., 0.}, {4., 0.03880764904386952}, 
             "RoundingRadius" -> 0]},
           
           ImageSizeCache->{{96.69047619047619, 101.5}, {-28.09992724264523, 
            42.024758424985244`}}],
          StatusArea[#, 0.03880764904386952]& ,
          TagBoxNote->"0.03880764904386952"],
         StyleBox["0.03880764904386952`", {
           GrayLevel[0]}, StripOnInput -> False]],
        Annotation[#, 
         Style[0.03880764904386952, {
           GrayLevel[0]}], "Tooltip"]& ], 
       TagBox[
        TooltipBox[
         TagBox[
          DynamicBox[{
            FEPrivate`If[
             CurrentValue["MouseOver"], 
             EdgeForm[{
               GrayLevel[0.5], 
               AbsoluteThickness[1.5], 
               Opacity[0.66]}], {}, {}], 
            
            RectangleBox[{4., 0.}, {6., 0.03362079740032496}, 
             "RoundingRadius" -> 0]},
           
           ImageSizeCache->{{100.5, 
            105.30952380952381`}, {-18.861040108324246`, 
            42.024758424985244`}}],
          StatusArea[#, 0.03362079740032496]& ,
          TagBoxNote->"0.03362079740032496"],
         StyleBox["0.03362079740032496`", {
           GrayLevel[0]}, StripOnInput -> False]],
        Annotation[#, 
         Style[0.03362079740032496, {
           GrayLevel[0]}], "Tooltip"]& ], 
       TagBox[
        TooltipBox[
         TagBox[
          DynamicBox[{
            FEPrivate`If[
             CurrentValue["MouseOver"], 
             EdgeForm[{
               GrayLevel[0.5], 
               AbsoluteThickness[1.5], 
               Opacity[0.66]}], {}, {}], 
            
            RectangleBox[{6., 0.}, {8., 0.02943382077240345}, 
             "RoundingRadius" -> 0]},
           
           ImageSizeCache->{{104.30952380952381`, 
            109.11904761904762`}, {-11.403143264956704`, 
            42.024758424985244`}}],
          StatusArea[#, 0.02943382077240345]& ,
          TagBoxNote->"0.02943382077240345"],
         StyleBox["0.02943382077240345`", {
           GrayLevel[0]}, StripOnInput -> False]],
        Annotation[#, 
         Style[0.02943382077240345, {
           GrayLevel[0]}], "Tooltip"]& ], 
       TagBox[
        TooltipBox[
         TagBox[
          DynamicBox[{
            FEPrivate`If[
             CurrentValue["MouseOver"], 
             EdgeForm[{
               GrayLevel[0.5], 
               AbsoluteThickness[1.5], 
               Opacity[0.66]}], {}, {}], 
            
            RectangleBox[{8., 0.}, {10., 0.026059242594675666`}, 
             "RoundingRadius" -> 0]},
           
           ImageSizeCache->{{108.11904761904762`, 
            112.92857142857143`}, {-5.39230103298884, 42.024758424985244`}}],
          StatusArea[#, 0.026059242594675666`]& ,
          TagBoxNote->"0.026059242594675666"],
         StyleBox["0.026059242594675666`", {
           GrayLevel[0]}, StripOnInput -> False]],
        Annotation[#, 
         Style[0.026059242594675666`, {
           GrayLevel[0]}], "Tooltip"]& ], 
       TagBox[
        TooltipBox[
         TagBox[
          DynamicBox[{
            FEPrivate`If[
             CurrentValue["MouseOver"], 
             EdgeForm[{
               GrayLevel[0.5], 
               AbsoluteThickness[1.5], 
               Opacity[0.66]}], {}, {}], 
            
            RectangleBox[{10., 0.}, {12., 0.02074740657417823}, 
             "RoundingRadius" -> 0]},
           
           ImageSizeCache->{{111.92857142857143`, 116.73809523809524`}, {
            4.069209887701319, 42.024758424985244`}}],
          StatusArea[#, 0.02074740657417823]& ,
          TagBoxNote->"0.02074740657417823"],
         StyleBox["0.02074740657417823`", {
           GrayLevel[0]}, StripOnInput -> False]],
        Annotation[#, 
         Style[0.02074740657417823, {
           GrayLevel[0]}], "Tooltip"]& ], 
       TagBox[
        TooltipBox[
         TagBox[
          DynamicBox[{
            FEPrivate`If[
             CurrentValue["MouseOver"], 
             EdgeForm[{
               GrayLevel[0.5], 
               AbsoluteThickness[1.5], 
               Opacity[0.66]}], {}, {}], 
            
            RectangleBox[{12., 0.}, {14., 0.0196225471816023}, 
             "RoundingRadius" -> 0]},
           
           ImageSizeCache->{{115.73809523809524`, 120.54761904761905`}, {
            6.072823965023943, 42.024758424985244`}}],
          StatusArea[#, 0.0196225471816023]& ,
          TagBoxNote->"0.0196225471816023"],
         StyleBox["0.0196225471816023`", {
           GrayLevel[0]}, StripOnInput -> False]],
        Annotation[#, 
         Style[0.0196225471816023, {
           GrayLevel[0]}], "Tooltip"]& ], 
       TagBox[
        TooltipBox[
         TagBox[
          DynamicBox[{
            FEPrivate`If[
             CurrentValue["MouseOver"], 
             EdgeForm[{
               GrayLevel[0.5], 
               AbsoluteThickness[1.5], 
               Opacity[0.66]}], {}, {}], 
            
            RectangleBox[{14., 0.}, {16., 0.013373328333958255`}, 
             "RoundingRadius" -> 0]},
           
           ImageSizeCache->{{119.54761904761905`, 124.35714285714286`}, {
            17.204013283482958`, 42.024758424985244`}}],
          StatusArea[#, 0.013373328333958255`]& ,
          TagBoxNote->"0.013373328333958255"],
         StyleBox["0.013373328333958255`", {
           GrayLevel[0]}, StripOnInput -> False]],
        Annotation[#, 
         Style[0.013373328333958255`, {
           GrayLevel[0]}], "Tooltip"]& ], 
       TagBox[
        TooltipBox[
         TagBox[
          DynamicBox[{
            FEPrivate`If[
             CurrentValue["MouseOver"], 
             EdgeForm[{
               GrayLevel[0.5], 
               AbsoluteThickness[1.5], 
               Opacity[0.66]}], {}, {}], 
            
            RectangleBox[{16., 0.}, {18., 0.008811398575178103}, 
             "RoundingRadius" -> 0]},
           
           ImageSizeCache->{{123.35714285714286`, 128.16666666666666`}, {
            25.329781485958037`, 42.024758424985244`}}],
          StatusArea[#, 0.008811398575178103]& ,
          TagBoxNote->"0.008811398575178103"],
         StyleBox["0.008811398575178103`", {
           GrayLevel[0]}, StripOnInput -> False]],
        Annotation[#, 
         Style[0.008811398575178103, {
           GrayLevel[0]}], "Tooltip"]& ], 
       TagBox[
        TooltipBox[
         TagBox[
          DynamicBox[{
            FEPrivate`If[
             CurrentValue["MouseOver"], 
             EdgeForm[{
               GrayLevel[0.5], 
               AbsoluteThickness[1.5], 
               Opacity[0.66]}], {}, {}], 
            
            RectangleBox[{18., 0.}, {20., 0.005936757905261842}, 
             "RoundingRadius" -> 0]},
           
           ImageSizeCache->{{127.16666666666666`, 131.97619047619048`}, {
            30.45012857244918, 42.024758424985244`}}],
          StatusArea[#, 0.005936757905261842]& ,
          TagBoxNote->"0.005936757905261842"],
         StyleBox["0.005936757905261842`", {
           GrayLevel[0]}, StripOnInput -> False]],
        Annotation[#, 
         Style[0.005936757905261842, {
           GrayLevel[0]}], "Tooltip"]& ], 
       TagBox[
        TooltipBox[
         TagBox[
          DynamicBox[{
            FEPrivate`If[
             CurrentValue["MouseOver"], 
             EdgeForm[{
               GrayLevel[0.5], 
               AbsoluteThickness[1.5], 
               Opacity[0.66]}], {}, {}], 
            
            RectangleBox[{20., 0.}, {22., 0.0024371953505811774`}, 
             "RoundingRadius" -> 0]},
           
           ImageSizeCache->{{130.97619047619048`, 135.78571428571428`}, {
            36.68359459078623, 42.024758424985244`}}],
          StatusArea[#, 0.0024371953505811774`]& ,
          TagBoxNote->"0.0024371953505811774"],
         StyleBox["0.0024371953505811774`", {
           GrayLevel[0]}, StripOnInput -> False]],
        Annotation[#, 
         Style[0.0024371953505811774`, {
           GrayLevel[0]}], "Tooltip"]& ], 
       TagBox[
        TooltipBox[
         TagBox[
          DynamicBox[{
            FEPrivate`If[
             CurrentValue["MouseOver"], 
             EdgeForm[{
               GrayLevel[0.5], 
               AbsoluteThickness[1.5], 
               Opacity[0.66]}], {}, {}], 
            
            RectangleBox[{22., 0.}, {24., 0.0024371953505811774`}, 
             "RoundingRadius" -> 0]},
           
           ImageSizeCache->{{134.78571428571428`, 139.5952380952381}, {
            36.68359459078623, 42.024758424985244`}}],
          StatusArea[#, 0.0024371953505811774`]& ,
          TagBoxNote->"0.0024371953505811774"],
         StyleBox["0.0024371953505811774`", {
           GrayLevel[0]}, StripOnInput -> False]],
        Annotation[#, 
         Style[0.0024371953505811774`, {
           GrayLevel[0]}], "Tooltip"]& ], 
       TagBox[
        TooltipBox[
         TagBox[
          DynamicBox[{
            FEPrivate`If[
             CurrentValue["MouseOver"], 
             EdgeForm[{
               GrayLevel[0.5], 
               AbsoluteThickness[1.5], 
               Opacity[0.66]}], {}, {}], 
            
            RectangleBox[{24., 0.}, {26., 0.0016247969003874516`}, 
             "RoundingRadius" -> 0]},
           
           ImageSizeCache->{{138.5952380952381, 143.4047619047619}, {
            38.1306492021859, 42.024758424985244`}}],
          StatusArea[#, 0.0016247969003874516`]& ,
          TagBoxNote->"0.0016247969003874516"],
         StyleBox["0.0016247969003874516`", {
           GrayLevel[0]}, StripOnInput -> False]],
        Annotation[#, 
         Style[0.0016247969003874516`, {
           GrayLevel[0]}], "Tooltip"]& ], 
       TagBox[
        TooltipBox[
         TagBox[
          DynamicBox[{
            FEPrivate`If[
             CurrentValue["MouseOver"], 
             EdgeForm[{
               GrayLevel[0.5], 
               AbsoluteThickness[1.5], 
               Opacity[0.66]}], {}, {}], 
            
            RectangleBox[{26., 0.}, {28., 0.0009373828271466067}, 
             "RoundingRadius" -> 0]},
           
           ImageSizeCache->{{142.4047619047619, 147.21428571428572`}, {
            39.35508002721639, 42.024758424985244`}}],
          StatusArea[#, 0.0009373828271466067]& ,
          TagBoxNote->"0.0009373828271466067"],
         StyleBox["0.0009373828271466067`", {
           GrayLevel[0]}, StripOnInput -> False]],
        Annotation[#, 
         Style[0.0009373828271466067, {
           GrayLevel[0]}], "Tooltip"]& ], 
       TagBox[
        TooltipBox[
         TagBox[
          DynamicBox[{
            FEPrivate`If[
             CurrentValue["MouseOver"], 
             EdgeForm[{
               GrayLevel[0.5], 
               AbsoluteThickness[1.5], 
               Opacity[0.66]}], {}, {}], 
            
            RectangleBox[{28., 0.}, {30., 0.0004374453193350831}, 
             "RoundingRadius" -> 0]},
           
           ImageSizeCache->{{146.21428571428572`, 151.02380952380952`}, {
            40.245575172693115`, 42.024758424985244`}}],
          StatusArea[#, 0.0004374453193350831]& ,
          TagBoxNote->"0.0004374453193350831"],
         StyleBox["0.0004374453193350831`", {
           GrayLevel[0]}, StripOnInput -> False]],
        Annotation[#, 
         Style[0.0004374453193350831, {
           GrayLevel[0]}], "Tooltip"]& ], 
       TagBox[
        TooltipBox[
         TagBox[
          DynamicBox[{
            FEPrivate`If[
             CurrentValue["MouseOver"], 
             EdgeForm[{
               GrayLevel[0.5], 
               AbsoluteThickness[1.5], 
               Opacity[0.66]}], {}, {}], 
            RectangleBox[{30., 0.}, {32., 0.0004374453193350831}, 
             "RoundingRadius" -> 0]},
           
           ImageSizeCache->{{150.02380952380952`, 154.83333333333334`}, {
            40.245575172693115`, 42.024758424985244`}}],
          StatusArea[#, 0.0004374453193350831]& ,
          TagBoxNote->"0.0004374453193350831"],
         StyleBox["0.0004374453193350831`", {
           GrayLevel[0]}, StripOnInput -> False]],
        Annotation[#, 
         Style[0.0004374453193350831, {
           GrayLevel[0]}], "Tooltip"]& ], 
       TagBox[
        TooltipBox[
         TagBox[
          DynamicBox[{
            FEPrivate`If[
             CurrentValue["MouseOver"], 
             EdgeForm[{
               GrayLevel[0.5], 
               AbsoluteThickness[1.5], 
               Opacity[0.66]}], {}, {}], 
            
            RectangleBox[{32., 0.}, {34., 0.00006249218847644044}, 
             "RoundingRadius" -> 0]},
           
           ImageSizeCache->{{153.83333333333334`, 158.64285714285714`}, {
            40.91344653180065, 42.024758424985244`}}],
          StatusArea[#, 0.00006249218847644044]& ,
          TagBoxNote->"0.00006249218847644044"],
         StyleBox["0.00006249218847644044`", {
           GrayLevel[0]}, StripOnInput -> False]],
        Annotation[#, 
         Style[0.00006249218847644044, {
           GrayLevel[0]}], "Tooltip"]& ]}, {}, {}}, {{}, {}, 
      {RGBColor[1, 0, 0], Thickness[Large], LineBox[CompressedData["
1:eJwV1nc81d8bAHBbWVmZKSMzZG+ecy+FbElk7wqRUAkZyUxCpGTdGyohlaK4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        "]]}}},
    AspectRatio->NCache[GoldenRatio^(-1), 0.6180339887498948],
    Axes->{True, True},
    AxesLabel->{
      FormBox["\"x2\"", TraditionalForm], 
      FormBox["\"freq\"", TraditionalForm]},
    AxesOrigin->{-38., 0},
    FrameTicks->{{Automatic, Automatic}, {Automatic, Automatic}},
    GridLines->{None, None},
    PlotRange->{{-38, 34}, {All, All}},
    PlotRangePadding->{{
       Scaled[0.02], 
       Scaled[0.02]}, {
       Scaled[0.02], 
       Scaled[0.1]}},
    Ticks->{Automatic, Automatic}]}], "}"}]], "Output",
 ImageSize->{389, 126},
 ImageMargins->{{0, 0}, {0, 0}},
 ImageRegion->{{0, 1}, {0, 1}},
 CellLabel->"Out[8]=",
 CellID->215573117]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 InterpretationBox[Cell[TextData[{
   "Neat Examples",
   "\[NonBreakingSpace]\[NonBreakingSpace]",
   Cell["(1)", "ExampleCount"]
  }], "ExampleSection"],
  $Line = 0; Null]], "ExampleSection",
 CellID->658589838],

Cell[BoxData[
 RowBox[{"Needs", "[", "\"\<Metropolis`\>\"", "]"}]], "Input",
 CellLabel->"In[1]:=",
 CellID->290007961],

Cell["\<\
This samples a Possion distribution, a truncated normal distribution, a \
bivariate normal distribution\
\>", "ExampleText",
 CellID->73418448],

Cell[CellGroupData[{

Cell[BoxData[{
 RowBox[{
  RowBox[{"Clear", "[", 
   RowBox[{
   "lambda", ",", "V", ",", "iV", ",", "funs", ",", "logpdf", ",", "scale", 
    ",", "cov", ",", "n", ",", "res"}], "]"}], ";"}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{"lambda", "=", "2"}], ";"}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{"V", "=", 
   RowBox[{"{", 
    RowBox[{
     RowBox[{"{", 
      RowBox[{"1", ",", "0.9"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"0.9", ",", "1"}], "}"}]}], "}"}]}], 
  ";"}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{"iV", "=", 
   RowBox[{"Inverse", "[", "V", "]"}]}], ";"}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{"funs", "=", 
   RowBox[{"{", 
    RowBox[{
     RowBox[{"Function", "[", 
      RowBox[{"x", ",", 
       RowBox[{"Log", "[", 
        FractionBox[
         RowBox[{
          SuperscriptBox["lambda", "x"], " ", 
          SuperscriptBox["\[ExponentialE]", 
           RowBox[{"-", "lambda"}]]}], 
         RowBox[{"x", " ", "!"}]], "]"}]}], "]"}], ",", 
     RowBox[{"Function", "[", 
      RowBox[{"x", ",", 
       RowBox[{
        RowBox[{"-", 
         RowBox[{"x", "^", "2"}]}], "/", "2"}]}], "]"}], ",", 
     RowBox[{"Function", "[", 
      RowBox[{"x", ",", 
       RowBox[{
        RowBox[{"-", 
         RowBox[{"x", ".", "iV", ".", "x"}]}], "/", "2"}]}], "]"}]}], "}"}]}],
   ";"}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{"logpdf", "=", 
   RowBox[{"Function", "[", 
    RowBox[{
     RowBox[{"{", 
      RowBox[{"x", ",", "indices"}], "}"}], ",", 
     RowBox[{
      RowBox[{
       RowBox[{
        RowBox[{"#", "[", 
         RowBox[{"[", "1", "]"}], "]"}], "[", 
        RowBox[{"#", "[", 
         RowBox[{"[", "2", "]"}], "]"}], "]"}], "&"}], "/@", 
      RowBox[{"Transpose", "[", 
       RowBox[{"{", 
        RowBox[{
         RowBox[{"funs", "[", 
          RowBox[{"[", "indices", "]"}], "]"}], ",", 
         RowBox[{"x", "[", 
          RowBox[{"[", "indices", "]"}], "]"}]}], "}"}], "]"}]}]}], 
    "\[IndentingNewLine]", "]"}]}], ";"}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{"scale", "=", 
   RowBox[{"{", 
    RowBox[{"lambda", ",", "3", ",", "3"}], "}"}]}], 
  ";"}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{"cov", "=", 
   RowBox[{"{", 
    RowBox[{
     RowBox[{"{", 
      RowBox[{"{", "1", "}"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"{", "1", "}"}], "}"}], ",", 
     RowBox[{"Inverse", "[", "iV", "]"}]}], "}"}]}], 
  ";"}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{"n", "=", "5000"}], ";"}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{"adpmax", "=", "2000"}], ";"}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{"x2min", "=", 
   RowBox[{"-", "1"}]}], ";"}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{"x", "=", 
   RowBox[{"{", 
    RowBox[{"0", ",", "0", ",", 
     RowBox[{"{", 
      RowBox[{"0", ",", "0"}], "}"}]}], "}"}]}], ";"}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{"adpset", "=", 
   RowBox[{"InitialProposal", "[", 
    RowBox[{
     RowBox[{"{", 
      RowBox[{"0.44", ",", "0.44", ",", "0.234"}], "}"}], ",", "x"}], "]"}]}],
   ";"}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{"res", "=", 
   RowBox[{"{", "}"}]}], ";"}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{
   RowBox[{"Do", "[", "\[IndentingNewLine]", 
    RowBox[{
     RowBox[{
      RowBox[{"temp", "=", 
       RowBox[{"BlockStudentTMetropolis", "[", 
        RowBox[{"logpdf", ",", "x", ",", "scale", ",", "cov", ",", "2", ",", 
         RowBox[{"IsReturnLast", "\[Rule]", "False"}], ",", 
         RowBox[{"IsDiscrete", "\[Rule]", 
          RowBox[{"{", 
           RowBox[{"True", ",", "False", ",", "False"}], "}"}]}], ",", 
         RowBox[{"DomainConstraint", "\[Rule]", 
          RowBox[{"(", 
           RowBox[{
            RowBox[{"{", 
             RowBox[{
              RowBox[{
               RowBox[{"#", "[", 
                RowBox[{"[", "1", "]"}], "]"}], "\[GreaterEqual]", "0"}], ",", 
              RowBox[{
               RowBox[{"#", "[", 
                RowBox[{"[", "2", "]"}], "]"}], ">", "x2min"}], ",", "True"}],
              "}"}], "&"}], ")"}]}]}], "]"}]}], ";", "\[IndentingNewLine]", 
      RowBox[{"adpset", "=", 
       RowBox[{"AdaptiveProposal", "[", 
        RowBox[{
         RowBox[{"TrueQ", "[", 
          RowBox[{"it", "\[LessEqual]", "adpmax"}], "]"}], ",", " ", 
         RowBox[{"Mean", "[", 
          RowBox[{"temp", "[", 
           RowBox[{"[", 
            RowBox[{"All", ",", 
             RowBox[{"-", "1"}]}], "]"}], "]"}], "]"}], ",", " ", 
         RowBox[{"5", "/", "adpmax"}], ",", " ", "adpset", ",", 
         RowBox[{"temp", "[", 
          RowBox[{"[", 
           RowBox[{"All", ",", "1"}], "]"}], "]"}]}], "]"}]}], ";", 
      "\[IndentingNewLine]", 
      RowBox[{
       RowBox[{"{", 
        RowBox[{"accept", ",", "scale", ",", " ", "cov"}], "}"}], "=", 
       RowBox[{"adpset", "[", 
        RowBox[{"[", 
         RowBox[{"{", 
          RowBox[{
           RowBox[{"-", "4"}], ",", 
           RowBox[{"-", "3"}], ",", 
           RowBox[{"-", "1"}]}], "}"}], "]"}], "]"}]}], ";", 
      "\[IndentingNewLine]", 
      RowBox[{"x", "=", 
       RowBox[{"temp", "[", 
        RowBox[{"[", 
         RowBox[{
          RowBox[{"-", "1"}], ",", "1"}], "]"}], "]"}]}], ";", 
      "\[IndentingNewLine]", 
      RowBox[{"AppendTo", "[", 
       RowBox[{"res", ",", 
        RowBox[{"{", 
         RowBox[{"x", ",", "accept", ",", "scale"}], "}"}]}], "]"}]}], ",", 
     RowBox[{"{", 
      RowBox[{"it", ",", " ", 
       RowBox[{"2", " ", "n"}]}], "}"}]}], "]"}], ";"}], "//", 
  "Timing"}]}], "Input",
 CellLabel->"In[2]:=",
 CellID->13864406],

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{"18.501000000000005`", ",", "Null"}], "}"}]], "Output",
 ImageSize->{99, 15},
 ImageMargins->{{0, 0}, {0, 0}},
 ImageRegion->{{0, 1}, {0, 1}},
 CellLabel->"Out[2]=",
 CellID->270182846]
}, Open  ]],

Cell["Accept ratito:", "ExampleText",
 CellID->311617669],

Cell[CellGroupData[{

Cell[BoxData[{
 RowBox[{"N", "[", 
  RowBox[{"Mean", "[", 
   RowBox[{"res", "[", 
    RowBox[{"[", 
     RowBox[{
      RowBox[{"adpmax", ";;"}], ",", "2"}], "]"}], "]"}], "]"}], 
  "]"}], "\[IndentingNewLine]", 
 RowBox[{"ListLinePlot", "[", 
  RowBox[{
   RowBox[{"Transpose", "[", 
    RowBox[{"res", "[", 
     RowBox[{"[", 
      RowBox[{"All", ",", "2"}], "]"}], "]"}], "]"}], ",", 
   RowBox[{"PlotRange", "\[Rule]", 
    RowBox[{"{", 
     RowBox[{"Automatic", ",", 
      RowBox[{"{", 
       RowBox[{"0.1", ",", "0.6"}], "}"}]}], "}"}]}], ",", 
   RowBox[{"AxesLabel", "\[Rule]", 
    RowBox[{"{", 
     RowBox[{"\"\<t\>\"", ",", "\"\<accept\>\""}], "}"}]}]}], "]"}]}], "Input",
 CellLabel->"In[3]:=",
 CellID->762797887],

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{
  "0.4316467153267613`", ",", "0.41813273380243526`", ",", 
   "0.21902151001005768`"}], "}"}]], "Output",
 ImageSize->{208, 15},
 ImageMargins->{{0, 0}, {0, 0}},
 ImageRegion->{{0, 1}, {0, 1}},
 CellLabel->"Out[3]=",
 CellID->2367316],

Cell[BoxData[
 GraphicsBox[
  TagBox[RasterBox[CompressedData["
1:eJztXQtTFFcWtmp/yf4NqjoaiRrXMm5Ukii6CZhMND6jgDwUwQf4QBkGRBQY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    "], {{0, 130}, {180, 0}}, {0, 255},
    ColorFunction->RGBColor],
   BoxForm`ImageTag["Byte", ColorSpace -> "RGB", Interleaving -> True],
   Selectable->False],
  BaseStyle->"ImageGraphics",
  ImageSize->Automatic,
  ImageSizeRaw->{180, 130},
  PlotRange->{{0, 180}, {0, 130}}]], "Output",
 ImageSize->{184, 130},
 ImageMargins->{{0, 0}, {0, 0}},
 ImageRegion->{{0, 1}, {0, 1}},
 CellLabel->"Out[3]=",
 CellID->695337132]
}, Open  ]],

Cell["Proposal scale", "ExampleText",
 CellID->805627902],

Cell[CellGroupData[{

Cell[BoxData[{
 RowBox[{"N", "[", 
  RowBox[{"Mean", "[", 
   RowBox[{"res", "[", 
    RowBox[{"[", 
     RowBox[{
      RowBox[{"adpmax", ";;"}], ",", "3"}], "]"}], "]"}], "]"}], 
  "]"}], "\[IndentingNewLine]", 
 RowBox[{"ListLinePlot", "[", 
  RowBox[{
   RowBox[{"Transpose", "[", 
    RowBox[{"res", "[", 
     RowBox[{"[", 
      RowBox[{"All", ",", "3"}], "]"}], "]"}], "]"}], ",", 
   RowBox[{"PlotRange", "\[Rule]", "All"}], ",", 
   RowBox[{"AxesLabel", "\[Rule]", 
    RowBox[{"{", 
     RowBox[{"\"\<t\>\"", ",", "\"\<scale\>\""}], "}"}]}]}], "]"}]}], "Input",
 CellLabel->"In[4]:=",
 CellID->807846078],

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{
  "1.102460551369479`", ",", "1.7980569650567706`", ",", 
   "2.718595885616843`"}], "}"}]], "Output",
 ImageSize->{180, 15},
 ImageMargins->{{0, 0}, {0, 0}},
 ImageRegion->{{0, 1}, {0, 1}},
 CellLabel->"Out[4]=",
 CellID->135472732],

Cell[BoxData[
 GraphicsBox[
  TagBox[RasterBox[CompressedData["
1:eJztnY1TVNcZxpnpX9J/g5kd0qJNaWptMzV1qrEpThk7KlQHWyTBYFAIirqC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    "], {{0, 131}, {180, 0}}, {0, 255},
    ColorFunction->RGBColor],
   BoxForm`ImageTag["Byte", ColorSpace -> "RGB", Interleaving -> True],
   Selectable->False],
  BaseStyle->"ImageGraphics",
  ImageSize->Automatic,
  ImageSizeRaw->{180, 131},
  PlotRange->{{0, 180}, {0, 131}}]], "Output",
 ImageSize->{184, 131},
 ImageMargins->{{0, 0}, {0, 0}},
 ImageRegion->{{0, 1}, {0, 1}},
 CellLabel->"Out[4]=",
 CellID->142531258]
}, Open  ]],

Cell["Time series", "ExampleText",
 CellID->193398859],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"Table", "[", 
  RowBox[{
   RowBox[{"ListLinePlot", "[", 
    RowBox[{
     RowBox[{"res", "[", 
      RowBox[{"[", 
       RowBox[{
        RowBox[{";;", ";;", "10"}], ",", "1", ",", "i"}], "]"}], "]"}], ",", 
     RowBox[{"PlotRange", "\[Rule]", "All"}], ",", 
     RowBox[{"AxesLabel", "\[Rule]", 
      RowBox[{"{", 
       RowBox[{"\"\<t\>\"", ",", 
        RowBox[{"\"\<x\>\"", "<>", 
         RowBox[{"ToString", "[", "i", "]"}]}]}], "}"}]}]}], "]"}], ",", 
   RowBox[{"{", 
    RowBox[{"i", ",", "2"}], "}"}]}], "]"}]], "Input",
 CellLabel->"In[5]:=",
 CellID->402775542],

Cell[BoxData[
 GraphicsBox[
  TagBox[RasterBox[CompressedData["
1:eJztnW2oHcd5x0/pd9PSFGxKSSucthAJEuoYokQ6JcJtFKVxnDa+TdM0RgFh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    "], {{0, 134}, {385, 0}}, {0, 255},
    ColorFunction->RGBColor],
   BoxForm`ImageTag["Byte", ColorSpace -> "RGB", Interleaving -> True],
   Selectable->False],
  BaseStyle->"ImageGraphics",
  ImageSize->Automatic,
  ImageSizeRaw->{385, 134},
  PlotRange->{{0, 385}, {0, 134}}]], "Output",
 ImageSize->{389, 134},
 ImageMargins->{{0, 0}, {0, 0}},
 ImageRegion->{{0, 1}, {0, 1}},
 CellLabel->"Out[5]=",
 CellID->540490369]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"Table", "[", 
  RowBox[{
   RowBox[{"ListLinePlot", "[", 
    RowBox[{
     RowBox[{"res", "[", 
      RowBox[{"[", 
       RowBox[{
        RowBox[{";;", ";;", "10"}], ",", "1", ",", 
        RowBox[{"-", "1"}], ",", "i"}], "]"}], "]"}], ",", 
     RowBox[{"PlotRange", "\[Rule]", "All"}], ",", 
     RowBox[{"AxesLabel", "\[Rule]", 
      RowBox[{"{", 
       RowBox[{"\"\<t\>\"", ",", 
        RowBox[{"\"\<x3\>\"", "<>", 
         RowBox[{"ToString", "[", "i", "]"}]}]}], "}"}]}]}], "]"}], ",", 
   RowBox[{"{", 
    RowBox[{"i", ",", "2"}], "}"}]}], "]"}]], "Input",
 CellLabel->"In[6]:=",
 CellID->223197944],

Cell[BoxData[
 GraphicsBox[
  TagBox[RasterBox[CompressedData["
1:eJztnW2oZVd5x2/p99ASSyNU1CGxpWYkgUmwUzNbHJM6TozV4NwPxpomcmvE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    "], {{0, 118}, {385, 0}}, {0, 255},
    ColorFunction->RGBColor],
   BoxForm`ImageTag["Byte", ColorSpace -> "RGB", Interleaving -> True],
   Selectable->False],
  BaseStyle->"ImageGraphics",
  ImageSize->Automatic,
  ImageSizeRaw->{385, 118},
  PlotRange->{{0, 385}, {0, 118}}]], "Output",
 ImageSize->{389, 118},
 ImageMargins->{{0, 0}, {0, 0}},
 ImageRegion->{{0, 1}, {0, 1}},
 CellLabel->"Out[6]=",
 CellID->235117103]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{
   RowBox[{"ListPlot", "[", 
    RowBox[{
     RowBox[{"res", "[", 
      RowBox[{"[", 
       RowBox[{
        RowBox[{";;", ";;", "10"}], ",", "1", ",", 
        RowBox[{"-", "1"}]}], "]"}], "]"}], ",", 
     RowBox[{"PlotRange", "\[Rule]", "All"}], ",", 
     RowBox[{"AxesLabel", "\[Rule]", 
      RowBox[{"{", 
       RowBox[{"\"\<x31\>\"", ",", "\"\<x32\>\""}], "}"}]}]}], "]"}], ",", 
   RowBox[{
    RowBox[{"Covariance", "[", 
     RowBox[{"res", "[", 
      RowBox[{"[", 
       RowBox[{
        RowBox[{";;", ";;", "10"}], ",", "1", ",", 
        RowBox[{"-", "1"}]}], "]"}], "]"}], "]"}], "//", "MatrixForm"}]}], 
  "}"}]], "Input",
 CellLabel->"In[7]:=",
 CellID->36445872],

Cell[BoxData[
 GraphicsBox[
  TagBox[RasterBox[CompressedData["
1:eJztnc+PHFcRxwdxjzhw4AriwCG5cOBgyfNjd72O147t2PEviLEdwSQIkUSQ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    "], {{0, 117}, {345, 0}}, {0, 255},
    ColorFunction->RGBColor],
   BoxForm`ImageTag["Byte", ColorSpace -> "RGB", Interleaving -> True],
   Selectable->False],
  BaseStyle->"ImageGraphics",
  ImageSize->Automatic,
  ImageSizeRaw->{345, 117},
  PlotRange->{{0, 345}, {0, 117}}]], "Output",
 ImageSize->{349, 117},
 ImageMargins->{{0, 0}, {0, 0}},
 ImageRegion->{{0, 1}, {0, 1}},
 CellLabel->"Out[7]=",
 CellID->233124039]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[{
 RowBox[{
  RowBox[{"data", "=", 
   RowBox[{"res", "[", 
    RowBox[{"[", 
     RowBox[{
      RowBox[{"adpmax", ";;"}], ",", "1", ",", "1"}], "]"}], "]"}]}], 
  ";"}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{"xls", "=", 
   RowBox[{"Range", "[", 
    RowBox[{
     RowBox[{"Min", "[", "data", "]"}], ",", 
     RowBox[{"Max", "[", "data", "]"}]}], "]"}]}], 
  ";"}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{"freq", "=", 
   RowBox[{
    RowBox[{
     RowBox[{"Count", "[", 
      RowBox[{"data", ",", "#"}], "]"}], "&"}], "/@", "xls"}]}], 
  ";"}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{"freq", "/=", 
   RowBox[{"Total", "[", "freq", "]"}]}], ";"}], "\[IndentingNewLine]", 
 RowBox[{"Show", "[", 
  RowBox[{
   RowBox[{"Plot", "[", 
    RowBox[{
     RowBox[{"Exp", "[", 
      RowBox[{
       RowBox[{"funs", "[", 
        RowBox[{"[", "1", "]"}], "]"}], "[", "x", "]"}], "]"}], ",", 
     RowBox[{"{", 
      RowBox[{"x", ",", 
       RowBox[{"Min", "[", "xls", "]"}], ",", 
       RowBox[{"Max", "[", "xls", "]"}]}], "}"}], ",", 
     RowBox[{"PlotStyle", "\[Rule]", "Red"}]}], "]"}], ",", 
   RowBox[{"ListPlot", "[", 
    RowBox[{
     RowBox[{"Transpose", "[", 
      RowBox[{"{", 
       RowBox[{"xls", ",", "freq"}], "}"}], "]"}], ",", 
     RowBox[{"Filling", "\[Rule]", "Axis"}], ",", 
     RowBox[{"PlotMarkers", "\[Rule]", "Automatic"}]}], "]"}], ",", 
   RowBox[{"PlotRange", "\[Rule]", "All"}]}], "]"}]}], "Input",
 CellLabel->"In[8]:=",
 CellID->182745918],

Cell[BoxData[
 GraphicsBox[
  TagBox[RasterBox[CompressedData["
1:eJztnb1vHEUYxk/iH0jp0nQuXVA4BssOQuJCZZAAF4EYY1mmSkgVhISLgFwg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    "], {{0, 114}, {180, 0}}, {0, 255},
    ColorFunction->RGBColor],
   BoxForm`ImageTag["Byte", ColorSpace -> "RGB", Interleaving -> True],
   Selectable->False],
  BaseStyle->"ImageGraphics",
  ImageSize->Automatic,
  ImageSizeRaw->{180, 114},
  PlotRange->{{0, 180}, {0, 114}}]], "Output",
 ImageSize->{184, 114},
 ImageMargins->{{0, 0}, {0, 0}},
 ImageRegion->{{0, 1}, {0, 1}},
 CellLabel->"Out[8]=",
 CellID->635841734]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"Show", "[", 
  RowBox[{
   RowBox[{
    RowBox[{"data", "=", 
     RowBox[{"res", "[", 
      RowBox[{"[", 
       RowBox[{
        RowBox[{"adpmax", ";;", ";;", "10"}], ",", "1", ",", "2"}], "]"}], 
      "]"}]}], ";", 
    RowBox[{"Histogram", "[", 
     RowBox[{"data", ",", "Automatic", ",", "\"\<ProbabilityDensity\>\""}], 
     "]"}]}], ",", 
   RowBox[{"Plot", "[", 
    RowBox[{
     RowBox[{
      RowBox[{"PDF", "[", 
       RowBox[{
        RowBox[{"NormalDistribution", "[", 
         RowBox[{"0", ",", "1"}], "]"}], ",", "x"}], "]"}], "/", 
      RowBox[{"(", 
       RowBox[{
        FractionBox["1", "2"], " ", 
        RowBox[{"Erfc", "[", 
         FractionBox["x2min", 
          SqrtBox["2"]], "]"}]}], ")"}]}], ",", 
     RowBox[{"{", 
      RowBox[{"x", ",", 
       RowBox[{"Min", "[", "data", "]"}], ",", 
       RowBox[{"Max", "[", "data", "]"}]}], "}"}], ",", 
     RowBox[{"PlotStyle", "\[Rule]", 
      RowBox[{"{", 
       RowBox[{"Red", ",", "Thick"}], "}"}]}], ",", 
     RowBox[{"PlotRange", "\[Rule]", "All"}], ",", 
     RowBox[{"AxesLabel", "\[Rule]", 
      RowBox[{"{", 
       RowBox[{"\"\<x2\>\"", ",", "\"\<freq\>\""}], "}"}]}]}], "]"}]}], 
  "]"}]], "Input",
 CellLabel->"In[9]:=",
 CellID->530831861],

Cell[BoxData[
 GraphicsBox[{
   {RGBColor[0.798413061722744, 0.824719615472648, 0.968322270542458], 
    EdgeForm[Opacity[0.7]], {}, 
    {RGBColor[0.798413061722744, 0.824719615472648, 0.968322270542458], 
     EdgeForm[Opacity[0.7]], 
     TagBox[
      TooltipBox[
       TagBox[
        DynamicBox[{
          FEPrivate`If[
           CurrentValue["MouseOver"], 
           EdgeForm[{
             GrayLevel[0.5], 
             AbsoluteThickness[1.5], 
             Opacity[0.66]}], {}, {}], 
          
          RectangleBox[{-1., 0.}, {-0.8, 0.2871410736579276}, 
           "RoundingRadius" -> 0]},
         ImageSizeCache->{{15.5, 23.75343228200371}, {-9.647050559684729, 
          40.27773676735785}}],
        StatusArea[#, 0.2871410736579276]& ,
        TagBoxNote->"0.2871410736579276"],
       StyleBox["0.2871410736579276`", {
         GrayLevel[0]}, StripOnInput -> False]],
      Annotation[#, 
       Style[0.2871410736579276, {
         GrayLevel[0]}], "Tooltip"]& ], 
     TagBox[
      TooltipBox[
       TagBox[
        DynamicBox[{
          FEPrivate`If[
           CurrentValue["MouseOver"], 
           EdgeForm[{
             GrayLevel[0.5], 
             AbsoluteThickness[1.5], 
             Opacity[0.66]}], {}, {}], 
          
          RectangleBox[{-0.8, 0.}, {-0.6, 0.3183520599250935}, 
           "RoundingRadius" -> 0]},
         ImageSizeCache->{{22.75343228200371, 
          31.006864564007426`}, {-14.964962225667591`, 40.27773676735785}}],
        StatusArea[#, 0.3183520599250935]& ,
        TagBoxNote->"0.3183520599250935"],
       StyleBox["0.3183520599250935`", {
         GrayLevel[0]}, StripOnInput -> False]],
      Annotation[#, 
       Style[0.3183520599250935, {
         GrayLevel[0]}], "Tooltip"]& ], 
     TagBox[
      TooltipBox[
       TagBox[
        DynamicBox[{
          FEPrivate`If[
           CurrentValue["MouseOver"], 
           EdgeForm[{
             GrayLevel[0.5], 
             AbsoluteThickness[1.5], 
             Opacity[0.66]}], {}, {}], 
          
          RectangleBox[{-0.6, 0.}, {-0.4, 0.34332084893882653`}, 
           "RoundingRadius" -> 0]},
         ImageSizeCache->{{30.006864564007426`, 
          38.26029684601114}, {-19.219291558453932`, 40.27773676735785}}],
        StatusArea[#, 0.34332084893882653`]& ,
        TagBoxNote->"0.34332084893882653"],
       StyleBox["0.34332084893882653`", {
         GrayLevel[0]}, StripOnInput -> False]],
      Annotation[#, 
       Style[0.34332084893882653`, {
         GrayLevel[0]}], "Tooltip"]& ], 
     TagBox[
      TooltipBox[
       TagBox[
        DynamicBox[{
          FEPrivate`If[
           CurrentValue["MouseOver"], 
           EdgeForm[{
             GrayLevel[0.5], 
             AbsoluteThickness[1.5], 
             Opacity[0.66]}], {}, {}], 
          
          RectangleBox[{-0.4, 0.}, {-0.2, 0.5305867665418227}, 
           "RoundingRadius" -> 0]},
         ImageSizeCache->{{37.26029684601114, 
          45.513729128014845`}, {-51.12676155435125, 40.27773676735785}}],
        StatusArea[#, 0.5305867665418227]& ,
        TagBoxNote->"0.5305867665418227"],
       StyleBox["0.5305867665418227`", {
         GrayLevel[0]}, StripOnInput -> False]],
      Annotation[#, 
       Style[0.5305867665418227, {
         GrayLevel[0]}], "Tooltip"]& ], 
     TagBox[
      TooltipBox[
       TagBox[
        DynamicBox[{
          FEPrivate`If[
           CurrentValue["MouseOver"], 
           EdgeForm[{
             GrayLevel[0.5], 
             AbsoluteThickness[1.5], 
             Opacity[0.66]}], {}, {}], 
          
          RectangleBox[{-0.2, 0.}, {0., 0.449438202247191}, "RoundingRadius" -> 
           0]},
         ImageSizeCache->{{44.513729128014845`, 
          52.76716141001856}, {-37.300191222795746`, 40.27773676735785}}],
        StatusArea[#, 0.449438202247191]& ,
        TagBoxNote->"0.449438202247191"],
       StyleBox["0.449438202247191`", {
         GrayLevel[0]}, StripOnInput -> False]],
      Annotation[#, 
       Style[0.449438202247191, {
         GrayLevel[0]}], "Tooltip"]& ], 
     TagBox[
      TooltipBox[
       TagBox[
        DynamicBox[{
          FEPrivate`If[
           CurrentValue["MouseOver"], 
           EdgeForm[{
             GrayLevel[0.5], 
             AbsoluteThickness[1.5], 
             Opacity[0.66]}], {}, {}], 
          
          RectangleBox[{0., 0.}, {0.2, 0.47440699126092384`}, 
           "RoundingRadius" -> 0]},
         ImageSizeCache->{{51.76716141001856, 
          60.020593692022274`}, {-41.55452055558206, 40.27773676735785}}],
        StatusArea[#, 0.47440699126092384`]& ,
        TagBoxNote->"0.47440699126092384"],
       StyleBox["0.47440699126092384`", {
         GrayLevel[0]}, StripOnInput -> False]],
      Annotation[#, 
       Style[0.47440699126092384`, {
         GrayLevel[0]}], "Tooltip"]& ], 
     TagBox[
      TooltipBox[
       TagBox[
        DynamicBox[{
          FEPrivate`If[
           CurrentValue["MouseOver"], 
           EdgeForm[{
             GrayLevel[0.5], 
             AbsoluteThickness[1.5], 
             Opacity[0.66]}], {}, {}], 
          
          RectangleBox[{0.2, 0.}, {0.4, 0.4556803995006242}, "RoundingRadius" -> 
           0]},
         ImageSizeCache->{{59.020593692022274`, 
          67.27402597402599}, {-38.363773555992324`, 40.27773676735785}}],
        StatusArea[#, 0.4556803995006242]& ,
        TagBoxNote->"0.4556803995006242"],
       StyleBox["0.4556803995006242`", {
         GrayLevel[0]}, StripOnInput -> False]],
      Annotation[#, 
       Style[0.4556803995006242, {
         GrayLevel[0]}], "Tooltip"]& ], 
     TagBox[
      TooltipBox[
       TagBox[
        DynamicBox[{
          FEPrivate`If[
           CurrentValue["MouseOver"], 
           EdgeForm[{
             GrayLevel[0.5], 
             AbsoluteThickness[1.5], 
             Opacity[0.66]}], {}, {}], 
          
          RectangleBox[{0.4, 0.}, {0.6, 0.4619225967540575}, "RoundingRadius" -> 
           0]},
         ImageSizeCache->{{66.27402597402599, 
          74.52745825602969}, {-39.42735588918892, 40.27773676735785}}],
        StatusArea[#, 0.4619225967540575]& ,
        TagBoxNote->"0.4619225967540575"],
       StyleBox["0.4619225967540575`", {
         GrayLevel[0]}, StripOnInput -> False]],
      Annotation[#, 
       Style[0.4619225967540575, {
         GrayLevel[0]}], "Tooltip"]& ], 
     TagBox[
      TooltipBox[
       TagBox[
        DynamicBox[{
          FEPrivate`If[
           CurrentValue["MouseOver"], 
           EdgeForm[{
             GrayLevel[0.5], 
             AbsoluteThickness[1.5], 
             Opacity[0.66]}], {}, {}], 
          
          RectangleBox[{0.6, 0.}, {0.8, 0.3620474406991259}, "RoundingRadius" -> 
           0]},
         ImageSizeCache->{{73.52745825602969, 
          81.7808905380334}, {-22.41003855804363, 40.27773676735785}}],
        StatusArea[#, 0.3620474406991259]& ,
        TagBoxNote->"0.3620474406991259"],
       StyleBox["0.3620474406991259`", {
         GrayLevel[0]}, StripOnInput -> False]],
      Annotation[#, 
       Style[0.3620474406991259, {
         GrayLevel[0]}], "Tooltip"]& ], 
     TagBox[
      TooltipBox[
       TagBox[
        DynamicBox[{
          FEPrivate`If[
           CurrentValue["MouseOver"], 
           EdgeForm[{
             GrayLevel[0.5], 
             AbsoluteThickness[1.5], 
             Opacity[0.66]}], {}, {}], 
          
          RectangleBox[{0.8, 0.}, {1., 0.268414481897628}, "RoundingRadius" -> 
           0]},
         ImageSizeCache->{{80.7808905380334, 
          89.03432282003712}, {-6.456303560095002, 40.27773676735785}}],
        StatusArea[#, 0.268414481897628]& ,
        TagBoxNote->"0.268414481897628"],
       StyleBox["0.268414481897628`", {
         GrayLevel[0]}, StripOnInput -> False]],
      Annotation[#, 
       Style[0.268414481897628, {
         GrayLevel[0]}], "Tooltip"]& ], 
     TagBox[
      TooltipBox[
       TagBox[
        DynamicBox[{
          FEPrivate`If[
           CurrentValue["MouseOver"], 
           EdgeForm[{
             GrayLevel[0.5], 
             AbsoluteThickness[1.5], 
             Opacity[0.66]}], {}, {}], 
          
          RectangleBox[{1., 0.}, {1.2, 0.23720349563046195`}, 
           "RoundingRadius" -> 0]},
         ImageSizeCache->{{88.03432282003712, 
          96.28775510204082}, {-1.1383918941121038`, 40.27773676735785}}],
        StatusArea[#, 0.23720349563046195`]& ,
        TagBoxNote->"0.23720349563046195"],
       StyleBox["0.23720349563046195`", {
         GrayLevel[0]}, StripOnInput -> False]],
      Annotation[#, 
       Style[0.23720349563046195`, {
         GrayLevel[0]}], "Tooltip"]& ], 
     TagBox[
      TooltipBox[
       TagBox[
        DynamicBox[{
          FEPrivate`If[
           CurrentValue["MouseOver"], 
           EdgeForm[{
             GrayLevel[0.5], 
             AbsoluteThickness[1.5], 
             Opacity[0.66]}], {}, {}], 
          
          RectangleBox[{1.2, 0.}, {1.4, 0.23096129837702875`}, 
           "RoundingRadius" -> 0]},
         ImageSizeCache->{{95.28775510204082, 
          103.54118738404455`}, {-0.07480956091553281, 40.27773676735785}}],
        StatusArea[#, 0.23096129837702875`]& ,
        TagBoxNote->"0.23096129837702875"],
       StyleBox["0.23096129837702875`", {
         GrayLevel[0]}, StripOnInput -> False]],
      Annotation[#, 
       Style[0.23096129837702875`, {
         GrayLevel[0]}], "Tooltip"]& ], 
     TagBox[
      TooltipBox[
       TagBox[
        DynamicBox[{
          FEPrivate`If[
           CurrentValue["MouseOver"], 
           EdgeForm[{
             GrayLevel[0.5], 
             AbsoluteThickness[1.5], 
             Opacity[0.66]}], {}, {}], 
          
          RectangleBox[{1.4, 0.}, {1.6, 0.1872659176029961}, "RoundingRadius" -> 
           0]},
         ImageSizeCache->{{102.54118738404455`, 110.79461966604825`}, {
          7.37026677146055, 40.27773676735785}}],
        StatusArea[#, 0.1872659176029961]& ,
        TagBoxNote->"0.1872659176029961"],
       StyleBox["0.1872659176029961`", {
         GrayLevel[0]}, StripOnInput -> False]],
      Annotation[#, 
       Style[0.1872659176029961, {
         GrayLevel[0]}], "Tooltip"]& ], 
     TagBox[
      TooltipBox[
       TagBox[
        DynamicBox[{
          FEPrivate`If[
           CurrentValue["MouseOver"], 
           EdgeForm[{
             GrayLevel[0.5], 
             AbsoluteThickness[1.5], 
             Opacity[0.66]}], {}, {}], 
          
          RectangleBox[{1.6, 0.}, {1.8, 0.12484394506866418`}, 
           "RoundingRadius" -> 0]},
         ImageSizeCache->{{109.79461966604825`, 118.04805194805196`}, {
          18.006090103426295`, 40.27773676735785}}],
        StatusArea[#, 0.12484394506866418`]& ,
        TagBoxNote->"0.12484394506866418"],
       StyleBox["0.12484394506866418`", {
         GrayLevel[0]}, StripOnInput -> False]],
      Annotation[#, 
       Style[0.12484394506866418`, {
         GrayLevel[0]}], "Tooltip"]& ], 
     TagBox[
      TooltipBox[
       TagBox[
        DynamicBox[{
          FEPrivate`If[
           CurrentValue["MouseOver"], 
           EdgeForm[{
             GrayLevel[0.5], 
             AbsoluteThickness[1.5], 
             Opacity[0.66]}], {}, {}], 
          
          RectangleBox[{1.8, 0.}, {2., 0.05617977528089888}, "RoundingRadius" -> 
           0]},
         ImageSizeCache->{{117.04805194805196`, 125.30148423005568`}, {
          29.70549576858865, 40.27773676735785}}],
        StatusArea[#, 0.05617977528089888]& ,
        TagBoxNote->"0.05617977528089888"],
       StyleBox["0.05617977528089888`", {
         GrayLevel[0]}, StripOnInput -> False]],
      Annotation[#, 
       Style[0.05617977528089888, {
         GrayLevel[0]}], "Tooltip"]& ], 
     TagBox[
      TooltipBox[
       TagBox[
        DynamicBox[{
          FEPrivate`If[
           CurrentValue["MouseOver"], 
           EdgeForm[{
             GrayLevel[0.5], 
             AbsoluteThickness[1.5], 
             Opacity[0.66]}], {}, {}], 
          
          RectangleBox[{2., 0.}, {2.2, 0.07490636704119844}, "RoundingRadius" -> 
           0]},
         ImageSizeCache->{{124.30148423005568`, 132.55491651205938`}, {
          26.514748768998928`, 40.27773676735785}}],
        StatusArea[#, 0.07490636704119844]& ,
        TagBoxNote->"0.07490636704119844"],
       StyleBox["0.07490636704119844`", {
         GrayLevel[0]}, StripOnInput -> False]],
      Annotation[#, 
       Style[0.07490636704119844, {
         GrayLevel[0]}], "Tooltip"]& ], 
     TagBox[
      TooltipBox[
       TagBox[
        DynamicBox[{
          FEPrivate`If[
           CurrentValue["MouseOver"], 
           EdgeForm[{
             GrayLevel[0.5], 
             AbsoluteThickness[1.5], 
             Opacity[0.66]}], {}, {}], 
          
          RectangleBox[{2.2, 0.}, {2.4, 0.043695380774032525`}, 
           "RoundingRadius" -> 0]},
         ImageSizeCache->{{131.55491651205938`, 139.80834879406308`}, {
          31.832660434981797`, 40.27773676735785}}],
        StatusArea[#, 0.043695380774032525`]& ,
        TagBoxNote->"0.043695380774032525"],
       StyleBox["0.043695380774032525`", {
         GrayLevel[0]}, StripOnInput -> False]],
      Annotation[#, 
       Style[0.043695380774032525`, {
         GrayLevel[0]}], "Tooltip"]& ], 
     TagBox[
      TooltipBox[
       TagBox[
        DynamicBox[{
          FEPrivate`If[
           CurrentValue["MouseOver"], 
           EdgeForm[{
             GrayLevel[0.5], 
             AbsoluteThickness[1.5], 
             Opacity[0.66]}], {}, {}], 
          
          RectangleBox[{2.4, 0.}, {2.6, 0.03745318352059922}, 
           "RoundingRadius" -> 0]},
         ImageSizeCache->{{138.80834879406308`, 147.06178107606684`}, {
          32.89624276817839, 40.27773676735785}}],
        StatusArea[#, 0.03745318352059922]& ,
        TagBoxNote->"0.03745318352059922"],
       StyleBox["0.03745318352059922`", {
         GrayLevel[0]}, StripOnInput -> False]],
      Annotation[#, 
       Style[0.03745318352059922, {
         GrayLevel[0]}], "Tooltip"]& ], 
     TagBox[
      TooltipBox[
       TagBox[
        DynamicBox[{
          FEPrivate`If[
           CurrentValue["MouseOver"], 
           EdgeForm[{
             GrayLevel[0.5], 
             AbsoluteThickness[1.5], 
             Opacity[0.66]}], {}, {}], 
          
          RectangleBox[{2.6, 0.}, {2.8, 0.043695380774032525`}, 
           "RoundingRadius" -> 0]},
         ImageSizeCache->{{146.06178107606684`, 154.31521335807054`}, {
          31.832660434981797`, 40.27773676735785}}],
        StatusArea[#, 0.043695380774032525`]& ,
        TagBoxNote->"0.043695380774032525"],
       StyleBox["0.043695380774032525`", {
         GrayLevel[0]}, StripOnInput -> False]],
      Annotation[#, 
       Style[0.043695380774032525`, {
         GrayLevel[0]}], "Tooltip"]& ], 
     TagBox[
      TooltipBox[
       TagBox[
        DynamicBox[{
          FEPrivate`If[
           CurrentValue["MouseOver"], 
           EdgeForm[{
             GrayLevel[0.5], 
             AbsoluteThickness[1.5], 
             Opacity[0.66]}], {}, {}], 
          
          RectangleBox[{2.8, 0.}, {3., 0.006242197253433203}, 
           "RoundingRadius" -> 0]},
         ImageSizeCache->{{153.31521335807054`, 161.56864564007424`}, {
          38.21415443416127, 40.27773676735785}}],
        StatusArea[#, 0.006242197253433203]& ,
        TagBoxNote->"0.006242197253433203"],
       StyleBox["0.006242197253433203`", {
         GrayLevel[0]}, StripOnInput -> False]],
      Annotation[#, 
       Style[0.006242197253433203, {
         GrayLevel[0]}], "Tooltip"]& ], {}, 
     TagBox[
      TooltipBox[
       TagBox[
        DynamicBox[{
          FEPrivate`If[
           CurrentValue["MouseOver"], 
           EdgeForm[{
             GrayLevel[0.5], 
             AbsoluteThickness[1.5], 
             Opacity[0.66]}], {}, {}], 
          
          RectangleBox[{3.2, 0.}, {3.4, 0.006242197253433218}, 
           "RoundingRadius" -> 0]},
         ImageSizeCache->{{167.82207792207794`, 176.07551020408164`}, {
          38.21415443416127, 40.27773676735785}}],
        StatusArea[#, 0.006242197253433218]& ,
        TagBoxNote->"0.006242197253433218"],
       StyleBox["0.006242197253433218`", {
         GrayLevel[0]}, StripOnInput -> False]],
      Annotation[#, 
       Style[0.006242197253433218, {
         GrayLevel[0]}], "Tooltip"]& ]}, {}, {}}, {{}, {}, 
    {RGBColor[1, 0, 0], Thickness[Large], LineBox[CompressedData["
1:eJwV1nc8lf/7B/CjYx0hn6SiYRalRbZ03YioyCgk9Q3HqKQyk5FNHCU7o6zk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      "]]}}},
  AspectRatio->NCache[GoldenRatio^(-1), 0.6180339887498948],
  Axes->{True, True},
  AxesOrigin->{-1., 0},
  FrameTicks->{{Automatic, Automatic}, {Automatic, Automatic}},
  GridLines->{None, None},
  PlotRange->NCache[{{-1, 
      Rational[17, 5]}, {All, All}}, {{-1, 3.4}, {All, All}}],
  PlotRangePadding->{{
     Scaled[0.02], 
     Scaled[0.02]}, {
     Scaled[0.02], 
     Scaled[0.1]}},
  Ticks->{Automatic, Automatic}]], "Output",
 ImageSize->{184, 117},
 ImageMargins->{{0, 0}, {0, 0}},
 ImageRegion->{{0, 1}, {0, 1}},
 CellLabel->"Out[9]=",
 CellID->92377784]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"Table", "[", 
  RowBox[{
   RowBox[{
    RowBox[{"data", "=", 
     RowBox[{"res", "[", 
      RowBox[{"[", 
       RowBox[{
        RowBox[{"adpmax", ";;", ";;", "10"}], ",", "1", ",", 
        RowBox[{"-", "1"}], ",", "i"}], "]"}], "]"}]}], ";", 
    "\[IndentingNewLine]", 
    RowBox[{"Show", "[", 
     RowBox[{
      RowBox[{"Histogram", "[", 
       RowBox[{"data", ",", "Automatic", ",", "\"\<ProbabilityDensity\>\""}], 
       "]"}], ",", 
      RowBox[{"Plot", "[", 
       RowBox[{
        RowBox[{"PDF", "[", 
         RowBox[{
          RowBox[{"NormalDistribution", "[", 
           RowBox[{"0", ",", "1"}], "]"}], ",", "x"}], "]"}], ",", 
        RowBox[{"{", 
         RowBox[{"x", ",", 
          RowBox[{"Min", "[", "data", "]"}], ",", 
          RowBox[{"Max", "[", "data", "]"}]}], "}"}], ",", 
        RowBox[{"PlotStyle", "\[Rule]", 
         RowBox[{"{", 
          RowBox[{"Red", ",", "Thick"}], "}"}]}], ",", 
        RowBox[{"PlotRange", "\[Rule]", "All"}], ",", 
        RowBox[{"AxesLabel", "\[Rule]", 
         RowBox[{"{", 
          RowBox[{
           RowBox[{"\"\<x3\>\"", "<>", 
            RowBox[{"ToString", "[", "i", "]"}]}], ",", "\"\<freq\>\""}], 
          "}"}]}]}], "]"}]}], "]"}]}], ",", 
   RowBox[{"{", 
    RowBox[{"i", ",", "2"}], "}"}]}], "]"}]], "Input",
 CellLabel->"In[10]:=",
 CellID->271797975],

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{
   GraphicsBox[{
     {RGBColor[0.798413061722744, 0.824719615472648, 0.968322270542458], 
      EdgeForm[Opacity[0.7]], {}, 
      {RGBColor[0.798413061722744, 0.824719615472648, 0.968322270542458], 
       EdgeForm[Opacity[0.7]], 
       TagBox[
        TooltipBox[
         TagBox[
          DynamicBox[{
            FEPrivate`If[
             CurrentValue["MouseOver"], 
             EdgeForm[{
               GrayLevel[0.5], 
               AbsoluteThickness[1.5], 
               Opacity[0.66]}], {}, {}], 
            
            RectangleBox[{-3.5, 0.}, {-3., 0.0024968789013732834`}, 
             "RoundingRadius" -> 0]},
           
           ImageSizeCache->{{15.5, 28.775039246467813`}, {38.71191761284229, 
            40.27773676735787}}],
          StatusArea[#, 0.0024968789013732834`]& ,
          TagBoxNote->"0.0024968789013732834"],
         StyleBox["0.0024968789013732834`", {
           GrayLevel[0]}, StripOnInput -> False]],
        Annotation[#, 
         Style[0.0024968789013732834`, {
           GrayLevel[0]}], "Tooltip"]& ], 
       TagBox[
        TooltipBox[
         TagBox[
          DynamicBox[{
            FEPrivate`If[
             CurrentValue["MouseOver"], 
             EdgeForm[{
               GrayLevel[0.5], 
               AbsoluteThickness[1.5], 
               Opacity[0.66]}], {}, {}], 
            
            RectangleBox[{-3., 0.}, {-2.5, 0.0024968789013732834`}, 
             "RoundingRadius" -> 0]},
           
           ImageSizeCache->{{27.775039246467813`, 41.05007849293563}, {
            38.71191761284229, 40.27773676735787}}],
          StatusArea[#, 0.0024968789013732834`]& ,
          TagBoxNote->"0.0024968789013732834"],
         StyleBox["0.0024968789013732834`", {
           GrayLevel[0]}, StripOnInput -> False]],
        Annotation[#, 
         Style[0.0024968789013732834`, {
           GrayLevel[0]}], "Tooltip"]& ], 
       TagBox[
        TooltipBox[
         TagBox[
          DynamicBox[{
            FEPrivate`If[
             CurrentValue["MouseOver"], 
             EdgeForm[{
               GrayLevel[0.5], 
               AbsoluteThickness[1.5], 
               Opacity[0.66]}], {}, {}], 
            
            RectangleBox[{-2.5, 0.}, {-2., 0.039950062421972535`}, 
             "RoundingRadius" -> 0]},
           
           ImageSizeCache->{{40.05007849293563, 53.32511773940345}, {
            30.224630295108575`, 40.27773676735787}}],
          StatusArea[#, 0.039950062421972535`]& ,
          TagBoxNote->"0.039950062421972535"],
         StyleBox["0.039950062421972535`", {
           GrayLevel[0]}, StripOnInput -> False]],
        Annotation[#, 
         Style[0.039950062421972535`, {
           GrayLevel[0]}], "Tooltip"]& ], 
       TagBox[
        TooltipBox[
         TagBox[
          DynamicBox[{
            FEPrivate`If[
             CurrentValue["MouseOver"], 
             EdgeForm[{
               GrayLevel[0.5], 
               AbsoluteThickness[1.5], 
               Opacity[0.66]}], {}, {}], 
            
            RectangleBox[{-2., 0.}, {-1.5, 0.09238451935081149}, 
             "RoundingRadius" -> 0]},
           
           ImageSizeCache->{{52.32511773940345, 65.60015698587128}, {
            18.34242805028137, 40.27773676735787}}],
          StatusArea[#, 0.09238451935081149]& ,
          TagBoxNote->"0.09238451935081149"],
         StyleBox["0.09238451935081149`", {
           GrayLevel[0]}, StripOnInput -> False]],
        Annotation[#, 
         Style[0.09238451935081149, {
           GrayLevel[0]}], "Tooltip"]& ], 
       TagBox[
        TooltipBox[
         TagBox[
          DynamicBox[{
            FEPrivate`If[
             CurrentValue["MouseOver"], 
             EdgeForm[{
               GrayLevel[0.5], 
               AbsoluteThickness[1.5], 
               Opacity[0.66]}], {}, {}], 
            RectangleBox[{-1.5, 0.}, {-1., 0.16978776529338327`}, 
             "RoundingRadius" -> 0]},
           
           ImageSizeCache->{{64.60015698587128, 77.87519623233909}, {
            0.8020342602983561, 40.27773676735787}}],
          StatusArea[#, 0.16978776529338327`]& ,
          TagBoxNote->"0.16978776529338327"],
         StyleBox["0.16978776529338327`", {
           GrayLevel[0]}, StripOnInput -> False]],
        Annotation[#, 
         Style[0.16978776529338327`, {
           GrayLevel[0]}], "Tooltip"]& ], 
       TagBox[
        TooltipBox[
         TagBox[
          DynamicBox[{
            FEPrivate`If[
             CurrentValue["MouseOver"], 
             EdgeForm[{
               GrayLevel[0.5], 
               AbsoluteThickness[1.5], 
               Opacity[0.66]}], {}, {}], 
            
            RectangleBox[{-1., 0.}, {-0.5, 0.3196004993757803}, 
             "RoundingRadius" -> 0]},
           
           ImageSizeCache->{{76.87519623233909, 
            90.1502354788069}, {-33.147115010636504`, 40.27773676735787}}],
          StatusArea[#, 0.3196004993757803]& ,
          TagBoxNote->"0.3196004993757803"],
         StyleBox["0.3196004993757803`", {
           GrayLevel[0]}, StripOnInput -> False]],
        Annotation[#, 
         Style[0.3196004993757803, {
           GrayLevel[0]}], "Tooltip"]& ], 
       TagBox[
        TooltipBox[
         TagBox[
          DynamicBox[{
            FEPrivate`If[
             CurrentValue["MouseOver"], 
             EdgeForm[{
               GrayLevel[0.5], 
               AbsoluteThickness[1.5], 
               Opacity[0.66]}], {}, {}], 
            
            RectangleBox[{-0.5, 0.}, {0., 0.3945068664169788}, 
             "RoundingRadius" -> 0]},
           
           ImageSizeCache->{{89.1502354788069, 
            102.42527472527473`}, {-50.12168964610394, 40.27773676735787}}],
          StatusArea[#, 0.3945068664169788]& ,
          TagBoxNote->"0.3945068664169788"],
         StyleBox["0.3945068664169788`", {
           GrayLevel[0]}, StripOnInput -> False]],
        Annotation[#, 
         Style[0.3945068664169788, {
           GrayLevel[0]}], "Tooltip"]& ], 
       TagBox[
        TooltipBox[
         TagBox[
          DynamicBox[{
            FEPrivate`If[
             CurrentValue["MouseOver"], 
             EdgeForm[{
               GrayLevel[0.5], 
               AbsoluteThickness[1.5], 
               Opacity[0.66]}], {}, {}], 
            
            RectangleBox[{0., 0.}, {0.5, 0.38202247191011235`}, 
             "RoundingRadius" -> 0]},
           
           ImageSizeCache->{{101.42527472527473`, 
            114.70031397174256`}, {-47.29259387352604, 40.27773676735787}}],
          StatusArea[#, 0.38202247191011235`]& ,
          TagBoxNote->"0.38202247191011235"],
         StyleBox["0.38202247191011235`", {
           GrayLevel[0]}, StripOnInput -> False]],
        Annotation[#, 
         Style[0.38202247191011235`, {
           GrayLevel[0]}], "Tooltip"]& ], 
       TagBox[
        TooltipBox[
         TagBox[
          DynamicBox[{
            FEPrivate`If[
             CurrentValue["MouseOver"], 
             EdgeForm[{
               GrayLevel[0.5], 
               AbsoluteThickness[1.5], 
               Opacity[0.66]}], {}, {}], 
            
            RectangleBox[{0.5, 0.}, {1., 0.2696629213483146}, 
             "RoundingRadius" -> 0]},
           
           ImageSizeCache->{{113.70031397174256`, 
            126.97535321821037`}, {-21.830731920324887`, 40.27773676735787}}],
          StatusArea[#, 0.2696629213483146]& ,
          TagBoxNote->"0.2696629213483146"],
         StyleBox["0.2696629213483146`", {
           GrayLevel[0]}, StripOnInput -> False]],
        Annotation[#, 
         Style[0.2696629213483146, {
           GrayLevel[0]}], "Tooltip"]& ], 
       TagBox[
        TooltipBox[
         TagBox[
          DynamicBox[{
            FEPrivate`If[
             CurrentValue["MouseOver"], 
             EdgeForm[{
               GrayLevel[0.5], 
               AbsoluteThickness[1.5], 
               Opacity[0.66]}], {}, {}], 
            
            RectangleBox[{1., 0.}, {1.5, 0.18726591760299627`}, 
             "RoundingRadius" -> 0]},
           
           ImageSizeCache->{{125.97535321821037`, 
            139.25039246467819`}, {-3.1586998213107123`, 40.27773676735787}}],
          StatusArea[#, 0.18726591760299627`]& ,
          TagBoxNote->"0.18726591760299627"],
         StyleBox["0.18726591760299627`", {
           GrayLevel[0]}, StripOnInput -> False]],
        Annotation[#, 
         Style[0.18726591760299627`, {
           GrayLevel[0]}], "Tooltip"]& ], 
       TagBox[
        TooltipBox[
         TagBox[
          DynamicBox[{
            FEPrivate`If[
             CurrentValue["MouseOver"], 
             EdgeForm[{
               GrayLevel[0.5], 
               AbsoluteThickness[1.5], 
               Opacity[0.66]}], {}, {}], 
            
            RectangleBox[{1.5, 0.}, {2., 0.09488139825218476}, 
             "RoundingRadius" -> 0]},
           
           ImageSizeCache->{{138.25039246467819`, 151.525431711146}, {
            17.776608895765794`, 40.27773676735787}}],
          StatusArea[#, 0.09488139825218476]& ,
          TagBoxNote->"0.09488139825218476"],
         StyleBox["0.09488139825218476`", {
           GrayLevel[0]}, StripOnInput -> False]],
        Annotation[#, 
         Style[0.09488139825218476, {
           GrayLevel[0]}], "Tooltip"]& ], 
       TagBox[
        TooltipBox[
         TagBox[
          DynamicBox[{
            FEPrivate`If[
             CurrentValue["MouseOver"], 
             EdgeForm[{
               GrayLevel[0.5], 
               AbsoluteThickness[1.5], 
               Opacity[0.66]}], {}, {}], 
            
            RectangleBox[{2., 0.}, {2.5, 0.039950062421972535`}, 
             "RoundingRadius" -> 0]},
           
           ImageSizeCache->{{150.525431711146, 163.80047095761384`}, {
            30.224630295108575`, 40.27773676735787}}],
          StatusArea[#, 0.039950062421972535`]& ,
          TagBoxNote->"0.039950062421972535"],
         StyleBox["0.039950062421972535`", {
           GrayLevel[0]}, StripOnInput -> False]],
        Annotation[#, 
         Style[0.039950062421972535`, {
           GrayLevel[0]}], "Tooltip"]& ], 
       TagBox[
        TooltipBox[
         TagBox[
          DynamicBox[{
            FEPrivate`If[
             CurrentValue["MouseOver"], 
             EdgeForm[{
               GrayLevel[0.5], 
               AbsoluteThickness[1.5], 
               Opacity[0.66]}], {}, {}], 
            
            RectangleBox[{2.5, 0.}, {3., 0.004993757802746567}, 
             "RoundingRadius" -> 0]},
           
           ImageSizeCache->{{162.80047095761384`, 176.07551020408164`}, {
            38.14609845832671, 40.27773676735787}}],
          StatusArea[#, 0.004993757802746567]& ,
          TagBoxNote->"0.004993757802746567"],
         StyleBox["0.004993757802746567`", {
           GrayLevel[0]}, StripOnInput -> False]],
        Annotation[#, 
         Style[0.004993757802746567, {
           GrayLevel[0]}], "Tooltip"]& ]}, {}, {}}, {{}, {}, 
      {RGBColor[1, 0, 0], Thickness[Large], LineBox[CompressedData["
1:eJwV13k8Fd0fB/DrbkQlUcRjKWVLEYoi3ynLkxQRimSPFLJli1xrlqwhS4sl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        "]]}}},
    AspectRatio->NCache[GoldenRatio^(-1), 0.6180339887498948],
    Axes->{True, True},
    AxesOrigin->{-3.5, 0.},
    FrameTicks->{{Automatic, Automatic}, {Automatic, Automatic}},
    GridLines->{None, None},
    PlotRange->NCache[{{
        Rational[-7, 2], 3}, {All, All}}, {{-3.5, 3}, {All, All}}],
    PlotRangePadding->{{
       Scaled[0.02], 
       Scaled[0.02]}, {
       Scaled[0.02], 
       Scaled[0.1]}},
    Ticks->{Automatic, Automatic}], ",", 
   GraphicsBox[{
     {RGBColor[0.798413061722744, 0.824719615472648, 0.968322270542458], 
      EdgeForm[Opacity[0.7]], {}, 
      {RGBColor[0.798413061722744, 0.824719615472648, 0.968322270542458], 
       EdgeForm[Opacity[0.7]], 
       TagBox[
        TooltipBox[
         TagBox[
          DynamicBox[{
            FEPrivate`If[
             CurrentValue["MouseOver"], 
             EdgeForm[{
               GrayLevel[0.5], 
               AbsoluteThickness[1.5], 
               Opacity[0.66]}], {}, {}], 
            
            RectangleBox[{-3.5, 0.}, {-3., 0.0024968789013732834`}, 
             "RoundingRadius" -> 0]},
           
           ImageSizeCache->{{15.5, 28.775039246467813`}, {38.74279890746611, 
            40.27773676735788}}],
          StatusArea[#, 0.0024968789013732834`]& ,
          TagBoxNote->"0.0024968789013732834"],
         StyleBox["0.0024968789013732834`", {
           GrayLevel[0]}, StripOnInput -> False]],
        Annotation[#, 
         Style[0.0024968789013732834`, {
           GrayLevel[0]}], "Tooltip"]& ], 
       TagBox[
        TooltipBox[
         TagBox[
          DynamicBox[{
            FEPrivate`If[
             CurrentValue["MouseOver"], 
             EdgeForm[{
               GrayLevel[0.5], 
               AbsoluteThickness[1.5], 
               Opacity[0.66]}], {}, {}], 
            
            RectangleBox[{-3., 0.}, {-2.5, 0.004993757802746567}, 
             "RoundingRadius" -> 0]},
           
           ImageSizeCache->{{27.775039246467813`, 41.05007849293563}, {
            38.20786104757434, 40.27773676735788}}],
          StatusArea[#, 0.004993757802746567]& ,
          TagBoxNote->"0.004993757802746567"],
         StyleBox["0.004993757802746567`", {
           GrayLevel[0]}, StripOnInput -> False]],
        Annotation[#, 
         Style[0.004993757802746567, {
           GrayLevel[0]}], "Tooltip"]& ], 
       TagBox[
        TooltipBox[
         TagBox[
          DynamicBox[{
            FEPrivate`If[
             CurrentValue["MouseOver"], 
             EdgeForm[{
               GrayLevel[0.5], 
               AbsoluteThickness[1.5], 
               Opacity[0.66]}], {}, {}], 
            
            RectangleBox[{-2.5, 0.}, {-2., 0.02746566791510612}, 
             "RoundingRadius" -> 0]},
           
           ImageSizeCache->{{40.05007849293563, 53.32511773940345}, {
            33.39342030854841, 40.27773676735788}}],
          StatusArea[#, 0.02746566791510612]& ,
          TagBoxNote->"0.02746566791510612"],
         StyleBox["0.02746566791510612`", {
           GrayLevel[0]}, StripOnInput -> False]],
        Annotation[#, 
         Style[0.02746566791510612, {
           GrayLevel[0]}], "Tooltip"]& ], 
       TagBox[
        TooltipBox[
         TagBox[
          DynamicBox[{
            FEPrivate`If[
             CurrentValue["MouseOver"], 
             EdgeForm[{
               GrayLevel[0.5], 
               AbsoluteThickness[1.5], 
               Opacity[0.66]}], {}, {}], 
            
            RectangleBox[{-2., 0.}, {-1.5, 0.09987515605493133}, 
             "RoundingRadius" -> 0]},
           
           ImageSizeCache->{{52.32511773940345, 65.60015698587128}, {
            17.880222371687083`, 40.27773676735788}}],
          StatusArea[#, 0.09987515605493133]& ,
          TagBoxNote->"0.09987515605493133"],
         StyleBox["0.09987515605493133`", {
           GrayLevel[0]}, StripOnInput -> False]],
        Annotation[#, 
         Style[0.09987515605493133, {
           GrayLevel[0]}], "Tooltip"]& ], 
       TagBox[
        TooltipBox[
         TagBox[
          DynamicBox[{
            FEPrivate`If[
             CurrentValue["MouseOver"], 
             EdgeForm[{
               GrayLevel[0.5], 
               AbsoluteThickness[1.5], 
               Opacity[0.66]}], {}, {}], 
            
            RectangleBox[{-1.5, 0.}, {-1., 0.1822721598002497}, 
             "RoundingRadius" -> 0]},
           
           ImageSizeCache->{{64.60015698587128, 77.87519623233909}, {
            0.22727299525867295`, 40.27773676735788}}],
          StatusArea[#, 0.1822721598002497]& ,
          TagBoxNote->"0.1822721598002497"],
         StyleBox["0.1822721598002497`", {
           GrayLevel[0]}, StripOnInput -> False]],
        Annotation[#, 
         Style[0.1822721598002497, {
           GrayLevel[0]}], "Tooltip"]& ], 
       TagBox[
        TooltipBox[
         TagBox[
          DynamicBox[{
            FEPrivate`If[
             CurrentValue["MouseOver"], 
             EdgeForm[{
               GrayLevel[0.5], 
               AbsoluteThickness[1.5], 
               Opacity[0.66]}], {}, {}], 
            
            RectangleBox[{-1., 0.}, {-0.5, 0.29213483146067415`}, 
             "RoundingRadius" -> 0]},
           
           ImageSizeCache->{{76.87519623233909, 
            90.1502354788069}, {-23.309992839979202`, 40.27773676735788}}],
          StatusArea[#, 0.29213483146067415`]& ,
          TagBoxNote->"0.29213483146067415"],
         StyleBox["0.29213483146067415`", {
           GrayLevel[0]}, StripOnInput -> False]],
        Annotation[#, 
         Style[0.29213483146067415`, {
           GrayLevel[0]}], "Tooltip"]& ], 
       TagBox[
        TooltipBox[
         TagBox[
          DynamicBox[{
            FEPrivate`If[
             CurrentValue["MouseOver"], 
             EdgeForm[{
               GrayLevel[0.5], 
               AbsoluteThickness[1.5], 
               Opacity[0.66]}], {}, {}], 
            
            RectangleBox[{-0.5, 0.}, {0., 0.42197253433208487`}, 
             "RoundingRadius" -> 0]},
           
           ImageSizeCache->{{89.1502354788069, 
            102.42527472527473`}, {-51.12676155435123, 40.27773676735788}}],
          StatusArea[#, 0.42197253433208487`]& ,
          TagBoxNote->"0.42197253433208487"],
         StyleBox["0.42197253433208487`", {
           GrayLevel[0]}, StripOnInput -> False]],
        Annotation[#, 
         Style[0.42197253433208487`, {
           GrayLevel[0]}], "Tooltip"]& ], 
       TagBox[
        TooltipBox[
         TagBox[
          DynamicBox[{
            FEPrivate`If[
             CurrentValue["MouseOver"], 
             EdgeForm[{
               GrayLevel[0.5], 
               AbsoluteThickness[1.5], 
               Opacity[0.66]}], {}, {}], 
            
            RectangleBox[{0., 0.}, {0.5, 0.38451935081148564`}, 
             "RoundingRadius" -> 0]},
           
           ImageSizeCache->{{101.42527472527473`, 
            114.70031397174256`}, {-43.10269365597469, 40.27773676735788}}],
          StatusArea[#, 0.38451935081148564`]& ,
          TagBoxNote->"0.38451935081148564"],
         StyleBox["0.38451935081148564`", {
           GrayLevel[0]}, StripOnInput -> False]],
        Annotation[#, 
         Style[0.38451935081148564`, {
           GrayLevel[0]}], "Tooltip"]& ], 
       TagBox[
        TooltipBox[
         TagBox[
          DynamicBox[{
            FEPrivate`If[
             CurrentValue["MouseOver"], 
             EdgeForm[{
               GrayLevel[0.5], 
               AbsoluteThickness[1.5], 
               Opacity[0.66]}], {}, {}], 
            
            RectangleBox[{0.5, 0.}, {1., 0.299625468164794}, "RoundingRadius" -> 
             0]},
           
           ImageSizeCache->{{113.70031397174256`, 
            126.97535321821037`}, {-24.91480641965451, 40.27773676735788}}],
          StatusArea[#, 0.299625468164794]& ,
          TagBoxNote->"0.299625468164794"],
         StyleBox["0.299625468164794`", {
           GrayLevel[0]}, StripOnInput -> False]],
        Annotation[#, 
         Style[0.299625468164794, {
           GrayLevel[0]}], "Tooltip"]& ], 
       TagBox[
        TooltipBox[
         TagBox[
          DynamicBox[{
            FEPrivate`If[
             CurrentValue["MouseOver"], 
             EdgeForm[{
               GrayLevel[0.5], 
               AbsoluteThickness[1.5], 
               Opacity[0.66]}], {}, {}], 
            
            RectangleBox[{1., 0.}, {1.5, 0.14232209737827714`}, 
             "RoundingRadius" -> 0]},
           
           ImageSizeCache->{{125.97535321821037`, 139.25039246467819`}, {
            8.786278753526997, 40.27773676735788}}],
          StatusArea[#, 0.14232209737827714`]& ,
          TagBoxNote->"0.14232209737827714"],
         StyleBox["0.14232209737827714`", {
           GrayLevel[0]}, StripOnInput -> False]],
        Annotation[#, 
         Style[0.14232209737827714`, {
           GrayLevel[0]}], "Tooltip"]& ], 
       TagBox[
        TooltipBox[
         TagBox[
          DynamicBox[{
            FEPrivate`If[
             CurrentValue["MouseOver"], 
             EdgeForm[{
               GrayLevel[0.5], 
               AbsoluteThickness[1.5], 
               Opacity[0.66]}], {}, {}], 
            
            RectangleBox[{1.5, 0.}, {2., 0.09238451935081149}, 
             "RoundingRadius" -> 0]},
           
           ImageSizeCache->{{138.25039246467819`, 151.525431711146}, {
            19.48503595136239, 40.27773676735788}}],
          StatusArea[#, 0.09238451935081149]& ,
          TagBoxNote->"0.09238451935081149"],
         StyleBox["0.09238451935081149`", {
           GrayLevel[0]}, StripOnInput -> False]],
        Annotation[#, 
         Style[0.09238451935081149, {
           GrayLevel[0]}], "Tooltip"]& ], 
       TagBox[
        TooltipBox[
         TagBox[
          DynamicBox[{
            FEPrivate`If[
             CurrentValue["MouseOver"], 
             EdgeForm[{
               GrayLevel[0.5], 
               AbsoluteThickness[1.5], 
               Opacity[0.66]}], {}, {}], 
            
            RectangleBox[{2., 0.}, {2.5, 0.04244694132334582}, 
             "RoundingRadius" -> 0]},
           
           ImageSizeCache->{{150.525431711146, 163.80047095761384`}, {
            30.18379314919779, 40.27773676735788}}],
          StatusArea[#, 0.04244694132334582]& ,
          TagBoxNote->"0.04244694132334582"],
         StyleBox["0.04244694132334582`", {
           GrayLevel[0]}, StripOnInput -> False]],
        Annotation[#, 
         Style[0.04244694132334582, {
           GrayLevel[0]}], "Tooltip"]& ], 
       TagBox[
        TooltipBox[
         TagBox[
          DynamicBox[{
            FEPrivate`If[
             CurrentValue["MouseOver"], 
             EdgeForm[{
               GrayLevel[0.5], 
               AbsoluteThickness[1.5], 
               Opacity[0.66]}], {}, {}], 
            
            RectangleBox[{2.5, 0.}, {3., 0.00749063670411985}, 
             "RoundingRadius" -> 0]},
           
           ImageSizeCache->{{162.80047095761384`, 176.07551020408164`}, {
            37.67292318768257, 40.27773676735788}}],
          StatusArea[#, 0.00749063670411985]& ,
          TagBoxNote->"0.00749063670411985"],
         StyleBox["0.00749063670411985`", {
           GrayLevel[0]}, StripOnInput -> False]],
        Annotation[#, 
         Style[0.00749063670411985, {
           GrayLevel[0]}], "Tooltip"]& ]}, {}, {}}, {{}, {}, 
      {RGBColor[1, 0, 0], Thickness[Large], LineBox[CompressedData["
1:eJwV1nc4Vv8bB3DPslJKQ0Uko7LKChn3SaJkF4mSUXZWsmXvkU2J7BmhkHmO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        "]]}}},
    AspectRatio->NCache[GoldenRatio^(-1), 0.6180339887498948],
    Axes->{True, True},
    AxesOrigin->{-3.5, 0.},
    FrameTicks->{{Automatic, Automatic}, {Automatic, Automatic}},
    GridLines->{None, None},
    PlotRange->NCache[{{
        Rational[-7, 2], 3}, {All, All}}, {{-3.5, 3}, {All, All}}],
    PlotRangePadding->{{
       Scaled[0.02], 
       Scaled[0.02]}, {
       Scaled[0.02], 
       Scaled[0.1]}},
    Ticks->{Automatic, Automatic}]}], "}"}]], "Output",
 ImageSize->{389, 119},
 ImageMargins->{{0, 0}, {0, 0}},
 ImageRegion->{{0, 1}, {0, 1}},
 CellLabel->"Out[10]=",
 CellID->85716915]
}, Open  ]]
}, Closed]]
}, Open  ]],

Cell[CellGroupData[{

Cell[TextData[ButtonBox["MORE ABOUT",
 BaseStyle->None,
 Appearance->{Automatic, None},
 Evaluator->None,
 Method->"Preemptive",
 ButtonFunction:>(FrontEndExecute[{
    FrontEnd`SelectionMove[
     FrontEnd`SelectedNotebook[], All, ButtonCell], 
    FrontEndToken["OpenCloseGroup"], 
    FrontEnd`SelectionMove[
     FrontEnd`SelectedNotebook[], After, 
     CellContents]}]& )]], "MoreAboutSection",
 CellID->38303248],

Cell[TextData[ButtonBox["Metropolis",
 BaseStyle->"Link",
 ButtonData->"paclet:Metropolis/guide/Metropolis"]], "MoreAbout",
 CellID->116142772]
}, Open  ]],

Cell[" ", "FooterCell"]
},
Saveable->False,
ScreenStyleEnvironment->"Working",
WindowSize->{725, 730},
WindowMargins->{{0, Automatic}, {Automatic, 0}},
WindowTitle->"AdaptiveProposal - Wolfram Mathematica",
TaggingRules->{
 "ModificationHighlight" -> False, 
  "Metadata" -> {
   "built" -> "{2013, 2, 10, 17, 35, 15.8514070}", "context" -> "Metropolis`",
     "keywords" -> {}, "index" -> True, "label" -> "Metropolis Paclet Symbol",
     "language" -> "en", "paclet" -> "Metropolis", "status" -> "None", 
    "summary" -> 
    "AdaptiveProposal[isadp, indaccept, adprate, adpset, xls] gives updated \
adpset. It calculates the recent accept rate, and the scale and the \
covariance of proposal distributions with adpative rate adprate if isadp. The \
xls is a list of random sample x, and the indaccept is the average accept \
probability of the corresponding prposal list.AdaptiveProposal[isadp, \
indacceptls, adprate, adpset, {xls_1, ..}] gives updated adpset for a set of \
random variables. It calculates the recent list of accept rates, and the \
scale and the covariance of proposal distributions with adpative rate adprate \
if isadp. The xls_i is a list of random sample x_i for i=1,.., and the \
indaccept is the corresponding list of average accept probabilities. ", 
    "synonyms" -> {}, "title" -> "AdaptiveProposal", "type" -> "Symbol", 
    "uri" -> "Metropolis/ref/AdaptiveProposal"}, "LinkTrails" -> "", 
  "SearchTextTranslated" -> ""},
CellContext->"Global`",
FrontEndVersion->"8.0 for Microsoft Windows (64-bit) (February 23, 2011)",
StyleDefinitions->Notebook[{
   Cell[
    StyleData[
    StyleDefinitions -> FrontEnd`FileName[{"Wolfram"}, "Reference.nb"]]], 
   Cell[
    StyleData["Input"], CellContext -> "Global`"], 
   Cell[
    StyleData["Output"], CellContext -> "Global`"]}, Visible -> False, 
  FrontEndVersion -> "8.0 for Microsoft Windows (64-bit) (February 23, 2011)",
   StyleDefinitions -> "Default.nb"]
]
(* End of Notebook Content *)

(* Internal cache information *)
(*CellTagsOutline
CellTagsIndex->{
 "PrimaryExamplesSection"->{
  Cell[6555, 195, 461, 13, 70, "PrimaryExamplesSection",
   CellTags->"PrimaryExamplesSection",
   CellID->184452762]}
 }
*)
(*CellTagsIndex
CellTagsIndex->{
 {"PrimaryExamplesSection", 246264, 5819}
 }
*)
(*NotebookFileOutline
Notebook[{
Cell[609, 21, 76, 1, 70, "SymbolColorBar"],
Cell[688, 24, 185, 5, 70, "LinkTrail"],
Cell[876, 31, 2343, 53, 70, "AnchorBarGrid",
 CellID->1],
Cell[CellGroupData[{
Cell[3244, 88, 59, 1, 70, "ObjectName",
 CellID->1224892054],
Cell[3306, 91, 2138, 60, 70, "Usage",
 CellID->982511436]
}, Open  ]],
Cell[CellGroupData[{
Cell[5481, 156, 580, 17, 70, "NotesSection",
 CellGroupingRules->{"SectionGrouping", 50},
 CellID->178741072],
Cell[6064, 175, 454, 15, 70, "Notes",
 CellID->1067943069]
}, Closed]],
Cell[CellGroupData[{
Cell[6555, 195, 461, 13, 70, "PrimaryExamplesSection",
 CellTags->"PrimaryExamplesSection",
 CellID->184452762],
Cell[CellGroupData[{
Cell[7041, 212, 148, 5, 70, "ExampleSection",
 CellID->202422922],
Cell[7192, 219, 131, 3, 70, "Input",
 CellID->307212672],
Cell[7326, 224, 118, 3, 70, "Input",
 CellID->22466094],
Cell[7447, 229, 84, 1, 70, "ExampleText",
 CellID->13303408],
Cell[CellGroupData[{
Cell[7556, 234, 4028, 120, 70, "Input",
 CellID->231556428],
Cell[11587, 356, 224, 7, 36, "Output",
 CellID->202682309]
}, Open  ]],
Cell[11826, 366, 57, 1, 70, "ExampleText",
 CellID->671059588],
Cell[CellGroupData[{
Cell[11908, 371, 692, 22, 70, "Input",
 CellID->245470297],
Cell[12603, 395, 177, 5, 36, "Output",
 CellID->270803938],
Cell[12783, 402, 3440, 64, 151, "Output",
 CellID->242756173]
}, Open  ]],
Cell[16238, 469, 57, 1, 70, "ExampleText",
 CellID->141799113],
Cell[CellGroupData[{
Cell[16320, 474, 575, 18, 70, "Input",
 CellID->119363098],
Cell[16898, 494, 176, 5, 36, "Output",
 CellID->159232310],
Cell[17077, 501, 3146, 59, 143, "Output",
 CellID->95788349]
}, Open  ]],
Cell[20238, 563, 54, 1, 70, "ExampleText",
 CellID->303467053],
Cell[CellGroupData[{
Cell[20317, 568, 602, 18, 47, "Input",
 CellID->30665817],
Cell[20922, 588, 13855, 235, 139, "Output",
 CellID->179800696]
}, Open  ]],
Cell[CellGroupData[{
Cell[34814, 828, 211, 7, 28, "Input",
 CellID->40892383],
Cell[35028, 837, 3401, 63, 140, "Output",
 CellID->3429041]
}, Open  ]],
Cell[38444, 903, 101, 2, 70, "ExampleText",
 CellID->57759205],
Cell[CellGroupData[{
Cell[38570, 909, 1716, 51, 70, "Input",
 CellID->13266113],
Cell[40289, 962, 79650, 1969, 70, "Output",
 CellID->215573117]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[119988, 2937, 228, 7, 70, "ExampleSection",
 CellID->658589838],
Cell[120219, 2946, 119, 3, 70, "Input",
 CellID->290007961],
Cell[120341, 2951, 153, 4, 70, "ExampleText",
 CellID->73418448],
Cell[CellGroupData[{
Cell[120519, 2959, 5610, 168, 70, "Input",
 CellID->13864406],
Cell[126132, 3129, 224, 7, 36, "Output",
 CellID->270182846]
}, Open  ]],
Cell[126371, 3139, 57, 1, 70, "ExampleText",
 CellID->311617669],
Cell[CellGroupData[{
Cell[126453, 3144, 732, 23, 70, "Input",
 CellID->762797887],
Cell[127188, 3169, 275, 9, 36, "Output",
 CellID->2367316],
Cell[127466, 3180, 5762, 102, 70, "Output",
 CellID->695337132]
}, Open  ]],
Cell[133243, 3285, 57, 1, 70, "ExampleText",
 CellID->805627902],
Cell[CellGroupData[{
Cell[133325, 3290, 615, 19, 70, "Input",
 CellID->807846078],
Cell[133943, 3311, 273, 9, 36, "Output",
 CellID->135472732],
Cell[134219, 3322, 4326, 78, 70, "Output",
 CellID->142531258]
}, Open  ]],
Cell[138560, 3403, 54, 1, 70, "ExampleText",
 CellID->193398859],
Cell[CellGroupData[{
Cell[138639, 3408, 603, 18, 70, "Input",
 CellID->402775542],
Cell[139245, 3428, 12061, 205, 70, "Output",
 CellID->540490369]
}, Open  ]],
Cell[CellGroupData[{
Cell[151343, 3638, 638, 19, 70, "Input",
 CellID->223197944],
Cell[151984, 3659, 13875, 235, 70, "Output",
 CellID->235117103]
}, Open  ]],
Cell[CellGroupData[{
Cell[165896, 3899, 726, 23, 70, "Input",
 CellID->36445872],
Cell[166625, 3924, 6238, 110, 70, "Output",
 CellID->233124039]
}, Open  ]],
Cell[CellGroupData[{
Cell[172900, 4039, 1503, 47, 70, "Input",
 CellID->182745918],
Cell[174406, 4088, 3407, 63, 70, "Output",
 CellID->635841734]
}, Open  ]],
Cell[CellGroupData[{
Cell[177850, 4156, 1265, 39, 70, "Input",
 CellID->530831861],
Cell[179118, 4197, 23583, 594, 70, "Output",
 CellID->92377784]
}, Open  ]],
Cell[CellGroupData[{
Cell[202738, 4796, 1372, 39, 70, "Input",
 CellID->271797975],
Cell[204113, 4837, 39439, 910, 70, "Output",
 CellID->85716915]
}, Open  ]]
}, Closed]]
}, Open  ]],
Cell[CellGroupData[{
Cell[243613, 5754, 419, 12, 70, "MoreAboutSection",
 CellID->38303248],
Cell[244035, 5768, 143, 3, 70, "MoreAbout",
 CellID->116142772]
}, Open  ]],
Cell[244193, 5774, 23, 0, 70, "FooterCell"]
}
]
*)

(* End of internal cache information *)

