(* Content-type: application/vnd.wolfram.mathematica *)

(*** Wolfram Notebook File ***)
(* http://www.wolfram.com/nb *)

(* CreatedBy='Mathematica 8.0' *)

(*CacheID: 234*)
(* Internal cache information:
NotebookFileLineBreakTest
NotebookFileLineBreakTest
NotebookDataPosition[       157,          7]
NotebookDataLength[    363453,       8967]
NotebookOptionsPosition[    357161,       8764]
NotebookOutlinePosition[    359351,       8818]
CellTagsIndexPosition[    359264,       8813]
WindowTitle->BlockAdaptiveMetropolis - Wolfram Mathematica
WindowFrame->Normal*)

(* Beginning of Notebook Content *)
Notebook[{
Cell[" ", "SymbolColorBar",
 CellMargins->{{Inherited, Inherited}, {-5, 0}}],

Cell[TextData[{
 ButtonBox["Metropolis",
  BaseStyle->{"Link", "LinkTrail"},
  ButtonData->"paclet:Metropolis/guide/Metropolis"],
 StyleBox[" > ", "LinkTrailSeparator"]
}], "LinkTrail"],

Cell[BoxData[GridBox[{
   {Cell["METROPOLIS PACKAGE PACLET SYMBOL", "PacletNameCell"], Cell[TextData[{
     Cell[BoxData[
      ActionMenuBox[
       FrameBox["\<\"More About \[RightGuillemet]\"\>",
        StripOnInput->False], {"\<\"Metropolis\"\>":>Documentation`HelpLookup[
       "paclet:Metropolis/guide/Metropolis"]},
       Appearance->None,
       MenuAppearance->Automatic]],
      LineSpacing->{1.4, 0}],
     "\[ThickSpace]\[ThickSpace]\[ThickSpace]\[ThickSpace]\[ThickSpace]\
\[ThickSpace]",
     Cell[BoxData[
      ActionMenuBox[
       FrameBox["\<\"URL \[RightGuillemet]\"\>",
        StripOnInput->
         False], {"\<\"Metropolis/ref/BlockAdaptiveMetropolis\"\>":>
       None, "\<\"Copy Mathematica url\"\>":>
       Module[{DocumentationSearch`Private`nb$}, 
        DocumentationSearch`Private`nb$ = NotebookPut[
           Notebook[{
             Cell["Metropolis/ref/BlockAdaptiveMetropolis"]}, Visible -> 
            False]]; SelectionMove[
         DocumentationSearch`Private`nb$, All, Notebook]; 
        FrontEndTokenExecute[DocumentationSearch`Private`nb$, "Copy"]; 
        NotebookClose[DocumentationSearch`Private`nb$]; Null], 
       Delimiter, "\<\"Copy web url\"\>":>
       Module[{DocumentationSearch`Private`nb$}, 
        DocumentationSearch`Private`nb$ = NotebookPut[
           Notebook[{
             Cell[
              BoxData[
               MakeBoxes[
                Hyperlink[
                "http://reference.wolfram.com/mathematica/Metropolis/ref/\
BlockAdaptiveMetropolis.html"], StandardForm]], "Input", TextClipboardType -> 
              "PlainText"]}, Visible -> False]]; 
        SelectionMove[DocumentationSearch`Private`nb$, All, Notebook]; 
        FrontEndTokenExecute[DocumentationSearch`Private`nb$, "Copy"]; 
        NotebookClose[DocumentationSearch`Private`nb$]; 
        Null], "\<\"Go to web url\"\>":>FrontEndExecute[{
         NotebookLocate[{
           URL[
            StringJoin[
            "http://reference.wolfram.com/mathematica/", 
             "Metropolis/ref/BlockAdaptiveMetropolis", ".html"]], None}]}]},
       Appearance->None,
       MenuAppearance->Automatic]],
      LineSpacing->{1.4, 0}]
    }], "AnchorBar"]}
  }]], "AnchorBarGrid",
 GridBoxOptions->{GridBoxItemSize->{"Columns" -> {
     Scaled[0.65], {
      Scaled[0.34]}}, "ColumnsIndexed" -> {}, "Rows" -> {{1.}}, 
   "RowsIndexed" -> {}}},
 CellID->1],

Cell[CellGroupData[{

Cell["BlockAdaptiveMetropolis", "ObjectName",
 CellID->1224892054],

Cell[BoxData[GridBox[{
   {"", Cell[TextData[{
     Cell[BoxData[
      RowBox[{"BlockAdaptiveMetropolis", "[", 
       RowBox[{
        StyleBox["logpdf", "TI"], ",", " ", 
        SubscriptBox[
         StyleBox["x", "TI"], 
         StyleBox["0", "TR"]], ",", " ", 
        StyleBox["size", "TI"], ",", " ", 
        StyleBox["stepsize", "TI"], ",", " ", 
        StyleBox["adpfraction", "TI"], ",", " ", 
        StyleBox["adprate", "TI"], ",", " ", 
        StyleBox["adpset", "TI"]}], "]"}]], "InlineFormula"],
     "\[LineSeparator]generates size updates for a list of conditional \
indepdent random variables starting from x0. Each update is given by the last \
update of  stepsize BlockStudentTMetropolis updatings. The target \
distribution p[x] is provided by logpdf[x, indices] = Log[p[x, indices]] for \
any subset x[[indices]] of x. The symmetric proprosal distribution is the \
student T distribution. The variance of the proposal distribution is adapted \
in the initial Floor[size adpfraction] iterations with rate adprate, and the \
adaptive parameter set adpset."
    }]]}
  }]], "Usage",
 GridBoxOptions->{
 GridBoxBackground->{
  "Columns" -> {{None}}, "ColumnsIndexed" -> {}, "Rows" -> {{None}}, 
   "RowsIndexed" -> {}}}]
}, Open  ]],

Cell[CellGroupData[{

Cell[TextData[Cell[BoxData[
 ButtonBox[
  FrameBox[
   StyleBox[
    RowBox[{"MORE", " ", "INFORMATION"}], "NotesFrameText"],
   StripOnInput->False],
  Appearance->{Automatic, None},
  BaseStyle->None,
  ButtonFunction:>(FrontEndExecute[{
     FrontEnd`SelectionMove[
      FrontEnd`SelectedNotebook[], All, ButtonCell], 
     FrontEndToken["OpenCloseGroup"], 
     FrontEnd`SelectionMove[
      FrontEnd`SelectedNotebook[], After, CellContents]}]& ),
  Evaluator->None,
  Method->"Preemptive"]]]], "NotesSection",
 CellGroupingRules->{"SectionGrouping", 50},
 CellID->87774412],

Cell[TextData[{
 "To use BlockAdaptiveMetropolis, you first need to load the ",
 Cell[BoxData[
  ButtonBox["Metropolis",
   BaseStyle->"Link",
   ButtonData->"paclet:Metropolis/guide/Metropolis"]], "InlineFormula"],
 " package using  Needs[\"Metropolis`\"]."
}], "Notes",
 CellID->115902181],

Cell[TextData[{
 "BlockAdaptiveMetropolis has options of ",
 Cell[BoxData[
  ButtonBox["Metropolis",
   BaseStyle->"Link",
   ButtonData->"paclet:Metropolis/ref/Metropolis"]], "InlineFormula"],
 " and ",
 Cell[BoxData[
  ButtonBox["StudentTDisplacements",
   BaseStyle->"Link",
   ButtonData->"paclet:Metropolis/ref/StudentTDisplacements"]], 
  "InlineFormula"]
}], "Notes",
 CellID->486766318],

Cell["\<\
Each update is given by {x, logpdf[x], 1/0 (accept/not), \
Min[1,Metropolis-Hastings ratio]}.\
\>", "Notes",
 CellID->894215170]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[ButtonBox["EXAMPLES",
 BaseStyle->None,
 Appearance->{Automatic, None},
 Evaluator->None,
 Method->"Preemptive",
 ButtonFunction:>(FrontEndExecute[{
    FrontEnd`SelectionMove[
     FrontEnd`SelectedNotebook[], All, ButtonCell], 
    FrontEndToken["OpenCloseGroup"], 
    FrontEnd`SelectionMove[
     FrontEnd`SelectedNotebook[], After, 
     CellContents]}]& )]], "PrimaryExamplesSection",
 CellTags->"PrimaryExamplesSection",
 CellID->217326441],

Cell[CellGroupData[{

Cell[TextData[{
 "Basic Examples",
 "\[NonBreakingSpace]\[NonBreakingSpace]",
 Cell["(1)", "ExampleCount"]
}], "ExampleSection",
 CellID->115992704],

Cell[BoxData[
 RowBox[{"Needs", "[", "\"\<Metropolis`\>\"", "]"}]], "Input",
 CellLabel->"In[1]:=",
 CellID->21081259],

Cell["This samples many independent normal distributions", "ExampleText",
 CellID->568556038],

Cell[CellGroupData[{

Cell[BoxData[{
 RowBox[{
  RowBox[{"Clear", "[", 
   RowBox[{
   "logpdf", ",", "scale", ",", "cov", ",", "m", ",", "mu", ",", "sd", ",", 
    "n", ",", "res"}], "]"}], ";"}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{"m", "=", "3"}], ";"}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{"mu", "=", 
   RowBox[{"RandomReal", "[", 
    RowBox[{
     RowBox[{"{", 
      RowBox[{"0", ",", "50"}], "}"}], ",", "m"}], "]"}]}], 
  ";"}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{"sd", "=", 
   RowBox[{"RandomReal", "[", 
    RowBox[{
     RowBox[{"{", 
      RowBox[{"1", ",", "5"}], "}"}], ",", "m"}], "]"}]}], 
  ";"}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{"logpdf", "=", 
   RowBox[{"Function", "[", 
    RowBox[{
     RowBox[{"{", 
      RowBox[{"x", ",", "indices"}], "}"}], ",", 
     RowBox[{
      RowBox[{"-", 
       RowBox[{
        RowBox[{"(", 
         RowBox[{
          RowBox[{"x", "[", 
           RowBox[{"[", "indices", "]"}], "]"}], "-", 
          RowBox[{"mu", "[", 
           RowBox[{"[", "indices", "]"}], "]"}]}], ")"}], "^", "2"}]}], "/", 
      RowBox[{"(", 
       RowBox[{"2", " ", 
        RowBox[{
         RowBox[{"sd", "[", 
          RowBox[{"[", "indices", "]"}], "]"}], "^", "2"}]}], ")"}]}]}], 
    "]"}]}], ";"}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{"scale", "=", 
   RowBox[{"3", " ", "sd"}]}], ";"}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{"cov", "=", 
   RowBox[{"Table", "[", 
    RowBox[{
     RowBox[{"{", 
      RowBox[{"{", "1", "}"}], "}"}], ",", 
     RowBox[{"{", "m", "}"}]}], "]"}]}], ";"}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{"n", "=", "5000"}], ";"}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{"adpmax", "=", "2000"}], ";"}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{"x0", "=", 
   RowBox[{"Table", "[", 
    RowBox[{"0", ",", 
     RowBox[{"{", "m", "}"}]}], "]"}]}], ";"}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{
   RowBox[{"res", "=", 
    RowBox[{"BlockAdaptiveMetropolis", "[", 
     RowBox[{"logpdf", ",", "x0", ",", 
      RowBox[{"2", " ", "n"}], ",", "1", ",", "0.5", ",", 
      RowBox[{"5", "/", "adpmax"}], ",", 
      RowBox[{"InitialProposal", "[", 
       RowBox[{
        RowBox[{"Table", "[", 
         RowBox[{"0.44", ",", 
          RowBox[{"{", "3", "}"}]}], "]"}], ",", "x0"}], "]"}], ",", 
      RowBox[{"IsReturnLast", "\[Rule]", "False"}]}], "]"}]}], ";"}], "//", 
  "Timing"}]}], "Input",
 CellLabel->"In[2]:=",
 CellID->73930184],

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{"11.091999999999999`", ",", "Null"}], "}"}]], "Output",
 ImageSize->{99, 15},
 ImageMargins->{{0, 0}, {0, 0}},
 ImageRegion->{{0, 1}, {0, 1}},
 CellLabel->"Out[2]=",
 CellID->327609976]
}, Open  ]],

Cell["Accept ratito:", "ExampleText",
 CellID->274231],

Cell[CellGroupData[{

Cell[BoxData[{
 RowBox[{"N", "[", 
  RowBox[{"Mean", "[", 
   RowBox[{"res", "[", 
    RowBox[{"[", 
     RowBox[{
      RowBox[{"adpmax", ";;"}], ",", 
      RowBox[{"{", 
       RowBox[{"3", ",", "4"}], "}"}]}], "]"}], "]"}], "]"}], 
  "]"}], "\[IndentingNewLine]", 
 RowBox[{"ListLinePlot", "[", 
  RowBox[{
   RowBox[{"Transpose", "[", 
    RowBox[{"res", "[", 
     RowBox[{"[", 
      RowBox[{"All", ",", "5", ",", 
       RowBox[{"-", "4"}]}], "]"}], "]"}], "]"}], ",", 
   RowBox[{"PlotRange", "\[Rule]", "All"}], ",", 
   RowBox[{"AxesLabel", "\[Rule]", 
    RowBox[{"{", 
     RowBox[{"\"\<t\>\"", ",", "\"\<accept\>\""}], "}"}]}]}], "]"}]}], "Input",
 CellLabel->"In[3]:=",
 CellID->44431038],

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{
   RowBox[{"{", 
    RowBox[{
    "0.48156480439945004`", ",", "0.4649418822647169`", ",", 
     "0.3859517560304962`"}], "}"}], ",", 
   RowBox[{"{", 
    RowBox[{
    "0.4826537911735692`", ",", "0.4632289677360046`", ",", 
     "0.3883633360739337`"}], "}"}]}], "}"}]], "Output",
 ImageSize->{437, 15},
 ImageMargins->{{0, 0}, {0, 0}},
 ImageRegion->{{0, 1}, {0, 1}},
 CellLabel->"Out[3]=",
 CellID->359490002],

Cell[BoxData[
 GraphicsBox[
  TagBox[RasterBox[CompressedData["
1:eJztnYtTFEcex6m6v+T+DarmUJFD9IxCQlAUEcSNKLpBieEROThe8hCX9w6P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    "], {{0, 114}, {180, 0}}, {0, 255},
    ColorFunction->RGBColor],
   BoxForm`ImageTag["Byte", ColorSpace -> "RGB", Interleaving -> True],
   Selectable->False],
  BaseStyle->"ImageGraphics",
  ImageSize->Automatic,
  ImageSizeRaw->{180, 114},
  PlotRange->{{0, 180}, {0, 114}}]], "Output",
 ImageSize->{184, 114},
 ImageMargins->{{0, 0}, {0, 0}},
 ImageRegion->{{0, 1}, {0, 1}},
 CellLabel->"Out[3]=",
 CellID->591258976]
}, Open  ]],

Cell["Proposal scale", "ExampleText",
 CellID->377051331],

Cell[CellGroupData[{

Cell[BoxData[{
 RowBox[{"N", "[", 
  RowBox[{"Mean", "[", 
   RowBox[{"res", "[", 
    RowBox[{"[", 
     RowBox[{
      RowBox[{"n", ";;"}], ",", "5", ",", 
      RowBox[{"-", "3"}]}], "]"}], "]"}], "]"}], "]"}], "\[IndentingNewLine]", 
 RowBox[{"ListLogPlot", "[", 
  RowBox[{
   RowBox[{"Transpose", "[", 
    RowBox[{"res", "[", 
     RowBox[{"[", 
      RowBox[{"All", ",", "5", ",", 
       RowBox[{"-", "3"}]}], "]"}], "]"}], "]"}], ",", 
   RowBox[{"PlotRange", "\[Rule]", "All"}], ",", 
   RowBox[{"Joined", "\[Rule]", "True"}], ",", 
   RowBox[{"AxesLabel", "\[Rule]", 
    RowBox[{"{", 
     RowBox[{"\"\<t\>\"", ",", "\"\<scale\>\""}], "}"}]}]}], "]"}]}], "Input",
 CellLabel->"In[4]:=",
 CellID->254894932],

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{
  "3.9395088935513494`", ",", "4.67610825548133`", ",", 
   "7.226914386803451`"}], "}"}]], "Output",
 ImageSize->{187, 15},
 ImageMargins->{{0, 0}, {0, 0}},
 ImageRegion->{{0, 1}, {0, 1}},
 CellLabel->"Out[4]=",
 CellID->58698215],

Cell[BoxData[
 GraphicsBox[
  TagBox[RasterBox[CompressedData["
1:eJztnYtTE0kex6m6v+T+DapSuoWsgiK6t5wPTl1ROXeX0z1FBATxRVAIkgIf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    "], {{0, 125}, {180, 0}}, {0, 255},
    ColorFunction->RGBColor],
   BoxForm`ImageTag["Byte", ColorSpace -> "RGB", Interleaving -> True],
   Selectable->False],
  BaseStyle->"ImageGraphics",
  ImageSize->Automatic,
  ImageSizeRaw->{180, 125},
  PlotRange->{{0, 180}, {0, 125}}]], "Output",
 ImageSize->{184, 125},
 ImageMargins->{{0, 0}, {0, 0}},
 ImageRegion->{{0, 1}, {0, 1}},
 CellLabel->"Out[4]=",
 CellID->52035]
}, Open  ]],

Cell["Time serie: ", "ExampleText",
 CellID->60397930],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"Table", "[", 
  RowBox[{
   RowBox[{"ListLinePlot", "[", 
    RowBox[{
     RowBox[{"res", "[", 
      RowBox[{"[", 
       RowBox[{
        RowBox[{";;", ";;", "10"}], ",", "1", ",", "i"}], "]"}], "]"}], ",", 
     RowBox[{"PlotRange", "\[Rule]", "All"}], ",", 
     RowBox[{"AxesLabel", "\[Rule]", 
      RowBox[{"{", 
       RowBox[{"\"\<t\>\"", ",", 
        RowBox[{"\"\<x\>\"", "<>", 
         RowBox[{"ToString", "[", "i", "]"}]}]}], "}"}]}]}], "]"}], ",", 
   RowBox[{"{", 
    RowBox[{"i", ",", "m"}], "}"}]}], "]"}]], "Input",
 CellLabel->"In[5]:=",
 CellID->111189838],

Cell[BoxData[
 GraphicsBox[
  TagBox[RasterBox[CompressedData["
1:eJztXV2vLMdVvYJfwt+wBG05fCRE4JgPnQcIDw5cJS+cB8hL8kAiPvwC0lGC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    "], {{0, 131}, {576, 0}}, {0, 255},
    ColorFunction->RGBColor],
   BoxForm`ImageTag["Byte", ColorSpace -> "RGB", Interleaving -> True],
   Selectable->False],
  BaseStyle->"ImageGraphics",
  ImageSize->Automatic,
  ImageSizeRaw->{576, 131},
  PlotRange->{{0, 576}, {0, 131}}]], "Output",
 ImageSize->{580, 131},
 ImageMargins->{{0, 0}, {0, 0}},
 ImageRegion->{{0, 1}, {0, 1}},
 CellLabel->"Out[5]=",
 CellID->376985250]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[{
 RowBox[{"Clear", "[", 
  RowBox[{"data", ",", "x", ",", "his", ",", "f"}], 
  "]"}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{"data", "=", 
   RowBox[{"Transpose", "[", 
    RowBox[{"res", "[", 
     RowBox[{"[", 
      RowBox[{
       RowBox[{"adpmax", ";;"}], ",", "1"}], "]"}], "]"}], "]"}]}], 
  ";"}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{"his", "=", 
   RowBox[{
    RowBox[{
     RowBox[{"Histogram", "[", 
      RowBox[{"#", ",", "Automatic", ",", "\"\<ProbabilityDensity\>\""}], 
      "]"}], "&"}], "/@", "data"}]}], ";"}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{"f", "=", 
   RowBox[{
    RowBox[{"logpdf", "[", 
     RowBox[{
      RowBox[{"{", 
       RowBox[{"x", ",", "x", ",", "x"}], "}"}], ",", "All"}], "]"}], "-", 
    RowBox[{"Log", "[", 
     RowBox[{"sd", " ", 
      RowBox[{"Sqrt", "[", 
       RowBox[{"2", " ", "Pi"}], "]"}]}], "]"}]}]}], 
  ";"}], "\[IndentingNewLine]", 
 RowBox[{"Table", "[", 
  RowBox[{
   RowBox[{"Show", "[", 
    RowBox[{
     RowBox[{"his", "[", 
      RowBox[{"[", "i", "]"}], "]"}], ",", 
     RowBox[{"Plot", "[", 
      RowBox[{
       RowBox[{"Exp", "[", 
        RowBox[{"f", "[", 
         RowBox[{"[", "i", "]"}], "]"}], "]"}], ",", 
       RowBox[{"{", 
        RowBox[{"x", ",", 
         RowBox[{"Min", "[", 
          RowBox[{"data", "[", 
           RowBox[{"[", "i", "]"}], "]"}], "]"}], ",", 
         RowBox[{"Max", "[", 
          RowBox[{"data", "[", 
           RowBox[{"[", "i", "]"}], "]"}], "]"}]}], "}"}], ",", 
       RowBox[{"PlotStyle", "\[Rule]", 
        RowBox[{"{", 
         RowBox[{"Red", ",", "Thick"}], "}"}]}], ",", 
       RowBox[{"PlotRange", "\[Rule]", "All"}], ",", 
       RowBox[{"AxesLabel", "\[Rule]", 
        RowBox[{"{", 
         RowBox[{
          RowBox[{"\"\<x\>\"", "<>", 
           RowBox[{"ToString", "[", "i", "]"}]}], ",", "\"\<freq\>\""}], 
         "}"}]}]}], "]"}]}], "]"}], ",", 
   RowBox[{"{", 
    RowBox[{"i", ",", "m"}], "}"}]}], "]"}]}], "Input",
 CellLabel->"In[6]:=",
 CellID->305510931],

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{
   GraphicsBox[{
     {RGBColor[0.798413061722744, 0.824719615472648, 0.968322270542458], 
      EdgeForm[Opacity[0.7]], {}, 
      {RGBColor[0.798413061722744, 0.824719615472648, 0.968322270542458], 
       EdgeForm[Opacity[0.7]], 
       TagBox[
        TooltipBox[
         TagBox[
          DynamicBox[{
            FEPrivate`If[
             CurrentValue["MouseOver"], 
             EdgeForm[{
               GrayLevel[0.5], 
               AbsoluteThickness[1.5], 
               Opacity[0.66]}], {}, {}], 
            
            RectangleBox[{26.8, 0.}, {27., 0.0024996875390576262`}, 
             "RoundingRadius" -> 0]},
           
           ImageSizeCache->{{20.5, 23.774669867947182`}, {37.35267961221971, 
            39.270267589618726`}}],
          StatusArea[#, 0.0024996875390576262`]& ,
          TagBoxNote->"0.0024996875390576262"],
         StyleBox["0.0024996875390576262`", {
           GrayLevel[0]}, StripOnInput -> False]],
        Annotation[#, 
         Style[0.0024996875390576262`, {
           GrayLevel[0]}], "Tooltip"]& ], {}, 
       TagBox[
        TooltipBox[
         TagBox[
          DynamicBox[{
            FEPrivate`If[
             CurrentValue["MouseOver"], 
             EdgeForm[{
               GrayLevel[0.5], 
               AbsoluteThickness[1.5], 
               Opacity[0.66]}], {}, {}], 
            
            RectangleBox[{27.2, 0.}, {27.4, 0.00187476565429322}, 
             "RoundingRadius" -> 0]},
           
           ImageSizeCache->{{25.049339735894364`, 28.324009603841546`}, {
            37.582076606569466`, 39.270267589618726`}}],
          StatusArea[#, 0.00187476565429322]& ,
          TagBoxNote->"0.00187476565429322"],
         StyleBox["0.00187476565429322`", {
           GrayLevel[0]}, StripOnInput -> False]],
        Annotation[#, 
         Style[0.00187476565429322, {
           GrayLevel[0]}], "Tooltip"]& ], {}, 
       TagBox[
        TooltipBox[
         TagBox[
          DynamicBox[{
            FEPrivate`If[
             CurrentValue["MouseOver"], 
             EdgeForm[{
               GrayLevel[0.5], 
               AbsoluteThickness[1.5], 
               Opacity[0.66]}], {}, {}], 
            
            RectangleBox[{27.6, 0.}, {27.8, 0.00187476565429322}, 
             "RoundingRadius" -> 0]},
           
           ImageSizeCache->{{29.598679471788728`, 32.87334933973591}, {
            37.582076606569466`, 39.270267589618726`}}],
          StatusArea[#, 0.00187476565429322]& ,
          TagBoxNote->"0.00187476565429322"],
         StyleBox["0.00187476565429322`", {
           GrayLevel[0]}, StripOnInput -> False]],
        Annotation[#, 
         Style[0.00187476565429322, {
           GrayLevel[0]}], "Tooltip"]& ], 
       TagBox[
        TooltipBox[
         TagBox[
          DynamicBox[{
            FEPrivate`If[
             CurrentValue["MouseOver"], 
             EdgeForm[{
               GrayLevel[0.5], 
               AbsoluteThickness[1.5], 
               Opacity[0.66]}], {}, {}], 
            
            RectangleBox[{27.8, 0.}, {28., 0.0024996875390576262`}, 
             "RoundingRadius" -> 0]},
           
           ImageSizeCache->{{31.87334933973591, 35.14801920768309}, {
            37.35267961221971, 39.270267589618726`}}],
          StatusArea[#, 0.0024996875390576262`]& ,
          TagBoxNote->"0.0024996875390576262"],
         StyleBox["0.0024996875390576262`", {
           GrayLevel[0]}, StripOnInput -> False]],
        Annotation[#, 
         Style[0.0024996875390576262`, {
           GrayLevel[0]}], "Tooltip"]& ], 
       TagBox[
        TooltipBox[
         TagBox[
          DynamicBox[{
            FEPrivate`If[
             CurrentValue["MouseOver"], 
             EdgeForm[{
               GrayLevel[0.5], 
               AbsoluteThickness[1.5], 
               Opacity[0.66]}], {}, {}], 
            
            RectangleBox[{28., 0.}, {28.2, 0.0024996875390576262`}, 
             "RoundingRadius" -> 0]},
           
           ImageSizeCache->{{34.14801920768309, 37.422689075630274`}, {
            37.35267961221971, 39.270267589618726`}}],
          StatusArea[#, 0.0024996875390576262`]& ,
          TagBoxNote->"0.0024996875390576262"],
         StyleBox["0.0024996875390576262`", {
           GrayLevel[0]}, StripOnInput -> False]],
        Annotation[#, 
         Style[0.0024996875390576262`, {
           GrayLevel[0]}], "Tooltip"]& ], 
       TagBox[
        TooltipBox[
         TagBox[
          DynamicBox[{
            FEPrivate`If[
             CurrentValue["MouseOver"], 
             EdgeForm[{
               GrayLevel[0.5], 
               AbsoluteThickness[1.5], 
               Opacity[0.66]}], {}, {}], 
            
            RectangleBox[{28.2, 0.}, {28.4, 0.0024996875390576262`}, 
             "RoundingRadius" -> 0]},
           
           ImageSizeCache->{{36.422689075630274`, 39.697358943577456`}, {
            37.35267961221971, 39.270267589618726`}}],
          StatusArea[#, 0.0024996875390576262`]& ,
          TagBoxNote->"0.0024996875390576262"],
         StyleBox["0.0024996875390576262`", {
           GrayLevel[0]}, StripOnInput -> False]],
        Annotation[#, 
         Style[0.0024996875390576262`, {
           GrayLevel[0]}], "Tooltip"]& ], {}, 
       TagBox[
        TooltipBox[
         TagBox[
          DynamicBox[{
            FEPrivate`If[
             CurrentValue["MouseOver"], 
             EdgeForm[{
               GrayLevel[0.5], 
               AbsoluteThickness[1.5], 
               Opacity[0.66]}], {}, {}], 
            
            RectangleBox[{28.6, 0.}, {28.8, 0.0012498437695288131`}, 
             "RoundingRadius" -> 0]},
           
           ImageSizeCache->{{40.97202881152464, 44.24669867947182}, {
            37.81147360091922, 39.270267589618726`}}],
          StatusArea[#, 0.0012498437695288131`]& ,
          TagBoxNote->"0.0012498437695288131"],
         StyleBox["0.0012498437695288131`", {
           GrayLevel[0]}, StripOnInput -> False]],
        Annotation[#, 
         Style[0.0012498437695288131`, {
           GrayLevel[0]}], "Tooltip"]& ], 
       TagBox[
        TooltipBox[
         TagBox[
          DynamicBox[{
            FEPrivate`If[
             CurrentValue["MouseOver"], 
             EdgeForm[{
               GrayLevel[0.5], 
               AbsoluteThickness[1.5], 
               Opacity[0.66]}], {}, {}], 
            
            RectangleBox[{28.8, 0.}, {29., 0.0024996875390576262`}, 
             "RoundingRadius" -> 0]},
           
           ImageSizeCache->{{43.24669867947182, 46.521368547419}, {
            37.35267961221971, 39.270267589618726`}}],
          StatusArea[#, 0.0024996875390576262`]& ,
          TagBoxNote->"0.0024996875390576262"],
         StyleBox["0.0024996875390576262`", {
           GrayLevel[0]}, StripOnInput -> False]],
        Annotation[#, 
         Style[0.0024996875390576262`, {
           GrayLevel[0]}], "Tooltip"]& ], 
       TagBox[
        TooltipBox[
         TagBox[
          DynamicBox[{
            FEPrivate`If[
             CurrentValue["MouseOver"], 
             EdgeForm[{
               GrayLevel[0.5], 
               AbsoluteThickness[1.5], 
               Opacity[0.66]}], {}, {}], 
            
            RectangleBox[{29., 0.}, {29.2, 0.006249218847644067}, 
             "RoundingRadius" -> 0]},
           
           ImageSizeCache->{{45.521368547419, 48.79603841536618}, {
            35.97629764612119, 39.270267589618726`}}],
          StatusArea[#, 0.006249218847644067]& ,
          TagBoxNote->"0.006249218847644067"],
         StyleBox["0.006249218847644067`", {
           GrayLevel[0]}, StripOnInput -> False]],
        Annotation[#, 
         Style[0.006249218847644067, {
           GrayLevel[0]}], "Tooltip"]& ], 
       TagBox[
        TooltipBox[
         TagBox[
          DynamicBox[{
            FEPrivate`If[
             CurrentValue["MouseOver"], 
             EdgeForm[{
               GrayLevel[0.5], 
               AbsoluteThickness[1.5], 
               Opacity[0.66]}], {}, {}], 
            
            RectangleBox[{29.2, 0.}, {29.4, 0.0093738282714661}, 
             "RoundingRadius" -> 0]},
           
           ImageSizeCache->{{47.79603841536618, 51.07070828331331}, {
            34.82931267437242, 39.270267589618726`}}],
          StatusArea[#, 0.0093738282714661]& ,
          TagBoxNote->"0.0093738282714661"],
         StyleBox["0.0093738282714661`", {
           GrayLevel[0]}, StripOnInput -> False]],
        Annotation[#, 
         Style[0.0093738282714661, {
           GrayLevel[0]}], "Tooltip"]& ], 
       TagBox[
        TooltipBox[
         TagBox[
          DynamicBox[{
            FEPrivate`If[
             CurrentValue["MouseOver"], 
             EdgeForm[{
               GrayLevel[0.5], 
               AbsoluteThickness[1.5], 
               Opacity[0.66]}], {}, {}], 
            
            RectangleBox[{29.4, 0.}, {29.6, 0.014373203349581098`}, 
             "RoundingRadius" -> 0]},
           
           ImageSizeCache->{{50.07070828331331, 53.34537815126055}, {
            32.99413671957448, 39.270267589618726`}}],
          StatusArea[#, 0.014373203349581098`]& ,
          TagBoxNote->"0.014373203349581098"],
         StyleBox["0.014373203349581098`", {
           GrayLevel[0]}, StripOnInput -> False]],
        Annotation[#, 
         Style[0.014373203349581098`, {
           GrayLevel[0]}], "Tooltip"]& ], 
       TagBox[
        TooltipBox[
         TagBox[
          DynamicBox[{
            FEPrivate`If[
             CurrentValue["MouseOver"], 
             EdgeForm[{
               GrayLevel[0.5], 
               AbsoluteThickness[1.5], 
               Opacity[0.66]}], {}, {}], 
            
            RectangleBox[{29.6, 0.}, {29.8, 0.016247969003874575`}, 
             "RoundingRadius" -> 0]},
           
           ImageSizeCache->{{52.34537815126055, 55.62004801920773}, {
            32.305945736525125`, 39.270267589618726`}}],
          StatusArea[#, 0.016247969003874575`]& ,
          TagBoxNote->"0.016247969003874575"],
         StyleBox["0.016247969003874575`", {
           GrayLevel[0]}, StripOnInput -> False]],
        Annotation[#, 
         Style[0.016247969003874575`, {
           GrayLevel[0]}], "Tooltip"]& ], 
       TagBox[
        TooltipBox[
         TagBox[
          DynamicBox[{
            FEPrivate`If[
             CurrentValue["MouseOver"], 
             EdgeForm[{
               GrayLevel[0.5], 
               AbsoluteThickness[1.5], 
               Opacity[0.66]}], {}, {}], 
            
            RectangleBox[{29.8, 0.}, {30., 0.01312335958005254}, 
             "RoundingRadius" -> 0]},
           
           ImageSizeCache->{{54.62004801920773, 57.894717887154854`}, {
            33.4529307082739, 39.270267589618726`}}],
          StatusArea[#, 0.01312335958005254]& ,
          TagBoxNote->"0.01312335958005254"],
         StyleBox["0.01312335958005254`", {
           GrayLevel[0]}, StripOnInput -> False]],
        Annotation[#, 
         Style[0.01312335958005254, {
           GrayLevel[0]}], "Tooltip"]& ], 
       TagBox[
        TooltipBox[
         TagBox[
          DynamicBox[{
            FEPrivate`If[
             CurrentValue["MouseOver"], 
             EdgeForm[{
               GrayLevel[0.5], 
               AbsoluteThickness[1.5], 
               Opacity[0.66]}], {}, {}], 
            
            RectangleBox[{30., 0.}, {30.2, 0.02624671916010508}, 
             "RoundingRadius" -> 0]},
           
           ImageSizeCache->{{56.894717887154854`, 60.169387755102036`}, {
            28.63559382692906, 39.270267589618726`}}],
          StatusArea[#, 0.02624671916010508]& ,
          TagBoxNote->"0.02624671916010508"],
         StyleBox["0.02624671916010508`", {
           GrayLevel[0]}, StripOnInput -> False]],
        Annotation[#, 
         Style[0.02624671916010508, {
           GrayLevel[0]}], "Tooltip"]& ], 
       TagBox[
        TooltipBox[
         TagBox[
          DynamicBox[{
            FEPrivate`If[
             CurrentValue["MouseOver"], 
             EdgeForm[{
               GrayLevel[0.5], 
               AbsoluteThickness[1.5], 
               Opacity[0.66]}], {}, {}], 
            
            RectangleBox[{30.2, 0.}, {30.4, 0.040619922509686435`}, 
             "RoundingRadius" -> 0]},
           
           ImageSizeCache->{{59.169387755102036`, 62.44405762304922}, {
            23.35946295688472, 39.270267589618726`}}],
          StatusArea[#, 0.040619922509686435`]& ,
          TagBoxNote->"0.040619922509686435"],
         StyleBox["0.040619922509686435`", {
           GrayLevel[0]}, StripOnInput -> False]],
        Annotation[#, 
         Style[0.040619922509686435`, {
           GrayLevel[0]}], "Tooltip"]& ], 
       TagBox[
        TooltipBox[
         TagBox[
          DynamicBox[{
            FEPrivate`If[
             CurrentValue["MouseOver"], 
             EdgeForm[{
               GrayLevel[0.5], 
               AbsoluteThickness[1.5], 
               Opacity[0.66]}], {}, {}], 
            
            RectangleBox[{30.4, 0.}, {30.6, 0.05436820397450241}, 
             "RoundingRadius" -> 0]},
           
           ImageSizeCache->{{61.44405762304922, 64.71872749099646}, {
            18.31272908119049, 39.270267589618726`}}],
          StatusArea[#, 0.05436820397450241]& ,
          TagBoxNote->"0.05436820397450241"],
         StyleBox["0.05436820397450241`", {
           GrayLevel[0]}, StripOnInput -> False]],
        Annotation[#, 
         Style[0.05436820397450241, {
           GrayLevel[0]}], "Tooltip"]& ], 
       TagBox[
        TooltipBox[
         TagBox[
          DynamicBox[{
            FEPrivate`If[
             CurrentValue["MouseOver"], 
             EdgeForm[{
               GrayLevel[0.5], 
               AbsoluteThickness[1.5], 
               Opacity[0.66]}], {}, {}], 
            
            RectangleBox[{30.6, 0.}, {30.8, 0.03124609423822033}, 
             "RoundingRadius" -> 0]},
           
           ImageSizeCache->{{63.71872749099646, 66.99339735894358}, {
            26.80041787213103, 39.270267589618726`}}],
          StatusArea[#, 0.03124609423822033]& ,
          TagBoxNote->"0.03124609423822033"],
         StyleBox["0.03124609423822033`", {
           GrayLevel[0]}, StripOnInput -> False]],
        Annotation[#, 
         Style[0.03124609423822033, {
           GrayLevel[0]}], "Tooltip"]& ], 
       TagBox[
        TooltipBox[
         TagBox[
          DynamicBox[{
            FEPrivate`If[
             CurrentValue["MouseOver"], 
             EdgeForm[{
               GrayLevel[0.5], 
               AbsoluteThickness[1.5], 
               Opacity[0.66]}], {}, {}], 
            
            RectangleBox[{30.8, 0.}, {31., 0.05686789151356101}, 
             "RoundingRadius" -> 0]},
           
           ImageSizeCache->{{65.99339735894358, 69.26806722689076}, {
            17.39514110379112, 39.270267589618726`}}],
          StatusArea[#, 0.05686789151356101]& ,
          TagBoxNote->"0.05686789151356101"],
         StyleBox["0.05686789151356101`", {
           GrayLevel[0]}, StripOnInput -> False]],
        Annotation[#, 
         Style[0.05686789151356101, {
           GrayLevel[0]}], "Tooltip"]& ], 
       TagBox[
        TooltipBox[
         TagBox[
          DynamicBox[{
            FEPrivate`If[
             CurrentValue["MouseOver"], 
             EdgeForm[{
               GrayLevel[0.5], 
               AbsoluteThickness[1.5], 
               Opacity[0.66]}], {}, {}], 
            
            RectangleBox[{31., 0.}, {31.2, 0.05936757905261863}, 
             "RoundingRadius" -> 0]},
           
           ImageSizeCache->{{68.26806722689076, 71.54273709483795}, {
            16.477553126392106`, 39.270267589618726`}}],
          StatusArea[#, 0.05936757905261863]& ,
          TagBoxNote->"0.05936757905261863"],
         StyleBox["0.05936757905261863`", {
           GrayLevel[0]}, StripOnInput -> False]],
        Annotation[#, 
         Style[0.05936757905261863, {
           GrayLevel[0]}], "Tooltip"]& ], 
       TagBox[
        TooltipBox[
         TagBox[
          DynamicBox[{
            FEPrivate`If[
             CurrentValue["MouseOver"], 
             EdgeForm[{
               GrayLevel[0.5], 
               AbsoluteThickness[1.5], 
               Opacity[0.66]}], {}, {}], 
            
            RectangleBox[{31.2, 0.}, {31.4, 0.08748906386701694}, 
             "RoundingRadius" -> 0]},
           
           ImageSizeCache->{{70.54273709483795, 73.81740696278513}, {
            6.154688380653177, 39.270267589618726`}}],
          StatusArea[#, 0.08748906386701694]& ,
          TagBoxNote->"0.08748906386701694"],
         StyleBox["0.08748906386701694`", {
           GrayLevel[0]}, StripOnInput -> False]],
        Annotation[#, 
         Style[0.08748906386701694, {
           GrayLevel[0]}], "Tooltip"]& ], 
       TagBox[
        TooltipBox[
         TagBox[
          DynamicBox[{
            FEPrivate`If[
             CurrentValue["MouseOver"], 
             EdgeForm[{
               GrayLevel[0.5], 
               AbsoluteThickness[1.5], 
               Opacity[0.66]}], {}, {}], 
            
            RectangleBox[{31.4, 0.}, {31.6, 0.1006124234470677}, 
             "RoundingRadius" -> 0]},
           
           ImageSizeCache->{{72.81740696278513, 76.09207683073231}, {
            1.3373514993090012`, 39.270267589618726`}}],
          StatusArea[#, 0.1006124234470677]& ,
          TagBoxNote->"0.1006124234470677"],
         StyleBox["0.1006124234470677`", {
           GrayLevel[0]}, StripOnInput -> False]],
        Annotation[#, 
         Style[0.1006124234470677, {
           GrayLevel[0]}], "Tooltip"]& ], 
       TagBox[
        TooltipBox[
         TagBox[
          DynamicBox[{
            FEPrivate`If[
             CurrentValue["MouseOver"], 
             EdgeForm[{
               GrayLevel[0.5], 
               AbsoluteThickness[1.5], 
               Opacity[0.66]}], {}, {}], 
            
            RectangleBox[{31.6, 0.}, {31.8, 0.12123484564429489`}, 
             "RoundingRadius" -> 0]},
           
           ImageSizeCache->{{75.09207683073231, 
            78.36674669867949}, {-6.23274931423353, 39.270267589618726`}}],
          StatusArea[#, 0.12123484564429489`]& ,
          TagBoxNote->"0.12123484564429489"],
         StyleBox["0.12123484564429489`", {
           GrayLevel[0]}, StripOnInput -> False]],
        Annotation[#, 
         Style[0.12123484564429489`, {
           GrayLevel[0]}], "Tooltip"]& ], 
       TagBox[
        TooltipBox[
         TagBox[
          DynamicBox[{
            FEPrivate`If[
             CurrentValue["MouseOver"], 
             EdgeForm[{
               GrayLevel[0.5], 
               AbsoluteThickness[1.5], 
               Opacity[0.66]}], {}, {}], 
            
            RectangleBox[{31.8, 0.}, {32., 0.14373203349581354`}, 
             "RoundingRadius" -> 0]},
           
           ImageSizeCache->{{77.36674669867949, 
            80.64141656662667}, {-14.491041110824675`, 39.270267589618726`}}],
          StatusArea[#, 0.14373203349581354`]& ,
          TagBoxNote->"0.14373203349581354"],
         StyleBox["0.14373203349581354`", {
           GrayLevel[0]}, StripOnInput -> False]],
        Annotation[#, 
         Style[0.14373203349581354`, {
           GrayLevel[0]}], "Tooltip"]& ], 
       TagBox[
        TooltipBox[
         TagBox[
          DynamicBox[{
            FEPrivate`If[
             CurrentValue["MouseOver"], 
             EdgeForm[{
               GrayLevel[0.5], 
               AbsoluteThickness[1.5], 
               Opacity[0.66]}], {}, {}], 
            
            RectangleBox[{32., 0.}, {32.2, 0.1131108611423556}, 
             "RoundingRadius" -> 0]},
           
           ImageSizeCache->{{79.64141656662667, 
            82.91608643457391}, {-3.250588387685994, 39.270267589618726`}}],
          StatusArea[#, 0.1131108611423556]& ,
          TagBoxNote->"0.1131108611423556"],
         StyleBox["0.1131108611423556`", {
           GrayLevel[0]}, StripOnInput -> False]],
        Annotation[#, 
         Style[0.1131108611423556, {
           GrayLevel[0]}], "Tooltip"]& ], 
       TagBox[
        TooltipBox[
         TagBox[
          DynamicBox[{
            FEPrivate`If[
             CurrentValue["MouseOver"], 
             EdgeForm[{
               GrayLevel[0.5], 
               AbsoluteThickness[1.5], 
               Opacity[0.66]}], {}, {}], 
            
            RectangleBox[{32.2, 0.}, {32.4, 0.13935758030246517`}, 
             "RoundingRadius" -> 0]},
           
           ImageSizeCache->{{81.91608643457391, 
            85.19075630252104}, {-12.885262150377308`, 39.270267589618726`}}],
          StatusArea[#, 0.13935758030246517`]& ,
          TagBoxNote->"0.13935758030246517"],
         StyleBox["0.13935758030246517`", {
           GrayLevel[0]}, StripOnInput -> False]],
        Annotation[#, 
         Style[0.13935758030246517`, {
           GrayLevel[0]}], "Tooltip"]& ], 
       TagBox[
        TooltipBox[
         TagBox[
          DynamicBox[{
            FEPrivate`If[
             CurrentValue["MouseOver"], 
             EdgeForm[{
               GrayLevel[0.5], 
               AbsoluteThickness[1.5], 
               Opacity[0.66]}], {}, {}], 
            RectangleBox[{32.4, 0.}, {32.6, 0.1687289088863868}, 
             "RoundingRadius" -> 0]},
           
           ImageSizeCache->{{84.19075630252104, 
            87.46542617046822}, {-23.666920884813734`, 39.270267589618726`}}],
          StatusArea[#, 0.1687289088863868]& ,
          TagBoxNote->"0.1687289088863868"],
         StyleBox["0.1687289088863868`", {
           GrayLevel[0]}, StripOnInput -> False]],
        Annotation[#, 
         Style[0.1687289088863868, {
           GrayLevel[0]}], "Tooltip"]& ], 
       TagBox[
        TooltipBox[
         TagBox[
          DynamicBox[{
            FEPrivate`If[
             CurrentValue["MouseOver"], 
             EdgeForm[{
               GrayLevel[0.5], 
               AbsoluteThickness[1.5], 
               Opacity[0.66]}], {}, {}], 
            
            RectangleBox[{32.6, 0.}, {32.8, 0.19497562804649834`}, 
             "RoundingRadius" -> 0]},
           
           ImageSizeCache->{{86.46542617046822, 
            89.74009603841534}, {-33.301594647505766`, 39.270267589618726`}}],
          StatusArea[#, 0.19497562804649834`]& ,
          TagBoxNote->"0.19497562804649834"],
         StyleBox["0.19497562804649834`", {
           GrayLevel[0]}, StripOnInput -> False]],
        Annotation[#, 
         Style[0.19497562804649834`, {
           GrayLevel[0]}], "Tooltip"]& ], 
       TagBox[
        TooltipBox[
         TagBox[
          DynamicBox[{
            FEPrivate`If[
             CurrentValue["MouseOver"], 
             EdgeForm[{
               GrayLevel[0.5], 
               AbsoluteThickness[1.5], 
               Opacity[0.66]}], {}, {}], 
            
            RectangleBox[{32.8, 0.}, {33., 0.17060367454067998`}, 
             "RoundingRadius" -> 0]},
           
           ImageSizeCache->{{88.74009603841534, 
            92.01476590636258}, {-24.355111867862973`, 39.270267589618726`}}],
          StatusArea[#, 0.17060367454067998`]& ,
          TagBoxNote->"0.17060367454067998"],
         StyleBox["0.17060367454067998`", {
           GrayLevel[0]}, StripOnInput -> False]],
        Annotation[#, 
         Style[0.17060367454067998`, {
           GrayLevel[0]}], "Tooltip"]& ], 
       TagBox[
        TooltipBox[
         TagBox[
          DynamicBox[{
            FEPrivate`If[
             CurrentValue["MouseOver"], 
             EdgeForm[{
               GrayLevel[0.5], 
               AbsoluteThickness[1.5], 
               Opacity[0.66]}], {}, {}], 
            
            RectangleBox[{33., 0.}, {33.2, 0.23122109736282634`}, 
             "RoundingRadius" -> 0]},
           
           ImageSizeCache->{{91.01476590636258, 
            94.28943577430977}, {-46.6066203197887, 39.270267589618726`}}],
          StatusArea[#, 0.23122109736282634`]& ,
          TagBoxNote->"0.23122109736282634"],
         StyleBox["0.23122109736282634`", {
           GrayLevel[0]}, StripOnInput -> False]],
        Annotation[#, 
         Style[0.23122109736282634`, {
           GrayLevel[0]}], "Tooltip"]& ], 
       TagBox[
        TooltipBox[
         TagBox[
          DynamicBox[{
            FEPrivate`If[
             CurrentValue["MouseOver"], 
             EdgeForm[{
               GrayLevel[0.5], 
               AbsoluteThickness[1.5], 
               Opacity[0.66]}], {}, {}], 
            
            RectangleBox[{33.2, 0.}, {33.4, 0.22622172228471923`}, 
             "RoundingRadius" -> 0]},
           
           ImageSizeCache->{{93.28943577430977, 
            96.56410564225689}, {-44.77144436499367, 39.270267589618726`}}],
          StatusArea[#, 0.22622172228471923`]& ,
          TagBoxNote->"0.22622172228471923"],
         StyleBox["0.22622172228471923`", {
           GrayLevel[0]}, StripOnInput -> False]],
        Annotation[#, 
         Style[0.22622172228471923`, {
           GrayLevel[0]}], "Tooltip"]& ], 
       TagBox[
        TooltipBox[
         TagBox[
          DynamicBox[{
            FEPrivate`If[
             CurrentValue["MouseOver"], 
             EdgeForm[{
               GrayLevel[0.5], 
               AbsoluteThickness[1.5], 
               Opacity[0.66]}], {}, {}], 
            
            RectangleBox[{33.4, 0.}, {33.6, 0.22497187851518238`}, 
             "RoundingRadius" -> 0]},
           
           ImageSizeCache->{{95.56410564225689, 
            98.83877551020413}, {-44.3126503762912, 39.270267589618726`}}],
          StatusArea[#, 0.22497187851518238`]& ,
          TagBoxNote->"0.22497187851518238"],
         StyleBox["0.22497187851518238`", {
           GrayLevel[0]}, StripOnInput -> False]],
        Annotation[#, 
         Style[0.22497187851518238`, {
           GrayLevel[0]}], "Tooltip"]& ], 
       TagBox[
        TooltipBox[
         TagBox[
          DynamicBox[{
            FEPrivate`If[
             CurrentValue["MouseOver"], 
             EdgeForm[{
               GrayLevel[0.5], 
               AbsoluteThickness[1.5], 
               Opacity[0.66]}], {}, {}], 
            
            RectangleBox[{33.6, 0.}, {33.8, 0.20622422197225787`}, 
             "RoundingRadius" -> 0]},
           
           ImageSizeCache->{{97.83877551020413, 
            101.11344537815125`}, {-37.43074054580141, 39.270267589618726`}}],
          StatusArea[#, 0.20622422197225787`]& ,
          TagBoxNote->"0.20622422197225787"],
         StyleBox["0.20622422197225787`", {
           GrayLevel[0]}, StripOnInput -> False]],
        Annotation[#, 
         Style[0.20622422197225787`, {
           GrayLevel[0]}], "Tooltip"]& ], 
       TagBox[
        TooltipBox[
         TagBox[
          DynamicBox[{
            FEPrivate`If[
             CurrentValue["MouseOver"], 
             EdgeForm[{
               GrayLevel[0.5], 
               AbsoluteThickness[1.5], 
               Opacity[0.66]}], {}, {}], 
            RectangleBox[{33.8, 0.}, {34., 0.21559805024371648`}, 
             "RoundingRadius" -> 0]},
           
           ImageSizeCache->{{100.11344537815125`, 
            103.3881152460985}, {-40.871695461044965`, 39.270267589618726`}}],
          StatusArea[#, 0.21559805024371648`]& ,
          TagBoxNote->"0.21559805024371648"],
         StyleBox["0.21559805024371648`", {
           GrayLevel[0]}, StripOnInput -> False]],
        Annotation[#, 
         Style[0.21559805024371648`, {
           GrayLevel[0]}], "Tooltip"]& ], 
       TagBox[
        TooltipBox[
         TagBox[
          DynamicBox[{
            FEPrivate`If[
             CurrentValue["MouseOver"], 
             EdgeForm[{
               GrayLevel[0.5], 
               AbsoluteThickness[1.5], 
               Opacity[0.66]}], {}, {}], 
            
            RectangleBox[{34., 0.}, {34.2, 0.19310086239219826`}, 
             "RoundingRadius" -> 0]},
           
           ImageSizeCache->{{102.3881152460985, 
            105.66278511404568`}, {-32.61340366445398, 39.270267589618726`}}],
          StatusArea[#, 0.19310086239219826`]& ,
          TagBoxNote->"0.19310086239219826"],
         StyleBox["0.19310086239219826`", {
           GrayLevel[0]}, StripOnInput -> False]],
        Annotation[#, 
         Style[0.19310086239219826`, {
           GrayLevel[0]}], "Tooltip"]& ], 
       TagBox[
        TooltipBox[
         TagBox[
          DynamicBox[{
            FEPrivate`If[
             CurrentValue["MouseOver"], 
             EdgeForm[{
               GrayLevel[0.5], 
               AbsoluteThickness[1.5], 
               Opacity[0.66]}], {}, {}], 
            
            RectangleBox[{34.2, 0.}, {34.4, 0.23872015998000756`}, 
             "RoundingRadius" -> 0]},
           
           ImageSizeCache->{{104.66278511404568`, 
            107.9374549819928}, {-49.359384251988814`, 39.270267589618726`}}],
          StatusArea[#, 0.23872015998000756`]& ,
          TagBoxNote->"0.23872015998000756"],
         StyleBox["0.23872015998000756`", {
           GrayLevel[0]}, StripOnInput -> False]],
        Annotation[#, 
         Style[0.23872015998000756`, {
           GrayLevel[0]}], "Tooltip"]& ], 
       TagBox[
        TooltipBox[
         TagBox[
          DynamicBox[{
            FEPrivate`If[
             CurrentValue["MouseOver"], 
             EdgeForm[{
               GrayLevel[0.5], 
               AbsoluteThickness[1.5], 
               Opacity[0.66]}], {}, {}], 
            
            RectangleBox[{34.4, 0.}, {34.6, 0.20747406574177935`}, 
             "RoundingRadius" -> 0]},
           
           ImageSizeCache->{{106.9374549819928, 
            110.21212484994004`}, {-37.889534534498225`, 
            39.270267589618726`}}],
          StatusArea[#, 0.20747406574177935`]& ,
          TagBoxNote->"0.20747406574177935"],
         StyleBox["0.20747406574177935`", {
           GrayLevel[0]}, StripOnInput -> False]],
        Annotation[#, 
         Style[0.20747406574177935`, {
           GrayLevel[0]}], "Tooltip"]& ], 
       TagBox[
        TooltipBox[
         TagBox[
          DynamicBox[{
            FEPrivate`If[
             CurrentValue["MouseOver"], 
             EdgeForm[{
               GrayLevel[0.5], 
               AbsoluteThickness[1.5], 
               Opacity[0.66]}], {}, {}], 
            
            RectangleBox[{34.6, 0.}, {34.8, 0.17997750281215233`}, 
             "RoundingRadius" -> 0]},
           
           ImageSizeCache->{{109.21212484994004`, 
            112.48679471788716`}, {-27.796066783111584`, 
            39.270267589618726`}}],
          StatusArea[#, 0.17997750281215233`]& ,
          TagBoxNote->"0.17997750281215233"],
         StyleBox["0.17997750281215233`", {
           GrayLevel[0]}, StripOnInput -> False]],
        Annotation[#, 
         Style[0.17997750281215233`, {
           GrayLevel[0]}], "Tooltip"]& ], 
       TagBox[
        TooltipBox[
         TagBox[
          DynamicBox[{
            FEPrivate`If[
             CurrentValue["MouseOver"], 
             EdgeForm[{
               GrayLevel[0.5], 
               AbsoluteThickness[1.5], 
               Opacity[0.66]}], {}, {}], 
            
            RectangleBox[{34.8, 0.}, {35., 0.1556055493063345}, 
             "RoundingRadius" -> 0]},
           
           ImageSizeCache->{{111.48679471788716`, 
            114.76146458583435`}, {-18.849584003468983`, 
            39.270267589618726`}}],
          StatusArea[#, 0.1556055493063345]& ,
          TagBoxNote->"0.1556055493063345"],
         StyleBox["0.1556055493063345`", {
           GrayLevel[0]}, StripOnInput -> False]],
        Annotation[#, 
         Style[0.1556055493063345, {
           GrayLevel[0]}], "Tooltip"]& ], 
       TagBox[
        TooltipBox[
         TagBox[
          DynamicBox[{
            FEPrivate`If[
             CurrentValue["MouseOver"], 
             EdgeForm[{
               GrayLevel[0.5], 
               AbsoluteThickness[1.5], 
               Opacity[0.66]}], {}, {}], 
            
            RectangleBox[{35., 0.}, {35.2, 0.17060367454067998`}, 
             "RoundingRadius" -> 0]},
           
           ImageSizeCache->{{113.76146458583435`, 
            117.03613445378159`}, {-24.355111867862973`, 
            39.270267589618726`}}],
          StatusArea[#, 0.17060367454067998`]& ,
          TagBoxNote->"0.17060367454067998"],
         StyleBox["0.17060367454067998`", {
           GrayLevel[0]}, StripOnInput -> False]],
        Annotation[#, 
         Style[0.17060367454067998`, {
           GrayLevel[0]}], "Tooltip"]& ], 
       TagBox[
        TooltipBox[
         TagBox[
          DynamicBox[{
            FEPrivate`If[
             CurrentValue["MouseOver"], 
             EdgeForm[{
               GrayLevel[0.5], 
               AbsoluteThickness[1.5], 
               Opacity[0.66]}], {}, {}], 
            
            RectangleBox[{35.2, 0.}, {35.4, 0.1731033620797437}, 
             "RoundingRadius" -> 0]},
           
           ImageSizeCache->{{116.03613445378159`, 
            119.31080432172871`}, {-25.272699845264228`, 
            39.270267589618726`}}],
          StatusArea[#, 0.1731033620797437]& ,
          TagBoxNote->"0.1731033620797437"],
         StyleBox["0.1731033620797437`", {
           GrayLevel[0]}, StripOnInput -> False]],
        Annotation[#, 
         Style[0.1731033620797437, {
           GrayLevel[0]}], "Tooltip"]& ], 
       TagBox[
        TooltipBox[
         TagBox[
          DynamicBox[{
            FEPrivate`If[
             CurrentValue["MouseOver"], 
             EdgeForm[{
               GrayLevel[0.5], 
               AbsoluteThickness[1.5], 
               Opacity[0.66]}], {}, {}], 
            
            RectangleBox[{35.4, 0.}, {35.6, 0.14435695538057539`}, 
             "RoundingRadius" -> 0]},
           
           ImageSizeCache->{{118.31080432172871`, 
            121.58547418967589`}, {-14.720438105173493`, 
            39.270267589618726`}}],
          StatusArea[#, 0.14435695538057539`]& ,
          TagBoxNote->"0.14435695538057539"],
         StyleBox["0.14435695538057539`", {
           GrayLevel[0]}, StripOnInput -> False]],
        Annotation[#, 
         Style[0.14435695538057539`, {
           GrayLevel[0]}], "Tooltip"]& ], 
       TagBox[
        TooltipBox[
         TagBox[
          DynamicBox[{
            FEPrivate`If[
             CurrentValue["MouseOver"], 
             EdgeForm[{
               GrayLevel[0.5], 
               AbsoluteThickness[1.5], 
               Opacity[0.66]}], {}, {}], 
            
            RectangleBox[{35.6, 0.}, {35.8, 0.13935758030246517`}, 
             "RoundingRadius" -> 0]},
           
           ImageSizeCache->{{120.58547418967589`, 
            123.86014405762302`}, {-12.885262150377308`, 
            39.270267589618726`}}],
          StatusArea[#, 0.13935758030246517`]& ,
          TagBoxNote->"0.13935758030246517"],
         StyleBox["0.13935758030246517`", {
           GrayLevel[0]}, StripOnInput -> False]],
        Annotation[#, 
         Style[0.13935758030246517`, {
           GrayLevel[0]}], "Tooltip"]& ], 
       TagBox[
        TooltipBox[
         TagBox[
          DynamicBox[{
            FEPrivate`If[
             CurrentValue["MouseOver"], 
             EdgeForm[{
               GrayLevel[0.5], 
               AbsoluteThickness[1.5], 
               Opacity[0.66]}], {}, {}], 
            
            RectangleBox[{35.8, 0.}, {36., 0.09436320459942374}, 
             "RoundingRadius" -> 0]},
           
           ImageSizeCache->{{122.86014405762302`, 126.13481392557026`}, {
            3.6313214428064953`, 39.270267589618726`}}],
          StatusArea[#, 0.09436320459942374]& ,
          TagBoxNote->"0.09436320459942374"],
         StyleBox["0.09436320459942374`", {
           GrayLevel[0]}, StripOnInput -> False]],
        Annotation[#, 
         Style[0.09436320459942374, {
           GrayLevel[0]}], "Tooltip"]& ], 
       TagBox[
        TooltipBox[
         TagBox[
          DynamicBox[{
            FEPrivate`If[
             CurrentValue["MouseOver"], 
             EdgeForm[{
               GrayLevel[0.5], 
               AbsoluteThickness[1.5], 
               Opacity[0.66]}], {}, {}], 
            
            RectangleBox[{36., 0.}, {36.2, 0.1006124234470677}, 
             "RoundingRadius" -> 0]},
           
           ImageSizeCache->{{125.13481392557026`, 128.4094837935175}, {
            1.3373514993090012`, 39.270267589618726`}}],
          StatusArea[#, 0.1006124234470677]& ,
          TagBoxNote->"0.1006124234470677"],
         StyleBox["0.1006124234470677`", {
           GrayLevel[0]}, StripOnInput -> False]],
        Annotation[#, 
         Style[0.1006124234470677, {
           GrayLevel[0]}], "Tooltip"]& ], 
       TagBox[
        TooltipBox[
         TagBox[
          DynamicBox[{
            FEPrivate`If[
             CurrentValue["MouseOver"], 
             EdgeForm[{
               GrayLevel[0.5], 
               AbsoluteThickness[1.5], 
               Opacity[0.66]}], {}, {}], 
            
            RectangleBox[{36.2, 0.}, {36.4, 0.08623922009748965}, 
             "RoundingRadius" -> 0]},
           
           ImageSizeCache->{{127.4094837935175, 130.68415366146462`}, {
            6.613482369352127, 39.270267589618726`}}],
          StatusArea[#, 0.08623922009748965]& ,
          TagBoxNote->"0.08623922009748965"],
         StyleBox["0.08623922009748965`", {
           GrayLevel[0]}, StripOnInput -> False]],
        Annotation[#, 
         Style[0.08623922009748965, {
           GrayLevel[0]}], "Tooltip"]& ], 
       TagBox[
        TooltipBox[
         TagBox[
          DynamicBox[{
            FEPrivate`If[
             CurrentValue["MouseOver"], 
             EdgeForm[{
               GrayLevel[0.5], 
               AbsoluteThickness[1.5], 
               Opacity[0.66]}], {}, {}], 
            
            RectangleBox[{36.4, 0.}, {36.6, 0.07186601674790549}, 
             "RoundingRadius" -> 0]},
           
           ImageSizeCache->{{129.68415366146462`, 132.9588235294118}, {
            11.889613239397494`, 39.270267589618726`}}],
          StatusArea[#, 0.07186601674790549]& ,
          TagBoxNote->"0.07186601674790549"],
         StyleBox["0.07186601674790549`", {
           GrayLevel[0]}, StripOnInput -> False]],
        Annotation[#, 
         Style[0.07186601674790549, {
           GrayLevel[0]}], "Tooltip"]& ], 
       TagBox[
        TooltipBox[
         TagBox[
          DynamicBox[{
            FEPrivate`If[
             CurrentValue["MouseOver"], 
             EdgeForm[{
               GrayLevel[0.5], 
               AbsoluteThickness[1.5], 
               Opacity[0.66]}], {}, {}], 
            
            RectangleBox[{36.6, 0.}, {36.8, 0.046869141357331334`}, 
             "RoundingRadius" -> 0]},
           
           ImageSizeCache->{{131.9588235294118, 135.23349339735893`}, {
            21.065493013386877`, 39.270267589618726`}}],
          StatusArea[#, 0.046869141357331334`]& ,
          TagBoxNote->"0.046869141357331334"],
         StyleBox["0.046869141357331334`", {
           GrayLevel[0]}, StripOnInput -> False]],
        Annotation[#, 
         Style[0.046869141357331334`, {
           GrayLevel[0]}], "Tooltip"]& ], 
       TagBox[
        TooltipBox[
         TagBox[
          DynamicBox[{
            FEPrivate`If[
             CurrentValue["MouseOver"], 
             EdgeForm[{
               GrayLevel[0.5], 
               AbsoluteThickness[1.5], 
               Opacity[0.66]}], {}, {}], 
            
            RectangleBox[{36.8, 0.}, {37., 0.044994375703036486`}, 
             "RoundingRadius" -> 0]},
           
           ImageSizeCache->{{134.23349339735893`, 137.50816326530617`}, {
            21.753683996436735`, 39.270267589618726`}}],
          StatusArea[#, 0.044994375703036486`]& ,
          TagBoxNote->"0.044994375703036486"],
         StyleBox["0.044994375703036486`", {
           GrayLevel[0]}, StripOnInput -> False]],
        Annotation[#, 
         Style[0.044994375703036486`, {
           GrayLevel[0]}], "Tooltip"]& ], 
       TagBox[
        TooltipBox[
         TagBox[
          DynamicBox[{
            FEPrivate`If[
             CurrentValue["MouseOver"], 
             EdgeForm[{
               GrayLevel[0.5], 
               AbsoluteThickness[1.5], 
               Opacity[0.66]}], {}, {}], 
            
            RectangleBox[{37., 0.}, {37.2, 0.051868516435444836`}, 
             "RoundingRadius" -> 0]},
           
           ImageSizeCache->{{136.50816326530617`, 139.78283313325335`}, {
            19.23031705858949, 39.270267589618726`}}],
          StatusArea[#, 0.051868516435444836`]& ,
          TagBoxNote->"0.051868516435444836"],
         StyleBox["0.051868516435444836`", {
           GrayLevel[0]}, StripOnInput -> False]],
        Annotation[#, 
         Style[0.051868516435444836`, {
           GrayLevel[0]}], "Tooltip"]& ], 
       TagBox[
        TooltipBox[
         TagBox[
          DynamicBox[{
            FEPrivate`If[
             CurrentValue["MouseOver"], 
             EdgeForm[{
               GrayLevel[0.5], 
               AbsoluteThickness[1.5], 
               Opacity[0.66]}], {}, {}], 
            
            RectangleBox[{37.2, 0.}, {37.4, 0.04124484439445157}, 
             "RoundingRadius" -> 0]},
           
           ImageSizeCache->{{138.78283313325335`, 142.05750300120047`}, {
            23.1300659625347, 39.270267589618726`}}],
          StatusArea[#, 0.04124484439445157]& ,
          TagBoxNote->"0.04124484439445157"],
         StyleBox["0.04124484439445157`", {
           GrayLevel[0]}, StripOnInput -> False]],
        Annotation[#, 
         Style[0.04124484439445157, {
           GrayLevel[0]}], "Tooltip"]& ], 
       TagBox[
        TooltipBox[
         TagBox[
          DynamicBox[{
            FEPrivate`If[
             CurrentValue["MouseOver"], 
             EdgeForm[{
               GrayLevel[0.5], 
               AbsoluteThickness[1.5], 
               Opacity[0.66]}], {}, {}], 
            
            RectangleBox[{37.4, 0.}, {37.6, 0.03249593800774857}, 
             "RoundingRadius" -> 0]},
           
           ImageSizeCache->{{141.05750300120047`, 144.3321728691477}, {
            26.341623883431737`, 39.270267589618726`}}],
          StatusArea[#, 0.03249593800774857]& ,
          TagBoxNote->"0.03249593800774857"],
         StyleBox["0.03249593800774857`", {
           GrayLevel[0]}, StripOnInput -> False]],
        Annotation[#, 
         Style[0.03249593800774857, {
           GrayLevel[0]}], "Tooltip"]& ], 
       TagBox[
        TooltipBox[
         TagBox[
          DynamicBox[{
            FEPrivate`If[
             CurrentValue["MouseOver"], 
             EdgeForm[{
               GrayLevel[0.5], 
               AbsoluteThickness[1.5], 
               Opacity[0.66]}], {}, {}], 
            
            RectangleBox[{37.6, 0.}, {37.8, 0.024371953505812293`}, 
             "RoundingRadius" -> 0]},
           
           ImageSizeCache->{{143.3321728691477, 146.60684273709484`}, {
            29.323784809978164`, 39.270267589618726`}}],
          StatusArea[#, 0.024371953505812293`]& ,
          TagBoxNote->"0.024371953505812293"],
         StyleBox["0.024371953505812293`", {
           GrayLevel[0]}, StripOnInput -> False]],
        Annotation[#, 
         Style[0.024371953505812293`, {
           GrayLevel[0]}], "Tooltip"]& ], 
       TagBox[
        TooltipBox[
         TagBox[
          DynamicBox[{
            FEPrivate`If[
             CurrentValue["MouseOver"], 
             EdgeForm[{
               GrayLevel[0.5], 
               AbsoluteThickness[1.5], 
               Opacity[0.66]}], {}, {}], 
            
            RectangleBox[{37.8, 0.}, {38., 0.024996875390575823`}, 
             "RoundingRadius" -> 0]},
           
           ImageSizeCache->{{145.60684273709484`, 148.88151260504202`}, {
            29.094387815628732`, 39.270267589618726`}}],
          StatusArea[#, 0.024996875390575823`]& ,
          TagBoxNote->"0.024996875390575823"],
         StyleBox["0.024996875390575823`", {
           GrayLevel[0]}, StripOnInput -> False]],
        Annotation[#, 
         Style[0.024996875390575823`, {
           GrayLevel[0]}], "Tooltip"]& ], 
       TagBox[
        TooltipBox[
         TagBox[
          DynamicBox[{
            FEPrivate`If[
             CurrentValue["MouseOver"], 
             EdgeForm[{
               GrayLevel[0.5], 
               AbsoluteThickness[1.5], 
               Opacity[0.66]}], {}, {}], 
            
            RectangleBox[{38., 0.}, {38.2, 0.016247969003874284`}, 
             "RoundingRadius" -> 0]},
           
           ImageSizeCache->{{147.88151260504202`, 151.15618247298926`}, {
            32.30594573652523, 39.270267589618726`}}],
          StatusArea[#, 0.016247969003874284`]& ,
          TagBoxNote->"0.016247969003874284"],
         StyleBox["0.016247969003874284`", {
           GrayLevel[0]}, StripOnInput -> False]],
        Annotation[#, 
         Style[0.016247969003874284`, {
           GrayLevel[0]}], "Tooltip"]& ], 
       TagBox[
        TooltipBox[
         TagBox[
          DynamicBox[{
            FEPrivate`If[
             CurrentValue["MouseOver"], 
             EdgeForm[{
               GrayLevel[0.5], 
               AbsoluteThickness[1.5], 
               Opacity[0.66]}], {}, {}], 
            
            RectangleBox[{38.2, 0.}, {38.4, 0.009373828271466267}, 
             "RoundingRadius" -> 0]},
           
           ImageSizeCache->{{150.15618247298926`, 153.43085234093638`}, {
            34.82931267437235, 39.270267589618726`}}],
          StatusArea[#, 0.009373828271466267]& ,
          TagBoxNote->"0.009373828271466267"],
         StyleBox["0.009373828271466267`", {
           GrayLevel[0]}, StripOnInput -> False]],
        Annotation[#, 
         Style[0.009373828271466267, {
           GrayLevel[0]}], "Tooltip"]& ], 
       TagBox[
        TooltipBox[
         TagBox[
          DynamicBox[{
            FEPrivate`If[
             CurrentValue["MouseOver"], 
             EdgeForm[{
               GrayLevel[0.5], 
               AbsoluteThickness[1.5], 
               Opacity[0.66]}], {}, {}], 
            
            RectangleBox[{38.4, 0.}, {38.6, 0.008123984501937142}, 
             "RoundingRadius" -> 0]},
           
           ImageSizeCache->{{152.43085234093638`, 155.70552220888362`}, {
            35.28810666307198, 39.270267589618726`}}],
          StatusArea[#, 0.008123984501937142]& ,
          TagBoxNote->"0.008123984501937142"],
         StyleBox["0.008123984501937142`", {
           GrayLevel[0]}, StripOnInput -> False]],
        Annotation[#, 
         Style[0.008123984501937142, {
           GrayLevel[0]}], "Tooltip"]& ], 
       TagBox[
        TooltipBox[
         TagBox[
          DynamicBox[{
            FEPrivate`If[
             CurrentValue["MouseOver"], 
             EdgeForm[{
               GrayLevel[0.5], 
               AbsoluteThickness[1.5], 
               Opacity[0.66]}], {}, {}], 
            
            RectangleBox[{38.6, 0.}, {38.8, 0.014998125234346028`}, 
             "RoundingRadius" -> 0]},
           
           ImageSizeCache->{{154.70552220888362`, 157.98019207683075`}, {
            32.76473972522453, 39.270267589618726`}}],
          StatusArea[#, 0.014998125234346028`]& ,
          TagBoxNote->"0.014998125234346028"],
         StyleBox["0.014998125234346028`", {
           GrayLevel[0]}, StripOnInput -> False]],
        Annotation[#, 
         Style[0.014998125234346028`, {
           GrayLevel[0]}], "Tooltip"]& ], 
       TagBox[
        TooltipBox[
         TagBox[
          DynamicBox[{
            FEPrivate`If[
             CurrentValue["MouseOver"], 
             EdgeForm[{
               GrayLevel[0.5], 
               AbsoluteThickness[1.5], 
               Opacity[0.66]}], {}, {}], 
            
            RectangleBox[{38.8, 0.}, {39., 0.006249218847643956}, 
             "RoundingRadius" -> 0]},
           
           ImageSizeCache->{{156.98019207683075`, 160.25486194477793`}, {
            35.976297646121225`, 39.270267589618726`}}],
          StatusArea[#, 0.006249218847643956]& ,
          TagBoxNote->"0.006249218847643956"],
         StyleBox["0.006249218847643956`", {
           GrayLevel[0]}, StripOnInput -> False]],
        Annotation[#, 
         Style[0.006249218847643956, {
           GrayLevel[0]}], "Tooltip"]& ], 
       TagBox[
        TooltipBox[
         TagBox[
          DynamicBox[{
            FEPrivate`If[
             CurrentValue["MouseOver"], 
             EdgeForm[{
               GrayLevel[0.5], 
               AbsoluteThickness[1.5], 
               Opacity[0.66]}], {}, {}], 
            
            RectangleBox[{39., 0.}, {39.2, 0.006249218847643956}, 
             "RoundingRadius" -> 0]},
           
           ImageSizeCache->{{159.25486194477793`, 162.52953181272517`}, {
            35.976297646121225`, 39.270267589618726`}}],
          StatusArea[#, 0.006249218847643956]& ,
          TagBoxNote->"0.006249218847643956"],
         StyleBox["0.006249218847643956`", {
           GrayLevel[0]}, StripOnInput -> False]],
        Annotation[#, 
         Style[0.006249218847643956, {
           GrayLevel[0]}], "Tooltip"]& ], 
       TagBox[
        TooltipBox[
         TagBox[
          DynamicBox[{
            FEPrivate`If[
             CurrentValue["MouseOver"], 
             EdgeForm[{
               GrayLevel[0.5], 
               AbsoluteThickness[1.5], 
               Opacity[0.66]}], {}, {}], 
            
            RectangleBox[{39.2, 0.}, {39.4, 0.009373828271466267}, 
             "RoundingRadius" -> 0]},
           
           ImageSizeCache->{{161.52953181272517`, 164.8042016806723}, {
            34.82931267437235, 39.270267589618726`}}],
          StatusArea[#, 0.009373828271466267]& ,
          TagBoxNote->"0.009373828271466267"],
         StyleBox["0.009373828271466267`", {
           GrayLevel[0]}, StripOnInput -> False]],
        Annotation[#, 
         Style[0.009373828271466267, {
           GrayLevel[0]}], "Tooltip"]& ], 
       TagBox[
        TooltipBox[
         TagBox[
          DynamicBox[{
            FEPrivate`If[
             CurrentValue["MouseOver"], 
             EdgeForm[{
               GrayLevel[0.5], 
               AbsoluteThickness[1.5], 
               Opacity[0.66]}], {}, {}], 
            
            RectangleBox[{39.4, 0.}, {39.6, 0.009373828271465934}, 
             "RoundingRadius" -> 0]},
           
           ImageSizeCache->{{163.8042016806723, 167.07887154861947`}, {
            34.82931267437248, 39.270267589618726`}}],
          StatusArea[#, 0.009373828271465934]& ,
          TagBoxNote->"0.009373828271465934"],
         StyleBox["0.009373828271465934`", {
           GrayLevel[0]}, StripOnInput -> False]],
        Annotation[#, 
         Style[0.009373828271465934, {
           GrayLevel[0]}], "Tooltip"]& ], 
       TagBox[
        TooltipBox[
         TagBox[
          DynamicBox[{
            FEPrivate`If[
             CurrentValue["MouseOver"], 
             EdgeForm[{
               GrayLevel[0.5], 
               AbsoluteThickness[1.5], 
               Opacity[0.66]}], {}, {}], 
            
            RectangleBox[{39.6, 0.}, {39.8, 0.006249218847644178}, 
             "RoundingRadius" -> 0]},
           
           ImageSizeCache->{{166.07887154861947`, 169.3535414165666}, {
            35.97629764612115, 39.270267589618726`}}],
          StatusArea[#, 0.006249218847644178]& ,
          TagBoxNote->"0.006249218847644178"],
         StyleBox["0.006249218847644178`", {
           GrayLevel[0]}, StripOnInput -> False]],
        Annotation[#, 
         Style[0.006249218847644178, {
           GrayLevel[0]}], "Tooltip"]& ], {}, 
       TagBox[
        TooltipBox[
         TagBox[
          DynamicBox[{
            FEPrivate`If[
             CurrentValue["MouseOver"], 
             EdgeForm[{
               GrayLevel[0.5], 
               AbsoluteThickness[1.5], 
               Opacity[0.66]}], {}, {}], 
            
            RectangleBox[{40., 0.}, {40.2, 0.001249843769528791}, 
             "RoundingRadius" -> 0]},
           
           ImageSizeCache->{{170.62821128451384`, 173.90288115246108`}, {
            37.81147360091923, 39.270267589618726`}}],
          StatusArea[#, 0.001249843769528791]& ,
          TagBoxNote->"0.001249843769528791"],
         StyleBox["0.001249843769528791`", {
           GrayLevel[0]}, StripOnInput -> False]],
        Annotation[#, 
         Style[0.001249843769528791, {
           GrayLevel[0]}], "Tooltip"]& ], 
       TagBox[
        TooltipBox[
         TagBox[
          DynamicBox[{
            FEPrivate`If[
             CurrentValue["MouseOver"], 
             EdgeForm[{
               GrayLevel[0.5], 
               AbsoluteThickness[1.5], 
               Opacity[0.66]}], {}, {}], 
            
            RectangleBox[{40.2, 0.}, {40.4, 0.0006249218847644177}, 
             "RoundingRadius" -> 0]},
           
           ImageSizeCache->{{172.90288115246108`, 176.1775510204082}, {
            38.04087059526897, 39.270267589618726`}}],
          StatusArea[#, 0.0006249218847644177]& ,
          TagBoxNote->"0.0006249218847644177"],
         StyleBox["0.0006249218847644177`", {
           GrayLevel[0]}, StripOnInput -> False]],
        Annotation[#, 
         Style[0.0006249218847644177, {
           GrayLevel[0]}], "Tooltip"]& ]}, {}, {}}, {{}, {}, 
      {RGBColor[1, 0, 0], Thickness[Large], LineBox[CompressedData["
1:eJwV13c81f8XB3BZoWyplNUw+lopwr3XOXcYUciorEJ2SoSQZGfPjJJRGYmG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        "]]}}},
    AspectRatio->NCache[GoldenRatio^(-1), 0.6180339887498948],
    Axes->{True, True},
    AxesOrigin->{26.8, 0},
    FrameTicks->{{Automatic, Automatic}, {Automatic, Automatic}},
    GridLines->{None, None},
    PlotRange->NCache[{{
        Rational[134, 5], 
        Rational[202, 5]}, {All, All}}, {{26.8, 40.4}, {All, All}}],
    PlotRangePadding->{{
       Scaled[0.02], 
       Scaled[0.02]}, {
       Scaled[0.02], 
       Scaled[0.1]}},
    Ticks->{Automatic, Automatic}], ",", 
   GraphicsBox[{
     {RGBColor[0.798413061722744, 0.824719615472648, 0.968322270542458], 
      EdgeForm[Opacity[0.7]], {}, 
      {RGBColor[0.798413061722744, 0.824719615472648, 0.968322270542458], 
       EdgeForm[Opacity[0.7]], 
       TagBox[
        TooltipBox[
         TagBox[
          DynamicBox[{
            FEPrivate`If[
             CurrentValue["MouseOver"], 
             EdgeForm[{
               GrayLevel[0.5], 
               AbsoluteThickness[1.5], 
               Opacity[0.66]}], {}, {}], 
            
            RectangleBox[{5.5, 0.}, {6., 0.0004999375078115235}, 
             "RoundingRadius" -> 0]},
           
           ImageSizeCache->{{20.499999999999993`, 26.655918367346935`}, {
            37.96459061027096, 39.187566823608385`}}],
          StatusArea[#, 0.0004999375078115235]& ,
          TagBoxNote->"0.0004999375078115235"],
         StyleBox["0.0004999375078115235`", {
           GrayLevel[0]}, StripOnInput -> False]],
        Annotation[#, 
         Style[0.0004999375078115235, {
           GrayLevel[0]}], "Tooltip"]& ], 
       TagBox[
        TooltipBox[
         TagBox[
          DynamicBox[{
            FEPrivate`If[
             CurrentValue["MouseOver"], 
             EdgeForm[{
               GrayLevel[0.5], 
               AbsoluteThickness[1.5], 
               Opacity[0.66]}], {}, {}], 
            
            RectangleBox[{6., 0.}, {6.5, 0.00024996875390576176`}, 
             "RoundingRadius" -> 0]},
           
           ImageSizeCache->{{25.655918367346935`, 31.811836734693877`}, {
            38.076078716939676`, 39.187566823608385`}}],
          StatusArea[#, 0.00024996875390576176`]& ,
          TagBoxNote->"0.00024996875390576176"],
         StyleBox["0.00024996875390576176`", {
           GrayLevel[0]}, StripOnInput -> False]],
        Annotation[#, 
         Style[0.00024996875390576176`, {
           GrayLevel[0]}], "Tooltip"]& ], 
       TagBox[
        TooltipBox[
         TagBox[
          DynamicBox[{
            FEPrivate`If[
             CurrentValue["MouseOver"], 
             EdgeForm[{
               GrayLevel[0.5], 
               AbsoluteThickness[1.5], 
               Opacity[0.66]}], {}, {}], 
            
            RectangleBox[{6.5, 0.}, {7., 0.0004999375078115235}, 
             "RoundingRadius" -> 0]},
           
           ImageSizeCache->{{30.811836734693877`, 36.967755102040805`}, {
            37.96459061027096, 39.187566823608385`}}],
          StatusArea[#, 0.0004999375078115235]& ,
          TagBoxNote->"0.0004999375078115235"],
         StyleBox["0.0004999375078115235`", {
           GrayLevel[0]}, StripOnInput -> False]],
        Annotation[#, 
         Style[0.0004999375078115235, {
           GrayLevel[0]}], "Tooltip"]& ], 
       TagBox[
        TooltipBox[
         TagBox[
          DynamicBox[{
            FEPrivate`If[
             CurrentValue["MouseOver"], 
             EdgeForm[{
               GrayLevel[0.5], 
               AbsoluteThickness[1.5], 
               Opacity[0.66]}], {}, {}], 
            
            RectangleBox[{7., 0.}, {7.5, 0.005749281339832521}, 
             "RoundingRadius" -> 0]},
           
           ImageSizeCache->{{35.967755102040805`, 42.12367346938775}, {
            35.623340370228014`, 39.187566823608385`}}],
          StatusArea[#, 0.005749281339832521]& ,
          TagBoxNote->"0.005749281339832521"],
         StyleBox["0.005749281339832521`", {
           GrayLevel[0]}, StripOnInput -> False]],
        Annotation[#, 
         Style[0.005749281339832521, {
           GrayLevel[0]}], "Tooltip"]& ], 
       TagBox[
        TooltipBox[
         TagBox[
          DynamicBox[{
            FEPrivate`If[
             CurrentValue["MouseOver"], 
             EdgeForm[{
               GrayLevel[0.5], 
               AbsoluteThickness[1.5], 
               Opacity[0.66]}], {}, {}], 
            
            RectangleBox[{7.5, 0.}, {8., 0.0054993125859267595`}, 
             "RoundingRadius" -> 0]},
           
           ImageSizeCache->{{41.12367346938775, 47.27959183673469}, {
            35.73482847689672, 39.187566823608385`}}],
          StatusArea[#, 0.0054993125859267595`]& ,
          TagBoxNote->"0.0054993125859267595"],
         StyleBox["0.0054993125859267595`", {
           GrayLevel[0]}, StripOnInput -> False]],
        Annotation[#, 
         Style[0.0054993125859267595`, {
           GrayLevel[0]}], "Tooltip"]& ], 
       TagBox[
        TooltipBox[
         TagBox[
          DynamicBox[{
            FEPrivate`If[
             CurrentValue["MouseOver"], 
             EdgeForm[{
               GrayLevel[0.5], 
               AbsoluteThickness[1.5], 
               Opacity[0.66]}], {}, {}], 
            
            RectangleBox[{8., 0.}, {8.5, 0.014248218972628422`}, 
             "RoundingRadius" -> 0]},
           
           ImageSizeCache->{{46.27959183673469, 52.43551020408163}, {
            31.83274474349181, 39.187566823608385`}}],
          StatusArea[#, 0.014248218972628422`]& ,
          TagBoxNote->"0.014248218972628422"],
         StyleBox["0.014248218972628422`", {
           GrayLevel[0]}, StripOnInput -> False]],
        Annotation[#, 
         Style[0.014248218972628422`, {
           GrayLevel[0]}], "Tooltip"]& ], 
       TagBox[
        TooltipBox[
         TagBox[
          DynamicBox[{
            FEPrivate`If[
             CurrentValue["MouseOver"], 
             EdgeForm[{
               GrayLevel[0.5], 
               AbsoluteThickness[1.5], 
               Opacity[0.66]}], {}, {}], 
            
            RectangleBox[{8.5, 0.}, {9., 0.024496937882764653`}, 
             "RoundingRadius" -> 0]},
           
           ImageSizeCache->{{51.43551020408163, 57.59142857142857}, {
            27.261732370074622`, 39.187566823608385`}}],
          StatusArea[#, 0.024496937882764653`]& ,
          TagBoxNote->"0.024496937882764653"],
         StyleBox["0.024496937882764653`", {
           GrayLevel[0]}, StripOnInput -> False]],
        Annotation[#, 
         Style[0.024496937882764653`, {
           GrayLevel[0]}], "Tooltip"]& ], 
       TagBox[
        TooltipBox[
         TagBox[
          DynamicBox[{
            FEPrivate`If[
             CurrentValue["MouseOver"], 
             EdgeForm[{
               GrayLevel[0.5], 
               AbsoluteThickness[1.5], 
               Opacity[0.66]}], {}, {}], 
            
            RectangleBox[{9., 0.}, {9.5, 0.031246094238220223`}, 
             "RoundingRadius" -> 0]},
           
           ImageSizeCache->{{56.59142857142857, 62.7473469387755}, {
            24.2515534900194, 39.187566823608385`}}],
          StatusArea[#, 0.031246094238220223`]& ,
          TagBoxNote->"0.031246094238220223"],
         StyleBox["0.031246094238220223`", {
           GrayLevel[0]}, StripOnInput -> False]],
        Annotation[#, 
         Style[0.031246094238220223`, {
           GrayLevel[0]}], "Tooltip"]& ], 
       TagBox[
        TooltipBox[
         TagBox[
          DynamicBox[{
            FEPrivate`If[
             CurrentValue["MouseOver"], 
             EdgeForm[{
               GrayLevel[0.5], 
               AbsoluteThickness[1.5], 
               Opacity[0.66]}], {}, {}], 
            
            RectangleBox[{9.5, 0.}, {10., 0.05524309461317335}, 
             "RoundingRadius" -> 0]},
           
           ImageSizeCache->{{61.7473469387755, 67.90326530612245}, {
            13.548695249823059`, 39.187566823608385`}}],
          StatusArea[#, 0.05524309461317335]& ,
          TagBoxNote->"0.05524309461317335"],
         StyleBox["0.05524309461317335`", {
           GrayLevel[0]}, StripOnInput -> False]],
        Annotation[#, 
         Style[0.05524309461317335, {
           GrayLevel[0]}], "Tooltip"]& ], 
       TagBox[
        TooltipBox[
         TagBox[
          DynamicBox[{
            FEPrivate`If[
             CurrentValue["MouseOver"], 
             EdgeForm[{
               GrayLevel[0.5], 
               AbsoluteThickness[1.5], 
               Opacity[0.66]}], {}, {}], 
            
            RectangleBox[{10., 0.}, {10.5, 0.07824021997250344}, 
             "RoundingRadius" -> 0]},
           
           ImageSizeCache->{{66.90326530612245, 73.05918367346939}, {
            3.2917894363015634`, 39.187566823608385`}}],
          StatusArea[#, 0.07824021997250344]& ,
          TagBoxNote->"0.07824021997250344"],
         StyleBox["0.07824021997250344`", {
           GrayLevel[0]}, StripOnInput -> False]],
        Annotation[#, 
         Style[0.07824021997250344, {
           GrayLevel[0]}], "Tooltip"]& ], 
       TagBox[
        TooltipBox[
         TagBox[
          DynamicBox[{
            FEPrivate`If[
             CurrentValue["MouseOver"], 
             EdgeForm[{
               GrayLevel[0.5], 
               AbsoluteThickness[1.5], 
               Opacity[0.66]}], {}, {}], 
            
            RectangleBox[{10.5, 0.}, {11., 0.11473565804274466`}, 
             "RoundingRadius" -> 0]},
           
           ImageSizeCache->{{72.05918367346939, 
            78.2151020408163}, {-12.985474137330371`, 39.187566823608385`}}],
          StatusArea[#, 0.11473565804274466`]& ,
          TagBoxNote->"0.11473565804274466"],
         StyleBox["0.11473565804274466`", {
           GrayLevel[0]}, StripOnInput -> False]],
        Annotation[#, 
         Style[0.11473565804274466`, {
           GrayLevel[0]}], "Tooltip"]& ], 
       TagBox[
        TooltipBox[
         TagBox[
          DynamicBox[{
            FEPrivate`If[
             CurrentValue["MouseOver"], 
             EdgeForm[{
               GrayLevel[0.5], 
               AbsoluteThickness[1.5], 
               Opacity[0.66]}], {}, {}], 
            
            RectangleBox[{11., 0.}, {11.5, 0.1557305336832896}, 
             "RoundingRadius" -> 0]},
           
           ImageSizeCache->{{77.2151020408163, 
            83.37102040816325}, {-31.269523630999124`, 39.187566823608385`}}],
          StatusArea[#, 0.1557305336832896]& ,
          TagBoxNote->"0.1557305336832896"],
         StyleBox["0.1557305336832896`", {
           GrayLevel[0]}, StripOnInput -> False]],
        Annotation[#, 
         Style[0.1557305336832896, {
           GrayLevel[0]}], "Tooltip"]& ], 
       TagBox[
        TooltipBox[
         TagBox[
          DynamicBox[{
            FEPrivate`If[
             CurrentValue["MouseOver"], 
             EdgeForm[{
               GrayLevel[0.5], 
               AbsoluteThickness[1.5], 
               Opacity[0.66]}], {}, {}], 
            
            RectangleBox[{11.5, 0.}, {12., 0.14298212723409573`}, 
             "RoundingRadius" -> 0]},
           
           ImageSizeCache->{{82.37102040816325, 
            88.52693877551019}, {-25.583630190894816`, 39.187566823608385`}}],
          StatusArea[#, 0.14298212723409573`]& ,
          TagBoxNote->"0.14298212723409573"],
         StyleBox["0.14298212723409573`", {
           GrayLevel[0]}, StripOnInput -> False]],
        Annotation[#, 
         Style[0.14298212723409573`, {
           GrayLevel[0]}], "Tooltip"]& ], 
       TagBox[
        TooltipBox[
         TagBox[
          DynamicBox[{
            FEPrivate`If[
             CurrentValue["MouseOver"], 
             EdgeForm[{
               GrayLevel[0.5], 
               AbsoluteThickness[1.5], 
               Opacity[0.66]}], {}, {}], 
            
            RectangleBox[{12., 0.}, {12.5, 0.16822897137857767`}, 
             "RoundingRadius" -> 0]},
           
           ImageSizeCache->{{87.52693877551019, 
            93.68285714285713}, {-36.8439289644347, 39.187566823608385`}}],
          StatusArea[#, 0.16822897137857767`]& ,
          TagBoxNote->"0.16822897137857767"],
         StyleBox["0.16822897137857767`", {
           GrayLevel[0]}, StripOnInput -> False]],
        Annotation[#, 
         Style[0.16822897137857767`, {
           GrayLevel[0]}], "Tooltip"]& ], 
       TagBox[
        TooltipBox[
         TagBox[
          DynamicBox[{
            FEPrivate`If[
             CurrentValue["MouseOver"], 
             EdgeForm[{
               GrayLevel[0.5], 
               AbsoluteThickness[1.5], 
               Opacity[0.66]}], {}, {}], 
            
            RectangleBox[{12.5, 0.}, {13., 0.1942257217847769}, 
             "RoundingRadius" -> 0]},
           
           ImageSizeCache->{{92.68285714285713, 
            98.83877551020407}, {-48.43869205798074, 39.187566823608385`}}],
          StatusArea[#, 0.1942257217847769]& ,
          TagBoxNote->"0.1942257217847769"],
         StyleBox["0.1942257217847769`", {
           GrayLevel[0]}, StripOnInput -> False]],
        Annotation[#, 
         Style[0.1942257217847769, {
           GrayLevel[0]}], "Tooltip"]& ], 
       TagBox[
        TooltipBox[
         TagBox[
          DynamicBox[{
            FEPrivate`If[
             CurrentValue["MouseOver"], 
             EdgeForm[{
               GrayLevel[0.5], 
               AbsoluteThickness[1.5], 
               Opacity[0.66]}], {}, {}], 
            
            RectangleBox[{13., 0.}, {13.5, 0.19647544056992877`}, 
             "RoundingRadius" -> 0]},
           
           ImageSizeCache->{{97.83877551020407, 
            103.99469387755101`}, {-49.442085017999155`, 
            39.187566823608385`}}],
          StatusArea[#, 0.19647544056992877`]& ,
          TagBoxNote->"0.19647544056992877"],
         StyleBox["0.19647544056992877`", {
           GrayLevel[0]}, StripOnInput -> False]],
        Annotation[#, 
         Style[0.19647544056992877`, {
           GrayLevel[0]}], "Tooltip"]& ], 
       TagBox[
        TooltipBox[
         TagBox[
          DynamicBox[{
            FEPrivate`If[
             CurrentValue["MouseOver"], 
             EdgeForm[{
               GrayLevel[0.5], 
               AbsoluteThickness[1.5], 
               Opacity[0.66]}], {}, {}], 
            
            RectangleBox[{13.5, 0.}, {14., 0.19247594050743658`}, 
             "RoundingRadius" -> 0]},
           
           ImageSizeCache->{{102.99469387755101`, 
            109.15061224489793`}, {-47.65827531129977, 39.187566823608385`}}],
          StatusArea[#, 0.19247594050743658`]& ,
          TagBoxNote->"0.19247594050743658"],
         StyleBox["0.19247594050743658`", {
           GrayLevel[0]}, StripOnInput -> False]],
        Annotation[#, 
         Style[0.19247594050743658`, {
           GrayLevel[0]}], "Tooltip"]& ], 
       TagBox[
        TooltipBox[
         TagBox[
          DynamicBox[{
            FEPrivate`If[
             CurrentValue["MouseOver"], 
             EdgeForm[{
               GrayLevel[0.5], 
               AbsoluteThickness[1.5], 
               Opacity[0.66]}], {}, {}], 
            
            RectangleBox[{14., 0.}, {14.5, 0.15123109611298588`}, 
             "RoundingRadius" -> 0]},
           
           ImageSizeCache->{{108.15061224489793`, 
            114.30653061224487`}, {-29.262737710962305`, 
            39.187566823608385`}}],
          StatusArea[#, 0.15123109611298588`]& ,
          TagBoxNote->"0.15123109611298588"],
         StyleBox["0.15123109611298588`", {
           GrayLevel[0]}, StripOnInput -> False]],
        Annotation[#, 
         Style[0.15123109611298588`, {
           GrayLevel[0]}], "Tooltip"]& ], 
       TagBox[
        TooltipBox[
         TagBox[
          DynamicBox[{
            FEPrivate`If[
             CurrentValue["MouseOver"], 
             EdgeForm[{
               GrayLevel[0.5], 
               AbsoluteThickness[1.5], 
               Opacity[0.66]}], {}, {}], 
            
            RectangleBox[{14.5, 0.}, {15., 0.11498562679665042`}, 
             "RoundingRadius" -> 0]},
           
           ImageSizeCache->{{113.30653061224487`, 
            119.46244897959181`}, {-13.09696224399908, 39.187566823608385`}}],
          StatusArea[#, 0.11498562679665042`]& ,
          TagBoxNote->"0.11498562679665042"],
         StyleBox["0.11498562679665042`", {
           GrayLevel[0]}, StripOnInput -> False]],
        Annotation[#, 
         Style[0.11498562679665042`, {
           GrayLevel[0]}], "Tooltip"]& ], 
       TagBox[
        TooltipBox[
         TagBox[
          DynamicBox[{
            FEPrivate`If[
             CurrentValue["MouseOver"], 
             EdgeForm[{
               GrayLevel[0.5], 
               AbsoluteThickness[1.5], 
               Opacity[0.66]}], {}, {}], 
            
            RectangleBox[{15., 0.}, {15.5, 0.1274840644919385}, 
             "RoundingRadius" -> 0]},
           
           ImageSizeCache->{{118.46244897959181`, 
            124.61836734693875`}, {-18.671367577434673`, 
            39.187566823608385`}}],
          StatusArea[#, 0.1274840644919385]& ,
          TagBoxNote->"0.1274840644919385"],
         StyleBox["0.1274840644919385`", {
           GrayLevel[0]}, StripOnInput -> False]],
        Annotation[#, 
         Style[0.1274840644919385, {
           GrayLevel[0]}], "Tooltip"]& ], 
       TagBox[
        TooltipBox[
         TagBox[
          DynamicBox[{
            FEPrivate`If[
             CurrentValue["MouseOver"], 
             EdgeForm[{
               GrayLevel[0.5], 
               AbsoluteThickness[1.5], 
               Opacity[0.66]}], {}, {}], 
            
            RectangleBox[{15.5, 0.}, {16., 0.07249093863267092}, 
             "RoundingRadius" -> 0]},
           
           ImageSizeCache->{{123.61836734693875`, 129.7742857142857}, {
            5.8560158896819345`, 39.187566823608385`}}],
          StatusArea[#, 0.07249093863267092]& ,
          TagBoxNote->"0.07249093863267092"],
         StyleBox["0.07249093863267092`", {
           GrayLevel[0]}, StripOnInput -> False]],
        Annotation[#, 
         Style[0.07249093863267092, {
           GrayLevel[0]}], "Tooltip"]& ], 
       TagBox[
        TooltipBox[
         TagBox[
          DynamicBox[{
            FEPrivate`If[
             CurrentValue["MouseOver"], 
             EdgeForm[{
               GrayLevel[0.5], 
               AbsoluteThickness[1.5], 
               Opacity[0.66]}], {}, {}], 
            
            RectangleBox[{16., 0.}, {16.5, 0.05349331333583302}, 
             "RoundingRadius" -> 0]},
           
           ImageSizeCache->{{128.7742857142857, 134.93020408163264`}, {
            14.329111996504043`, 39.187566823608385`}}],
          StatusArea[#, 0.05349331333583302]& ,
          TagBoxNote->"0.05349331333583302"],
         StyleBox["0.05349331333583302`", {
           GrayLevel[0]}, StripOnInput -> False]],
        Annotation[#, 
         Style[0.05349331333583302, {
           GrayLevel[0]}], "Tooltip"]& ], 
       TagBox[
        TooltipBox[
         TagBox[
          DynamicBox[{
            FEPrivate`If[
             CurrentValue["MouseOver"], 
             EdgeForm[{
               GrayLevel[0.5], 
               AbsoluteThickness[1.5], 
               Opacity[0.66]}], {}, {}], 
            
            RectangleBox[{16.5, 0.}, {17., 0.045244344456942885`}, 
             "RoundingRadius" -> 0]},
           
           ImageSizeCache->{{133.93020408163264`, 140.08612244897958`}, {
            18.008219516571533`, 39.187566823608385`}}],
          StatusArea[#, 0.045244344456942885`]& ,
          TagBoxNote->"0.045244344456942885"],
         StyleBox["0.045244344456942885`", {
           GrayLevel[0]}, StripOnInput -> False]],
        Annotation[#, 
         Style[0.045244344456942885`, {
           GrayLevel[0]}], "Tooltip"]& ], 
       TagBox[
        TooltipBox[
         TagBox[
          DynamicBox[{
            FEPrivate`If[
             CurrentValue["MouseOver"], 
             EdgeForm[{
               GrayLevel[0.5], 
               AbsoluteThickness[1.5], 
               Opacity[0.66]}], {}, {}], 
            
            RectangleBox[{17., 0.}, {17.5, 0.028746406699162605`}, 
             "RoundingRadius" -> 0]},
           
           ImageSizeCache->{{139.08612244897958`, 145.24204081632652`}, {
            25.36643455670652, 39.187566823608385`}}],
          StatusArea[#, 0.028746406699162605`]& ,
          TagBoxNote->"0.028746406699162605"],
         StyleBox["0.028746406699162605`", {
           GrayLevel[0]}, StripOnInput -> False]],
        Annotation[#, 
         Style[0.028746406699162605`, {
           GrayLevel[0]}], "Tooltip"]& ], 
       TagBox[
        TooltipBox[
         TagBox[
          DynamicBox[{
            FEPrivate`If[
             CurrentValue["MouseOver"], 
             EdgeForm[{
               GrayLevel[0.5], 
               AbsoluteThickness[1.5], 
               Opacity[0.66]}], {}, {}], 
            
            RectangleBox[{17.5, 0.}, {18., 0.009998750156230472}, 
             "RoundingRadius" -> 0]},
           ImageSizeCache->{{144.24204081632652`, 150.39795918367346`}, {
            33.72804255685991, 39.187566823608385`}}],
          StatusArea[#, 0.009998750156230472]& ,
          TagBoxNote->"0.009998750156230472"],
         StyleBox["0.009998750156230472`", {
           GrayLevel[0]}, StripOnInput -> False]],
        Annotation[#, 
         Style[0.009998750156230472, {
           GrayLevel[0]}], "Tooltip"]& ], 
       TagBox[
        TooltipBox[
         TagBox[
          DynamicBox[{
            FEPrivate`If[
             CurrentValue["MouseOver"], 
             EdgeForm[{
               GrayLevel[0.5], 
               AbsoluteThickness[1.5], 
               Opacity[0.66]}], {}, {}], 
            
            RectangleBox[{18., 0.}, {18.5, 0.008248968878890138}, 
             "RoundingRadius" -> 0]},
           
           ImageSizeCache->{{149.39795918367346`, 155.55387755102038`}, {
            34.508459303540896`, 39.187566823608385`}}],
          StatusArea[#, 0.008248968878890138]& ,
          TagBoxNote->"0.008248968878890138"],
         StyleBox["0.008248968878890138`", {
           GrayLevel[0]}, StripOnInput -> False]],
        Annotation[#, 
         Style[0.008248968878890138, {
           GrayLevel[0]}], "Tooltip"]& ], 
       TagBox[
        TooltipBox[
         TagBox[
          DynamicBox[{
            FEPrivate`If[
             CurrentValue["MouseOver"], 
             EdgeForm[{
               GrayLevel[0.5], 
               AbsoluteThickness[1.5], 
               Opacity[0.66]}], {}, {}], 
            
            RectangleBox[{18.5, 0.}, {19., 0.003999500062492188}, 
             "RoundingRadius" -> 0]},
           
           ImageSizeCache->{{154.55387755102038`, 160.70979591836732`}, {
            36.403757116909, 39.187566823608385`}}],
          StatusArea[#, 0.003999500062492188]& ,
          TagBoxNote->"0.003999500062492188"],
         StyleBox["0.003999500062492188`", {
           GrayLevel[0]}, StripOnInput -> False]],
        Annotation[#, 
         Style[0.003999500062492188, {
           GrayLevel[0]}], "Tooltip"]& ], 
       TagBox[
        TooltipBox[
         TagBox[
          DynamicBox[{
            FEPrivate`If[
             CurrentValue["MouseOver"], 
             EdgeForm[{
               GrayLevel[0.5], 
               AbsoluteThickness[1.5], 
               Opacity[0.66]}], {}, {}], 
            
            RectangleBox[{19., 0.}, {19.5, 0.0029996250468691415`}, 
             "RoundingRadius" -> 0]},
           
           ImageSizeCache->{{159.70979591836732`, 165.86571428571426`}, {
            36.84970954358384, 39.187566823608385`}}],
          StatusArea[#, 0.0029996250468691415`]& ,
          TagBoxNote->"0.0029996250468691415"],
         StyleBox["0.0029996250468691415`", {
           GrayLevel[0]}, StripOnInput -> False]],
        Annotation[#, 
         Style[0.0029996250468691415`, {
           GrayLevel[0]}], "Tooltip"]& ], {}, 
       TagBox[
        TooltipBox[
         TagBox[
          DynamicBox[{
            FEPrivate`If[
             CurrentValue["MouseOver"], 
             EdgeForm[{
               GrayLevel[0.5], 
               AbsoluteThickness[1.5], 
               Opacity[0.66]}], {}, {}], 
            
            RectangleBox[{20., 0.}, {20.5, 0.00024996875390576176`}, 
             "RoundingRadius" -> 0]},
           
           ImageSizeCache->{{170.0216326530612, 176.17755102040815`}, {
            38.076078716939676`, 39.187566823608385`}}],
          StatusArea[#, 0.00024996875390576176`]& ,
          TagBoxNote->"0.00024996875390576176"],
         StyleBox["0.00024996875390576176`", {
           GrayLevel[0]}, StripOnInput -> False]],
        Annotation[#, 
         Style[0.00024996875390576176`, {
           GrayLevel[0]}], "Tooltip"]& ]}, {}, {}}, {{}, {}, 
      {RGBColor[1, 0, 0], Thickness[Large], LineBox[CompressedData["
1:eJwVlnc81W8Ux5URJUlkc8m4w7r3SviVc4xuFMlMMloie8uWvVcyUmYklUqS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        "]]}}},
    AspectRatio->NCache[GoldenRatio^(-1), 0.6180339887498948],
    Axes->{True, True},
    AxesOrigin->{5.5, 0},
    FrameTicks->{{Automatic, Automatic}, {Automatic, Automatic}},
    GridLines->{None, None},
    PlotRange->NCache[{{
        Rational[11, 2], 
        Rational[41, 2]}, {All, All}}, {{5.5, 20.5}, {All, All}}],
    PlotRangePadding->{{
       Scaled[0.02], 
       Scaled[0.02]}, {
       Scaled[0.02], 
       Scaled[0.1]}},
    Ticks->{Automatic, Automatic}], ",", 
   GraphicsBox[{
     {RGBColor[0.798413061722744, 0.824719615472648, 0.968322270542458], 
      EdgeForm[Opacity[0.7]], {}, 
      {RGBColor[0.798413061722744, 0.824719615472648, 0.968322270542458], 
       EdgeForm[Opacity[0.7]], 
       TagBox[
        TooltipBox[
         TagBox[
          DynamicBox[{
            FEPrivate`If[
             CurrentValue["MouseOver"], 
             EdgeForm[{
               GrayLevel[0.5], 
               AbsoluteThickness[1.5], 
               Opacity[0.66]}], {}, {}], 
            
            RectangleBox[{26.5, 0.}, {27., 0.00024996875390576176`}, 
             "RoundingRadius" -> 0]},
           
           ImageSizeCache->{{20.5, 25.27262319561973}, {38.0473593806618, 
            39.18756682360838}}],
          StatusArea[#, 0.00024996875390576176`]& ,
          TagBoxNote->"0.00024996875390576176"],
         StyleBox["0.00024996875390576176`", {
           GrayLevel[0]}, StripOnInput -> False]],
        Annotation[#, 
         Style[0.00024996875390576176`, {
           GrayLevel[0]}], "Tooltip"]& ], 
       TagBox[
        TooltipBox[
         TagBox[
          DynamicBox[{
            FEPrivate`If[
             CurrentValue["MouseOver"], 
             EdgeForm[{
               GrayLevel[0.5], 
               AbsoluteThickness[1.5], 
               Opacity[0.66]}], {}, {}], 
            
            RectangleBox[{27., 0.}, {27.5, 0.000999875015623047}, 
             "RoundingRadius" -> 0]},
           
           ImageSizeCache->{{24.27262319561973, 29.04524639123943}, {
            37.62673705182209, 39.18756682360838}}],
          StatusArea[#, 0.000999875015623047]& ,
          TagBoxNote->"0.000999875015623047"],
         StyleBox["0.000999875015623047`", {
           GrayLevel[0]}, StripOnInput -> False]],
        Annotation[#, 
         Style[0.000999875015623047, {
           GrayLevel[0]}], "Tooltip"]& ], 
       TagBox[
        TooltipBox[
         TagBox[
          DynamicBox[{
            FEPrivate`If[
             CurrentValue["MouseOver"], 
             EdgeForm[{
               GrayLevel[0.5], 
               AbsoluteThickness[1.5], 
               Opacity[0.66]}], {}, {}], 
            
            RectangleBox[{27.5, 0.}, {28., 0.00024996875390576176`}, 
             "RoundingRadius" -> 0]},
           
           ImageSizeCache->{{28.04524639123943, 32.81786958685913}, {
            38.0473593806618, 39.18756682360838}}],
          StatusArea[#, 0.00024996875390576176`]& ,
          TagBoxNote->"0.00024996875390576176"],
         StyleBox["0.00024996875390576176`", {
           GrayLevel[0]}, StripOnInput -> False]],
        Annotation[#, 
         Style[0.00024996875390576176`, {
           GrayLevel[0]}], "Tooltip"]& ], 
       TagBox[
        TooltipBox[
         TagBox[
          DynamicBox[{
            FEPrivate`If[
             CurrentValue["MouseOver"], 
             EdgeForm[{
               GrayLevel[0.5], 
               AbsoluteThickness[1.5], 
               Opacity[0.66]}], {}, {}], 
            
            RectangleBox[{28., 0.}, {28.5, 0.000999875015623047}, 
             "RoundingRadius" -> 0]},
           
           ImageSizeCache->{{31.817869586859132`, 36.59049278247886}, {
            37.62673705182209, 39.18756682360838}}],
          StatusArea[#, 0.000999875015623047]& ,
          TagBoxNote->"0.000999875015623047"],
         StyleBox["0.000999875015623047`", {
           GrayLevel[0]}, StripOnInput -> False]],
        Annotation[#, 
         Style[0.000999875015623047, {
           GrayLevel[0]}], "Tooltip"]& ], 
       TagBox[
        TooltipBox[
         TagBox[
          DynamicBox[{
            FEPrivate`If[
             CurrentValue["MouseOver"], 
             EdgeForm[{
               GrayLevel[0.5], 
               AbsoluteThickness[1.5], 
               Opacity[0.66]}], {}, {}], 
            
            RectangleBox[{28.5, 0.}, {29., 0.001999750031246094}, 
             "RoundingRadius" -> 0]},
           
           ImageSizeCache->{{35.59049278247886, 40.36311597809856}, {
            37.0659072800358, 39.18756682360838}}],
          StatusArea[#, 0.001999750031246094]& ,
          TagBoxNote->"0.001999750031246094"],
         StyleBox["0.001999750031246094`", {
           GrayLevel[0]}, StripOnInput -> False]],
        Annotation[#, 
         Style[0.001999750031246094, {
           GrayLevel[0]}], "Tooltip"]& ], 
       TagBox[
        TooltipBox[
         TagBox[
          DynamicBox[{
            FEPrivate`If[
             CurrentValue["MouseOver"], 
             EdgeForm[{
               GrayLevel[0.5], 
               AbsoluteThickness[1.5], 
               Opacity[0.66]}], {}, {}], 
            
            RectangleBox[{29., 0.}, {29.5, 0.004749406324209474}, 
             "RoundingRadius" -> 0]},
           
           ImageSizeCache->{{39.36311597809856, 44.135739173718264`}, {
            35.52362540762351, 39.18756682360838}}],
          StatusArea[#, 0.004749406324209474]& ,
          TagBoxNote->"0.004749406324209474"],
         StyleBox["0.004749406324209474`", {
           GrayLevel[0]}, StripOnInput -> False]],
        Annotation[#, 
         Style[0.004749406324209474, {
           GrayLevel[0]}], "Tooltip"]& ], 
       TagBox[
        TooltipBox[
         TagBox[
          DynamicBox[{
            FEPrivate`If[
             CurrentValue["MouseOver"], 
             EdgeForm[{
               GrayLevel[0.5], 
               AbsoluteThickness[1.5], 
               Opacity[0.66]}], {}, {}], 
            
            RectangleBox[{29.5, 0.}, {30., 0.008998875140607425}, 
             "RoundingRadius" -> 0]},
           
           ImageSizeCache->{{43.135739173718264`, 47.90836236933799}, {
            33.140098877531784`, 39.18756682360838}}],
          StatusArea[#, 0.008998875140607425]& ,
          TagBoxNote->"0.008998875140607425"],
         StyleBox["0.008998875140607425`", {
           GrayLevel[0]}, StripOnInput -> False]],
        Annotation[#, 
         Style[0.008998875140607425, {
           GrayLevel[0]}], "Tooltip"]& ], 
       TagBox[
        TooltipBox[
         TagBox[
          DynamicBox[{
            FEPrivate`If[
             CurrentValue["MouseOver"], 
             EdgeForm[{
               GrayLevel[0.5], 
               AbsoluteThickness[1.5], 
               Opacity[0.66]}], {}, {}], 
            
            RectangleBox[{30., 0.}, {30.5, 0.007999000124984376}, 
             "RoundingRadius" -> 0]},
           
           ImageSizeCache->{{46.90836236933799, 51.680985564957695`}, {
            33.70092864931807, 39.18756682360838}}],
          StatusArea[#, 0.007999000124984376]& ,
          TagBoxNote->"0.007999000124984376"],
         StyleBox["0.007999000124984376`", {
           GrayLevel[0]}, StripOnInput -> False]],
        Annotation[#, 
         Style[0.007999000124984376, {
           GrayLevel[0]}], "Tooltip"]& ], 
       TagBox[
        TooltipBox[
         TagBox[
          DynamicBox[{
            FEPrivate`If[
             CurrentValue["MouseOver"], 
             EdgeForm[{
               GrayLevel[0.5], 
               AbsoluteThickness[1.5], 
               Opacity[0.66]}], {}, {}], 
            
            RectangleBox[{30.5, 0.}, {31., 0.007749031371078615}, 
             "RoundingRadius" -> 0]},
           
           ImageSizeCache->{{50.680985564957695`, 55.453608760577396`}, {
            33.84113609226465, 39.18756682360838}}],
          StatusArea[#, 0.007749031371078615]& ,
          TagBoxNote->"0.007749031371078615"],
         StyleBox["0.007749031371078615`", {
           GrayLevel[0]}, StripOnInput -> False]],
        Annotation[#, 
         Style[0.007749031371078615, {
           GrayLevel[0]}], "Tooltip"]& ], 
       TagBox[
        TooltipBox[
         TagBox[
          DynamicBox[{
            FEPrivate`If[
             CurrentValue["MouseOver"], 
             EdgeForm[{
               GrayLevel[0.5], 
               AbsoluteThickness[1.5], 
               Opacity[0.66]}], {}, {}], 
            
            RectangleBox[{31., 0.}, {31.5, 0.020497437820272465`}, 
             "RoundingRadius" -> 0]},
           
           ImageSizeCache->{{54.453608760577396`, 59.226231956197125`}, {
            26.69055650198947, 39.18756682360838}}],
          StatusArea[#, 0.020497437820272465`]& ,
          TagBoxNote->"0.020497437820272465"],
         StyleBox["0.020497437820272465`", {
           GrayLevel[0]}, StripOnInput -> False]],
        Annotation[#, 
         Style[0.020497437820272465`, {
           GrayLevel[0]}], "Tooltip"]& ], 
       TagBox[
        TooltipBox[
         TagBox[
          DynamicBox[{
            FEPrivate`If[
             CurrentValue["MouseOver"], 
             EdgeForm[{
               GrayLevel[0.5], 
               AbsoluteThickness[1.5], 
               Opacity[0.66]}], {}, {}], 
            
            RectangleBox[{31.5, 0.}, {32., 0.03849518810148731}, 
             "RoundingRadius" -> 0]},
           
           ImageSizeCache->{{58.226231956197125`, 62.99885515181683}, {
            16.59562060983628, 39.18756682360838}}],
          StatusArea[#, 0.03849518810148731]& ,
          TagBoxNote->"0.03849518810148731"],
         StyleBox["0.03849518810148731`", {
           GrayLevel[0]}, StripOnInput -> False]],
        Annotation[#, 
         Style[0.03849518810148731, {
           GrayLevel[0]}], "Tooltip"]& ], 
       TagBox[
        TooltipBox[
         TagBox[
          DynamicBox[{
            FEPrivate`If[
             CurrentValue["MouseOver"], 
             EdgeForm[{
               GrayLevel[0.5], 
               AbsoluteThickness[1.5], 
               Opacity[0.66]}], {}, {}], 
            
            RectangleBox[{32., 0.}, {32.5, 0.0339957505311836}, 
             "RoundingRadius" -> 0]},
           
           ImageSizeCache->{{61.99885515181683, 66.77147834743653}, {
            19.11935458287458, 39.18756682360838}}],
          StatusArea[#, 0.0339957505311836]& ,
          TagBoxNote->"0.0339957505311836"],
         StyleBox["0.0339957505311836`", {
           GrayLevel[0]}, StripOnInput -> False]],
        Annotation[#, 
         Style[0.0339957505311836, {
           GrayLevel[0]}], "Tooltip"]& ], 
       TagBox[
        TooltipBox[
         TagBox[
          DynamicBox[{
            FEPrivate`If[
             CurrentValue["MouseOver"], 
             EdgeForm[{
               GrayLevel[0.5], 
               AbsoluteThickness[1.5], 
               Opacity[0.66]}], {}, {}], 
            
            RectangleBox[{32.5, 0.}, {33., 0.06524184476940383}, 
             "RoundingRadius" -> 0]},
           
           ImageSizeCache->{{65.77147834743653, 70.54410154305626}, {
            1.5934242145530675`, 39.18756682360838}}],
          StatusArea[#, 0.06524184476940383]& ,
          TagBoxNote->"0.06524184476940383"],
         StyleBox["0.06524184476940383`", {
           GrayLevel[0]}, StripOnInput -> False]],
        Annotation[#, 
         Style[0.06524184476940383, {
           GrayLevel[0]}], "Tooltip"]& ], 
       TagBox[
        TooltipBox[
         TagBox[
          DynamicBox[{
            FEPrivate`If[
             CurrentValue["MouseOver"], 
             EdgeForm[{
               GrayLevel[0.5], 
               AbsoluteThickness[1.5], 
               Opacity[0.66]}], {}, {}], 
            
            RectangleBox[{33., 0.}, {33.5, 0.06899137607799025}, 
             "RoundingRadius" -> 0]},
           
           ImageSizeCache->{{69.54410154305626, 
            74.31672473867596}, {-0.5096874296455098, 39.18756682360838}}],
          StatusArea[#, 0.06899137607799025]& ,
          TagBoxNote->"0.06899137607799025"],
         StyleBox["0.06899137607799025`", {
           GrayLevel[0]}, StripOnInput -> False]],
        Annotation[#, 
         Style[0.06899137607799025, {
           GrayLevel[0]}], "Tooltip"]& ], 
       TagBox[
        TooltipBox[
         TagBox[
          DynamicBox[{
            FEPrivate`If[
             CurrentValue["MouseOver"], 
             EdgeForm[{
               GrayLevel[0.5], 
               AbsoluteThickness[1.5], 
               Opacity[0.66]}], {}, {}], 
            
            RectangleBox[{33.5, 0.}, {34., 0.07174103237095363}, 
             "RoundingRadius" -> 0]},
           
           ImageSizeCache->{{73.31672473867596, 
            78.08934793429569}, {-2.051969302057799, 39.18756682360838}}],
          StatusArea[#, 0.07174103237095363]& ,
          TagBoxNote->"0.07174103237095363"],
         StyleBox["0.07174103237095363`", {
           GrayLevel[0]}, StripOnInput -> False]],
        Annotation[#, 
         Style[0.07174103237095363, {
           GrayLevel[0]}], "Tooltip"]& ], 
       TagBox[
        TooltipBox[
         TagBox[
          DynamicBox[{
            FEPrivate`If[
             CurrentValue["MouseOver"], 
             EdgeForm[{
               GrayLevel[0.5], 
               AbsoluteThickness[1.5], 
               Opacity[0.66]}], {}, {}], 
            
            RectangleBox[{34., 0.}, {34.5, 0.10998625171853518`}, 
             "RoundingRadius" -> 0]},
           
           ImageSizeCache->{{77.08934793429569, 
            81.86197112991539}, {-23.503708072883327`, 39.18756682360838}}],
          StatusArea[#, 0.10998625171853518`]& ,
          TagBoxNote->"0.10998625171853518"],
         StyleBox["0.10998625171853518`", {
           GrayLevel[0]}, StripOnInput -> False]],
        Annotation[#, 
         Style[0.10998625171853518`, {
           GrayLevel[0]}], "Tooltip"]& ], 
       TagBox[
        TooltipBox[
         TagBox[
          DynamicBox[{
            FEPrivate`If[
             CurrentValue["MouseOver"], 
             EdgeForm[{
               GrayLevel[0.5], 
               AbsoluteThickness[1.5], 
               Opacity[0.66]}], {}, {}], 
            
            RectangleBox[{34.5, 0.}, {35., 0.12198475190601175`}, 
             "RoundingRadius" -> 0]},
           
           ImageSizeCache->{{80.86197112991539, 
            85.63459432553509}, {-30.233665334318786`, 39.18756682360838}}],
          StatusArea[#, 0.12198475190601175`]& ,
          TagBoxNote->"0.12198475190601175"],
         StyleBox["0.12198475190601175`", {
           GrayLevel[0]}, StripOnInput -> False]],
        Annotation[#, 
         Style[0.12198475190601175`, {
           GrayLevel[0]}], "Tooltip"]& ], 
       TagBox[
        TooltipBox[
         TagBox[
          DynamicBox[{
            FEPrivate`If[
             CurrentValue["MouseOver"], 
             EdgeForm[{
               GrayLevel[0.5], 
               AbsoluteThickness[1.5], 
               Opacity[0.66]}], {}, {}], 
            RectangleBox[{35., 0.}, {35.5, 0.14698162729658792`}, 
             "RoundingRadius" -> 0]},
           
           ImageSizeCache->{{84.63459432553509, 
            89.40721752115479}, {-44.25440962897599, 39.18756682360838}}],
          StatusArea[#, 0.14698162729658792`]& ,
          TagBoxNote->"0.14698162729658792"],
         StyleBox["0.14698162729658792`", {
           GrayLevel[0]}, StripOnInput -> False]],
        Annotation[#, 
         Style[0.14698162729658792`, {
           GrayLevel[0]}], "Tooltip"]& ], 
       TagBox[
        TooltipBox[
         TagBox[
          DynamicBox[{
            FEPrivate`If[
             CurrentValue["MouseOver"], 
             EdgeForm[{
               GrayLevel[0.5], 
               AbsoluteThickness[1.5], 
               Opacity[0.66]}], {}, {}], 
            
            RectangleBox[{35.5, 0.}, {36., 0.15273090863642044`}, 
             "RoundingRadius" -> 0]},
           
           ImageSizeCache->{{88.40721752115479, 
            93.17984071677449}, {-47.479180816747146`, 39.18756682360838}}],
          StatusArea[#, 0.15273090863642044`]& ,
          TagBoxNote->"0.15273090863642044"],
         StyleBox["0.15273090863642044`", {
           GrayLevel[0]}, StripOnInput -> False]],
        Annotation[#, 
         Style[0.15273090863642044`, {
           GrayLevel[0]}], "Tooltip"]& ], 
       TagBox[
        TooltipBox[
         TagBox[
          DynamicBox[{
            FEPrivate`If[
             CurrentValue["MouseOver"], 
             EdgeForm[{
               GrayLevel[0.5], 
               AbsoluteThickness[1.5], 
               Opacity[0.66]}], {}, {}], 
            
            RectangleBox[{36., 0.}, {36.5, 0.15223097112860892`}, 
             "RoundingRadius" -> 0]},
           
           ImageSizeCache->{{92.17984071677449, 
            96.95246391239425}, {-47.19876593085401, 39.18756682360838}}],
          StatusArea[#, 0.15223097112860892`]& ,
          TagBoxNote->"0.15223097112860892"],
         StyleBox["0.15223097112860892`", {
           GrayLevel[0]}, StripOnInput -> False]],
        Annotation[#, 
         Style[0.15223097112860892`, {
           GrayLevel[0]}], "Tooltip"]& ], 
       TagBox[
        TooltipBox[
         TagBox[
          DynamicBox[{
            FEPrivate`If[
             CurrentValue["MouseOver"], 
             EdgeForm[{
               GrayLevel[0.5], 
               AbsoluteThickness[1.5], 
               Opacity[0.66]}], {}, {}], 
            
            RectangleBox[{36.5, 0.}, {37., 0.15623047119110112`}, 
             "RoundingRadius" -> 0]},
           
           ImageSizeCache->{{95.95246391239425, 
            100.72508710801395`}, {-49.44208501799916, 39.18756682360838}}],
          StatusArea[#, 0.15623047119110112`]& ,
          TagBoxNote->"0.15623047119110112"],
         StyleBox["0.15623047119110112`", {
           GrayLevel[0]}, StripOnInput -> False]],
        Annotation[#, 
         Style[0.15623047119110112`, {
           GrayLevel[0]}], "Tooltip"]& ], 
       TagBox[
        TooltipBox[
         TagBox[
          DynamicBox[{
            FEPrivate`If[
             CurrentValue["MouseOver"], 
             EdgeForm[{
               GrayLevel[0.5], 
               AbsoluteThickness[1.5], 
               Opacity[0.66]}], {}, {}], 
            
            RectangleBox[{37., 0.}, {37.5, 0.14073240844894389`}, 
             "RoundingRadius" -> 0]},
           
           ImageSizeCache->{{99.72508710801395, 
            104.49771030363365`}, {-40.74922355531169, 39.18756682360838}}],
          StatusArea[#, 0.14073240844894389`]& ,
          TagBoxNote->"0.14073240844894389"],
         StyleBox["0.14073240844894389`", {
           GrayLevel[0]}, StripOnInput -> False]],
        Annotation[#, 
         Style[0.14073240844894389`, {
           GrayLevel[0]}], "Tooltip"]& ], 
       TagBox[
        TooltipBox[
         TagBox[
          DynamicBox[{
            FEPrivate`If[
             CurrentValue["MouseOver"], 
             EdgeForm[{
               GrayLevel[0.5], 
               AbsoluteThickness[1.5], 
               Opacity[0.66]}], {}, {}], 
            
            RectangleBox[{37.5, 0.}, {38., 0.14598175228096488`}, 
             "RoundingRadius" -> 0]},
           
           ImageSizeCache->{{103.49771030363365`, 
            108.27033349925335`}, {-43.693579857189704`, 39.18756682360838}}],
          StatusArea[#, 0.14598175228096488`]& ,
          TagBoxNote->"0.14598175228096488"],
         StyleBox["0.14598175228096488`", {
           GrayLevel[0]}, StripOnInput -> False]],
        Annotation[#, 
         Style[0.14598175228096488`, {
           GrayLevel[0]}], "Tooltip"]& ], 
       TagBox[
        TooltipBox[
         TagBox[
          DynamicBox[{
            FEPrivate`If[
             CurrentValue["MouseOver"], 
             EdgeForm[{
               GrayLevel[0.5], 
               AbsoluteThickness[1.5], 
               Opacity[0.66]}], {}, {}], 
            
            RectangleBox[{38., 0.}, {38.5, 0.11673540807399076`}, 
             "RoundingRadius" -> 0]},
           
           ImageSizeCache->{{107.27033349925335`, 
            112.04295669487306`}, {-27.28930903244077, 39.18756682360838}}],
          StatusArea[#, 0.11673540807399076`]& ,
          TagBoxNote->"0.11673540807399076"],
         StyleBox["0.11673540807399076`", {
           GrayLevel[0]}, StripOnInput -> False]],
        Annotation[#, 
         Style[0.11673540807399076`, {
           GrayLevel[0]}], "Tooltip"]& ], 
       TagBox[
        TooltipBox[
         TagBox[
          DynamicBox[{
            FEPrivate`If[
             CurrentValue["MouseOver"], 
             EdgeForm[{
               GrayLevel[0.5], 
               AbsoluteThickness[1.5], 
               Opacity[0.66]}], {}, {}], 
            
            RectangleBox[{38.5, 0.}, {39., 0.10498687664041995`}, 
             "RoundingRadius" -> 0]},
           
           ImageSizeCache->{{111.04295669487306`, 
            115.81557989049276`}, {-20.699559213951886`, 39.18756682360838}}],
          StatusArea[#, 0.10498687664041995`]& ,
          TagBoxNote->"0.10498687664041995"],
         StyleBox["0.10498687664041995`", {
           GrayLevel[0]}, StripOnInput -> False]],
        Annotation[#, 
         Style[0.10498687664041995`, {
           GrayLevel[0]}], "Tooltip"]& ], 
       TagBox[
        TooltipBox[
         TagBox[
          DynamicBox[{
            FEPrivate`If[
             CurrentValue["MouseOver"], 
             EdgeForm[{
               GrayLevel[0.5], 
               AbsoluteThickness[1.5], 
               Opacity[0.66]}], {}, {}], 
            
            RectangleBox[{39., 0.}, {39.5, 0.08123984501937258}, 
             "RoundingRadius" -> 0]},
           
           ImageSizeCache->{{114.81557989049276`, 
            119.58820308611251`}, {-7.379852134027537, 39.18756682360838}}],
          StatusArea[#, 0.08123984501937258]& ,
          TagBoxNote->"0.08123984501937258"],
         StyleBox["0.08123984501937258`", {
           GrayLevel[0]}, StripOnInput -> False]],
        Annotation[#, 
         Style[0.08123984501937258, {
           GrayLevel[0]}], "Tooltip"]& ], 
       TagBox[
        TooltipBox[
         TagBox[
          DynamicBox[{
            FEPrivate`If[
             CurrentValue["MouseOver"], 
             EdgeForm[{
               GrayLevel[0.5], 
               AbsoluteThickness[1.5], 
               Opacity[0.66]}], {}, {}], 
            
            RectangleBox[{39.5, 0.}, {40., 0.06749156355455568}, 
             "RoundingRadius" -> 0]},
           
           ImageSizeCache->{{118.58820308611251`, 123.36082628173222`}, {
            0.3315572280339225, 39.18756682360838}}],
          StatusArea[#, 0.06749156355455568]& ,
          TagBoxNote->"0.06749156355455568"],
         StyleBox["0.06749156355455568`", {
           GrayLevel[0]}, StripOnInput -> False]],
        Annotation[#, 
         Style[0.06749156355455568, {
           GrayLevel[0]}], "Tooltip"]& ], 
       TagBox[
        TooltipBox[
         TagBox[
          DynamicBox[{
            FEPrivate`If[
             CurrentValue["MouseOver"], 
             EdgeForm[{
               GrayLevel[0.5], 
               AbsoluteThickness[1.5], 
               Opacity[0.66]}], {}, {}], 
            
            RectangleBox[{40., 0.}, {40.5, 0.06099237595300588}, 
             "RoundingRadius" -> 0]},
           
           ImageSizeCache->{{122.36082628173222`, 127.13344947735192`}, {
            3.976950744644796, 39.18756682360838}}],
          StatusArea[#, 0.06099237595300588]& ,
          TagBoxNote->"0.06099237595300588"],
         StyleBox["0.06099237595300588`", {
           GrayLevel[0]}, StripOnInput -> False]],
        Annotation[#, 
         Style[0.06099237595300588, {
           GrayLevel[0]}], "Tooltip"]& ], 
       TagBox[
        TooltipBox[
         TagBox[
          DynamicBox[{
            FEPrivate`If[
             CurrentValue["MouseOver"], 
             EdgeForm[{
               GrayLevel[0.5], 
               AbsoluteThickness[1.5], 
               Opacity[0.66]}], {}, {}], 
            
            RectangleBox[{40.5, 0.}, {41., 0.03874515685539308}, 
             "RoundingRadius" -> 0]},
           
           ImageSizeCache->{{126.13344947735192`, 130.90607267297162`}, {
            16.45541316688971, 39.18756682360838}}],
          StatusArea[#, 0.03874515685539308]& ,
          TagBoxNote->"0.03874515685539308"],
         StyleBox["0.03874515685539308`", {
           GrayLevel[0]}, StripOnInput -> False]],
        Annotation[#, 
         Style[0.03874515685539308, {
           GrayLevel[0]}], "Tooltip"]& ], 
       TagBox[
        TooltipBox[
         TagBox[
          DynamicBox[{
            FEPrivate`If[
             CurrentValue["MouseOver"], 
             EdgeForm[{
               GrayLevel[0.5], 
               AbsoluteThickness[1.5], 
               Opacity[0.66]}], {}, {}], 
            
            RectangleBox[{41., 0.}, {41.5, 0.023497062867141608`}, 
             "RoundingRadius" -> 0]},
           
           ImageSizeCache->{{129.90607267297162`, 134.67869586859132`}, {
            25.008067186630605`, 39.18756682360838}}],
          StatusArea[#, 0.023497062867141608`]& ,
          TagBoxNote->"0.023497062867141608"],
         StyleBox["0.023497062867141608`", {
           GrayLevel[0]}, StripOnInput -> False]],
        Annotation[#, 
         Style[0.023497062867141608`, {
           GrayLevel[0]}], "Tooltip"]& ], 
       TagBox[
        TooltipBox[
         TagBox[
          DynamicBox[{
            FEPrivate`If[
             CurrentValue["MouseOver"], 
             EdgeForm[{
               GrayLevel[0.5], 
               AbsoluteThickness[1.5], 
               Opacity[0.66]}], {}, {}], 
            
            RectangleBox[{41.5, 0.}, {42., 0.017747781527309086`}, 
             "RoundingRadius" -> 0]},
           
           ImageSizeCache->{{133.67869586859132`, 138.45131906421102`}, {
            28.232838374401762`, 39.18756682360838}}],
          StatusArea[#, 0.017747781527309086`]& ,
          TagBoxNote->"0.017747781527309086"],
         StyleBox["0.017747781527309086`", {
           GrayLevel[0]}, StripOnInput -> False]],
        Annotation[#, 
         Style[0.017747781527309086`, {
           GrayLevel[0]}], "Tooltip"]& ], 
       TagBox[
        TooltipBox[
         TagBox[
          DynamicBox[{
            FEPrivate`If[
             CurrentValue["MouseOver"], 
             EdgeForm[{
               GrayLevel[0.5], 
               AbsoluteThickness[1.5], 
               Opacity[0.66]}], {}, {}], 
            
            RectangleBox[{42., 0.}, {42.5, 0.012498437695288089`}, 
             "RoundingRadius" -> 0]},
           
           ImageSizeCache->{{137.45131906421102`, 142.22394225983078`}, {
            31.177194676279775`, 39.18756682360838}}],
          StatusArea[#, 0.012498437695288089`]& ,
          TagBoxNote->"0.012498437695288089"],
         StyleBox["0.012498437695288089`", {
           GrayLevel[0]}, StripOnInput -> False]],
        Annotation[#, 
         Style[0.012498437695288089`, {
           GrayLevel[0]}], "Tooltip"]& ], 
       TagBox[
        TooltipBox[
         TagBox[
          DynamicBox[{
            FEPrivate`If[
             CurrentValue["MouseOver"], 
             EdgeForm[{
               GrayLevel[0.5], 
               AbsoluteThickness[1.5], 
               Opacity[0.66]}], {}, {}], 
            
            RectangleBox[{42.5, 0.}, {43., 0.003499562554680665}, 
             "RoundingRadius" -> 0]},
           
           ImageSizeCache->{{141.22394225983078`, 145.99656545545048`}, {
            36.22466262235637, 39.18756682360838}}],
          StatusArea[#, 0.003499562554680665]& ,
          TagBoxNote->"0.003499562554680665"],
         StyleBox["0.003499562554680665`", {
           GrayLevel[0]}, StripOnInput -> False]],
        Annotation[#, 
         Style[0.003499562554680665, {
           GrayLevel[0]}], "Tooltip"]& ], 
       TagBox[
        TooltipBox[
         TagBox[
          DynamicBox[{
            FEPrivate`If[
             CurrentValue["MouseOver"], 
             EdgeForm[{
               GrayLevel[0.5], 
               AbsoluteThickness[1.5], 
               Opacity[0.66]}], {}, {}], 
            
            RectangleBox[{43., 0.}, {43.5, 0.00424946881639795}, 
             "RoundingRadius" -> 0]},
           
           ImageSizeCache->{{144.99656545545048`, 149.76918865107018`}, {
            35.80404029351665, 39.18756682360838}}],
          StatusArea[#, 0.00424946881639795]& ,
          TagBoxNote->"0.00424946881639795"],
         StyleBox["0.00424946881639795`", {
           GrayLevel[0]}, StripOnInput -> False]],
        Annotation[#, 
         Style[0.00424946881639795, {
           GrayLevel[0]}], "Tooltip"]& ], 
       TagBox[
        TooltipBox[
         TagBox[
          DynamicBox[{
            FEPrivate`If[
             CurrentValue["MouseOver"], 
             EdgeForm[{
               GrayLevel[0.5], 
               AbsoluteThickness[1.5], 
               Opacity[0.66]}], {}, {}], 
            
            RectangleBox[{43.5, 0.}, {44., 0.003499562554680665}, 
             "RoundingRadius" -> 0]},
           
           ImageSizeCache->{{148.76918865107018`, 153.54181184668988`}, {
            36.22466262235637, 39.18756682360838}}],
          StatusArea[#, 0.003499562554680665]& ,
          TagBoxNote->"0.003499562554680665"],
         StyleBox["0.003499562554680665`", {
           GrayLevel[0]}, StripOnInput -> False]],
        Annotation[#, 
         Style[0.003499562554680665, {
           GrayLevel[0]}], "Tooltip"]& ], 
       TagBox[
        TooltipBox[
         TagBox[
          DynamicBox[{
            FEPrivate`If[
             CurrentValue["MouseOver"], 
             EdgeForm[{
               GrayLevel[0.5], 
               AbsoluteThickness[1.5], 
               Opacity[0.66]}], {}, {}], 
            
            RectangleBox[{44., 0.}, {44.5, 0.001999750031246094}, 
             "RoundingRadius" -> 0]},
           
           ImageSizeCache->{{152.54181184668988`, 157.31443504230958`}, {
            37.0659072800358, 39.18756682360838}}],
          StatusArea[#, 0.001999750031246094]& ,
          TagBoxNote->"0.001999750031246094"],
         StyleBox["0.001999750031246094`", {
           GrayLevel[0]}, StripOnInput -> False]],
        Annotation[#, 
         Style[0.001999750031246094, {
           GrayLevel[0]}], "Tooltip"]& ], 
       TagBox[
        TooltipBox[
         TagBox[
          DynamicBox[{
            FEPrivate`If[
             CurrentValue["MouseOver"], 
             EdgeForm[{
               GrayLevel[0.5], 
               AbsoluteThickness[1.5], 
               Opacity[0.66]}], {}, {}], 
            
            RectangleBox[{44.5, 0.}, {45., 0.0017497812773403325`}, 
             "RoundingRadius" -> 0]},
           
           ImageSizeCache->{{156.31443504230958`, 161.08705823792928`}, {
            37.20611472298238, 39.18756682360838}}],
          StatusArea[#, 0.0017497812773403325`]& ,
          TagBoxNote->"0.0017497812773403325"],
         StyleBox["0.0017497812773403325`", {
           GrayLevel[0]}, StripOnInput -> False]],
        Annotation[#, 
         Style[0.0017497812773403325`, {
           GrayLevel[0]}], "Tooltip"]& ], 
       TagBox[
        TooltipBox[
         TagBox[
          DynamicBox[{
            FEPrivate`If[
             CurrentValue["MouseOver"], 
             EdgeForm[{
               GrayLevel[0.5], 
               AbsoluteThickness[1.5], 
               Opacity[0.66]}], {}, {}], 
            
            RectangleBox[{45., 0.}, {45.5, 0.0007499062617172854}, 
             "RoundingRadius" -> 0]},
           
           ImageSizeCache->{{160.08705823792928`, 164.85968143354904`}, {
            37.766944494768666`, 39.18756682360838}}],
          StatusArea[#, 0.0007499062617172854]& ,
          TagBoxNote->"0.0007499062617172854"],
         StyleBox["0.0007499062617172854`", {
           GrayLevel[0]}, StripOnInput -> False]],
        Annotation[#, 
         Style[0.0007499062617172854, {
           GrayLevel[0]}], "Tooltip"]& ], {}, {}, 
       TagBox[
        TooltipBox[
         TagBox[
          DynamicBox[{
            FEPrivate`If[
             CurrentValue["MouseOver"], 
             EdgeForm[{
               GrayLevel[0.5], 
               AbsoluteThickness[1.5], 
               Opacity[0.66]}], {}, {}], 
            
            RectangleBox[{46.5, 0.}, {47., 0.0004999375078115235}, 
             "RoundingRadius" -> 0]},
           
           ImageSizeCache->{{171.40492782478844`, 176.17755102040815`}, {
            37.907151937715234`, 39.18756682360838}}],
          StatusArea[#, 0.0004999375078115235]& ,
          TagBoxNote->"0.0004999375078115235"],
         StyleBox["0.0004999375078115235`", {
           GrayLevel[0]}, StripOnInput -> False]],
        Annotation[#, 
         Style[0.0004999375078115235, {
           GrayLevel[0]}], "Tooltip"]& ]}, {}, {}}, {{}, {}, 
      {RGBColor[1, 0, 0], Thickness[Large], LineBox[CompressedData["
1:eJwV13k4lG8XB3AKSVLZKW2KqGwhzIxz5jEzlkJZihLZImWXJNkLZamQQlki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        "]]}}},
    AspectRatio->NCache[GoldenRatio^(-1), 0.6180339887498948],
    Axes->{True, True},
    AxesOrigin->{26.5, 0},
    FrameTicks->{{Automatic, Automatic}, {Automatic, Automatic}},
    GridLines->{None, None},
    PlotRange->NCache[{{
        Rational[53, 2], 47}, {All, All}}, {{26.5, 47}, {All, All}}],
    PlotRangePadding->{{
       Scaled[0.02], 
       Scaled[0.02]}, {
       Scaled[0.02], 
       Scaled[0.1]}},
    Ticks->{Automatic, Automatic}]}], "}"}]], "Output",
 ImageSize->{580, 116},
 ImageMargins->{{0, 0}, {0, 0}},
 ImageRegion->{{0, 1}, {0, 1}},
 CellLabel->"Out[6]=",
 CellID->547936853]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 InterpretationBox[Cell[TextData[{
   "Neat Examples",
   "\[NonBreakingSpace]\[NonBreakingSpace]",
   Cell["(1)", "ExampleCount"]
  }], "ExampleSection"],
  $Line = 0; Null]], "ExampleSection",
 CellID->65407496],

Cell[BoxData[
 RowBox[{"Needs", "[", "\"\<Metropolis`\>\"", "]"}]], "Input",
 CellLabel->"In[1]:=",
 CellID->433947188],

Cell["\<\
This samples a Possion distribution, a truncated normal distribution, a \
bivariate normal distribution\
\>", "ExampleText",
 CellID->73418448],

Cell[CellGroupData[{

Cell[BoxData[{
 RowBox[{
  RowBox[{"Clear", "[", 
   RowBox[{
   "lambda", ",", "vv", ",", "ivv", ",", "funs", ",", "logpdf", ",", "scale", 
    ",", "cov", ",", "n", ",", "res"}], "]"}], ";"}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{"lambda", "=", "2"}], ";"}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{"vv", "=", 
   RowBox[{"{", 
    RowBox[{
     RowBox[{"{", 
      RowBox[{"1", ",", "0.99"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"0.99", ",", "1"}], "}"}]}], "}"}]}], 
  ";"}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{"ivv", "=", 
   RowBox[{"Inverse", "[", "vv", "]"}]}], ";"}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{"funs", "=", 
   RowBox[{"{", 
    RowBox[{
     RowBox[{"Function", "[", 
      RowBox[{"x", ",", 
       RowBox[{"Log", "[", 
        FractionBox[
         RowBox[{
          SuperscriptBox["lambda", "x"], " ", 
          SuperscriptBox["\[ExponentialE]", 
           RowBox[{"-", "lambda"}]]}], 
         RowBox[{"x", " ", "!"}]], "]"}]}], "]"}], ",", 
     RowBox[{"Function", "[", 
      RowBox[{"x", ",", 
       RowBox[{
        RowBox[{"-", 
         RowBox[{"x", "^", "2"}]}], "/", "2"}]}], "]"}], ",", 
     RowBox[{"Function", "[", 
      RowBox[{"x", ",", 
       RowBox[{
        RowBox[{"-", 
         RowBox[{"x", ".", "ivv", ".", "x"}]}], "/", "2"}]}], "]"}]}], 
    "}"}]}], ";"}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{"logpdf", "=", 
   RowBox[{"Function", "[", 
    RowBox[{
     RowBox[{"{", 
      RowBox[{"x", ",", "indices"}], "}"}], ",", 
     RowBox[{
      RowBox[{
       RowBox[{
        RowBox[{"#", "[", 
         RowBox[{"[", "1", "]"}], "]"}], "[", 
        RowBox[{"#", "[", 
         RowBox[{"[", "2", "]"}], "]"}], "]"}], "&"}], "/@", 
      RowBox[{"Transpose", "[", 
       RowBox[{"{", 
        RowBox[{
         RowBox[{"funs", "[", 
          RowBox[{"[", "indices", "]"}], "]"}], ",", 
         RowBox[{"x", "[", 
          RowBox[{"[", "indices", "]"}], "]"}]}], "}"}], "]"}]}]}], 
    "\[IndentingNewLine]", "]"}]}], ";"}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{"scale", "=", 
   RowBox[{"{", 
    RowBox[{"lambda", ",", "3", ",", "3"}], "}"}]}], 
  ";"}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{"cov", "=", 
   RowBox[{"{", 
    RowBox[{
     RowBox[{"{", 
      RowBox[{"{", "1", "}"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"{", "1", "}"}], "}"}], ",", 
     RowBox[{"IdentityMatrix", "[", "2", "]"}]}], "}"}]}], 
  ";"}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{"n", "=", "5000"}], ";"}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{"adpmax", "=", "2000"}], ";"}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{"x2min", "=", 
   RowBox[{"-", "0.5"}]}], ";"}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{"x0", "=", 
   RowBox[{"{", 
    RowBox[{"0", ",", "0", ",", 
     RowBox[{"{", 
      RowBox[{"0", ",", "0"}], "}"}]}], "}"}]}], ";"}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{
   RowBox[{"res", "=", 
    RowBox[{"BlockAdaptiveMetropolis", "[", 
     RowBox[{"logpdf", ",", "x0", ",", 
      RowBox[{"2", " ", "n"}], ",", "2", ",", " ", "0.5", ",", 
      RowBox[{"5", "/", " ", "adpmax"}], ",", 
      RowBox[{"InitialProposal", "[", 
       RowBox[{
        RowBox[{"{", 
         RowBox[{"0.44", ",", "0.44", ",", "0.234"}], "}"}], ",", "x0"}], 
       "]"}], ",", 
      RowBox[{"IsReturnLast", "\[Rule]", "False"}], ",", 
      RowBox[{"IsDiscrete", "\[Rule]", 
       RowBox[{"{", 
        RowBox[{"True", ",", "False", ",", "False"}], "}"}]}], ",", 
      RowBox[{"SamplingPattern", "\[Rule]", 
       RowBox[{"{", 
        RowBox[{"\"\<Cholesky\>\"", ",", "3", ",", "False"}], "}"}]}], ",", 
      RowBox[{"DomainConstraint", "\[Rule]", 
       RowBox[{"(", 
        RowBox[{
         RowBox[{"{", 
          RowBox[{
           RowBox[{
            RowBox[{"#", "[", 
             RowBox[{"[", "1", "]"}], "]"}], "\[GreaterEqual]", "0"}], ",", 
           RowBox[{
            RowBox[{"#", "[", 
             RowBox[{"[", "2", "]"}], "]"}], ">", "x2min"}], ",", "True"}], 
          "}"}], "&"}], ")"}]}]}], "]"}]}], ";"}], "//", "Timing"}]}], "Input",
 CellLabel->"In[2]:=",
 CellID->756940083],

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{"17.721000000000004`", ",", "Null"}], "}"}]], "Output",
 ImageSize->{99, 15},
 ImageMargins->{{0, 0}, {0, 0}},
 ImageRegion->{{0, 1}, {0, 1}},
 CellLabel->"Out[2]=",
 CellID->77302206]
}, Open  ]],

Cell["Accept ratio:", "ExampleText",
 CellID->24993314],

Cell[CellGroupData[{

Cell[BoxData[{
 RowBox[{"N", "[", 
  RowBox[{"Mean", "[", 
   RowBox[{"res", "[", 
    RowBox[{"[", 
     RowBox[{
      RowBox[{"n", ";;"}], ",", 
      RowBox[{"{", 
       RowBox[{"3", ",", "4"}], "}"}]}], "]"}], "]"}], "]"}], 
  "]"}], "\[IndentingNewLine]", 
 RowBox[{"ListLinePlot", "[", 
  RowBox[{
   RowBox[{"Transpose", "[", 
    RowBox[{"res", "[", 
     RowBox[{"[", 
      RowBox[{"All", ",", "5", ",", 
       RowBox[{"-", "4"}]}], "]"}], "]"}], "]"}], ",", 
   RowBox[{"PlotRange", "\[Rule]", "All"}], ",", 
   RowBox[{"AxesLabel", "\[Rule]", 
    RowBox[{"{", 
     RowBox[{"\"\<t\>\"", ",", "\"\<accept\>\""}], "}"}]}]}], "]"}]}], "Input",
 CellLabel->"In[3]:=",
 CellID->265777506],

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{
   RowBox[{"{", 
    RowBox[{
    "0.5821835632873426`", ",", "0.4318136372725455`", ",", 
     "0.23205358928214356`"}], "}"}], ",", 
   RowBox[{"{", 
    RowBox[{
    "0.4510185606647135`", ",", "0.4376746314625508`", ",", 
     "0.2329883384210813`"}], "}"}]}], "}"}]], "Output",
 ImageSize->{437, 15},
 ImageMargins->{{0, 0}, {0, 0}},
 ImageRegion->{{0, 1}, {0, 1}},
 CellLabel->"Out[3]=",
 CellID->23471563],

Cell[BoxData[
 GraphicsBox[
  TagBox[RasterBox[CompressedData["
1:eJztnYtTVNcdx5npX9J/g5kbsdQHPkJ0ijrxQbTtSoLFaQNKAOUpKMWwYXkp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    "], {{0, 119}, {180, 0}}, {0, 255},
    ColorFunction->RGBColor],
   BoxForm`ImageTag["Byte", ColorSpace -> "RGB", Interleaving -> True],
   Selectable->False],
  BaseStyle->"ImageGraphics",
  ImageSize->Automatic,
  ImageSizeRaw->{180, 119},
  PlotRange->{{0, 180}, {0, 119}}]], "Output",
 ImageSize->{184, 119},
 ImageMargins->{{0, 0}, {0, 0}},
 ImageRegion->{{0, 1}, {0, 1}},
 CellLabel->"Out[3]=",
 CellID->590140632]
}, Open  ]],

Cell["Proposal scale", "ExampleText",
 CellID->12498044],

Cell[CellGroupData[{

Cell[BoxData[{
 RowBox[{"N", "[", 
  RowBox[{"Mean", "[", 
   RowBox[{"res", "[", 
    RowBox[{"[", 
     RowBox[{
      RowBox[{"n", ";;"}], ",", "5", ",", 
      RowBox[{"-", "3"}]}], "]"}], "]"}], "]"}], "]"}], "\[IndentingNewLine]", 
 RowBox[{"ListLogPlot", "[", 
  RowBox[{
   RowBox[{"Transpose", "[", 
    RowBox[{"res", "[", 
     RowBox[{"[", 
      RowBox[{"All", ",", "5", ",", 
       RowBox[{"-", "3"}]}], "]"}], "]"}], "]"}], ",", 
   RowBox[{"PlotRange", "\[Rule]", "All"}], ",", 
   RowBox[{"Joined", "\[Rule]", "True"}], ",", 
   RowBox[{"AxesLabel", "\[Rule]", 
    RowBox[{"{", 
     RowBox[{"\"\<t\>\"", ",", "\"\<scale\>\""}], "}"}]}]}], "]"}]}], "Input",
 CellLabel->"In[4]:=",
 CellID->1468033],

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{
  "1.8000313273576973`", ",", "1.2799303921940384`", ",", 
   "2.6366003957313433`"}], "}"}]], "Output",
 ImageSize->{180, 15},
 ImageMargins->{{0, 0}, {0, 0}},
 ImageRegion->{{0, 1}, {0, 1}},
 CellLabel->"Out[4]=",
 CellID->12145802],

Cell[BoxData[
 GraphicsBox[
  TagBox[RasterBox[CompressedData["
1:eJztnYtPU1kex0n2L9l/g6RxFUdER1HHx+yw6wM3+BgD6ygOi7qg2CJQoRah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    "], {{0, 128}, {180, 0}}, {0, 255},
    ColorFunction->RGBColor],
   BoxForm`ImageTag["Byte", ColorSpace -> "RGB", Interleaving -> True],
   Selectable->False],
  BaseStyle->"ImageGraphics",
  ImageSize->Automatic,
  ImageSizeRaw->{180, 128},
  PlotRange->{{0, 180}, {0, 128}}]], "Output",
 ImageSize->{184, 128},
 ImageMargins->{{0, 0}, {0, 0}},
 ImageRegion->{{0, 1}, {0, 1}},
 CellLabel->"Out[4]=",
 CellID->5639925]
}, Open  ]],

Cell["Time series", "ExampleText",
 CellID->193398859],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"Table", "[", 
  RowBox[{
   RowBox[{"ListLinePlot", "[", 
    RowBox[{
     RowBox[{"res", "[", 
      RowBox[{"[", 
       RowBox[{"All", ",", "1", ",", "i"}], "]"}], "]"}], ",", 
     RowBox[{"PlotRange", "\[Rule]", "All"}], ",", 
     RowBox[{"AxesLabel", "\[Rule]", 
      RowBox[{"{", 
       RowBox[{"\"\<t\>\"", ",", 
        RowBox[{"\"\<x\>\"", "<>", 
         RowBox[{"ToString", "[", "i", "]"}]}]}], "}"}]}]}], "]"}], ",", 
   RowBox[{"{", 
    RowBox[{"i", ",", "2"}], "}"}]}], "]"}]], "Input",
 CellLabel->"In[5]:=",
 CellID->402775542],

Cell[BoxData[
 GraphicsBox[
  TagBox[RasterBox[CompressedData["
1:eJztnX2MFVWaxu9m/zezmZ2MZt3MLEHGsGh0QcO44hVa0KYR+RIYpIcFtIOf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    "], {{0, 134}, {385, 0}}, {0, 255},
    ColorFunction->RGBColor],
   BoxForm`ImageTag["Byte", ColorSpace -> "RGB", Interleaving -> True],
   Selectable->False],
  BaseStyle->"ImageGraphics",
  ImageSize->Automatic,
  ImageSizeRaw->{385, 134},
  PlotRange->{{0, 385}, {0, 134}}]], "Output",
 ImageSize->{389, 134},
 ImageMargins->{{0, 0}, {0, 0}},
 ImageRegion->{{0, 1}, {0, 1}},
 CellLabel->"Out[5]=",
 CellID->187378485]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"Table", "[", 
  RowBox[{
   RowBox[{"ListLinePlot", "[", 
    RowBox[{
     RowBox[{"res", "[", 
      RowBox[{"[", 
       RowBox[{"All", ",", "1", ",", 
        RowBox[{"-", "1"}], ",", "i"}], "]"}], "]"}], ",", 
     RowBox[{"PlotRange", "\[Rule]", "All"}], ",", 
     RowBox[{"AxesLabel", "\[Rule]", 
      RowBox[{"{", 
       RowBox[{"\"\<t\>\"", ",", 
        RowBox[{"\"\<x3\>\"", "<>", 
         RowBox[{"ToString", "[", "i", "]"}]}]}], "}"}]}]}], "]"}], ",", 
   RowBox[{"{", 
    RowBox[{"i", ",", "2"}], "}"}]}], "]"}]], "Input",
 CellLabel->"In[6]:=",
 CellID->223197944],

Cell[BoxData[
 GraphicsBox[
  TagBox[RasterBox[CompressedData["
1:eJztnWusHdV1x131e6RKrQRSqraI5EvhS5oiREUOjh88zMM2YCgpBWxyAwnE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    "], {{0, 118}, {385, 0}}, {0, 255},
    ColorFunction->RGBColor],
   BoxForm`ImageTag["Byte", ColorSpace -> "RGB", Interleaving -> True],
   Selectable->False],
  BaseStyle->"ImageGraphics",
  ImageSize->Automatic,
  ImageSizeRaw->{385, 118},
  PlotRange->{{0, 385}, {0, 118}}]], "Output",
 ImageSize->{389, 118},
 ImageMargins->{{0, 0}, {0, 0}},
 ImageRegion->{{0, 1}, {0, 1}},
 CellLabel->"Out[6]=",
 CellID->628946702]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{
   RowBox[{"ListPlot", "[", 
    RowBox[{
     RowBox[{"res", "[", 
      RowBox[{"[", 
       RowBox[{
        RowBox[{"adpmax", ";;"}], ",", "1", ",", 
        RowBox[{"-", "1"}]}], "]"}], "]"}], ",", 
     RowBox[{"PlotRange", "\[Rule]", "All"}], ",", 
     RowBox[{"AxesLabel", "\[Rule]", 
      RowBox[{"{", 
       RowBox[{"\"\<x31\>\"", ",", "\"\<x32\>\""}], "}"}]}]}], "]"}], ",", 
   RowBox[{
    RowBox[{"Covariance", "[", 
     RowBox[{"res", "[", 
      RowBox[{"[", 
       RowBox[{
        RowBox[{"adpmax", ";;"}], ",", "1", ",", 
        RowBox[{"-", "1"}]}], "]"}], "]"}], "]"}], "//", "MatrixForm"}]}], 
  "}"}]], "Input",
 CellLabel->"In[7]:=",
 CellID->36445872],

Cell[BoxData[
 GraphicsBox[
  TagBox[RasterBox[CompressedData["
1:eJztnc9v3MYVx7foPedeC7S35NqDAO0vSZYjK5ah1pbtpEhsOHQPhlMU8bHo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    "], {{0, 117}, {345, 0}}, {0, 255},
    ColorFunction->RGBColor],
   BoxForm`ImageTag["Byte", ColorSpace -> "RGB", Interleaving -> True],
   Selectable->False],
  BaseStyle->"ImageGraphics",
  ImageSize->Automatic,
  ImageSizeRaw->{345, 117},
  PlotRange->{{0, 345}, {0, 117}}]], "Output",
 ImageSize->{349, 117},
 ImageMargins->{{0, 0}, {0, 0}},
 ImageRegion->{{0, 1}, {0, 1}},
 CellLabel->"Out[7]=",
 CellID->69605284]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[{
 RowBox[{
  RowBox[{"data", "=", 
   RowBox[{"res", "[", 
    RowBox[{"[", 
     RowBox[{
      RowBox[{"adpmax", ";;"}], ",", "1", ",", "1"}], "]"}], "]"}]}], 
  ";"}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{"xls", "=", 
   RowBox[{"Range", "[", 
    RowBox[{
     RowBox[{"Min", "[", "data", "]"}], ",", 
     RowBox[{"Max", "[", "data", "]"}]}], "]"}]}], 
  ";"}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{"freq", "=", 
   RowBox[{
    RowBox[{
     RowBox[{"Count", "[", 
      RowBox[{"data", ",", "#"}], "]"}], "&"}], "/@", "xls"}]}], 
  ";"}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{"freq", "/=", 
   RowBox[{"Total", "[", "freq", "]"}]}], ";"}], "\[IndentingNewLine]", 
 RowBox[{"Show", "[", 
  RowBox[{
   RowBox[{"Plot", "[", 
    RowBox[{
     RowBox[{"Exp", "[", 
      RowBox[{
       RowBox[{"funs", "[", 
        RowBox[{"[", "1", "]"}], "]"}], "[", "x", "]"}], "]"}], ",", 
     RowBox[{"{", 
      RowBox[{"x", ",", 
       RowBox[{"Min", "[", "xls", "]"}], ",", 
       RowBox[{"Max", "[", "xls", "]"}]}], "}"}], ",", 
     RowBox[{"PlotStyle", "\[Rule]", "Red"}]}], "]"}], ",", 
   RowBox[{"ListPlot", "[", 
    RowBox[{
     RowBox[{"Transpose", "[", 
      RowBox[{"{", 
       RowBox[{"xls", ",", "freq"}], "}"}], "]"}], ",", 
     RowBox[{"Filling", "\[Rule]", "Axis"}], ",", 
     RowBox[{"PlotMarkers", "\[Rule]", "Automatic"}]}], "]"}], ",", 
   RowBox[{"PlotRange", "\[Rule]", "All"}]}], "]"}]}], "Input",
 CellLabel->"In[8]:=",
 CellID->182745918],

Cell[BoxData[
 GraphicsBox[
  TagBox[RasterBox[CompressedData["
1:eJztnc2LHEUYhwf8B3Lc43rb4+JBJpGY8QMZLxIFdYVgxjUsK4hGL8aLI0TZ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    "], {{0, 114}, {180, 0}}, {0, 255},
    ColorFunction->RGBColor],
   BoxForm`ImageTag["Byte", ColorSpace -> "RGB", Interleaving -> True],
   Selectable->False],
  BaseStyle->"ImageGraphics",
  ImageSize->Automatic,
  ImageSizeRaw->{180, 114},
  PlotRange->{{0, 180}, {0, 114}}]], "Output",
 ImageSize->{184, 114},
 ImageMargins->{{0, 0}, {0, 0}},
 ImageRegion->{{0, 1}, {0, 1}},
 CellLabel->"Out[8]=",
 CellID->580742033]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"Show", "[", 
  RowBox[{
   RowBox[{
    RowBox[{"data", "=", 
     RowBox[{"res", "[", 
      RowBox[{"[", 
       RowBox[{
        RowBox[{"adpmax", ";;"}], ",", "1", ",", "2"}], "]"}], "]"}]}], ";", 
    RowBox[{"Histogram", "[", 
     RowBox[{"data", ",", "Automatic", ",", "\"\<ProbabilityDensity\>\""}], 
     "]"}]}], ",", 
   RowBox[{"Plot", "[", 
    RowBox[{
     RowBox[{
      RowBox[{"PDF", "[", 
       RowBox[{
        RowBox[{"NormalDistribution", "[", 
         RowBox[{"0", ",", "1"}], "]"}], ",", "x"}], "]"}], "/", 
      RowBox[{"(", 
       RowBox[{
        FractionBox["1", "2"], " ", 
        RowBox[{"Erfc", "[", 
         FractionBox["x2min", 
          SqrtBox["2"]], "]"}]}], ")"}]}], ",", 
     RowBox[{"{", 
      RowBox[{"x", ",", 
       RowBox[{"Min", "[", "data", "]"}], ",", 
       RowBox[{"Max", "[", "data", "]"}]}], "}"}], ",", 
     RowBox[{"PlotStyle", "\[Rule]", 
      RowBox[{"{", 
       RowBox[{"Red", ",", "Thick"}], "}"}]}], ",", 
     RowBox[{"PlotRange", "\[Rule]", "All"}], ",", 
     RowBox[{"AxesLabel", "\[Rule]", 
      RowBox[{"{", 
       RowBox[{"\"\<x2\>\"", ",", "\"\<freq\>\""}], "}"}]}]}], "]"}]}], 
  "]"}]], "Input",
 CellLabel->"In[9]:=",
 CellID->530831861],

Cell[BoxData[
 GraphicsBox[{
   {RGBColor[0.798413061722744, 0.824719615472648, 0.968322270542458], 
    EdgeForm[Opacity[0.7]], {}, 
    {RGBColor[0.798413061722744, 0.824719615472648, 0.968322270542458], 
     EdgeForm[Opacity[0.7]], 
     TagBox[
      TooltipBox[
       TagBox[
        DynamicBox[{
          FEPrivate`If[
           CurrentValue["MouseOver"], 
           EdgeForm[{
             GrayLevel[0.5], 
             AbsoluteThickness[1.5], 
             Opacity[0.66]}], {}, {}], 
          
          RectangleBox[{-0.5, 0.}, {-0.4, 0.5161854768153982}, 
           "RoundingRadius" -> 0]},
         ImageSizeCache->{{15.5, 19.824489795918367`}, {-31.63270512028454, 
          42.00748516544671}}],
        StatusArea[#, 0.5161854768153982]& ,
        TagBoxNote->"0.5161854768153982"],
       StyleBox["0.5161854768153982`", {
         GrayLevel[0]}, StripOnInput -> False]],
      Annotation[#, 
       Style[0.5161854768153982, {
         GrayLevel[0]}], "Tooltip"]& ], 
     TagBox[
      TooltipBox[
       TagBox[
        DynamicBox[{
          FEPrivate`If[
           CurrentValue["MouseOver"], 
           EdgeForm[{
             GrayLevel[0.5], 
             AbsoluteThickness[1.5], 
             Opacity[0.66]}], {}, {}], 
          
          RectangleBox[{-0.4, 0.}, {-0.3, 0.5824271966004247}, 
           "RoundingRadius" -> 0]},
         ImageSizeCache->{{18.824489795918367`, 
          23.148979591836735`}, {-40.954569733223366`, 42.00748516544671}}],
        StatusArea[#, 0.5824271966004247]& ,
        TagBoxNote->"0.5824271966004247"],
       StyleBox["0.5824271966004247`", {
         GrayLevel[0]}, StripOnInput -> False]],
      Annotation[#, 
       Style[0.5824271966004247, {
         GrayLevel[0]}], "Tooltip"]& ], 
     TagBox[
      TooltipBox[
       TagBox[
        DynamicBox[{
          FEPrivate`If[
           CurrentValue["MouseOver"], 
           EdgeForm[{
             GrayLevel[0.5], 
             AbsoluteThickness[1.5], 
             Opacity[0.66]}], {}, {}], 
          
          RectangleBox[{-0.3, 0.}, {-0.2, 0.6211723534558181}, 
           "RoundingRadius" -> 0]},
         ImageSizeCache->{{22.148979591836735`, 
          26.473469387755102`}, {-46.40698111060276, 42.00748516544671}}],
        StatusArea[#, 0.6211723534558181]& ,
        TagBoxNote->"0.6211723534558181"],
       StyleBox["0.6211723534558181`", {
         GrayLevel[0]}, StripOnInput -> False]],
      Annotation[#, 
       Style[0.6211723534558181, {
         GrayLevel[0]}], "Tooltip"]& ], 
     TagBox[
      TooltipBox[
       TagBox[
        DynamicBox[{
          FEPrivate`If[
           CurrentValue["MouseOver"], 
           EdgeForm[{
             GrayLevel[0.5], 
             AbsoluteThickness[1.5], 
             Opacity[0.66]}], {}, {}], 
          
          RectangleBox[{-0.2, 0.}, {-0.1, 0.6061742282214723}, 
           "RoundingRadius" -> 0]},
         ImageSizeCache->{{25.473469387755102`, 
          29.79795918367347}, {-44.296370254842984`, 42.00748516544671}}],
        StatusArea[#, 0.6061742282214723]& ,
        TagBoxNote->"0.6061742282214723"],
       StyleBox["0.6061742282214723`", {
         GrayLevel[0]}, StripOnInput -> False]],
      Annotation[#, 
       Style[0.6061742282214723, {
         GrayLevel[0]}], "Tooltip"]& ], 
     TagBox[
      TooltipBox[
       TagBox[
        DynamicBox[{
          FEPrivate`If[
           CurrentValue["MouseOver"], 
           EdgeForm[{
             GrayLevel[0.5], 
             AbsoluteThickness[1.5], 
             Opacity[0.66]}], {}, {}], 
          
          RectangleBox[{-0.1, 0.}, {0., 0.6424196975378078}, "RoundingRadius" -> 
           0]},
         ImageSizeCache->{{28.79795918367347, 
          33.12244897959184}, {-49.39701315626239, 42.00748516544671}}],
        StatusArea[#, 0.6424196975378078]& ,
        TagBoxNote->"0.6424196975378078"],
       StyleBox["0.6424196975378078`", {
         GrayLevel[0]}, StripOnInput -> False]],
      Annotation[#, 
       Style[0.6424196975378078, {
         GrayLevel[0]}], "Tooltip"]& ], 
     TagBox[
      TooltipBox[
       TagBox[
        DynamicBox[{
          FEPrivate`If[
           CurrentValue["MouseOver"], 
           EdgeForm[{
             GrayLevel[0.5], 
             AbsoluteThickness[1.5], 
             Opacity[0.66]}], {}, {}], 
          
          RectangleBox[{0., 0.}, {0.1, 0.5986751656042995}, "RoundingRadius" -> 
           0]},
         ImageSizeCache->{{32.12244897959184, 
          36.446938775510205`}, {-43.241064826963125`, 42.00748516544671}}],
        StatusArea[#, 0.5986751656042995]& ,
        TagBoxNote->"0.5986751656042995"],
       StyleBox["0.5986751656042995`", {
         GrayLevel[0]}, StripOnInput -> False]],
      Annotation[#, 
       Style[0.5986751656042995, {
         GrayLevel[0]}], "Tooltip"]& ], 
     TagBox[
      TooltipBox[
       TagBox[
        DynamicBox[{
          FEPrivate`If[
           CurrentValue["MouseOver"], 
           EdgeForm[{
             GrayLevel[0.5], 
             AbsoluteThickness[1.5], 
             Opacity[0.66]}], {}, {}], 
          
          RectangleBox[{0.1, 0.}, {0.2, 0.5211848518935134}, "RoundingRadius" -> 
           0]},
         ImageSizeCache->{{35.446938775510205`, 
          39.77142857142857}, {-32.33624207220444, 42.00748516544671}}],
        StatusArea[#, 0.5211848518935134]& ,
        TagBoxNote->"0.5211848518935134"],
       StyleBox["0.5211848518935134`", {
         GrayLevel[0]}, StripOnInput -> False]],
      Annotation[#, 
       Style[0.5211848518935134, {
         GrayLevel[0]}], "Tooltip"]& ], 
     TagBox[
      TooltipBox[
       TagBox[
        DynamicBox[{
          FEPrivate`If[
           CurrentValue["MouseOver"], 
           EdgeForm[{
             GrayLevel[0.5], 
             AbsoluteThickness[1.5], 
             Opacity[0.66]}], {}, {}], 
          
          RectangleBox[{0.2, 0.}, {0.3, 0.5361829771278591}, "RoundingRadius" -> 
           0]},
         ImageSizeCache->{{38.77142857142857, 
          43.09591836734694}, {-34.4468529279642, 42.00748516544671}}],
        StatusArea[#, 0.5361829771278591]& ,
        TagBoxNote->"0.5361829771278591"],
       StyleBox["0.5361829771278591`", {
         GrayLevel[0]}, StripOnInput -> False]],
      Annotation[#, 
       Style[0.5361829771278591, {
         GrayLevel[0]}], "Tooltip"]& ], 
     TagBox[
      TooltipBox[
       TagBox[
        DynamicBox[{
          FEPrivate`If[
           CurrentValue["MouseOver"], 
           EdgeForm[{
             GrayLevel[0.5], 
             AbsoluteThickness[1.5], 
             Opacity[0.66]}], {}, {}], 
          
          RectangleBox[{0.3, 0.}, {0.4, 0.49618797650293694`}, 
           "RoundingRadius" -> 0]},
         ImageSizeCache->{{42.09591836734694, 
          46.42040816326531}, {-28.818557312604838`, 42.00748516544671}}],
        StatusArea[#, 0.49618797650293694`]& ,
        TagBoxNote->"0.49618797650293694"],
       StyleBox["0.49618797650293694`", {
         GrayLevel[0]}, StripOnInput -> False]],
      Annotation[#, 
       Style[0.49618797650293694`, {
         GrayLevel[0]}], "Tooltip"]& ], 
     TagBox[
      TooltipBox[
       TagBox[
        DynamicBox[{
          FEPrivate`If[
           CurrentValue["MouseOver"], 
           EdgeForm[{
             GrayLevel[0.5], 
             AbsoluteThickness[1.5], 
             Opacity[0.66]}], {}, {}], 
          
          RectangleBox[{0.4, 0.}, {0.5, 0.5599300087489065}, "RoundingRadius" -> 
           0]},
         ImageSizeCache->{{45.42040816326531, 
          49.744897959183675`}, {-37.788653449583805`, 42.00748516544671}}],
        StatusArea[#, 0.5599300087489065]& ,
        TagBoxNote->"0.5599300087489065"],
       StyleBox["0.5599300087489065`", {
         GrayLevel[0]}, StripOnInput -> False]],
      Annotation[#, 
       Style[0.5599300087489065, {
         GrayLevel[0]}], "Tooltip"]& ], 
     TagBox[
      TooltipBox[
       TagBox[
        DynamicBox[{
          FEPrivate`If[
           CurrentValue["MouseOver"], 
           EdgeForm[{
             GrayLevel[0.5], 
             AbsoluteThickness[1.5], 
             Opacity[0.66]}], {}, {}], 
          
          RectangleBox[{0.5, 0.}, {0.6, 0.511186101737283}, "RoundingRadius" -> 
           0]},
         ImageSizeCache->{{48.744897959183675`, 
          53.06938775510204}, {-30.929168168364626`, 42.00748516544671}}],
        StatusArea[#, 0.511186101737283]& ,
        TagBoxNote->"0.511186101737283"],
       StyleBox["0.511186101737283`", {
         GrayLevel[0]}, StripOnInput -> False]],
      Annotation[#, 
       Style[0.511186101737283, {
         GrayLevel[0]}], "Tooltip"]& ], 
     TagBox[
      TooltipBox[
       TagBox[
        DynamicBox[{
          FEPrivate`If[
           CurrentValue["MouseOver"], 
           EdgeForm[{
             GrayLevel[0.5], 
             AbsoluteThickness[1.5], 
             Opacity[0.66]}], {}, {}], 
          
          RectangleBox[{0.6, 0.}, {0.7, 0.46119235095613054`}, 
           "RoundingRadius" -> 0]},
         ImageSizeCache->{{52.06938775510204, 
          56.39387755102041}, {-23.89379864916546, 42.00748516544671}}],
        StatusArea[#, 0.46119235095613054`]& ,
        TagBoxNote->"0.46119235095613054"],
       StyleBox["0.46119235095613054`", {
         GrayLevel[0]}, StripOnInput -> False]],
      Annotation[#, 
       Style[0.46119235095613054`, {
         GrayLevel[0]}], "Tooltip"]& ], 
     TagBox[
      TooltipBox[
       TagBox[
        DynamicBox[{
          FEPrivate`If[
           CurrentValue["MouseOver"], 
           EdgeForm[{
             GrayLevel[0.5], 
             AbsoluteThickness[1.5], 
             Opacity[0.66]}], {}, {}], 
          
          RectangleBox[{0.7, 0.}, {0.8, 0.39370078740157444`}, 
           "RoundingRadius" -> 0]},
         ImageSizeCache->{{55.39387755102041, 
          59.71836734693878}, {-14.396049798246551`, 42.00748516544671}}],
        StatusArea[#, 0.39370078740157444`]& ,
        TagBoxNote->"0.39370078740157444"],
       StyleBox["0.39370078740157444`", {
         GrayLevel[0]}, StripOnInput -> False]],
      Annotation[#, 
       Style[0.39370078740157444`, {
         GrayLevel[0]}], "Tooltip"]& ], 
     TagBox[
      TooltipBox[
       TagBox[
        DynamicBox[{
          FEPrivate`If[
           CurrentValue["MouseOver"], 
           EdgeForm[{
             GrayLevel[0.5], 
             AbsoluteThickness[1.5], 
             Opacity[0.66]}], {}, {}], 
          
          RectangleBox[{0.8, 0.}, {0.9, 0.3487064116985378}, "RoundingRadius" -> 
           0]},
         ImageSizeCache->{{58.71836734693878, 
          63.042857142857144`}, {-8.06421723096738, 42.00748516544671}}],
        StatusArea[#, 0.3487064116985378]& ,
        TagBoxNote->"0.3487064116985378"],
       StyleBox["0.3487064116985378`", {
         GrayLevel[0]}, StripOnInput -> False]],
      Annotation[#, 
       Style[0.3487064116985378, {
         GrayLevel[0]}], "Tooltip"]& ], 
     TagBox[
      TooltipBox[
       TagBox[
        DynamicBox[{
          FEPrivate`If[
           CurrentValue["MouseOver"], 
           EdgeForm[{
             GrayLevel[0.5], 
             AbsoluteThickness[1.5], 
             Opacity[0.66]}], {}, {}], 
          
          RectangleBox[{0.9, 0.}, {1., 0.3299587551556056}, "RoundingRadius" -> 
           0]},
         ImageSizeCache->{{62.042857142857144`, 
          66.36734693877551}, {-5.425953661267691, 42.00748516544671}}],
        StatusArea[#, 0.3299587551556056]& ,
        TagBoxNote->"0.3299587551556056"],
       StyleBox["0.3299587551556056`", {
         GrayLevel[0]}, StripOnInput -> False]],
      Annotation[#, 
       Style[0.3299587551556056, {
         GrayLevel[0]}], "Tooltip"]& ], 
     TagBox[
      TooltipBox[
       TagBox[
        DynamicBox[{
          FEPrivate`If[
           CurrentValue["MouseOver"], 
           EdgeForm[{
             GrayLevel[0.5], 
             AbsoluteThickness[1.5], 
             Opacity[0.66]}], {}, {}], 
          
          RectangleBox[{1., 0.}, {1.1, 0.32620922384701884`}, 
           "RoundingRadius" -> 0]},
         ImageSizeCache->{{65.36734693877551, 
          69.69183673469388}, {-4.898300947327712, 42.00748516544671}}],
        StatusArea[#, 0.32620922384701884`]& ,
        TagBoxNote->"0.32620922384701884"],
       StyleBox["0.32620922384701884`", {
         GrayLevel[0]}, StripOnInput -> False]],
      Annotation[#, 
       Style[0.32620922384701884`, {
         GrayLevel[0]}], "Tooltip"]& ], 
     TagBox[
      TooltipBox[
       TagBox[
        DynamicBox[{
          FEPrivate`If[
           CurrentValue["MouseOver"], 
           EdgeForm[{
             GrayLevel[0.5], 
             AbsoluteThickness[1.5], 
             Opacity[0.66]}], {}, {}], 
          
          RectangleBox[{1.1, 0.}, {1.2, 0.2799650043744536}, "RoundingRadius" -> 
           0]},
         ImageSizeCache->{{68.69183673469388, 73.01632653061225}, {
          1.6094158579314097`, 42.00748516544671}}],
        StatusArea[#, 0.2799650043744536]& ,
        TagBoxNote->"0.2799650043744536"],
       StyleBox["0.2799650043744536`", {
         GrayLevel[0]}, StripOnInput -> False]],
      Annotation[#, 
       Style[0.2799650043744536, {
         GrayLevel[0]}], "Tooltip"]& ], 
     TagBox[
      TooltipBox[
       TagBox[
        DynamicBox[{
          FEPrivate`If[
           CurrentValue["MouseOver"], 
           EdgeForm[{
             GrayLevel[0.5], 
             AbsoluteThickness[1.5], 
             Opacity[0.66]}], {}, {}], 
          
          RectangleBox[{1.2, 0.}, {1.3, 0.2874640669916258}, "RoundingRadius" -> 
           0]},
         ImageSizeCache->{{72.01632653061225, 76.34081632653061}, {
          0.5541104300516295, 42.00748516544671}}],
        StatusArea[#, 0.2874640669916258]& ,
        TagBoxNote->"0.2874640669916258"],
       StyleBox["0.2874640669916258`", {
         GrayLevel[0]}, StripOnInput -> False]],
      Annotation[#, 
       Style[0.2874640669916258, {
         GrayLevel[0]}], "Tooltip"]& ], 
     TagBox[
      TooltipBox[
       TagBox[
        DynamicBox[{
          FEPrivate`If[
           CurrentValue["MouseOver"], 
           EdgeForm[{
             GrayLevel[0.5], 
             AbsoluteThickness[1.5], 
             Opacity[0.66]}], {}, {}], 
          
          RectangleBox[{1.3, 0.}, {1.4, 0.28371453568304}, "RoundingRadius" -> 
           0]},
         ImageSizeCache->{{75.34081632653061, 79.66530612244898}, {
          1.0817631439914734`, 42.00748516544671}}],
        StatusArea[#, 0.28371453568304]& ,
        TagBoxNote->"0.28371453568304"],
       StyleBox["0.28371453568304`", {
         GrayLevel[0]}, StripOnInput -> False]],
      Annotation[#, 
       Style[0.28371453568304, {
         GrayLevel[0]}], "Tooltip"]& ], 
     TagBox[
      TooltipBox[
       TagBox[
        DynamicBox[{
          FEPrivate`If[
           CurrentValue["MouseOver"], 
           EdgeForm[{
             GrayLevel[0.5], 
             AbsoluteThickness[1.5], 
             Opacity[0.66]}], {}, {}], 
          
          RectangleBox[{1.4, 0.}, {1.5, 0.19247594050743638`}, 
           "RoundingRadius" -> 0]},
         ImageSizeCache->{{78.66530612244898, 82.98979591836735}, {
          13.921312516530005`, 42.00748516544671}}],
        StatusArea[#, 0.19247594050743638`]& ,
        TagBoxNote->"0.19247594050743638"],
       StyleBox["0.19247594050743638`", {
         GrayLevel[0]}, StripOnInput -> False]],
      Annotation[#, 
       Style[0.19247594050743638`, {
         GrayLevel[0]}], "Tooltip"]& ], 
     TagBox[
      TooltipBox[
       TagBox[
        DynamicBox[{
          FEPrivate`If[
           CurrentValue["MouseOver"], 
           EdgeForm[{
             GrayLevel[0.5], 
             AbsoluteThickness[1.5], 
             Opacity[0.66]}], {}, {}], 
          
          RectangleBox[{1.5, 0.}, {1.6, 0.17997750281214833`}, 
           "RoundingRadius" -> 0]},
         ImageSizeCache->{{81.98979591836735, 86.31428571428572}, {
          15.680154896329789`, 42.00748516544671}}],
        StatusArea[#, 0.17997750281214833`]& ,
        TagBoxNote->"0.17997750281214833"],
       StyleBox["0.17997750281214833`", {
         GrayLevel[0]}, StripOnInput -> False]],
      Annotation[#, 
       Style[0.17997750281214833`, {
         GrayLevel[0]}], "Tooltip"]& ], 
     TagBox[
      TooltipBox[
       TagBox[
        DynamicBox[{
          FEPrivate`If[
           CurrentValue["MouseOver"], 
           EdgeForm[{
             GrayLevel[0.5], 
             AbsoluteThickness[1.5], 
             Opacity[0.66]}], {}, {}], 
          
          RectangleBox[{1.6, 0.}, {1.7, 0.1162354705661794}, "RoundingRadius" -> 
           0]},
         ImageSizeCache->{{85.31428571428572, 89.63877551020408}, {
          24.65025103330866, 42.00748516544671}}],
        StatusArea[#, 0.1162354705661794]& ,
        TagBoxNote->"0.1162354705661794"],
       StyleBox["0.1162354705661794`", {
         GrayLevel[0]}, StripOnInput -> False]],
      Annotation[#, 
       Style[0.1162354705661794, {
         GrayLevel[0]}], "Tooltip"]& ], 
     TagBox[
      TooltipBox[
       TagBox[
        DynamicBox[{
          FEPrivate`If[
           CurrentValue["MouseOver"], 
           EdgeForm[{
             GrayLevel[0.5], 
             AbsoluteThickness[1.5], 
             Opacity[0.66]}], {}, {}], 
          
          RectangleBox[{1.7, 0.}, {1.8, 0.12248468941382315`}, 
           "RoundingRadius" -> 0]},
         ImageSizeCache->{{88.63877551020408, 92.96326530612245}, {
          23.77082984340881, 42.00748516544671}}],
        StatusArea[#, 0.12248468941382315`]& ,
        TagBoxNote->"0.12248468941382315"],
       StyleBox["0.12248468941382315`", {
         GrayLevel[0]}, StripOnInput -> False]],
      Annotation[#, 
       Style[0.12248468941382315`, {
         GrayLevel[0]}], "Tooltip"]& ], 
     TagBox[
      TooltipBox[
       TagBox[
        DynamicBox[{
          FEPrivate`If[
           CurrentValue["MouseOver"], 
           EdgeForm[{
             GrayLevel[0.5], 
             AbsoluteThickness[1.5], 
             Opacity[0.66]}], {}, {}], 
          
          RectangleBox[{1.8, 0.}, {1.9, 0.1049868766404201}, "RoundingRadius" -> 
           0]},
         ImageSizeCache->{{91.96326530612245, 96.28775510204082}, {
          26.233209175128472`, 42.00748516544671}}],
        StatusArea[#, 0.1049868766404201]& ,
        TagBoxNote->"0.1049868766404201"],
       StyleBox["0.1049868766404201`", {
         GrayLevel[0]}, StripOnInput -> False]],
      Annotation[#, 
       Style[0.1049868766404201, {
         GrayLevel[0]}], "Tooltip"]& ], 
     TagBox[
      TooltipBox[
       TagBox[
        DynamicBox[{
          FEPrivate`If[
           CurrentValue["MouseOver"], 
           EdgeForm[{
             GrayLevel[0.5], 
             AbsoluteThickness[1.5], 
             Opacity[0.66]}], {}, {}], 
          
          RectangleBox[{1.9, 0.}, {2., 0.09373828271466059}, "RoundingRadius" -> 
           0]},
         ImageSizeCache->{{95.28775510204082, 99.61224489795919}, {
          27.816167316948313`, 42.00748516544671}}],
        StatusArea[#, 0.09373828271466059]& ,
        TagBoxNote->"0.09373828271466059"],
       StyleBox["0.09373828271466059`", {
         GrayLevel[0]}, StripOnInput -> False]],
      Annotation[#, 
       Style[0.09373828271466059, {
         GrayLevel[0]}], "Tooltip"]& ], 
     TagBox[
      TooltipBox[
       TagBox[
        DynamicBox[{
          FEPrivate`If[
           CurrentValue["MouseOver"], 
           EdgeForm[{
             GrayLevel[0.5], 
             AbsoluteThickness[1.5], 
             Opacity[0.66]}], {}, {}], 
          
          RectangleBox[{2., 0.}, {2.1, 0.057492813398325154`}, 
           "RoundingRadius" -> 0]},
         ImageSizeCache->{{98.61224489795919, 102.93673469387755`}, {
          32.916810218367694`, 42.00748516544671}}],
        StatusArea[#, 0.057492813398325154`]& ,
        TagBoxNote->"0.057492813398325154"],
       StyleBox["0.057492813398325154`", {
         GrayLevel[0]}, StripOnInput -> False]],
      Annotation[#, 
       Style[0.057492813398325154`, {
         GrayLevel[0]}], "Tooltip"]& ], 
     TagBox[
      TooltipBox[
       TagBox[
        DynamicBox[{
          FEPrivate`If[
           CurrentValue["MouseOver"], 
           EdgeForm[{
             GrayLevel[0.5], 
             AbsoluteThickness[1.5], 
             Opacity[0.66]}], {}, {}], 
          
          RectangleBox[{2.1, 0.}, {2.2, 0.053743282089738734`}, 
           "RoundingRadius" -> 0]},
         ImageSizeCache->{{101.93673469387755`, 106.26122448979592`}, {
          33.44446293230763, 42.00748516544671}}],
        StatusArea[#, 0.053743282089738734`]& ,
        TagBoxNote->"0.053743282089738734"],
       StyleBox["0.053743282089738734`", {
         GrayLevel[0]}, StripOnInput -> False]],
      Annotation[#, 
       Style[0.053743282089738734`, {
         GrayLevel[0]}], "Tooltip"]& ], 
     TagBox[
      TooltipBox[
       TagBox[
        DynamicBox[{
          FEPrivate`If[
           CurrentValue["MouseOver"], 
           EdgeForm[{
             GrayLevel[0.5], 
             AbsoluteThickness[1.5], 
             Opacity[0.66]}], {}, {}], 
          
          RectangleBox[{2.2, 0.}, {2.3, 0.051243594550681344`}, 
           "RoundingRadius" -> 0]},
         ImageSizeCache->{{105.26122448979592`, 109.58571428571429`}, {
          33.79623140826755, 42.00748516544671}}],
        StatusArea[#, 0.051243594550681344`]& ,
        TagBoxNote->"0.051243594550681344"],
       StyleBox["0.051243594550681344`", {
         GrayLevel[0]}, StripOnInput -> False]],
      Annotation[#, 
       Style[0.051243594550681344`, {
         GrayLevel[0]}], "Tooltip"]& ], 
     TagBox[
      TooltipBox[
       TagBox[
        DynamicBox[{
          FEPrivate`If[
           CurrentValue["MouseOver"], 
           EdgeForm[{
             GrayLevel[0.5], 
             AbsoluteThickness[1.5], 
             Opacity[0.66]}], {}, {}], 
          
          RectangleBox[{2.3, 0.}, {2.4, 0.034995625546806616`}, 
           "RoundingRadius" -> 0]},
         ImageSizeCache->{{108.58571428571429`, 112.91020408163266`}, {
          36.08272650200731, 42.00748516544671}}],
        StatusArea[#, 0.034995625546806616`]& ,
        TagBoxNote->"0.034995625546806616"],
       StyleBox["0.034995625546806616`", {
         GrayLevel[0]}, StripOnInput -> False]],
      Annotation[#, 
       Style[0.034995625546806616`, {
         GrayLevel[0]}], "Tooltip"]& ], 
     TagBox[
      TooltipBox[
       TagBox[
        DynamicBox[{
          FEPrivate`If[
           CurrentValue["MouseOver"], 
           EdgeForm[{
             GrayLevel[0.5], 
             AbsoluteThickness[1.5], 
             Opacity[0.66]}], {}, {}], 
          
          RectangleBox[{2.4, 0.}, {2.5, 0.02249718785151854}, 
           "RoundingRadius" -> 0]},
         ImageSizeCache->{{111.91020408163266`, 116.23469387755102`}, {
          37.84156888180709, 42.00748516544671}}],
        StatusArea[#, 0.02249718785151854]& ,
        TagBoxNote->"0.02249718785151854"],
       StyleBox["0.02249718785151854`", {
         GrayLevel[0]}, StripOnInput -> False]],
      Annotation[#, 
       Style[0.02249718785151854, {
         GrayLevel[0]}], "Tooltip"]& ], 
     TagBox[
      TooltipBox[
       TagBox[
        DynamicBox[{
          FEPrivate`If[
           CurrentValue["MouseOver"], 
           EdgeForm[{
             GrayLevel[0.5], 
             AbsoluteThickness[1.5], 
             Opacity[0.66]}], {}, {}], 
          
          RectangleBox[{2.5, 0.}, {2.6, 0.017497812773403308`}, 
           "RoundingRadius" -> 0]},
         ImageSizeCache->{{115.23469387755102`, 119.55918367346939`}, {
          38.54510583372701, 42.00748516544671}}],
        StatusArea[#, 0.017497812773403308`]& ,
        TagBoxNote->"0.017497812773403308"],
       StyleBox["0.017497812773403308`", {
         GrayLevel[0]}, StripOnInput -> False]],
      Annotation[#, 
       Style[0.017497812773403308`, {
         GrayLevel[0]}], "Tooltip"]& ], 
     TagBox[
      TooltipBox[
       TagBox[
        DynamicBox[{
          FEPrivate`If[
           CurrentValue["MouseOver"], 
           EdgeForm[{
             GrayLevel[0.5], 
             AbsoluteThickness[1.5], 
             Opacity[0.66]}], {}, {}], 
          
          RectangleBox[{2.6, 0.}, {2.7, 0.01124859392575927}, 
           "RoundingRadius" -> 0]},
         ImageSizeCache->{{118.55918367346939`, 122.88367346938776`}, {
          39.4245270236269, 42.00748516544671}}],
        StatusArea[#, 0.01124859392575927]& ,
        TagBoxNote->"0.01124859392575927"],
       StyleBox["0.01124859392575927`", {
         GrayLevel[0]}, StripOnInput -> False]],
      Annotation[#, 
       Style[0.01124859392575927, {
         GrayLevel[0]}], "Tooltip"]& ], {}, 
     TagBox[
      TooltipBox[
       TagBox[
        DynamicBox[{
          FEPrivate`If[
           CurrentValue["MouseOver"], 
           EdgeForm[{
             GrayLevel[0.5], 
             AbsoluteThickness[1.5], 
             Opacity[0.66]}], {}, {}], 
          
          RectangleBox[{2.8, 0.}, {2.9, 0.004999375078115231}, 
           "RoundingRadius" -> 0]},
         ImageSizeCache->{{125.20816326530613`, 129.5326530612245}, {
          40.303948213526795`, 42.00748516544671}}],
        StatusArea[#, 0.004999375078115231]& ,
        TagBoxNote->"0.004999375078115231"],
       StyleBox["0.004999375078115231`", {
         GrayLevel[0]}, StripOnInput -> False]],
      Annotation[#, 
       Style[0.004999375078115231, {
         GrayLevel[0]}], "Tooltip"]& ], 
     TagBox[
      TooltipBox[
       TagBox[
        DynamicBox[{
          FEPrivate`If[
           CurrentValue["MouseOver"], 
           EdgeForm[{
             GrayLevel[0.5], 
             AbsoluteThickness[1.5], 
             Opacity[0.66]}], {}, {}], 
          
          RectangleBox[{2.9, 0.}, {3., 0.009998750156230462}, 
           "RoundingRadius" -> 0]},
         ImageSizeCache->{{128.5326530612245, 132.85714285714286`}, {
          39.60041126160688, 42.00748516544671}}],
        StatusArea[#, 0.009998750156230462]& ,
        TagBoxNote->"0.009998750156230462"],
       StyleBox["0.009998750156230462`", {
         GrayLevel[0]}, StripOnInput -> False]],
      Annotation[#, 
       Style[0.009998750156230462, {
         GrayLevel[0]}], "Tooltip"]& ], 
     TagBox[
      TooltipBox[
       TagBox[
        DynamicBox[{
          FEPrivate`If[
           CurrentValue["MouseOver"], 
           EdgeForm[{
             GrayLevel[0.5], 
             AbsoluteThickness[1.5], 
             Opacity[0.66]}], {}, {}], 
          
          RectangleBox[{3., 0.}, {3.1, 0.008748906386701654}, 
           "RoundingRadius" -> 0]},
         ImageSizeCache->{{131.85714285714286`, 136.18163265306123`}, {
          39.77629549958686, 42.00748516544671}}],
        StatusArea[#, 0.008748906386701654]& ,
        TagBoxNote->"0.008748906386701654"],
       StyleBox["0.008748906386701654`", {
         GrayLevel[0]}, StripOnInput -> False]],
      Annotation[#, 
       Style[0.008748906386701654, {
         GrayLevel[0]}], "Tooltip"]& ], 
     TagBox[
      TooltipBox[
       TagBox[
        DynamicBox[{
          FEPrivate`If[
           CurrentValue["MouseOver"], 
           EdgeForm[{
             GrayLevel[0.5], 
             AbsoluteThickness[1.5], 
             Opacity[0.66]}], {}, {}], 
          
          RectangleBox[{3.1, 0.}, {3.2, 0.0037495313085864237`}, 
           "RoundingRadius" -> 0]},
         ImageSizeCache->{{135.18163265306123`, 139.5061224489796}, {
          40.47983245150677, 42.00748516544671}}],
        StatusArea[#, 0.0037495313085864237`]& ,
        TagBoxNote->"0.0037495313085864237"],
       StyleBox["0.0037495313085864237`", {
         GrayLevel[0]}, StripOnInput -> False]],
      Annotation[#, 
       Style[0.0037495313085864237`, {
         GrayLevel[0]}], "Tooltip"]& ], 
     TagBox[
      TooltipBox[
       TagBox[
        DynamicBox[{
          FEPrivate`If[
           CurrentValue["MouseOver"], 
           EdgeForm[{
             GrayLevel[0.5], 
             AbsoluteThickness[1.5], 
             Opacity[0.66]}], {}, {}], 
          
          RectangleBox[{3.2, 0.}, {3.3, 0.0049993750781152525`}, 
           "RoundingRadius" -> 0]},
         ImageSizeCache->{{138.5061224489796, 142.83061224489796`}, {
          40.303948213526795`, 42.00748516544671}}],
        StatusArea[#, 0.0049993750781152525`]& ,
        TagBoxNote->"0.0049993750781152525"],
       StyleBox["0.0049993750781152525`", {
         GrayLevel[0]}, StripOnInput -> False]],
      Annotation[#, 
       Style[0.0049993750781152525`, {
         GrayLevel[0]}], "Tooltip"]& ], 
     TagBox[
      TooltipBox[
       TagBox[
        DynamicBox[{
          FEPrivate`If[
           CurrentValue["MouseOver"], 
           EdgeForm[{
             GrayLevel[0.5], 
             AbsoluteThickness[1.5], 
             Opacity[0.66]}], {}, {}], 
          
          RectangleBox[{3.3, 0.}, {3.4, 0.0024996875390576154`}, 
           "RoundingRadius" -> 0]},
         ImageSizeCache->{{141.83061224489796`, 146.15510204081633`}, {
          40.65571668948675, 42.00748516544671}}],
        StatusArea[#, 0.0024996875390576154`]& ,
        TagBoxNote->"0.0024996875390576154"],
       StyleBox["0.0024996875390576154`", {
         GrayLevel[0]}, StripOnInput -> False]],
      Annotation[#, 
       Style[0.0024996875390576154`, {
         GrayLevel[0]}], "Tooltip"]& ], {}, {}, {}, 
     TagBox[
      TooltipBox[
       TagBox[
        DynamicBox[{
          FEPrivate`If[
           CurrentValue["MouseOver"], 
           EdgeForm[{
             GrayLevel[0.5], 
             AbsoluteThickness[1.5], 
             Opacity[0.66]}], {}, {}], 
          
          RectangleBox[{3.7, 0.}, {3.8, 0.0024996875390576262`}, 
           "RoundingRadius" -> 0]},
         ImageSizeCache->{{155.12857142857143`, 159.4530612244898}, {
          40.65571668948675, 42.00748516544671}}],
        StatusArea[#, 0.0024996875390576262`]& ,
        TagBoxNote->"0.0024996875390576262"],
       StyleBox["0.0024996875390576262`", {
         GrayLevel[0]}, StripOnInput -> False]],
      Annotation[#, 
       Style[0.0024996875390576262`, {
         GrayLevel[0]}], "Tooltip"]& ], {}, {}, {}, {}, 
     TagBox[
      TooltipBox[
       TagBox[
        DynamicBox[{
          FEPrivate`If[
           CurrentValue["MouseOver"], 
           EdgeForm[{
             GrayLevel[0.5], 
             AbsoluteThickness[1.5], 
             Opacity[0.66]}], {}, {}], 
          
          RectangleBox[{4.2, 0.}, {4.3, 0.0012498437695288131`}, 
           "RoundingRadius" -> 0]},
         ImageSizeCache->{{171.75102040816327`, 176.07551020408164`}, {
          40.83160092746673, 42.00748516544671}}],
        StatusArea[#, 0.0012498437695288131`]& ,
        TagBoxNote->"0.0012498437695288131"],
       StyleBox["0.0012498437695288131`", {
         GrayLevel[0]}, StripOnInput -> False]],
      Annotation[#, 
       Style[0.0012498437695288131`, {
         GrayLevel[0]}], "Tooltip"]& ]}, {}, {}}, {{}, {}, 
    {RGBColor[1, 0, 0], Thickness[Large], LineBox[CompressedData["
1:eJwV13c81d8fB3Djuua9Lpkpe5TsEYnepzKSrMxKSTIyIlvIzMyo6CsVQtll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      "]]}}},
  AspectRatio->NCache[GoldenRatio^(-1), 0.6180339887498948],
  Axes->{True, True},
  AxesOrigin->{-0.5, 0},
  FrameTicks->{{Automatic, Automatic}, {Automatic, Automatic}},
  GridLines->{None, None},
  PlotRange->NCache[{{
      Rational[-1, 2], 
      Rational[43, 10]}, {All, All}}, {{-0.5, 4.3}, {All, All}}],
  PlotRangePadding->{{
     Scaled[0.02], 
     Scaled[0.02]}, {
     Scaled[0.02], 
     Scaled[0.1]}},
  Ticks->{Automatic, Automatic}]], "Output",
 ImageSize->{184, 121},
 ImageMargins->{{0, 0}, {0, 0}},
 ImageRegion->{{0, 1}, {0, 1}},
 CellLabel->"Out[9]=",
 CellID->193848635]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"Table", "[", 
  RowBox[{
   RowBox[{
    RowBox[{"data", "=", 
     RowBox[{"res", "[", 
      RowBox[{"[", 
       RowBox[{
        RowBox[{"adpmax", ";;"}], ",", "1", ",", 
        RowBox[{"-", "1"}], ",", "i"}], "]"}], "]"}]}], ";", 
    "\[IndentingNewLine]", 
    RowBox[{"Show", "[", 
     RowBox[{
      RowBox[{"Histogram", "[", 
       RowBox[{"data", ",", "Automatic", ",", "\"\<ProbabilityDensity\>\""}], 
       "]"}], ",", 
      RowBox[{"Plot", "[", 
       RowBox[{
        RowBox[{"PDF", "[", 
         RowBox[{
          RowBox[{"NormalDistribution", "[", 
           RowBox[{"0", ",", "1"}], "]"}], ",", "x"}], "]"}], ",", 
        RowBox[{"{", 
         RowBox[{"x", ",", 
          RowBox[{"Min", "[", "data", "]"}], ",", 
          RowBox[{"Max", "[", "data", "]"}]}], "}"}], ",", 
        RowBox[{"PlotStyle", "\[Rule]", 
         RowBox[{"{", 
          RowBox[{"Red", ",", "Thick"}], "}"}]}], ",", 
        RowBox[{"PlotRange", "\[Rule]", "All"}], ",", 
        RowBox[{"AxesLabel", "\[Rule]", 
         RowBox[{"{", 
          RowBox[{
           RowBox[{"\"\<x3\>\"", "<>", 
            RowBox[{"ToString", "[", "i", "]"}]}], ",", "\"\<freq\>\""}], 
          "}"}]}]}], "]"}]}], "]"}]}], ",", 
   RowBox[{"{", 
    RowBox[{"i", ",", "2"}], "}"}]}], "]"}]], "Input",
 CellLabel->"In[10]:=",
 CellID->271797975],

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{
   GraphicsBox[{
     {RGBColor[0.798413061722744, 0.824719615472648, 0.968322270542458], 
      EdgeForm[Opacity[0.7]], {}, 
      {RGBColor[0.798413061722744, 0.824719615472648, 0.968322270542458], 
       EdgeForm[Opacity[0.7]], 
       TagBox[
        TooltipBox[
         TagBox[
          DynamicBox[{
            FEPrivate`If[
             CurrentValue["MouseOver"], 
             EdgeForm[{
               GrayLevel[0.5], 
               AbsoluteThickness[1.5], 
               Opacity[0.66]}], {}, {}], 
            
            RectangleBox[{-3.8, 0.}, {-3.6, 0.0006249218847644053}, 
             "RoundingRadius" -> 0]},
           
           ImageSizeCache->{{15.499999999999986`, 20.93265306122447}, {
            39.15094223955742, 40.27773676735786}}],
          StatusArea[#, 0.0006249218847644053]& ,
          TagBoxNote->"0.0006249218847644053"],
         StyleBox["0.0006249218847644053`", {
           GrayLevel[0]}, StripOnInput -> False]],
        Annotation[#, 
         Style[0.0006249218847644053, {
           GrayLevel[0]}], "Tooltip"]& ], {}, 
       TagBox[
        TooltipBox[
         TagBox[
          DynamicBox[{
            FEPrivate`If[
             CurrentValue["MouseOver"], 
             EdgeForm[{
               GrayLevel[0.5], 
               AbsoluteThickness[1.5], 
               Opacity[0.66]}], {}, {}], 
            
            RectangleBox[{-3.4, 0.}, {-3.2, 0.0018747656542932162`}, 
             "RoundingRadius" -> 0]},
           
           ImageSizeCache->{{24.365306122448956`, 29.79795918367344}, {
            38.897353183956554`, 40.27773676735786}}],
          StatusArea[#, 0.0018747656542932162`]& ,
          TagBoxNote->"0.0018747656542932162"],
         StyleBox["0.0018747656542932162`", {
           GrayLevel[0]}, StripOnInput -> False]],
        Annotation[#, 
         Style[0.0018747656542932162`, {
           GrayLevel[0]}], "Tooltip"]& ], 
       TagBox[
        TooltipBox[
         TagBox[
          DynamicBox[{
            FEPrivate`If[
             CurrentValue["MouseOver"], 
             EdgeForm[{
               GrayLevel[0.5], 
               AbsoluteThickness[1.5], 
               Opacity[0.66]}], {}, {}], 
            
            RectangleBox[{-3.2, 0.}, {-3., 0.006874140732408443}, 
             "RoundingRadius" -> 0]},
           
           ImageSizeCache->{{28.79795918367344, 34.23061224489794}, {
            37.88299696155309, 40.27773676735786}}],
          StatusArea[#, 0.006874140732408443]& ,
          TagBoxNote->"0.006874140732408443"],
         StyleBox["0.006874140732408443`", {
           GrayLevel[0]}, StripOnInput -> False]],
        Annotation[#, 
         Style[0.006874140732408443, {
           GrayLevel[0]}], "Tooltip"]& ], 
       TagBox[
        TooltipBox[
         TagBox[
          DynamicBox[{
            FEPrivate`If[
             CurrentValue["MouseOver"], 
             EdgeForm[{
               GrayLevel[0.5], 
               AbsoluteThickness[1.5], 
               Opacity[0.66]}], {}, {}], 
            
            RectangleBox[{-3., 0.}, {-2.8, 0.006874140732408443}, 
             "RoundingRadius" -> 0]},
           
           ImageSizeCache->{{33.23061224489794, 38.66326530612243}, {
            37.88299696155309, 40.27773676735786}}],
          StatusArea[#, 0.006874140732408443]& ,
          TagBoxNote->"0.006874140732408443"],
         StyleBox["0.006874140732408443`", {
           GrayLevel[0]}, StripOnInput -> False]],
        Annotation[#, 
         Style[0.006874140732408443, {
           GrayLevel[0]}], "Tooltip"]& ], 
       TagBox[
        TooltipBox[
         TagBox[
          DynamicBox[{
            FEPrivate`If[
             CurrentValue["MouseOver"], 
             EdgeForm[{
               GrayLevel[0.5], 
               AbsoluteThickness[1.5], 
               Opacity[0.66]}], {}, {}], 
            RectangleBox[{-2.8, 0.}, {-2.6, 0.007499062617172865}, 
             "RoundingRadius" -> 0]},
           
           ImageSizeCache->{{37.66326530612243, 43.09591836734692}, {
            37.756202433752655`, 40.27773676735786}}],
          StatusArea[#, 0.007499062617172865]& ,
          TagBoxNote->"0.007499062617172865"],
         StyleBox["0.007499062617172865`", {
           GrayLevel[0]}, StripOnInput -> False]],
        Annotation[#, 
         Style[0.007499062617172865, {
           GrayLevel[0]}], "Tooltip"]& ], 
       TagBox[
        TooltipBox[
         TagBox[
          DynamicBox[{
            FEPrivate`If[
             CurrentValue["MouseOver"], 
             EdgeForm[{
               GrayLevel[0.5], 
               AbsoluteThickness[1.5], 
               Opacity[0.66]}], {}, {}], 
            
            RectangleBox[{-2.6, 0.}, {-2.4, 0.013123359580052481`}, 
             "RoundingRadius" -> 0]},
           
           ImageSizeCache->{{42.09591836734692, 47.52857142857141}, {
            36.61505168354876, 40.27773676735786}}],
          StatusArea[#, 0.013123359580052481`]& ,
          TagBoxNote->"0.013123359580052481"],
         StyleBox["0.013123359580052481`", {
           GrayLevel[0]}, StripOnInput -> False]],
        Annotation[#, 
         Style[0.013123359580052481`, {
           GrayLevel[0]}], "Tooltip"]& ], 
       TagBox[
        TooltipBox[
         TagBox[
          DynamicBox[{
            FEPrivate`If[
             CurrentValue["MouseOver"], 
             EdgeForm[{
               GrayLevel[0.5], 
               AbsoluteThickness[1.5], 
               Opacity[0.66]}], {}, {}], 
            
            RectangleBox[{-2.4, 0.}, {-2.2, 0.03624546931633551}, 
             "RoundingRadius" -> 0]},
           
           ImageSizeCache->{{46.52857142857141, 51.961224489795896`}, {
            31.923654154932706`, 40.27773676735786}}],
          StatusArea[#, 0.03624546931633551]& ,
          TagBoxNote->"0.03624546931633551"],
         StyleBox["0.03624546931633551`", {
           GrayLevel[0]}, StripOnInput -> False]],
        Annotation[#, 
         Style[0.03624546931633551, {
           GrayLevel[0]}], "Tooltip"]& ], 
       TagBox[
        TooltipBox[
         TagBox[
          DynamicBox[{
            FEPrivate`If[
             CurrentValue["MouseOver"], 
             EdgeForm[{
               GrayLevel[0.5], 
               AbsoluteThickness[1.5], 
               Opacity[0.66]}], {}, {}], 
            
            RectangleBox[{-2.2, 0.}, {-2., 0.024371953505811752`}, 
             "RoundingRadius" -> 0]},
           
           ImageSizeCache->{{50.961224489795896`, 56.39387755102039}, {
            34.33275018314096, 40.27773676735786}}],
          StatusArea[#, 0.024371953505811752`]& ,
          TagBoxNote->"0.024371953505811752"],
         StyleBox["0.024371953505811752`", {
           GrayLevel[0]}, StripOnInput -> False]],
        Annotation[#, 
         Style[0.024371953505811752`, {
           GrayLevel[0]}], "Tooltip"]& ], 
       TagBox[
        TooltipBox[
         TagBox[
          DynamicBox[{
            FEPrivate`If[
             CurrentValue["MouseOver"], 
             EdgeForm[{
               GrayLevel[0.5], 
               AbsoluteThickness[1.5], 
               Opacity[0.66]}], {}, {}], 
            
            RectangleBox[{-2., 0.}, {-1.8, 0.07624046994125735}, 
             "RoundingRadius" -> 0]},
           
           ImageSizeCache->{{55.39387755102039, 60.82653061224488}, {
            23.80880437570497, 40.27773676735786}}],
          StatusArea[#, 0.07624046994125735]& ,
          TagBoxNote->"0.07624046994125735"],
         StyleBox["0.07624046994125735`", {
           GrayLevel[0]}, StripOnInput -> False]],
        Annotation[#, 
         Style[0.07624046994125735, {
           GrayLevel[0]}], "Tooltip"]& ], 
       TagBox[
        TooltipBox[
         TagBox[
          DynamicBox[{
            FEPrivate`If[
             CurrentValue["MouseOver"], 
             EdgeForm[{
               GrayLevel[0.5], 
               AbsoluteThickness[1.5], 
               Opacity[0.66]}], {}, {}], 
            
            RectangleBox[{-1.8, 0.}, {-1.6, 0.10686164229471319`}, 
             "RoundingRadius" -> 0]},
           
           ImageSizeCache->{{59.82653061224488, 65.25918367346937}, {
            17.595872513483727`, 40.27773676735786}}],
          StatusArea[#, 0.10686164229471319`]& ,
          TagBoxNote->"0.10686164229471319"],
         StyleBox["0.10686164229471319`", {
           GrayLevel[0]}, StripOnInput -> False]],
        Annotation[#, 
         Style[0.10686164229471319`, {
           GrayLevel[0]}], "Tooltip"]& ], 
       TagBox[
        TooltipBox[
         TagBox[
          DynamicBox[{
            FEPrivate`If[
             CurrentValue["MouseOver"], 
             EdgeForm[{
               GrayLevel[0.5], 
               AbsoluteThickness[1.5], 
               Opacity[0.66]}], {}, {}], 
            
            RectangleBox[{-1.6, 0.}, {-1.4, 0.17122859642544666`}, 
             "RoundingRadius" -> 0]},
           
           ImageSizeCache->{{64.25918367346937, 69.69183673469387}, {
            4.536036150039116, 40.27773676735786}}],
          StatusArea[#, 0.17122859642544666`]& ,
          TagBoxNote->"0.17122859642544666"],
         StyleBox["0.17122859642544666`", {
           GrayLevel[0]}, StripOnInput -> False]],
        Annotation[#, 
         Style[0.17122859642544666`, {
           GrayLevel[0]}], "Tooltip"]& ], 
       TagBox[
        TooltipBox[
         TagBox[
          DynamicBox[{
            FEPrivate`If[
             CurrentValue["MouseOver"], 
             EdgeForm[{
               GrayLevel[0.5], 
               AbsoluteThickness[1.5], 
               Opacity[0.66]}], {}, {}], 
            RectangleBox[{-1.4, 0.}, {-1.2, 0.17310336207974006`}, 
             "RoundingRadius" -> 0]},
           
           ImageSizeCache->{{68.69183673469387, 74.12448979591835}, {
            4.155652566637777, 40.27773676735786}}],
          StatusArea[#, 0.17310336207974006`]& ,
          TagBoxNote->"0.17310336207974006"],
         StyleBox["0.17310336207974006`", {
           GrayLevel[0]}, StripOnInput -> False]],
        Annotation[#, 
         Style[0.17310336207974006`, {
           GrayLevel[0]}], "Tooltip"]& ], 
       TagBox[
        TooltipBox[
         TagBox[
          DynamicBox[{
            FEPrivate`If[
             CurrentValue["MouseOver"], 
             EdgeForm[{
               GrayLevel[0.5], 
               AbsoluteThickness[1.5], 
               Opacity[0.66]}], {}, {}], 
            
            RectangleBox[{-1.2, 0.}, {-1., 0.24059492563429577`}, 
             "RoundingRadius" -> 0]},
           
           ImageSizeCache->{{73.12448979591835, 
            78.55714285714284}, {-9.53815643580905, 40.27773676735786}}],
          StatusArea[#, 0.24059492563429577`]& ,
          TagBoxNote->"0.24059492563429577"],
         StyleBox["0.24059492563429577`", {
           GrayLevel[0]}, StripOnInput -> False]],
        Annotation[#, 
         Style[0.24059492563429577`, {
           GrayLevel[0]}], "Tooltip"]& ], 
       TagBox[
        TooltipBox[
         TagBox[
          DynamicBox[{
            FEPrivate`If[
             CurrentValue["MouseOver"], 
             EdgeForm[{
               GrayLevel[0.5], 
               AbsoluteThickness[1.5], 
               Opacity[0.66]}], {}, {}], 
            
            RectangleBox[{-1., 0.}, {-0.8, 0.23997000374953137`}, 
             "RoundingRadius" -> 0]},
           
           ImageSizeCache->{{77.55714285714284, 
            82.98979591836732}, {-9.411361908008615, 40.27773676735786}}],
          StatusArea[#, 0.23997000374953137`]& ,
          TagBoxNote->"0.23997000374953137"],
         StyleBox["0.23997000374953137`", {
           GrayLevel[0]}, StripOnInput -> False]],
        Annotation[#, 
         Style[0.23997000374953137`, {
           GrayLevel[0]}], "Tooltip"]& ], 
       TagBox[
        TooltipBox[
         TagBox[
          DynamicBox[{
            FEPrivate`If[
             CurrentValue["MouseOver"], 
             EdgeForm[{
               GrayLevel[0.5], 
               AbsoluteThickness[1.5], 
               Opacity[0.66]}], {}, {}], 
            
            RectangleBox[{-0.8, 0.}, {-0.6, 0.3224596925384326}, 
             "RoundingRadius" -> 0]},
           
           ImageSizeCache->{{81.98979591836732, 
            87.42244897959182}, {-26.148239577665798`, 40.27773676735786}}],
          StatusArea[#, 0.3224596925384326]& ,
          TagBoxNote->"0.3224596925384326"],
         StyleBox["0.3224596925384326`", {
           GrayLevel[0]}, StripOnInput -> False]],
        Annotation[#, 
         Style[0.3224596925384326, {
           GrayLevel[0]}], "Tooltip"]& ], 
       TagBox[
        TooltipBox[
         TagBox[
          DynamicBox[{
            FEPrivate`If[
             CurrentValue["MouseOver"], 
             EdgeForm[{
               GrayLevel[0.5], 
               AbsoluteThickness[1.5], 
               Opacity[0.66]}], {}, {}], 
            
            RectangleBox[{-0.6, 0.}, {-0.4, 0.3187101612298463}, 
             "RoundingRadius" -> 0]},
           
           ImageSizeCache->{{86.42244897959182, 
            91.8551020408163}, {-25.387472410863232`, 40.27773676735786}}],
          StatusArea[#, 0.3187101612298463]& ,
          TagBoxNote->"0.3187101612298463"],
         StyleBox["0.3187101612298463`", {
           GrayLevel[0]}, StripOnInput -> False]],
        Annotation[#, 
         Style[0.3187101612298463, {
           GrayLevel[0]}], "Tooltip"]& ], 
       TagBox[
        TooltipBox[
         TagBox[
          DynamicBox[{
            FEPrivate`If[
             CurrentValue["MouseOver"], 
             EdgeForm[{
               GrayLevel[0.5], 
               AbsoluteThickness[1.5], 
               Opacity[0.66]}], {}, {}], 
            
            RectangleBox[{-0.4, 0.}, {-0.2, 0.4036995375578053}, 
             "RoundingRadius" -> 0]},
           
           ImageSizeCache->{{90.8551020408163, 
            96.2877551020408}, {-42.63152819172219, 40.27773676735786}}],
          StatusArea[#, 0.4036995375578053]& ,
          TagBoxNote->"0.4036995375578053"],
         StyleBox["0.4036995375578053`", {
           GrayLevel[0]}, StripOnInput -> False]],
        Annotation[#, 
         Style[0.4036995375578053, {
           GrayLevel[0]}], "Tooltip"]& ], 
       TagBox[
        TooltipBox[
         TagBox[
          DynamicBox[{
            FEPrivate`If[
             CurrentValue["MouseOver"], 
             EdgeForm[{
               GrayLevel[0.5], 
               AbsoluteThickness[1.5], 
               Opacity[0.66]}], {}, {}], 
            
            RectangleBox[{-0.2, 0.}, {0., 0.3818272715910511}, 
             "RoundingRadius" -> 0]},
           
           ImageSizeCache->{{95.2877551020408, 
            100.72040816326529`}, {-38.19371971870701, 40.27773676735786}}],
          StatusArea[#, 0.3818272715910511]& ,
          TagBoxNote->"0.3818272715910511"],
         StyleBox["0.3818272715910511`", {
           GrayLevel[0]}, StripOnInput -> False]],
        Annotation[#, 
         Style[0.3818272715910511, {
           GrayLevel[0]}], "Tooltip"]& ], 
       TagBox[
        TooltipBox[
         TagBox[
          DynamicBox[{
            FEPrivate`If[
             CurrentValue["MouseOver"], 
             EdgeForm[{
               GrayLevel[0.5], 
               AbsoluteThickness[1.5], 
               Opacity[0.66]}], {}, {}], 
            
            RectangleBox[{0., 0.}, {0.2, 0.3843269591301087}, 
             "RoundingRadius" -> 0]},
           ImageSizeCache->{{99.72040816326529, 
            105.15306122448978`}, {-38.70089782990873, 40.27773676735786}}],
          StatusArea[#, 0.3843269591301087]& ,
          TagBoxNote->"0.3843269591301087"],
         StyleBox["0.3843269591301087`", {
           GrayLevel[0]}, StripOnInput -> False]],
        Annotation[#, 
         Style[0.3843269591301087, {
           GrayLevel[0]}], "Tooltip"]& ], 
       TagBox[
        TooltipBox[
         TagBox[
          DynamicBox[{
            FEPrivate`If[
             CurrentValue["MouseOver"], 
             EdgeForm[{
               GrayLevel[0.5], 
               AbsoluteThickness[1.5], 
               Opacity[0.66]}], {}, {}], 
            
            RectangleBox[{0.2, 0.}, {0.4, 0.44556930383702037`}, 
             "RoundingRadius" -> 0]},
           
           ImageSizeCache->{{104.15306122448978`, 
            109.58571428571427`}, {-51.126761554351226`, 40.27773676735786}}],
          StatusArea[#, 0.44556930383702037`]& ,
          TagBoxNote->"0.44556930383702037"],
         StyleBox["0.44556930383702037`", {
           GrayLevel[0]}, StripOnInput -> False]],
        Annotation[#, 
         Style[0.44556930383702037`, {
           GrayLevel[0]}], "Tooltip"]& ], 
       TagBox[
        TooltipBox[
         TagBox[
          DynamicBox[{
            FEPrivate`If[
             CurrentValue["MouseOver"], 
             EdgeForm[{
               GrayLevel[0.5], 
               AbsoluteThickness[1.5], 
               Opacity[0.66]}], {}, {}], 
            
            RectangleBox[{0.4, 0.}, {0.6, 0.3524559430071242}, 
             "RoundingRadius" -> 0]},
           
           ImageSizeCache->{{108.58571428571427`, 
            114.01836734693876`}, {-32.23437691208665, 40.27773676735786}}],
          StatusArea[#, 0.3524559430071242]& ,
          TagBoxNote->"0.3524559430071242"],
         StyleBox["0.3524559430071242`", {
           GrayLevel[0]}, StripOnInput -> False]],
        Annotation[#, 
         Style[0.3524559430071242, {
           GrayLevel[0]}], "Tooltip"]& ], 
       TagBox[
        TooltipBox[
         TagBox[
          DynamicBox[{
            FEPrivate`If[
             CurrentValue["MouseOver"], 
             EdgeForm[{
               GrayLevel[0.5], 
               AbsoluteThickness[1.5], 
               Opacity[0.66]}], {}, {}], 
            
            RectangleBox[{0.6, 0.}, {0.8, 0.28683914510686154`}, 
             "RoundingRadius" -> 0]},
           
           ImageSizeCache->{{113.01836734693876`, 
            118.45102040816326`}, {-18.920951493041095`, 40.27773676735786}}],
          StatusArea[#, 0.28683914510686154`]& ,
          TagBoxNote->"0.28683914510686154"],
         StyleBox["0.28683914510686154`", {
           GrayLevel[0]}, StripOnInput -> False]],
        Annotation[#, 
         Style[0.28683914510686154`, {
           GrayLevel[0]}], "Tooltip"]& ], 
       TagBox[
        TooltipBox[
         TagBox[
          DynamicBox[{
            FEPrivate`If[
             CurrentValue["MouseOver"], 
             EdgeForm[{
               GrayLevel[0.5], 
               AbsoluteThickness[1.5], 
               Opacity[0.66]}], {}, {}], 
            
            RectangleBox[{0.8, 0.}, {1., 0.24121984751906017`}, 
             "RoundingRadius" -> 0]},
           
           ImageSizeCache->{{117.45102040816326`, 
            122.88367346938774`}, {-9.664950963609478, 40.27773676735786}}],
          StatusArea[#, 0.24121984751906017`]& ,
          TagBoxNote->"0.24121984751906017"],
         StyleBox["0.24121984751906017`", {
           GrayLevel[0]}, StripOnInput -> False]],
        Annotation[#, 
         Style[0.24121984751906017`, {
           GrayLevel[0]}], "Tooltip"]& ], 
       TagBox[
        TooltipBox[
         TagBox[
          DynamicBox[{
            FEPrivate`If[
             CurrentValue["MouseOver"], 
             EdgeForm[{
               GrayLevel[0.5], 
               AbsoluteThickness[1.5], 
               Opacity[0.66]}], {}, {}], 
            
            RectangleBox[{1., 0.}, {1.2, 0.18622672165979254`}, 
             "RoundingRadius" -> 0]},
           
           ImageSizeCache->{{121.88367346938774`, 127.31632653061223`}, {
            1.492967482828675, 40.27773676735786}}],
          StatusArea[#, 0.18622672165979254`]& ,
          TagBoxNote->"0.18622672165979254"],
         StyleBox["0.18622672165979254`", {
           GrayLevel[0]}, StripOnInput -> False]],
        Annotation[#, 
         Style[0.18622672165979254`, {
           GrayLevel[0]}], "Tooltip"]& ], 
       TagBox[
        TooltipBox[
         TagBox[
          DynamicBox[{
            FEPrivate`If[
             CurrentValue["MouseOver"], 
             EdgeForm[{
               GrayLevel[0.5], 
               AbsoluteThickness[1.5], 
               Opacity[0.66]}], {}, {}], 
            
            RectangleBox[{1.2, 0.}, {1.4, 0.19810023747031624`}, 
             "RoundingRadius" -> 0]},
           
           ImageSizeCache->{{126.31632653061223`, 
            131.74897959183673`}, {-0.9161285453795642, 40.27773676735786}}],
          StatusArea[#, 0.19810023747031624`]& ,
          TagBoxNote->"0.19810023747031624"],
         StyleBox["0.19810023747031624`", {
           GrayLevel[0]}, StripOnInput -> False]],
        Annotation[#, 
         Style[0.19810023747031624`, {
           GrayLevel[0]}], "Tooltip"]& ], 
       TagBox[
        TooltipBox[
         TagBox[
          DynamicBox[{
            FEPrivate`If[
             CurrentValue["MouseOver"], 
             EdgeForm[{
               GrayLevel[0.5], 
               AbsoluteThickness[1.5], 
               Opacity[0.66]}], {}, {}], 
            
            RectangleBox[{1.4, 0.}, {1.6, 0.13560804899387563`}, 
             "RoundingRadius" -> 0]},
           
           ImageSizeCache->{{130.74897959183673`, 136.18163265306123`}, {
            11.763324234663816`, 40.27773676735786}}],
          StatusArea[#, 0.13560804899387563`]& ,
          TagBoxNote->"0.13560804899387563"],
         StyleBox["0.13560804899387563`", {
           GrayLevel[0]}, StripOnInput -> False]],
        Annotation[#, 
         Style[0.13560804899387563`, {
           GrayLevel[0]}], "Tooltip"]& ], 
       TagBox[
        TooltipBox[
         TagBox[
          DynamicBox[{
            FEPrivate`If[
             CurrentValue["MouseOver"], 
             EdgeForm[{
               GrayLevel[0.5], 
               AbsoluteThickness[1.5], 
               Opacity[0.66]}], {}, {}], 
            
            RectangleBox[{1.6, 0.}, {1.8, 0.08998875140607426}, 
             "RoundingRadius" -> 0]},
           
           ImageSizeCache->{{135.18163265306123`, 140.6142857142857}, {
            21.01932476409543, 40.27773676735786}}],
          StatusArea[#, 0.08998875140607426]& ,
          TagBoxNote->"0.08998875140607426"],
         StyleBox["0.08998875140607426`", {
           GrayLevel[0]}, StripOnInput -> False]],
        Annotation[#, 
         Style[0.08998875140607426, {
           GrayLevel[0]}], "Tooltip"]& ], 
       TagBox[
        TooltipBox[
         TagBox[
          DynamicBox[{
            FEPrivate`If[
             CurrentValue["MouseOver"], 
             EdgeForm[{
               GrayLevel[0.5], 
               AbsoluteThickness[1.5], 
               Opacity[0.66]}], {}, {}], 
            
            RectangleBox[{1.8, 0.}, {2., 0.04749406324209474}, 
             "RoundingRadius" -> 0]},
           
           ImageSizeCache->{{139.6142857142857, 145.0469387755102}, {
            29.641352654524912`, 40.27773676735786}}],
          StatusArea[#, 0.04749406324209474]& ,
          TagBoxNote->"0.04749406324209474"],
         StyleBox["0.04749406324209474`", {
           GrayLevel[0]}, StripOnInput -> False]],
        Annotation[#, 
         Style[0.04749406324209474, {
           GrayLevel[0]}], "Tooltip"]& ], 
       TagBox[
        TooltipBox[
         TagBox[
          DynamicBox[{
            FEPrivate`If[
             CurrentValue["MouseOver"], 
             EdgeForm[{
               GrayLevel[0.5], 
               AbsoluteThickness[1.5], 
               Opacity[0.66]}], {}, {}], 
            
            RectangleBox[{2., 0.}, {2.2, 0.02999625046869139}, 
             "RoundingRadius" -> 0]},
           
           ImageSizeCache->{{144.0469387755102, 149.4795918367347}, {
            33.19159943293705, 40.27773676735786}}],
          StatusArea[#, 0.02999625046869139]& ,
          TagBoxNote->"0.02999625046869139"],
         StyleBox["0.02999625046869139`", {
           GrayLevel[0]}, StripOnInput -> False]],
        Annotation[#, 
         Style[0.02999625046869139, {
           GrayLevel[0]}], "Tooltip"]& ], 
       TagBox[
        TooltipBox[
         TagBox[
          DynamicBox[{
            FEPrivate`If[
             CurrentValue["MouseOver"], 
             EdgeForm[{
               GrayLevel[0.5], 
               AbsoluteThickness[1.5], 
               Opacity[0.66]}], {}, {}], 
            
            RectangleBox[{2.2, 0.}, {2.4, 0.02937132858392705}, 
             "RoundingRadius" -> 0]},
           
           ImageSizeCache->{{148.4795918367347, 153.91224489795917`}, {
            33.318393960737474`, 40.27773676735786}}],
          StatusArea[#, 0.02937132858392705]& ,
          TagBoxNote->"0.02937132858392705"],
         StyleBox["0.02937132858392705`", {
           GrayLevel[0]}, StripOnInput -> False]],
        Annotation[#, 
         Style[0.02937132858392705, {
           GrayLevel[0]}], "Tooltip"]& ], 
       TagBox[
        TooltipBox[
         TagBox[
          DynamicBox[{
            FEPrivate`If[
             CurrentValue["MouseOver"], 
             EdgeForm[{
               GrayLevel[0.5], 
               AbsoluteThickness[1.5], 
               Opacity[0.66]}], {}, {}], 
            
            RectangleBox[{2.4, 0.}, {2.6, 0.014373203349581289`}, 
             "RoundingRadius" -> 0]},
           
           ImageSizeCache->{{152.91224489795917`, 158.34489795918367`}, {
            36.36146262794789, 40.27773676735786}}],
          StatusArea[#, 0.014373203349581289`]& ,
          TagBoxNote->"0.014373203349581289"],
         StyleBox["0.014373203349581289`", {
           GrayLevel[0]}, StripOnInput -> False]],
        Annotation[#, 
         Style[0.014373203349581289`, {
           GrayLevel[0]}], "Tooltip"]& ], 
       TagBox[
        TooltipBox[
         TagBox[
          DynamicBox[{
            FEPrivate`If[
             CurrentValue["MouseOver"], 
             EdgeForm[{
               GrayLevel[0.5], 
               AbsoluteThickness[1.5], 
               Opacity[0.66]}], {}, {}], 
            
            RectangleBox[{2.6, 0.}, {2.8, 0.011248593925759297`}, 
             "RoundingRadius" -> 0]},
           
           ImageSizeCache->{{157.34489795918367`, 162.77755102040814`}, {
            36.995435266950054`, 40.27773676735786}}],
          StatusArea[#, 0.011248593925759297`]& ,
          TagBoxNote->"0.011248593925759297"],
         StyleBox["0.011248593925759297`", {
           GrayLevel[0]}, StripOnInput -> False]],
        Annotation[#, 
         Style[0.011248593925759297`, {
           GrayLevel[0]}], "Tooltip"]& ], 
       TagBox[
        TooltipBox[
         TagBox[
          DynamicBox[{
            FEPrivate`If[
             CurrentValue["MouseOver"], 
             EdgeForm[{
               GrayLevel[0.5], 
               AbsoluteThickness[1.5], 
               Opacity[0.66]}], {}, {}], 
            
            RectangleBox[{2.8, 0.}, {3., 0.009998750156230462}, 
             "RoundingRadius" -> 0]},
           
           ImageSizeCache->{{161.77755102040814`, 167.21020408163264`}, {
            37.24902432255092, 40.27773676735786}}],
          StatusArea[#, 0.009998750156230462]& ,
          TagBoxNote->"0.009998750156230462"],
         StyleBox["0.009998750156230462`", {
           GrayLevel[0]}, StripOnInput -> False]],
        Annotation[#, 
         Style[0.009998750156230462, {
           GrayLevel[0]}], "Tooltip"]& ], 
       TagBox[
        TooltipBox[
         TagBox[
          DynamicBox[{
            FEPrivate`If[
             CurrentValue["MouseOver"], 
             EdgeForm[{
               GrayLevel[0.5], 
               AbsoluteThickness[1.5], 
               Opacity[0.66]}], {}, {}], 
            
            RectangleBox[{3., 0.}, {3.2, 0.0006249218847644038}, 
             "RoundingRadius" -> 0]},
           
           ImageSizeCache->{{166.21020408163264`, 171.64285714285714`}, {
            39.15094223955742, 40.27773676735786}}],
          StatusArea[#, 0.0006249218847644038]& ,
          TagBoxNote->"0.0006249218847644038"],
         StyleBox["0.0006249218847644038`", {
           GrayLevel[0]}, StripOnInput -> False]],
        Annotation[#, 
         Style[0.0006249218847644038, {
           GrayLevel[0]}], "Tooltip"]& ], 
       TagBox[
        TooltipBox[
         TagBox[
          DynamicBox[{
            FEPrivate`If[
             CurrentValue["MouseOver"], 
             EdgeForm[{
               GrayLevel[0.5], 
               AbsoluteThickness[1.5], 
               Opacity[0.66]}], {}, {}], 
            
            RectangleBox[{3.2, 0.}, {3.4, 0.0043744531933508374`}, 
             "RoundingRadius" -> 0]},
           
           ImageSizeCache->{{170.64285714285714`, 176.07551020408164`}, {
            38.39017507275482, 40.27773676735786}}],
          StatusArea[#, 0.0043744531933508374`]& ,
          TagBoxNote->"0.0043744531933508374"],
         StyleBox["0.0043744531933508374`", {
           GrayLevel[0]}, StripOnInput -> False]],
        Annotation[#, 
         Style[0.0043744531933508374`, {
           GrayLevel[0]}], "Tooltip"]& ]}, {}, {}}, {{}, {}, 
      {RGBColor[1, 0, 0], Thickness[Large], LineBox[CompressedData["
1:eJwV13k0VV8bB3DDdScuSqlESaVUREoZ8hxThszJEELUT5KpTKEMGRMZEpI5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        "]]}}},
    AspectRatio->NCache[GoldenRatio^(-1), 0.6180339887498948],
    Axes->{True, True},
    AxesOrigin->{-3.8, 0},
    FrameTicks->{{Automatic, Automatic}, {Automatic, Automatic}},
    GridLines->{None, None},
    PlotRange->NCache[{{
        Rational[-19, 5], 
        Rational[17, 5]}, {All, All}}, {{-3.8, 3.4}, {All, All}}],
    PlotRangePadding->{{
       Scaled[0.02], 
       Scaled[0.02]}, {
       Scaled[0.02], 
       Scaled[0.1]}},
    Ticks->{Automatic, Automatic}], ",", 
   GraphicsBox[{
     {RGBColor[0.798413061722744, 0.824719615472648, 0.968322270542458], 
      EdgeForm[Opacity[0.7]], {}, 
      {RGBColor[0.798413061722744, 0.824719615472648, 0.968322270542458], 
       EdgeForm[Opacity[0.7]], 
       TagBox[
        TooltipBox[
         TagBox[
          DynamicBox[{
            FEPrivate`If[
             CurrentValue["MouseOver"], 
             EdgeForm[{
               GrayLevel[0.5], 
               AbsoluteThickness[1.5], 
               Opacity[0.66]}], {}, {}], 
            RectangleBox[{-3.8, 0.}, {-3.6, 0.0006249218847644053}, 
             "RoundingRadius" -> 0]},
           
           ImageSizeCache->{{15.5, 20.812851627137334`}, {39.1511198229297, 
            40.277736767357865`}}],
          StatusArea[#, 0.0006249218847644053]& ,
          TagBoxNote->"0.0006249218847644053"],
         StyleBox["0.0006249218847644053`", {
           GrayLevel[0]}, StripOnInput -> False]],
        Annotation[#, 
         Style[0.0006249218847644053, {
           GrayLevel[0]}], "Tooltip"]& ], {}, 
       TagBox[
        TooltipBox[
         TagBox[
          DynamicBox[{
            FEPrivate`If[
             CurrentValue["MouseOver"], 
             EdgeForm[{
               GrayLevel[0.5], 
               AbsoluteThickness[1.5], 
               Opacity[0.66]}], {}, {}], 
            
            RectangleBox[{-3.4, 0.}, {-3.2, 0.0006249218847644053}, 
             "RoundingRadius" -> 0]},
           
           ImageSizeCache->{{24.12570325427467, 29.438554881412017`}, {
            39.1511198229297, 40.277736767357865`}}],
          StatusArea[#, 0.0006249218847644053]& ,
          TagBoxNote->"0.0006249218847644053"],
         StyleBox["0.0006249218847644053`", {
           GrayLevel[0]}, StripOnInput -> False]],
        Annotation[#, 
         Style[0.0006249218847644053, {
           GrayLevel[0]}], "Tooltip"]& ], 
       TagBox[
        TooltipBox[
         TagBox[
          DynamicBox[{
            FEPrivate`If[
             CurrentValue["MouseOver"], 
             EdgeForm[{
               GrayLevel[0.5], 
               AbsoluteThickness[1.5], 
               Opacity[0.66]}], {}, {}], 
            
            RectangleBox[{-3.2, 0.}, {-3., 0.009998750156230462}, 
             "RoundingRadius" -> 0]},
           
           ImageSizeCache->{{28.438554881412017`, 33.75140650854935}, {
            37.25186565650724, 40.277736767357865`}}],
          StatusArea[#, 0.009998750156230462]& ,
          TagBoxNote->"0.009998750156230462"],
         StyleBox["0.009998750156230462`", {
           GrayLevel[0]}, StripOnInput -> False]],
        Annotation[#, 
         Style[0.009998750156230462, {
           GrayLevel[0]}], "Tooltip"]& ], 
       TagBox[
        TooltipBox[
         TagBox[
          DynamicBox[{
            FEPrivate`If[
             CurrentValue["MouseOver"], 
             EdgeForm[{
               GrayLevel[0.5], 
               AbsoluteThickness[1.5], 
               Opacity[0.66]}], {}, {}], 
            
            RectangleBox[{-3., 0.}, {-2.8, 0.004999375078115231}, 
             "RoundingRadius" -> 0]},
           
           ImageSizeCache->{{32.75140650854935, 38.0642581356867}, {
            38.26480121193256, 40.277736767357865`}}],
          StatusArea[#, 0.004999375078115231]& ,
          TagBoxNote->"0.004999375078115231"],
         StyleBox["0.004999375078115231`", {
           GrayLevel[0]}, StripOnInput -> False]],
        Annotation[#, 
         Style[0.004999375078115231, {
           GrayLevel[0]}], "Tooltip"]& ], 
       TagBox[
        TooltipBox[
         TagBox[
          DynamicBox[{
            FEPrivate`If[
             CurrentValue["MouseOver"], 
             EdgeForm[{
               GrayLevel[0.5], 
               AbsoluteThickness[1.5], 
               Opacity[0.66]}], {}, {}], 
            
            RectangleBox[{-2.8, 0.}, {-2.6, 0.009998750156230484}, 
             "RoundingRadius" -> 0]},
           
           ImageSizeCache->{{37.0642581356867, 42.377109762824034`}, {
            37.251865656507235`, 40.277736767357865`}}],
          StatusArea[#, 0.009998750156230484]& ,
          TagBoxNote->"0.009998750156230484"],
         StyleBox["0.009998750156230484`", {
           GrayLevel[0]}, StripOnInput -> False]],
        Annotation[#, 
         Style[0.009998750156230484, {
           GrayLevel[0]}], "Tooltip"]& ], 
       TagBox[
        TooltipBox[
         TagBox[
          DynamicBox[{
            FEPrivate`If[
             CurrentValue["MouseOver"], 
             EdgeForm[{
               GrayLevel[0.5], 
               AbsoluteThickness[1.5], 
               Opacity[0.66]}], {}, {}], 
            
            RectangleBox[{-2.6, 0.}, {-2.4, 0.009373828271466059}, 
             "RoundingRadius" -> 0]},
           
           ImageSizeCache->{{41.377109762824034`, 46.689961389961375`}, {
            37.378482600935406`, 40.277736767357865`}}],
          StatusArea[#, 0.009373828271466059]& ,
          TagBoxNote->"0.009373828271466059"],
         StyleBox["0.009373828271466059`", {
           GrayLevel[0]}, StripOnInput -> False]],
        Annotation[#, 
         Style[0.009373828271466059, {
           GrayLevel[0]}], "Tooltip"]& ], 
       TagBox[
        TooltipBox[
         TagBox[
          DynamicBox[{
            FEPrivate`If[
             CurrentValue["MouseOver"], 
             EdgeForm[{
               GrayLevel[0.5], 
               AbsoluteThickness[1.5], 
               Opacity[0.66]}], {}, {}], 
            
            RectangleBox[{-2.4, 0.}, {-2.2, 0.024371953505811807`}, 
             "RoundingRadius" -> 0]},
           
           ImageSizeCache->{{45.689961389961375`, 51.00281301709872}, {
            34.33967593465946, 40.277736767357865`}}],
          StatusArea[#, 0.024371953505811807`]& ,
          TagBoxNote->"0.024371953505811807"],
         StyleBox["0.024371953505811807`", {
           GrayLevel[0]}, StripOnInput -> False]],
        Annotation[#, 
         Style[0.024371953505811807`, {
           GrayLevel[0]}], "Tooltip"]& ], 
       TagBox[
        TooltipBox[
         TagBox[
          DynamicBox[{
            FEPrivate`If[
             CurrentValue["MouseOver"], 
             EdgeForm[{
               GrayLevel[0.5], 
               AbsoluteThickness[1.5], 
               Opacity[0.66]}], {}, {}], 
            
            RectangleBox[{-2.2, 0.}, {-2., 0.040619922509686254`}, 
             "RoundingRadius" -> 0]},
           
           ImageSizeCache->{{50.00281301709872, 55.31566464423606}, {
            31.04763537952721, 40.277736767357865`}}],
          StatusArea[#, 0.040619922509686254`]& ,
          TagBoxNote->"0.040619922509686254"],
         StyleBox["0.040619922509686254`", {
           GrayLevel[0]}, StripOnInput -> False]],
        Annotation[#, 
         Style[0.040619922509686254`, {
           GrayLevel[0]}], "Tooltip"]& ], 
       TagBox[
        TooltipBox[
         TagBox[
          DynamicBox[{
            FEPrivate`If[
             CurrentValue["MouseOver"], 
             EdgeForm[{
               GrayLevel[0.5], 
               AbsoluteThickness[1.5], 
               Opacity[0.66]}], {}, {}], 
            
            RectangleBox[{-2., 0.}, {-1.8, 0.06061742282214724}, 
             "RoundingRadius" -> 0]},
           
           ImageSizeCache->{{54.31566464423606, 59.6285162713734}, {
            26.995893157825954`, 40.277736767357865`}}],
          StatusArea[#, 0.06061742282214724]& ,
          TagBoxNote->"0.06061742282214724"],
         StyleBox["0.06061742282214724`", {
           GrayLevel[0]}, StripOnInput -> False]],
        Annotation[#, 
         Style[0.06061742282214724, {
           GrayLevel[0]}], "Tooltip"]& ], 
       TagBox[
        TooltipBox[
         TagBox[
          DynamicBox[{
            FEPrivate`If[
             CurrentValue["MouseOver"], 
             EdgeForm[{
               GrayLevel[0.5], 
               AbsoluteThickness[1.5], 
               Opacity[0.66]}], {}, {}], 
            
            RectangleBox[{-1.8, 0.}, {-1.6, 0.1243594550681165}, 
             "RoundingRadius" -> 0]},
           
           ImageSizeCache->{{58.6285162713734, 63.94136789851074}, {
            14.080964826153224`, 40.277736767357865`}}],
          StatusArea[#, 0.1243594550681165]& ,
          TagBoxNote->"0.1243594550681165"],
         StyleBox["0.1243594550681165`", {
           GrayLevel[0]}, StripOnInput -> False]],
        Annotation[#, 
         Style[0.1243594550681165, {
           GrayLevel[0]}], "Tooltip"]& ], 
       TagBox[
        TooltipBox[
         TagBox[
          DynamicBox[{
            FEPrivate`If[
             CurrentValue["MouseOver"], 
             EdgeForm[{
               GrayLevel[0.5], 
               AbsoluteThickness[1.5], 
               Opacity[0.66]}], {}, {}], 
            
            RectangleBox[{-1.6, 0.}, {-1.4, 0.15185601799775014`}, 
             "RoundingRadius" -> 0]},
           
           ImageSizeCache->{{62.94136789851074, 68.25421952564808}, {
            8.50981927131404, 40.277736767357865`}}],
          StatusArea[#, 0.15185601799775014`]& ,
          TagBoxNote->"0.15185601799775014"],
         StyleBox["0.15185601799775014`", {
           GrayLevel[0]}, StripOnInput -> False]],
        Annotation[#, 
         Style[0.15185601799775014`, {
           GrayLevel[0]}], "Tooltip"]& ], 
       TagBox[
        TooltipBox[
         TagBox[
          DynamicBox[{
            FEPrivate`If[
             CurrentValue["MouseOver"], 
             EdgeForm[{
               GrayLevel[0.5], 
               AbsoluteThickness[1.5], 
               Opacity[0.66]}], {}, {}], 
            
            RectangleBox[{-1.4, 0.}, {-1.2, 0.18747656542932137`}, 
             "RoundingRadius" -> 0]},
           
           ImageSizeCache->{{67.25421952564808, 72.56707115278542}, {
            1.2926534389086584`, 40.277736767357865`}}],
          StatusArea[#, 0.18747656542932137`]& ,
          TagBoxNote->"0.18747656542932137"],
         StyleBox["0.18747656542932137`", {
           GrayLevel[0]}, StripOnInput -> False]],
        Annotation[#, 
         Style[0.18747656542932137`, {
           GrayLevel[0]}], "Tooltip"]& ], 
       TagBox[
        TooltipBox[
         TagBox[
          DynamicBox[{
            FEPrivate`If[
             CurrentValue["MouseOver"], 
             EdgeForm[{
               GrayLevel[0.5], 
               AbsoluteThickness[1.5], 
               Opacity[0.66]}], {}, {}], 
            
            RectangleBox[{-1.2, 0.}, {-1., 0.23184601924759407`}, 
             "RoundingRadius" -> 0]},
           
           ImageSizeCache->{{71.56707115278542, 
            76.87992277992277}, {-7.697149615490986, 40.277736767357865`}}],
          StatusArea[#, 0.23184601924759407`]& ,
          TagBoxNote->"0.23184601924759407"],
         StyleBox["0.23184601924759407`", {
           GrayLevel[0]}, StripOnInput -> False]],
        Annotation[#, 
         Style[0.23184601924759407`, {
           GrayLevel[0]}], "Tooltip"]& ], 
       TagBox[
        TooltipBox[
         TagBox[
          DynamicBox[{
            FEPrivate`If[
             CurrentValue["MouseOver"], 
             EdgeForm[{
               GrayLevel[0.5], 
               AbsoluteThickness[1.5], 
               Opacity[0.66]}], {}, {}], 
            
            RectangleBox[{-1., 0.}, {-0.8, 0.2624671916010499}, 
             "RoundingRadius" -> 0]},
           
           ImageSizeCache->{{75.87992277992277, 
            81.1927744070601}, {-13.901379892471027`, 40.277736767357865`}}],
          StatusArea[#, 0.2624671916010499]& ,
          TagBoxNote->"0.2624671916010499"],
         StyleBox["0.2624671916010499`", {
           GrayLevel[0]}, StripOnInput -> False]],
        Annotation[#, 
         Style[0.2624671916010499, {
           GrayLevel[0]}], "Tooltip"]& ], 
       TagBox[
        TooltipBox[
         TagBox[
          DynamicBox[{
            FEPrivate`If[
             CurrentValue["MouseOver"], 
             EdgeForm[{
               GrayLevel[0.5], 
               AbsoluteThickness[1.5], 
               Opacity[0.66]}], {}, {}], 
            RectangleBox[{-0.8, 0.}, {-0.6, 0.3068366454193225}, 
             "RoundingRadius" -> 0]},
           
           ImageSizeCache->{{80.1927744070601, 
            85.50562603419745}, {-22.891182946870643`, 40.277736767357865`}}],
          StatusArea[#, 0.3068366454193225]& ,
          TagBoxNote->"0.3068366454193225"],
         StyleBox["0.3068366454193225`", {
           GrayLevel[0]}, StripOnInput -> False]],
        Annotation[#, 
         Style[0.3068366454193225, {
           GrayLevel[0]}], "Tooltip"]& ], 
       TagBox[
        TooltipBox[
         TagBox[
          DynamicBox[{
            FEPrivate`If[
             CurrentValue["MouseOver"], 
             EdgeForm[{
               GrayLevel[0.5], 
               AbsoluteThickness[1.5], 
               Opacity[0.66]}], {}, {}], 
            
            RectangleBox[{-0.6, 0.}, {-0.4, 0.3668291463567055}, 
             "RoundingRadius" -> 0]},
           
           ImageSizeCache->{{84.50562603419745, 
            89.81847766133478}, {-35.046409611974426`, 40.277736767357865`}}],
          StatusArea[#, 0.3668291463567055]& ,
          TagBoxNote->"0.3668291463567055"],
         StyleBox["0.3668291463567055`", {
           GrayLevel[0]}, StripOnInput -> False]],
        Annotation[#, 
         Style[0.3668291463567055, {
           GrayLevel[0]}], "Tooltip"]& ], 
       TagBox[
        TooltipBox[
         TagBox[
          DynamicBox[{
            FEPrivate`If[
             CurrentValue["MouseOver"], 
             EdgeForm[{
               GrayLevel[0.5], 
               AbsoluteThickness[1.5], 
               Opacity[0.66]}], {}, {}], 
            
            RectangleBox[{-0.4, 0.}, {-0.2, 0.3555805524309461}, 
             "RoundingRadius" -> 0]},
           
           ImageSizeCache->{{88.81847766133478, 
            94.13132928847213}, {-32.76730461226745, 40.277736767357865`}}],
          StatusArea[#, 0.3555805524309461]& ,
          TagBoxNote->"0.3555805524309461"],
         StyleBox["0.3555805524309461`", {
           GrayLevel[0]}, StripOnInput -> False]],
        Annotation[#, 
         Style[0.3555805524309461, {
           GrayLevel[0]}], "Tooltip"]& ], 
       TagBox[
        TooltipBox[
         TagBox[
          DynamicBox[{
            FEPrivate`If[
             CurrentValue["MouseOver"], 
             EdgeForm[{
               GrayLevel[0.5], 
               AbsoluteThickness[1.5], 
               Opacity[0.66]}], {}, {}], 
            
            RectangleBox[{-0.2, 0.}, {0., 0.389326334208224}, 
             "RoundingRadius" -> 0]},
           
           ImageSizeCache->{{93.13132928847213, 
            98.44418091560946}, {-39.604619611388316`, 40.277736767357865`}}],
          StatusArea[#, 0.389326334208224]& ,
          TagBoxNote->"0.389326334208224"],
         StyleBox["0.389326334208224`", {
           GrayLevel[0]}, StripOnInput -> False]],
        Annotation[#, 
         Style[0.389326334208224, {
           GrayLevel[0]}], "Tooltip"]& ], 
       TagBox[
        TooltipBox[
         TagBox[
          DynamicBox[{
            FEPrivate`If[
             CurrentValue["MouseOver"], 
             EdgeForm[{
               GrayLevel[0.5], 
               AbsoluteThickness[1.5], 
               Opacity[0.66]}], {}, {}], 
            
            RectangleBox[{0., 0.}, {0.2, 0.38870141232345956`}, 
             "RoundingRadius" -> 0]},
           
           ImageSizeCache->{{97.44418091560946, 
            102.7570325427468}, {-39.47800266696015, 40.277736767357865`}}],
          StatusArea[#, 0.38870141232345956`]& ,
          TagBoxNote->"0.38870141232345956"],
         StyleBox["0.38870141232345956`", {
           GrayLevel[0]}, StripOnInput -> False]],
        Annotation[#, 
         Style[0.38870141232345956`, {
           GrayLevel[0]}], "Tooltip"]& ], 
       TagBox[
        TooltipBox[
         TagBox[
          DynamicBox[{
            FEPrivate`If[
             CurrentValue["MouseOver"], 
             EdgeForm[{
               GrayLevel[0.5], 
               AbsoluteThickness[1.5], 
               Opacity[0.66]}], {}, {}], 
            
            RectangleBox[{0.2, 0.}, {0.4, 0.4461942257217848}, 
             "RoundingRadius" -> 0]},
           
           ImageSizeCache->{{101.7570325427468, 
            107.06988416988415`}, {-51.12676155435123, 40.277736767357865`}}],
          StatusArea[#, 0.4461942257217848]& ,
          TagBoxNote->"0.4461942257217848"],
         StyleBox["0.4461942257217848`", {
           GrayLevel[0]}, StripOnInput -> False]],
        Annotation[#, 
         Style[0.4461942257217848, {
           GrayLevel[0]}], "Tooltip"]& ], 
       TagBox[
        TooltipBox[
         TagBox[
          DynamicBox[{
            FEPrivate`If[
             CurrentValue["MouseOver"], 
             EdgeForm[{
               GrayLevel[0.5], 
               AbsoluteThickness[1.5], 
               Opacity[0.66]}], {}, {}], 
            
            RectangleBox[{0.4, 0.}, {0.6, 0.3349581302337209}, 
             "RoundingRadius" -> 0]},
           
           ImageSizeCache->{{106.06988416988415`, 
            111.38273579702148`}, {-28.58894544613807, 40.277736767357865`}}],
          StatusArea[#, 0.3349581302337209]& ,
          TagBoxNote->"0.3349581302337209"],
         StyleBox["0.3349581302337209`", {
           GrayLevel[0]}, StripOnInput -> False]],
        Annotation[#, 
         Style[0.3349581302337209, {
           GrayLevel[0]}], "Tooltip"]& ], 
       TagBox[
        TooltipBox[
         TagBox[
          DynamicBox[{
            FEPrivate`If[
             CurrentValue["MouseOver"], 
             EdgeForm[{
               GrayLevel[0.5], 
               AbsoluteThickness[1.5], 
               Opacity[0.66]}], {}, {}], 
            
            RectangleBox[{0.6, 0.}, {0.8, 0.2962129733783276}, 
             "RoundingRadius" -> 0]},
           ImageSizeCache->{{110.38273579702148`, 
            115.69558742415883`}, {-20.738694891591848`, 
            40.277736767357865`}}],
          StatusArea[#, 0.2962129733783276]& ,
          TagBoxNote->"0.2962129733783276"],
         StyleBox["0.2962129733783276`", {
           GrayLevel[0]}, StripOnInput -> False]],
        Annotation[#, 
         Style[0.2962129733783276, {
           GrayLevel[0]}], "Tooltip"]& ], 
       TagBox[
        TooltipBox[
         TagBox[
          DynamicBox[{
            FEPrivate`If[
             CurrentValue["MouseOver"], 
             EdgeForm[{
               GrayLevel[0.5], 
               AbsoluteThickness[1.5], 
               Opacity[0.66]}], {}, {}], 
            
            RectangleBox[{0.8, 0.}, {1., 0.2593425821772279}, 
             "RoundingRadius" -> 0]},
           
           ImageSizeCache->{{114.69558742415883`, 
            120.00843905129616`}, {-13.268295170330205`, 
            40.277736767357865`}}],
          StatusArea[#, 0.2593425821772279]& ,
          TagBoxNote->"0.2593425821772279"],
         StyleBox["0.2593425821772279`", {
           GrayLevel[0]}, StripOnInput -> False]],
        Annotation[#, 
         Style[0.2593425821772279, {
           GrayLevel[0]}], "Tooltip"]& ], 
       TagBox[
        TooltipBox[
         TagBox[
          DynamicBox[{
            FEPrivate`If[
             CurrentValue["MouseOver"], 
             EdgeForm[{
               GrayLevel[0.5], 
               AbsoluteThickness[1.5], 
               Opacity[0.66]}], {}, {}], 
            
            RectangleBox[{1., 0.}, {1.2, 0.19247594050743658`}, 
             "RoundingRadius" -> 0]},
           
           ImageSizeCache->{{119.00843905129616`, 124.32129067843351`}, {
            0.27971788348335025`, 40.277736767357865`}}],
          StatusArea[#, 0.19247594050743658`]& ,
          TagBoxNote->"0.19247594050743658"],
         StyleBox["0.19247594050743658`", {
           GrayLevel[0]}, StripOnInput -> False]],
        Annotation[#, 
         Style[0.19247594050743658`, {
           GrayLevel[0]}], "Tooltip"]& ], 
       TagBox[
        TooltipBox[
         TagBox[
          DynamicBox[{
            FEPrivate`If[
             CurrentValue["MouseOver"], 
             EdgeForm[{
               GrayLevel[0.5], 
               AbsoluteThickness[1.5], 
               Opacity[0.66]}], {}, {}], 
            
            RectangleBox[{1.2, 0.}, {1.4, 0.1649793775778028}, 
             "RoundingRadius" -> 0]},
           
           ImageSizeCache->{{123.32129067843351`, 128.63414230557083`}, {
            5.850863438322563, 40.277736767357865`}}],
          StatusArea[#, 0.1649793775778028]& ,
          TagBoxNote->"0.1649793775778028"],
         StyleBox["0.1649793775778028`", {
           GrayLevel[0]}, StripOnInput -> False]],
        Annotation[#, 
         Style[0.1649793775778028, {
           GrayLevel[0]}], "Tooltip"]& ], 
       TagBox[
        TooltipBox[
         TagBox[
          DynamicBox[{
            FEPrivate`If[
             CurrentValue["MouseOver"], 
             EdgeForm[{
               GrayLevel[0.5], 
               AbsoluteThickness[1.5], 
               Opacity[0.66]}], {}, {}], 
            
            RectangleBox[{1.4, 0.}, {1.6, 0.14060742407199087`}, 
             "RoundingRadius" -> 0]},
           
           ImageSizeCache->{{127.63414230557083`, 132.94699393270818`}, {
            10.788924271020992`, 40.277736767357865`}}],
          StatusArea[#, 0.14060742407199087`]& ,
          TagBoxNote->"0.14060742407199087"],
         StyleBox["0.14060742407199087`", {
           GrayLevel[0]}, StripOnInput -> False]],
        Annotation[#, 
         Style[0.14060742407199087`, {
           GrayLevel[0]}], "Tooltip"]& ], 
       TagBox[
        TooltipBox[
         TagBox[
          DynamicBox[{
            FEPrivate`If[
             CurrentValue["MouseOver"], 
             EdgeForm[{
               GrayLevel[0.5], 
               AbsoluteThickness[1.5], 
               Opacity[0.66]}], {}, {}], 
            
            RectangleBox[{1.6, 0.}, {1.8, 0.0824896887889014}, 
             "RoundingRadius" -> 0]},
           
           ImageSizeCache->{{131.94699393270818`, 137.25984555984553`}, {
            22.564300102840214`, 40.277736767357865`}}],
          StatusArea[#, 0.0824896887889014]& ,
          TagBoxNote->"0.0824896887889014"],
         StyleBox["0.0824896887889014`", {
           GrayLevel[0]}, StripOnInput -> False]],
        Annotation[#, 
         Style[0.0824896887889014, {
           GrayLevel[0]}], "Tooltip"]& ], 
       TagBox[
        TooltipBox[
         TagBox[
          DynamicBox[{
            FEPrivate`If[
             CurrentValue["MouseOver"], 
             EdgeForm[{
               GrayLevel[0.5], 
               AbsoluteThickness[1.5], 
               Opacity[0.66]}], {}, {}], 
            
            RectangleBox[{1.8, 0.}, {2., 0.059367579052618434`}, 
             "RoundingRadius" -> 0]},
           
           ImageSizeCache->{{136.25984555984553`, 141.57269718698288`}, {
            27.249127046682283`, 40.277736767357865`}}],
          StatusArea[#, 0.059367579052618434`]& ,
          TagBoxNote->"0.059367579052618434"],
         StyleBox["0.059367579052618434`", {
           GrayLevel[0]}, StripOnInput -> False]],
        Annotation[#, 
         Style[0.059367579052618434`, {
           GrayLevel[0]}], "Tooltip"]& ], 
       TagBox[
        TooltipBox[
         TagBox[
          DynamicBox[{
            FEPrivate`If[
             CurrentValue["MouseOver"], 
             EdgeForm[{
               GrayLevel[0.5], 
               AbsoluteThickness[1.5], 
               Opacity[0.66]}], {}, {}], 
            
            RectangleBox[{2., 0.}, {2.2, 0.0318710161229846}, 
             "RoundingRadius" -> 0]},
           ImageSizeCache->{{140.57269718698288`, 145.88554881412023`}, {
            32.82027260152151, 40.277736767357865`}}],
          StatusArea[#, 0.0318710161229846]& ,
          TagBoxNote->"0.0318710161229846"],
         StyleBox["0.0318710161229846`", {
           GrayLevel[0]}, StripOnInput -> False]],
        Annotation[#, 
         Style[0.0318710161229846, {
           GrayLevel[0]}], "Tooltip"]& ], 
       TagBox[
        TooltipBox[
         TagBox[
          DynamicBox[{
            FEPrivate`If[
             CurrentValue["MouseOver"], 
             EdgeForm[{
               GrayLevel[0.5], 
               AbsoluteThickness[1.5], 
               Opacity[0.66]}], {}, {}], 
            
            RectangleBox[{2.2, 0.}, {2.4, 0.024996875390576215`}, 
             "RoundingRadius" -> 0]},
           
           ImageSizeCache->{{144.88554881412023`, 150.19840044125755`}, {
            34.213058990231296`, 40.277736767357865`}}],
          StatusArea[#, 0.024996875390576215`]& ,
          TagBoxNote->"0.024996875390576215"],
         StyleBox["0.024996875390576215`", {
           GrayLevel[0]}, StripOnInput -> False]],
        Annotation[#, 
         Style[0.024996875390576215`, {
           GrayLevel[0]}], "Tooltip"]& ], 
       TagBox[
        TooltipBox[
         TagBox[
          DynamicBox[{
            FEPrivate`If[
             CurrentValue["MouseOver"], 
             EdgeForm[{
               GrayLevel[0.5], 
               AbsoluteThickness[1.5], 
               Opacity[0.66]}], {}, {}], 
            
            RectangleBox[{2.4, 0.}, {2.6, 0.011873515810523674`}, 
             "RoundingRadius" -> 0]},
           
           ImageSizeCache->{{149.19840044125755`, 154.5112520683949}, {
            36.87201482322275, 40.277736767357865`}}],
          StatusArea[#, 0.011873515810523674`]& ,
          TagBoxNote->"0.011873515810523674"],
         StyleBox["0.011873515810523674`", {
           GrayLevel[0]}, StripOnInput -> False]],
        Annotation[#, 
         Style[0.011873515810523674`, {
           GrayLevel[0]}], "Tooltip"]& ], 
       TagBox[
        TooltipBox[
         TagBox[
          DynamicBox[{
            FEPrivate`If[
             CurrentValue["MouseOver"], 
             EdgeForm[{
               GrayLevel[0.5], 
               AbsoluteThickness[1.5], 
               Opacity[0.66]}], {}, {}], 
            
            RectangleBox[{2.6, 0.}, {2.8, 0.01749781277340335}, 
             "RoundingRadius" -> 0]},
           
           ImageSizeCache->{{153.5112520683949, 158.82410369553224`}, {
            35.73246232336927, 40.277736767357865`}}],
          StatusArea[#, 0.01749781277340335]& ,
          TagBoxNote->"0.01749781277340335"],
         StyleBox["0.01749781277340335`", {
           GrayLevel[0]}, StripOnInput -> False]],
        Annotation[#, 
         Style[0.01749781277340335, {
           GrayLevel[0]}], "Tooltip"]& ], 
       TagBox[
        TooltipBox[
         TagBox[
          DynamicBox[{
            FEPrivate`If[
             CurrentValue["MouseOver"], 
             EdgeForm[{
               GrayLevel[0.5], 
               AbsoluteThickness[1.5], 
               Opacity[0.66]}], {}, {}], 
            
            RectangleBox[{2.8, 0.}, {3., 0.005624296962879635}, 
             "RoundingRadius" -> 0]},
           
           ImageSizeCache->{{157.82410369553224`, 163.13695532266956`}, {
            38.13818426750439, 40.277736767357865`}}],
          StatusArea[#, 0.005624296962879635]& ,
          TagBoxNote->"0.005624296962879635"],
         StyleBox["0.005624296962879635`", {
           GrayLevel[0]}, StripOnInput -> False]],
        Annotation[#, 
         Style[0.005624296962879635, {
           GrayLevel[0]}], "Tooltip"]& ], 
       TagBox[
        TooltipBox[
         TagBox[
          DynamicBox[{
            FEPrivate`If[
             CurrentValue["MouseOver"], 
             EdgeForm[{
               GrayLevel[0.5], 
               AbsoluteThickness[1.5], 
               Opacity[0.66]}], {}, {}], 
            
            RectangleBox[{3., 0.}, {3.2, 0.0037495313085864237`}, 
             "RoundingRadius" -> 0]},
           
           ImageSizeCache->{{162.13695532266956`, 167.4498069498069}, {
            38.51803510078888, 40.277736767357865`}}],
          StatusArea[#, 0.0037495313085864237`]& ,
          TagBoxNote->"0.0037495313085864237"],
         StyleBox["0.0037495313085864237`", {
           GrayLevel[0]}, StripOnInput -> False]],
        Annotation[#, 
         Style[0.0037495313085864237`, {
           GrayLevel[0]}], "Tooltip"]& ], {}, 
       TagBox[
        TooltipBox[
         TagBox[
          DynamicBox[{
            FEPrivate`If[
             CurrentValue["MouseOver"], 
             EdgeForm[{
               GrayLevel[0.5], 
               AbsoluteThickness[1.5], 
               Opacity[0.66]}], {}, {}], 
            
            RectangleBox[{3.4, 0.}, {3.6, 0.0012498437695288077`}, 
             "RoundingRadius" -> 0]},
           
           ImageSizeCache->{{170.76265857694426`, 176.07551020408158`}, {
            39.024502878501536`, 40.277736767357865`}}],
          StatusArea[#, 0.0012498437695288077`]& ,
          TagBoxNote->"0.0012498437695288077"],
         StyleBox["0.0012498437695288077`", {
           GrayLevel[0]}, StripOnInput -> False]],
        Annotation[#, 
         Style[0.0012498437695288077`, {
           GrayLevel[0]}], "Tooltip"]& ]}, {}, {}}, {{}, {}, 
      {RGBColor[1, 0, 0], Thickness[Large], LineBox[CompressedData["
1:eJwV13k4VO8XAPDBYMxMSkSLkH0LSWSpc7OU/WsPRShrCSH7TrYUIUmELEV2
osK9QpaQrDMilKWUtSj7b35/3efz3Pvc5bznnPfcY3ZuRvb0OByOQofD/f+Y
eyskTHOGjH1lT49ukVCDm6nFTW2TZKzWW01jykgNFOuG6S6MkTFxepUyPX81
6N2WDtcYIGNrbIlBux1qsBMzEa76joz5uaV//XVDHSyfnotUySJjkwe/HLdp
0gC2LoZoGTMyJvm95+al55owuiDdVmZIxto8iisqqJrwnO0ys7QeGZvLma23
I2nB2YvV0cfVydiUIT5q2E0Lbny9FiMuS8Ymrgje5j6jDe3/WmMFWclY2PMU
r4wfOhAqFHP34HsS5vQ9SJZy7z9AnooPyTeRMB5Hc7k/n/4D3KFuXtN6EsZy
8EdgB9EAQsn7q5IqSdhw2xFpqf9o/p1BJT4lYfeNznw58pnmpjLhbR8ShnJK
mS3uGgKibOR+xJOEnXT4KdUpZAS4mj+vFW+SMGGc+MC8jhGEPlfUu32NhLUb
tt5mS6f5frPnogEJa+1inRmWN4ZQq2FsUoyEafIY7+sLNQFk2I+IEyJhzJm1
ad7FJoAz5Dbh4SNhS9+nCS5DJhCqbjtryUnCJDuDOy4cN4VQiZ97BuhJGK72
df/0F5rXdyzffyZimY+3iCJGFyEkVXi1KIGIWYo09zQGWoJ2aPNe7lgihq/v
vUtXZgkc123EEyKJ2JnzG4uPJy3hOfL4ilsAEeMsT67qv3AJen/u6TzpTMSO
CnjUXD94GfhUV7Pq1YlYZNjR5eJOK8AWmjV7Nlmwk8UD8jlvbSCOanMV/rJg
e4IovpUjNmDSsh1UvsKCjUwGz/xdt4Hv6QpVD36wYK09b27zKdrCPo3ioxeH
WbBaxc3Rpde2YJvxYGW8kgXzbroe1N9qB/Satk+WnVmwc0/09lEJ9iCY0fXY
z54F05kxdtWWsofzCwrpdLYsWP9H+Y6/xvYQn8L6kM2cBWO57t7JnG0PHJNv
7584z4I5JjXxMSk7gLA/Z4QHPwvWdTn87ftAR9B++cF5eYSA6WmwdqAnXODG
rryT3xABU+aQ5bS44gL3jHId6PoIGJFfNEAxwQX6//leZesgYHHBVYRPP1zg
soaw1YlaAvajffSlfcF1uPkl1MAjmYDV5lrK+h93hQf75BWWdQhY9e/jDUzK
7vDiDRccvEDAQvQ4m7fN3QG7tn4eVAkYW2b6XmEfd5ivqzdLOE3A4n+53idV
u4OmrdptMSECppO61nhCxgN2KgxqbHeYsQO8mu6isrfAxfj6yb5yZiyHzqLk
jbUXhG3rKq8XM2N90e2c9hFe8KhQSo2vkBkbPDn9Q+u5F7RuLhvdzGTGag7/
wreseAFfnu8tYhwzpuZEzg+P84ahP1EVqleZsbz935TdWm+D2sNs6aoDzFhh
QXOT6nU/uFYRSDjOxoxFbFqc+ZziB5Fd5pMFZGbMMtzjXlmjH7ynZ0t+zMCM
3XC53E1m9wdtt7C1sGUm7NYxx8ScRn8w0LRrMOhmws4ppQV6HgsE6w0BnYVI
JqxoyNEF5QqBkAM4QadQJsz9UUGCjVYIZMuMbk0GMGHRrZLD5wNCYNIhpXTw
FhPmt2XDOTUeAtf68fsbbJiwgOeFy3JmoXD95TQlXoUJo1vcI+slFga+1oUO
Yn8YsVeWzYdWf4bDvd51TdVFRixhVmzyCykCnqnqSlyaY8TipzQ1/0pEQI/w
8uLdCUYMU7X8WH0jAgQWlfyWuhix0fuT9U5LtPMhPfG1+YxYoNZ+rylcFAjk
rJZrXGTEJAbZSa8uRMNpds1kayNGLHz1IvOEazToRz329tFjxCjbH+ogJRp8
nUHphRoj5iKfLpIzGQ09J2KbyTKM2MH3M5UrwTHg+457qJ/AiP3+WD6Fb4mF
nm9qm7Zv8Ni5TK5v9/3ugkrjtaakGjzmbZhsyVZwF148iopuKsdjb18XrHT3
3YVIvTb2Y4V4jDNqr9e6ZAIo1WpLTiTjMXeNKH2+qQQojDO0snbFY9fExoy+
Wt2H0BNXUEs+PPZqNtfoFpIEC6TQqLgjeKxXrfVwtmUSXJrJ1nnDicey/Du0
F72SQD7j6/ChPXhMW6vYZuVFEvzCOyxR1hkwBcZy6r0DD8CccuPYxX4GLLSS
vl/l9wM4EeIfYXyHAePIVI/U60iB552ph4fCGDD2a92pQ9MpwMdZUWEexIB9
bLt/KpQ+FfaVzI5beTFgs1trsadUUmFhxETFyY4BU0KXb0pUpEKRvPRq4FkG
LEJFIORSzkMQXPjmULBGjyU6y2rFZDyCJ4q726Ir9NiUAO6XWf0jYI86nFo8
T49VMjP5aI09Anpuw+byKXrs+5meise86TCp1cj7to8ei3unUraUnw5Z+Y8o
H0vpMd/X+t7hrx/DQSs97XUHekyL+qpwY+sJRO5+mTtsR495xl2wKxDKhKUc
93gVK3rs0/lfhqH6mdA+k9wVYkyPZQ0csUSzM8HPfUSfCaHH8m3ne10vZMFI
hKMx+yF67LT2T5uizKfwpCjs8vEPdBi5LjBRUyEHCHrs2/rv6TCtYNGZXP0c
8FrMy3RvosM+1SX7cDrkgK5c+5eqWjpsbextolZqDmw3sNoq5dFhQ7+z3Sf/
5ID1pwz7C0F02G5KkPdMXS7w/au5aStNh50t2JtXZ5IHD0mqE4PidNjXPyLp
Ka55QObtMdQWpsPUTsQbx9zJg3WNGbmTR+mwnw/qdGrr8qAvmXMTT6LDLEJe
lcvx5EOEtE/08xkcVhNEvV43nw8zDqezFzNx2KtRrgmlp4XwcvDNpyAyDruu
tBvXql8M/oYHRFgIOKyJkmJp6loMWt1ugSkMOKzMs7hvJ74YPsRq6Zut7qKD
39q/ZnUUw1mno+vln3bR8oThuxznX4KgUKu+Q9wuSlInS35RK4HlLI6N3s0d
dO75QquocRnEPagyKPiyjUod/NktOVkB+06bsm4ObKMXbXPrbFYr4OHY2of/
Pmyj7Il/JvJYKiFHVOnCeu022uNm8s9KthLqGrEzuknbqPVD6+zkyEqYnesW
X1HbRvM/o+l7JatAXe07/uyLLVRmLlfXKroadn5zvx703ET58KE/qt1qwW+k
MctVdh0dSnLqG0logFPKdyEOW0NDgmVPqKW8A3J7QttByz9o/WqItmHDe/hF
bevae3QFXbdgM3Oc7YTDXZaHua4som0lb/SyUj7C9YWHqeNDc2iAw6fMPWuf
ABjFoyvmp1G2b1df3pIZgK0VMWunkWmUn2tIo+30ANRNiJ3ibZtGCS8mfnOd
GwCZerFvd7OnUb/TvG4lhgPA5ykGTsbTqO22CPn5rQGgnxT9y/N6CqWiLHKD
VQPQUi/ieDfyG0qU3Cc1IjsIWl5CFxyPTKJ7yiWfzh8bAinhvXbOLJPo1qqg
4APRIdhP+Rd4/e8EmuMpm3dCeghGlLsq3fon0LT87mBblSG4weDJ4xs/gWa5
++xGmA0Bj1DYT7a/4yiTeIP6x9ghKLat1TStHUM7bZzUzH8NgVdrXLBF8hia
VsRob7gyBCpi1tVWbmPoW3OlPvV/Q9CzhOdzEBlDJVy8/A/jh2ElxOjv7bRR
VIvhknvokWE4nTWf/8jnM1r7SsFDVmsYWkYEGEYVqGhPgFchKXsYTqy9iKTs
o6IZwQMvq/OHIYtNhmnwBwWt/PT6nFnxMPhqqrD0ZFDQoSYN+ohXwyDxymQv
tjuMjtidUwzvGoYHSVFH8tqG0A3/tRv7/w6DrebsyRsXB1CWEZ0xo/MU6Lnq
WuMkM4BWMVj20+tQQCnkt7w9ywB6UcbpedF/FOB4tato/bYfvc3p93nSnALt
AgcRA95+VGl8emTehQIyu5q6crOf0MytYUbfBApsXTft+Dffg7oaYbhTXRTw
qSKM8uT3oEWHgs4//kiBlY23C+qXe9Dt3eJ/f/soMBvLz5HU2Y3+WU5MfUKl
QF/horVYYReq2W9kETdDgeffYv9Y2HSiiW1Z1qs7FBCQOMMcytWJNijVrY/T
USHr1tKhgp4OtJBf4dM7PBWS6S7CikoHSj91SdyDSIVgXsG4uEPtqGxqqqL7
ASqYXPp/v2xF+S2c76yKUeFjrofsZGwrqt0xUmwnSQXtOUEN5nOtKNuesak2
KSqAX7yLcVkLqkO2i/E/SQXxNPOan/HNaKF0tVmoChXyv5Da2dSa0TrVRmrJ
WSrwCaMjChvv0N26Out+hAoHaoRwkU7v0BiFsw5MGlTA9a9oc2s0obi3HifJ
elQIOFxgpbqFofxHyxPW9amwZmvh7lSFoRFEwZEvBlT4uYSmVh/D0N5iG9Uk
EyoUHz3clL7VgK53dfCkXqKCWar+r42br1HqieJULicqGEmHxwtM1KGn07hO
BzpTQb+jRlzXsA6NHjqHp7pQ4cI2t3PmyVr0IKmFI8CVCqftf07Dv2pUH2n+
Y3OLCnK7PFFOztUoR4108x1PKsikGwkmjVSh/B67FXleVBDtfm33tb4STWMo
+NV+mwqHTsWMR4aVo/eNXr2p9KfFJ8tn23K6DG3ML3dPD6CCQN8X5ZdtpeiG
Km12CKQCh1JJnf7dl2jaxSRBnmAqkG9y/H16sxjdllVWnKaZMTfg1LJBETrJ
X3o1L4QWDxbtyuQDz9G8A8f/EsOosHCmYmnqbwH6zGvOupLmWY+D0vIj+ei1
3H6qYTgVKNSZYkrWM3TcaozJL4IKvXv058TCclFzm7laXCQV2s/ViAZczUEb
uLf8Qmm+Zcyso2yQhV5kNZd2iqLC1beDbWMcT9Canx3Hemk2EchTD6Wmo3kD
9wVk7lBBPf5WE39WGoryG8jF0HzqN3K21S4VVU+vM6bQLHxp71tHkWQ0xT44
jC+aClzNYwrEX4loIa91gw3N+8zMnS1sE1DvziPM6TQvC63aZc3FoGYD4dYd
NPetJl3+5hmJ6nE5NC/T7BRv9dJgIwQ1pT49tT+GCpISaU9HyD5oRhdbjTjN
X1JzhDrO3kCL9uafU6a5o0js4Js2E/T+ce3PajSbvBw/pTKqB6tH18PUab5X
jBfZNHQGSY2KU2doxr/SzPZ86A1SXn5rx2lGOPeW/IgIhuTvV1o4aT77dm60
rC4CAnX9n/6lvc/ln3rfXKRjwO3VdHQvzQFHKn4I0eY4sRfvg3Jo7hYi+rrf
S4Qb7FphLjSPNihvdnc8gIr/sh9I0PzT1DVYgjEVJgSYK6do8dqYz6KLRdJg
EUPHU2km3umNnAlIh5B37NxA8yEeeoJ6bQbYS5k4TtDWQ/TVyficlUw46UVt
8qOZu1A2epczG/JvMEiSaBb0/KBZKpcDBywUC1Jo6ysB14hWRrkw+PWHDCfN
SsOpCW8S8uA/n8d+dLT84Jt5tFzQnA9B4ZePWNPyh3k1wzR5vQB4nfKdKmj5
NbA/9+gN+xcQT+0iqIRS4e2x/DDzjCJ4G6Fh5UnLz1yZ59Pqn4rhnKqfTQ4t
f930S0uOnikF96evbk/S8t3MqoKN6FkGOPaX8b9p9cAjP6IUYFIOUufkjLdo
9cKoMZblGFcBcg/Ud5Z9aetv980BWasCaVzqlCut3jTmph7JJ1XDza+E0lO0
eqzzmOmUlKwBP5XCk388qPA09IfUIbtXUDe46GvqRoUbWYtry911EJLQwPGC
Vv/jQsui351eA5e8znfh//eHkhXLLwxvgOMef2SGAxUU61cbOxXfwsp64BHn
q1RgGtm88yy/AU78qN1IvUwFf9vtunSkEXS/Cah2WlJh/vvO3P3PjfCixt9m
1ZwKc5uhIo0EDELz18RkTGn1RoyMLzrbBEdNKsVP0/pblmiccURRM0w3jw8n
K1PB/FfcjZ2/zWDliEuRV6QCW3l8lJ9GC8gp1un0ylMhUiGh1m2iBf68c637
foIK188nHrnM+R54RO/1lYjQ+tW1h9/kwttBnGstKp6NCoZcS7/CO9uhF2NP
+MRKBZdOrbXe/R1Q+h/b171kKmTIbBNcn3UA/X7eUG8mKuxsXZXKb+mEjDq5
4akNCrSknvA7wNwN0hfeWpl9pcCYZnz41f+6QTusP45/nAJrm1PxFWndIPG9
SXb2MwVE7B5l6Yn2QBunwa7FIAXipHabo7Q+AsO+1cnpNtp+ddWBebO1F14F
BaWnF1NA7gw7mX+uF/Y52C4uFFKgiRPbp8X6CX5kjv9WyaPASMehw2lmnyDn
eThn8xMKkGS6JU/OfoK7LJWC/99Pb26dNLpO6IeaK9x3Y1wpcDKVLvOz9iAY
vWYmWolSgLt205R4axCGvu6/7CpAATx1lVUxfRDuLOMKvHkoMMA9F/pwdhAU
tvYo3OSggHde3zXDyCHYvGo1s7s7DLVVzyTbGoaBI7fBrGhwGJT7NOorZUYA
b7exoBYwDGp74z7Hco7Dt/aNq1PlQ+CysO823ZlxwCnM65wrGoLE7rR9flfH
Ibmd72/asyH4Ep+v4VQ+Dp303ypkH9LmociJiwJCEyCgRe/FEzAEQ7itJjp0
Ar4HSbEPqg0BcUP2YePKJJywwU8t9A6C56/ss6ctp0DwyM+rdp8HQL6NlGo5
9B3af8f66k71gXcJr7dn1Tw8nGz51fqoF/rVbUvGeJah1Ozd44CeLvh0Z6fu
vvlv2P4jTbmX2g4J+u4sDe9WgTM0iNsRaQHVpBejYQz/oIOtwTr1HgbVFrV/
mrw24KN3YP5F5jfAk3n3fnr9Fty4hM3hLlVDwLFLwvGtW6AQsXTl0flqoOSL
NQT2bMF+lcbNM7LVkFz6fu7KxBbYVB/JqWKpBgKGOy+M34bslUpvzddVsPLV
a6tSZxtCo5BaziNV8F7UyrlrZBua9Z5R5X9UgGv1cbXdfztQyy+oP2JVBl/j
9rZV0u8CmzJifEixDMxsl7UdyLtA2FHf8uEog7OsNUbdvLvArvh7X8qHUmB1
UrHLOL8LOvxWfOeVS6H8iE6YQsouJHiPucTwlYD/uerxWi4cEjCnun6Kvhga
n386ulSJQ6wW7Np9eQthgb+d1acWhyQeVtyVwxfSvr9xd/stDqkOn2Mk/iiA
4KTiSVIrDpmu5jXeqSyAs/5R+SLDOKSqLewJq2YBoDqKx69s4hBpBQvt5tv5
gC1kq/So0SFOf720hGeewZJT2nFTTTrEY0msaqr7GfB9S+AZ1aVDVOZMrd/U
PIPQYX/cnCkdUiisp1Ec9QwQzKSF0YkOUSXN/8gRfgZYEkH3zF065F9Q/ZHh
G7nQJOd+6eUAHTIrUuyOJ+bA1sUl1RQqHaL4c+Zo20o2nA5wFw/8QoeMbHgI
P/ucDeVN7us63+mQ0var+MKSbMjS90j7uUmHzCxK7h01ygZ/p1v94gL0CNPd
9EsUtacg+8RL54UHPRL4jT3YJO8J3ET/yCbdpkeejCQenw94Ai++eh32C6BH
RGwb5KqNnwCfmPcPzSh6JEud3NDF8ARYX3lHf39EjxCsHKkCDhkw9/F2swhG
jyguSF/WOfkYnjH4KRewMiBZ2vP2MQtpcKdRHHzYGZAz14Tx2l1p4OI3qqp5
kAGRvybAKleUBjKLtJH0GAMC9s4ToQ5p0DjCZCF1igFZL2AdXZ94CJSKlNuv
LjEg//ZrSumOpQKrTXnF++cMiGrsb4v4lWRYPmxXk1bCgIj9OeBweDAZBgbZ
XztVMiCtnRlYd20yZOj4YMR6BoS/jXE8OygZRBXOftT/yIAI+ZbLrxKTQY21
69fQKgMSuvfDYSXJB+BbPysyq4pHdNlH2ExjEsGQ6aml/QU8si9hiXvcPRHE
DcwSvungkS8GZdpRFokw9q1lZdwEj5hOHHohJJEIqqTcBooDHnm954ON49P7
QLa8bNwZh0cq5jq2x5TvQfbf3uCSPjyieDv9ncuTePA7F1MhMYxHLvI85BLx
igejeJh68RmPeJW2RR/QjQdG3lKtgik8Et4ruRG9FQcu5+M5nq7hEdu1+oWG
K3Egl6pRlHiYEZGbt8xRl46FDtk3g552jAjjkJbD7eU7UCq0uWPkyIiYXm7J
Uei6Aw8OqojK3mBEuvk45iUL78DlnQb/JW9GZDM+qqfS6g4sdjTxusYyIjvv
6iKEuqPggG2Hs0M5I9KfG2XvXRMJtvcp2+Y7jIiEvv1v5aJwOB9+SOQ0AxMi
HKrOxh4fDuLelgZcBCbkQ+uuNdeNcPhjOZo7yMaEKJZ31xZIhUO08MQFIyEm
5LbarcKRmjAoaZhN0tFlQqKD0XMuXaGw/nNN6OxjJmTk7iJ690QwDLieCe7M
YkLCZIUI0nTBULoYMWT2jAkRO0IQOfApCK793hft9pIJ0eklVhZ7BEHPhvj3
7EYmxI7RN269JhDyWK4UMXxjQszeoZ6D2gHwn0ibVIcEMxLPur8/4pkviL/Y
E20qw4zo9Xh1sgb4Al7CZHxSjhnBDaUYDxj5whupyfsbZ2jm3aZjZfAFQfnN
JQkDZsRHak1SzsEHNtSlq+55MSOH+7qc9yrchny7NAWTBmbEwVyBUYLOC0QD
977saGJGOuzntI5PeEJRajQfvGdG6oTRgSuoJ5S2+7CIf2RGTLHUjybBnlAn
Zf55d4IZ4aw7+MNn5xZ0bRwMKcYTkGSbrj435lvw58Hj9/S6BESHWTb+sZI7
3H65X9nXgICg9y6myPO6w3prXNm8Ce36d5e/bDK4w84//0cUKwJSurDQTP/E
DQhXLrmUuhGQD5jNyFTbTeCW4Ga1TCYg0qwPe4OlXEG9OdO0YoSAOBM7m3Xk
XKAv6ikWPU5AHs1yRysxuICNZo6E9RQB0c048CagzxkCu/PoSAsEpOh+QfJv
D2eoHi4uuUbHgijxi5x8We0Egr9eMx4UZUHeeMu2fNZ0BIYDQzXB3iyI3Ij9
tMCLa5A4PMxn6s+CHGy0Yo0MuwY8j6nxEiEsSNCDkhZRy2ugxDtmS4lhQdRY
lM6eJV+DW2JTe2SfsCAWqfyq9zyvwtczK/bT71gQm1wlxuv6dvDOgZVTdx8R
UdZTe+MpawPn6w6Uf+QgIr0m1T9bOWygk+WotvEhImJaFtj/YOEK9L2UCLbk
JyLS2o6W7/2uwOSK5oyTHBFZsg7lqU6xBlxIxKsocyJi1q4aIjN6GSDtnxma
TUQu6zQb75RYwLvvu0uq+UTEWnrh2a8gCzivxBz//gUR2dyrMCGpbwH6YwfQ
7koiQmA+eiJn0RysBU6KjLYQEeKG/6SOvDkEl7n+/fediJzU/W8I6TOD+rav
abInSMgpqoBc+DkTKHPzSFORJyE6TrOu+pwmkMuFSzuvTEIea+Kmb/40hljH
ow8tNUgIrsFcveihMZgxm6eEW5AQVvl/jJtLRrB4vjuxL4yE5Do7yMVWG8LX
hUuJo3dIyMLUnUKze4Yw9HDu/kw8Ccl0/PnYzckQ6meY72+kkpDW3L2LYUcN
IeaOagJ/EQmZEZlbPBRvAPzva2Nv9ZGQC478VL1ZfThw83xs4DAJOfpO8Or4
M30gcA7G3BklIfg/uvT/bPRh0X4l+vEMCfnqwfP6De0/rZ7x+J13GyTkrlL9
9+sjumCqkRu+X4CMdH6QXCuf1wbNeZlwblEy8jlf0UujUhtUUtEw4eNk5GtJ
5n0jH23gnx4NVVIgI4lSb2X66bVhIZIrxE6HjGjEqjua8WrBpGRB8A0DMuJ9
PYbt0awmDAzIBd82JSNHJ3Sydco14a2gUVDcFTKyId4rwqimCaVdE4Ep12jP
c2YJZ9mjCbleboFZzmTk+0RZRODwBUjl3gl4fpOMEOsqbKxyL0BMy92ASk/a
/W+e0y92vQD/A29qMmQ=
        "]]}}},
    AspectRatio->NCache[GoldenRatio^(-1), 0.6180339887498948],
    Axes->{True, True},
    AxesOrigin->{-3.8, 0},
    FrameTicks->{{Automatic, Automatic}, {Automatic, Automatic}},
    GridLines->{None, None},
    PlotRange->NCache[{{
        Rational[-19, 5], 
        Rational[18, 5]}, {All, All}}, {{-3.8, 3.6}, {All, All}}],
    PlotRangePadding->{{
       Scaled[0.02], 
       Scaled[0.02]}, {
       Scaled[0.02], 
       Scaled[0.1]}},
    Ticks->{Automatic, Automatic}]}], "}"}]], "Output",
 ImageSize->{389, 119},
 ImageMargins->{{0, 0}, {0, 0}},
 ImageRegion->{{0, 1}, {0, 1}},
 CellLabel->"Out[10]=",
 CellID->154096072]
}, Open  ]]
}, Closed]]
}, Open  ]],

Cell[CellGroupData[{

Cell[TextData[ButtonBox["MORE ABOUT",
 BaseStyle->None,
 Appearance->{Automatic, None},
 Evaluator->None,
 Method->"Preemptive",
 ButtonFunction:>(FrontEndExecute[{
    FrontEnd`SelectionMove[
     FrontEnd`SelectedNotebook[], All, ButtonCell], 
    FrontEndToken["OpenCloseGroup"], 
    FrontEnd`SelectionMove[
     FrontEnd`SelectedNotebook[], After, 
     CellContents]}]& )]], "MoreAboutSection",
 CellID->38303248],

Cell[TextData[ButtonBox["Metropolis",
 BaseStyle->"Link",
 ButtonData->"paclet:Metropolis/guide/Metropolis"]], "MoreAbout",
 CellID->612972363]
}, Open  ]],

Cell[" ", "FooterCell"]
},
Saveable->False,
ScreenStyleEnvironment->"Working",
WindowSize->{725, 730},
WindowMargins->{{0, Automatic}, {Automatic, 0}},
WindowTitle->"BlockAdaptiveMetropolis - Wolfram Mathematica",
TaggingRules->{
 "ModificationHighlight" -> False, 
  "Metadata" -> {
   "built" -> "{2013, 2, 10, 17, 35, 23.0998216}", "context" -> "Metropolis`",
     "keywords" -> {}, "index" -> True, "label" -> 
    "Metropolis Package Paclet Symbol", "language" -> "en", "paclet" -> 
    "Metropolis Package", "status" -> "None", "summary" -> 
    "BlockAdaptiveMetropolis[logpdf, x_0, size, stepsize, adpfraction, \
adprate, adpset] generates size updates for a list of conditional indepdent \
random variables starting from x0. Each update is given by the last update of \
stepsize BlockStudentTMetropolis updatings. The target distribution p[x] is \
provided by logpdf[x, indices] = Log[p[x, indices]] for any subset \
x[[indices]] of x. The symmetric proprosal distribution is the student T \
distribution. The variance of the proposal distribution is adapted in the \
initial Floor[size adpfraction] iterations with rate adprate, and the \
adaptive parameter set adpset.", "synonyms" -> {}, "title" -> 
    "BlockAdaptiveMetropolis", "type" -> "Symbol", "uri" -> 
    "Metropolis/ref/BlockAdaptiveMetropolis"}, "LinkTrails" -> "", 
  "SearchTextTranslated" -> ""},
CellContext->"Global`",
FrontEndVersion->"8.0 for Microsoft Windows (64-bit) (February 23, 2011)",
StyleDefinitions->Notebook[{
   Cell[
    StyleData[
    StyleDefinitions -> FrontEnd`FileName[{"Wolfram"}, "Reference.nb"]]], 
   Cell[
    StyleData["Input"], CellContext -> "Global`"], 
   Cell[
    StyleData["Output"], CellContext -> "Global`"]}, Visible -> False, 
  FrontEndVersion -> "8.0 for Microsoft Windows (64-bit) (February 23, 2011)",
   StyleDefinitions -> "Default.nb"]
]
(* End of Notebook Content *)

(* Internal cache information *)
(*CellTagsOutline
CellTagsIndex->{
 "PrimaryExamplesSection"->{
  Cell[6109, 178, 461, 13, 70, "PrimaryExamplesSection",
   CellTags->"PrimaryExamplesSection",
   CellID->217326441]}
 }
*)
(*CellTagsIndex
CellTagsIndex->{
 {"PrimaryExamplesSection", 359121, 8806}
 }
*)
(*NotebookFileOutline
Notebook[{
Cell[616, 21, 76, 1, 70, "SymbolColorBar"],
Cell[695, 24, 185, 5, 70, "LinkTrail"],
Cell[883, 31, 2403, 55, 70, "AnchorBarGrid",
 CellID->1],
Cell[CellGroupData[{
Cell[3311, 90, 66, 1, 70, "ObjectName",
 CellID->1224892054],
Cell[3380, 93, 1244, 27, 70, "Usage"]
}, Open  ]],
Cell[CellGroupData[{
Cell[4661, 125, 579, 17, 70, "NotesSection",
 CellGroupingRules->{"SectionGrouping", 50},
 CellID->87774412],
Cell[5243, 144, 291, 8, 70, "Notes",
 CellID->115902181],
Cell[5537, 154, 394, 13, 70, "Notes",
 CellID->486766318],
Cell[5934, 169, 138, 4, 70, "Notes",
 CellID->894215170]
}, Closed]],
Cell[CellGroupData[{
Cell[6109, 178, 461, 13, 70, "PrimaryExamplesSection",
 CellTags->"PrimaryExamplesSection",
 CellID->217326441],
Cell[CellGroupData[{
Cell[6595, 195, 148, 5, 70, "ExampleSection",
 CellID->115992704],
Cell[6746, 202, 118, 3, 70, "Input",
 CellID->21081259],
Cell[6867, 207, 93, 1, 70, "ExampleText",
 CellID->568556038],
Cell[CellGroupData[{
Cell[6985, 212, 2449, 77, 70, "Input",
 CellID->73930184],
Cell[9437, 291, 224, 7, 36, "Output",
 CellID->327609976]
}, Open  ]],
Cell[9676, 301, 54, 1, 70, "ExampleText",
 CellID->274231],
Cell[CellGroupData[{
Cell[9755, 306, 703, 22, 70, "Input",
 CellID->44431038],
Cell[10461, 330, 453, 15, 36, "Output",
 CellID->359490002],
Cell[10917, 347, 4404, 80, 135, "Output",
 CellID->591258976]
}, Open  ]],
Cell[15336, 430, 57, 1, 70, "ExampleText",
 CellID->377051331],
Cell[CellGroupData[{
Cell[15418, 435, 719, 21, 70, "Input",
 CellID->254894932],
Cell[16140, 458, 271, 9, 36, "Output",
 CellID->58698215],
Cell[16414, 469, 4318, 78, 146, "Output",
 CellID->52035]
}, Open  ]],
Cell[20747, 550, 54, 1, 70, "ExampleText",
 CellID->60397930],
Cell[CellGroupData[{
Cell[20826, 555, 603, 18, 47, "Input",
 CellID->111189838],
Cell[21432, 575, 14822, 250, 140, "Output",
 CellID->376985250]
}, Open  ]],
Cell[CellGroupData[{
Cell[36291, 830, 2037, 63, 70, "Input",
 CellID->305510931],
Cell[38331, 895, 142701, 3603, 70, "Output",
 CellID->547936853]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[181081, 4504, 227, 7, 70, "ExampleSection",
 CellID->65407496],
Cell[181311, 4513, 119, 3, 70, "Input",
 CellID->433947188],
Cell[181433, 4518, 153, 4, 70, "ExampleText",
 CellID->73418448],
Cell[CellGroupData[{
Cell[181611, 4526, 4139, 125, 70, "Input",
 CellID->756940083],
Cell[185753, 4653, 223, 7, 36, "Output",
 CellID->77302206]
}, Open  ]],
Cell[185991, 4663, 55, 1, 70, "ExampleText",
 CellID->24993314],
Cell[CellGroupData[{
Cell[186071, 4668, 699, 22, 70, "Input",
 CellID->265777506],
Cell[186773, 4692, 452, 15, 36, "Output",
 CellID->23471563],
Cell[187228, 4709, 4522, 82, 70, "Output",
 CellID->590140632]
}, Open  ]],
Cell[191765, 4794, 56, 1, 70, "ExampleText",
 CellID->12498044],
Cell[CellGroupData[{
Cell[191846, 4799, 717, 21, 70, "Input",
 CellID->1468033],
Cell[192566, 4822, 274, 9, 36, "Output",
 CellID->12145802],
Cell[192843, 4833, 4638, 84, 70, "Output",
 CellID->5639925]
}, Open  ]],
Cell[197496, 4920, 54, 1, 70, "ExampleText",
 CellID->193398859],
Cell[CellGroupData[{
Cell[197575, 4925, 573, 17, 70, "Input",
 CellID->402775542],
Cell[198151, 4944, 10524, 180, 70, "Output",
 CellID->187378485]
}, Open  ]],
Cell[CellGroupData[{
Cell[208712, 5129, 608, 18, 70, "Input",
 CellID->223197944],
Cell[209323, 5149, 13566, 230, 70, "Output",
 CellID->628946702]
}, Open  ]],
Cell[CellGroupData[{
Cell[222926, 5384, 722, 23, 70, "Input",
 CellID->36445872],
Cell[223651, 5409, 4809, 86, 70, "Output",
 CellID->69605284]
}, Open  ]],
Cell[CellGroupData[{
Cell[228497, 5500, 1503, 47, 70, "Input",
 CellID->182745918],
Cell[230003, 5549, 3375, 63, 70, "Output",
 CellID->580742033]
}, Open  ]],
Cell[CellGroupData[{
Cell[233415, 5617, 1246, 38, 70, "Input",
 CellID->530831861],
Cell[234664, 5657, 40754, 1053, 70, "Output",
 CellID->193848635]
}, Open  ]],
Cell[CellGroupData[{
Cell[275455, 6715, 1360, 39, 70, "Input",
 CellID->271797975],
Cell[276818, 6756, 79675, 1979, 70, "Output",
 CellID->154096072]
}, Open  ]]
}, Closed]]
}, Open  ]],
Cell[CellGroupData[{
Cell[356554, 8742, 419, 12, 70, "MoreAboutSection",
 CellID->38303248],
Cell[356976, 8756, 143, 3, 70, "MoreAbout",
 CellID->612972363]
}, Open  ]],
Cell[357134, 8762, 23, 0, 70, "FooterCell"]
}
]
*)

(* End of internal cache information *)

