(* Content-type: application/vnd.wolfram.mathematica *)

(*** Wolfram Notebook File ***)
(* http://www.wolfram.com/nb *)

(* CreatedBy='Mathematica 8.0' *)

(*CacheID: 234*)
(* Internal cache information:
NotebookFileLineBreakTest
NotebookFileLineBreakTest
NotebookDataPosition[       157,          7]
NotebookDataLength[    281321,       6867]
NotebookOptionsPosition[    275940,       6692]
NotebookOutlinePosition[    277903,       6742]
CellTagsIndexPosition[    277816,       6737]
WindowTitle->BlockStudentTMetropolis - Wolfram Mathematica
WindowFrame->Normal*)

(* Beginning of Notebook Content *)
Notebook[{
Cell[" ", "SymbolColorBar",
 CellMargins->{{Inherited, Inherited}, {-5, 0}}],

Cell[TextData[{
 ButtonBox["Metropolis",
  BaseStyle->{"Link", "LinkTrail"},
  ButtonData->"paclet:Metropolis/guide/Metropolis"],
 StyleBox[" > ", "LinkTrailSeparator"]
}], "LinkTrail"],

Cell[BoxData[GridBox[{
   {Cell["METROPOLIS PACKAGE PACLET SYMBOL", "PacletNameCell"], Cell[TextData[{
     Cell[BoxData[
      ActionMenuBox[
       FrameBox["\<\"More About \[RightGuillemet]\"\>",
        StripOnInput->False], {"\<\"Metropolis\"\>":>Documentation`HelpLookup[
       "paclet:Metropolis/guide/Metropolis"]},
       Appearance->None,
       MenuAppearance->Automatic]],
      LineSpacing->{1.4, 0}],
     "\[ThickSpace]\[ThickSpace]\[ThickSpace]\[ThickSpace]\[ThickSpace]\
\[ThickSpace]",
     Cell[BoxData[
      ActionMenuBox[
       FrameBox["\<\"URL \[RightGuillemet]\"\>",
        StripOnInput->
         False], {"\<\"Metropolis/ref/BlockStudentTMetropolis\"\>":>
       None, "\<\"Copy Mathematica url\"\>":>
       Module[{DocumentationSearch`Private`nb$}, 
        DocumentationSearch`Private`nb$ = NotebookPut[
           Notebook[{
             Cell["Metropolis/ref/BlockStudentTMetropolis"]}, Visible -> 
            False]]; SelectionMove[
         DocumentationSearch`Private`nb$, All, Notebook]; 
        FrontEndTokenExecute[DocumentationSearch`Private`nb$, "Copy"]; 
        NotebookClose[DocumentationSearch`Private`nb$]; Null], 
       Delimiter, "\<\"Copy web url\"\>":>
       Module[{DocumentationSearch`Private`nb$}, 
        DocumentationSearch`Private`nb$ = NotebookPut[
           Notebook[{
             Cell[
              BoxData[
               MakeBoxes[
                Hyperlink[
                "http://reference.wolfram.com/mathematica/Metropolis/ref/\
BlockStudentTMetropolis.html"], StandardForm]], "Input", TextClipboardType -> 
              "PlainText"]}, Visible -> False]]; 
        SelectionMove[DocumentationSearch`Private`nb$, All, Notebook]; 
        FrontEndTokenExecute[DocumentationSearch`Private`nb$, "Copy"]; 
        NotebookClose[DocumentationSearch`Private`nb$]; 
        Null], "\<\"Go to web url\"\>":>FrontEndExecute[{
         NotebookLocate[{
           URL[
            StringJoin[
            "http://reference.wolfram.com/mathematica/", 
             "Metropolis/ref/BlockStudentTMetropolis", ".html"]], None}]}]},
       Appearance->None,
       MenuAppearance->Automatic]],
      LineSpacing->{1.4, 0}]
    }], "AnchorBar"]}
  }]], "AnchorBarGrid",
 GridBoxOptions->{GridBoxItemSize->{"Columns" -> {
     Scaled[0.65], {
      Scaled[0.34]}}, "ColumnsIndexed" -> {}, "Rows" -> {{1.}}, 
   "RowsIndexed" -> {}}},
 CellID->1],

Cell[CellGroupData[{

Cell["BlockStudentTMetropolis", "ObjectName",
 CellID->1224892054],

Cell[BoxData[GridBox[{
   {"", Cell[TextData[{
     Cell[BoxData[
      RowBox[{"BlockStudentTMetropolis", "[", 
       RowBox[{
        StyleBox["logpdf", "TI"], ",", " ", 
        SubscriptBox[
         StyleBox["x", "TI"], 
         StyleBox["0", "TR"]], ",", " ", 
        StyleBox["s", "TI"], ",", " ", 
        ButtonBox["V",
         BaseStyle->"Link",
         ButtonData->"paclet:ref/V"], ",", " ", 
        StyleBox["n", "TI"]}], "]"}]], "InlineFormula"],
     "\[LineSeparator]generates n updates for a set of conditional indepdent \
random variables starting from x0. The target distribution p[x] is provided \
by logpdf[x, indices] = Log[p[x, indices]] for any subset x[[indices]] of x. \
The symmetric proprosal distribution is the student T distribution with mean \
being the current value, (co)variance s^2 V, and degree infinity."
    }]]}
  }]], "Usage",
 GridBoxOptions->{
 GridBoxBackground->{
  "Columns" -> {{None}}, "ColumnsIndexed" -> {}, "Rows" -> {{None}}, 
   "RowsIndexed" -> {}}}]
}, Open  ]],

Cell[CellGroupData[{

Cell[TextData[Cell[BoxData[
 ButtonBox[
  FrameBox[
   StyleBox[
    RowBox[{"MORE", " ", "INFORMATION"}], "NotesFrameText"],
   StripOnInput->False],
  Appearance->{Automatic, None},
  BaseStyle->None,
  ButtonFunction:>(FrontEndExecute[{
     FrontEnd`SelectionMove[
      FrontEnd`SelectedNotebook[], All, ButtonCell], 
     FrontEndToken["OpenCloseGroup"], 
     FrontEnd`SelectionMove[
      FrontEnd`SelectedNotebook[], After, CellContents]}]& ),
  Evaluator->None,
  Method->"Preemptive"]]]], "NotesSection",
 CellGroupingRules->{"SectionGrouping", 50},
 CellID->200745645],

Cell[TextData[{
 "To use BlockStudentTMetropolis, you first need to load the ",
 Cell[BoxData[
  ButtonBox["Metropolis",
   BaseStyle->"Link",
   ButtonData->"paclet:Metropolis/guide/Metropolis"]], "InlineFormula"],
 " package using  Needs[\"Metropolis`\"]."
}], "Notes",
 CellID->115902181],

Cell[TextData[{
 "BlockStudentTMetropolis has options of ",
 Cell[BoxData[
  ButtonBox["Metropolis",
   BaseStyle->"Link",
   ButtonData->"paclet:Metropolis/ref/Metropolis"]], "InlineFormula"],
 " and ",
 Cell[BoxData[
  ButtonBox["StudentTDisplacements",
   BaseStyle->"Link",
   ButtonData->"paclet:Metropolis/ref/StudentTDisplacements"]], 
  "InlineFormula"]
}], "Notes",
 CellID->486766318],

Cell["\<\
Each update is given by {x, logpdf[x], 1/0 (accept/not), \
Min[1,Metropolis-Hastings ratio]}.\
\>", "Notes",
 CellID->894215170]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[ButtonBox["EXAMPLES",
 BaseStyle->None,
 Appearance->{Automatic, None},
 Evaluator->None,
 Method->"Preemptive",
 ButtonFunction:>(FrontEndExecute[{
    FrontEnd`SelectionMove[
     FrontEnd`SelectedNotebook[], All, ButtonCell], 
    FrontEndToken["OpenCloseGroup"], 
    FrontEnd`SelectionMove[
     FrontEnd`SelectedNotebook[], After, 
     CellContents]}]& )]], "PrimaryExamplesSection",
 CellTags->"PrimaryExamplesSection",
 CellID->5934563],

Cell[CellGroupData[{

Cell[TextData[{
 "Basic Examples",
 "\[NonBreakingSpace]\[NonBreakingSpace]",
 Cell["(1)", "ExampleCount"]
}], "ExampleSection",
 CellID->1464540],

Cell[BoxData[
 RowBox[{"Needs", "[", "\"\<Metropolis`\>\"", "]"}]], "Input",
 CellLabel->"In[1]:=",
 CellID->16158053],

Cell["This samples many independent normal distributions", "ExampleText",
 CellID->75377655],

Cell[CellGroupData[{

Cell[BoxData[{
 RowBox[{
  RowBox[{"Clear", "[", 
   RowBox[{
   "logpdf", ",", "scale", ",", "cov", ",", "m", ",", "mu", ",", "sd", ",", 
    "n", ",", "res"}], "]"}], ";"}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{"m", "=", "3"}], ";"}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{"mu", "=", 
   RowBox[{"RandomReal", "[", 
    RowBox[{
     RowBox[{"{", 
      RowBox[{"0", ",", "50"}], "}"}], ",", "m"}], "]"}]}], 
  ";"}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{"sd", "=", 
   RowBox[{"RandomReal", "[", 
    RowBox[{
     RowBox[{"{", 
      RowBox[{"1", ",", "5"}], "}"}], ",", "m"}], "]"}]}], 
  ";"}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{"logpdf", "=", 
   RowBox[{"Function", "[", 
    RowBox[{
     RowBox[{"{", 
      RowBox[{"x", ",", "indices"}], "}"}], ",", 
     RowBox[{
      RowBox[{"-", 
       RowBox[{
        RowBox[{"(", 
         RowBox[{
          RowBox[{"x", "[", 
           RowBox[{"[", "indices", "]"}], "]"}], "-", 
          RowBox[{"mu", "[", 
           RowBox[{"[", "indices", "]"}], "]"}]}], ")"}], "^", "2"}]}], "/", 
      RowBox[{"(", 
       RowBox[{"2", " ", 
        RowBox[{
         RowBox[{"sd", "[", 
          RowBox[{"[", "indices", "]"}], "]"}], "^", "2"}]}], ")"}]}]}], 
    "]"}]}], ";"}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{"scale", "=", 
   RowBox[{"3", " ", "sd"}]}], ";"}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{"cov", "=", 
   RowBox[{"Table", "[", 
    RowBox[{
     RowBox[{"{", 
      RowBox[{"{", "1", "}"}], "}"}], ",", 
     RowBox[{"{", "m", "}"}]}], "]"}]}], ";"}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{"n", "=", "5000"}], ";"}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{
   RowBox[{"res", "=", 
    RowBox[{"BlockStudentTMetropolis", "[", 
     RowBox[{"logpdf", ",", 
      RowBox[{"Table", "[", 
       RowBox[{"0", ",", 
        RowBox[{"{", "m", "}"}]}], "]"}], ",", "scale", ",", "cov", ",", 
      RowBox[{"2", " ", "n"}], ",", 
      RowBox[{"IsReturnLast", "\[Rule]", "False"}]}], "]"}]}], ";"}], "//", 
  "Timing"}]}], "Input",
 CellLabel->"In[2]:=",
 CellID->223834254],

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{"2.2620000000000005`", ",", "Null"}], "}"}]], "Output",
 ImageSize->{92, 15},
 ImageMargins->{{0, 0}, {0, 0}},
 ImageRegion->{{0, 1}, {0, 1}},
 CellLabel->"Out[2]=",
 CellID->220518226]
}, Open  ]],

Cell["Mean accept ratio:", "ExampleText",
 CellID->12821957],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"N", "[", 
  RowBox[{"Mean", "[", 
   RowBox[{"res", "[", 
    RowBox[{"[", 
     RowBox[{
      RowBox[{"n", ";;"}], ",", 
      RowBox[{
       RowBox[{"-", "2"}], ";;"}]}], "]"}], "]"}], "]"}], "]"}]], "Input",
 CellLabel->"In[3]:=",
 CellID->891160780],

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{
   RowBox[{"{", 
    RowBox[{
    "0.3847230553889222`", ",", "0.365126974605079`", ",", 
     "0.3755248950209958`"}], "}"}], ",", 
   RowBox[{"{", 
    RowBox[{
    "0.3832183359879331`", ",", "0.3699798456087501`", ",", 
     "0.3739273499813077`"}], "}"}]}], "}"}]], "Output",
 ImageSize->{430, 15},
 ImageMargins->{{0, 0}, {0, 0}},
 ImageRegion->{{0, 1}, {0, 1}},
 CellLabel->"Out[3]=",
 CellID->132162376]
}, Open  ]],

Cell["Time serie: ", "ExampleText",
 CellID->60397930],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"Table", "[", 
  RowBox[{
   RowBox[{"ListLinePlot", "[", 
    RowBox[{
     RowBox[{"res", "[", 
      RowBox[{"[", 
       RowBox[{
        RowBox[{";;", ";;", "10"}], ",", "1", ",", "i"}], "]"}], "]"}], ",", 
     RowBox[{"PlotRange", "\[Rule]", "All"}], ",", 
     RowBox[{"AxesLabel", "\[Rule]", 
      RowBox[{"{", 
       RowBox[{"\"\<t\>\"", ",", 
        RowBox[{"\"\<x\>\"", "<>", 
         RowBox[{"ToString", "[", "i", "]"}]}]}], "}"}]}]}], "]"}], ",", 
   RowBox[{"{", 
    RowBox[{"i", ",", "m"}], "}"}]}], "]"}]], "Input",
 CellLabel->"In[4]:=",
 CellID->111189838],

Cell[BoxData[
 GraphicsBox[
  TagBox[RasterBox[CompressedData["
1:eJztXVuvJcdZHcEv4W9YitqEEEIcCA6XQ0j8gIOG5CGZFy4SFykIwbwQNEpe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    "], {{0, 131}, {576, 0}}, {0, 255},
    ColorFunction->RGBColor],
   BoxForm`ImageTag["Byte", ColorSpace -> "RGB", Interleaving -> True],
   Selectable->False],
  BaseStyle->"ImageGraphics",
  ImageSize->Automatic,
  ImageSizeRaw->{576, 131},
  PlotRange->{{0, 576}, {0, 131}}]], "Output",
 ImageSize->{580, 131},
 ImageMargins->{{0, 0}, {0, 0}},
 ImageRegion->{{0, 1}, {0, 1}},
 CellLabel->"Out[4]=",
 CellID->141115893]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[{
 RowBox[{"Clear", "[", 
  RowBox[{"data", ",", "x", ",", "his", ",", "f"}], 
  "]"}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{"data", "=", 
   RowBox[{"Transpose", "[", 
    RowBox[{"res", "[", 
     RowBox[{"[", 
      RowBox[{
       RowBox[{"n", ";;"}], ",", "1"}], "]"}], "]"}], "]"}]}], 
  ";"}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{"his", "=", 
   RowBox[{
    RowBox[{
     RowBox[{"Histogram", "[", 
      RowBox[{"#", ",", "Automatic", ",", "\"\<ProbabilityDensity\>\""}], 
      "]"}], "&"}], "/@", "data"}]}], ";"}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{"f", "=", 
   RowBox[{
    RowBox[{"logpdf", "[", 
     RowBox[{
      RowBox[{"{", 
       RowBox[{"x", ",", "x", ",", "x"}], "}"}], ",", "All"}], "]"}], "-", 
    RowBox[{"Log", "[", 
     RowBox[{"sd", " ", 
      RowBox[{"Sqrt", "[", 
       RowBox[{"2", " ", "Pi"}], "]"}]}], "]"}]}]}], 
  ";"}], "\[IndentingNewLine]", 
 RowBox[{"Table", "[", 
  RowBox[{
   RowBox[{"Show", "[", 
    RowBox[{
     RowBox[{"his", "[", 
      RowBox[{"[", "i", "]"}], "]"}], ",", 
     RowBox[{"Plot", "[", 
      RowBox[{
       RowBox[{"Exp", "[", 
        RowBox[{"f", "[", 
         RowBox[{"[", "i", "]"}], "]"}], "]"}], ",", 
       RowBox[{"{", 
        RowBox[{"x", ",", 
         RowBox[{"Min", "[", 
          RowBox[{"data", "[", 
           RowBox[{"[", "i", "]"}], "]"}], "]"}], ",", 
         RowBox[{"Max", "[", 
          RowBox[{"data", "[", 
           RowBox[{"[", "i", "]"}], "]"}], "]"}]}], "}"}], ",", 
       RowBox[{"PlotStyle", "\[Rule]", 
        RowBox[{"{", 
         RowBox[{"Red", ",", "Thick"}], "}"}]}], ",", 
       RowBox[{"PlotRange", "\[Rule]", "All"}], ",", 
       RowBox[{"AxesLabel", "\[Rule]", 
        RowBox[{"{", 
         RowBox[{
          RowBox[{"\"\<x\>\"", "<>", 
           RowBox[{"ToString", "[", "i", "]"}]}], ",", "\"\<freq\>\""}], 
         "}"}]}]}], "]"}]}], "]"}], ",", 
   RowBox[{"{", 
    RowBox[{"i", ",", "m"}], "}"}]}], "]"}]}], "Input",
 CellLabel->"In[5]:=",
 CellID->45324905],

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{
   GraphicsBox[{
     {RGBColor[0.798413061722744, 0.824719615472648, 0.968322270542458], 
      EdgeForm[Opacity[0.7]], {}, 
      {RGBColor[0.798413061722744, 0.824719615472648, 0.968322270542458], 
       EdgeForm[Opacity[0.7]], 
       TagBox[
        TooltipBox[
         TagBox[
          DynamicBox[{
            FEPrivate`If[
             CurrentValue["MouseOver"], 
             EdgeForm[{
               GrayLevel[0.5], 
               AbsoluteThickness[1.5], 
               Opacity[0.66]}], {}, {}], 
            
            RectangleBox[{23.5, 0.}, {24., 0.0007998400319936012}, 
             "RoundingRadius" -> 0]},
           
           ImageSizeCache->{{20.5, 25.00888888888889}, {37.526259136342176`, 
            39.00684910726837}}],
          StatusArea[#, 0.0007998400319936012]& ,
          TagBoxNote->"0.0007998400319936012"],
         StyleBox["0.0007998400319936012`", {
           GrayLevel[0]}, StripOnInput -> False]],
        Annotation[#, 
         Style[0.0007998400319936012, {
           GrayLevel[0]}], "Tooltip"]& ], {}, 
       TagBox[
        TooltipBox[
         TagBox[
          DynamicBox[{
            FEPrivate`If[
             CurrentValue["MouseOver"], 
             EdgeForm[{
               GrayLevel[0.5], 
               AbsoluteThickness[1.5], 
               Opacity[0.66]}], {}, {}], 
            
            RectangleBox[{24.5, 0.}, {25., 0.0023995200959808036`}, 
             "RoundingRadius" -> 0]},
           
           ImageSizeCache->{{27.51777777777778, 32.02666666666667}, {
            36.56507919448979, 39.00684910726837}}],
          StatusArea[#, 0.0023995200959808036`]& ,
          TagBoxNote->"0.0023995200959808036"],
         StyleBox["0.0023995200959808036`", {
           GrayLevel[0]}, StripOnInput -> False]],
        Annotation[#, 
         Style[0.0023995200959808036`, {
           GrayLevel[0]}], "Tooltip"]& ], 
       TagBox[
        TooltipBox[
         TagBox[
          DynamicBox[{
            FEPrivate`If[
             CurrentValue["MouseOver"], 
             EdgeForm[{
               GrayLevel[0.5], 
               AbsoluteThickness[1.5], 
               Opacity[0.66]}], {}, {}], 
            
            RectangleBox[{25., 0.}, {25.5, 0.001999600079984003}, 
             "RoundingRadius" -> 0]},
           
           ImageSizeCache->{{31.02666666666667, 35.53555555555556}, {
            36.805374179952885`, 39.00684910726837}}],
          StatusArea[#, 0.001999600079984003]& ,
          TagBoxNote->"0.001999600079984003"],
         StyleBox["0.001999600079984003`", {
           GrayLevel[0]}, StripOnInput -> False]],
        Annotation[#, 
         Style[0.001999600079984003, {
           GrayLevel[0]}], "Tooltip"]& ], 
       TagBox[
        TooltipBox[
         TagBox[
          DynamicBox[{
            FEPrivate`If[
             CurrentValue["MouseOver"], 
             EdgeForm[{
               GrayLevel[0.5], 
               AbsoluteThickness[1.5], 
               Opacity[0.66]}], {}, {}], 
            
            RectangleBox[{25.5, 0.}, {26., 0.003599280143971206}, 
             "RoundingRadius" -> 0]},
           
           ImageSizeCache->{{34.53555555555556, 39.04444444444445}, {
            35.84419423810049, 39.00684910726837}}],
          StatusArea[#, 0.003599280143971206]& ,
          TagBoxNote->"0.003599280143971206"],
         StyleBox["0.003599280143971206`", {
           GrayLevel[0]}, StripOnInput -> False]],
        Annotation[#, 
         Style[0.003599280143971206, {
           GrayLevel[0]}], "Tooltip"]& ], 
       TagBox[
        TooltipBox[
         TagBox[
          DynamicBox[{
            FEPrivate`If[
             CurrentValue["MouseOver"], 
             EdgeForm[{
               GrayLevel[0.5], 
               AbsoluteThickness[1.5], 
               Opacity[0.66]}], {}, {}], 
            RectangleBox[{26., 0.}, {26.5, 0.001999600079984003}, 
             "RoundingRadius" -> 0]},
           
           ImageSizeCache->{{38.04444444444445, 42.55333333333334}, {
            36.805374179952885`, 39.00684910726837}}],
          StatusArea[#, 0.001999600079984003]& ,
          TagBoxNote->"0.001999600079984003"],
         StyleBox["0.001999600079984003`", {
           GrayLevel[0]}, StripOnInput -> False]],
        Annotation[#, 
         Style[0.001999600079984003, {
           GrayLevel[0]}], "Tooltip"]& ], 
       TagBox[
        TooltipBox[
         TagBox[
          DynamicBox[{
            FEPrivate`If[
             CurrentValue["MouseOver"], 
             EdgeForm[{
               GrayLevel[0.5], 
               AbsoluteThickness[1.5], 
               Opacity[0.66]}], {}, {}], 
            
            RectangleBox[{26.5, 0.}, {27., 0.0027994401119776045`}, 
             "RoundingRadius" -> 0]},
           
           ImageSizeCache->{{41.55333333333334, 46.0622222222222}, {
            36.32478420902669, 39.00684910726837}}],
          StatusArea[#, 0.0027994401119776045`]& ,
          TagBoxNote->"0.0027994401119776045"],
         StyleBox["0.0027994401119776045`", {
           GrayLevel[0]}, StripOnInput -> False]],
        Annotation[#, 
         Style[0.0027994401119776045`, {
           GrayLevel[0]}], "Tooltip"]& ], 
       TagBox[
        TooltipBox[
         TagBox[
          DynamicBox[{
            FEPrivate`If[
             CurrentValue["MouseOver"], 
             EdgeForm[{
               GrayLevel[0.5], 
               AbsoluteThickness[1.5], 
               Opacity[0.66]}], {}, {}], 
            
            RectangleBox[{27., 0.}, {27.5, 0.009998000399920015}, 
             "RoundingRadius" -> 0]},
           
           ImageSizeCache->{{45.0622222222222, 49.571111111111094`}, {
            31.999474470690934`, 39.00684910726837}}],
          StatusArea[#, 0.009998000399920015]& ,
          TagBoxNote->"0.009998000399920015"],
         StyleBox["0.009998000399920015`", {
           GrayLevel[0]}, StripOnInput -> False]],
        Annotation[#, 
         Style[0.009998000399920015, {
           GrayLevel[0]}], "Tooltip"]& ], 
       TagBox[
        TooltipBox[
         TagBox[
          DynamicBox[{
            FEPrivate`If[
             CurrentValue["MouseOver"], 
             EdgeForm[{
               GrayLevel[0.5], 
               AbsoluteThickness[1.5], 
               Opacity[0.66]}], {}, {}], 
            
            RectangleBox[{27.5, 0.}, {28., 0.007998400319936013}, 
             "RoundingRadius" -> 0]},
           
           ImageSizeCache->{{48.571111111111094`, 53.079999999999984`}, {
            33.20094939800642, 39.00684910726837}}],
          StatusArea[#, 0.007998400319936013]& ,
          TagBoxNote->"0.007998400319936013"],
         StyleBox["0.007998400319936013`", {
           GrayLevel[0]}, StripOnInput -> False]],
        Annotation[#, 
         Style[0.007998400319936013, {
           GrayLevel[0]}], "Tooltip"]& ], 
       TagBox[
        TooltipBox[
         TagBox[
          DynamicBox[{
            FEPrivate`If[
             CurrentValue["MouseOver"], 
             EdgeForm[{
               GrayLevel[0.5], 
               AbsoluteThickness[1.5], 
               Opacity[0.66]}], {}, {}], 
            
            RectangleBox[{28., 0.}, {28.5, 0.021595680863827234`}, 
             "RoundingRadius" -> 0]},
           
           ImageSizeCache->{{52.079999999999984`, 56.588888888888874`}, {
            25.030919892261103`, 39.00684910726837}}],
          StatusArea[#, 0.021595680863827234`]& ,
          TagBoxNote->"0.021595680863827234"],
         StyleBox["0.021595680863827234`", {
           GrayLevel[0]}, StripOnInput -> False]],
        Annotation[#, 
         Style[0.021595680863827234`, {
           GrayLevel[0]}], "Tooltip"]& ], 
       TagBox[
        TooltipBox[
         TagBox[
          DynamicBox[{
            FEPrivate`If[
             CurrentValue["MouseOver"], 
             EdgeForm[{
               GrayLevel[0.5], 
               AbsoluteThickness[1.5], 
               Opacity[0.66]}], {}, {}], 
            
            RectangleBox[{28.5, 0.}, {29., 0.027994401119776045`}, 
             "RoundingRadius" -> 0]},
           
           ImageSizeCache->{{55.588888888888874`, 60.097777777777765`}, {
            21.18620012485154, 39.00684910726837}}],
          StatusArea[#, 0.027994401119776045`]& ,
          TagBoxNote->"0.027994401119776045"],
         StyleBox["0.027994401119776045`", {
           GrayLevel[0]}, StripOnInput -> False]],
        Annotation[#, 
         Style[0.027994401119776045`, {
           GrayLevel[0]}], "Tooltip"]& ], 
       TagBox[
        TooltipBox[
         TagBox[
          DynamicBox[{
            FEPrivate`If[
             CurrentValue["MouseOver"], 
             EdgeForm[{
               GrayLevel[0.5], 
               AbsoluteThickness[1.5], 
               Opacity[0.66]}], {}, {}], 
            
            RectangleBox[{29., 0.}, {29.5, 0.02959408118376325}, 
             "RoundingRadius" -> 0]},
           
           ImageSizeCache->{{59.097777777777765`, 63.606666666666655`}, {
            20.22502018299915, 39.00684910726837}}],
          StatusArea[#, 0.02959408118376325]& ,
          TagBoxNote->"0.02959408118376325"],
         StyleBox["0.02959408118376325`", {
           GrayLevel[0]}, StripOnInput -> False]],
        Annotation[#, 
         Style[0.02959408118376325, {
           GrayLevel[0]}], "Tooltip"]& ], 
       TagBox[
        TooltipBox[
         TagBox[
          DynamicBox[{
            FEPrivate`If[
             CurrentValue["MouseOver"], 
             EdgeForm[{
               GrayLevel[0.5], 
               AbsoluteThickness[1.5], 
               Opacity[0.66]}], {}, {}], 
            
            RectangleBox[{29.5, 0.}, {30., 0.039192161567686463`}, 
             "RoundingRadius" -> 0]},
           
           ImageSizeCache->{{62.606666666666655`, 67.11555555555555}, {
            14.457940531884805`, 39.00684910726837}}],
          StatusArea[#, 0.039192161567686463`]& ,
          TagBoxNote->"0.039192161567686463"],
         StyleBox["0.039192161567686463`", {
           GrayLevel[0]}, StripOnInput -> False]],
        Annotation[#, 
         Style[0.039192161567686463`, {
           GrayLevel[0]}], "Tooltip"]& ], 
       TagBox[
        TooltipBox[
         TagBox[
          DynamicBox[{
            FEPrivate`If[
             CurrentValue["MouseOver"], 
             EdgeForm[{
               GrayLevel[0.5], 
               AbsoluteThickness[1.5], 
               Opacity[0.66]}], {}, {}], 
            
            RectangleBox[{30., 0.}, {30.5, 0.03839232153569286}, 
             "RoundingRadius" -> 0]},
           
           ImageSizeCache->{{66.11555555555555, 70.62444444444444}, {
            14.938530502811005`, 39.00684910726837}}],
          StatusArea[#, 0.03839232153569286]& ,
          TagBoxNote->"0.03839232153569286"],
         StyleBox["0.03839232153569286`", {
           GrayLevel[0]}, StripOnInput -> False]],
        Annotation[#, 
         Style[0.03839232153569286, {
           GrayLevel[0]}], "Tooltip"]& ], 
       TagBox[
        TooltipBox[
         TagBox[
          DynamicBox[{
            FEPrivate`If[
             CurrentValue["MouseOver"], 
             EdgeForm[{
               GrayLevel[0.5], 
               AbsoluteThickness[1.5], 
               Opacity[0.66]}], {}, {}], 
            
            RectangleBox[{30.5, 0.}, {31., 0.047190561887622474`}, 
             "RoundingRadius" -> 0]},
           
           ImageSizeCache->{{69.62444444444444, 74.13333333333333}, {
            9.652040822622855, 39.00684910726837}}],
          StatusArea[#, 0.047190561887622474`]& ,
          TagBoxNote->"0.047190561887622474"],
         StyleBox["0.047190561887622474`", {
           GrayLevel[0]}, StripOnInput -> False]],
        Annotation[#, 
         Style[0.047190561887622474`, {
           GrayLevel[0]}], "Tooltip"]& ], 
       TagBox[
        TooltipBox[
         TagBox[
          DynamicBox[{
            FEPrivate`If[
             CurrentValue["MouseOver"], 
             EdgeForm[{
               GrayLevel[0.5], 
               AbsoluteThickness[1.5], 
               Opacity[0.66]}], {}, {}], 
            
            RectangleBox[{31., 0.}, {31.5, 0.0595880823835233}, 
             "RoundingRadius" -> 0]},
           
           ImageSizeCache->{{73.13333333333333, 77.64222222222222}, {
            2.2028962732668305`, 39.00684910726837}}],
          StatusArea[#, 0.0595880823835233]& ,
          TagBoxNote->"0.0595880823835233"],
         StyleBox["0.0595880823835233`", {
           GrayLevel[0]}, StripOnInput -> False]],
        Annotation[#, 
         Style[0.0595880823835233, {
           GrayLevel[0]}], "Tooltip"]& ], 
       TagBox[
        TooltipBox[
         TagBox[
          DynamicBox[{
            FEPrivate`If[
             CurrentValue["MouseOver"], 
             EdgeForm[{
               GrayLevel[0.5], 
               AbsoluteThickness[1.5], 
               Opacity[0.66]}], {}, {}], 
            
            RectangleBox[{31.5, 0.}, {32., 0.07958408318336332}, 
             "RoundingRadius" -> 0]},
           
           ImageSizeCache->{{76.64222222222222, 
            81.1511111111111}, {-9.811852999888053, 39.00684910726837}}],
          StatusArea[#, 0.07958408318336332]& ,
          TagBoxNote->"0.07958408318336332"],
         StyleBox["0.07958408318336332`", {
           GrayLevel[0]}, StripOnInput -> False]],
        Annotation[#, 
         Style[0.07958408318336332, {
           GrayLevel[0]}], "Tooltip"]& ], 
       TagBox[
        TooltipBox[
         TagBox[
          DynamicBox[{
            FEPrivate`If[
             CurrentValue["MouseOver"], 
             EdgeForm[{
               GrayLevel[0.5], 
               AbsoluteThickness[1.5], 
               Opacity[0.66]}], {}, {}], 
            
            RectangleBox[{32., 0.}, {32.5, 0.06638672265546891}, 
             "RoundingRadius" -> 0]},
           
           ImageSizeCache->{{80.1511111111111, 84.66}, {-1.8821184796058361`, 
            39.00684910726837}}],
          StatusArea[#, 0.06638672265546891]& ,
          TagBoxNote->"0.06638672265546891"],
         StyleBox["0.06638672265546891`", {
           GrayLevel[0]}, StripOnInput -> False]],
        Annotation[#, 
         Style[0.06638672265546891, {
           GrayLevel[0]}], "Tooltip"]& ], 
       TagBox[
        TooltipBox[
         TagBox[
          DynamicBox[{
            FEPrivate`If[
             CurrentValue["MouseOver"], 
             EdgeForm[{
               GrayLevel[0.5], 
               AbsoluteThickness[1.5], 
               Opacity[0.66]}], {}, {}], 
            
            RectangleBox[{32.5, 0.}, {33., 0.09238152369526095}, 
             "RoundingRadius" -> 0]},
           
           ImageSizeCache->{{83.66, 88.16888888888889}, {-17.501292534707176`,
             39.00684910726837}}],
          StatusArea[#, 0.09238152369526095]& ,
          TagBoxNote->"0.09238152369526095"],
         StyleBox["0.09238152369526095`", {
           GrayLevel[0]}, StripOnInput -> False]],
        Annotation[#, 
         Style[0.09238152369526095, {
           GrayLevel[0]}], "Tooltip"]& ], 
       TagBox[
        TooltipBox[
         TagBox[
          DynamicBox[{
            FEPrivate`If[
             CurrentValue["MouseOver"], 
             EdgeForm[{
               GrayLevel[0.5], 
               AbsoluteThickness[1.5], 
               Opacity[0.66]}], {}, {}], 
            
            RectangleBox[{33., 0.}, {33.5, 0.10397920415916817`}, 
             "RoundingRadius" -> 0]},
           
           ImageSizeCache->{{87.16888888888889, 
            91.67777777777778}, {-24.469847113137007`, 39.00684910726837}}],
          StatusArea[#, 0.10397920415916817`]& ,
          TagBoxNote->"0.10397920415916817"],
         StyleBox["0.10397920415916817`", {
           GrayLevel[0]}, StripOnInput -> False]],
        Annotation[#, 
         Style[0.10397920415916817`, {
           GrayLevel[0]}], "Tooltip"]& ], 
       TagBox[
        TooltipBox[
         TagBox[
          DynamicBox[{
            FEPrivate`If[
             CurrentValue["MouseOver"], 
             EdgeForm[{
               GrayLevel[0.5], 
               AbsoluteThickness[1.5], 
               Opacity[0.66]}], {}, {}], 
            
            RectangleBox[{33.5, 0.}, {34., 0.1299740051989602}, 
             "RoundingRadius" -> 0]},
           
           ImageSizeCache->{{90.67777777777778, 
            95.18666666666667}, {-40.089021168238354`, 39.00684910726837}}],
          StatusArea[#, 0.1299740051989602]& ,
          TagBoxNote->"0.1299740051989602"],
         StyleBox["0.1299740051989602`", {
           GrayLevel[0]}, StripOnInput -> False]],
        Annotation[#, 
         Style[0.1299740051989602, {
           GrayLevel[0]}], "Tooltip"]& ], 
       TagBox[
        TooltipBox[
         TagBox[
          DynamicBox[{
            FEPrivate`If[
             CurrentValue["MouseOver"], 
             EdgeForm[{
               GrayLevel[0.5], 
               AbsoluteThickness[1.5], 
               Opacity[0.66]}], {}, {}], 
            
            RectangleBox[{34., 0.}, {34.5, 0.09638072385522896}, 
             "RoundingRadius" -> 0]},
           
           ImageSizeCache->{{94.18666666666667, 
            98.69555555555556}, {-19.90424238933815, 39.00684910726837}}],
          StatusArea[#, 0.09638072385522896]& ,
          TagBoxNote->"0.09638072385522896"],
         StyleBox["0.09638072385522896`", {
           GrayLevel[0]}, StripOnInput -> False]],
        Annotation[#, 
         Style[0.09638072385522896, {
           GrayLevel[0]}], "Tooltip"]& ], 
       TagBox[
        TooltipBox[
         TagBox[
          DynamicBox[{
            FEPrivate`If[
             CurrentValue["MouseOver"], 
             EdgeForm[{
               GrayLevel[0.5], 
               AbsoluteThickness[1.5], 
               Opacity[0.66]}], {}, {}], 
            
            RectangleBox[{34.5, 0.}, {35., 0.1247750449910018}, 
             "RoundingRadius" -> 0]},
           
           ImageSizeCache->{{97.69555555555556, 
            102.20444444444445`}, {-36.96518635721808, 39.00684910726837}}],
          StatusArea[#, 0.1247750449910018]& ,
          TagBoxNote->"0.1247750449910018"],
         StyleBox["0.1247750449910018`", {
           GrayLevel[0]}, StripOnInput -> False]],
        Annotation[#, 
         Style[0.1247750449910018, {
           GrayLevel[0]}], "Tooltip"]& ], 
       TagBox[
        TooltipBox[
         TagBox[
          DynamicBox[{
            FEPrivate`If[
             CurrentValue["MouseOver"], 
             EdgeForm[{
               GrayLevel[0.5], 
               AbsoluteThickness[1.5], 
               Opacity[0.66]}], {}, {}], 
            
            RectangleBox[{35., 0.}, {35.5, 0.14557088582283542`}, 
             "RoundingRadius" -> 0]},
           
           ImageSizeCache->{{101.20444444444445`, 
            105.71333333333334`}, {-49.460525601299146`, 39.00684910726837}}],
          StatusArea[#, 0.14557088582283542`]& ,
          TagBoxNote->"0.14557088582283542"],
         StyleBox["0.14557088582283542`", {
           GrayLevel[0]}, StripOnInput -> False]],
        Annotation[#, 
         Style[0.14557088582283542`, {
           GrayLevel[0]}], "Tooltip"]& ], 
       TagBox[
        TooltipBox[
         TagBox[
          DynamicBox[{
            FEPrivate`If[
             CurrentValue["MouseOver"], 
             EdgeForm[{
               GrayLevel[0.5], 
               AbsoluteThickness[1.5], 
               Opacity[0.66]}], {}, {}], 
            
            RectangleBox[{35.5, 0.}, {36., 0.11357728454309138`}, 
             "RoundingRadius" -> 0]},
           
           ImageSizeCache->{{104.71333333333334`, 
            109.22222222222223`}, {-30.236926764251344`, 39.00684910726837}}],
          StatusArea[#, 0.11357728454309138`]& ,
          TagBoxNote->"0.11357728454309138"],
         StyleBox["0.11357728454309138`", {
           GrayLevel[0]}, StripOnInput -> False]],
        Annotation[#, 
         Style[0.11357728454309138`, {
           GrayLevel[0]}], "Tooltip"]& ], 
       TagBox[
        TooltipBox[
         TagBox[
          DynamicBox[{
            FEPrivate`If[
             CurrentValue["MouseOver"], 
             EdgeForm[{
               GrayLevel[0.5], 
               AbsoluteThickness[1.5], 
               Opacity[0.66]}], {}, {}], 
            
            RectangleBox[{36., 0.}, {36.5, 0.13917216556688664`}, 
             "RoundingRadius" -> 0]},
           
           ImageSizeCache->{{108.22222222222223`, 
            112.73111111111112`}, {-45.6158058338896, 39.00684910726837}}],
          StatusArea[#, 0.13917216556688664`]& ,
          TagBoxNote->"0.13917216556688664"],
         StyleBox["0.13917216556688664`", {
           GrayLevel[0]}, StripOnInput -> False]],
        Annotation[#, 
         Style[0.13917216556688664`, {
           GrayLevel[0]}], "Tooltip"]& ], 
       TagBox[
        TooltipBox[
         TagBox[
          DynamicBox[{
            FEPrivate`If[
             CurrentValue["MouseOver"], 
             EdgeForm[{
               GrayLevel[0.5], 
               AbsoluteThickness[1.5], 
               Opacity[0.66]}], {}, {}], 
            
            RectangleBox[{36.5, 0.}, {37., 0.11597680463907219`}, 
             "RoundingRadius" -> 0]},
           
           ImageSizeCache->{{111.73111111111112`, 
            116.23999999999998`}, {-31.67869667702994, 39.00684910726837}}],
          StatusArea[#, 0.11597680463907219`]& ,
          TagBoxNote->"0.11597680463907219"],
         StyleBox["0.11597680463907219`", {
           GrayLevel[0]}, StripOnInput -> False]],
        Annotation[#, 
         Style[0.11597680463907219`, {
           GrayLevel[0]}], "Tooltip"]& ], 
       TagBox[
        TooltipBox[
         TagBox[
          DynamicBox[{
            FEPrivate`If[
             CurrentValue["MouseOver"], 
             EdgeForm[{
               GrayLevel[0.5], 
               AbsoluteThickness[1.5], 
               Opacity[0.66]}], {}, {}], 
            
            RectangleBox[{37., 0.}, {37.5, 0.09078184363127374}, 
             "RoundingRadius" -> 0]},
           
           ImageSizeCache->{{115.23999999999998`, 
            119.7488888888889}, {-16.540112592854783`, 39.00684910726837}}],
          StatusArea[#, 0.09078184363127374]& ,
          TagBoxNote->"0.09078184363127374"],
         StyleBox["0.09078184363127374`", {
           GrayLevel[0]}, StripOnInput -> False]],
        Annotation[#, 
         Style[0.09078184363127374, {
           GrayLevel[0]}], "Tooltip"]& ], 
       TagBox[
        TooltipBox[
         TagBox[
          DynamicBox[{
            FEPrivate`If[
             CurrentValue["MouseOver"], 
             EdgeForm[{
               GrayLevel[0.5], 
               AbsoluteThickness[1.5], 
               Opacity[0.66]}], {}, {}], 
            
            RectangleBox[{37.5, 0.}, {38., 0.08478304339132174}, 
             "RoundingRadius" -> 0]},
           
           ImageSizeCache->{{118.7488888888889, 
            123.25777777777776`}, {-12.935687810908327`, 39.00684910726837}}],
          StatusArea[#, 0.08478304339132174]& ,
          TagBoxNote->"0.08478304339132174"],
         StyleBox["0.08478304339132174`", {
           GrayLevel[0]}, StripOnInput -> False]],
        Annotation[#, 
         Style[0.08478304339132174, {
           GrayLevel[0]}], "Tooltip"]& ], 
       TagBox[
        TooltipBox[
         TagBox[
          DynamicBox[{
            FEPrivate`If[
             CurrentValue["MouseOver"], 
             EdgeForm[{
               GrayLevel[0.5], 
               AbsoluteThickness[1.5], 
               Opacity[0.66]}], {}, {}], 
            
            RectangleBox[{38., 0.}, {38.5, 0.0655868826234753}, 
             "RoundingRadius" -> 0]},
           
           ImageSizeCache->{{122.25777777777776`, 
            126.76666666666668`}, {-1.4015285086796325`, 39.00684910726837}}],
          StatusArea[#, 0.0655868826234753]& ,
          TagBoxNote->"0.0655868826234753"],
         StyleBox["0.0655868826234753`", {
           GrayLevel[0]}, StripOnInput -> False]],
        Annotation[#, 
         Style[0.0655868826234753, {
           GrayLevel[0]}], "Tooltip"]& ], 
       TagBox[
        TooltipBox[
         TagBox[
          DynamicBox[{
            FEPrivate`If[
             CurrentValue["MouseOver"], 
             EdgeForm[{
               GrayLevel[0.5], 
               AbsoluteThickness[1.5], 
               Opacity[0.66]}], {}, {}], 
            
            RectangleBox[{38.5, 0.}, {39., 0.060387922415516895`}, 
             "RoundingRadius" -> 0]},
           
           ImageSizeCache->{{125.76666666666668`, 130.27555555555554`}, {
            1.722306302340634, 39.00684910726837}}],
          StatusArea[#, 0.060387922415516895`]& ,
          TagBoxNote->"0.060387922415516895"],
         StyleBox["0.060387922415516895`", {
           GrayLevel[0]}, StripOnInput -> False]],
        Annotation[#, 
         Style[0.060387922415516895`, {
           GrayLevel[0]}], "Tooltip"]& ], 
       TagBox[
        TooltipBox[
         TagBox[
          DynamicBox[{
            FEPrivate`If[
             CurrentValue["MouseOver"], 
             EdgeForm[{
               GrayLevel[0.5], 
               AbsoluteThickness[1.5], 
               Opacity[0.66]}], {}, {}], 
            
            RectangleBox[{39., 0.}, {39.5, 0.03559288142371526}, 
             "RoundingRadius" -> 0]},
           
           ImageSizeCache->{{129.27555555555554`, 133.78444444444446`}, {
            16.620595401052686`, 39.00684910726837}}],
          StatusArea[#, 0.03559288142371526]& ,
          TagBoxNote->"0.03559288142371526"],
         StyleBox["0.03559288142371526`", {
           GrayLevel[0]}, StripOnInput -> False]],
        Annotation[#, 
         Style[0.03559288142371526, {
           GrayLevel[0]}], "Tooltip"]& ], 
       TagBox[
        TooltipBox[
         TagBox[
          DynamicBox[{
            FEPrivate`If[
             CurrentValue["MouseOver"], 
             EdgeForm[{
               GrayLevel[0.5], 
               AbsoluteThickness[1.5], 
               Opacity[0.66]}], {}, {}], 
            
            RectangleBox[{39.5, 0.}, {40., 0.03719256148770246}, 
             "RoundingRadius" -> 0]},
           
           ImageSizeCache->{{132.78444444444446`, 137.29333333333332`}, {
            15.659415459200293`, 39.00684910726837}}],
          StatusArea[#, 0.03719256148770246]& ,
          TagBoxNote->"0.03719256148770246"],
         StyleBox["0.03719256148770246`", {
           GrayLevel[0]}, StripOnInput -> False]],
        Annotation[#, 
         Style[0.03719256148770246, {
           GrayLevel[0]}], "Tooltip"]& ], 
       TagBox[
        TooltipBox[
         TagBox[
          DynamicBox[{
            FEPrivate`If[
             CurrentValue["MouseOver"], 
             EdgeForm[{
               GrayLevel[0.5], 
               AbsoluteThickness[1.5], 
               Opacity[0.66]}], {}, {}], 
            
            RectangleBox[{40., 0.}, {40.5, 0.03159368126374725}, 
             "RoundingRadius" -> 0]},
           
           ImageSizeCache->{{136.29333333333332`, 140.80222222222224`}, {
            19.02354525568366, 39.00684910726837}}],
          StatusArea[#, 0.03159368126374725]& ,
          TagBoxNote->"0.03159368126374725"],
         StyleBox["0.03159368126374725`", {
           GrayLevel[0]}, StripOnInput -> False]],
        Annotation[#, 
         Style[0.03159368126374725, {
           GrayLevel[0]}], "Tooltip"]& ], 
       TagBox[
        TooltipBox[
         TagBox[
          DynamicBox[{
            FEPrivate`If[
             CurrentValue["MouseOver"], 
             EdgeForm[{
               GrayLevel[0.5], 
               AbsoluteThickness[1.5], 
               Opacity[0.66]}], {}, {}], 
            
            RectangleBox[{40.5, 0.}, {41., 0.017996400719856028`}, 
             "RoundingRadius" -> 0]},
           
           ImageSizeCache->{{139.80222222222224`, 144.3111111111111}, {
            27.19357476142898, 39.00684910726837}}],
          StatusArea[#, 0.017996400719856028`]& ,
          TagBoxNote->"0.017996400719856028"],
         StyleBox["0.017996400719856028`", {
           GrayLevel[0]}, StripOnInput -> False]],
        Annotation[#, 
         Style[0.017996400719856028`, {
           GrayLevel[0]}], "Tooltip"]& ], 
       TagBox[
        TooltipBox[
         TagBox[
          DynamicBox[{
            FEPrivate`If[
             CurrentValue["MouseOver"], 
             EdgeForm[{
               GrayLevel[0.5], 
               AbsoluteThickness[1.5], 
               Opacity[0.66]}], {}, {}], 
            
            RectangleBox[{41., 0.}, {41.5, 0.01919616076784643}, 
             "RoundingRadius" -> 0]},
           
           ImageSizeCache->{{143.3111111111111, 147.82000000000002`}, {
            26.47268980503969, 39.00684910726837}}],
          StatusArea[#, 0.01919616076784643]& ,
          TagBoxNote->"0.01919616076784643"],
         StyleBox["0.01919616076784643`", {
           GrayLevel[0]}, StripOnInput -> False]],
        Annotation[#, 
         Style[0.01919616076784643, {
           GrayLevel[0]}], "Tooltip"]& ], 
       TagBox[
        TooltipBox[
         TagBox[
          DynamicBox[{
            FEPrivate`If[
             CurrentValue["MouseOver"], 
             EdgeForm[{
               GrayLevel[0.5], 
               AbsoluteThickness[1.5], 
               Opacity[0.66]}], {}, {}], 
            
            RectangleBox[{41.5, 0.}, {42., 0.01919616076784643}, 
             "RoundingRadius" -> 0]},
           
           ImageSizeCache->{{146.82000000000002`, 151.32888888888888`}, {
            26.47268980503969, 39.00684910726837}}],
          StatusArea[#, 0.01919616076784643]& ,
          TagBoxNote->"0.01919616076784643"],
         StyleBox["0.01919616076784643`", {
           GrayLevel[0]}, StripOnInput -> False]],
        Annotation[#, 
         Style[0.01919616076784643, {
           GrayLevel[0]}], "Tooltip"]& ], 
       TagBox[
        TooltipBox[
         TagBox[
          DynamicBox[{
            FEPrivate`If[
             CurrentValue["MouseOver"], 
             EdgeForm[{
               GrayLevel[0.5], 
               AbsoluteThickness[1.5], 
               Opacity[0.66]}], {}, {}], 
            
            RectangleBox[{42., 0.}, {42.5, 0.011197760447910418`}, 
             "RoundingRadius" -> 0]},
           
           ImageSizeCache->{{150.32888888888888`, 154.8377777777778}, {
            31.27858951430164, 39.00684910726837}}],
          StatusArea[#, 0.011197760447910418`]& ,
          TagBoxNote->"0.011197760447910418"],
         StyleBox["0.011197760447910418`", {
           GrayLevel[0]}, StripOnInput -> False]],
        Annotation[#, 
         Style[0.011197760447910418`, {
           GrayLevel[0]}], "Tooltip"]& ], 
       TagBox[
        TooltipBox[
         TagBox[
          DynamicBox[{
            FEPrivate`If[
             CurrentValue["MouseOver"], 
             EdgeForm[{
               GrayLevel[0.5], 
               AbsoluteThickness[1.5], 
               Opacity[0.66]}], {}, {}], 
            
            RectangleBox[{42.5, 0.}, {43., 0.007598480303939212}, 
             "RoundingRadius" -> 0]},
           
           ImageSizeCache->{{153.8377777777778, 158.34666666666666`}, {
            33.44124438346952, 39.00684910726837}}],
          StatusArea[#, 0.007598480303939212]& ,
          TagBoxNote->"0.007598480303939212"],
         StyleBox["0.007598480303939212`", {
           GrayLevel[0]}, StripOnInput -> False]],
        Annotation[#, 
         Style[0.007598480303939212, {
           GrayLevel[0]}], "Tooltip"]& ], 
       TagBox[
        TooltipBox[
         TagBox[
          DynamicBox[{
            FEPrivate`If[
             CurrentValue["MouseOver"], 
             EdgeForm[{
               GrayLevel[0.5], 
               AbsoluteThickness[1.5], 
               Opacity[0.66]}], {}, {}], 
            
            RectangleBox[{43., 0.}, {43.5, 0.00599880023995201}, 
             "RoundingRadius" -> 0]},
           
           ImageSizeCache->{{157.34666666666666`, 161.85555555555553`}, {
            34.40242432532191, 39.00684910726837}}],
          StatusArea[#, 0.00599880023995201]& ,
          TagBoxNote->"0.00599880023995201"],
         StyleBox["0.00599880023995201`", {
           GrayLevel[0]}, StripOnInput -> False]],
        Annotation[#, 
         Style[0.00599880023995201, {
           GrayLevel[0]}], "Tooltip"]& ], 
       TagBox[
        TooltipBox[
         TagBox[
          DynamicBox[{
            FEPrivate`If[
             CurrentValue["MouseOver"], 
             EdgeForm[{
               GrayLevel[0.5], 
               AbsoluteThickness[1.5], 
               Opacity[0.66]}], {}, {}], 
            
            RectangleBox[{43.5, 0.}, {44., 0.005598880223955209}, 
             "RoundingRadius" -> 0]},
           
           ImageSizeCache->{{160.85555555555553`, 165.36444444444444`}, {
            34.642719310785004`, 39.00684910726837}}],
          StatusArea[#, 0.005598880223955209]& ,
          TagBoxNote->"0.005598880223955209"],
         StyleBox["0.005598880223955209`", {
           GrayLevel[0]}, StripOnInput -> False]],
        Annotation[#, 
         Style[0.005598880223955209, {
           GrayLevel[0]}], "Tooltip"]& ], 
       TagBox[
        TooltipBox[
         TagBox[
          DynamicBox[{
            FEPrivate`If[
             CurrentValue["MouseOver"], 
             EdgeForm[{
               GrayLevel[0.5], 
               AbsoluteThickness[1.5], 
               Opacity[0.66]}], {}, {}], 
            
            RectangleBox[{44., 0.}, {44.5, 0.003199360127974405}, 
             "RoundingRadius" -> 0]},
           
           ImageSizeCache->{{164.36444444444444`, 168.8733333333333}, {
            36.084489223563594`, 39.00684910726837}}],
          StatusArea[#, 0.003199360127974405]& ,
          TagBoxNote->"0.003199360127974405"],
         StyleBox["0.003199360127974405`", {
           GrayLevel[0]}, StripOnInput -> False]],
        Annotation[#, 
         Style[0.003199360127974405, {
           GrayLevel[0]}], "Tooltip"]& ], 
       TagBox[
        TooltipBox[
         TagBox[
          DynamicBox[{
            FEPrivate`If[
             CurrentValue["MouseOver"], 
             EdgeForm[{
               GrayLevel[0.5], 
               AbsoluteThickness[1.5], 
               Opacity[0.66]}], {}, {}], 
            
            RectangleBox[{44.5, 0.}, {45., 0.0007998400319936012}, 
             "RoundingRadius" -> 0]},
           
           ImageSizeCache->{{167.8733333333333, 172.38222222222223`}, {
            37.526259136342176`, 39.00684910726837}}],
          StatusArea[#, 0.0007998400319936012]& ,
          TagBoxNote->"0.0007998400319936012"],
         StyleBox["0.0007998400319936012`", {
           GrayLevel[0]}, StripOnInput -> False]],
        Annotation[#, 
         Style[0.0007998400319936012, {
           GrayLevel[0]}], "Tooltip"]& ], 
       TagBox[
        TooltipBox[
         TagBox[
          DynamicBox[{
            FEPrivate`If[
             CurrentValue["MouseOver"], 
             EdgeForm[{
               GrayLevel[0.5], 
               AbsoluteThickness[1.5], 
               Opacity[0.66]}], {}, {}], 
            
            RectangleBox[{45., 0.}, {45.5, 0.0003999200159968006}, 
             "RoundingRadius" -> 0]},
           
           ImageSizeCache->{{171.38222222222223`, 175.8911111111111}, {
            37.76655412180528, 39.00684910726837}}],
          StatusArea[#, 0.0003999200159968006]& ,
          TagBoxNote->"0.0003999200159968006"],
         StyleBox["0.0003999200159968006`", {
           GrayLevel[0]}, StripOnInput -> False]],
        Annotation[#, 
         Style[0.0003999200159968006, {
           GrayLevel[0]}], "Tooltip"]& ]}, {}, {}}, {{}, {}, 
      {RGBColor[1, 0, 0], Thickness[Large], LineBox[CompressedData["
1:eJwVl3k8VO8Xx+0ha6soRGVJkXXGDOdcZsaaXVHUF1kie6KQfV+yV1IKUdkK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        "]]}}},
    AspectRatio->NCache[GoldenRatio^(-1), 0.6180339887498948],
    Axes->{True, True},
    AxesOrigin->{23.5, 0},
    FrameTicks->{{Automatic, Automatic}, {Automatic, Automatic}},
    GridLines->{None, None},
    PlotRange->NCache[{{
        Rational[47, 2], 
        Rational[91, 2]}, {All, All}}, {{23.5, 45.5}, {All, All}}],
    PlotRangePadding->{{
       Scaled[0.02], 
       Scaled[0.02]}, {
       Scaled[0.02], 
       Scaled[0.1]}},
    Ticks->{Automatic, Automatic}], ",", 
   GraphicsBox[{
     {RGBColor[0.798413061722744, 0.824719615472648, 0.968322270542458], 
      EdgeForm[Opacity[0.7]], {}, 
      {RGBColor[0.798413061722744, 0.824719615472648, 0.968322270542458], 
       EdgeForm[Opacity[0.7]], 
       TagBox[
        TooltipBox[
         TagBox[
          DynamicBox[{
            FEPrivate`If[
             CurrentValue["MouseOver"], 
             EdgeForm[{
               GrayLevel[0.5], 
               AbsoluteThickness[1.5], 
               Opacity[0.66]}], {}, {}], 
            
            RectangleBox[{14., 0.}, {14.5, 0.0007998400319936012}, 
             "RoundingRadius" -> 0]},
           
           ImageSizeCache->{{20.499999999999986`, 24.932608695652164`}, {
            37.54321718568554, 39.054975020858926`}}],
          StatusArea[#, 0.0007998400319936012]& ,
          TagBoxNote->"0.0007998400319936012"],
         StyleBox["0.0007998400319936012`", {
           GrayLevel[0]}, StripOnInput -> False]],
        Annotation[#, 
         Style[0.0007998400319936012, {
           GrayLevel[0]}], "Tooltip"]& ], 
       TagBox[
        TooltipBox[
         TagBox[
          DynamicBox[{
            FEPrivate`If[
             CurrentValue["MouseOver"], 
             EdgeForm[{
               GrayLevel[0.5], 
               AbsoluteThickness[1.5], 
               Opacity[0.66]}], {}, {}], 
            
            RectangleBox[{14.5, 0.}, {15., 0.0023995200959808036`}, 
             "RoundingRadius" -> 0]},
           
           ImageSizeCache->{{23.932608695652164`, 28.36521739130434}, {
            36.519701515338774`, 39.054975020858926`}}],
          StatusArea[#, 0.0023995200959808036`]& ,
          TagBoxNote->"0.0023995200959808036"],
         StyleBox["0.0023995200959808036`", {
           GrayLevel[0]}, StripOnInput -> False]],
        Annotation[#, 
         Style[0.0023995200959808036`, {
           GrayLevel[0]}], "Tooltip"]& ], 
       TagBox[
        TooltipBox[
         TagBox[
          DynamicBox[{
            FEPrivate`If[
             CurrentValue["MouseOver"], 
             EdgeForm[{
               GrayLevel[0.5], 
               AbsoluteThickness[1.5], 
               Opacity[0.66]}], {}, {}], 
            RectangleBox[{15., 0.}, {15.5, 0.0015996800639872025`}, 
             "RoundingRadius" -> 0]},
           
           ImageSizeCache->{{27.36521739130434, 31.797826086956505`}, {
            37.03145935051216, 39.054975020858926`}}],
          StatusArea[#, 0.0015996800639872025`]& ,
          TagBoxNote->"0.0015996800639872025"],
         StyleBox["0.0015996800639872025`", {
           GrayLevel[0]}, StripOnInput -> False]],
        Annotation[#, 
         Style[0.0015996800639872025`, {
           GrayLevel[0]}], "Tooltip"]& ], 
       TagBox[
        TooltipBox[
         TagBox[
          DynamicBox[{
            FEPrivate`If[
             CurrentValue["MouseOver"], 
             EdgeForm[{
               GrayLevel[0.5], 
               AbsoluteThickness[1.5], 
               Opacity[0.66]}], {}, {}], 
            
            RectangleBox[{15.5, 0.}, {16., 0.0003999200159968006}, 
             "RoundingRadius" -> 0]},
           
           ImageSizeCache->{{30.797826086956505`, 35.23043478260868}, {
            37.799096103272234`, 39.054975020858926`}}],
          StatusArea[#, 0.0003999200159968006]& ,
          TagBoxNote->"0.0003999200159968006"],
         StyleBox["0.0003999200159968006`", {
           GrayLevel[0]}, StripOnInput -> False]],
        Annotation[#, 
         Style[0.0003999200159968006, {
           GrayLevel[0]}], "Tooltip"]& ], {}, 
       TagBox[
        TooltipBox[
         TagBox[
          DynamicBox[{
            FEPrivate`If[
             CurrentValue["MouseOver"], 
             EdgeForm[{
               GrayLevel[0.5], 
               AbsoluteThickness[1.5], 
               Opacity[0.66]}], {}, {}], 
            
            RectangleBox[{16.5, 0.}, {17., 0.0027994401119776045`}, 
             "RoundingRadius" -> 0]},
           
           ImageSizeCache->{{37.66304347826086, 42.095652173913024`}, {
            36.263822597752075`, 39.054975020858926`}}],
          StatusArea[#, 0.0027994401119776045`]& ,
          TagBoxNote->"0.0027994401119776045"],
         StyleBox["0.0027994401119776045`", {
           GrayLevel[0]}, StripOnInput -> False]],
        Annotation[#, 
         Style[0.0027994401119776045`, {
           GrayLevel[0]}], "Tooltip"]& ], 
       TagBox[
        TooltipBox[
         TagBox[
          DynamicBox[{
            FEPrivate`If[
             CurrentValue["MouseOver"], 
             EdgeForm[{
               GrayLevel[0.5], 
               AbsoluteThickness[1.5], 
               Opacity[0.66]}], {}, {}], 
            
            RectangleBox[{17., 0.}, {17.5, 0.003999200159968006}, 
             "RoundingRadius" -> 0]},
           
           ImageSizeCache->{{41.095652173913024`, 45.5282608695652}, {
            35.496185844992, 39.054975020858926`}}],
          StatusArea[#, 0.003999200159968006]& ,
          TagBoxNote->"0.003999200159968006"],
         StyleBox["0.003999200159968006`", {
           GrayLevel[0]}, StripOnInput -> False]],
        Annotation[#, 
         Style[0.003999200159968006, {
           GrayLevel[0]}], "Tooltip"]& ], 
       TagBox[
        TooltipBox[
         TagBox[
          DynamicBox[{
            FEPrivate`If[
             CurrentValue["MouseOver"], 
             EdgeForm[{
               GrayLevel[0.5], 
               AbsoluteThickness[1.5], 
               Opacity[0.66]}], {}, {}], 
            
            RectangleBox[{17.5, 0.}, {18., 0.006798640271945611}, 
             "RoundingRadius" -> 0]},
           
           ImageSizeCache->{{44.5282608695652, 48.96086956521738}, {
            33.705033421885155`, 39.054975020858926`}}],
          StatusArea[#, 0.006798640271945611]& ,
          TagBoxNote->"0.006798640271945611"],
         StyleBox["0.006798640271945611`", {
           GrayLevel[0]}, StripOnInput -> False]],
        Annotation[#, 
         Style[0.006798640271945611, {
           GrayLevel[0]}], "Tooltip"]& ], 
       TagBox[
        TooltipBox[
         TagBox[
          DynamicBox[{
            FEPrivate`If[
             CurrentValue["MouseOver"], 
             EdgeForm[{
               GrayLevel[0.5], 
               AbsoluteThickness[1.5], 
               Opacity[0.66]}], {}, {}], 
            
            RectangleBox[{18., 0.}, {18.5, 0.009998000399920015}, 
             "RoundingRadius" -> 0]},
           
           ImageSizeCache->{{47.96086956521738, 52.39347826086956}, {
            31.65800208119161, 39.054975020858926`}}],
          StatusArea[#, 0.009998000399920015]& ,
          TagBoxNote->"0.009998000399920015"],
         StyleBox["0.009998000399920015`", {
           GrayLevel[0]}, StripOnInput -> False]],
        Annotation[#, 
         Style[0.009998000399920015, {
           GrayLevel[0]}], "Tooltip"]& ], 
       TagBox[
        TooltipBox[
         TagBox[
          DynamicBox[{
            FEPrivate`If[
             CurrentValue["MouseOver"], 
             EdgeForm[{
               GrayLevel[0.5], 
               AbsoluteThickness[1.5], 
               Opacity[0.66]}], {}, {}], 
            
            RectangleBox[{18.5, 0.}, {19., 0.008398320335932814}, 
             "RoundingRadius" -> 0]},
           
           ImageSizeCache->{{51.39347826086956, 55.826086956521735`}, {
            32.68151775153838, 39.054975020858926`}}],
          StatusArea[#, 0.008398320335932814]& ,
          TagBoxNote->"0.008398320335932814"],
         StyleBox["0.008398320335932814`", {
           GrayLevel[0]}, StripOnInput -> False]],
        Annotation[#, 
         Style[0.008398320335932814, {
           GrayLevel[0]}], "Tooltip"]& ], 
       TagBox[
        TooltipBox[
         TagBox[
          DynamicBox[{
            FEPrivate`If[
             CurrentValue["MouseOver"], 
             EdgeForm[{
               GrayLevel[0.5], 
               AbsoluteThickness[1.5], 
               Opacity[0.66]}], {}, {}], 
            
            RectangleBox[{19., 0.}, {19.5, 0.022795440911817635`}, 
             "RoundingRadius" -> 0]},
           
           ImageSizeCache->{{54.826086956521735`, 59.2586956521739}, {
            23.469876718417453`, 39.054975020858926`}}],
          StatusArea[#, 0.022795440911817635`]& ,
          TagBoxNote->"0.022795440911817635"],
         StyleBox["0.022795440911817635`", {
           GrayLevel[0]}, StripOnInput -> False]],
        Annotation[#, 
         Style[0.022795440911817635`, {
           GrayLevel[0]}], "Tooltip"]& ], 
       TagBox[
        TooltipBox[
         TagBox[
          DynamicBox[{
            FEPrivate`If[
             CurrentValue["MouseOver"], 
             EdgeForm[{
               GrayLevel[0.5], 
               AbsoluteThickness[1.5], 
               Opacity[0.66]}], {}, {}], 
            
            RectangleBox[{19.5, 0.}, {20., 0.021595680863827234`}, 
             "RoundingRadius" -> 0]},
           
           ImageSizeCache->{{58.2586956521739, 62.69130434782606}, {
            24.23751347117753, 39.054975020858926`}}],
          StatusArea[#, 0.021595680863827234`]& ,
          TagBoxNote->"0.021595680863827234"],
         StyleBox["0.021595680863827234`", {
           GrayLevel[0]}, StripOnInput -> False]],
        Annotation[#, 
         Style[0.021595680863827234`, {
           GrayLevel[0]}], "Tooltip"]& ], 
       TagBox[
        TooltipBox[
         TagBox[
          DynamicBox[{
            FEPrivate`If[
             CurrentValue["MouseOver"], 
             EdgeForm[{
               GrayLevel[0.5], 
               AbsoluteThickness[1.5], 
               Opacity[0.66]}], {}, {}], 
            
            RectangleBox[{20., 0.}, {20.5, 0.030393921215756847`}, 
             "RoundingRadius" -> 0]},
           
           ImageSizeCache->{{61.69130434782606, 66.12391304347825}, {
            18.608177284270294`, 39.054975020858926`}}],
          StatusArea[#, 0.030393921215756847`]& ,
          TagBoxNote->"0.030393921215756847"],
         StyleBox["0.030393921215756847`", {
           GrayLevel[0]}, StripOnInput -> False]],
        Annotation[#, 
         Style[0.030393921215756847`, {
           GrayLevel[0]}], "Tooltip"]& ], 
       TagBox[
        TooltipBox[
         TagBox[
          DynamicBox[{
            FEPrivate`If[
             CurrentValue["MouseOver"], 
             EdgeForm[{
               GrayLevel[0.5], 
               AbsoluteThickness[1.5], 
               Opacity[0.66]}], {}, {}], 
            
            RectangleBox[{20.5, 0.}, {21., 0.04879024195160968}, 
             "RoundingRadius" -> 0]},
           
           ImageSizeCache->{{65.12391304347825, 69.55652173913042}, {
            6.837747075282429, 39.054975020858926`}}],
          StatusArea[#, 0.04879024195160968]& ,
          TagBoxNote->"0.04879024195160968"],
         StyleBox["0.04879024195160968`", {
           GrayLevel[0]}, StripOnInput -> False]],
        Annotation[#, 
         Style[0.04879024195160968, {
           GrayLevel[0]}], "Tooltip"]& ], 
       TagBox[
        TooltipBox[
         TagBox[
          DynamicBox[{
            FEPrivate`If[
             CurrentValue["MouseOver"], 
             EdgeForm[{
               GrayLevel[0.5], 
               AbsoluteThickness[1.5], 
               Opacity[0.66]}], {}, {}], 
            
            RectangleBox[{21., 0.}, {21.5, 0.05118976204759048}, 
             "RoundingRadius" -> 0]},
           
           ImageSizeCache->{{68.55652173913042, 72.98913043478258}, {
            5.302473569762277, 39.054975020858926`}}],
          StatusArea[#, 0.05118976204759048]& ,
          TagBoxNote->"0.05118976204759048"],
         StyleBox["0.05118976204759048`", {
           GrayLevel[0]}, StripOnInput -> False]],
        Annotation[#, 
         Style[0.05118976204759048, {
           GrayLevel[0]}], "Tooltip"]& ], 
       TagBox[
        TooltipBox[
         TagBox[
          DynamicBox[{
            FEPrivate`If[
             CurrentValue["MouseOver"], 
             EdgeForm[{
               GrayLevel[0.5], 
               AbsoluteThickness[1.5], 
               Opacity[0.66]}], {}, {}], 
            
            RectangleBox[{21.5, 0.}, {22., 0.05158968206358728}, 
             "RoundingRadius" -> 0]},
           
           ImageSizeCache->{{71.98913043478258, 76.42173913043477}, {
            5.046594652175585, 39.054975020858926`}}],
          StatusArea[#, 0.05158968206358728]& ,
          TagBoxNote->"0.05158968206358728"],
         StyleBox["0.05158968206358728`", {
           GrayLevel[0]}, StripOnInput -> False]],
        Annotation[#, 
         Style[0.05158968206358728, {
           GrayLevel[0]}], "Tooltip"]& ], 
       TagBox[
        TooltipBox[
         TagBox[
          DynamicBox[{
            FEPrivate`If[
             CurrentValue["MouseOver"], 
             EdgeForm[{
               GrayLevel[0.5], 
               AbsoluteThickness[1.5], 
               Opacity[0.66]}], {}, {}], 
            
            RectangleBox[{22., 0.}, {22.5, 0.06998600279944012}, 
             "RoundingRadius" -> 0]},
           
           ImageSizeCache->{{75.42173913043477, 
            79.85434782608694}, {-6.723835556812276, 39.054975020858926`}}],
          StatusArea[#, 0.06998600279944012]& ,
          TagBoxNote->"0.06998600279944012"],
         StyleBox["0.06998600279944012`", {
           GrayLevel[0]}, StripOnInput -> False]],
        Annotation[#, 
         Style[0.06998600279944012, {
           GrayLevel[0]}], "Tooltip"]& ], 
       TagBox[
        TooltipBox[
         TagBox[
          DynamicBox[{
            FEPrivate`If[
             CurrentValue["MouseOver"], 
             EdgeForm[{
               GrayLevel[0.5], 
               AbsoluteThickness[1.5], 
               Opacity[0.66]}], {}, {}], 
            
            RectangleBox[{22.5, 0.}, {23., 0.08278344331133773}, 
             "RoundingRadius" -> 0]},
           
           ImageSizeCache->{{78.85434782608694, 
            83.28695652173913}, {-14.911960919586434`, 39.054975020858926`}}],
          StatusArea[#, 0.08278344331133773]& ,
          TagBoxNote->"0.08278344331133773"],
         StyleBox["0.08278344331133773`", {
           GrayLevel[0]}, StripOnInput -> False]],
        Annotation[#, 
         Style[0.08278344331133773, {
           GrayLevel[0]}], "Tooltip"]& ], 
       TagBox[
        TooltipBox[
         TagBox[
          DynamicBox[{
            FEPrivate`If[
             CurrentValue["MouseOver"], 
             EdgeForm[{
               GrayLevel[0.5], 
               AbsoluteThickness[1.5], 
               Opacity[0.66]}], {}, {}], 
            
            RectangleBox[{23., 0.}, {23.5, 0.08998200359928014}, 
             "RoundingRadius" -> 0]},
           
           ImageSizeCache->{{82.28695652173913, 
            86.71956521739129}, {-19.517781436146898`, 39.054975020858926`}}],
          StatusArea[#, 0.08998200359928014]& ,
          TagBoxNote->"0.08998200359928014"],
         StyleBox["0.08998200359928014`", {
           GrayLevel[0]}, StripOnInput -> False]],
        Annotation[#, 
         Style[0.08998200359928014, {
           GrayLevel[0]}], "Tooltip"]& ], 
       TagBox[
        TooltipBox[
         TagBox[
          DynamicBox[{
            FEPrivate`If[
             CurrentValue["MouseOver"], 
             EdgeForm[{
               GrayLevel[0.5], 
               AbsoluteThickness[1.5], 
               Opacity[0.66]}], {}, {}], 
            
            RectangleBox[{23.5, 0.}, {24., 0.09478104379124175}, 
             "RoundingRadius" -> 0]},
           
           ImageSizeCache->{{85.71956521739129, 
            90.15217391304346}, {-22.58832844718721, 39.054975020858926`}}],
          StatusArea[#, 0.09478104379124175]& ,
          TagBoxNote->"0.09478104379124175"],
         StyleBox["0.09478104379124175`", {
           GrayLevel[0]}, StripOnInput -> False]],
        Annotation[#, 
         Style[0.09478104379124175, {
           GrayLevel[0]}], "Tooltip"]& ], 
       TagBox[
        TooltipBox[
         TagBox[
          DynamicBox[{
            FEPrivate`If[
             CurrentValue["MouseOver"], 
             EdgeForm[{
               GrayLevel[0.5], 
               AbsoluteThickness[1.5], 
               Opacity[0.66]}], {}, {}], 
            
            RectangleBox[{24., 0.}, {24.5, 0.08438312337532493}, 
             "RoundingRadius" -> 0]},
           
           ImageSizeCache->{{89.15217391304346, 
            93.58478260869565}, {-15.935476589933202`, 39.054975020858926`}}],
          StatusArea[#, 0.08438312337532493]& ,
          TagBoxNote->"0.08438312337532493"],
         StyleBox["0.08438312337532493`", {
           GrayLevel[0]}, StripOnInput -> False]],
        Annotation[#, 
         Style[0.08438312337532493, {
           GrayLevel[0]}], "Tooltip"]& ], 
       TagBox[
        TooltipBox[
         TagBox[
          DynamicBox[{
            FEPrivate`If[
             CurrentValue["MouseOver"], 
             EdgeForm[{
               GrayLevel[0.5], 
               AbsoluteThickness[1.5], 
               Opacity[0.66]}], {}, {}], 
            
            RectangleBox[{24.5, 0.}, {25., 0.1191761647670466}, 
             "RoundingRadius" -> 0]},
           
           ImageSizeCache->{{92.58478260869565, 
            97.01739130434781}, {-38.196942419975464`, 39.054975020858926`}}],
          StatusArea[#, 0.1191761647670466]& ,
          TagBoxNote->"0.1191761647670466"],
         StyleBox["0.1191761647670466`", {
           GrayLevel[0]}, StripOnInput -> False]],
        Annotation[#, 
         Style[0.1191761647670466, {
           GrayLevel[0]}], "Tooltip"]& ], 
       TagBox[
        TooltipBox[
         TagBox[
          DynamicBox[{
            FEPrivate`If[
             CurrentValue["MouseOver"], 
             EdgeForm[{
               GrayLevel[0.5], 
               AbsoluteThickness[1.5], 
               Opacity[0.66]}], {}, {}], 
            
            RectangleBox[{25., 0.}, {25.5, 0.13637272545490903`}, 
             "RoundingRadius" -> 0]},
           
           ImageSizeCache->{{96.01739130434781, 
            100.44999999999997`}, {-49.19973587620324, 39.054975020858926`}}],
          StatusArea[#, 0.13637272545490903`]& ,
          TagBoxNote->"0.13637272545490903"],
         StyleBox["0.13637272545490903`", {
           GrayLevel[0]}, StripOnInput -> False]],
        Annotation[#, 
         Style[0.13637272545490903`, {
           GrayLevel[0]}], "Tooltip"]& ], 
       TagBox[
        TooltipBox[
         TagBox[
          DynamicBox[{
            FEPrivate`If[
             CurrentValue["MouseOver"], 
             EdgeForm[{
               GrayLevel[0.5], 
               AbsoluteThickness[1.5], 
               Opacity[0.66]}], {}, {}], 
            
            RectangleBox[{25.5, 0.}, {26., 0.1367726454709058}, 
             "RoundingRadius" -> 0]},
           
           ImageSizeCache->{{99.44999999999997, 
            103.88260869565217`}, {-49.45561479378993, 39.054975020858926`}}],
          StatusArea[#, 0.1367726454709058]& ,
          TagBoxNote->"0.1367726454709058"],
         StyleBox["0.1367726454709058`", {
           GrayLevel[0]}, StripOnInput -> False]],
        Annotation[#, 
         Style[0.1367726454709058, {
           GrayLevel[0]}], "Tooltip"]& ], 
       TagBox[
        TooltipBox[
         TagBox[
          DynamicBox[{
            FEPrivate`If[
             CurrentValue["MouseOver"], 
             EdgeForm[{
               GrayLevel[0.5], 
               AbsoluteThickness[1.5], 
               Opacity[0.66]}], {}, {}], 
            RectangleBox[{26., 0.}, {26.5, 0.08918216356728655}, 
             "RoundingRadius" -> 0]},
           
           ImageSizeCache->{{102.88260869565217`, 
            107.31521739130433`}, {-19.006023600973514`, 
            39.054975020858926`}}],
          StatusArea[#, 0.08918216356728655]& ,
          TagBoxNote->"0.08918216356728655"],
         StyleBox["0.08918216356728655`", {
           GrayLevel[0]}, StripOnInput -> False]],
        Annotation[#, 
         Style[0.08918216356728655, {
           GrayLevel[0]}], "Tooltip"]& ], 
       TagBox[
        TooltipBox[
         TagBox[
          DynamicBox[{
            FEPrivate`If[
             CurrentValue["MouseOver"], 
             EdgeForm[{
               GrayLevel[0.5], 
               AbsoluteThickness[1.5], 
               Opacity[0.66]}], {}, {}], 
            
            RectangleBox[{26.5, 0.}, {27., 0.1227754449110178}, 
             "RoundingRadius" -> 0]},
           
           ImageSizeCache->{{106.31521739130433`, 
            110.7478260869565}, {-40.499852678255685`, 39.054975020858926`}}],
          StatusArea[#, 0.1227754449110178]& ,
          TagBoxNote->"0.1227754449110178"],
         StyleBox["0.1227754449110178`", {
           GrayLevel[0]}, StripOnInput -> False]],
        Annotation[#, 
         Style[0.1227754449110178, {
           GrayLevel[0]}], "Tooltip"]& ], 
       TagBox[
        TooltipBox[
         TagBox[
          DynamicBox[{
            FEPrivate`If[
             CurrentValue["MouseOver"], 
             EdgeForm[{
               GrayLevel[0.5], 
               AbsoluteThickness[1.5], 
               Opacity[0.66]}], {}, {}], 
            
            RectangleBox[{27., 0.}, {27.5, 0.10277944411117776`}, 
             "RoundingRadius" -> 0]},
           
           ImageSizeCache->{{109.7478260869565, 
            114.18043478260869`}, {-27.705906798921056`, 
            39.054975020858926`}}],
          StatusArea[#, 0.10277944411117776`]& ,
          TagBoxNote->"0.10277944411117776"],
         StyleBox["0.10277944411117776`", {
           GrayLevel[0]}, StripOnInput -> False]],
        Annotation[#, 
         Style[0.10277944411117776`, {
           GrayLevel[0]}], "Tooltip"]& ], 
       TagBox[
        TooltipBox[
         TagBox[
          DynamicBox[{
            FEPrivate`If[
             CurrentValue["MouseOver"], 
             EdgeForm[{
               GrayLevel[0.5], 
               AbsoluteThickness[1.5], 
               Opacity[0.66]}], {}, {}], 
            
            RectangleBox[{27.5, 0.}, {28., 0.0627874425114977}, 
             "RoundingRadius" -> 0]},
           
           ImageSizeCache->{{113.18043478260869`, 
            117.61304347826085`}, {-2.118015040251805, 39.054975020858926`}}],
          StatusArea[#, 0.0627874425114977]& ,
          TagBoxNote->"0.0627874425114977"],
         StyleBox["0.0627874425114977`", {
           GrayLevel[0]}, StripOnInput -> False]],
        Annotation[#, 
         Style[0.0627874425114977, {
           GrayLevel[0]}], "Tooltip"]& ], 
       TagBox[
        TooltipBox[
         TagBox[
          DynamicBox[{
            FEPrivate`If[
             CurrentValue["MouseOver"], 
             EdgeForm[{
               GrayLevel[0.5], 
               AbsoluteThickness[1.5], 
               Opacity[0.66]}], {}, {}], 
            
            RectangleBox[{28., 0.}, {28.5, 0.09998000399920016}, 
             "RoundingRadius" -> 0]},
           
           ImageSizeCache->{{116.61304347826085`, 
            121.04565217391301`}, {-25.91475437581422, 39.054975020858926`}}],
          StatusArea[#, 0.09998000399920016]& ,
          TagBoxNote->"0.09998000399920016"],
         StyleBox["0.09998000399920016`", {
           GrayLevel[0]}, StripOnInput -> False]],
        Annotation[#, 
         Style[0.09998000399920016, {
           GrayLevel[0]}], "Tooltip"]& ], 
       TagBox[
        TooltipBox[
         TagBox[
          DynamicBox[{
            FEPrivate`If[
             CurrentValue["MouseOver"], 
             EdgeForm[{
               GrayLevel[0.5], 
               AbsoluteThickness[1.5], 
               Opacity[0.66]}], {}, {}], 
            
            RectangleBox[{28.5, 0.}, {29., 0.09318136372725455}, 
             "RoundingRadius" -> 0]},
           
           ImageSizeCache->{{120.04565217391301`, 
            124.4782608695652}, {-21.56481277684044, 39.054975020858926`}}],
          StatusArea[#, 0.09318136372725455]& ,
          TagBoxNote->"0.09318136372725455"],
         StyleBox["0.09318136372725455`", {
           GrayLevel[0]}, StripOnInput -> False]],
        Annotation[#, 
         Style[0.09318136372725455, {
           GrayLevel[0]}], "Tooltip"]& ], 
       TagBox[
        TooltipBox[
         TagBox[
          DynamicBox[{
            FEPrivate`If[
             CurrentValue["MouseOver"], 
             EdgeForm[{
               GrayLevel[0.5], 
               AbsoluteThickness[1.5], 
               Opacity[0.66]}], {}, {}], 
            
            RectangleBox[{29., 0.}, {29.5, 0.0619876024795041}, 
             "RoundingRadius" -> 0]},
           
           ImageSizeCache->{{123.4782608695652, 
            127.91086956521737`}, {-1.6062572050784212`, 
            39.054975020858926`}}],
          StatusArea[#, 0.0619876024795041]& ,
          TagBoxNote->"0.0619876024795041"],
         StyleBox["0.0619876024795041`", {
           GrayLevel[0]}, StripOnInput -> False]],
        Annotation[#, 
         Style[0.0619876024795041, {
           GrayLevel[0]}], "Tooltip"]& ], 
       TagBox[
        TooltipBox[
         TagBox[
          DynamicBox[{
            FEPrivate`If[
             CurrentValue["MouseOver"], 
             EdgeForm[{
               GrayLevel[0.5], 
               AbsoluteThickness[1.5], 
               Opacity[0.66]}], {}, {}], 
            RectangleBox[{29.5, 0.}, {30., 0.0655868826234753}, 
             "RoundingRadius" -> 0]},
           
           ImageSizeCache->{{126.91086956521737`, 
            131.34347826086957`}, {-3.9091674633586493`, 
            39.054975020858926`}}],
          StatusArea[#, 0.0655868826234753]& ,
          TagBoxNote->"0.0655868826234753"],
         StyleBox["0.0655868826234753`", {
           GrayLevel[0]}, StripOnInput -> False]],
        Annotation[#, 
         Style[0.0655868826234753, {
           GrayLevel[0]}], "Tooltip"]& ], 
       TagBox[
        TooltipBox[
         TagBox[
          DynamicBox[{
            FEPrivate`If[
             CurrentValue["MouseOver"], 
             EdgeForm[{
               GrayLevel[0.5], 
               AbsoluteThickness[1.5], 
               Opacity[0.66]}], {}, {}], 
            
            RectangleBox[{30., 0.}, {30.5, 0.055188962207558485`}, 
             "RoundingRadius" -> 0]},
           
           ImageSizeCache->{{130.34347826086957`, 134.77608695652174`}, {
            2.7436843938953572`, 39.054975020858926`}}],
          StatusArea[#, 0.055188962207558485`]& ,
          TagBoxNote->"0.055188962207558485"],
         StyleBox["0.055188962207558485`", {
           GrayLevel[0]}, StripOnInput -> False]],
        Annotation[#, 
         Style[0.055188962207558485`, {
           GrayLevel[0]}], "Tooltip"]& ], 
       TagBox[
        TooltipBox[
         TagBox[
          DynamicBox[{
            FEPrivate`If[
             CurrentValue["MouseOver"], 
             EdgeForm[{
               GrayLevel[0.5], 
               AbsoluteThickness[1.5], 
               Opacity[0.66]}], {}, {}], 
            
            RectangleBox[{30.5, 0.}, {31., 0.051989602079584085`}, 
             "RoundingRadius" -> 0]},
           
           ImageSizeCache->{{133.77608695652174`, 138.2086956521739}, {
            4.790715734588893, 39.054975020858926`}}],
          StatusArea[#, 0.051989602079584085`]& ,
          TagBoxNote->"0.051989602079584085"],
         StyleBox["0.051989602079584085`", {
           GrayLevel[0]}, StripOnInput -> False]],
        Annotation[#, 
         Style[0.051989602079584085`, {
           GrayLevel[0]}], "Tooltip"]& ], 
       TagBox[
        TooltipBox[
         TagBox[
          DynamicBox[{
            FEPrivate`If[
             CurrentValue["MouseOver"], 
             EdgeForm[{
               GrayLevel[0.5], 
               AbsoluteThickness[1.5], 
               Opacity[0.66]}], {}, {}], 
            
            RectangleBox[{31., 0.}, {31.5, 0.027994401119776045`}, 
             "RoundingRadius" -> 0]},
           
           ImageSizeCache->{{137.2086956521739, 141.64130434782606`}, {
            20.143450789790446`, 39.054975020858926`}}],
          StatusArea[#, 0.027994401119776045`]& ,
          TagBoxNote->"0.027994401119776045"],
         StyleBox["0.027994401119776045`", {
           GrayLevel[0]}, StripOnInput -> False]],
        Annotation[#, 
         Style[0.027994401119776045`, {
           GrayLevel[0]}], "Tooltip"]& ], 
       TagBox[
        TooltipBox[
         TagBox[
          DynamicBox[{
            FEPrivate`If[
             CurrentValue["MouseOver"], 
             EdgeForm[{
               GrayLevel[0.5], 
               AbsoluteThickness[1.5], 
               Opacity[0.66]}], {}, {}], 
            
            RectangleBox[{31.5, 0.}, {32., 0.02559488102379524}, 
             "RoundingRadius" -> 0]},
           
           ImageSizeCache->{{140.64130434782606`, 145.07391304347823`}, {
            21.6787242953106, 39.054975020858926`}}],
          StatusArea[#, 0.02559488102379524]& ,
          TagBoxNote->"0.02559488102379524"],
         StyleBox["0.02559488102379524`", {
           GrayLevel[0]}, StripOnInput -> False]],
        Annotation[#, 
         Style[0.02559488102379524, {
           GrayLevel[0]}], "Tooltip"]& ], 
       TagBox[
        TooltipBox[
         TagBox[
          DynamicBox[{
            FEPrivate`If[
             CurrentValue["MouseOver"], 
             EdgeForm[{
               GrayLevel[0.5], 
               AbsoluteThickness[1.5], 
               Opacity[0.66]}], {}, {}], 
            
            RectangleBox[{32., 0.}, {32.5, 0.010797840431913617`}, 
             "RoundingRadius" -> 0]},
           
           ImageSizeCache->{{144.07391304347823`, 148.5065217391304}, {
            31.146244246018227`, 39.054975020858926`}}],
          StatusArea[#, 0.010797840431913617`]& ,
          TagBoxNote->"0.010797840431913617"],
         StyleBox["0.010797840431913617`", {
           GrayLevel[0]}, StripOnInput -> False]],
        Annotation[#, 
         Style[0.010797840431913617`, {
           GrayLevel[0]}], "Tooltip"]& ], 
       TagBox[
        TooltipBox[
         TagBox[
          DynamicBox[{
            FEPrivate`If[
             CurrentValue["MouseOver"], 
             EdgeForm[{
               GrayLevel[0.5], 
               AbsoluteThickness[1.5], 
               Opacity[0.66]}], {}, {}], 
            
            RectangleBox[{32.5, 0.}, {33., 0.014397120575884824`}, 
             "RoundingRadius" -> 0]},
           
           ImageSizeCache->{{147.5065217391304, 151.9391304347826}, {
            28.843333987737992`, 39.054975020858926`}}],
          StatusArea[#, 0.014397120575884824`]& ,
          TagBoxNote->"0.014397120575884824"],
         StyleBox["0.014397120575884824`", {
           GrayLevel[0]}, StripOnInput -> False]],
        Annotation[#, 
         Style[0.014397120575884824`, {
           GrayLevel[0]}], "Tooltip"]& ], 
       TagBox[
        TooltipBox[
         TagBox[
          DynamicBox[{
            FEPrivate`If[
             CurrentValue["MouseOver"], 
             EdgeForm[{
               GrayLevel[0.5], 
               AbsoluteThickness[1.5], 
               Opacity[0.66]}], {}, {}], 
            
            RectangleBox[{33., 0.}, {33.5, 0.010397920415916816`}, 
             "RoundingRadius" -> 0]},
           
           ImageSizeCache->{{150.9391304347826, 155.37173913043478`}, {
            31.40212316360492, 39.054975020858926`}}],
          StatusArea[#, 0.010397920415916816`]& ,
          TagBoxNote->"0.010397920415916816"],
         StyleBox["0.010397920415916816`", {
           GrayLevel[0]}, StripOnInput -> False]],
        Annotation[#, 
         Style[0.010397920415916816`, {
           GrayLevel[0]}], "Tooltip"]& ], 
       TagBox[
        TooltipBox[
         TagBox[
          DynamicBox[{
            FEPrivate`If[
             CurrentValue["MouseOver"], 
             EdgeForm[{
               GrayLevel[0.5], 
               AbsoluteThickness[1.5], 
               Opacity[0.66]}], {}, {}], 
            
            RectangleBox[{33.5, 0.}, {34., 0.014397120575884824`}, 
             "RoundingRadius" -> 0]},
           
           ImageSizeCache->{{154.37173913043478`, 158.80434782608694`}, {
            28.843333987737992`, 39.054975020858926`}}],
          StatusArea[#, 0.014397120575884824`]& ,
          TagBoxNote->"0.014397120575884824"],
         StyleBox["0.014397120575884824`", {
           GrayLevel[0]}, StripOnInput -> False]],
        Annotation[#, 
         Style[0.014397120575884824`, {
           GrayLevel[0]}], "Tooltip"]& ], 
       TagBox[
        TooltipBox[
         TagBox[
          DynamicBox[{
            FEPrivate`If[
             CurrentValue["MouseOver"], 
             EdgeForm[{
               GrayLevel[0.5], 
               AbsoluteThickness[1.5], 
               Opacity[0.66]}], {}, {}], 
            
            RectangleBox[{34., 0.}, {34.5, 0.007598480303939212}, 
             "RoundingRadius" -> 0]},
           
           ImageSizeCache->{{157.80434782608694`, 162.2369565217391}, {
            33.19327558671177, 39.054975020858926`}}],
          StatusArea[#, 0.007598480303939212]& ,
          TagBoxNote->"0.007598480303939212"],
         StyleBox["0.007598480303939212`", {
           GrayLevel[0]}, StripOnInput -> False]],
        Annotation[#, 
         Style[0.007598480303939212, {
           GrayLevel[0]}], "Tooltip"]& ], 
       TagBox[
        TooltipBox[
         TagBox[
          DynamicBox[{
            FEPrivate`If[
             CurrentValue["MouseOver"], 
             EdgeForm[{
               GrayLevel[0.5], 
               AbsoluteThickness[1.5], 
               Opacity[0.66]}], {}, {}], 
            
            RectangleBox[{34.5, 0.}, {35., 0.0011997600479904018`}, 
             "RoundingRadius" -> 0]},
           
           ImageSizeCache->{{161.2369565217391, 165.66956521739127`}, {
            37.28733826809885, 39.054975020858926`}}],
          StatusArea[#, 0.0011997600479904018`]& ,
          TagBoxNote->"0.0011997600479904018"],
         StyleBox["0.0011997600479904018`", {
           GrayLevel[0]}, StripOnInput -> False]],
        Annotation[#, 
         Style[0.0011997600479904018`, {
           GrayLevel[0]}], "Tooltip"]& ], 
       TagBox[
        TooltipBox[
         TagBox[
          DynamicBox[{
            FEPrivate`If[
             CurrentValue["MouseOver"], 
             EdgeForm[{
               GrayLevel[0.5], 
               AbsoluteThickness[1.5], 
               Opacity[0.66]}], {}, {}], 
            
            RectangleBox[{35., 0.}, {35.5, 0.0015996800639872025`}, 
             "RoundingRadius" -> 0]},
           
           ImageSizeCache->{{164.66956521739127`, 169.1021739130435}, {
            37.03145935051216, 39.054975020858926`}}],
          StatusArea[#, 0.0015996800639872025`]& ,
          TagBoxNote->"0.0015996800639872025"],
         StyleBox["0.0015996800639872025`", {
           GrayLevel[0]}, StripOnInput -> False]],
        Annotation[#, 
         Style[0.0015996800639872025`, {
           GrayLevel[0]}], "Tooltip"]& ], 
       TagBox[
        TooltipBox[
         TagBox[
          DynamicBox[{
            FEPrivate`If[
             CurrentValue["MouseOver"], 
             EdgeForm[{
               GrayLevel[0.5], 
               AbsoluteThickness[1.5], 
               Opacity[0.66]}], {}, {}], 
            
            RectangleBox[{35.5, 0.}, {36., 0.001999600079984003}, 
             "RoundingRadius" -> 0]},
           
           ImageSizeCache->{{168.1021739130435, 172.53478260869565`}, {
            36.775580432925466`, 39.054975020858926`}}],
          StatusArea[#, 0.001999600079984003]& ,
          TagBoxNote->"0.001999600079984003"],
         StyleBox["0.001999600079984003`", {
           GrayLevel[0]}, StripOnInput -> False]],
        Annotation[#, 
         Style[0.001999600079984003, {
           GrayLevel[0]}], "Tooltip"]& ], 
       TagBox[
        TooltipBox[
         TagBox[
          DynamicBox[{
            FEPrivate`If[
             CurrentValue["MouseOver"], 
             EdgeForm[{
               GrayLevel[0.5], 
               AbsoluteThickness[1.5], 
               Opacity[0.66]}], {}, {}], 
            
            RectangleBox[{36., 0.}, {36.5, 0.0007998400319936012}, 
             "RoundingRadius" -> 0]},
           
           ImageSizeCache->{{171.53478260869565`, 175.9673913043478}, {
            37.54321718568554, 39.054975020858926`}}],
          StatusArea[#, 0.0007998400319936012]& ,
          TagBoxNote->"0.0007998400319936012"],
         StyleBox["0.0007998400319936012`", {
           GrayLevel[0]}, StripOnInput -> False]],
        Annotation[#, 
         Style[0.0007998400319936012, {
           GrayLevel[0]}], "Tooltip"]& ]}, {}, {}}, {{}, {}, 
      {RGBColor[1, 0, 0], Thickness[Large], LineBox[CompressedData["
1:eJwd1nc8V9//AHArUSEjJBpGIeNtxPu+3+n1eiNUkpBZyCgKWZmR4iMZhZBV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        "]]}}},
    AspectRatio->NCache[GoldenRatio^(-1), 0.6180339887498948],
    Axes->{True, True},
    AxesOrigin->{14., 0},
    FrameTicks->{{Automatic, Automatic}, {Automatic, Automatic}},
    GridLines->{None, None},
    PlotRange->NCache[{{14, 
        Rational[73, 2]}, {All, All}}, {{14, 36.5}, {All, All}}],
    PlotRangePadding->{{
       Scaled[0.02], 
       Scaled[0.02]}, {
       Scaled[0.02], 
       Scaled[0.1]}},
    Ticks->{Automatic, Automatic}], ",", 
   GraphicsBox[{
     {RGBColor[0.798413061722744, 0.824719615472648, 0.968322270542458], 
      EdgeForm[Opacity[0.7]], {}, 
      {RGBColor[0.798413061722744, 0.824719615472648, 0.968322270542458], 
       EdgeForm[Opacity[0.7]], 
       TagBox[
        TooltipBox[
         TagBox[
          DynamicBox[{
            FEPrivate`If[
             CurrentValue["MouseOver"], 
             EdgeForm[{
               GrayLevel[0.5], 
               AbsoluteThickness[1.5], 
               Opacity[0.66]}], {}, {}], 
            
            RectangleBox[{27.2, 0.}, {27.4, 0.0029994001199760154`}, 
             "RoundingRadius" -> 0]},
           
           ImageSizeCache->{{15.499999999999886`, 19.895223621363357`}, {
            38.52015717248321, 40.27773676735785}}],
          StatusArea[#, 0.0029994001199760154`]& ,
          TagBoxNote->"0.0029994001199760154"],
         StyleBox["0.0029994001199760154`", {
           GrayLevel[0]}, StripOnInput -> False]],
        Annotation[#, 
         Style[0.0029994001199760154`, {
           GrayLevel[0]}], "Tooltip"]& ], {}, 
       TagBox[
        TooltipBox[
         TagBox[
          DynamicBox[{
            FEPrivate`If[
             CurrentValue["MouseOver"], 
             EdgeForm[{
               GrayLevel[0.5], 
               AbsoluteThickness[1.5], 
               Opacity[0.66]}], {}, {}], 
            
            RectangleBox[{27.6, 0.}, {27.8, 0.0029994001199760154`}, 
             "RoundingRadius" -> 0]},
           
           ImageSizeCache->{{22.290447242726827`, 26.68567086409024}, {
            38.52015717248321, 40.27773676735785}}],
          StatusArea[#, 0.0029994001199760154`]& ,
          TagBoxNote->"0.0029994001199760154"],
         StyleBox["0.0029994001199760154`", {
           GrayLevel[0]}, StripOnInput -> False]],
        Annotation[#, 
         Style[0.0029994001199760154`, {
           GrayLevel[0]}], "Tooltip"]& ], 
       TagBox[
        TooltipBox[
         TagBox[
          DynamicBox[{
            FEPrivate`If[
             CurrentValue["MouseOver"], 
             EdgeForm[{
               GrayLevel[0.5], 
               AbsoluteThickness[1.5], 
               Opacity[0.66]}], {}, {}], 
            
            RectangleBox[{27.8, 0.}, {28., 0.000999800039992005}, 
             "RoundingRadius" -> 0]},
           
           ImageSizeCache->{{25.68567086409024, 30.080894485453655`}, {
            39.02521023573297, 40.27773676735785}}],
          StatusArea[#, 0.000999800039992005]& ,
          TagBoxNote->"0.000999800039992005"],
         StyleBox["0.000999800039992005`", {
           GrayLevel[0]}, StripOnInput -> False]],
        Annotation[#, 
         Style[0.000999800039992005, {
           GrayLevel[0]}], "Tooltip"]& ], {}, 
       TagBox[
        TooltipBox[
         TagBox[
          DynamicBox[{
            FEPrivate`If[
             CurrentValue["MouseOver"], 
             EdgeForm[{
               GrayLevel[0.5], 
               AbsoluteThickness[1.5], 
               Opacity[0.66]}], {}, {}], 
            
            RectangleBox[{28.2, 0.}, {28.4, 0.008998200359928047}, 
             "RoundingRadius" -> 0]},
           
           ImageSizeCache->{{32.47611810681707, 36.87134172818054}, {
            37.00499798273392, 40.27773676735785}}],
          StatusArea[#, 0.008998200359928047]& ,
          TagBoxNote->"0.008998200359928047"],
         StyleBox["0.008998200359928047`", {
           GrayLevel[0]}, StripOnInput -> False]],
        Annotation[#, 
         Style[0.008998200359928047, {
           GrayLevel[0]}], "Tooltip"]& ], 
       TagBox[
        TooltipBox[
         TagBox[
          DynamicBox[{
            FEPrivate`If[
             CurrentValue["MouseOver"], 
             EdgeForm[{
               GrayLevel[0.5], 
               AbsoluteThickness[1.5], 
               Opacity[0.66]}], {}, {}], 
            
            RectangleBox[{28.4, 0.}, {28.6, 0.0019996000799839746`}, 
             "RoundingRadius" -> 0]},
           ImageSizeCache->{{35.87134172818054, 40.26656534954401}, {
            38.7726837041081, 40.27773676735785}}],
          StatusArea[#, 0.0019996000799839746`]& ,
          TagBoxNote->"0.0019996000799839746"],
         StyleBox["0.0019996000799839746`", {
           GrayLevel[0]}, StripOnInput -> False]],
        Annotation[#, 
         Style[0.0019996000799839746`, {
           GrayLevel[0]}], "Tooltip"]& ], 
       TagBox[
        TooltipBox[
         TagBox[
          DynamicBox[{
            FEPrivate`If[
             CurrentValue["MouseOver"], 
             EdgeForm[{
               GrayLevel[0.5], 
               AbsoluteThickness[1.5], 
               Opacity[0.66]}], {}, {}], 
            
            RectangleBox[{28.6, 0.}, {28.8, 0.00799840031993604}, 
             "RoundingRadius" -> 0]},
           
           ImageSizeCache->{{39.26656534954401, 43.66178897090742}, {
            37.2575245143588, 40.27773676735785}}],
          StatusArea[#, 0.00799840031993604]& ,
          TagBoxNote->"0.00799840031993604"],
         StyleBox["0.00799840031993604`", {
           GrayLevel[0]}, StripOnInput -> False]],
        Annotation[#, 
         Style[0.00799840031993604, {
           GrayLevel[0]}], "Tooltip"]& ], 
       TagBox[
        TooltipBox[
         TagBox[
          DynamicBox[{
            FEPrivate`If[
             CurrentValue["MouseOver"], 
             EdgeForm[{
               GrayLevel[0.5], 
               AbsoluteThickness[1.5], 
               Opacity[0.66]}], {}, {}], 
            
            RectangleBox[{28.8, 0.}, {29., 0.00799840031993604}, 
             "RoundingRadius" -> 0]},
           
           ImageSizeCache->{{42.66178897090742, 47.05701259227084}, {
            37.2575245143588, 40.27773676735785}}],
          StatusArea[#, 0.00799840031993604]& ,
          TagBoxNote->"0.00799840031993604"],
         StyleBox["0.00799840031993604`", {
           GrayLevel[0]}, StripOnInput -> False]],
        Annotation[#, 
         Style[0.00799840031993604, {
           GrayLevel[0]}], "Tooltip"]& ], 
       TagBox[
        TooltipBox[
         TagBox[
          DynamicBox[{
            FEPrivate`If[
             CurrentValue["MouseOver"], 
             EdgeForm[{
               GrayLevel[0.5], 
               AbsoluteThickness[1.5], 
               Opacity[0.66]}], {}, {}], 
            
            RectangleBox[{29., 0.}, {29.2, 0.021995600879824113`}, 
             "RoundingRadius" -> 0]},
           
           ImageSizeCache->{{46.05701259227084, 50.45223621363431}, {
            33.722153071610464`, 40.27773676735785}}],
          StatusArea[#, 0.021995600879824113`]& ,
          TagBoxNote->"0.021995600879824113"],
         StyleBox["0.021995600879824113`", {
           GrayLevel[0]}, StripOnInput -> False]],
        Annotation[#, 
         Style[0.021995600879824113`, {
           GrayLevel[0]}], "Tooltip"]& ], 
       TagBox[
        TooltipBox[
         TagBox[
          DynamicBox[{
            FEPrivate`If[
             CurrentValue["MouseOver"], 
             EdgeForm[{
               GrayLevel[0.5], 
               AbsoluteThickness[1.5], 
               Opacity[0.66]}], {}, {}], 
            
            RectangleBox[{29.2, 0.}, {29.4, 0.017996400719856094`}, 
             "RoundingRadius" -> 0]},
           
           ImageSizeCache->{{49.45223621363431, 53.84745983499772}, {
            34.732259198109986`, 40.27773676735785}}],
          StatusArea[#, 0.017996400719856094`]& ,
          TagBoxNote->"0.017996400719856094"],
         StyleBox["0.017996400719856094`", {
           GrayLevel[0]}, StripOnInput -> False]],
        Annotation[#, 
         Style[0.017996400719856094`, {
           GrayLevel[0]}], "Tooltip"]& ], 
       TagBox[
        TooltipBox[
         TagBox[
          DynamicBox[{
            FEPrivate`If[
             CurrentValue["MouseOver"], 
             EdgeForm[{
               GrayLevel[0.5], 
               AbsoluteThickness[1.5], 
               Opacity[0.66]}], {}, {}], 
            
            RectangleBox[{29.4, 0.}, {29.6, 0.04299140171965545}, 
             "RoundingRadius" -> 0]},
           
           ImageSizeCache->{{52.84745983499772, 57.24268345636119}, {
            28.41909590748815, 40.27773676735785}}],
          StatusArea[#, 0.04299140171965545]& ,
          TagBoxNote->"0.04299140171965545"],
         StyleBox["0.04299140171965545`", {
           GrayLevel[0]}, StripOnInput -> False]],
        Annotation[#, 
         Style[0.04299140171965545, {
           GrayLevel[0]}], "Tooltip"]& ], 
       TagBox[
        TooltipBox[
         TagBox[
          DynamicBox[{
            FEPrivate`If[
             CurrentValue["MouseOver"], 
             EdgeForm[{
               GrayLevel[0.5], 
               AbsoluteThickness[1.5], 
               Opacity[0.66]}], {}, {}], 
            
            RectangleBox[{29.6, 0.}, {29.8, 0.06698660267946435}, 
             "RoundingRadius" -> 0]},
           
           ImageSizeCache->{{56.24268345636119, 60.637907077724606`}, {
            22.358459148490805`, 40.27773676735785}}],
          StatusArea[#, 0.06698660267946435]& ,
          TagBoxNote->"0.06698660267946435"],
         StyleBox["0.06698660267946435`", {
           GrayLevel[0]}, StripOnInput -> False]],
        Annotation[#, 
         Style[0.06698660267946435, {
           GrayLevel[0]}], "Tooltip"]& ], 
       TagBox[
        TooltipBox[
         TagBox[
          DynamicBox[{
            FEPrivate`If[
             CurrentValue["MouseOver"], 
             EdgeForm[{
               GrayLevel[0.5], 
               AbsoluteThickness[1.5], 
               Opacity[0.66]}], {}, {}], 
            
            RectangleBox[{29.8, 0.}, {30., 0.06798640271945636}, 
             "RoundingRadius" -> 0]},
           
           ImageSizeCache->{{59.637907077724606`, 64.03313069908808}, {
            22.10593261686592, 40.27773676735785}}],
          StatusArea[#, 0.06798640271945636]& ,
          TagBoxNote->"0.06798640271945636"],
         StyleBox["0.06798640271945636`", {
           GrayLevel[0]}, StripOnInput -> False]],
        Annotation[#, 
         Style[0.06798640271945636, {
           GrayLevel[0]}], "Tooltip"]& ], 
       TagBox[
        TooltipBox[
         TagBox[
          DynamicBox[{
            FEPrivate`If[
             CurrentValue["MouseOver"], 
             EdgeForm[{
               GrayLevel[0.5], 
               AbsoluteThickness[1.5], 
               Opacity[0.66]}], {}, {}], 
            
            RectangleBox[{30., 0.}, {30.2, 0.10897820435912856`}, 
             "RoundingRadius" -> 0]},
           
           ImageSizeCache->{{63.033130699088076`, 67.42835432045143}, {
            11.752344820245792`, 40.27773676735785}}],
          StatusArea[#, 0.10897820435912856`]& ,
          TagBoxNote->"0.10897820435912856"],
         StyleBox["0.10897820435912856`", {
           GrayLevel[0]}, StripOnInput -> False]],
        Annotation[#, 
         Style[0.10897820435912856`, {
           GrayLevel[0]}], "Tooltip"]& ], 
       TagBox[
        TooltipBox[
         TagBox[
          DynamicBox[{
            FEPrivate`If[
             CurrentValue["MouseOver"], 
             EdgeForm[{
               GrayLevel[0.5], 
               AbsoluteThickness[1.5], 
               Opacity[0.66]}], {}, {}], 
            
            RectangleBox[{30.2, 0.}, {30.4, 0.10097980403919253`}, 
             "RoundingRadius" -> 0]},
           
           ImageSizeCache->{{66.42835432045143, 70.8235779418149}, {
            13.77255707324484, 40.27773676735785}}],
          StatusArea[#, 0.10097980403919253`]& ,
          TagBoxNote->"0.10097980403919253"],
         StyleBox["0.10097980403919253`", {
           GrayLevel[0]}, StripOnInput -> False]],
        Annotation[#, 
         Style[0.10097980403919253`, {
           GrayLevel[0]}], "Tooltip"]& ], 
       TagBox[
        TooltipBox[
         TagBox[
          DynamicBox[{
            FEPrivate`If[
             CurrentValue["MouseOver"], 
             EdgeForm[{
               GrayLevel[0.5], 
               AbsoluteThickness[1.5], 
               Opacity[0.66]}], {}, {}], 
            
            RectangleBox[{30.4, 0.}, {30.6, 0.1329734053189343}, 
             "RoundingRadius" -> 0]},
           
           ImageSizeCache->{{69.8235779418149, 74.21880156317837}, {
            5.691708061249237, 40.27773676735785}}],
          StatusArea[#, 0.1329734053189343]& ,
          TagBoxNote->"0.1329734053189343"],
         StyleBox["0.1329734053189343`", {
           GrayLevel[0]}, StripOnInput -> False]],
        Annotation[#, 
         Style[0.1329734053189343, {
           GrayLevel[0]}], "Tooltip"]& ], 
       TagBox[
        TooltipBox[
         TagBox[
          DynamicBox[{
            FEPrivate`If[
             CurrentValue["MouseOver"], 
             EdgeForm[{
               GrayLevel[0.5], 
               AbsoluteThickness[1.5], 
               Opacity[0.66]}], {}, {}], 
            
            RectangleBox[{30.6, 0.}, {30.8, 0.18996200759848098`}, 
             "RoundingRadius" -> 0]},
           
           ImageSizeCache->{{73.21880156317837, 
            77.61402518454184}, {-8.702304241369589, 40.27773676735785}}],
          StatusArea[#, 0.18996200759848098`]& ,
          TagBoxNote->"0.18996200759848098"],
         StyleBox["0.18996200759848098`", {
           GrayLevel[0]}, StripOnInput -> False]],
        Annotation[#, 
         Style[0.18996200759848098`, {
           GrayLevel[0]}], "Tooltip"]& ], 
       TagBox[
        TooltipBox[
         TagBox[
          DynamicBox[{
            FEPrivate`If[
             CurrentValue["MouseOver"], 
             EdgeForm[{
               GrayLevel[0.5], 
               AbsoluteThickness[1.5], 
               Opacity[0.66]}], {}, {}], 
            
            RectangleBox[{30.8, 0.}, {31., 0.21495700859828112`}, 
             "RoundingRadius" -> 0]},
           
           ImageSizeCache->{{76.61402518454184, 
            81.0092488059052}, {-15.015467531991625`, 40.27773676735785}}],
          StatusArea[#, 0.21495700859828112`]& ,
          TagBoxNote->"0.21495700859828112"],
         StyleBox["0.21495700859828112`", {
           GrayLevel[0]}, StripOnInput -> False]],
        Annotation[#, 
         Style[0.21495700859828112`, {
           GrayLevel[0]}], "Tooltip"]& ], 
       TagBox[
        TooltipBox[
         TagBox[
          DynamicBox[{
            FEPrivate`If[
             CurrentValue["MouseOver"], 
             EdgeForm[{
               GrayLevel[0.5], 
               AbsoluteThickness[1.5], 
               Opacity[0.66]}], {}, {}], 
            
            RectangleBox[{31., 0.}, {31.2, 0.27894421115776946`}, 
             "RoundingRadius" -> 0]},
           
           ImageSizeCache->{{80.0092488059052, 
            84.40447242726867}, {-31.177165555984033`, 40.27773676735785}}],
          StatusArea[#, 0.27894421115776946`]& ,
          TagBoxNote->"0.27894421115776946"],
         StyleBox["0.27894421115776946`", {
           GrayLevel[0]}, StripOnInput -> False]],
        Annotation[#, 
         Style[0.27894421115776946`, {
           GrayLevel[0]}], "Tooltip"]& ], 
       TagBox[
        TooltipBox[
         TagBox[
          DynamicBox[{
            FEPrivate`If[
             CurrentValue["MouseOver"], 
             EdgeForm[{
               GrayLevel[0.5], 
               AbsoluteThickness[1.5], 
               Opacity[0.66]}], {}, {}], 
            
            RectangleBox[{31.2, 0.}, {31.4, 0.24695060987802525`}, 
             "RoundingRadius" -> 0]},
           
           ImageSizeCache->{{83.40447242726867, 
            87.79969604863214}, {-23.096316543987818`, 40.27773676735785}}],
          StatusArea[#, 0.24695060987802525`]& ,
          TagBoxNote->"0.24695060987802525"],
         StyleBox["0.24695060987802525`", {
           GrayLevel[0]}, StripOnInput -> False]],
        Annotation[#, 
         Style[0.24695060987802525`, {
           GrayLevel[0]}], "Tooltip"]& ], 
       TagBox[
        TooltipBox[
         TagBox[
          DynamicBox[{
            FEPrivate`If[
             CurrentValue["MouseOver"], 
             EdgeForm[{
               GrayLevel[0.5], 
               AbsoluteThickness[1.5], 
               Opacity[0.66]}], {}, {}], 
            
            RectangleBox[{31.4, 0.}, {31.6, 0.3009398120375882}, 
             "RoundingRadius" -> 0]},
           
           ImageSizeCache->{{86.79969604863214, 
            91.19491966999561}, {-36.732749251730056`, 40.27773676735785}}],
          StatusArea[#, 0.3009398120375882]& ,
          TagBoxNote->"0.3009398120375882"],
         StyleBox["0.3009398120375882`", {
           GrayLevel[0]}, StripOnInput -> False]],
        Annotation[#, 
         Style[0.3009398120375882, {
           GrayLevel[0]}], "Tooltip"]& ], 
       TagBox[
        TooltipBox[
         TagBox[
          DynamicBox[{
            FEPrivate`If[
             CurrentValue["MouseOver"], 
             EdgeForm[{
               GrayLevel[0.5], 
               AbsoluteThickness[1.5], 
               Opacity[0.66]}], {}, {}], 
            
            RectangleBox[{31.6, 0.}, {31.8, 0.33993201359728176`}, 
             "RoundingRadius" -> 0]},
           
           ImageSizeCache->{{90.19491966999561, 
            94.59014329135897}, {-46.58128398510179, 40.27773676735785}}],
          StatusArea[#, 0.33993201359728176`]& ,
          TagBoxNote->"0.33993201359728176"],
         StyleBox["0.33993201359728176`", {
           GrayLevel[0]}, StripOnInput -> False]],
        Annotation[#, 
         Style[0.33993201359728176`, {
           GrayLevel[0]}], "Tooltip"]& ], 
       TagBox[
        TooltipBox[
         TagBox[
          DynamicBox[{
            FEPrivate`If[
             CurrentValue["MouseOver"], 
             EdgeForm[{
               GrayLevel[0.5], 
               AbsoluteThickness[1.5], 
               Opacity[0.66]}], {}, {}], 
            
            RectangleBox[{31.8, 0.}, {32., 0.3149370125974816}, 
             "RoundingRadius" -> 0]},
           
           ImageSizeCache->{{93.59014329135897, 
            97.98536691272244}, {-40.268120694479734`, 40.27773676735785}}],
          StatusArea[#, 0.3149370125974816]& ,
          TagBoxNote->"0.3149370125974816"],
         StyleBox["0.3149370125974816`", {
           GrayLevel[0]}, StripOnInput -> False]],
        Annotation[#, 
         Style[0.3149370125974816, {
           GrayLevel[0]}], "Tooltip"]& ], 
       TagBox[
        TooltipBox[
         TagBox[
          DynamicBox[{
            FEPrivate`If[
             CurrentValue["MouseOver"], 
             EdgeForm[{
               GrayLevel[0.5], 
               AbsoluteThickness[1.5], 
               Opacity[0.66]}], {}, {}], 
            
            RectangleBox[{32., 0.}, {32.2, 0.3089382123575241}, 
             "RoundingRadius" -> 0]},
           
           ImageSizeCache->{{96.98536691272244, 
            101.38059053408591`}, {-38.752961504729086`, 40.27773676735785}}],
          StatusArea[#, 0.3089382123575241]& ,
          TagBoxNote->"0.3089382123575241"],
         StyleBox["0.3089382123575241`", {
           GrayLevel[0]}, StripOnInput -> False]],
        Annotation[#, 
         Style[0.3089382123575241, {
           GrayLevel[0]}], "Tooltip"]& ], 
       TagBox[
        TooltipBox[
         TagBox[
          DynamicBox[{
            FEPrivate`If[
             CurrentValue["MouseOver"], 
             EdgeForm[{
               GrayLevel[0.5], 
               AbsoluteThickness[1.5], 
               Opacity[0.66]}], {}, {}], 
            RectangleBox[{32.2, 0.}, {32.4, 0.3299340131973676}, 
             "RoundingRadius" -> 0]},
           
           ImageSizeCache->{{100.38059053408591`, 
            104.77581415544927`}, {-44.05601866885446, 40.27773676735785}}],
          StatusArea[#, 0.3299340131973676]& ,
          TagBoxNote->"0.3299340131973676"],
         StyleBox["0.3299340131973676`", {
           GrayLevel[0]}, StripOnInput -> False]],
        Annotation[#, 
         Style[0.3299340131973676, {
           GrayLevel[0]}], "Tooltip"]& ], 
       TagBox[
        TooltipBox[
         TagBox[
          DynamicBox[{
            FEPrivate`If[
             CurrentValue["MouseOver"], 
             EdgeForm[{
               GrayLevel[0.5], 
               AbsoluteThickness[1.5], 
               Opacity[0.66]}], {}, {}], 
            
            RectangleBox[{32.4, 0.}, {32.6, 0.30493901219755615`}, 
             "RoundingRadius" -> 0]},
           
           ImageSizeCache->{{103.77581415544927`, 
            108.17103777681274`}, {-37.742855378229564`, 40.27773676735785}}],
          StatusArea[#, 0.30493901219755615`]& ,
          TagBoxNote->"0.30493901219755615"],
         StyleBox["0.30493901219755615`", {
           GrayLevel[0]}, StripOnInput -> False]],
        Annotation[#, 
         Style[0.30493901219755615`, {
           GrayLevel[0]}], "Tooltip"]& ], 
       TagBox[
        TooltipBox[
         TagBox[
          DynamicBox[{
            FEPrivate`If[
             CurrentValue["MouseOver"], 
             EdgeForm[{
               GrayLevel[0.5], 
               AbsoluteThickness[1.5], 
               Opacity[0.66]}], {}, {}], 
            
            RectangleBox[{32.6, 0.}, {32.8, 0.3579284143171442}, 
             "RoundingRadius" -> 0]},
           
           ImageSizeCache->{{107.17103777681274`, 
            111.5662613981761}, {-51.12676155435125, 40.27773676735785}}],
          StatusArea[#, 0.3579284143171442]& ,
          TagBoxNote->"0.3579284143171442"],
         StyleBox["0.3579284143171442`", {
           GrayLevel[0]}, StripOnInput -> False]],
        Annotation[#, 
         Style[0.3579284143171442, {
           GrayLevel[0]}], "Tooltip"]& ], 
       TagBox[
        TooltipBox[
         TagBox[
          DynamicBox[{
            FEPrivate`If[
             CurrentValue["MouseOver"], 
             EdgeForm[{
               GrayLevel[0.5], 
               AbsoluteThickness[1.5], 
               Opacity[0.66]}], {}, {}], 
            
            RectangleBox[{32.8, 0.}, {33., 0.29894021195760423`}, 
             "RoundingRadius" -> 0]},
           
           ImageSizeCache->{{110.5662613981761, 
            114.96148501953968`}, {-36.22769618848031, 40.27773676735785}}],
          StatusArea[#, 0.29894021195760423`]& ,
          TagBoxNote->"0.29894021195760423"],
         StyleBox["0.29894021195760423`", {
           GrayLevel[0]}, StripOnInput -> False]],
        Annotation[#, 
         Style[0.29894021195760423`, {
           GrayLevel[0]}], "Tooltip"]& ], 
       TagBox[
        TooltipBox[
         TagBox[
          DynamicBox[{
            FEPrivate`If[
             CurrentValue["MouseOver"], 
             EdgeForm[{
               GrayLevel[0.5], 
               AbsoluteThickness[1.5], 
               Opacity[0.66]}], {}, {}], 
            
            RectangleBox[{33., 0.}, {33.2, 0.20195960807838148`}, 
             "RoundingRadius" -> 0]},
           
           ImageSizeCache->{{113.96148501953968`, 
            118.35670864090315`}, {-11.732622620867268`, 40.27773676735785}}],
          StatusArea[#, 0.20195960807838148`]& ,
          TagBoxNote->"0.20195960807838148"],
         StyleBox["0.20195960807838148`", {
           GrayLevel[0]}, StripOnInput -> False]],
        Annotation[#, 
         Style[0.20195960807838148`, {
           GrayLevel[0]}], "Tooltip"]& ], 
       TagBox[
        TooltipBox[
         TagBox[
          DynamicBox[{
            FEPrivate`If[
             CurrentValue["MouseOver"], 
             EdgeForm[{
               GrayLevel[0.5], 
               AbsoluteThickness[1.5], 
               Opacity[0.66]}], {}, {}], 
            
            RectangleBox[{33.2, 0.}, {33.4, 0.18396320735853222`}, 
             "RoundingRadius" -> 0]},
           
           ImageSizeCache->{{117.35670864090315`, 
            121.75193226226651`}, {-7.18714505162113, 40.27773676735785}}],
          StatusArea[#, 0.18396320735853222`]& ,
          TagBoxNote->"0.18396320735853222"],
         StyleBox["0.18396320735853222`", {
           GrayLevel[0]}, StripOnInput -> False]],
        Annotation[#, 
         Style[0.18396320735853222`, {
           GrayLevel[0]}], "Tooltip"]& ], 
       TagBox[
        TooltipBox[
         TagBox[
          DynamicBox[{
            FEPrivate`If[
             CurrentValue["MouseOver"], 
             EdgeForm[{
               GrayLevel[0.5], 
               AbsoluteThickness[1.5], 
               Opacity[0.66]}], {}, {}], 
            
            RectangleBox[{33.4, 0.}, {33.6, 0.14597080583883015`}, 
             "RoundingRadius" -> 0]},
           
           ImageSizeCache->{{120.75193226226651`, 125.14715588362998`}, {
            2.408863150125839, 40.27773676735785}}],
          StatusArea[#, 0.14597080583883015`]& ,
          TagBoxNote->"0.14597080583883015"],
         StyleBox["0.14597080583883015`", {
           GrayLevel[0]}, StripOnInput -> False]],
        Annotation[#, 
         Style[0.14597080583883015`, {
           GrayLevel[0]}], "Tooltip"]& ], 
       TagBox[
        TooltipBox[
         TagBox[
          DynamicBox[{
            FEPrivate`If[
             CurrentValue["MouseOver"], 
             EdgeForm[{
               GrayLevel[0.5], 
               AbsoluteThickness[1.5], 
               Opacity[0.66]}], {}, {}], 
            
            RectangleBox[{33.6, 0.}, {33.8, 0.11597680463907466`}, 
             "RoundingRadius" -> 0]},
           
           ImageSizeCache->{{124.14715588362998`, 128.54237950499333`}, {
            9.984659098871102, 40.27773676735785}}],
          StatusArea[#, 0.11597680463907466`]& ,
          TagBoxNote->"0.11597680463907466"],
         StyleBox["0.11597680463907466`", {
           GrayLevel[0]}, StripOnInput -> False]],
        Annotation[#, 
         Style[0.11597680463907466`, {
           GrayLevel[0]}], "Tooltip"]& ], 
       TagBox[
        TooltipBox[
         TagBox[
          DynamicBox[{
            FEPrivate`If[
             CurrentValue["MouseOver"], 
             EdgeForm[{
               GrayLevel[0.5], 
               AbsoluteThickness[1.5], 
               Opacity[0.66]}], {}, {}], 
            
            RectangleBox[{33.8, 0.}, {34., 0.06698660267946316}, 
             "RoundingRadius" -> 0]},
           
           ImageSizeCache->{{127.54237950499333`, 131.9376031263568}, {
            22.358459148491104`, 40.27773676735785}}],
          StatusArea[#, 0.06698660267946316]& ,
          TagBoxNote->"0.06698660267946316"],
         StyleBox["0.06698660267946316`", {
           GrayLevel[0]}, StripOnInput -> False]],
        Annotation[#, 
         Style[0.06698660267946316, {
           GrayLevel[0]}], "Tooltip"]& ], 
       TagBox[
        TooltipBox[
         TagBox[
          DynamicBox[{
            FEPrivate`If[
             CurrentValue["MouseOver"], 
             EdgeForm[{
               GrayLevel[0.5], 
               AbsoluteThickness[1.5], 
               Opacity[0.66]}], {}, {}], 
            
            RectangleBox[{34., 0.}, {34.2, 0.07398520295940707}, 
             "RoundingRadius" -> 0]},
           
           ImageSizeCache->{{130.9376031263568, 135.33282674772028`}, {
            20.590773427116968`, 40.27773676735785}}],
          StatusArea[#, 0.07398520295940707]& ,
          TagBoxNote->"0.07398520295940707"],
         StyleBox["0.07398520295940707`", {
           GrayLevel[0]}, StripOnInput -> False]],
        Annotation[#, 
         Style[0.07398520295940707, {
           GrayLevel[0]}], "Tooltip"]& ], 
       TagBox[
        TooltipBox[
         TagBox[
          DynamicBox[{
            FEPrivate`If[
             CurrentValue["MouseOver"], 
             EdgeForm[{
               GrayLevel[0.5], 
               AbsoluteThickness[1.5], 
               Opacity[0.66]}], {}, {}], 
            
            RectangleBox[{34.2, 0.}, {34.4, 0.03899220155968889}, 
             "RoundingRadius" -> 0]},
           
           ImageSizeCache->{{134.33282674772028`, 138.72805036908363`}, {
            29.42920203398731, 40.27773676735785}}],
          StatusArea[#, 0.03899220155968889]& ,
          TagBoxNote->"0.03899220155968889"],
         StyleBox["0.03899220155968889`", {
           GrayLevel[0]}, StripOnInput -> False]],
        Annotation[#, 
         Style[0.03899220155968889, {
           GrayLevel[0]}], "Tooltip"]& ], 
       TagBox[
        TooltipBox[
         TagBox[
          DynamicBox[{
            FEPrivate`If[
             CurrentValue["MouseOver"], 
             EdgeForm[{
               GrayLevel[0.5], 
               AbsoluteThickness[1.5], 
               Opacity[0.66]}], {}, {}], 
            
            RectangleBox[{34.4, 0.}, {34.6, 0.019996000799839746`}, 
             "RoundingRadius" -> 0]},
           
           ImageSizeCache->{{137.72805036908363`, 142.12327399044722`}, {
            34.22720613486032, 40.27773676735785}}],
          StatusArea[#, 0.019996000799839746`]& ,
          TagBoxNote->"0.019996000799839746"],
         StyleBox["0.019996000799839746`", {
           GrayLevel[0]}, StripOnInput -> False]],
        Annotation[#, 
         Style[0.019996000799839746`, {
           GrayLevel[0]}], "Tooltip"]& ], 
       TagBox[
        TooltipBox[
         TagBox[
          DynamicBox[{
            FEPrivate`If[
             CurrentValue["MouseOver"], 
             EdgeForm[{
               GrayLevel[0.5], 
               AbsoluteThickness[1.5], 
               Opacity[0.66]}], {}, {}], 
            
            RectangleBox[{34.6, 0.}, {34.8, 0.012997400519896297`}, 
             "RoundingRadius" -> 0]},
           
           ImageSizeCache->{{141.12327399044722`, 145.51849761181057`}, {
            35.99489185623434, 40.27773676735785}}],
          StatusArea[#, 0.012997400519896297`]& ,
          TagBoxNote->"0.012997400519896297"],
         StyleBox["0.012997400519896297`", {
           GrayLevel[0]}, StripOnInput -> False]],
        Annotation[#, 
         Style[0.012997400519896297`, {
           GrayLevel[0]}], "Tooltip"]& ], 
       TagBox[
        TooltipBox[
         TagBox[
          DynamicBox[{
            FEPrivate`If[
             CurrentValue["MouseOver"], 
             EdgeForm[{
               GrayLevel[0.5], 
               AbsoluteThickness[1.5], 
               Opacity[0.66]}], {}, {}], 
            
            RectangleBox[{34.8, 0.}, {35., 0.032993401319735584`}, 
             "RoundingRadius" -> 0]},
           
           ImageSizeCache->{{144.51849761181057`, 148.91372123317404`}, {
            30.94436122373692, 40.27773676735785}}],
          StatusArea[#, 0.032993401319735584`]& ,
          TagBoxNote->"0.032993401319735584"],
         StyleBox["0.032993401319735584`", {
           GrayLevel[0]}, StripOnInput -> False]],
        Annotation[#, 
         Style[0.032993401319735584`, {
           GrayLevel[0]}], "Tooltip"]& ], 
       TagBox[
        TooltipBox[
         TagBox[
          DynamicBox[{
            FEPrivate`If[
             CurrentValue["MouseOver"], 
             EdgeForm[{
               GrayLevel[0.5], 
               AbsoluteThickness[1.5], 
               Opacity[0.66]}], {}, {}], 
            
            RectangleBox[{35., 0.}, {35.2, 0.015996800639871796`}, 
             "RoundingRadius" -> 0]},
           
           ImageSizeCache->{{147.91372123317404`, 152.30894485453751`}, {
            35.237312261359826`, 40.27773676735785}}],
          StatusArea[#, 0.015996800639871796`]& ,
          TagBoxNote->"0.015996800639871796"],
         StyleBox["0.015996800639871796`", {
           GrayLevel[0]}, StripOnInput -> False]],
        Annotation[#, 
         Style[0.015996800639871796`, {
           GrayLevel[0]}], "Tooltip"]& ], 
       TagBox[
        TooltipBox[
         TagBox[
          DynamicBox[{
            FEPrivate`If[
             CurrentValue["MouseOver"], 
             EdgeForm[{
               GrayLevel[0.5], 
               AbsoluteThickness[1.5], 
               Opacity[0.66]}], {}, {}], 
            
            RectangleBox[{35.2, 0.}, {35.4, 0.0019996000799840457`}, 
             "RoundingRadius" -> 0]},
           
           ImageSizeCache->{{151.30894485453751`, 155.70416847590087`}, {
            38.77268370410808, 40.27773676735785}}],
          StatusArea[#, 0.0019996000799840457`]& ,
          TagBoxNote->"0.0019996000799840457"],
         StyleBox["0.0019996000799840457`", {
           GrayLevel[0]}, StripOnInput -> False]],
        Annotation[#, 
         Style[0.0019996000799840457`, {
           GrayLevel[0]}], "Tooltip"]& ], 
       TagBox[
        TooltipBox[
         TagBox[
          DynamicBox[{
            FEPrivate`If[
             CurrentValue["MouseOver"], 
             EdgeForm[{
               GrayLevel[0.5], 
               AbsoluteThickness[1.5], 
               Opacity[0.66]}], {}, {}], 
            
            RectangleBox[{35.4, 0.}, {35.6, 0.002999400119975962}, 
             "RoundingRadius" -> 0]},
           
           ImageSizeCache->{{154.70416847590087`, 159.09939209726434`}, {
            38.52015717248322, 40.27773676735785}}],
          StatusArea[#, 0.002999400119975962]& ,
          TagBoxNote->"0.002999400119975962"],
         StyleBox["0.002999400119975962`", {
           GrayLevel[0]}, StripOnInput -> False]],
        Annotation[#, 
         Style[0.002999400119975962, {
           GrayLevel[0]}], "Tooltip"]& ], 
       TagBox[
        TooltipBox[
         TagBox[
          DynamicBox[{
            FEPrivate`If[
             CurrentValue["MouseOver"], 
             EdgeForm[{
               GrayLevel[0.5], 
               AbsoluteThickness[1.5], 
               Opacity[0.66]}], {}, {}], 
            
            RectangleBox[{35.6, 0.}, {35.8, 0.0019996000799840457`}, 
             "RoundingRadius" -> 0]},
           
           ImageSizeCache->{{158.09939209726434`, 162.4946157186277}, {
            38.77268370410808, 40.27773676735785}}],
          StatusArea[#, 0.0019996000799840457`]& ,
          TagBoxNote->"0.0019996000799840457"],
         StyleBox["0.0019996000799840457`", {
           GrayLevel[0]}, StripOnInput -> False]],
        Annotation[#, 
         Style[0.0019996000799840457`, {
           GrayLevel[0]}], "Tooltip"]& ], {}, {}, {}, 
       TagBox[
        TooltipBox[
         TagBox[
          DynamicBox[{
            FEPrivate`If[
             CurrentValue["MouseOver"], 
             EdgeForm[{
               GrayLevel[0.5], 
               AbsoluteThickness[1.5], 
               Opacity[0.66]}], {}, {}], 
            
            RectangleBox[{36.4, 0.}, {36.6, 0.0009998000399919873}, 
             "RoundingRadius" -> 0]},
           
           ImageSizeCache->{{171.6802865827181, 176.07551020408158`}, {
            39.02521023573298, 40.27773676735785}}],
          StatusArea[#, 0.0009998000399919873]& ,
          TagBoxNote->"0.0009998000399919873"],
         StyleBox["0.0009998000399919873`", {
           GrayLevel[0]}, StripOnInput -> False]],
        Annotation[#, 
         Style[0.0009998000399919873, {
           GrayLevel[0]}], "Tooltip"]& ]}, {}, {}}, {{}, {}, 
      {RGBColor[1, 0, 0], Thickness[Large], LineBox[CompressedData["
1:eJwV13k8VO8XB/CiIlFEKtlFsifZZsY5MyRF9ixF2clXJELZklKhqBCKiGQn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        "]]}}},
    AspectRatio->NCache[GoldenRatio^(-1), 0.6180339887498948],
    Axes->{True, True},
    AxesOrigin->{27.2, 0},
    FrameTicks->{{Automatic, Automatic}, {Automatic, Automatic}},
    GridLines->{None, None},
    PlotRange->NCache[{{
        Rational[136, 5], 
        Rational[183, 5]}, {All, All}}, {{27.2, 36.6}, {All, All}}],
    PlotRangePadding->{{
       Scaled[0.02], 
       Scaled[0.02]}, {
       Scaled[0.02], 
       Scaled[0.1]}},
    Ticks->{Automatic, Automatic}]}], "}"}]], "Output",
 ImageSize->{580, 119},
 ImageMargins->{{0, 0}, {0, 0}},
 ImageRegion->{{0, 1}, {0, 1}},
 CellLabel->"Out[5]=",
 CellID->36970811]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 InterpretationBox[Cell[TextData[{
   "Neat Examples",
   "\[NonBreakingSpace]\[NonBreakingSpace]",
   Cell["(1)", "ExampleCount"]
  }], "ExampleSection"],
  $Line = 0; Null]], "ExampleSection",
 CellID->589267740],

Cell[BoxData[
 RowBox[{"Needs", "[", "\"\<Metropolis`\>\"", "]"}]], "Input",
 CellLabel->"In[1]:=",
 CellID->192544099],

Cell["\<\
This samples a Possion distribution, a truncated normal distribution, a \
bivariate normal distribution\
\>", "ExampleText",
 CellID->73418448],

Cell[CellGroupData[{

Cell[BoxData[{
 RowBox[{
  RowBox[{"Clear", "[", 
   RowBox[{
   "lambda", ",", "V", ",", "iV", ",", "funs", ",", "logpdf", ",", "scale", 
    ",", "cov", ",", "n", ",", "res"}], "]"}], ";"}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{"lambda", "=", "2"}], ";"}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{"V", "=", 
   RowBox[{"{", 
    RowBox[{
     RowBox[{"{", 
      RowBox[{"1", ",", "0.9"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"0.9", ",", "1"}], "}"}]}], "}"}]}], 
  ";"}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{"iV", "=", 
   RowBox[{"Inverse", "[", "V", "]"}]}], ";"}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{"funs", "=", 
   RowBox[{"{", 
    RowBox[{
     RowBox[{"Function", "[", 
      RowBox[{"x", ",", 
       RowBox[{"Log", "[", 
        FractionBox[
         RowBox[{
          SuperscriptBox["lambda", "x"], " ", 
          SuperscriptBox["\[ExponentialE]", 
           RowBox[{"-", "lambda"}]]}], 
         RowBox[{"x", " ", "!"}]], "]"}]}], "]"}], ",", 
     RowBox[{"Function", "[", 
      RowBox[{"x", ",", 
       RowBox[{
        RowBox[{"-", 
         RowBox[{"x", "^", "2"}]}], "/", "2"}]}], "]"}], ",", 
     RowBox[{"Function", "[", 
      RowBox[{"x", ",", 
       RowBox[{
        RowBox[{"-", 
         RowBox[{"x", ".", "iV", ".", "x"}]}], "/", "2"}]}], "]"}]}], "}"}]}],
   ";"}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{"logpdf", "=", 
   RowBox[{"Function", "[", 
    RowBox[{
     RowBox[{"{", 
      RowBox[{"x", ",", "indices"}], "}"}], ",", 
     RowBox[{
      RowBox[{
       RowBox[{
        RowBox[{"#", "[", 
         RowBox[{"[", "1", "]"}], "]"}], "[", 
        RowBox[{"#", "[", 
         RowBox[{"[", "2", "]"}], "]"}], "]"}], "&"}], "/@", 
      RowBox[{"Transpose", "[", 
       RowBox[{"{", 
        RowBox[{
         RowBox[{"funs", "[", 
          RowBox[{"[", "indices", "]"}], "]"}], ",", 
         RowBox[{"x", "[", 
          RowBox[{"[", "indices", "]"}], "]"}]}], "}"}], "]"}]}]}], 
    "\[IndentingNewLine]", "]"}]}], ";"}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{"scale", "=", 
   RowBox[{"{", 
    RowBox[{"lambda", ",", "3", ",", "3"}], "}"}]}], 
  ";"}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{"cov", "=", 
   RowBox[{"{", 
    RowBox[{
     RowBox[{"{", 
      RowBox[{"{", "1", "}"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"{", "1", "}"}], "}"}], ",", 
     RowBox[{"Inverse", "[", "iV", "]"}]}], "}"}]}], 
  ";"}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{"n", "=", "5000"}], ";"}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{"x2min", "=", 
   RowBox[{"-", "1"}]}], ";"}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{
   RowBox[{"res", "=", 
    RowBox[{"BlockStudentTMetropolis", "[", 
     RowBox[{"logpdf", ",", 
      RowBox[{"{", 
       RowBox[{"0", ",", "0", ",", 
        RowBox[{"{", 
         RowBox[{"0", ",", "0"}], "}"}]}], "}"}], ",", "scale", ",", "cov", 
      ",", 
      RowBox[{"2", " ", "n"}], ",", 
      RowBox[{"IsReturnLast", "\[Rule]", "False"}], ",", 
      RowBox[{"IsDiscrete", "\[Rule]", 
       RowBox[{"{", 
        RowBox[{"True", ",", "False", ",", "False"}], "}"}]}], ",", 
      RowBox[{"DomainConstraint", "\[Rule]", 
       RowBox[{"(", 
        RowBox[{
         RowBox[{"{", 
          RowBox[{
           RowBox[{
            RowBox[{"#", "[", 
             RowBox[{"[", "1", "]"}], "]"}], "\[GreaterEqual]", "0"}], ",", 
           RowBox[{
            RowBox[{"#", "[", 
             RowBox[{"[", "2", "]"}], "]"}], ">", "x2min"}], ",", "True"}], 
          "}"}], "&"}], ")"}]}]}], "]"}]}], ";"}], "//", "Timing"}]}], "Input",
 CellLabel->"In[2]:=",
 CellID->756940083],

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{"3.182000000000002`", ",", "Null"}], "}"}]], "Output",
 ImageSize->{92, 15},
 ImageMargins->{{0, 0}, {0, 0}},
 ImageRegion->{{0, 1}, {0, 1}},
 CellLabel->"Out[2]=",
 CellID->693900897]
}, Open  ]],

Cell["Mean accept ratio:", "ExampleText",
 CellID->24993314],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"N", "[", 
  RowBox[{"Mean", "[", 
   RowBox[{"res", "[", 
    RowBox[{"[", 
     RowBox[{
      RowBox[{"n", ";;"}], ",", 
      RowBox[{
       RowBox[{"-", "2"}], ";;"}]}], "]"}], "]"}], "]"}], "]"}]], "Input",
 CellLabel->"In[3]:=",
 CellID->265777506],

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{
   RowBox[{"{", 
    RowBox[{
    "0.6030793841231754`", ",", "0.3073385322935413`", ",", 
     "0.2731453709258148`"}], "}"}], ",", 
   RowBox[{"{", 
    RowBox[{
    "0.46002951072209713`", ",", "0.3040189839224292`", ",", 
     "0.2739215362229127`"}], "}"}]}], "}"}]], "Output",
 ImageSize->{430, 15},
 ImageMargins->{{0, 0}, {0, 0}},
 ImageRegion->{{0, 1}, {0, 1}},
 CellLabel->"Out[3]=",
 CellID->307896011]
}, Open  ]],

Cell["Time series", "ExampleText",
 CellID->193398859],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"Table", "[", 
  RowBox[{
   RowBox[{"ListLinePlot", "[", 
    RowBox[{
     RowBox[{"res", "[", 
      RowBox[{"[", 
       RowBox[{
        RowBox[{";;", ";;", "10"}], ",", "1", ",", "i"}], "]"}], "]"}], ",", 
     RowBox[{"PlotRange", "\[Rule]", "All"}], ",", 
     RowBox[{"AxesLabel", "\[Rule]", 
      RowBox[{"{", 
       RowBox[{"\"\<t\>\"", ",", 
        RowBox[{"\"\<x\>\"", "<>", 
         RowBox[{"ToString", "[", "i", "]"}]}]}], "}"}]}]}], "]"}], ",", 
   RowBox[{"{", 
    RowBox[{"i", ",", "2"}], "}"}]}], "]"}]], "Input",
 CellLabel->"In[4]:=",
 CellID->402775542],

Cell[BoxData[
 GraphicsBox[
  TagBox[RasterBox[CompressedData["
1:eJztnV2oHdd1x0/pe16agk0oSYVDIZGNXRITlMgnjevGjpKmMY4vSTAEqRE2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    "], {{0, 134}, {385, 0}}, {0, 255},
    ColorFunction->RGBColor],
   BoxForm`ImageTag["Byte", ColorSpace -> "RGB", Interleaving -> True],
   Selectable->False],
  BaseStyle->"ImageGraphics",
  ImageSize->Automatic,
  ImageSizeRaw->{385, 134},
  PlotRange->{{0, 385}, {0, 134}}]], "Output",
 ImageSize->{389, 134},
 ImageMargins->{{0, 0}, {0, 0}},
 ImageRegion->{{0, 1}, {0, 1}},
 CellLabel->"Out[4]=",
 CellID->214186837]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"Table", "[", 
  RowBox[{
   RowBox[{"ListLinePlot", "[", 
    RowBox[{
     RowBox[{"res", "[", 
      RowBox[{"[", 
       RowBox[{
        RowBox[{";;", ";;", "10"}], ",", "1", ",", 
        RowBox[{"-", "1"}], ",", "i"}], "]"}], "]"}], ",", 
     RowBox[{"PlotRange", "\[Rule]", "All"}], ",", 
     RowBox[{"AxesLabel", "\[Rule]", 
      RowBox[{"{", 
       RowBox[{"\"\<t\>\"", ",", 
        RowBox[{"\"\<x3\>\"", "<>", 
         RowBox[{"ToString", "[", "i", "]"}]}]}], "}"}]}]}], "]"}], ",", 
   RowBox[{"{", 
    RowBox[{"i", ",", "2"}], "}"}]}], "]"}]], "Input",
 CellLabel->"In[5]:=",
 CellID->223197944],

Cell[BoxData[
 GraphicsBox[
  TagBox[RasterBox[CompressedData["
1:eJztnW2oJsd152/Y74awa7BIlmyE8oIjGXsli6Aw6hhZRqMxspWs/awTOzGy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    "], {{0, 118}, {385, 0}}, {0, 255},
    ColorFunction->RGBColor],
   BoxForm`ImageTag["Byte", ColorSpace -> "RGB", Interleaving -> True],
   Selectable->False],
  BaseStyle->"ImageGraphics",
  ImageSize->Automatic,
  ImageSizeRaw->{385, 118},
  PlotRange->{{0, 385}, {0, 118}}]], "Output",
 ImageSize->{389, 118},
 ImageMargins->{{0, 0}, {0, 0}},
 ImageRegion->{{0, 1}, {0, 1}},
 CellLabel->"Out[5]=",
 CellID->201295919]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{
   RowBox[{"ListPlot", "[", 
    RowBox[{
     RowBox[{"res", "[", 
      RowBox[{"[", 
       RowBox[{
        RowBox[{";;", ";;", "10"}], ",", "1", ",", 
        RowBox[{"-", "1"}]}], "]"}], "]"}], ",", 
     RowBox[{"PlotRange", "\[Rule]", "All"}], ",", 
     RowBox[{"AxesLabel", "\[Rule]", 
      RowBox[{"{", 
       RowBox[{"\"\<x31\>\"", ",", "\"\<x32\>\""}], "}"}]}]}], "]"}], ",", 
   RowBox[{
    RowBox[{"Covariance", "[", 
     RowBox[{"res", "[", 
      RowBox[{"[", 
       RowBox[{
        RowBox[{";;", ";;", "10"}], ",", "1", ",", 
        RowBox[{"-", "1"}]}], "]"}], "]"}], "]"}], "//", "MatrixForm"}]}], 
  "}"}]], "Input",
 CellLabel->"In[6]:=",
 CellID->36445872],

Cell[BoxData[
 GraphicsBox[
  TagBox[RasterBox[CompressedData["
1:eJztnc9vJFcRx424R5G4cAQJKRciwSUHS55fHnt3be+u7U02m7ALxgKHCNhE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    "], {{0, 117}, {345, 0}}, {0, 255},
    ColorFunction->RGBColor],
   BoxForm`ImageTag["Byte", ColorSpace -> "RGB", Interleaving -> True],
   Selectable->False],
  BaseStyle->"ImageGraphics",
  ImageSize->Automatic,
  ImageSizeRaw->{345, 117},
  PlotRange->{{0, 345}, {0, 117}}]], "Output",
 ImageSize->{349, 117},
 ImageMargins->{{0, 0}, {0, 0}},
 ImageRegion->{{0, 1}, {0, 1}},
 CellLabel->"Out[6]=",
 CellID->432403975]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[{
 RowBox[{
  RowBox[{"data", "=", 
   RowBox[{"res", "[", 
    RowBox[{"[", 
     RowBox[{
      RowBox[{"n", ";;"}], ",", "1", ",", "1"}], "]"}], "]"}]}], 
  ";"}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{"xls", "=", 
   RowBox[{"Range", "[", 
    RowBox[{
     RowBox[{"Min", "[", "data", "]"}], ",", 
     RowBox[{"Max", "[", "data", "]"}]}], "]"}]}], 
  ";"}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{"freq", "=", 
   RowBox[{
    RowBox[{
     RowBox[{"Count", "[", 
      RowBox[{"data", ",", "#"}], "]"}], "&"}], "/@", "xls"}]}], 
  ";"}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{"freq", "/=", 
   RowBox[{"Total", "[", "freq", "]"}]}], ";"}], "\[IndentingNewLine]", 
 RowBox[{"Show", "[", 
  RowBox[{
   RowBox[{"Plot", "[", 
    RowBox[{
     RowBox[{"Exp", "[", 
      RowBox[{
       RowBox[{"funs", "[", 
        RowBox[{"[", "1", "]"}], "]"}], "[", "x", "]"}], "]"}], ",", 
     RowBox[{"{", 
      RowBox[{"x", ",", 
       RowBox[{"Min", "[", "xls", "]"}], ",", 
       RowBox[{"Max", "[", "xls", "]"}]}], "}"}], ",", 
     RowBox[{"PlotStyle", "\[Rule]", "Red"}]}], "]"}], ",", 
   RowBox[{"ListPlot", "[", 
    RowBox[{
     RowBox[{"Transpose", "[", 
      RowBox[{"{", 
       RowBox[{"xls", ",", "freq"}], "}"}], "]"}], ",", 
     RowBox[{"Filling", "\[Rule]", "Axis"}], ",", 
     RowBox[{"PlotMarkers", "\[Rule]", "Automatic"}]}], "]"}], ",", 
   RowBox[{"PlotRange", "\[Rule]", "All"}]}], "]"}]}], "Input",
 CellLabel->"In[7]:=",
 CellID->182745918],

Cell[BoxData[
 GraphicsBox[
  TagBox[RasterBox[CompressedData["
1:eJztnc1rJEUUwAf8B3LMcb3luHjKBpYdRTCCh7isEkQ0xhDGk1EUVhBGURnQ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    "], {{0, 114}, {180, 0}}, {0, 255},
    ColorFunction->RGBColor],
   BoxForm`ImageTag["Byte", ColorSpace -> "RGB", Interleaving -> True],
   Selectable->False],
  BaseStyle->"ImageGraphics",
  ImageSize->Automatic,
  ImageSizeRaw->{180, 114},
  PlotRange->{{0, 180}, {0, 114}}]], "Output",
 ImageSize->{184, 114},
 ImageMargins->{{0, 0}, {0, 0}},
 ImageRegion->{{0, 1}, {0, 1}},
 CellLabel->"Out[7]=",
 CellID->1021282114]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"Show", "[", 
  RowBox[{
   RowBox[{
    RowBox[{"data", "=", 
     RowBox[{"res", "[", 
      RowBox[{"[", 
       RowBox[{
        RowBox[{";;", ";;", "10"}], ",", "1", ",", "2"}], "]"}], "]"}]}], ";", 
    RowBox[{"Histogram", "[", 
     RowBox[{"data", ",", "Automatic", ",", "\"\<ProbabilityDensity\>\""}], 
     "]"}]}], ",", 
   RowBox[{"Plot", "[", 
    RowBox[{
     RowBox[{
      RowBox[{"PDF", "[", 
       RowBox[{
        RowBox[{"NormalDistribution", "[", 
         RowBox[{"0", ",", "1"}], "]"}], ",", "x"}], "]"}], "/", 
      RowBox[{"(", 
       RowBox[{
        FractionBox["1", "2"], " ", 
        RowBox[{"Erfc", "[", 
         FractionBox["x2min", 
          SqrtBox["2"]], "]"}]}], ")"}]}], ",", 
     RowBox[{"{", 
      RowBox[{"x", ",", 
       RowBox[{"Min", "[", "data", "]"}], ",", 
       RowBox[{"Max", "[", "data", "]"}]}], "}"}], ",", 
     RowBox[{"PlotStyle", "\[Rule]", 
      RowBox[{"{", 
       RowBox[{"Red", ",", "Thick"}], "}"}]}], ",", 
     RowBox[{"PlotRange", "\[Rule]", "All"}], ",", 
     RowBox[{"AxesLabel", "\[Rule]", 
      RowBox[{"{", 
       RowBox[{"\"\<x2\>\"", ",", "\"\<freq\>\""}], "}"}]}]}], "]"}]}], 
  "]"}]], "Input",
 CellLabel->"In[8]:=",
 CellID->530831861],

Cell[BoxData[
 GraphicsBox[{
   {RGBColor[0.798413061722744, 0.824719615472648, 0.968322270542458], 
    EdgeForm[Opacity[0.7]], {}, 
    {RGBColor[0.798413061722744, 0.824719615472648, 0.968322270542458], 
     EdgeForm[Opacity[0.7]], 
     TagBox[
      TooltipBox[
       TagBox[
        DynamicBox[{
          FEPrivate`If[
           CurrentValue["MouseOver"], 
           EdgeForm[{
             GrayLevel[0.5], 
             AbsoluteThickness[1.5], 
             Opacity[0.66]}], {}, {}], 
          
          RectangleBox[{-1., 0.}, {-0.8, 0.31000000000000005`}, 
           "RoundingRadius" -> 0]},
         ImageSizeCache->{{15.500000000000007`, 
          23.75343228200372}, {-15.140504940855408`, 40.27773676735786}}],
        StatusArea[#, 0.31000000000000005`]& ,
        TagBoxNote->"0.31000000000000005"],
       StyleBox["0.31000000000000005`", {
         GrayLevel[0]}, StripOnInput -> False]],
      Annotation[#, 
       Style[0.31000000000000005`, {
         GrayLevel[0]}], "Tooltip"]& ], 
     TagBox[
      TooltipBox[
       TagBox[
        DynamicBox[{
          FEPrivate`If[
           CurrentValue["MouseOver"], 
           EdgeForm[{
             GrayLevel[0.5], 
             AbsoluteThickness[1.5], 
             Opacity[0.66]}], {}, {}], 
          
          RectangleBox[{-0.8, 0.}, {-0.6, 0.3549999999999998}, 
           "RoundingRadius" -> 0]},
         ImageSizeCache->{{22.75343228200372, 
          31.006864564007433`}, {-23.039927124305677`, 40.27773676735786}}],
        StatusArea[#, 0.3549999999999998]& ,
        TagBoxNote->"0.3549999999999998"],
       StyleBox["0.3549999999999998`", {
         GrayLevel[0]}, StripOnInput -> False]],
      Annotation[#, 
       Style[0.3549999999999998, {
         GrayLevel[0]}], "Tooltip"]& ], 
     TagBox[
      TooltipBox[
       TagBox[
        DynamicBox[{
          FEPrivate`If[
           CurrentValue["MouseOver"], 
           EdgeForm[{
             GrayLevel[0.5], 
             AbsoluteThickness[1.5], 
             Opacity[0.66]}], {}, {}], 
          
          RectangleBox[{-0.6, 0.}, {-0.4, 0.39000000000000007`}, 
           "RoundingRadius" -> 0]},
         ImageSizeCache->{{30.006864564007433`, 
          38.260296846011144`}, {-29.183922155878186`, 40.27773676735786}}],
        StatusArea[#, 0.39000000000000007`]& ,
        TagBoxNote->"0.39000000000000007"],
       StyleBox["0.39000000000000007`", {
         GrayLevel[0]}, StripOnInput -> False]],
      Annotation[#, 
       Style[0.39000000000000007`, {
         GrayLevel[0]}], "Tooltip"]& ], 
     TagBox[
      TooltipBox[
       TagBox[
        DynamicBox[{
          FEPrivate`If[
           CurrentValue["MouseOver"], 
           EdgeForm[{
             GrayLevel[0.5], 
             AbsoluteThickness[1.5], 
             Opacity[0.66]}], {}, {}], 
          RectangleBox[{-0.4, 0.}, {-0.2, 0.375}, "RoundingRadius" -> 0]},
         ImageSizeCache->{{37.260296846011144`, 
          45.51372912801485}, {-26.550781428061406`, 40.27773676735786}}],
        StatusArea[#, 0.375]& ,
        TagBoxNote->"0.375"],
       StyleBox["0.375`", {
         GrayLevel[0]}, StripOnInput -> False]],
      Annotation[#, 
       Style[0.375, {
         GrayLevel[0]}], "Tooltip"]& ], 
     TagBox[
      TooltipBox[
       TagBox[
        DynamicBox[{
          FEPrivate`If[
           CurrentValue["MouseOver"], 
           EdgeForm[{
             GrayLevel[0.5], 
             AbsoluteThickness[1.5], 
             Opacity[0.66]}], {}, {}], 
          RectangleBox[{-0.2, 0.}, {0., 0.515}, "RoundingRadius" -> 0]},
         ImageSizeCache->{{44.51372912801485, 
          52.767161410018566`}, {-51.126761554351255`, 40.27773676735786}}],
        StatusArea[#, 0.515]& ,
        TagBoxNote->"0.515"],
       StyleBox["0.515`", {
         GrayLevel[0]}, StripOnInput -> False]],
      Annotation[#, 
       Style[0.515, {
         GrayLevel[0]}], "Tooltip"]& ], 
     TagBox[
      TooltipBox[
       TagBox[
        DynamicBox[{
          FEPrivate`If[
           CurrentValue["MouseOver"], 
           EdgeForm[{
             GrayLevel[0.5], 
             AbsoluteThickness[1.5], 
             Opacity[0.66]}], {}, {}], 
          
          RectangleBox[{0., 0.}, {0.2, 0.46499999999999997`}, 
           "RoundingRadius" -> 0]},
         ImageSizeCache->{{51.767161410018566`, 
          60.02059369202228}, {-42.349625794962016`, 40.27773676735786}}],
        StatusArea[#, 0.46499999999999997`]& ,
        TagBoxNote->"0.46499999999999997"],
       StyleBox["0.46499999999999997`", {
         GrayLevel[0]}, StripOnInput -> False]],
      Annotation[#, 
       Style[0.46499999999999997`, {
         GrayLevel[0]}], "Tooltip"]& ], 
     TagBox[
      TooltipBox[
       TagBox[
        DynamicBox[{
          FEPrivate`If[
           CurrentValue["MouseOver"], 
           EdgeForm[{
             GrayLevel[0.5], 
             AbsoluteThickness[1.5], 
             Opacity[0.66]}], {}, {}], 
          RectangleBox[{0.2, 0.}, {0.4, 0.49}, "RoundingRadius" -> 0]},
         ImageSizeCache->{{59.02059369202228, 
          67.27402597402599}, {-46.738193674656635`, 40.27773676735786}}],
        StatusArea[#, 0.49]& ,
        TagBoxNote->"0.49"],
       StyleBox["0.49`", {
         GrayLevel[0]}, StripOnInput -> False]],
      Annotation[#, 
       Style[0.49, {
         GrayLevel[0]}], "Tooltip"]& ], 
     TagBox[
      TooltipBox[
       TagBox[
        DynamicBox[{
          FEPrivate`If[
           CurrentValue["MouseOver"], 
           EdgeForm[{
             GrayLevel[0.5], 
             AbsoluteThickness[1.5], 
             Opacity[0.66]}], {}, {}], 
          RectangleBox[{0.4, 0.}, {0.6, 0.4250000000000001}, "RoundingRadius" -> 
           0]},
         ImageSizeCache->{{66.27402597402599, 
          74.5274582560297}, {-35.32791718745066, 40.27773676735786}}],
        StatusArea[#, 0.4250000000000001]& ,
        TagBoxNote->"0.4250000000000001"],
       StyleBox["0.4250000000000001`", {
         GrayLevel[0]}, StripOnInput -> False]],
      Annotation[#, 
       Style[0.4250000000000001, {
         GrayLevel[0]}], "Tooltip"]& ], 
     TagBox[
      TooltipBox[
       TagBox[
        DynamicBox[{
          FEPrivate`If[
           CurrentValue["MouseOver"], 
           EdgeForm[{
             GrayLevel[0.5], 
             AbsoluteThickness[1.5], 
             Opacity[0.66]}], {}, {}], 
          
          RectangleBox[{0.6, 0.}, {0.8, 0.3249999999999999}, "RoundingRadius" -> 
           0]},
         ImageSizeCache->{{73.5274582560297, 
          81.78089053803342}, {-17.773645668672152`, 40.27773676735786}}],
        StatusArea[#, 0.3249999999999999]& ,
        TagBoxNote->"0.3249999999999999"],
       StyleBox["0.3249999999999999`", {
         GrayLevel[0]}, StripOnInput -> False]],
      Annotation[#, 
       Style[0.3249999999999999, {
         GrayLevel[0]}], "Tooltip"]& ], 
     TagBox[
      TooltipBox[
       TagBox[
        DynamicBox[{
          FEPrivate`If[
           CurrentValue["MouseOver"], 
           EdgeForm[{
             GrayLevel[0.5], 
             AbsoluteThickness[1.5], 
             Opacity[0.66]}], {}, {}], 
          
          RectangleBox[{0.8, 0.}, {1., 0.35500000000000004`}, 
           "RoundingRadius" -> 0]},
         ImageSizeCache->{{80.78089053803342, 
          89.03432282003712}, {-23.039927124305713`, 40.27773676735786}}],
        StatusArea[#, 0.35500000000000004`]& ,
        TagBoxNote->"0.35500000000000004"],
       StyleBox["0.35500000000000004`", {
         GrayLevel[0]}, StripOnInput -> False]],
      Annotation[#, 
       Style[0.35500000000000004`, {
         GrayLevel[0]}], "Tooltip"]& ], 
     TagBox[
      TooltipBox[
       TagBox[
        DynamicBox[{
          FEPrivate`If[
           CurrentValue["MouseOver"], 
           EdgeForm[{
             GrayLevel[0.5], 
             AbsoluteThickness[1.5], 
             Opacity[0.66]}], {}, {}], 
          
          RectangleBox[{1., 0.}, {1.2, 0.31500000000000006`}, 
           "RoundingRadius" -> 0]},
         ImageSizeCache->{{88.03432282003712, 
          96.28775510204083}, {-16.018218516794327`, 40.27773676735786}}],
        StatusArea[#, 0.31500000000000006`]& ,
        TagBoxNote->"0.31500000000000006"],
       StyleBox["0.31500000000000006`", {
         GrayLevel[0]}, StripOnInput -> False]],
      Annotation[#, 
       Style[0.31500000000000006`, {
         GrayLevel[0]}], "Tooltip"]& ], 
     TagBox[
      TooltipBox[
       TagBox[
        DynamicBox[{
          FEPrivate`If[
           CurrentValue["MouseOver"], 
           EdgeForm[{
             GrayLevel[0.5], 
             AbsoluteThickness[1.5], 
             Opacity[0.66]}], {}, {}], 
          
          RectangleBox[{1.2, 0.}, {1.4, 0.20000000000000004`}, 
           "RoundingRadius" -> 0]},
         ImageSizeCache->{{95.28775510204083, 103.54118738404455`}, {
          4.169193729800909, 40.27773676735786}}],
        StatusArea[#, 0.20000000000000004`]& ,
        TagBoxNote->"0.20000000000000004"],
       StyleBox["0.20000000000000004`", {
         GrayLevel[0]}, StripOnInput -> False]],
      Annotation[#, 
       Style[0.20000000000000004`, {
         GrayLevel[0]}], "Tooltip"]& ], 
     TagBox[
      TooltipBox[
       TagBox[
        DynamicBox[{
          FEPrivate`If[
           CurrentValue["MouseOver"], 
           EdgeForm[{
             GrayLevel[0.5], 
             AbsoluteThickness[1.5], 
             Opacity[0.66]}], {}, {}], 
          
          RectangleBox[{1.4, 0.}, {1.6, 0.12499999999999989`}, 
           "RoundingRadius" -> 0]},
         ImageSizeCache->{{102.54118738404455`, 110.79461966604826`}, {
          17.33489736888479, 40.27773676735786}}],
        StatusArea[#, 0.12499999999999989`]& ,
        TagBoxNote->"0.12499999999999989"],
       StyleBox["0.12499999999999989`", {
         GrayLevel[0]}, StripOnInput -> False]],
      Annotation[#, 
       Style[0.12499999999999989`, {
         GrayLevel[0]}], "Tooltip"]& ], 
     TagBox[
      TooltipBox[
       TagBox[
        DynamicBox[{
          FEPrivate`If[
           CurrentValue["MouseOver"], 
           EdgeForm[{
             GrayLevel[0.5], 
             AbsoluteThickness[1.5], 
             Opacity[0.66]}], {}, {}], 
          
          RectangleBox[{1.6, 0.}, {1.8, 0.11500000000000002`}, 
           "RoundingRadius" -> 0]},
         ImageSizeCache->{{109.79461966604826`, 118.04805194805198`}, {
          19.090324520762614`, 40.27773676735786}}],
        StatusArea[#, 0.11500000000000002`]& ,
        TagBoxNote->"0.11500000000000002"],
       StyleBox["0.11500000000000002`", {
         GrayLevel[0]}, StripOnInput -> False]],
      Annotation[#, 
       Style[0.11500000000000002`, {
         GrayLevel[0]}], "Tooltip"]& ], 
     TagBox[
      TooltipBox[
       TagBox[
        DynamicBox[{
          FEPrivate`If[
           CurrentValue["MouseOver"], 
           EdgeForm[{
             GrayLevel[0.5], 
             AbsoluteThickness[1.5], 
             Opacity[0.66]}], {}, {}], 
          
          RectangleBox[{1.8, 0.}, {2., 0.10500000000000002`}, 
           "RoundingRadius" -> 0]},
         ImageSizeCache->{{117.04805194805198`, 125.30148423005568`}, {
          20.84575167264046, 40.27773676735786}}],
        StatusArea[#, 0.10500000000000002`]& ,
        TagBoxNote->"0.10500000000000002"],
       StyleBox["0.10500000000000002`", {
         GrayLevel[0]}, StripOnInput -> False]],
      Annotation[#, 
       Style[0.10500000000000002`, {
         GrayLevel[0]}], "Tooltip"]& ], 
     TagBox[
      TooltipBox[
       TagBox[
        DynamicBox[{
          FEPrivate`If[
           CurrentValue["MouseOver"], 
           EdgeForm[{
             GrayLevel[0.5], 
             AbsoluteThickness[1.5], 
             Opacity[0.66]}], {}, {}], 
          
          RectangleBox[{2., 0.}, {2.2, 0.049999999999999954`}, 
           "RoundingRadius" -> 0]},
         ImageSizeCache->{{124.30148423005568`, 132.5549165120594}, {
          30.500601007968633`, 40.27773676735786}}],
        StatusArea[#, 0.049999999999999954`]& ,
        TagBoxNote->"0.049999999999999954"],
       StyleBox["0.049999999999999954`", {
         GrayLevel[0]}, StripOnInput -> False]],
      Annotation[#, 
       Style[0.049999999999999954`, {
         GrayLevel[0]}], "Tooltip"]& ], 
     TagBox[
      TooltipBox[
       TagBox[
        DynamicBox[{
          FEPrivate`If[
           CurrentValue["MouseOver"], 
           EdgeForm[{
             GrayLevel[0.5], 
             AbsoluteThickness[1.5], 
             Opacity[0.66]}], {}, {}], 
          
          RectangleBox[{2.2, 0.}, {2.4, 0.020000000000000028`}, 
           "RoundingRadius" -> 0]},
         ImageSizeCache->{{131.5549165120594, 139.8083487940631}, {
          35.76688246360216, 40.27773676735786}}],
        StatusArea[#, 0.020000000000000028`]& ,
        TagBoxNote->"0.020000000000000028"],
       StyleBox["0.020000000000000028`", {
         GrayLevel[0]}, StripOnInput -> False]],
      Annotation[#, 
       Style[0.020000000000000028`, {
         GrayLevel[0]}], "Tooltip"]& ], 
     TagBox[
      TooltipBox[
       TagBox[
        DynamicBox[{
          FEPrivate`If[
           CurrentValue["MouseOver"], 
           EdgeForm[{
             GrayLevel[0.5], 
             AbsoluteThickness[1.5], 
             Opacity[0.66]}], {}, {}], 
          
          RectangleBox[{2.4, 0.}, {2.6, 0.03499999999999997}, 
           "RoundingRadius" -> 0]},
         ImageSizeCache->{{138.8083487940631, 147.06178107606684`}, {
          33.1337417357854, 40.27773676735786}}],
        StatusArea[#, 0.03499999999999997]& ,
        TagBoxNote->"0.03499999999999997"],
       StyleBox["0.03499999999999997`", {
         GrayLevel[0]}, StripOnInput -> False]],
      Annotation[#, 
       Style[0.03499999999999997, {
         GrayLevel[0]}], "Tooltip"]& ], 
     TagBox[
      TooltipBox[
       TagBox[
        DynamicBox[{
          FEPrivate`If[
           CurrentValue["MouseOver"], 
           EdgeForm[{
             GrayLevel[0.5], 
             AbsoluteThickness[1.5], 
             Opacity[0.66]}], {}, {}], 
          
          RectangleBox[{2.6, 0.}, {2.8, 0.020000000000000028`}, 
           "RoundingRadius" -> 0]},
         ImageSizeCache->{{146.06178107606684`, 154.31521335807054`}, {
          35.76688246360216, 40.27773676735786}}],
        StatusArea[#, 0.020000000000000028`]& ,
        TagBoxNote->"0.020000000000000028"],
       StyleBox["0.020000000000000028`", {
         GrayLevel[0]}, StripOnInput -> False]],
      Annotation[#, 
       Style[0.020000000000000028`, {
         GrayLevel[0]}], "Tooltip"]& ], {}, {}, 
     TagBox[
      TooltipBox[
       TagBox[
        DynamicBox[{
          FEPrivate`If[
           CurrentValue["MouseOver"], 
           EdgeForm[{
             GrayLevel[0.5], 
             AbsoluteThickness[1.5], 
             Opacity[0.66]}], {}, {}], 
          
          RectangleBox[{3.2, 0.}, {3.4, 0.010000000000000014`}, 
           "RoundingRadius" -> 0]},
         ImageSizeCache->{{167.82207792207797`, 176.07551020408167`}, {
          37.52230961548001, 40.27773676735786}}],
        StatusArea[#, 0.010000000000000014`]& ,
        TagBoxNote->"0.010000000000000014"],
       StyleBox["0.010000000000000014`", {
         GrayLevel[0]}, StripOnInput -> False]],
      Annotation[#, 
       Style[0.010000000000000014`, {
         GrayLevel[0]}], "Tooltip"]& ]}, {}, {}}, {{}, {}, 
    {RGBColor[1, 0, 0], Thickness[Large], LineBox[CompressedData["
1:eJwV1nk4FW8bB3BlO5YoColUSqGfnUS6J4QsFSqlVGQpW/aske1Ysh1FtpCl
UNasxUPZI5JjO+dMKoqyVPaEd96/5vpcz1wzz9zXc3/v2Wt128RmMxMTk/Ym
Jqb/XyXXxX1/9cwi+9iv6cUhfXDf2GtGpmUWWQjaWOGRfTCT227lVDeLbFLU
9vDE90G5gZPBj9xZ1OZ6PNApvQ/UUqpExnxmUaaBcI50ZR/oK51Cg+Kz6AFF
Vjv3Wx842juzoDszKHyfxrj/qY9QPFATGyM2jTovU/h9efphImHBUkv5ByLb
9s6rVlDhxLGANif/7yjs09eGcJlB+CnR7K31YQxV3uJqV4wbAv358gn5qc9o
JecWz/bfwxBvOUyVdMVRRVzTpdGdNIhyHq2ov4Sj6LqDxUm7aRDi9z3BWBNH
xXLd8vriNPB6uGDkw48jF+5h9OwwDa50bGttr2Sg1XLVVi2ggaS8fpXdXzpq
YWxemb1BgzdMr5LyQmjI0m4r944iGixkppvtTRpC4+fupCvL0eFP/qb4fYFD
aA/vtIiQEh1mXti2i98cQvbvhu8tHaHDt1dyRyXUhtBHBeu550CHgcFmYWl8
EE18ijD6c5oOVbzTDCWJQTQb2ZMz7kgH70ANa70qKrK6Ja5hk08Hj/An6acy
qagwvjv9cwEdXGLYqfoRVJRdaVBp/oION9M+nDQyp6Kqn8lcWi/pcKna5qDJ
ej+SP+srOdVEB7XZmB+XdfpROY8TTyadDv+u4i63B/qQaJzX8F1eBrjWU43+
NfahRGYB7DIfA77t6paKLOpDKidCCpV2MKB36NXYk6A+xNTPTcaFGZBnknJx
QKoPPRlf0uaRYMBpnXOYRuAHlJ/PR2KoMyDzv05erkO96Dt/oZ+8DQO232+a
SubrRZ2/LTKf2zGAj+qaeOpvD+ITGSjcb8+A76fPsAfk9qB///V5ct1mQLwm
5+zX5ffo1qxdS4U3Az5LBjWUZXcjgdWaFOdoBoQuO1gYzXWiimJKI7mYAamK
J3PamzvR49LJM+9KGVDivHtSK6kTDTZoinFVMGBkrNdD7Wgnamf28Q2sZoDc
B+X7BwM7UOJaNad6IwPoBRt1m7jaEVsqUP/1MuDP2BBTAL0Naa24rDN/ZAD7
nnKd5RdtqDBU05ydygCFJOu+mbNtqO1QVv7mYQZEBHdM0pJa0ZxvjP+7UQYo
XU4UqtrXgqw4j3nkzTJAP8nxqvx8M1qdZHa8/ZsB1z6czH3e0owilwKilOcY
EKW7LJtzsxmF1TSqVi0yYFTRQje++C06T7uiG7nGgPvcB73s1d6gQ1mVjzw5
cXiiy/R6nOsN4l06VbSJG4fq4OFNlowmJCi9eyRyCw5flqNjzIKaULr07qjY
rTiojv/K1W5tRNmmK/tuCeDQ6uf5VCO3AYWT12qs9uJAPePaYmXUgEY8MyF1
Hw5j4k5fyYv16PiRF+bvxXHY3GUt1qdXj6brbQwPS+CgIXIu2Xb6FWqZX+dD
Uji8fK1AjlOpRf8x9AZAEYe38TJ5Lz/VIGG/BA9MCYc+a6m3wxE1qOaNYraG
Mg6/uPdtiNOqkSddMve/Izgctth2pyawCu058Fh3TA2HnLVZ28/tFShZLLmh
6AQO5R9+hrK5VaBzq0oinpo4NOZ9fyK9qwI1Dusyq2vhgBt9wj2dypHQ4mvH
19o4CD9+f4GTrww1VNXmPdTFQdKt00P2VSnSud6lq6VH1EOnlXLOuhQZl5m6
ThMuovtlaDiVoJy3A5eO6OOwL8A8xrbqOZoYjfXNNsQBO2DSUUp6jqxzd5so
GuFwtfsU66p5EaJvv13SRDhF9GhA3EYB0qiL/fvxNA5VLXK1w8YFKMJw47vZ
GRz6nQ4tiOc+Q5067uZDhLc2CDrV6D1F7haV3e/P4iBjy1uwOS0fPXWWea9t
jIMhD/u44XQeSv3AZFpNOOLq4pXPCbloR+H2yQQTHDbW+w05adlopvK3pug5
HOTvvL6M5LJRtO1YrifhG7M59h7hWaiMA3/eSTgOjW+tv/EYZTt5Fdufx2G+
5KXMQd4MVCI0+rSS8MWsEMOEujQ0FfVNZ43wnnt7I2y2paCj/Q6ewRdwCHX9
ldfzOhlV+jawNBKesERvj95MQqbk80f/EjY0if2cw/8QdRaI8sub4VCqabGx
BSWiqmIWijXh7YqHRb3tKSglcKXyAWFv8VW1LzsS0M99U3ebCNcefFD2KDgO
HXjdPfWDcDBJYdpx+D5qio7f2HoRB/3JnkMn5KKQ/uzBakXC/J1O1jvIZHTV
MoLflDC9kCtrkhGKqnyebr9NODe6gFavFIwkp/3qyISHhmJl9gsGoqqYNZYM
wrdSO/O6mn1QhP6B+WLCfy+ziXq6eaBF2WlyPeHt6qGH1DFnRLY729hOWKVh
VeB+2g20eMQgo5dwcfA3zrppY6TCSttNJexdY9q0hyEF7mfXtQcId7moq2/O
uwjcNS95+wmfVJaQlVuyhcdL63ffE+a7ZJY12+0Cq99xSgvhe8m7JHX3eMHh
QDODWsK/qJ/KHrv5gVuv47MCwnKcJqJ+VkFwnHt/YRJhO8mXn/O5QkApKORs
EOHHegL5fS/DYH8yJc2W8ICdt/26RQSwDV+KPkWYhzwiI8UeDcs/vohJ/n8/
+cfmzpfGgHKc6iVWwg1WPbTGw/HAk2d7BCfqb+7G1HDhXgLMHvKtqyC8eE8+
a4pKAW2le1/CCFMSrIKDpR5A6FpUyTnCnaXNOsUfk8BLaVJ1gjgPto0Lh7QP
PYK+TMuF54Q390pwjfinwMXm/WbOhI/ORvSwSqRDmusnzu/Eeetfry1L882A
hzizxWPCLjw/E+V7HsPe7q5zJv9fT5J4LvM6C4oVypVLifP8ZDdk+Nhmw7Ib
i5g5YZd8s9jmrU/gRJttMRNh7qoIF3ObHDhw5yPSMsVhRCPbMo83FxQDRTRH
if551lJr8qs2F9bn/tj7ENam/lAK58mHpPIPpdlEv/nNG66UVT8DWqTZtk6i
X9NDXtV5uRSAxLPyCYzwaz4pf3XJQtCncPdWEP2+Lse+0ZxaBIk5qT0xRB4E
OzWxDAUUAz06/BGfAbHff7Kt6SolELdLafEWkS9vox+TLWdLQJil2ab+FFEf
qkaYZ2kppNan+F0k8kh7Tpi0sbccgn1HIq+dJL5fre+m5d1yiH1jdDeZyDeb
4MiOtyPl0DowEf6OyL/wbctREYkVsLRzgHqQyMc2WeoWPtZK4NPLci87TvSP
Yxz/gYlqSK+I/Eci8nakQseDrF0DhbjAlc0qODisrvVPZtVAybgvdYnI6/tR
jknFl2phUv+m4KACDt3PTu1UfVcHzkqUr7YyOJwdZxYzKK4H0+VPG+/24zB6
+FXgC84GoJzvkSMT88LVw22U164BRNdag48T84TC/Dm7XwzB7FRPdLYYDj+2
XuMr0WuExtUs7KgwDtKyxx6q1jVBo9bY8xIeoh773f1NxptgXGaTpC4xz4aE
Cm84bn0Dv23pRjQuIo82Cylk2b6B/KK6xBUS0f/Uuffs/G9B16BHlIcZh1j/
ItKgQzNQF5vSk4l5+qdD2N9TtA3OjY5zytEY0N5gfCNOrw20JZyeWBDz+3FF
hH6Bexs4pCrOkwcZYJCxKMToaIMLAkvtfcT8z3Xpq9T2aoe44q96J7oYcEEw
apa/pwMOmskqBL9mQK31ilV5UBdQbXgtH6UywJp2ssG+pAtk2hUjZR4xgMeE
slMc7wKhe2V/Gh8S6yDdm6jeDfuvR8eMxDNgq7CFhudiN1wZa18bDWfAzZ5G
QVXHHmh/FJGn7sYALlHtb7j0B/jttllbVocB1y/+4FfX+gAs+TnRZ7QYUJUY
fyLZ/AM0cg0pO2LEOicj/UzkBzCY8a9KVSPWlzxNG799gLVpSkSXDAOs+p42
Zmf3weuxfe6hAgyoI3OlWQv2g6ddJTL6SoctHabOpjL9sFFKluMepRPPSz+h
ebIfZHNS9NuI/0e2mMMTYu79IKhwJVBxgA4mFCNl2vt+WHhfu2WsnQ6T6XE9
xuFU+FsS0J9P/J8KlfOzwMIAWMWOrrzyoIMXfZfTzo/DsGPGquD2Cg3Otn5I
858Yhv82xF88W6CBVCm589PaMBTpxWz/9JsGeMjcwfxDI5BoeYBf+wcNdKTf
fVG4OwLs5//7PTNCgx2+vhcNJWnAOLOAdbyiwUvBYa3AQDoEPREaeORHgzmT
JOFx6U/wwOGs3M+5EWAJvp4H/l+gQYW1eaV4GJx/SkkLyY/DKn+Jrsi1Idjv
ZvRnUGwCDvq90xAZHICImlKnKaafsH9X9CTlBBW6eTsu3C+dhvUNzcacVx9B
9k16I4fRL6g8LentyNsHE6El3PwFv0HwuMbF+qBe8H76pVxn/Q9k+vVsimR6
D8VDTStOLvMQzNQlsH+wE85R/JZGOhZgk0PMxQDmdsgy//xQHVuCO9UCInai
LdAcbBfbWrIMUW5/+UM838BFxXaNZLa/ELP9wdZwqUag/OmPuey7ChQWv7+Z
ba/gmNrrBzbUf+BaOX8nYaYa6j+/XUiRXIe/O0flayJewljynaWm1A1Y5i3U
vXqiDDR+bh5K9mHCorYxz+rseQG0bVmhekqbMF8qP48b5zMo8y/kTv+yCXui
tEPD4U8OuAu2NIXGbMboA2JfHt7NgohaZd3vOszY5WOro1xPUmGvfO2VtFPM
mOzjqfM37qdC3bNjbmcMmbHlmyJWHV6pMJ2slVFtzIxlcd/e32SQCiZexnPk
K8yYd1hM7NXFFNil6JR5yI0Zq1C4PaF/OgWKn+cu3cpgxhTEQy7Vkh4BNZP/
2dQcM6b3/td5p/yHcP/2UO7IIjOWrzxYXUx5CJqQkd2+woz1GAce3bj7EEpw
ibTcDWbs0pkln26zhxApqhZ7hYsFu3RMeNNuzoegkXbdo2sfC3blv4w6susD
yEt6gT03ZsHE306/1zRIBM8Y3WGHEhYs6IhXatihBJjKaJQyLGfBsvZqhT0S
SgCr4qP+hytZsJKWYpYaUgKc7pEWm65jwVS4uXNEJuNBYttWG+dWFuzXyfVI
v8J4GHow9MuFwYKZ6bLfipSNB7X0WyQvLlZMRUnrlFVYLDAVRR8JsmPFKtpe
8x0WjAayi9t2WXtWrKfoium39SjgUbn0m+HIimU21xlWfIsCkSaJInU3VsxM
2vhYWFUUqA427V7yZ8VMG0X72M2iwJV5mdWZwoo5ujetB6RFwpfL1v2X61kx
Yabya6PyEdDMreF6hJ8Nm7l8K0yYEgY3pWyU9wiwYWVfhavu+ofBFr2YFdJO
Niyo4m31gm0YXAhh3BvZzYbZ2J9e2XUsDL6vBFACpNiw6No+Vr3vocD5vaHi
7Qk2rMttlsZ3IhTONGGLZ1zYsH+Ujps7N4fAiKe2381uNszr+Lesw9VBEKc3
0inWy4apZQlFsz0NAu1dLsKDfWyYpGHJgZmkIChpTKs9OcSG8b/9ffKbVxCE
cs0t7fvKhnWM75H0ORIEMtnZHoxlNoxHXuGOplcgBHetO5nsZ8fmdqlvc38U
AJLitdeP+bFjHbp+3vsO+sJjFWbWbwHs2AUnG0bWFl/g0zcqiAtix55yuGsq
z/vAP5fPv7+EsWP2KLM7s8kHehs4QiLj2TEV1jQZXwsf8DY3zxvIY8ee32m9
qZLiDa0Jq5MuvewY26aRget77sANpuPueQdImCUX+ZuqvQcsOnSsRxwiYSwT
R84zn/OAiMFzUY7SJKxfIeDUVw0PeF7skK0oT8J2BBzw/7rNA+avpPQ0HSNh
k7qzol/r3CG8dl4aNyVhrL2LAr1b3aHQrWhMIJiEpW4Pl5185wrHcBWXv6Ek
DJQFvmvWuML7U29WGWQS9u/6K+bSXFeY2zPMlxdDwvawrkk1BxD397CdUEwh
YVdHyuz75FyhW9oq40wpCSv60pnmneoCv8aELkTgJKyMM/PR30VnSGTlc1L7
TMIs6iXsUgecQUWCK3TqKwm7pjhCuVjlDP52a2VnJ0lY2xRrlqyXM5B+fN6y
c56EdXtRJk4uOcHe2cKWAg4OzP3HHdH1dUcwWTmm/E6JA5PQNQtfEneAhZ0q
hgFHODAmJ8VGJXYHeKQme0NWjQM7d4vbNuaHPeB+exMSgQNrGXJPiy+zB4d1
1unL+hyYxe1QVTpmD6HMPblT1ziwa7ofWz2sb0HlFsvtPNEc2L2Xq8F8yA7+
nNZx2hrLgY2qb95If2IHMvHSrXwJHBhvhrmmTrgdPONbvCOYzIFNOPgUzRvZ
QapgNG1PDgf2i+z1tR63haC9lU8UX3Fgh0e0tk2y2oKBMof8pZ8cWGbn2tKn
W9YQ4TUTeXmGA3u3s/4ft7E1NFd//GLxmwPLpUicP69qDcfVHj+wWuLA+FJX
ko+wW4McKC47MHNiFT57oyH/BgicskB3d3FiO5NISTqTVmAapSl0bzcnJnbn
SZlInxXEvzvoGrKXE7Ml7wznfmUFHKf/7Is4yIlFyhcsqMZYwT8TcniCIidG
Nv5+5I2CFag+cPyUqMKJOYp7W90QsQJPqrFq0lFO7Hrz30BRNiuYMRP5kQqc
2Mjlcwr0YUuQTtmklaHJiQm5pSWNvrWEmyPf0jJPEu8Lfnl/7YUl5O7qms/W
48QOCUf8U3hkCZ+vlBnlGnBiUTv53gcFW8L/AAJe/hE=
      "]]}}},
  AspectRatio->NCache[GoldenRatio^(-1), 0.6180339887498948],
  Axes->{True, True},
  AxesOrigin->{-1., 0},
  FrameTicks->{{Automatic, Automatic}, {Automatic, Automatic}},
  GridLines->{None, None},
  PlotRange->NCache[{{-1, 
      Rational[17, 5]}, {All, All}}, {{-1, 3.4}, {All, All}}],
  PlotRangePadding->{{
     Scaled[0.02], 
     Scaled[0.02]}, {
     Scaled[0.02], 
     Scaled[0.1]}},
  Ticks->{Automatic, Automatic}]], "Output",
 ImageSize->{184, 117},
 ImageMargins->{{0, 0}, {0, 0}},
 ImageRegion->{{0, 1}, {0, 1}},
 CellLabel->"Out[8]=",
 CellID->158480250]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"Table", "[", 
  RowBox[{
   RowBox[{
    RowBox[{"data", "=", 
     RowBox[{"res", "[", 
      RowBox[{"[", 
       RowBox[{
        RowBox[{";;", ";;", "10"}], ",", "1", ",", 
        RowBox[{"-", "1"}], ",", "i"}], "]"}], "]"}]}], ";", 
    "\[IndentingNewLine]", 
    RowBox[{"Show", "[", 
     RowBox[{
      RowBox[{"Histogram", "[", 
       RowBox[{"data", ",", "Automatic", ",", "\"\<ProbabilityDensity\>\""}], 
       "]"}], ",", 
      RowBox[{"Plot", "[", 
       RowBox[{
        RowBox[{"PDF", "[", 
         RowBox[{
          RowBox[{"NormalDistribution", "[", 
           RowBox[{"0", ",", "1"}], "]"}], ",", "x"}], "]"}], ",", 
        RowBox[{"{", 
         RowBox[{"x", ",", 
          RowBox[{"Min", "[", "data", "]"}], ",", 
          RowBox[{"Max", "[", "data", "]"}]}], "}"}], ",", 
        RowBox[{"PlotStyle", "\[Rule]", 
         RowBox[{"{", 
          RowBox[{"Red", ",", "Thick"}], "}"}]}], ",", 
        RowBox[{"PlotRange", "\[Rule]", "All"}], ",", 
        RowBox[{"AxesLabel", "\[Rule]", 
         RowBox[{"{", 
          RowBox[{
           RowBox[{"\"\<x3\>\"", "<>", 
            RowBox[{"ToString", "[", "i", "]"}]}], ",", "\"\<freq\>\""}], 
          "}"}]}]}], "]"}]}], "]"}]}], ",", 
   RowBox[{"{", 
    RowBox[{"i", ",", "2"}], "}"}]}], "]"}]], "Input",
 CellLabel->"In[9]:=",
 CellID->271797975],

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{
   GraphicsBox[{
     {RGBColor[0.798413061722744, 0.824719615472648, 0.968322270542458], 
      EdgeForm[Opacity[0.7]], {}, 
      {RGBColor[0.798413061722744, 0.824719615472648, 0.968322270542458], 
       EdgeForm[Opacity[0.7]], 
       TagBox[
        TooltipBox[
         TagBox[
          DynamicBox[{
            FEPrivate`If[
             CurrentValue["MouseOver"], 
             EdgeForm[{
               GrayLevel[0.5], 
               AbsoluteThickness[1.5], 
               Opacity[0.66]}], {}, {}], 
            RectangleBox[{-3.5, 0.}, {-3., 0.004}, "RoundingRadius" -> 0]},
           
           ImageSizeCache->{{15.499999999999986`, 28.775039246467813`}, {
            38.400023191418924`, 40.27773676735785}}],
          StatusArea[#, 0.004]& ,
          TagBoxNote->"0.004"],
         StyleBox["0.004`", {
           GrayLevel[0]}, StripOnInput -> False]],
        Annotation[#, 
         Style[0.004, {
           GrayLevel[0]}], "Tooltip"]& ], 
       TagBox[
        TooltipBox[
         TagBox[
          DynamicBox[{
            FEPrivate`If[
             CurrentValue["MouseOver"], 
             EdgeForm[{
               GrayLevel[0.5], 
               AbsoluteThickness[1.5], 
               Opacity[0.66]}], {}, {}], 
            RectangleBox[{-3., 0.}, {-2.5, 0.004}, "RoundingRadius" -> 0]},
           
           ImageSizeCache->{{27.775039246467813`, 41.050078492935626`}, {
            38.400023191418924`, 40.27773676735785}}],
          StatusArea[#, 0.004]& ,
          TagBoxNote->"0.004"],
         StyleBox["0.004`", {
           GrayLevel[0]}, StripOnInput -> False]],
        Annotation[#, 
         Style[0.004, {
           GrayLevel[0]}], "Tooltip"]& ], 
       TagBox[
        TooltipBox[
         TagBox[
          DynamicBox[{
            FEPrivate`If[
             CurrentValue["MouseOver"], 
             EdgeForm[{
               GrayLevel[0.5], 
               AbsoluteThickness[1.5], 
               Opacity[0.66]}], {}, {}], 
            RectangleBox[{-2.5, 0.}, {-2., 0.03}, "RoundingRadius" -> 0]},
           
           ImageSizeCache->{{40.050078492935626`, 53.325117739403446`}, {
            32.69488494781593, 40.27773676735785}}],
          StatusArea[#, 0.03]& ,
          TagBoxNote->"0.03"],
         StyleBox["0.03`", {
           GrayLevel[0]}, StripOnInput -> False]],
        Annotation[#, 
         Style[0.03, {
           GrayLevel[0]}], "Tooltip"]& ], 
       TagBox[
        TooltipBox[
         TagBox[
          DynamicBox[{
            FEPrivate`If[
             CurrentValue["MouseOver"], 
             EdgeForm[{
               GrayLevel[0.5], 
               AbsoluteThickness[1.5], 
               Opacity[0.66]}], {}, {}], 
            RectangleBox[{-2., 0.}, {-1.5, 0.08}, "RoundingRadius" -> 0]},
           
           ImageSizeCache->{{52.325117739403446`, 65.60015698587127}, {
            21.723465248579387`, 40.27773676735785}}],
          StatusArea[#, 0.08]& ,
          TagBoxNote->"0.08"],
         StyleBox["0.08`", {
           GrayLevel[0]}, StripOnInput -> False]],
        Annotation[#, 
         Style[0.08, {
           GrayLevel[0]}], "Tooltip"]& ], 
       TagBox[
        TooltipBox[
         TagBox[
          DynamicBox[{
            FEPrivate`If[
             CurrentValue["MouseOver"], 
             EdgeForm[{
               GrayLevel[0.5], 
               AbsoluteThickness[1.5], 
               Opacity[0.66]}], {}, {}], 
            RectangleBox[{-1.5, 0.}, {-1., 0.176}, "RoundingRadius" -> 0]},
           
           ImageSizeCache->{{64.60015698587127, 77.87519623233908}, {
            0.6583394260452309, 40.27773676735785}}],
          StatusArea[#, 0.176]& ,
          TagBoxNote->"0.176"],
         StyleBox["0.176`", {
           GrayLevel[0]}, StripOnInput -> False]],
        Annotation[#, 
         Style[0.176, {
           GrayLevel[0]}], "Tooltip"]& ], 
       TagBox[
        TooltipBox[
         TagBox[
          DynamicBox[{
            FEPrivate`If[
             CurrentValue["MouseOver"], 
             EdgeForm[{
               GrayLevel[0.5], 
               AbsoluteThickness[1.5], 
               Opacity[0.66]}], {}, {}], 
            RectangleBox[{-1., 0.}, {-0.5, 0.268}, "RoundingRadius" -> 0]},
           
           ImageSizeCache->{{76.87519623233908, 
            90.1502354788069}, {-19.529072820550006`, 40.27773676735785}}],
          StatusArea[#, 0.268]& ,
          TagBoxNote->"0.268"],
         StyleBox["0.268`", {
           GrayLevel[0]}, StripOnInput -> False]],
        Annotation[#, 
         Style[0.268, {
           GrayLevel[0]}], "Tooltip"]& ], 
       TagBox[
        TooltipBox[
         TagBox[
          DynamicBox[{
            FEPrivate`If[
             CurrentValue["MouseOver"], 
             EdgeForm[{
               GrayLevel[0.5], 
               AbsoluteThickness[1.5], 
               Opacity[0.66]}], {}, {}], 
            RectangleBox[{-0.5, 0.}, {0., 0.412}, "RoundingRadius" -> 0]},
           
           ImageSizeCache->{{89.1502354788069, 
            102.42527472527472`}, {-51.12676155435123, 40.27773676735785}}],
          StatusArea[#, 0.412]& ,
          TagBoxNote->"0.412"],
         StyleBox["0.412`", {
           GrayLevel[0]}, StripOnInput -> False]],
        Annotation[#, 
         Style[0.412, {
           GrayLevel[0]}], "Tooltip"]& ], 
       TagBox[
        TooltipBox[
         TagBox[
          DynamicBox[{
            FEPrivate`If[
             CurrentValue["MouseOver"], 
             EdgeForm[{
               GrayLevel[0.5], 
               AbsoluteThickness[1.5], 
               Opacity[0.66]}], {}, {}], 
            RectangleBox[{0., 0.}, {0.5, 0.384}, "RoundingRadius" -> 0]},
           
           ImageSizeCache->{{101.42527472527472`, 
            114.70031397174253`}, {-44.982766522778775`, 40.27773676735785}}],
          StatusArea[#, 0.384]& ,
          TagBoxNote->"0.384"],
         StyleBox["0.384`", {
           GrayLevel[0]}, StripOnInput -> False]],
        Annotation[#, 
         Style[0.384, {
           GrayLevel[0]}], "Tooltip"]& ], 
       TagBox[
        TooltipBox[
         TagBox[
          DynamicBox[{
            FEPrivate`If[
             CurrentValue["MouseOver"], 
             EdgeForm[{
               GrayLevel[0.5], 
               AbsoluteThickness[1.5], 
               Opacity[0.66]}], {}, {}], 
            RectangleBox[{0.5, 0.}, {1., 0.258}, "RoundingRadius" -> 0]},
           
           ImageSizeCache->{{113.70031397174253`, 
            126.97535321821036`}, {-17.334788880702696`, 40.27773676735785}}],
          StatusArea[#, 0.258]& ,
          TagBoxNote->"0.258"],
         StyleBox["0.258`", {
           GrayLevel[0]}, StripOnInput -> False]],
        Annotation[#, 
         Style[0.258, {
           GrayLevel[0]}], "Tooltip"]& ], 
       TagBox[
        TooltipBox[
         TagBox[
          DynamicBox[{
            FEPrivate`If[
             CurrentValue["MouseOver"], 
             EdgeForm[{
               GrayLevel[0.5], 
               AbsoluteThickness[1.5], 
               Opacity[0.66]}], {}, {}], 
            RectangleBox[{1., 0.}, {1.5, 0.202}, "RoundingRadius" -> 0]},
           
           ImageSizeCache->{{125.97535321821036`, 
            139.25039246467816`}, {-5.046798817557772, 40.27773676735785}}],
          StatusArea[#, 0.202]& ,
          TagBoxNote->"0.202"],
         StyleBox["0.202`", {
           GrayLevel[0]}, StripOnInput -> False]],
        Annotation[#, 
         Style[0.202, {
           GrayLevel[0]}], "Tooltip"]& ], 
       TagBox[
        TooltipBox[
         TagBox[
          DynamicBox[{
            FEPrivate`If[
             CurrentValue["MouseOver"], 
             EdgeForm[{
               GrayLevel[0.5], 
               AbsoluteThickness[1.5], 
               Opacity[0.66]}], {}, {}], 
            RectangleBox[{1.5, 0.}, {2., 0.13}, "RoundingRadius" -> 0]},
           
           ImageSizeCache->{{138.25039246467816`, 151.52543171114598`}, {
            10.752045549342846`, 40.27773676735785}}],
          StatusArea[#, 0.13]& ,
          TagBoxNote->"0.13"],
         StyleBox["0.13`", {
           GrayLevel[0]}, StripOnInput -> False]],
        Annotation[#, 
         Style[0.13, {
           GrayLevel[0]}], "Tooltip"]& ], 
       TagBox[
        TooltipBox[
         TagBox[
          DynamicBox[{
            FEPrivate`If[
             CurrentValue["MouseOver"], 
             EdgeForm[{
               GrayLevel[0.5], 
               AbsoluteThickness[1.5], 
               Opacity[0.66]}], {}, {}], 
            RectangleBox[{2., 0.}, {2.5, 0.04}, "RoundingRadius" -> 0]},
           
           ImageSizeCache->{{150.52543171114598`, 163.8004709576138}, {
            30.50060100796862, 40.27773676735785}}],
          StatusArea[#, 0.04]& ,
          TagBoxNote->"0.04"],
         StyleBox["0.04`", {
           GrayLevel[0]}, StripOnInput -> False]],
        Annotation[#, 
         Style[0.04, {
           GrayLevel[0]}], "Tooltip"]& ], 
       TagBox[
        TooltipBox[
         TagBox[
          DynamicBox[{
            FEPrivate`If[
             CurrentValue["MouseOver"], 
             EdgeForm[{
               GrayLevel[0.5], 
               AbsoluteThickness[1.5], 
               Opacity[0.66]}], {}, {}], 
            RectangleBox[{2.5, 0.}, {3., 0.012}, "RoundingRadius" -> 0]},
           
           ImageSizeCache->{{162.8004709576138, 176.07551020408164`}, {
            36.644596039541085`, 40.27773676735785}}],
          StatusArea[#, 0.012]& ,
          TagBoxNote->"0.012"],
         StyleBox["0.012`", {
           GrayLevel[0]}, StripOnInput -> False]],
        Annotation[#, 
         Style[0.012, {
           GrayLevel[0]}], "Tooltip"]& ]}, {}, {}}, {{}, {}, 
      {RGBColor[1, 0, 0], Thickness[Large], LineBox[CompressedData["
1:eJwV1nc8VW8cB/DrLiMhoYGkIbOQkNT3KCMjpEFZEcrKioxkZsseGWXLFiGS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        "]]}}},
    AspectRatio->NCache[GoldenRatio^(-1), 0.6180339887498948],
    Axes->{True, True},
    AxesOrigin->{-3.5, 0.},
    FrameTicks->{{Automatic, Automatic}, {Automatic, Automatic}},
    GridLines->{None, None},
    PlotRange->NCache[{{
        Rational[-7, 2], 3}, {All, All}}, {{-3.5, 3}, {All, All}}],
    PlotRangePadding->{{
       Scaled[0.02], 
       Scaled[0.02]}, {
       Scaled[0.02], 
       Scaled[0.1]}},
    Ticks->{Automatic, Automatic}], ",", 
   GraphicsBox[{
     {RGBColor[0.798413061722744, 0.824719615472648, 0.968322270542458], 
      EdgeForm[Opacity[0.7]], {}, 
      {RGBColor[0.798413061722744, 0.824719615472648, 0.968322270542458], 
       EdgeForm[Opacity[0.7]], 
       TagBox[
        TooltipBox[
         TagBox[
          DynamicBox[{
            FEPrivate`If[
             CurrentValue["MouseOver"], 
             EdgeForm[{
               GrayLevel[0.5], 
               AbsoluteThickness[1.5], 
               Opacity[0.66]}], {}, {}], 
            RectangleBox[{-3., 0.}, {-2.5, 0.006}, "RoundingRadius" -> 0]},
           
           ImageSizeCache->{{15.500000000000014`, 28.775039246467834`}, {
            37.93509570317408, 40.27773676735788}}],
          StatusArea[#, 0.006]& ,
          TagBoxNote->"0.006"],
         StyleBox["0.006`", {
           GrayLevel[0]}, StripOnInput -> False]],
        Annotation[#, 
         Style[0.006, {
           GrayLevel[0]}], "Tooltip"]& ], 
       TagBox[
        TooltipBox[
         TagBox[
          DynamicBox[{
            FEPrivate`If[
             CurrentValue["MouseOver"], 
             EdgeForm[{
               GrayLevel[0.5], 
               AbsoluteThickness[1.5], 
               Opacity[0.66]}], {}, {}], 
            RectangleBox[{-2.5, 0.}, {-2., 0.036}, "RoundingRadius" -> 0]},
           
           ImageSizeCache->{{27.775039246467834`, 41.050078492935654`}, {
            31.221890382255086`, 40.27773676735788}}],
          StatusArea[#, 0.036]& ,
          TagBoxNote->"0.036"],
         StyleBox["0.036`", {
           GrayLevel[0]}, StripOnInput -> False]],
        Annotation[#, 
         Style[0.036, {
           GrayLevel[0]}], "Tooltip"]& ], 
       TagBox[
        TooltipBox[
         TagBox[
          DynamicBox[{
            FEPrivate`If[
             CurrentValue["MouseOver"], 
             EdgeForm[{
               GrayLevel[0.5], 
               AbsoluteThickness[1.5], 
               Opacity[0.66]}], {}, {}], 
            RectangleBox[{-2., 0.}, {-1.5, 0.07}, "RoundingRadius" -> 0]},
           
           ImageSizeCache->{{40.050078492935654`, 53.325117739403474`}, {
            23.61359101854689, 40.27773676735788}}],
          StatusArea[#, 0.07]& ,
          TagBoxNote->"0.07"],
         StyleBox["0.07`", {
           GrayLevel[0]}, StripOnInput -> False]],
        Annotation[#, 
         Style[0.07, {
           GrayLevel[0]}], "Tooltip"]& ], 
       TagBox[
        TooltipBox[
         TagBox[
          DynamicBox[{
            FEPrivate`If[
             CurrentValue["MouseOver"], 
             EdgeForm[{
               GrayLevel[0.5], 
               AbsoluteThickness[1.5], 
               Opacity[0.66]}], {}, {}], 
            RectangleBox[{-1.5, 0.}, {-1., 0.18}, "RoundingRadius" -> 0]},
           
           ImageSizeCache->{{52.325117739403474`, 
            65.6001569858713}, {-1.0014951581560894`, 40.27773676735788}}],
          StatusArea[#, 0.18]& ,
          TagBoxNote->"0.18"],
         StyleBox["0.18`", {
           GrayLevel[0]}, StripOnInput -> False]],
        Annotation[#, 
         Style[0.18, {
           GrayLevel[0]}], "Tooltip"]& ], 
       TagBox[
        TooltipBox[
         TagBox[
          DynamicBox[{
            FEPrivate`If[
             CurrentValue["MouseOver"], 
             EdgeForm[{
               GrayLevel[0.5], 
               AbsoluteThickness[1.5], 
               Opacity[0.66]}], {}, {}], 
            RectangleBox[{-1., 0.}, {-0.5, 0.306}, "RoundingRadius" -> 0]},
           ImageSizeCache->{{64.6001569858713, 
            77.8751962323391}, {-29.196957506015863`, 40.27773676735788}}],
          StatusArea[#, 0.306]& ,
          TagBoxNote->"0.306"],
         StyleBox["0.306`", {
           GrayLevel[0]}, StripOnInput -> False]],
        Annotation[#, 
         Style[0.306, {
           GrayLevel[0]}], "Tooltip"]& ], 
       TagBox[
        TooltipBox[
         TagBox[
          DynamicBox[{
            FEPrivate`If[
             CurrentValue["MouseOver"], 
             EdgeForm[{
               GrayLevel[0.5], 
               AbsoluteThickness[1.5], 
               Opacity[0.66]}], {}, {}], 
            RectangleBox[{-0.5, 0.}, {0., 0.36}, "RoundingRadius" -> 0]},
           
           ImageSizeCache->{{76.8751962323391, 
            90.15023547880693}, {-41.28072708367006, 40.27773676735788}}],
          StatusArea[#, 0.36]& ,
          TagBoxNote->"0.36"],
         StyleBox["0.36`", {
           GrayLevel[0]}, StripOnInput -> False]],
        Annotation[#, 
         Style[0.36, {
           GrayLevel[0]}], "Tooltip"]& ], 
       TagBox[
        TooltipBox[
         TagBox[
          DynamicBox[{
            FEPrivate`If[
             CurrentValue["MouseOver"], 
             EdgeForm[{
               GrayLevel[0.5], 
               AbsoluteThickness[1.5], 
               Opacity[0.66]}], {}, {}], 
            RectangleBox[{0., 0.}, {0.5, 0.404}, "RoundingRadius" -> 0]},
           
           ImageSizeCache->{{89.15023547880693, 
            102.42527472527476`}, {-51.12676155435126, 40.27773676735788}}],
          StatusArea[#, 0.404]& ,
          TagBoxNote->"0.404"],
         StyleBox["0.404`", {
           GrayLevel[0]}, StripOnInput -> False]],
        Annotation[#, 
         Style[0.404, {
           GrayLevel[0]}], "Tooltip"]& ], 
       TagBox[
        TooltipBox[
         TagBox[
          DynamicBox[{
            FEPrivate`If[
             CurrentValue["MouseOver"], 
             EdgeForm[{
               GrayLevel[0.5], 
               AbsoluteThickness[1.5], 
               Opacity[0.66]}], {}, {}], 
            RectangleBox[{0.5, 0.}, {1., 0.246}, "RoundingRadius" -> 0]},
           
           ImageSizeCache->{{101.42527472527476`, 
            114.70031397174257`}, {-15.770546864177874`, 40.27773676735788}}],
          StatusArea[#, 0.246]& ,
          TagBoxNote->"0.246"],
         StyleBox["0.246`", {
           GrayLevel[0]}, StripOnInput -> False]],
        Annotation[#, 
         Style[0.246, {
           GrayLevel[0]}], "Tooltip"]& ], 
       TagBox[
        TooltipBox[
         TagBox[
          DynamicBox[{
            FEPrivate`If[
             CurrentValue["MouseOver"], 
             EdgeForm[{
               GrayLevel[0.5], 
               AbsoluteThickness[1.5], 
               Opacity[0.66]}], {}, {}], 
            RectangleBox[{1., 0.}, {1.5, 0.22}, "RoundingRadius" -> 0]},
           
           ImageSizeCache->{{113.70031397174257`, 
            126.97535321821039`}, {-9.952435586048082, 40.27773676735788}}],
          StatusArea[#, 0.22]& ,
          TagBoxNote->"0.22"],
         StyleBox["0.22`", {
           GrayLevel[0]}, StripOnInput -> False]],
        Annotation[#, 
         Style[0.22, {
           GrayLevel[0]}], "Tooltip"]& ], 
       TagBox[
        TooltipBox[
         TagBox[
          DynamicBox[{
            FEPrivate`If[
             CurrentValue["MouseOver"], 
             EdgeForm[{
               GrayLevel[0.5], 
               AbsoluteThickness[1.5], 
               Opacity[0.66]}], {}, {}], 
            RectangleBox[{1.5, 0.}, {2., 0.11}, "RoundingRadius" -> 0]},
           
           ImageSizeCache->{{125.97535321821039`, 139.2503924646782}, {
            14.662650590654899`, 40.27773676735788}}],
          StatusArea[#, 0.11]& ,
          TagBoxNote->"0.11"],
         StyleBox["0.11`", {
           GrayLevel[0]}, StripOnInput -> False]],
        Annotation[#, 
         Style[0.11, {
           GrayLevel[0]}], "Tooltip"]& ], 
       TagBox[
        TooltipBox[
         TagBox[
          DynamicBox[{
            FEPrivate`If[
             CurrentValue["MouseOver"], 
             EdgeForm[{
               GrayLevel[0.5], 
               AbsoluteThickness[1.5], 
               Opacity[0.66]}], {}, {}], 
            RectangleBox[{2., 0.}, {2.5, 0.042}, "RoundingRadius" -> 0]},
           
           ImageSizeCache->{{138.2503924646782, 151.52543171114604`}, {
            29.879249318071288`, 40.27773676735788}}],
          StatusArea[#, 0.042]& ,
          TagBoxNote->"0.042"],
         StyleBox["0.042`", {
           GrayLevel[0]}, StripOnInput -> False]],
        Annotation[#, 
         Style[0.042, {
           GrayLevel[0]}], "Tooltip"]& ], 
       TagBox[
        TooltipBox[
         TagBox[
          DynamicBox[{
            FEPrivate`If[
             CurrentValue["MouseOver"], 
             EdgeForm[{
               GrayLevel[0.5], 
               AbsoluteThickness[1.5], 
               Opacity[0.66]}], {}, {}], 
            RectangleBox[{2.5, 0.}, {3., 0.016}, "RoundingRadius" -> 0]},
           
           ImageSizeCache->{{150.52543171114604`, 163.80047095761387`}, {
            35.69736059620108, 40.27773676735788}}],
          StatusArea[#, 0.016]& ,
          TagBoxNote->"0.016"],
         StyleBox["0.016`", {
           GrayLevel[0]}, StripOnInput -> False]],
        Annotation[#, 
         Style[0.016, {
           GrayLevel[0]}], "Tooltip"]& ], 
       TagBox[
        TooltipBox[
         TagBox[
          DynamicBox[{
            FEPrivate`If[
             CurrentValue["MouseOver"], 
             EdgeForm[{
               GrayLevel[0.5], 
               AbsoluteThickness[1.5], 
               Opacity[0.66]}], {}, {}], 
            RectangleBox[{3., 0.}, {3.5, 0.004}, "RoundingRadius" -> 0]},
           
           ImageSizeCache->{{162.80047095761387`, 176.07551020408167`}, {
            38.38264272456868, 40.27773676735788}}],
          StatusArea[#, 0.004]& ,
          TagBoxNote->"0.004"],
         StyleBox["0.004`", {
           GrayLevel[0]}, StripOnInput -> False]],
        Annotation[#, 
         Style[0.004, {
           GrayLevel[0]}], "Tooltip"]& ]}, {}, {}}, {{}, {}, 
      {RGBColor[1, 0, 0], Thickness[Large], LineBox[CompressedData["
1:eJwV1nc4V+8bB/DP5FMZUSGkxLchSpK03IesooQSUqns7F2ZmSF7V/ZMssvs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        "]]}}},
    AspectRatio->NCache[GoldenRatio^(-1), 0.6180339887498948],
    Axes->{True, True},
    AxesOrigin->{-3., 0.},
    FrameTicks->{{Automatic, Automatic}, {Automatic, Automatic}},
    GridLines->{None, None},
    PlotRange->NCache[{{-3, 
        Rational[7, 2]}, {All, All}}, {{-3, 3.5}, {All, All}}],
    PlotRangePadding->{{
       Scaled[0.02], 
       Scaled[0.02]}, {
       Scaled[0.02], 
       Scaled[0.1]}},
    Ticks->{Automatic, Automatic}]}], "}"}]], "Output",
 ImageSize->{389, 119},
 ImageMargins->{{0, 0}, {0, 0}},
 ImageRegion->{{0, 1}, {0, 1}},
 CellLabel->"Out[9]=",
 CellID->109766786]
}, Open  ]]
}, Closed]]
}, Open  ]],

Cell[CellGroupData[{

Cell[TextData[ButtonBox["MORE ABOUT",
 BaseStyle->None,
 Appearance->{Automatic, None},
 Evaluator->None,
 Method->"Preemptive",
 ButtonFunction:>(FrontEndExecute[{
    FrontEnd`SelectionMove[
     FrontEnd`SelectedNotebook[], All, ButtonCell], 
    FrontEndToken["OpenCloseGroup"], 
    FrontEnd`SelectionMove[
     FrontEnd`SelectedNotebook[], After, 
     CellContents]}]& )]], "MoreAboutSection",
 CellID->38303248],

Cell[TextData[ButtonBox["Metropolis",
 BaseStyle->"Link",
 ButtonData->"paclet:Metropolis/guide/Metropolis"]], "MoreAbout",
 CellID->198223970]
}, Open  ]],

Cell[" ", "FooterCell"]
},
Saveable->False,
ScreenStyleEnvironment->"Working",
WindowSize->{725, 730},
WindowMargins->{{0, Automatic}, {Automatic, 0}},
WindowTitle->"BlockStudentTMetropolis - Wolfram Mathematica",
TaggingRules->{
 "ModificationHighlight" -> False, 
  "Metadata" -> {
   "built" -> "{2013, 2, 10, 17, 35, 30.3322352}", "context" -> "Metropolis`",
     "keywords" -> {}, "index" -> True, "label" -> 
    "Metropolis Package Paclet Symbol", "language" -> "en", "paclet" -> 
    "Metropolis Package", "status" -> "None", "summary" -> 
    "BlockStudentTMetropolis[logpdf, x_0, s, V, n] generates n updates for a \
set of conditional indepdent random variables starting from x0. The target \
distribution p[x] is provided by logpdf[x, indices] = Log[p[x, indices]] for \
any subset x[[indices]] of x. The symmetric proprosal distribution is the \
student T distribution with mean being the current value, (co)variance s^2 V, \
and degree infinity.", "synonyms" -> {}, "title" -> "BlockStudentTMetropolis",
     "type" -> "Symbol", "uri" -> "Metropolis/ref/BlockStudentTMetropolis"}, 
  "LinkTrails" -> "", "SearchTextTranslated" -> ""},
CellContext->"Global`",
FrontEndVersion->"8.0 for Microsoft Windows (64-bit) (February 23, 2011)",
StyleDefinitions->Notebook[{
   Cell[
    StyleData[
    StyleDefinitions -> FrontEnd`FileName[{"Wolfram"}, "Reference.nb"]]], 
   Cell[
    StyleData["Input"], CellContext -> "Global`"], 
   Cell[
    StyleData["Output"], CellContext -> "Global`"]}, Visible -> False, 
  FrontEndVersion -> "8.0 for Microsoft Windows (64-bit) (February 23, 2011)",
   StyleDefinitions -> "Default.nb"]
]
(* End of Notebook Content *)

(* Internal cache information *)
(*CellTagsOutline
CellTagsIndex->{
 "PrimaryExamplesSection"->{
  Cell[5875, 175, 459, 13, 70, "PrimaryExamplesSection",
   CellTags->"PrimaryExamplesSection",
   CellID->5934563]}
 }
*)
(*CellTagsIndex
CellTagsIndex->{
 {"PrimaryExamplesSection", 277675, 6730}
 }
*)
(*NotebookFileOutline
Notebook[{
Cell[616, 21, 76, 1, 70, "SymbolColorBar"],
Cell[695, 24, 185, 5, 70, "LinkTrail"],
Cell[883, 31, 2403, 55, 70, "AnchorBarGrid",
 CellID->1],
Cell[CellGroupData[{
Cell[3311, 90, 66, 1, 70, "ObjectName",
 CellID->1224892054],
Cell[3380, 93, 1009, 24, 70, "Usage"]
}, Open  ]],
Cell[CellGroupData[{
Cell[4426, 122, 580, 17, 70, "NotesSection",
 CellGroupingRules->{"SectionGrouping", 50},
 CellID->200745645],
Cell[5009, 141, 291, 8, 70, "Notes",
 CellID->115902181],
Cell[5303, 151, 394, 13, 70, "Notes",
 CellID->486766318],
Cell[5700, 166, 138, 4, 70, "Notes",
 CellID->894215170]
}, Closed]],
Cell[CellGroupData[{
Cell[5875, 175, 459, 13, 70, "PrimaryExamplesSection",
 CellTags->"PrimaryExamplesSection",
 CellID->5934563],
Cell[CellGroupData[{
Cell[6359, 192, 146, 5, 70, "ExampleSection",
 CellID->1464540],
Cell[6508, 199, 118, 3, 70, "Input",
 CellID->16158053],
Cell[6629, 204, 92, 1, 70, "ExampleText",
 CellID->75377655],
Cell[CellGroupData[{
Cell[6746, 209, 2088, 67, 70, "Input",
 CellID->223834254],
Cell[8837, 278, 224, 7, 36, "Output",
 CellID->220518226]
}, Open  ]],
Cell[9076, 288, 60, 1, 70, "ExampleText",
 CellID->12821957],
Cell[CellGroupData[{
Cell[9161, 293, 279, 10, 70, "Input",
 CellID->891160780],
Cell[9443, 305, 451, 15, 36, "Output",
 CellID->132162376]
}, Open  ]],
Cell[9909, 323, 54, 1, 70, "ExampleText",
 CellID->60397930],
Cell[CellGroupData[{
Cell[9988, 328, 603, 18, 47, "Input",
 CellID->111189838],
Cell[10594, 348, 14794, 250, 140, "Output",
 CellID->141115893]
}, Open  ]],
Cell[CellGroupData[{
Cell[25425, 603, 2031, 63, 70, "Input",
 CellID->45324905],
Cell[27459, 668, 138735, 3507, 70, "Output",
 CellID->36970811]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[166243, 4181, 228, 7, 70, "ExampleSection",
 CellID->589267740],
Cell[166474, 4190, 119, 3, 70, "Input",
 CellID->192544099],
Cell[166596, 4195, 153, 4, 70, "ExampleText",
 CellID->73418448],
Cell[CellGroupData[{
Cell[166774, 4203, 3631, 113, 70, "Input",
 CellID->756940083],
Cell[170408, 4318, 223, 7, 36, "Output",
 CellID->693900897]
}, Open  ]],
Cell[170646, 4328, 60, 1, 70, "ExampleText",
 CellID->24993314],
Cell[CellGroupData[{
Cell[170731, 4333, 279, 10, 70, "Input",
 CellID->265777506],
Cell[171013, 4345, 453, 15, 36, "Output",
 CellID->307896011]
}, Open  ]],
Cell[171481, 4363, 54, 1, 70, "ExampleText",
 CellID->193398859],
Cell[CellGroupData[{
Cell[171560, 4368, 603, 18, 70, "Input",
 CellID->402775542],
Cell[172166, 4388, 12179, 207, 70, "Output",
 CellID->214186837]
}, Open  ]],
Cell[CellGroupData[{
Cell[184382, 4600, 638, 19, 70, "Input",
 CellID->223197944],
Cell[185023, 4621, 15115, 255, 70, "Output",
 CellID->201295919]
}, Open  ]],
Cell[CellGroupData[{
Cell[200175, 4881, 726, 23, 70, "Input",
 CellID->36445872],
Cell[200904, 4906, 6766, 118, 70, "Output",
 CellID->432403975]
}, Open  ]],
Cell[CellGroupData[{
Cell[207707, 5029, 1498, 47, 70, "Input",
 CellID->182745918],
Cell[209208, 5078, 3441, 64, 70, "Output",
 CellID->1021282114]
}, Open  ]],
Cell[CellGroupData[{
Cell[212686, 5147, 1248, 38, 70, "Input",
 CellID->530831861],
Cell[213937, 5187, 22854, 569, 70, "Output",
 CellID->158480250]
}, Open  ]],
Cell[CellGroupData[{
Cell[236828, 5761, 1361, 39, 70, "Input",
 CellID->271797975],
Cell[238192, 5802, 37080, 861, 70, "Output",
 CellID->109766786]
}, Open  ]]
}, Closed]]
}, Open  ]],
Cell[CellGroupData[{
Cell[275333, 6670, 419, 12, 70, "MoreAboutSection",
 CellID->38303248],
Cell[275755, 6684, 143, 3, 70, "MoreAbout",
 CellID->198223970]
}, Open  ]],
Cell[275913, 6690, 23, 0, 70, "FooterCell"]
}
]
*)

(* End of internal cache information *)

