(* Content-type: application/vnd.wolfram.mathematica *)

(*** Wolfram Notebook File ***)
(* http://www.wolfram.com/nb *)

(* CreatedBy='Mathematica 8.0' *)

(*CacheID: 234*)
(* Internal cache information:
NotebookFileLineBreakTest
NotebookFileLineBreakTest
NotebookDataPosition[       157,          7]
NotebookDataLength[    198029,       5112]
NotebookOptionsPosition[    193102,       4949]
NotebookOutlinePosition[    194900,       4998]
CellTagsIndexPosition[    194813,       4993]
WindowTitle->Metropolis - Wolfram Mathematica
WindowFrame->Normal*)

(* Beginning of Notebook Content *)
Notebook[{
Cell[" ", "SymbolColorBar",
 CellMargins->{{Inherited, Inherited}, {-5, 0}}],

Cell[TextData[{
 ButtonBox["Metropolis",
  BaseStyle->{"Link", "LinkTrail"},
  ButtonData->"paclet:Metropolis/guide/Metropolis"],
 StyleBox[" > ", "LinkTrailSeparator"]
}], "LinkTrail"],

Cell[BoxData[GridBox[{
   {Cell["METROPOLIS PACLET SYMBOL", "PacletNameCell"], Cell[TextData[{
     Cell[BoxData[
      ActionMenuBox[
       FrameBox["\<\"More About \[RightGuillemet]\"\>",
        StripOnInput->False], {"\<\"Metropolis\"\>":>Documentation`HelpLookup[
       "paclet:Metropolis/guide/Metropolis"]},
       Appearance->None,
       MenuAppearance->Automatic]],
      LineSpacing->{1.4, 0}],
     "\[ThickSpace]\[ThickSpace]\[ThickSpace]\[ThickSpace]\[ThickSpace]\
\[ThickSpace]",
     Cell[BoxData[
      ActionMenuBox[
       FrameBox["\<\"URL \[RightGuillemet]\"\>",
        StripOnInput->False], {"\<\"Metropolis/ref/Metropolis\"\>":>
       None, "\<\"Copy Mathematica url\"\>":>
       Module[{DocumentationSearch`Private`nb$}, 
        DocumentationSearch`Private`nb$ = NotebookPut[
           Notebook[{
             Cell["Metropolis/ref/Metropolis"]}, Visible -> False]]; 
        SelectionMove[DocumentationSearch`Private`nb$, All, Notebook]; 
        FrontEndTokenExecute[DocumentationSearch`Private`nb$, "Copy"]; 
        NotebookClose[DocumentationSearch`Private`nb$]; Null], 
       Delimiter, "\<\"Copy web url\"\>":>
       Module[{DocumentationSearch`Private`nb$}, 
        DocumentationSearch`Private`nb$ = NotebookPut[
           Notebook[{
             Cell[
              BoxData[
               MakeBoxes[
                Hyperlink[
                "http://reference.wolfram.com/mathematica/Metropolis/ref/\
Metropolis.html"], StandardForm]], "Input", TextClipboardType -> 
              "PlainText"]}, Visible -> False]]; 
        SelectionMove[DocumentationSearch`Private`nb$, All, Notebook]; 
        FrontEndTokenExecute[DocumentationSearch`Private`nb$, "Copy"]; 
        NotebookClose[DocumentationSearch`Private`nb$]; 
        Null], "\<\"Go to web url\"\>":>FrontEndExecute[{
         NotebookLocate[{
           URL[
            StringJoin[
            "http://reference.wolfram.com/mathematica/", 
             "Metropolis/ref/Metropolis", ".html"]], None}]}]},
       Appearance->None,
       MenuAppearance->Automatic]],
      LineSpacing->{1.4, 0}]
    }], "AnchorBar"]}
  }]], "AnchorBarGrid",
 GridBoxOptions->{GridBoxItemSize->{"Columns" -> {
     Scaled[0.65], {
      Scaled[0.34]}}, "ColumnsIndexed" -> {}, "Rows" -> {{1.}}, 
   "RowsIndexed" -> {}}},
 CellID->1],

Cell[CellGroupData[{

Cell["Metropolis", "ObjectName",
 CellID->1224892054],

Cell[BoxData[GridBox[{
   {"", Cell[TextData[{
     Cell[BoxData[
      RowBox[{"Metropolis", "[", 
       RowBox[{
       "logpdf", ",", "logpdfjump", ",", "randomjump", ",", "x0", ",", "n"}], 
       "]"}]], "InlineFormula"],
     " \[LineSeparator]generates n Metropolis updates starting from x0. The \
target distribution is p[x] with the proposal distribution  J[xt, x] \
randomized by x = randomjump[xt] given current xt. Here logpdf[x] = Log[p[x]] \
and logpdfjump[xt,x] = Log[J[xt,x]]"
    }]]}
  }]], "Usage",
 GridBoxOptions->{
 GridBoxBackground->{
  "Columns" -> {{None}}, "ColumnsIndexed" -> {}, "Rows" -> {{None}}, 
   "RowsIndexed" -> {}}},
 CellID->982511436]
}, Open  ]],

Cell[CellGroupData[{

Cell[TextData[Cell[BoxData[
 ButtonBox[
  FrameBox[
   StyleBox[
    RowBox[{"MORE", " ", "INFORMATION"}], "NotesFrameText"],
   StripOnInput->False],
  Appearance->{Automatic, None},
  BaseStyle->None,
  ButtonFunction:>(FrontEndExecute[{
     FrontEnd`SelectionMove[
      FrontEnd`SelectedNotebook[], All, ButtonCell], 
     FrontEndToken["OpenCloseGroup"], 
     FrontEnd`SelectionMove[
      FrontEnd`SelectedNotebook[], After, CellContents]}]& ),
  Evaluator->None,
  Method->"Preemptive"]]]], "NotesSection",
 CellGroupingRules->{"SectionGrouping", 50},
 CellID->234101828],

Cell["\<\
To use  Metropolis, you first need to load the Metropolis package using  \
Needs[\"Metropolis`\"].\
\>", "Notes",
 CellID->85410339],

Cell["The following options can be given: ", "Notes",
 CellID->182314388],

Cell[BoxData[GridBox[{
   {Cell["      ", "TableRowIcon"], 
    ButtonBox["DomainConstraint",
     BaseStyle->"Link",
     ButtonData->"paclet:Metropolis/ref/DomainConstraint"], 
    RowBox[{
     RowBox[{"True", " ", "&"}], " "}], Cell[TextData[Cell[TextData[Cell[
    "random variable x satifies DomainConstraint[x]", "TableText"]], 
     "TableText"]], "TableText"]},
   {Cell["      ", "TableRowIcon"], 
    ButtonBox["StartLogpdf",
     BaseStyle->"Link",
     ButtonData->"paclet:Metropolis/ref/StartLogpdf"], "None", Cell[TextData[
    Cell[TextData[Cell[BoxData[
    Cell["provide Log[p[x0]]", "TableText"]]]], "TableText"]], "TableText"]},
   {Cell["      ", "TableRowIcon"], 
    ButtonBox["IsReturnLast",
     BaseStyle->"Link",
     ButtonData->"paclet:Metropolis/ref/IsReturnLast"], "True", Cell[TextData[
    Cell["whether return the last Metropolis update", "TableText"]], 
     "TableText"]},
   {Cell["      ", "TableRowIcon"], 
    ButtonBox["PrintDetails",
     BaseStyle->"Link",
     ButtonData->"paclet:Metropolis/ref/PrintDetails"], "False", Cell[
    TextData[Cell["whether print details", "TableText"]], "TableText"]}
  }]], "3ColumnTableMod",
 GridBoxOptions->{
 GridBoxBackground->{
  "Columns" -> {{None}}, "ColumnsIndexed" -> {}, 
   "Rows" -> {None, None, None, {None}}, "RowsIndexed" -> {}},
 GridBoxDividers->{"Rows" -> {{True, True, True, True, True}}}},
 CellID->497869101],

Cell["\<\
Each update is given by {x, logpdf[x], 1/0 (acccept/not), \
Min[1,Metropolis-Hastings ratio]}.\
\>", "Notes",
 CellID->588891594]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[ButtonBox["EXAMPLES",
 BaseStyle->None,
 Appearance->{Automatic, None},
 Evaluator->None,
 Method->"Preemptive",
 ButtonFunction:>(FrontEndExecute[{
    FrontEnd`SelectionMove[
     FrontEnd`SelectedNotebook[], All, ButtonCell], 
    FrontEndToken["OpenCloseGroup"], 
    FrontEnd`SelectionMove[
     FrontEnd`SelectedNotebook[], After, 
     CellContents]}]& )]], "PrimaryExamplesSection",
 CellTags->"PrimaryExamplesSection",
 CellID->42576746],

Cell[CellGroupData[{

Cell[TextData[{
 "Basic Examples",
 "\[NonBreakingSpace]\[NonBreakingSpace]",
 Cell["(1)", "ExampleCount"]
}], "ExampleSection",
 CellID->121590637],

Cell[BoxData[
 RowBox[{"Needs", "[", "\"\<Metropolis`\>\"", "]"}]], "Input",
 CellLabel->"In[1]:=",
 CellID->223020744],

Cell["This samples the standard normal distribuiton", "ExampleText",
 CellID->34393979],

Cell[CellGroupData[{

Cell[BoxData[{
 RowBox[{"Clear", "[", 
  RowBox[{
  "scale", ",", " ", "n", ",", "logpdf", ",", "logpdfjump", ",", "randomjump",
    ",", "res"}], "]"}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{"scale", "=", "2.234"}], ";"}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{"n", "=", "5000"}], ";"}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{"logpdf", "=", 
   RowBox[{"Function", "[", 
    RowBox[{"x", ",", 
     RowBox[{"Log", "[", 
      RowBox[{"PDF", "[", 
       RowBox[{
        RowBox[{"NormalDistribution", "[", 
         RowBox[{"0.", ",", "1"}], "]"}], ",", "x"}], "]"}], "]"}]}], "]"}]}],
   ";"}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{"logpdfjump", "=", 
   RowBox[{"If", "[", 
    RowBox[{"True", ",", 
     RowBox[{"(", 
      RowBox[{"0", "&"}], ")"}], ",", 
     RowBox[{"Function", "[", 
      RowBox[{
       RowBox[{"{", 
        RowBox[{"xt", ",", "x"}], "}"}], ",", 
       RowBox[{"Log", "[", 
        RowBox[{"PDF", "[", 
         RowBox[{
          RowBox[{"NormalDistribution", "[", 
           RowBox[{"xt", ",", "scale"}], "]"}], ",", "x"}], "]"}], "]"}]}], 
      "]"}]}], "\[IndentingNewLine]", "]"}]}], ";"}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{"randomjump", "=", 
   RowBox[{"Function", "[", 
    RowBox[{"xt", ",", 
     RowBox[{"RandomReal", "[", 
      RowBox[{"NormalDistribution", "[", 
       RowBox[{"xt", ",", "scale"}], "]"}], "]"}]}], "]"}]}], 
  ";"}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{
   RowBox[{"res", "=", 
    RowBox[{"Metropolis", "[", 
     RowBox[{"logpdf", ",", "logpdfjump", ",", "randomjump", ",", "0", ",", 
      RowBox[{"2", " ", "n"}], ",", 
      RowBox[{"IsReturnLast", "\[Rule]", "False"}]}], "]"}]}], ";"}], "//", 
  "Timing"}]}], "Input",
 CellLabel->"In[2]:=",
 CellID->294782483],

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{"1.6529999999999998`", ",", "Null"}], "}"}]], "Output",
 ImageSize->{92, 15},
 ImageMargins->{{0, 0}, {0, 0}},
 ImageRegion->{{0, 1}, {0, 1}},
 CellLabel->"Out[2]=",
 CellID->2875901]
}, Open  ]],

Cell["Mean accept ratio:", "ExampleText",
 CellID->12821957],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"N", "[", 
  RowBox[{"Mean", "[", 
   RowBox[{"res", "[", 
    RowBox[{"[", 
     RowBox[{
      RowBox[{"n", ";;"}], ",", 
      RowBox[{
       RowBox[{"-", "2"}], ";;"}]}], "]"}], "]"}], "]"}], "]"}]], "Input",
 CellLabel->"In[3]:=",
 CellID->891160780],

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{"0.4751049790041992`", ",", "0.4751869404756622`"}], 
  "}"}]], "Output",
 ImageSize->{141, 15},
 ImageMargins->{{0, 0}, {0, 0}},
 ImageRegion->{{0, 1}, {0, 1}},
 CellLabel->"Out[3]=",
 CellID->184204190]
}, Open  ]],

Cell["Time serie: ", "ExampleText",
 CellID->60397930],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"ListLinePlot", "[", 
  RowBox[{
   RowBox[{"res", "[", 
    RowBox[{"[", 
     RowBox[{
      RowBox[{";;", ";;", "10"}], ",", "1"}], "]"}], "]"}], ",", 
   RowBox[{"PlotRange", "\[Rule]", "All"}], ",", 
   RowBox[{"AxesLabel", "\[Rule]", 
    RowBox[{"{", 
     RowBox[{"\"\<t\>\"", ",", "\"\<x\>\""}], "}"}]}]}], "]"}]], "Input",
 CellLabel->"In[4]:=",
 CellID->111189838],

Cell[BoxData[
 GraphicsBox[
  TagBox[RasterBox[CompressedData["
1:eJztnV+IJcUVxgfy7osJ7GKCOuyGQGaDixsflsxc/ySQzQSjJu5iojExZIgQ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    "], {{0, 116}, {180, 0}}, {0, 255},
    ColorFunction->RGBColor],
   BoxForm`ImageTag["Byte", ColorSpace -> "RGB", Interleaving -> True],
   Selectable->False],
  BaseStyle->"ImageGraphics",
  ImageSize->Automatic,
  ImageSizeRaw->{180, 116},
  PlotRange->{{0, 180}, {0, 116}}]], "Output",
 ImageSize->{184, 116},
 ImageMargins->{{0, 0}, {0, 0}},
 ImageRegion->{{0, 1}, {0, 1}},
 CellLabel->"Out[4]=",
 CellID->697109720]
}, Open  ]],

Cell["Simulated distribution vs target distribution (red line):", \
"ExampleText",
 CellID->77108731],

Cell[CellGroupData[{

Cell[BoxData[{
 RowBox[{"Clear", "[", "data", "]"}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{"data", "=", 
   RowBox[{"res", "[", 
    RowBox[{"[", 
     RowBox[{
      RowBox[{"n", ";;"}], ",", "1"}], "]"}], "]"}]}], 
  ";"}], "\[IndentingNewLine]", 
 RowBox[{"Show", "[", 
  RowBox[{
   RowBox[{"Histogram", "[", 
    RowBox[{"data", ",", "Automatic", ",", "\"\<ProbabilityDensity\>\""}], 
    "]"}], ",", 
   RowBox[{"Plot", "[", 
    RowBox[{
     RowBox[{"Exp", "[", 
      RowBox[{"logpdf", "[", "x", "]"}], "]"}], ",", 
     RowBox[{"{", 
      RowBox[{"x", ",", 
       RowBox[{"Min", "[", "data", "]"}], ",", 
       RowBox[{"Max", "[", "data", "]"}]}], "}"}], ",", 
     RowBox[{"PlotStyle", "\[Rule]", 
      RowBox[{"{", 
       RowBox[{"Red", ",", "Thick"}], "}"}]}]}], "]"}], ",", 
   RowBox[{"AxesLabel", "\[Rule]", 
    RowBox[{"{", 
     RowBox[{"\"\<x\>\"", ",", "\"\<freq\>\""}], "}"}]}]}], "]"}]}], "Input",
 CellLabel->"In[5]:=",
 CellID->66917885],

Cell[BoxData[
 GraphicsBox[{
   {RGBColor[0.798413061722744, 0.824719615472648, 0.968322270542458], 
    EdgeForm[Opacity[0.7]], {}, 
    {RGBColor[0.798413061722744, 0.824719615472648, 0.968322270542458], 
     EdgeForm[Opacity[0.7]], 
     TagBox[
      TooltipBox[
       TagBox[
        DynamicBox[{
          FEPrivate`If[
           CurrentValue["MouseOver"], 
           EdgeForm[{
             GrayLevel[0.5], 
             AbsoluteThickness[1.5], 
             Opacity[0.66]}], {}, {}], 
          
          RectangleBox[{-3., 0.}, {-2.8, 0.003999200159968003}, 
           "RoundingRadius" -> 0]},
         ImageSizeCache->{{15.500000000000014`, 20.9526901669759}, {
          43.386961758154506`, 45.2050983124842}}],
        StatusArea[#, 0.003999200159968003]& ,
        TagBoxNote->"0.003999200159968003"],
       StyleBox["0.003999200159968003`", {
         GrayLevel[0]}, StripOnInput -> False]],
      Annotation[#, 
       Style[0.003999200159968003, {
         GrayLevel[0]}], "Tooltip"]& ], 
     TagBox[
      TooltipBox[
       TagBox[
        DynamicBox[{
          FEPrivate`If[
           CurrentValue["MouseOver"], 
           EdgeForm[{
             GrayLevel[0.5], 
             AbsoluteThickness[1.5], 
             Opacity[0.66]}], {}, {}], 
          
          RectangleBox[{-2.8, 0.}, {-2.6, 0.012997400519896039`}, 
           "RoundingRadius" -> 0]},
         ImageSizeCache->{{19.9526901669759, 25.40538033395177}, {
          41.54615451091267, 45.2050983124842}}],
        StatusArea[#, 0.012997400519896039`]& ,
        TagBoxNote->"0.012997400519896039"],
       StyleBox["0.012997400519896039`", {
         GrayLevel[0]}, StripOnInput -> False]],
      Annotation[#, 
       Style[0.012997400519896039`, {
         GrayLevel[0]}], "Tooltip"]& ], 
     TagBox[
      TooltipBox[
       TagBox[
        DynamicBox[{
          FEPrivate`If[
           CurrentValue["MouseOver"], 
           EdgeForm[{
             GrayLevel[0.5], 
             AbsoluteThickness[1.5], 
             Opacity[0.66]}], {}, {}], 
          
          RectangleBox[{-2.6, 0.}, {-2.4, 0.021995600879824016`}, 
           "RoundingRadius" -> 0]},
         ImageSizeCache->{{24.40538033395177, 29.858070500927653`}, {
          39.70534726367085, 45.2050983124842}}],
        StatusArea[#, 0.021995600879824016`]& ,
        TagBoxNote->"0.021995600879824016"],
       StyleBox["0.021995600879824016`", {
         GrayLevel[0]}, StripOnInput -> False]],
      Annotation[#, 
       Style[0.021995600879824016`, {
         GrayLevel[0]}], "Tooltip"]& ], 
     TagBox[
      TooltipBox[
       TagBox[
        DynamicBox[{
          FEPrivate`If[
           CurrentValue["MouseOver"], 
           EdgeForm[{
             GrayLevel[0.5], 
             AbsoluteThickness[1.5], 
             Opacity[0.66]}], {}, {}], 
          
          RectangleBox[{-2.4, 0.}, {-2.2, 0.03499300139972011}, 
           "RoundingRadius" -> 0]},
         ImageSizeCache->{{28.858070500927653`, 34.31076066790353}, {
          37.04640346209931, 45.2050983124842}}],
        StatusArea[#, 0.03499300139972011]& ,
        TagBoxNote->"0.03499300139972011"],
       StyleBox["0.03499300139972011`", {
         GrayLevel[0]}, StripOnInput -> False]],
      Annotation[#, 
       Style[0.03499300139972011, {
         GrayLevel[0]}], "Tooltip"]& ], 
     TagBox[
      TooltipBox[
       TagBox[
        DynamicBox[{
          FEPrivate`If[
           CurrentValue["MouseOver"], 
           EdgeForm[{
             GrayLevel[0.5], 
             AbsoluteThickness[1.5], 
             Opacity[0.66]}], {}, {}], 
          
          RectangleBox[{-2.2, 0.}, {-2., 0.04499100179964003}, 
           "RoundingRadius" -> 0]},
         ImageSizeCache->{{33.31076066790353, 38.763450834879414`}, {
          35.001062076275076`, 45.2050983124842}}],
        StatusArea[#, 0.04499100179964003]& ,
        TagBoxNote->"0.04499100179964003"],
       StyleBox["0.04499100179964003`", {
         GrayLevel[0]}, StripOnInput -> False]],
      Annotation[#, 
       Style[0.04499100179964003, {
         GrayLevel[0]}], "Tooltip"]& ], 
     TagBox[
      TooltipBox[
       TagBox[
        DynamicBox[{
          FEPrivate`If[
           CurrentValue["MouseOver"], 
           EdgeForm[{
             GrayLevel[0.5], 
             AbsoluteThickness[1.5], 
             Opacity[0.66]}], {}, {}], 
          
          RectangleBox[{-2., 0.}, {-1.8, 0.06898620275944813}, 
           "RoundingRadius" -> 0]},
         ImageSizeCache->{{37.763450834879414`, 43.21614100185529}, {
          30.09224275029686, 45.2050983124842}}],
        StatusArea[#, 0.06898620275944813]& ,
        TagBoxNote->"0.06898620275944813"],
       StyleBox["0.06898620275944813`", {
         GrayLevel[0]}, StripOnInput -> False]],
      Annotation[#, 
       Style[0.06898620275944813, {
         GrayLevel[0]}], "Tooltip"]& ], 
     TagBox[
      TooltipBox[
       TagBox[
        DynamicBox[{
          FEPrivate`If[
           CurrentValue["MouseOver"], 
           EdgeForm[{
             GrayLevel[0.5], 
             AbsoluteThickness[1.5], 
             Opacity[0.66]}], {}, {}], 
          
          RectangleBox[{-1.8, 0.}, {-1.6, 0.05798840231953611}, 
           "RoundingRadius" -> 0]},
         ImageSizeCache->{{42.21614100185529, 47.668831168831176`}, {
          32.34211827470354, 45.2050983124842}}],
        StatusArea[#, 0.05798840231953611]& ,
        TagBoxNote->"0.05798840231953611"],
       StyleBox["0.05798840231953611`", {
         GrayLevel[0]}, StripOnInput -> False]],
      Annotation[#, 
       Style[0.05798840231953611, {
         GrayLevel[0]}], "Tooltip"]& ], 
     TagBox[
      TooltipBox[
       TagBox[
        DynamicBox[{
          FEPrivate`If[
           CurrentValue["MouseOver"], 
           EdgeForm[{
             GrayLevel[0.5], 
             AbsoluteThickness[1.5], 
             Opacity[0.66]}], {}, {}], 
          
          RectangleBox[{-1.6, 0.}, {-1.4, 0.11497700459908008`}, 
           "RoundingRadius" -> 0]},
         ImageSizeCache->{{46.668831168831176`, 52.12152133580706}, {
          20.683672375505328`, 45.2050983124842}}],
        StatusArea[#, 0.11497700459908008`]& ,
        TagBoxNote->"0.11497700459908008"],
       StyleBox["0.11497700459908008`", {
         GrayLevel[0]}, StripOnInput -> False]],
      Annotation[#, 
       Style[0.11497700459908008`, {
         GrayLevel[0]}], "Tooltip"]& ], 
     TagBox[
      TooltipBox[
       TagBox[
        DynamicBox[{
          FEPrivate`If[
           CurrentValue["MouseOver"], 
           EdgeForm[{
             GrayLevel[0.5], 
             AbsoluteThickness[1.5], 
             Opacity[0.66]}], {}, {}], 
          
          RectangleBox[{-1.4, 0.}, {-1.2, 0.16396720655868827`}, 
           "RoundingRadius" -> 0]},
         ImageSizeCache->{{51.12152133580706, 56.57421150278294}, {
          10.661499584966478`, 45.2050983124842}}],
        StatusArea[#, 0.16396720655868827`]& ,
        TagBoxNote->"0.16396720655868827"],
       StyleBox["0.16396720655868827`", {
         GrayLevel[0]}, StripOnInput -> False]],
      Annotation[#, 
       Style[0.16396720655868827`, {
         GrayLevel[0]}], "Tooltip"]& ], 
     TagBox[
      TooltipBox[
       TagBox[
        DynamicBox[{
          FEPrivate`If[
           CurrentValue["MouseOver"], 
           EdgeForm[{
             GrayLevel[0.5], 
             AbsoluteThickness[1.5], 
             Opacity[0.66]}], {}, {}], 
          
          RectangleBox[{-1.2, 0.}, {-1., 0.20095980803839233`}, 
           "RoundingRadius" -> 0]},
         ImageSizeCache->{{55.57421150278294, 61.026901669758814`}, {
          3.093736457416739, 45.2050983124842}}],
        StatusArea[#, 0.20095980803839233`]& ,
        TagBoxNote->"0.20095980803839233"],
       StyleBox["0.20095980803839233`", {
         GrayLevel[0]}, StripOnInput -> False]],
      Annotation[#, 
       Style[0.20095980803839233`, {
         GrayLevel[0]}], "Tooltip"]& ], 
     TagBox[
      TooltipBox[
       TagBox[
        DynamicBox[{
          FEPrivate`If[
           CurrentValue["MouseOver"], 
           EdgeForm[{
             GrayLevel[0.5], 
             AbsoluteThickness[1.5], 
             Opacity[0.66]}], {}, {}], 
          
          RectangleBox[{-1., 0.}, {-0.8, 0.2699460107978405}, 
           "RoundingRadius" -> 0]},
         ImageSizeCache->{{60.026901669758814`, 
          65.4795918367347}, {-11.019119104770603`, 45.2050983124842}}],
        StatusArea[#, 0.2699460107978405]& ,
        TagBoxNote->"0.2699460107978405"],
       StyleBox["0.2699460107978405`", {
         GrayLevel[0]}, StripOnInput -> False]],
      Annotation[#, 
       Style[0.2699460107978405, {
         GrayLevel[0]}], "Tooltip"]& ], 
     TagBox[
      TooltipBox[
       TagBox[
        DynamicBox[{
          FEPrivate`If[
           CurrentValue["MouseOver"], 
           EdgeForm[{
             GrayLevel[0.5], 
             AbsoluteThickness[1.5], 
             Opacity[0.66]}], {}, {}], 
          
          RectangleBox[{-0.8, 0.}, {-0.6, 0.2589482103579283}, 
           "RoundingRadius" -> 0]},
         ImageSizeCache->{{64.4795918367347, 
          69.93228200371058}, {-8.769243580363892, 45.2050983124842}}],
        StatusArea[#, 0.2589482103579283]& ,
        TagBoxNote->"0.2589482103579283"],
       StyleBox["0.2589482103579283`", {
         GrayLevel[0]}, StripOnInput -> False]],
      Annotation[#, 
       Style[0.2589482103579283, {
         GrayLevel[0]}], "Tooltip"]& ], 
     TagBox[
      TooltipBox[
       TagBox[
        DynamicBox[{
          FEPrivate`If[
           CurrentValue["MouseOver"], 
           EdgeForm[{
             GrayLevel[0.5], 
             AbsoluteThickness[1.5], 
             Opacity[0.66]}], {}, {}], 
          
          RectangleBox[{-0.6, 0.}, {-0.4, 0.36692661467706467`}, 
           "RoundingRadius" -> 0]},
         ImageSizeCache->{{68.93228200371058, 
          74.38497217068647}, {-30.85893054726585, 45.2050983124842}}],
        StatusArea[#, 0.36692661467706467`]& ,
        TagBoxNote->"0.36692661467706467"],
       StyleBox["0.36692661467706467`", {
         GrayLevel[0]}, StripOnInput -> False]],
      Annotation[#, 
       Style[0.36692661467706467`, {
         GrayLevel[0]}], "Tooltip"]& ], 
     TagBox[
      TooltipBox[
       TagBox[
        DynamicBox[{
          FEPrivate`If[
           CurrentValue["MouseOver"], 
           EdgeForm[{
             GrayLevel[0.5], 
             AbsoluteThickness[1.5], 
             Opacity[0.66]}], {}, {}], 
          
          RectangleBox[{-0.4, 0.}, {-0.2, 0.3889222155568886}, 
           "RoundingRadius" -> 0]},
         ImageSizeCache->{{73.38497217068647, 
          78.83766233766234}, {-35.35868159607919, 45.2050983124842}}],
        StatusArea[#, 0.3889222155568886]& ,
        TagBoxNote->"0.3889222155568886"],
       StyleBox["0.3889222155568886`", {
         GrayLevel[0]}, StripOnInput -> False]],
      Annotation[#, 
       Style[0.3889222155568886, {
         GrayLevel[0]}], "Tooltip"]& ], 
     TagBox[
      TooltipBox[
       TagBox[
        DynamicBox[{
          FEPrivate`If[
           CurrentValue["MouseOver"], 
           EdgeForm[{
             GrayLevel[0.5], 
             AbsoluteThickness[1.5], 
             Opacity[0.66]}], {}, {}], 
          
          RectangleBox[{-0.2, 0.}, {0., 0.38792241551689666`}, 
           "RoundingRadius" -> 0]},
         ImageSizeCache->{{77.83766233766234, 
          83.29035250463822}, {-35.15414745749677, 45.2050983124842}}],
        StatusArea[#, 0.38792241551689666`]& ,
        TagBoxNote->"0.38792241551689666"],
       StyleBox["0.38792241551689666`", {
         GrayLevel[0]}, StripOnInput -> False]],
      Annotation[#, 
       Style[0.38792241551689666`, {
         GrayLevel[0]}], "Tooltip"]& ], 
     TagBox[
      TooltipBox[
       TagBox[
        DynamicBox[{
          FEPrivate`If[
           CurrentValue["MouseOver"], 
           EdgeForm[{
             GrayLevel[0.5], 
             AbsoluteThickness[1.5], 
             Opacity[0.66]}], {}, {}], 
          
          RectangleBox[{0., 0.}, {0.2, 0.39392121575684863`}, 
           "RoundingRadius" -> 0]},
         ImageSizeCache->{{82.29035250463822, 
          87.7430426716141}, {-36.381352288991316`, 45.2050983124842}}],
        StatusArea[#, 0.39392121575684863`]& ,
        TagBoxNote->"0.39392121575684863"],
       StyleBox["0.39392121575684863`", {
         GrayLevel[0]}, StripOnInput -> False]],
      Annotation[#, 
       Style[0.39392121575684863`, {
         GrayLevel[0]}], "Tooltip"]& ], 
     TagBox[
      TooltipBox[
       TagBox[
        DynamicBox[{
          FEPrivate`If[
           CurrentValue["MouseOver"], 
           EdgeForm[{
             GrayLevel[0.5], 
             AbsoluteThickness[1.5], 
             Opacity[0.66]}], {}, {}], 
          
          RectangleBox[{0.2, 0.}, {0.4, 0.40691861627674464`}, 
           "RoundingRadius" -> 0]},
         ImageSizeCache->{{86.7430426716141, 
          92.19573283858998}, {-39.04029609056284, 45.2050983124842}}],
        StatusArea[#, 0.40691861627674464`]& ,
        TagBoxNote->"0.40691861627674464"],
       StyleBox["0.40691861627674464`", {
         GrayLevel[0]}, StripOnInput -> False]],
      Annotation[#, 
       Style[0.40691861627674464`, {
         GrayLevel[0]}], "Tooltip"]& ], 
     TagBox[
      TooltipBox[
       TagBox[
        DynamicBox[{
          FEPrivate`If[
           CurrentValue["MouseOver"], 
           EdgeForm[{
             GrayLevel[0.5], 
             AbsoluteThickness[1.5], 
             Opacity[0.66]}], {}, {}], 
          
          RectangleBox[{0.4, 0.}, {0.6, 0.4069186162767447}, "RoundingRadius" -> 
           0]},
         ImageSizeCache->{{91.19573283858998, 
          96.64842300556586}, {-39.040296090562855`, 45.2050983124842}}],
        StatusArea[#, 0.4069186162767447]& ,
        TagBoxNote->"0.4069186162767447"],
       StyleBox["0.4069186162767447`", {
         GrayLevel[0]}, StripOnInput -> False]],
      Annotation[#, 
       Style[0.4069186162767447, {
         GrayLevel[0]}], "Tooltip"]& ], 
     TagBox[
      TooltipBox[
       TagBox[
        DynamicBox[{
          FEPrivate`If[
           CurrentValue["MouseOver"], 
           EdgeForm[{
             GrayLevel[0.5], 
             AbsoluteThickness[1.5], 
             Opacity[0.66]}], {}, {}], 
          
          RectangleBox[{0.6, 0.}, {0.8, 0.30293941211757636`}, 
           "RoundingRadius" -> 0]},
         ImageSizeCache->{{95.64842300556586, 
          101.10111317254174`}, {-17.768745677990594`, 45.2050983124842}}],
        StatusArea[#, 0.30293941211757636`]& ,
        TagBoxNote->"0.30293941211757636"],
       StyleBox["0.30293941211757636`", {
         GrayLevel[0]}, StripOnInput -> False]],
      Annotation[#, 
       Style[0.30293941211757636`, {
         GrayLevel[0]}], "Tooltip"]& ], 
     TagBox[
      TooltipBox[
       TagBox[
        DynamicBox[{
          FEPrivate`If[
           CurrentValue["MouseOver"], 
           EdgeForm[{
             GrayLevel[0.5], 
             AbsoluteThickness[1.5], 
             Opacity[0.66]}], {}, {}], 
          
          RectangleBox[{0.8, 0.}, {1., 0.25594881023795246`}, 
           "RoundingRadius" -> 0]},
         ImageSizeCache->{{100.10111317254174`, 
          105.55380333951763`}, {-8.155641164616654, 45.2050983124842}}],
        StatusArea[#, 0.25594881023795246`]& ,
        TagBoxNote->"0.25594881023795246"],
       StyleBox["0.25594881023795246`", {
         GrayLevel[0]}, StripOnInput -> False]],
      Annotation[#, 
       Style[0.25594881023795246`, {
         GrayLevel[0]}], "Tooltip"]& ], 
     TagBox[
      TooltipBox[
       TagBox[
        DynamicBox[{
          FEPrivate`If[
           CurrentValue["MouseOver"], 
           EdgeForm[{
             GrayLevel[0.5], 
             AbsoluteThickness[1.5], 
             Opacity[0.66]}], {}, {}], 
          
          RectangleBox[{1., 0.}, {1.2, 0.2219556088782244}, "RoundingRadius" -> 
           0]},
         ImageSizeCache->{{104.55380333951763`, 
          110.00649350649351`}, {-1.2014804528141951`, 45.2050983124842}}],
        StatusArea[#, 0.2219556088782244]& ,
        TagBoxNote->"0.2219556088782244"],
       StyleBox["0.2219556088782244`", {
         GrayLevel[0]}, StripOnInput -> False]],
      Annotation[#, 
       Style[0.2219556088782244, {
         GrayLevel[0]}], "Tooltip"]& ], 
     TagBox[
      TooltipBox[
       TagBox[
        DynamicBox[{
          FEPrivate`If[
           CurrentValue["MouseOver"], 
           EdgeForm[{
             GrayLevel[0.5], 
             AbsoluteThickness[1.5], 
             Opacity[0.66]}], {}, {}], 
          
          RectangleBox[{1.2, 0.}, {1.4, 0.20895820835832837`}, 
           "RoundingRadius" -> 0]},
         ImageSizeCache->{{109.00649350649351`, 114.45918367346938`}, {
          1.4574633487573365`, 45.2050983124842}}],
        StatusArea[#, 0.20895820835832837`]& ,
        TagBoxNote->"0.20895820835832837"],
       StyleBox["0.20895820835832837`", {
         GrayLevel[0]}, StripOnInput -> False]],
      Annotation[#, 
       Style[0.20895820835832837`, {
         GrayLevel[0]}], "Tooltip"]& ], 
     TagBox[
      TooltipBox[
       TagBox[
        DynamicBox[{
          FEPrivate`If[
           CurrentValue["MouseOver"], 
           EdgeForm[{
             GrayLevel[0.5], 
             AbsoluteThickness[1.5], 
             Opacity[0.66]}], {}, {}], 
          
          RectangleBox[{1.4, 0.}, {1.6, 0.12297540491901608`}, 
           "RoundingRadius" -> 0]},
         ImageSizeCache->{{113.45918367346938`, 118.91187384044527`}, {
          19.047399266845932`, 45.2050983124842}}],
        StatusArea[#, 0.12297540491901608`]& ,
        TagBoxNote->"0.12297540491901608"],
       StyleBox["0.12297540491901608`", {
         GrayLevel[0]}, StripOnInput -> False]],
      Annotation[#, 
       Style[0.12297540491901608`, {
         GrayLevel[0]}], "Tooltip"]& ], 
     TagBox[
      TooltipBox[
       TagBox[
        DynamicBox[{
          FEPrivate`If[
           CurrentValue["MouseOver"], 
           EdgeForm[{
             GrayLevel[0.5], 
             AbsoluteThickness[1.5], 
             Opacity[0.66]}], {}, {}], 
          
          RectangleBox[{1.6, 0.}, {1.8, 0.08498300339932015}, 
           "RoundingRadius" -> 0]},
         ImageSizeCache->{{117.91187384044527`, 123.36456400742115`}, {
          26.81969653297806, 45.2050983124842}}],
        StatusArea[#, 0.08498300339932015]& ,
        TagBoxNote->"0.08498300339932015"],
       StyleBox["0.08498300339932015`", {
         GrayLevel[0]}, StripOnInput -> False]],
      Annotation[#, 
       Style[0.08498300339932015, {
         GrayLevel[0]}], "Tooltip"]& ], 
     TagBox[
      TooltipBox[
       TagBox[
        DynamicBox[{
          FEPrivate`If[
           CurrentValue["MouseOver"], 
           EdgeForm[{
             GrayLevel[0.5], 
             AbsoluteThickness[1.5], 
             Opacity[0.66]}], {}, {}], 
          
          RectangleBox[{1.8, 0.}, {2., 0.06598680263947211}, "RoundingRadius" -> 
           0]},
         ImageSizeCache->{{122.36456400742115`, 127.81725417439702`}, {
          30.705845166044142`, 45.2050983124842}}],
        StatusArea[#, 0.06598680263947211]& ,
        TagBoxNote->"0.06598680263947211"],
       StyleBox["0.06598680263947211`", {
         GrayLevel[0]}, StripOnInput -> False]],
      Annotation[#, 
       Style[0.06598680263947211, {
         GrayLevel[0]}], "Tooltip"]& ], 
     TagBox[
      TooltipBox[
       TagBox[
        DynamicBox[{
          FEPrivate`If[
           CurrentValue["MouseOver"], 
           EdgeForm[{
             GrayLevel[0.5], 
             AbsoluteThickness[1.5], 
             Opacity[0.66]}], {}, {}], 
          
          RectangleBox[{2., 0.}, {2.2, 0.03599280143971203}, "RoundingRadius" -> 
           0]},
         ImageSizeCache->{{126.81725417439702`, 132.2699443413729}, {
          36.8418693235169, 45.2050983124842}}],
        StatusArea[#, 0.03599280143971203]& ,
        TagBoxNote->"0.03599280143971203"],
       StyleBox["0.03599280143971203`", {
         GrayLevel[0]}, StripOnInput -> False]],
      Annotation[#, 
       Style[0.03599280143971203, {
         GrayLevel[0]}], "Tooltip"]& ], 
     TagBox[
      TooltipBox[
       TagBox[
        DynamicBox[{
          FEPrivate`If[
           CurrentValue["MouseOver"], 
           EdgeForm[{
             GrayLevel[0.5], 
             AbsoluteThickness[1.5], 
             Opacity[0.66]}], {}, {}], 
          
          RectangleBox[{2.2, 0.}, {2.4, 0.03099380123975209}, 
           "RoundingRadius" -> 0]},
         ImageSizeCache->{{131.2699443413729, 136.7226345083488}, {
          37.86454001642902, 45.2050983124842}}],
        StatusArea[#, 0.03099380123975209]& ,
        TagBoxNote->"0.03099380123975209"],
       StyleBox["0.03099380123975209`", {
         GrayLevel[0]}, StripOnInput -> False]],
      Annotation[#, 
       Style[0.03099380123975209, {
         GrayLevel[0]}], "Tooltip"]& ], 
     TagBox[
      TooltipBox[
       TagBox[
        DynamicBox[{
          FEPrivate`If[
           CurrentValue["MouseOver"], 
           EdgeForm[{
             GrayLevel[0.5], 
             AbsoluteThickness[1.5], 
             Opacity[0.66]}], {}, {}], 
          
          RectangleBox[{2.4, 0.}, {2.6, 0.028994201159768022`}, 
           "RoundingRadius" -> 0]},
         ImageSizeCache->{{135.7226345083488, 141.17532467532467`}, {
          38.27360829359388, 45.2050983124842}}],
        StatusArea[#, 0.028994201159768022`]& ,
        TagBoxNote->"0.028994201159768022"],
       StyleBox["0.028994201159768022`", {
         GrayLevel[0]}, StripOnInput -> False]],
      Annotation[#, 
       Style[0.028994201159768022`, {
         GrayLevel[0]}], "Tooltip"]& ], 
     TagBox[
      TooltipBox[
       TagBox[
        DynamicBox[{
          FEPrivate`If[
           CurrentValue["MouseOver"], 
           EdgeForm[{
             GrayLevel[0.5], 
             AbsoluteThickness[1.5], 
             Opacity[0.66]}], {}, {}], 
          
          RectangleBox[{2.6, 0.}, {2.8, 0.01699660067986405}, 
           "RoundingRadius" -> 0]},
         ImageSizeCache->{{140.17532467532467`, 145.62801484230056`}, {
          40.728017956582974`, 45.2050983124842}}],
        StatusArea[#, 0.01699660067986405]& ,
        TagBoxNote->"0.01699660067986405"],
       StyleBox["0.01699660067986405`", {
         GrayLevel[0]}, StripOnInput -> False]],
      Annotation[#, 
       Style[0.01699660067986405, {
         GrayLevel[0]}], "Tooltip"]& ], 
     TagBox[
      TooltipBox[
       TagBox[
        DynamicBox[{
          FEPrivate`If[
           CurrentValue["MouseOver"], 
           EdgeForm[{
             GrayLevel[0.5], 
             AbsoluteThickness[1.5], 
             Opacity[0.66]}], {}, {}], 
          
          RectangleBox[{2.8, 0.}, {3., 0.003999200159968003}, 
           "RoundingRadius" -> 0]},
         ImageSizeCache->{{144.62801484230056`, 150.08070500927641`}, {
          43.386961758154506`, 45.2050983124842}}],
        StatusArea[#, 0.003999200159968003]& ,
        TagBoxNote->"0.003999200159968003"],
       StyleBox["0.003999200159968003`", {
         GrayLevel[0]}, StripOnInput -> False]],
      Annotation[#, 
       Style[0.003999200159968003, {
         GrayLevel[0]}], "Tooltip"]& ], {}, 
     TagBox[
      TooltipBox[
       TagBox[
        DynamicBox[{
          FEPrivate`If[
           CurrentValue["MouseOver"], 
           EdgeForm[{
             GrayLevel[0.5], 
             AbsoluteThickness[1.5], 
             Opacity[0.66]}], {}, {}], 
          
          RectangleBox[{3.2, 0.}, {3.4, 0.007998400319936023}, 
           "RoundingRadius" -> 0]},
         ImageSizeCache->{{153.53339517625233`, 158.98608534322818`}, {
          42.5688252038248, 45.2050983124842}}],
        StatusArea[#, 0.007998400319936023]& ,
        TagBoxNote->"0.007998400319936023"],
       StyleBox["0.007998400319936023`", {
         GrayLevel[0]}, StripOnInput -> False]],
      Annotation[#, 
       Style[0.007998400319936023, {
         GrayLevel[0]}], "Tooltip"]& ], 
     TagBox[
      TooltipBox[
       TagBox[
        DynamicBox[{
          FEPrivate`If[
           CurrentValue["MouseOver"], 
           EdgeForm[{
             GrayLevel[0.5], 
             AbsoluteThickness[1.5], 
             Opacity[0.66]}], {}, {}], 
          
          RectangleBox[{3.4, 0.}, {3.6, 0.004999000199960003}, 
           "RoundingRadius" -> 0]},
         ImageSizeCache->{{157.98608534322818`, 163.4387755102041}, {
          43.18242761957208, 45.2050983124842}}],
        StatusArea[#, 0.004999000199960003]& ,
        TagBoxNote->"0.004999000199960003"],
       StyleBox["0.004999000199960003`", {
         GrayLevel[0]}, StripOnInput -> False]],
      Annotation[#, 
       Style[0.004999000199960003, {
         GrayLevel[0]}], "Tooltip"]& ]}, {}, {}}, {{}, {}, 
    {RGBColor[1, 0, 0], Thickness[Large], LineBox[CompressedData["
1:eJwV13c8Ve8fAHDude/l0iJlRimjJElmfE4hOzOlYfOVki2RvbNHyCxbsleI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      "]]}}},
  AspectRatio->NCache[GoldenRatio^(-1), 0.6180339887498948],
  Axes->{True, True},
  AxesLabel->{
    FormBox["\"x\"", TraditionalForm], 
    FormBox["\"freq\"", TraditionalForm]},
  AxesOrigin->{-3., 0},
  FrameTicks->{{Automatic, Automatic}, {Automatic, Automatic}},
  GridLines->{None, None},
  PlotRange->NCache[{{-3, 
      Rational[18, 5]}, {All, All}}, {{-3, 3.6}, {All, All}}],
  PlotRangePadding->{{
     Scaled[0.02], 
     Scaled[0.02]}, {
     Scaled[0.02], 
     Scaled[0.1]}},
  Ticks->{Automatic, Automatic}]], "Output",
 ImageSize->{184, 127},
 ImageMargins->{{0, 0}, {0, 0}},
 ImageRegion->{{0, 1}, {0, 1}},
 CellLabel->"Out[5]=",
 CellID->261084222]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 InterpretationBox[Cell[TextData[{
   "Neat Examples",
   "\[NonBreakingSpace]\[NonBreakingSpace]",
   Cell["(1)", "ExampleCount"]
  }], "ExampleSection"],
  $Line = 0; Null]], "ExampleSection",
 CellID->589267740],

Cell["This samples bivariate truncted student T distribution", "ExampleText",
 CellID->73418448],

Cell[CellGroupData[{

Cell[BoxData[{
 RowBox[{"Clear", "[", 
  RowBox[{
  "rho", ",", "deg", ",", "scale", ",", "n", ",", "xmin", ",", "logpdf", ",", 
   "logpdfjump", ",", "randomjump", ",", "res"}], 
  "]"}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{"rho", "=", "0.99"}], ";"}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{"deg", "=", "3"}], ";"}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{"scale", "=", 
   RowBox[{"2.234", "/", 
    RowBox[{"Sqrt", "[", "2", "]"}]}]}], ";"}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{"n", "=", "5000"}], ";"}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{
   RowBox[{"xmin", "=", 
    RowBox[{"-", "1"}]}], ";"}], " ", 
  RowBox[{"(*", 
   RowBox[{"<", "0"}], "*)"}]}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{"logpdf", "=", 
   RowBox[{"Function", "[", 
    RowBox[{"x", ",", 
     RowBox[{"Log", "[", 
      RowBox[{"PDF", "[", 
       RowBox[{
        RowBox[{"MultivariateTDistribution", "[", 
         RowBox[{
          RowBox[{"{", 
           RowBox[{
            RowBox[{"{", 
             RowBox[{"1", ",", "rho"}], "}"}], ",", 
            RowBox[{"{", 
             RowBox[{"rho", ",", "1"}], "}"}]}], "}"}], ",", "deg"}], "]"}], 
        ",", "x"}], "]"}], "]"}]}], "]"}]}], ";"}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{"logpdfjump", "=", 
   RowBox[{"If", "[", 
    RowBox[{"True", ",", 
     RowBox[{"(", 
      RowBox[{"0", "&"}], ")"}], ",", 
     RowBox[{"Function", "[", 
      RowBox[{
       RowBox[{"{", 
        RowBox[{"xt", ",", "x"}], "}"}], ",", 
       RowBox[{"Log", "[", 
        RowBox[{"PDF", "[", 
         RowBox[{
          RowBox[{"MultinormalDistribution", "[", 
           RowBox[{"xt", ",", 
            RowBox[{
             RowBox[{"scale", "^", "2"}], " ", 
             RowBox[{"{", 
              RowBox[{
               RowBox[{"{", 
                RowBox[{"1", ",", "rho"}], "}"}], ",", 
               RowBox[{"{", 
                RowBox[{"rho", ",", "1"}], "}"}]}], "}"}]}]}], "]"}], ",", 
          "x"}], "]"}], "]"}]}], "]"}]}], "\[IndentingNewLine]", "]"}]}], 
  ";"}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{"randomjump", "=", 
   RowBox[{"Function", "[", 
    RowBox[{"xt", ",", 
     RowBox[{"RandomReal", "[", 
      RowBox[{"MultinormalDistribution", "[", 
       RowBox[{"xt", ",", 
        RowBox[{
         RowBox[{"scale", "^", "2"}], " ", 
         RowBox[{"{", 
          RowBox[{
           RowBox[{"{", 
            RowBox[{"1", ",", "rho"}], "}"}], ",", 
           RowBox[{"{", 
            RowBox[{"rho", ",", "1"}], "}"}]}], "}"}]}]}], "]"}], "]"}]}], 
    "]"}]}], ";"}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{
   RowBox[{"res", "=", 
    RowBox[{"Metropolis", "[", 
     RowBox[{"logpdf", ",", "logpdfjump", ",", "randomjump", ",", 
      RowBox[{"{", 
       RowBox[{"0", ",", "0"}], "}"}], ",", 
      RowBox[{"2", " ", "n"}], ",", 
      RowBox[{"IsReturnLast", "\[Rule]", "False"}], ",", 
      RowBox[{"DomainConstraint", "\[Rule]", 
       RowBox[{"(", 
        RowBox[{
         RowBox[{
          RowBox[{
           RowBox[{"#", "[", 
            RowBox[{"[", "1", "]"}], "]"}], "\[GreaterEqual]", "xmin"}], "&&", 
          RowBox[{
           RowBox[{"#", "[", 
            RowBox[{"[", "2", "]"}], "]"}], "\[GreaterEqual]", "xmin"}]}], 
         "&"}], ")"}]}]}], "]"}]}], ";"}], "//", "Timing"}]}], "Input",
 CellLabel->"In[1]:=",
 CellID->251956668],

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{"4.134`", ",", "Null"}], "}"}]], "Output",
 ImageSize->{92, 15},
 ImageMargins->{{0, 0}, {0, 0}},
 ImageRegion->{{0, 1}, {0, 1}},
 CellLabel->"Out[1]=",
 CellID->439528226]
}, Open  ]],

Cell["Mean accept ratio:", "ExampleText",
 CellID->183647242],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"N", "[", 
  RowBox[{"Mean", "[", 
   RowBox[{"res", "[", 
    RowBox[{"[", 
     RowBox[{
      RowBox[{"n", ";;"}], ",", 
      RowBox[{"-", "1"}]}], "]"}], "]"}], "]"}], "]"}]], "Input",
 CellLabel->"In[2]:=",
 CellID->265777506],

Cell[BoxData["0.36538885612558836`"], "Output",
 ImageSize->{60, 15},
 ImageMargins->{{0, 0}, {0, 0}},
 ImageRegion->{{0, 1}, {0, 1}},
 CellLabel->"Out[2]=",
 CellID->738619732]
}, Open  ]],

Cell["Time series", "ExampleText",
 CellID->193398859],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"Table", "[", 
  RowBox[{
   RowBox[{"ListLinePlot", "[", 
    RowBox[{
     RowBox[{"res", "[", 
      RowBox[{"[", 
       RowBox[{"All", ",", "1", ",", "i"}], "]"}], "]"}], ",", 
     RowBox[{"PlotRange", "\[Rule]", "All"}], ",", 
     RowBox[{"AxesLabel", "\[Rule]", 
      RowBox[{"{", 
       RowBox[{"\"\<t\>\"", ",", 
        RowBox[{"\"\<x\>\"", "<>", 
         RowBox[{"ToString", "[", "i", "]"}]}]}], "}"}]}]}], "]"}], ",", 
   RowBox[{"{", 
    RowBox[{"i", ",", "2"}], "}"}]}], "]"}]], "Input",
 CellLabel->"In[3]:=",
 CellID->309231733],

Cell[BoxData[
 GraphicsBox[
  TagBox[RasterBox[CompressedData["
1:eJztnX3MHVWdx5/N/r/ZjSYrMf4h8S0rGFFKXAxcoQVaSgFbysNLUWhrLWrp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    "], {{0, 124}, {385, 0}}, {0, 255},
    ColorFunction->RGBColor],
   BoxForm`ImageTag["Byte", ColorSpace -> "RGB", Interleaving -> True],
   Selectable->False],
  BaseStyle->"ImageGraphics",
  ImageSize->Automatic,
  ImageSizeRaw->{385, 124},
  PlotRange->{{0, 385}, {0, 124}}]], "Output",
 ImageSize->{389, 124},
 ImageMargins->{{0, 0}, {0, 0}},
 ImageRegion->{{0, 1}, {0, 1}},
 CellLabel->"Out[3]=",
 CellID->66405875]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"ListPlot", "[", 
  RowBox[{
   RowBox[{"res", "[", 
    RowBox[{"[", 
     RowBox[{
      RowBox[{"n", ";;"}], ",", "1"}], "]"}], "]"}], ",", 
   RowBox[{"PlotRange", "\[Rule]", "All"}], ",", 
   RowBox[{"AxesLabel", "\[Rule]", 
    RowBox[{"{", 
     RowBox[{"\"\<x1\>\"", ",", "\"\<x2\>\""}], "}"}]}]}], "]"}]], "Input",
 CellLabel->"In[4]:=",
 CellID->357276416],

Cell[BoxData[
 GraphicsBox[
  TagBox[RasterBox[CompressedData["
1:eJztnVtv1EYYhlfqL+nfQEr2kAMhSQkhnAXioGpbpAoQJeUCoV71oKqV9qKV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    "], {{0, 123}, {180, 0}}, {0, 255},
    ColorFunction->RGBColor],
   BoxForm`ImageTag["Byte", ColorSpace -> "RGB", Interleaving -> True],
   Selectable->False],
  BaseStyle->"ImageGraphics",
  ImageSize->Automatic,
  ImageSizeRaw->{180, 123},
  PlotRange->{{0, 180}, {0, 123}}]], "Output",
 ImageSize->{184, 123},
 ImageMargins->{{0, 0}, {0, 0}},
 ImageRegion->{{0, 1}, {0, 1}},
 CellLabel->"Out[4]=",
 CellID->2079456]
}, Open  ]],

Cell["\<\
Marginal simulated distribution vs marginal target distribution (red line):\
\>", "ExampleText",
 CellID->395721816],

Cell[CellGroupData[{

Cell[BoxData[{
 RowBox[{
  RowBox[{"Clear", "[", 
   RowBox[{"x", ",", "f0", ",", "c", ",", "f", ",", "data"}], "]"}], 
  ";"}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{
   RowBox[{
    RowBox[{"f0", "[", "x_", "]"}], "=", 
    RowBox[{"Integrate", "[", 
     RowBox[{
      RowBox[{"Exp", "[", 
       RowBox[{"logpdf", "[", 
        RowBox[{"{", 
         RowBox[{"x1", ",", "x"}], "}"}], "]"}], "]"}], ",", 
      RowBox[{"{", 
       RowBox[{"x1", ",", "xmin", ",", "Infinity"}], "}"}], ",", 
      RowBox[{"Assumptions", "->", 
       RowBox[{"x", "\[GreaterEqual]", "xmin"}]}]}], "]"}]}], ";"}], "//", 
  "Timing"}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{
   RowBox[{"c", "=", 
    RowBox[{"NIntegrate", "[", 
     RowBox[{
      RowBox[{"f0", "[", "x", "]"}], ",", 
      RowBox[{"{", 
       RowBox[{"x", ",", "xmin", ",", "Infinity"}], "}"}]}], "]"}]}], ";"}], 
  " "}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{
   RowBox[{"f", "[", "x_", "]"}], ":=", 
   RowBox[{
    RowBox[{"f0", "[", "x", "]"}], "/", "c"}]}], ";"}], "\[IndentingNewLine]", 
 RowBox[{"Table", "[", "\[IndentingNewLine]", 
  RowBox[{
   RowBox[{
    RowBox[{"data", "=", 
     RowBox[{"res", "[", 
      RowBox[{"[", 
       RowBox[{
        RowBox[{"n", ";;"}], ",", "1", ",", "i"}], "]"}], "]"}]}], ";", 
    "\[IndentingNewLine]", 
    RowBox[{"Show", "[", 
     RowBox[{
      RowBox[{"Histogram", "[", 
       RowBox[{"data", ",", "Automatic", ",", "\"\<ProbabilityDensity\>\""}], 
       "]"}], ",", 
      RowBox[{"Plot", "[", 
       RowBox[{
        RowBox[{"f", "[", "x", "]"}], ",", 
        RowBox[{"{", 
         RowBox[{"x", ",", 
          RowBox[{"Min", "[", "data", "]"}], ",", 
          RowBox[{"Max", "[", "data", "]"}]}], "}"}], ",", 
        RowBox[{"PlotRange", "\[Rule]", "All"}], ",", 
        RowBox[{"PlotStyle", "\[Rule]", 
         RowBox[{"{", 
          RowBox[{"Red", ",", "Thick"}], "}"}]}]}], "]"}], ",", 
      RowBox[{"AxesLabel", "\[Rule]", 
       RowBox[{"{", 
        RowBox[{
         RowBox[{"\"\<x\>\"", "<>", 
          RowBox[{"ToString", "[", "i", "]"}]}], ",", "\"\<freq\>\""}], 
        "}"}]}]}], "]"}]}], ",", 
   RowBox[{"{", 
    RowBox[{"i", ",", "2"}], "}"}]}], "]"}]}], "Input",
 CellLabel->"In[5]:=",
 CellID->10673344],

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{"6.162000000000002`", ",", "Null"}], "}"}]], "Output",
 ImageSize->{92, 15},
 ImageMargins->{{0, 0}, {0, 0}},
 ImageRegion->{{0, 1}, {0, 1}},
 CellLabel->"Out[5]=",
 CellID->32820248],

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{
   GraphicsBox[{
     {RGBColor[0.798413061722744, 0.824719615472648, 0.968322270542458], 
      EdgeForm[Opacity[0.7]], {}, 
      {RGBColor[0.798413061722744, 0.824719615472648, 0.968322270542458], 
       EdgeForm[Opacity[0.7]], 
       TagBox[
        TooltipBox[
         TagBox[
          DynamicBox[{
            FEPrivate`If[
             CurrentValue["MouseOver"], 
             EdgeForm[{
               GrayLevel[0.5], 
               AbsoluteThickness[1.5], 
               Opacity[0.66]}], {}, {}], 
            
            RectangleBox[{-1., 0.}, {-0.8, 0.25194961007798444`}, 
             "RoundingRadius" -> 0]},
           
           ImageSizeCache->{{15.500000000000007`, 18.529154518950442`}, {
            6.70814546643264, 44.11492836873474}}],
          StatusArea[#, 0.25194961007798444`]& ,
          TagBoxNote->"0.25194961007798444"],
         StyleBox["0.25194961007798444`", {
           GrayLevel[0]}, StripOnInput -> False]],
        Annotation[#, 
         Style[0.25194961007798444`, {
           GrayLevel[0]}], "Tooltip"]& ], 
       TagBox[
        TooltipBox[
         TagBox[
          DynamicBox[{
            FEPrivate`If[
             CurrentValue["MouseOver"], 
             EdgeForm[{
               GrayLevel[0.5], 
               AbsoluteThickness[1.5], 
               Opacity[0.66]}], {}, {}], 
            
            RectangleBox[{-0.8, 0.}, {-0.6, 0.21595680863827227`}, 
             "RoundingRadius" -> 0]},
           
           ImageSizeCache->{{17.529154518950442`, 20.55830903790088}, {
            11.909114452475812`, 44.11492836873474}}],
          StatusArea[#, 0.21595680863827227`]& ,
          TagBoxNote->"0.21595680863827227"],
         StyleBox["0.21595680863827227`", {
           GrayLevel[0]}, StripOnInput -> False]],
        Annotation[#, 
         Style[0.21595680863827227`, {
           GrayLevel[0]}], "Tooltip"]& ], 
       TagBox[
        TooltipBox[
         TagBox[
          DynamicBox[{
            FEPrivate`If[
             CurrentValue["MouseOver"], 
             EdgeForm[{
               GrayLevel[0.5], 
               AbsoluteThickness[1.5], 
               Opacity[0.66]}], {}, {}], 
            
            RectangleBox[{-0.6, 0.}, {-0.4, 0.3799240151969607}, 
             "RoundingRadius" -> 0]},
           
           ImageSizeCache->{{19.55830903790088, 
            22.58746355685132}, {-11.784188706165253`, 44.11492836873474}}],
          StatusArea[#, 0.3799240151969607]& ,
          TagBoxNote->"0.3799240151969607"],
         StyleBox["0.3799240151969607`", {
           GrayLevel[0]}, StripOnInput -> False]],
        Annotation[#, 
         Style[0.3799240151969607, {
           GrayLevel[0]}], "Tooltip"]& ], 
       TagBox[
        TooltipBox[
         TagBox[
          DynamicBox[{
            FEPrivate`If[
             CurrentValue["MouseOver"], 
             EdgeForm[{
               GrayLevel[0.5], 
               AbsoluteThickness[1.5], 
               Opacity[0.66]}], {}, {}], 
            
            RectangleBox[{-0.4, 0.}, {-0.2, 0.4159168166366727}, 
             "RoundingRadius" -> 0]},
           
           ImageSizeCache->{{21.58746355685132, 
            24.616618075801753`}, {-16.985157692208404`, 44.11492836873474}}],
          StatusArea[#, 0.4159168166366727]& ,
          TagBoxNote->"0.4159168166366727"],
         StyleBox["0.4159168166366727`", {
           GrayLevel[0]}, StripOnInput -> False]],
        Annotation[#, 
         Style[0.4159168166366727, {
           GrayLevel[0]}], "Tooltip"]& ], 
       TagBox[
        TooltipBox[
         TagBox[
          DynamicBox[{
            FEPrivate`If[
             CurrentValue["MouseOver"], 
             EdgeForm[{
               GrayLevel[0.5], 
               AbsoluteThickness[1.5], 
               Opacity[0.66]}], {}, {}], 
            RectangleBox[{-0.2, 0.}, {0., 0.43591281743651267`}, 
             "RoundingRadius" -> 0]},
           
           ImageSizeCache->{{23.616618075801753`, 
            26.64577259475219}, {-19.874584906676816`, 44.11492836873474}}],
          StatusArea[#, 0.43591281743651267`]& ,
          TagBoxNote->"0.43591281743651267"],
         StyleBox["0.43591281743651267`", {
           GrayLevel[0]}, StripOnInput -> False]],
        Annotation[#, 
         Style[0.43591281743651267`, {
           GrayLevel[0]}], "Tooltip"]& ], 
       TagBox[
        TooltipBox[
         TagBox[
          DynamicBox[{
            FEPrivate`If[
             CurrentValue["MouseOver"], 
             EdgeForm[{
               GrayLevel[0.5], 
               AbsoluteThickness[1.5], 
               Opacity[0.66]}], {}, {}], 
            
            RectangleBox[{0., 0.}, {0.2, 0.5568886222755449}, 
             "RoundingRadius" -> 0]},
           
           ImageSizeCache->{{25.64577259475219, 
            28.67492711370263}, {-37.35561955421076, 44.11492836873474}}],
          StatusArea[#, 0.5568886222755449]& ,
          TagBoxNote->"0.5568886222755449"],
         StyleBox["0.5568886222755449`", {
           GrayLevel[0]}, StripOnInput -> False]],
        Annotation[#, 
         Style[0.5568886222755449, {
           GrayLevel[0]}], "Tooltip"]& ], 
       TagBox[
        TooltipBox[
         TagBox[
          DynamicBox[{
            FEPrivate`If[
             CurrentValue["MouseOver"], 
             EdgeForm[{
               GrayLevel[0.5], 
               AbsoluteThickness[1.5], 
               Opacity[0.66]}], {}, {}], 
            
            RectangleBox[{0.2, 0.}, {0.4, 0.42191561687662466`}, 
             "RoundingRadius" -> 0]},
           
           ImageSizeCache->{{27.67492711370263, 
            30.704081632653065`}, {-17.85198585654892, 44.11492836873474}}],
          StatusArea[#, 0.42191561687662466`]& ,
          TagBoxNote->"0.42191561687662466"],
         StyleBox["0.42191561687662466`", {
           GrayLevel[0]}, StripOnInput -> False]],
        Annotation[#, 
         Style[0.42191561687662466`, {
           GrayLevel[0]}], "Tooltip"]& ], 
       TagBox[
        TooltipBox[
         TagBox[
          DynamicBox[{
            FEPrivate`If[
             CurrentValue["MouseOver"], 
             EdgeForm[{
               GrayLevel[0.5], 
               AbsoluteThickness[1.5], 
               Opacity[0.66]}], {}, {}], 
            
            RectangleBox[{0.4, 0.}, {0.6, 0.2999400119976005}, 
             "RoundingRadius" -> 0]},
           
           ImageSizeCache->{{29.704081632653065`, 
            32.733236151603506`}, {-0.22647984829156798`, 44.11492836873474}}],
          StatusArea[#, 0.2999400119976005]& ,
          TagBoxNote->"0.2999400119976005"],
         StyleBox["0.2999400119976005`", {
           GrayLevel[0]}, StripOnInput -> False]],
        Annotation[#, 
         Style[0.2999400119976005, {
           GrayLevel[0]}], "Tooltip"]& ], 
       TagBox[
        TooltipBox[
         TagBox[
          DynamicBox[{
            FEPrivate`If[
             CurrentValue["MouseOver"], 
             EdgeForm[{
               GrayLevel[0.5], 
               AbsoluteThickness[1.5], 
               Opacity[0.66]}], {}, {}], 
            
            RectangleBox[{0.6, 0.}, {0.8, 0.3229354129174164}, 
             "RoundingRadius" -> 0]},
           
           ImageSizeCache->{{31.733236151603506`, 
            34.76239067055394}, {-3.549321144930225, 44.11492836873474}}],
          StatusArea[#, 0.3229354129174164]& ,
          TagBoxNote->"0.3229354129174164"],
         StyleBox["0.3229354129174164`", {
           GrayLevel[0]}, StripOnInput -> False]],
        Annotation[#, 
         Style[0.3229354129174164, {
           GrayLevel[0]}], "Tooltip"]& ], 
       TagBox[
        TooltipBox[
         TagBox[
          DynamicBox[{
            FEPrivate`If[
             CurrentValue["MouseOver"], 
             EdgeForm[{
               GrayLevel[0.5], 
               AbsoluteThickness[1.5], 
               Opacity[0.66]}], {}, {}], 
            
            RectangleBox[{0.8, 0.}, {1., 0.2869426114777045}, 
             "RoundingRadius" -> 0]},
           
           ImageSizeCache->{{33.76239067055394, 36.791545189504376`}, {
            1.6516478411129043`, 44.11492836873474}}],
          StatusArea[#, 0.2869426114777045]& ,
          TagBoxNote->"0.2869426114777045"],
         StyleBox["0.2869426114777045`", {
           GrayLevel[0]}, StripOnInput -> False]],
        Annotation[#, 
         Style[0.2869426114777045, {
           GrayLevel[0]}], "Tooltip"]& ], 
       TagBox[
        TooltipBox[
         TagBox[
          DynamicBox[{
            FEPrivate`If[
             CurrentValue["MouseOver"], 
             EdgeForm[{
               GrayLevel[0.5], 
               AbsoluteThickness[1.5], 
               Opacity[0.66]}], {}, {}], 
            
            RectangleBox[{1., 0.}, {1.2, 0.25394921015796845`}, 
             "RoundingRadius" -> 0]},
           
           ImageSizeCache->{{35.791545189504376`, 38.82069970845481}, {
            6.419202744985796, 44.11492836873474}}],
          StatusArea[#, 0.25394921015796845`]& ,
          TagBoxNote->"0.25394921015796845"],
         StyleBox["0.25394921015796845`", {
           GrayLevel[0]}, StripOnInput -> False]],
        Annotation[#, 
         Style[0.25394921015796845`, {
           GrayLevel[0]}], "Tooltip"]& ], 
       TagBox[
        TooltipBox[
         TagBox[
          DynamicBox[{
            FEPrivate`If[
             CurrentValue["MouseOver"], 
             EdgeForm[{
               GrayLevel[0.5], 
               AbsoluteThickness[1.5], 
               Opacity[0.66]}], {}, {}], 
            
            RectangleBox[{1.2, 0.}, {1.4, 0.23195360927814443`}, 
             "RoundingRadius" -> 0]},
           
           ImageSizeCache->{{37.82069970845481, 40.84985422740525}, {
            9.59757268090106, 44.11492836873474}}],
          StatusArea[#, 0.23195360927814443`]& ,
          TagBoxNote->"0.23195360927814443"],
         StyleBox["0.23195360927814443`", {
           GrayLevel[0]}, StripOnInput -> False]],
        Annotation[#, 
         Style[0.23195360927814443`, {
           GrayLevel[0]}], "Tooltip"]& ], 
       TagBox[
        TooltipBox[
         TagBox[
          DynamicBox[{
            FEPrivate`If[
             CurrentValue["MouseOver"], 
             EdgeForm[{
               GrayLevel[0.5], 
               AbsoluteThickness[1.5], 
               Opacity[0.66]}], {}, {}], 
            
            RectangleBox[{1.4, 0.}, {1.6, 0.12097580483903209`}, 
             "RoundingRadius" -> 0]},
           
           ImageSizeCache->{{39.84985422740525, 42.879008746355694`}, {
            25.633893721200813`, 44.11492836873474}}],
          StatusArea[#, 0.12097580483903209`]& ,
          TagBoxNote->"0.12097580483903209"],
         StyleBox["0.12097580483903209`", {
           GrayLevel[0]}, StripOnInput -> False]],
        Annotation[#, 
         Style[0.12097580483903209`, {
           GrayLevel[0]}], "Tooltip"]& ], 
       TagBox[
        TooltipBox[
         TagBox[
          DynamicBox[{
            FEPrivate`If[
             CurrentValue["MouseOver"], 
             EdgeForm[{
               GrayLevel[0.5], 
               AbsoluteThickness[1.5], 
               Opacity[0.66]}], {}, {}], 
            
            RectangleBox[{1.6, 0.}, {1.8, 0.14097180563887224`}, 
             "RoundingRadius" -> 0]},
           
           ImageSizeCache->{{41.879008746355694`, 44.90816326530613}, {
            22.744466506732376`, 44.11492836873474}}],
          StatusArea[#, 0.14097180563887224`]& ,
          TagBoxNote->"0.14097180563887224"],
         StyleBox["0.14097180563887224`", {
           GrayLevel[0]}, StripOnInput -> False]],
        Annotation[#, 
         Style[0.14097180563887224`, {
           GrayLevel[0]}], "Tooltip"]& ], 
       TagBox[
        TooltipBox[
         TagBox[
          DynamicBox[{
            FEPrivate`If[
             CurrentValue["MouseOver"], 
             EdgeForm[{
               GrayLevel[0.5], 
               AbsoluteThickness[1.5], 
               Opacity[0.66]}], {}, {}], 
            
            RectangleBox[{1.8, 0.}, {2., 0.1119776044791042}, 
             "RoundingRadius" -> 0]},
           
           ImageSizeCache->{{43.90816326530613, 46.937317784256564`}, {
            26.934135967711583`, 44.11492836873474}}],
          StatusArea[#, 0.1119776044791042]& ,
          TagBoxNote->"0.1119776044791042"],
         StyleBox["0.1119776044791042`", {
           GrayLevel[0]}, StripOnInput -> False]],
        Annotation[#, 
         Style[0.1119776044791042, {
           GrayLevel[0]}], "Tooltip"]& ], 
       TagBox[
        TooltipBox[
         TagBox[
          DynamicBox[{
            FEPrivate`If[
             CurrentValue["MouseOver"], 
             EdgeForm[{
               GrayLevel[0.5], 
               AbsoluteThickness[1.5], 
               Opacity[0.66]}], {}, {}], 
            
            RectangleBox[{2., 0.}, {2.2, 0.10297940411917607`}, 
             "RoundingRadius" -> 0]},
           
           ImageSizeCache->{{45.937317784256564`, 48.966472303207}, {
            28.23437821422239, 44.11492836873474}}],
          StatusArea[#, 0.10297940411917607`]& ,
          TagBoxNote->"0.10297940411917607"],
         StyleBox["0.10297940411917607`", {
           GrayLevel[0]}, StripOnInput -> False]],
        Annotation[#, 
         Style[0.10297940411917607`, {
           GrayLevel[0]}], "Tooltip"]& ], 
       TagBox[
        TooltipBox[
         TagBox[
          DynamicBox[{
            FEPrivate`If[
             CurrentValue["MouseOver"], 
             EdgeForm[{
               GrayLevel[0.5], 
               AbsoluteThickness[1.5], 
               Opacity[0.66]}], {}, {}], 
            
            RectangleBox[{2.2, 0.}, {2.4, 0.0629874025194962}, 
             "RoundingRadius" -> 0]},
           
           ImageSizeCache->{{47.966472303207, 50.99562682215743}, {
            34.0132326431592, 44.11492836873474}}],
          StatusArea[#, 0.0629874025194962]& ,
          TagBoxNote->"0.0629874025194962"],
         StyleBox["0.0629874025194962`", {
           GrayLevel[0]}, StripOnInput -> False]],
        Annotation[#, 
         Style[0.0629874025194962, {
           GrayLevel[0]}], "Tooltip"]& ], 
       TagBox[
        TooltipBox[
         TagBox[
          DynamicBox[{
            FEPrivate`If[
             CurrentValue["MouseOver"], 
             EdgeForm[{
               GrayLevel[0.5], 
               AbsoluteThickness[1.5], 
               Opacity[0.66]}], {}, {}], 
            
            RectangleBox[{2.4, 0.}, {2.6, 0.03899220155968803}, 
             "RoundingRadius" -> 0]},
           
           ImageSizeCache->{{49.99562682215743, 53.024781341107875`}, {
            37.48054530052132, 44.11492836873474}}],
          StatusArea[#, 0.03899220155968803]& ,
          TagBoxNote->"0.03899220155968803"],
         StyleBox["0.03899220155968803`", {
           GrayLevel[0]}, StripOnInput -> False]],
        Annotation[#, 
         Style[0.03899220155968803, {
           GrayLevel[0]}], "Tooltip"]& ], 
       TagBox[
        TooltipBox[
         TagBox[
          DynamicBox[{
            FEPrivate`If[
             CurrentValue["MouseOver"], 
             EdgeForm[{
               GrayLevel[0.5], 
               AbsoluteThickness[1.5], 
               Opacity[0.66]}], {}, {}], 
            
            RectangleBox[{2.6, 0.}, {2.8, 0.059988002399520186`}, 
             "RoundingRadius" -> 0]},
           
           ImageSizeCache->{{52.024781341107875`, 55.05393586005831}, {
            34.446646725329465`, 44.11492836873474}}],
          StatusArea[#, 0.059988002399520186`]& ,
          TagBoxNote->"0.059988002399520186"],
         StyleBox["0.059988002399520186`", {
           GrayLevel[0]}, StripOnInput -> False]],
        Annotation[#, 
         Style[0.059988002399520186`, {
           GrayLevel[0]}], "Tooltip"]& ], 
       TagBox[
        TooltipBox[
         TagBox[
          DynamicBox[{
            FEPrivate`If[
             CurrentValue["MouseOver"], 
             EdgeForm[{
               GrayLevel[0.5], 
               AbsoluteThickness[1.5], 
               Opacity[0.66]}], {}, {}], 
            
            RectangleBox[{2.8, 0.}, {3., 0.04399120175964803}, 
             "RoundingRadius" -> 0]},
           
           ImageSizeCache->{{54.05393586005831, 57.08309037900875}, {
            36.75818849690422, 44.11492836873474}}],
          StatusArea[#, 0.04399120175964803]& ,
          TagBoxNote->"0.04399120175964803"],
         StyleBox["0.04399120175964803`", {
           GrayLevel[0]}, StripOnInput -> False]],
        Annotation[#, 
         Style[0.04399120175964803, {
           GrayLevel[0]}], "Tooltip"]& ], 
       TagBox[
        TooltipBox[
         TagBox[
          DynamicBox[{
            FEPrivate`If[
             CurrentValue["MouseOver"], 
             EdgeForm[{
               GrayLevel[0.5], 
               AbsoluteThickness[1.5], 
               Opacity[0.66]}], {}, {}], 
            
            RectangleBox[{3., 0.}, {3.2, 0.028994201159768022`}, 
             "RoundingRadius" -> 0]},
           
           ImageSizeCache->{{56.08309037900875, 59.11224489795919}, {
            38.925258907755534`, 44.11492836873474}}],
          StatusArea[#, 0.028994201159768022`]& ,
          TagBoxNote->"0.028994201159768022"],
         StyleBox["0.028994201159768022`", {
           GrayLevel[0]}, StripOnInput -> False]],
        Annotation[#, 
         Style[0.028994201159768022`, {
           GrayLevel[0]}], "Tooltip"]& ], 
       TagBox[
        TooltipBox[
         TagBox[
          DynamicBox[{
            FEPrivate`If[
             CurrentValue["MouseOver"], 
             EdgeForm[{
               GrayLevel[0.5], 
               AbsoluteThickness[1.5], 
               Opacity[0.66]}], {}, {}], 
            
            RectangleBox[{3.2, 0.}, {3.4, 0.02399520095980807}, 
             "RoundingRadius" -> 0]},
           
           ImageSizeCache->{{58.11224489795919, 61.14139941690962}, {
            39.64761571137263, 44.11492836873474}}],
          StatusArea[#, 0.02399520095980807]& ,
          TagBoxNote->"0.02399520095980807"],
         StyleBox["0.02399520095980807`", {
           GrayLevel[0]}, StripOnInput -> False]],
        Annotation[#, 
         Style[0.02399520095980807, {
           GrayLevel[0]}], "Tooltip"]& ], 
       TagBox[
        TooltipBox[
         TagBox[
          DynamicBox[{
            FEPrivate`If[
             CurrentValue["MouseOver"], 
             EdgeForm[{
               GrayLevel[0.5], 
               AbsoluteThickness[1.5], 
               Opacity[0.66]}], {}, {}], 
            
            RectangleBox[{3.4, 0.}, {3.6, 0.021995600879824016`}, 
             "RoundingRadius" -> 0]},
           
           ImageSizeCache->{{60.14139941690962, 63.170553935860056`}, {
            39.93655843281948, 44.11492836873474}}],
          StatusArea[#, 0.021995600879824016`]& ,
          TagBoxNote->"0.021995600879824016"],
         StyleBox["0.021995600879824016`", {
           GrayLevel[0]}, StripOnInput -> False]],
        Annotation[#, 
         Style[0.021995600879824016`, {
           GrayLevel[0]}], "Tooltip"]& ], 
       TagBox[
        TooltipBox[
         TagBox[
          DynamicBox[{
            FEPrivate`If[
             CurrentValue["MouseOver"], 
             EdgeForm[{
               GrayLevel[0.5], 
               AbsoluteThickness[1.5], 
               Opacity[0.66]}], {}, {}], 
            
            RectangleBox[{3.6, 0.}, {3.8, 0.024995000999800072`}, 
             "RoundingRadius" -> 0]},
           
           ImageSizeCache->{{62.170553935860056`, 65.19970845481049}, {
            39.50314435064921, 44.11492836873474}}],
          StatusArea[#, 0.024995000999800072`]& ,
          TagBoxNote->"0.024995000999800072"],
         StyleBox["0.024995000999800072`", {
           GrayLevel[0]}, StripOnInput -> False]],
        Annotation[#, 
         Style[0.024995000999800072`, {
           GrayLevel[0]}], "Tooltip"]& ], 
       TagBox[
        TooltipBox[
         TagBox[
          DynamicBox[{
            FEPrivate`If[
             CurrentValue["MouseOver"], 
             EdgeForm[{
               GrayLevel[0.5], 
               AbsoluteThickness[1.5], 
               Opacity[0.66]}], {}, {}], 
            
            RectangleBox[{3.8, 0.}, {4., 0.01399720055988801}, 
             "RoundingRadius" -> 0]},
           
           ImageSizeCache->{{64.19970845481049, 67.22886297376094}, {
            41.09232931860684, 44.11492836873474}}],
          StatusArea[#, 0.01399720055988801]& ,
          TagBoxNote->"0.01399720055988801"],
         StyleBox["0.01399720055988801`", {
           GrayLevel[0]}, StripOnInput -> False]],
        Annotation[#, 
         Style[0.01399720055988801, {
           GrayLevel[0]}], "Tooltip"]& ], 
       TagBox[
        TooltipBox[
         TagBox[
          DynamicBox[{
            FEPrivate`If[
             CurrentValue["MouseOver"], 
             EdgeForm[{
               GrayLevel[0.5], 
               AbsoluteThickness[1.5], 
               Opacity[0.66]}], {}, {}], 
            
            RectangleBox[{4., 0.}, {4.2, 0.005998800239952004}, 
             "RoundingRadius" -> 0]},
           
           ImageSizeCache->{{66.22886297376094, 69.25801749271137}, {
            42.24810020439421, 44.11492836873474}}],
          StatusArea[#, 0.005998800239952004]& ,
          TagBoxNote->"0.005998800239952004"],
         StyleBox["0.005998800239952004`", {
           GrayLevel[0]}, StripOnInput -> False]],
        Annotation[#, 
         Style[0.005998800239952004, {
           GrayLevel[0]}], "Tooltip"]& ], 
       TagBox[
        TooltipBox[
         TagBox[
          DynamicBox[{
            FEPrivate`If[
             CurrentValue["MouseOver"], 
             EdgeForm[{
               GrayLevel[0.5], 
               AbsoluteThickness[1.5], 
               Opacity[0.66]}], {}, {}], 
            
            RectangleBox[{4.2, 0.}, {4.4, 0.007998400319936006}, 
             "RoundingRadius" -> 0]},
           
           ImageSizeCache->{{68.25801749271137, 71.28717201166181}, {
            41.95915748294737, 44.11492836873474}}],
          StatusArea[#, 0.007998400319936006]& ,
          TagBoxNote->"0.007998400319936006"],
         StyleBox["0.007998400319936006`", {
           GrayLevel[0]}, StripOnInput -> False]],
        Annotation[#, 
         Style[0.007998400319936006, {
           GrayLevel[0]}], "Tooltip"]& ], 
       TagBox[
        TooltipBox[
         TagBox[
          DynamicBox[{
            FEPrivate`If[
             CurrentValue["MouseOver"], 
             EdgeForm[{
               GrayLevel[0.5], 
               AbsoluteThickness[1.5], 
               Opacity[0.66]}], {}, {}], 
            
            RectangleBox[{4.4, 0.}, {4.6, 0.005998800239952031}, 
             "RoundingRadius" -> 0]},
           
           ImageSizeCache->{{70.28717201166181, 73.31632653061224}, {
            42.248100204394206`, 44.11492836873474}}],
          StatusArea[#, 0.005998800239952031]& ,
          TagBoxNote->"0.005998800239952031"],
         StyleBox["0.005998800239952031`", {
           GrayLevel[0]}, StripOnInput -> False]],
        Annotation[#, 
         Style[0.005998800239952031, {
           GrayLevel[0]}], "Tooltip"]& ], 
       TagBox[
        TooltipBox[
         TagBox[
          DynamicBox[{
            FEPrivate`If[
             CurrentValue["MouseOver"], 
             EdgeForm[{
               GrayLevel[0.5], 
               AbsoluteThickness[1.5], 
               Opacity[0.66]}], {}, {}], 
            
            RectangleBox[{4.6, 0.}, {4.8, 0.01299740051989601}, 
             "RoundingRadius" -> 0]},
           
           ImageSizeCache->{{72.31632653061224, 75.34548104956268}, {
            41.236800679330265`, 44.11492836873474}}],
          StatusArea[#, 0.01299740051989601]& ,
          TagBoxNote->"0.01299740051989601"],
         StyleBox["0.01299740051989601`", {
           GrayLevel[0]}, StripOnInput -> False]],
        Annotation[#, 
         Style[0.01299740051989601, {
           GrayLevel[0]}], "Tooltip"]& ], 
       TagBox[
        TooltipBox[
         TagBox[
          DynamicBox[{
            FEPrivate`If[
             CurrentValue["MouseOver"], 
             EdgeForm[{
               GrayLevel[0.5], 
               AbsoluteThickness[1.5], 
               Opacity[0.66]}], {}, {}], 
            
            RectangleBox[{4.8, 0.}, {5., 0.006998600279944005}, 
             "RoundingRadius" -> 0]},
           
           ImageSizeCache->{{74.34548104956268, 77.37463556851313}, {
            42.10362884367079, 44.11492836873474}}],
          StatusArea[#, 0.006998600279944005]& ,
          TagBoxNote->"0.006998600279944005"],
         StyleBox["0.006998600279944005`", {
           GrayLevel[0]}, StripOnInput -> False]],
        Annotation[#, 
         Style[0.006998600279944005, {
           GrayLevel[0]}], "Tooltip"]& ], 
       TagBox[
        TooltipBox[
         TagBox[
          DynamicBox[{
            FEPrivate`If[
             CurrentValue["MouseOver"], 
             EdgeForm[{
               GrayLevel[0.5], 
               AbsoluteThickness[1.5], 
               Opacity[0.66]}], {}, {}], 
            
            RectangleBox[{5., 0.}, {5.2, 0.007998400319936006}, 
             "RoundingRadius" -> 0]},
           
           ImageSizeCache->{{76.37463556851313, 79.40379008746356}, {
            41.95915748294737, 44.11492836873474}}],
          StatusArea[#, 0.007998400319936006]& ,
          TagBoxNote->"0.007998400319936006"],
         StyleBox["0.007998400319936006`", {
           GrayLevel[0]}, StripOnInput -> False]],
        Annotation[#, 
         Style[0.007998400319936006, {
           GrayLevel[0]}], "Tooltip"]& ], 
       TagBox[
        TooltipBox[
         TagBox[
          DynamicBox[{
            FEPrivate`If[
             CurrentValue["MouseOver"], 
             EdgeForm[{
               GrayLevel[0.5], 
               AbsoluteThickness[1.5], 
               Opacity[0.66]}], {}, {}], 
            
            RectangleBox[{5.2, 0.}, {5.4, 0.008998200359928007}, 
             "RoundingRadius" -> 0]},
           
           ImageSizeCache->{{78.40379008746356, 81.432944606414}, {
            41.81468612222395, 44.11492836873474}}],
          StatusArea[#, 0.008998200359928007]& ,
          TagBoxNote->"0.008998200359928007"],
         StyleBox["0.008998200359928007`", {
           GrayLevel[0]}, StripOnInput -> False]],
        Annotation[#, 
         Style[0.008998200359928007, {
           GrayLevel[0]}], "Tooltip"]& ], 
       TagBox[
        TooltipBox[
         TagBox[
          DynamicBox[{
            FEPrivate`If[
             CurrentValue["MouseOver"], 
             EdgeForm[{
               GrayLevel[0.5], 
               AbsoluteThickness[1.5], 
               Opacity[0.66]}], {}, {}], 
            
            RectangleBox[{5.4, 0.}, {5.6, 0.0029994001199760154`}, 
             "RoundingRadius" -> 0]},
           
           ImageSizeCache->{{80.432944606414, 83.46209912536443}, {
            42.68151428656447, 44.11492836873474}}],
          StatusArea[#, 0.0029994001199760154`]& ,
          TagBoxNote->"0.0029994001199760154"],
         StyleBox["0.0029994001199760154`", {
           GrayLevel[0]}, StripOnInput -> False]],
        Annotation[#, 
         Style[0.0029994001199760154`, {
           GrayLevel[0]}], "Tooltip"]& ], 
       TagBox[
        TooltipBox[
         TagBox[
          DynamicBox[{
            FEPrivate`If[
             CurrentValue["MouseOver"], 
             EdgeForm[{
               GrayLevel[0.5], 
               AbsoluteThickness[1.5], 
               Opacity[0.66]}], {}, {}], 
            
            RectangleBox[{5.6, 0.}, {5.8, 0.003999200159968003}, 
             "RoundingRadius" -> 0]},
           
           ImageSizeCache->{{82.46209912536443, 85.49125364431487}, {
            42.53704292584106, 44.11492836873474}}],
          StatusArea[#, 0.003999200159968003]& ,
          TagBoxNote->"0.003999200159968003"],
         StyleBox["0.003999200159968003`", {
           GrayLevel[0]}, StripOnInput -> False]],
        Annotation[#, 
         Style[0.003999200159968003, {
           GrayLevel[0]}], "Tooltip"]& ], 
       TagBox[
        TooltipBox[
         TagBox[
          DynamicBox[{
            FEPrivate`If[
             CurrentValue["MouseOver"], 
             EdgeForm[{
               GrayLevel[0.5], 
               AbsoluteThickness[1.5], 
               Opacity[0.66]}], {}, {}], 
            
            RectangleBox[{5.8, 0.}, {6., 0.005998800239952004}, 
             "RoundingRadius" -> 0]},
           
           ImageSizeCache->{{84.49125364431487, 87.5204081632653}, {
            42.24810020439421, 44.11492836873474}}],
          StatusArea[#, 0.005998800239952004]& ,
          TagBoxNote->"0.005998800239952004"],
         StyleBox["0.005998800239952004`", {
           GrayLevel[0]}, StripOnInput -> False]],
        Annotation[#, 
         Style[0.005998800239952004, {
           GrayLevel[0]}], "Tooltip"]& ], 
       TagBox[
        TooltipBox[
         TagBox[
          DynamicBox[{
            FEPrivate`If[
             CurrentValue["MouseOver"], 
             EdgeForm[{
               GrayLevel[0.5], 
               AbsoluteThickness[1.5], 
               Opacity[0.66]}], {}, {}], 
            
            RectangleBox[{6., 0.}, {6.2, 0.002999400119976002}, 
             "RoundingRadius" -> 0]},
           
           ImageSizeCache->{{86.5204081632653, 89.54956268221575}, {
            42.68151428656448, 44.11492836873474}}],
          StatusArea[#, 0.002999400119976002]& ,
          TagBoxNote->"0.002999400119976002"],
         StyleBox["0.002999400119976002`", {
           GrayLevel[0]}, StripOnInput -> False]],
        Annotation[#, 
         Style[0.002999400119976002, {
           GrayLevel[0]}], "Tooltip"]& ], 
       TagBox[
        TooltipBox[
         TagBox[
          DynamicBox[{
            FEPrivate`If[
             CurrentValue["MouseOver"], 
             EdgeForm[{
               GrayLevel[0.5], 
               AbsoluteThickness[1.5], 
               Opacity[0.66]}], {}, {}], 
            
            RectangleBox[{6.2, 0.}, {6.4, 0.002999400119976002}, 
             "RoundingRadius" -> 0]},
           
           ImageSizeCache->{{88.54956268221575, 91.57871720116619}, {
            42.68151428656448, 44.11492836873474}}],
          StatusArea[#, 0.002999400119976002]& ,
          TagBoxNote->"0.002999400119976002"],
         StyleBox["0.002999400119976002`", {
           GrayLevel[0]}, StripOnInput -> False]],
        Annotation[#, 
         Style[0.002999400119976002, {
           GrayLevel[0]}], "Tooltip"]& ], 
       TagBox[
        TooltipBox[
         TagBox[
          DynamicBox[{
            FEPrivate`If[
             CurrentValue["MouseOver"], 
             EdgeForm[{
               GrayLevel[0.5], 
               AbsoluteThickness[1.5], 
               Opacity[0.66]}], {}, {}], 
            RectangleBox[{6.4, 0.}, {6.6, 0.000999800039992005}, 
             "RoundingRadius" -> 0]},
           
           ImageSizeCache->{{90.57871720116619, 93.60787172011662}, {
            42.970457008011316`, 44.11492836873474}}],
          StatusArea[#, 0.000999800039992005]& ,
          TagBoxNote->"0.000999800039992005"],
         StyleBox["0.000999800039992005`", {
           GrayLevel[0]}, StripOnInput -> False]],
        Annotation[#, 
         Style[0.000999800039992005, {
           GrayLevel[0]}], "Tooltip"]& ], 
       TagBox[
        TooltipBox[
         TagBox[
          DynamicBox[{
            FEPrivate`If[
             CurrentValue["MouseOver"], 
             EdgeForm[{
               GrayLevel[0.5], 
               AbsoluteThickness[1.5], 
               Opacity[0.66]}], {}, {}], 
            
            RectangleBox[{6.6, 0.}, {6.8, 0.005998800239952004}, 
             "RoundingRadius" -> 0]},
           
           ImageSizeCache->{{92.60787172011662, 95.63702623906705}, {
            42.24810020439421, 44.11492836873474}}],
          StatusArea[#, 0.005998800239952004]& ,
          TagBoxNote->"0.005998800239952004"],
         StyleBox["0.005998800239952004`", {
           GrayLevel[0]}, StripOnInput -> False]],
        Annotation[#, 
         Style[0.005998800239952004, {
           GrayLevel[0]}], "Tooltip"]& ], 
       TagBox[
        TooltipBox[
         TagBox[
          DynamicBox[{
            FEPrivate`If[
             CurrentValue["MouseOver"], 
             EdgeForm[{
               GrayLevel[0.5], 
               AbsoluteThickness[1.5], 
               Opacity[0.66]}], {}, {}], 
            
            RectangleBox[{6.8, 0.}, {7., 0.002999400119976002}, 
             "RoundingRadius" -> 0]},
           
           ImageSizeCache->{{94.63702623906705, 97.66618075801749}, {
            42.68151428656448, 44.11492836873474}}],
          StatusArea[#, 0.002999400119976002]& ,
          TagBoxNote->"0.002999400119976002"],
         StyleBox["0.002999400119976002`", {
           GrayLevel[0]}, StripOnInput -> False]],
        Annotation[#, 
         Style[0.002999400119976002, {
           GrayLevel[0]}], "Tooltip"]& ], 
       TagBox[
        TooltipBox[
         TagBox[
          DynamicBox[{
            FEPrivate`If[
             CurrentValue["MouseOver"], 
             EdgeForm[{
               GrayLevel[0.5], 
               AbsoluteThickness[1.5], 
               Opacity[0.66]}], {}, {}], 
            
            RectangleBox[{7., 0.}, {7.2, 0.002999400119976002}, 
             "RoundingRadius" -> 0]},
           
           ImageSizeCache->{{96.66618075801749, 99.69533527696792}, {
            42.68151428656448, 44.11492836873474}}],
          StatusArea[#, 0.002999400119976002]& ,
          TagBoxNote->"0.002999400119976002"],
         StyleBox["0.002999400119976002`", {
           GrayLevel[0]}, StripOnInput -> False]],
        Annotation[#, 
         Style[0.002999400119976002, {
           GrayLevel[0]}], "Tooltip"]& ], 
       TagBox[
        TooltipBox[
         TagBox[
          DynamicBox[{
            FEPrivate`If[
             CurrentValue["MouseOver"], 
             EdgeForm[{
               GrayLevel[0.5], 
               AbsoluteThickness[1.5], 
               Opacity[0.66]}], {}, {}], 
            
            RectangleBox[{7.2, 0.}, {7.4, 0.0019996000799840014`}, 
             "RoundingRadius" -> 0]},
           
           ImageSizeCache->{{98.69533527696792, 101.72448979591837`}, {
            42.825985647287894`, 44.11492836873474}}],
          StatusArea[#, 0.0019996000799840014`]& ,
          TagBoxNote->"0.0019996000799840014"],
         StyleBox["0.0019996000799840014`", {
           GrayLevel[0]}, StripOnInput -> False]],
        Annotation[#, 
         Style[0.0019996000799840014`, {
           GrayLevel[0]}], "Tooltip"]& ], {}, 
       TagBox[
        TooltipBox[
         TagBox[
          DynamicBox[{
            FEPrivate`If[
             CurrentValue["MouseOver"], 
             EdgeForm[{
               GrayLevel[0.5], 
               AbsoluteThickness[1.5], 
               Opacity[0.66]}], {}, {}], 
            
            RectangleBox[{7.6, 0.}, {7.8, 0.002999400119976002}, 
             "RoundingRadius" -> 0]},
           
           ImageSizeCache->{{102.7536443148688, 105.78279883381924`}, {
            42.68151428656448, 44.11492836873474}}],
          StatusArea[#, 0.002999400119976002]& ,
          TagBoxNote->"0.002999400119976002"],
         StyleBox["0.002999400119976002`", {
           GrayLevel[0]}, StripOnInput -> False]],
        Annotation[#, 
         Style[0.002999400119976002, {
           GrayLevel[0]}], "Tooltip"]& ], {}, 
       TagBox[
        TooltipBox[
         TagBox[
          DynamicBox[{
            FEPrivate`If[
             CurrentValue["MouseOver"], 
             EdgeForm[{
               GrayLevel[0.5], 
               AbsoluteThickness[1.5], 
               Opacity[0.66]}], {}, {}], 
            
            RectangleBox[{8., 0.}, {8.2, 0.00399920015996802}, 
             "RoundingRadius" -> 0]},
           
           ImageSizeCache->{{106.81195335276968`, 109.84110787172011`}, {
            42.53704292584105, 44.11492836873474}}],
          StatusArea[#, 0.00399920015996802]& ,
          TagBoxNote->"0.00399920015996802"],
         StyleBox["0.00399920015996802`", {
           GrayLevel[0]}, StripOnInput -> False]],
        Annotation[#, 
         Style[0.00399920015996802, {
           GrayLevel[0]}], "Tooltip"]& ], 
       TagBox[
        TooltipBox[
         TagBox[
          DynamicBox[{
            FEPrivate`If[
             CurrentValue["MouseOver"], 
             EdgeForm[{
               GrayLevel[0.5], 
               AbsoluteThickness[1.5], 
               Opacity[0.66]}], {}, {}], 
            
            RectangleBox[{8.2, 0.}, {8.4, 0.0029994001199759885`}, 
             "RoundingRadius" -> 0]},
           
           ImageSizeCache->{{108.84110787172011`, 111.87026239067056`}, {
            42.68151428656448, 44.11492836873474}}],
          StatusArea[#, 0.0029994001199759885`]& ,
          TagBoxNote->"0.0029994001199759885"],
         StyleBox["0.0029994001199759885`", {
           GrayLevel[0]}, StripOnInput -> False]],
        Annotation[#, 
         Style[0.0029994001199759885`, {
           GrayLevel[0]}], "Tooltip"]& ], 
       TagBox[
        TooltipBox[
         TagBox[
          DynamicBox[{
            FEPrivate`If[
             CurrentValue["MouseOver"], 
             EdgeForm[{
               GrayLevel[0.5], 
               AbsoluteThickness[1.5], 
               Opacity[0.66]}], {}, {}], 
            
            RectangleBox[{8.4, 0.}, {8.6, 0.004999000199960026}, 
             "RoundingRadius" -> 0]},
           
           ImageSizeCache->{{110.87026239067056`, 113.89941690962098`}, {
            42.39257156511763, 44.11492836873474}}],
          StatusArea[#, 0.004999000199960026]& ,
          TagBoxNote->"0.004999000199960026"],
         StyleBox["0.004999000199960026`", {
           GrayLevel[0]}, StripOnInput -> False]],
        Annotation[#, 
         Style[0.004999000199960026, {
           GrayLevel[0]}], "Tooltip"]& ], {}, {}, {}, 
       TagBox[
        TooltipBox[
         TagBox[
          DynamicBox[{
            FEPrivate`If[
             CurrentValue["MouseOver"], 
             EdgeForm[{
               GrayLevel[0.5], 
               AbsoluteThickness[1.5], 
               Opacity[0.66]}], {}, {}], 
            
            RectangleBox[{9.2, 0.}, {9.4, 0.0009998000399919962}, 
             "RoundingRadius" -> 0]},
           
           ImageSizeCache->{{118.9868804664723, 122.01603498542273`}, {
            42.970457008011316`, 44.11492836873474}}],
          StatusArea[#, 0.0009998000399919962]& ,
          TagBoxNote->"0.0009998000399919962"],
         StyleBox["0.0009998000399919962`", {
           GrayLevel[0]}, StripOnInput -> False]],
        Annotation[#, 
         Style[0.0009998000399919962, {
           GrayLevel[0]}], "Tooltip"]& ], 
       TagBox[
        TooltipBox[
         TagBox[
          DynamicBox[{
            FEPrivate`If[
             CurrentValue["MouseOver"], 
             EdgeForm[{
               GrayLevel[0.5], 
               AbsoluteThickness[1.5], 
               Opacity[0.66]}], {}, {}], 
            
            RectangleBox[{9.4, 0.}, {9.6, 0.004999000199960026}, 
             "RoundingRadius" -> 0]},
           
           ImageSizeCache->{{121.01603498542273`, 124.04518950437317`}, {
            42.39257156511763, 44.11492836873474}}],
          StatusArea[#, 0.004999000199960026]& ,
          TagBoxNote->"0.004999000199960026"],
         StyleBox["0.004999000199960026`", {
           GrayLevel[0]}, StripOnInput -> False]],
        Annotation[#, 
         Style[0.004999000199960026, {
           GrayLevel[0]}], "Tooltip"]& ], {}, 
       TagBox[
        TooltipBox[
         TagBox[
          DynamicBox[{
            FEPrivate`If[
             CurrentValue["MouseOver"], 
             EdgeForm[{
               GrayLevel[0.5], 
               AbsoluteThickness[1.5], 
               Opacity[0.66]}], {}, {}], 
            
            RectangleBox[{9.8, 0.}, {10., 0.004999000199960026}, 
             "RoundingRadius" -> 0]},
           
           ImageSizeCache->{{125.07434402332362`, 128.10349854227405`}, {
            42.39257156511763, 44.11492836873474}}],
          StatusArea[#, 0.004999000199960026]& ,
          TagBoxNote->"0.004999000199960026"],
         StyleBox["0.004999000199960026`", {
           GrayLevel[0]}, StripOnInput -> False]],
        Annotation[#, 
         Style[0.004999000199960026, {
           GrayLevel[0]}], "Tooltip"]& ], 
       TagBox[
        TooltipBox[
         TagBox[
          DynamicBox[{
            FEPrivate`If[
             CurrentValue["MouseOver"], 
             EdgeForm[{
               GrayLevel[0.5], 
               AbsoluteThickness[1.5], 
               Opacity[0.66]}], {}, {}], 
            
            RectangleBox[{10., 0.}, {10.2, 0.000999800039992005}, 
             "RoundingRadius" -> 0]},
           
           ImageSizeCache->{{127.10349854227405`, 130.1326530612245}, {
            42.970457008011316`, 44.11492836873474}}],
          StatusArea[#, 0.000999800039992005]& ,
          TagBoxNote->"0.000999800039992005"],
         StyleBox["0.000999800039992005`", {
           GrayLevel[0]}, StripOnInput -> False]],
        Annotation[#, 
         Style[0.000999800039992005, {
           GrayLevel[0]}], "Tooltip"]& ], 
       TagBox[
        TooltipBox[
         TagBox[
          DynamicBox[{
            FEPrivate`If[
             CurrentValue["MouseOver"], 
             EdgeForm[{
               GrayLevel[0.5], 
               AbsoluteThickness[1.5], 
               Opacity[0.66]}], {}, {}], 
            
            RectangleBox[{10.2, 0.}, {10.4, 0.0009998000399919962}, 
             "RoundingRadius" -> 0]},
           
           ImageSizeCache->{{129.1326530612245, 132.16180758017492`}, {
            42.970457008011316`, 44.11492836873474}}],
          StatusArea[#, 0.0009998000399919962]& ,
          TagBoxNote->"0.0009998000399919962"],
         StyleBox["0.0009998000399919962`", {
           GrayLevel[0]}, StripOnInput -> False]],
        Annotation[#, 
         Style[0.0009998000399919962, {
           GrayLevel[0]}], "Tooltip"]& ], 
       TagBox[
        TooltipBox[
         TagBox[
          DynamicBox[{
            FEPrivate`If[
             CurrentValue["MouseOver"], 
             EdgeForm[{
               GrayLevel[0.5], 
               AbsoluteThickness[1.5], 
               Opacity[0.66]}], {}, {}], 
            
            RectangleBox[{10.4, 0.}, {10.6, 0.00199960007998401}, 
             "RoundingRadius" -> 0]},
           
           ImageSizeCache->{{131.16180758017492`, 134.19096209912536`}, {
            42.825985647287894`, 44.11492836873474}}],
          StatusArea[#, 0.00199960007998401]& ,
          TagBoxNote->"0.00199960007998401"],
         StyleBox["0.00199960007998401`", {
           GrayLevel[0]}, StripOnInput -> False]],
        Annotation[#, 
         Style[0.00199960007998401, {
           GrayLevel[0]}], "Tooltip"]& ], 
       TagBox[
        TooltipBox[
         TagBox[
          DynamicBox[{
            FEPrivate`If[
             CurrentValue["MouseOver"], 
             EdgeForm[{
               GrayLevel[0.5], 
               AbsoluteThickness[1.5], 
               Opacity[0.66]}], {}, {}], 
            
            RectangleBox[{10.6, 0.}, {10.8, 0.0019996000799839923`}, 
             "RoundingRadius" -> 0]},
           
           ImageSizeCache->{{133.19096209912536`, 136.22011661807582`}, {
            42.8259856472879, 44.11492836873474}}],
          StatusArea[#, 0.0019996000799839923`]& ,
          TagBoxNote->"0.0019996000799839923"],
         StyleBox["0.0019996000799839923`", {
           GrayLevel[0]}, StripOnInput -> False]],
        Annotation[#, 
         Style[0.0019996000799839923`, {
           GrayLevel[0]}], "Tooltip"]& ], 
       TagBox[
        TooltipBox[
         TagBox[
          DynamicBox[{
            FEPrivate`If[
             CurrentValue["MouseOver"], 
             EdgeForm[{
               GrayLevel[0.5], 
               AbsoluteThickness[1.5], 
               Opacity[0.66]}], {}, {}], 
            
            RectangleBox[{10.8, 0.}, {11., 0.00199960007998401}, 
             "RoundingRadius" -> 0]},
           
           ImageSizeCache->{{135.22011661807582`, 138.24927113702626`}, {
            42.825985647287894`, 44.11492836873474}}],
          StatusArea[#, 0.00199960007998401]& ,
          TagBoxNote->"0.00199960007998401"],
         StyleBox["0.00199960007998401`", {
           GrayLevel[0]}, StripOnInput -> False]],
        Annotation[#, 
         Style[0.00199960007998401, {
           GrayLevel[0]}], "Tooltip"]& ], 
       TagBox[
        TooltipBox[
         TagBox[
          DynamicBox[{
            FEPrivate`If[
             CurrentValue["MouseOver"], 
             EdgeForm[{
               GrayLevel[0.5], 
               AbsoluteThickness[1.5], 
               Opacity[0.66]}], {}, {}], 
            
            RectangleBox[{11., 0.}, {11.2, 0.000999800039992005}, 
             "RoundingRadius" -> 0]},
           
           ImageSizeCache->{{137.24927113702626`, 140.27842565597666`}, {
            42.970457008011316`, 44.11492836873474}}],
          StatusArea[#, 0.000999800039992005]& ,
          TagBoxNote->"0.000999800039992005"],
         StyleBox["0.000999800039992005`", {
           GrayLevel[0]}, StripOnInput -> False]],
        Annotation[#, 
         Style[0.000999800039992005, {
           GrayLevel[0]}], "Tooltip"]& ], {}, 
       TagBox[
        TooltipBox[
         TagBox[
          DynamicBox[{
            FEPrivate`If[
             CurrentValue["MouseOver"], 
             EdgeForm[{
               GrayLevel[0.5], 
               AbsoluteThickness[1.5], 
               Opacity[0.66]}], {}, {}], 
            
            RectangleBox[{11.4, 0.}, {11.6, 0.000999800039992005}, 
             "RoundingRadius" -> 0]},
           
           ImageSizeCache->{{141.30758017492712`, 144.33673469387756`}, {
            42.970457008011316`, 44.11492836873474}}],
          StatusArea[#, 0.000999800039992005]& ,
          TagBoxNote->"0.000999800039992005"],
         StyleBox["0.000999800039992005`", {
           GrayLevel[0]}, StripOnInput -> False]],
        Annotation[#, 
         Style[0.000999800039992005, {
           GrayLevel[0]}], "Tooltip"]& ], 
       TagBox[
        TooltipBox[
         TagBox[
          DynamicBox[{
            FEPrivate`If[
             CurrentValue["MouseOver"], 
             EdgeForm[{
               GrayLevel[0.5], 
               AbsoluteThickness[1.5], 
               Opacity[0.66]}], {}, {}], 
            
            RectangleBox[{11.6, 0.}, {11.8, 0.0009998000399919962}, 
             "RoundingRadius" -> 0]},
           
           ImageSizeCache->{{143.33673469387756`, 146.365889212828}, {
            42.970457008011316`, 44.11492836873474}}],
          StatusArea[#, 0.0009998000399919962]& ,
          TagBoxNote->"0.0009998000399919962"],
         StyleBox["0.0009998000399919962`", {
           GrayLevel[0]}, StripOnInput -> False]],
        Annotation[#, 
         Style[0.0009998000399919962, {
           GrayLevel[0]}], "Tooltip"]& ], {}, {}, {}, {}, {}, 
       TagBox[
        TooltipBox[
         TagBox[
          DynamicBox[{
            FEPrivate`If[
             CurrentValue["MouseOver"], 
             EdgeForm[{
               GrayLevel[0.5], 
               AbsoluteThickness[1.5], 
               Opacity[0.66]}], {}, {}], 
            
            RectangleBox[{12.8, 0.}, {13., 0.000999800039992005}, 
             "RoundingRadius" -> 0]},
           
           ImageSizeCache->{{155.5116618075802, 158.54081632653063`}, {
            42.970457008011316`, 44.11492836873474}}],
          StatusArea[#, 0.000999800039992005]& ,
          TagBoxNote->"0.000999800039992005"],
         StyleBox["0.000999800039992005`", {
           GrayLevel[0]}, StripOnInput -> False]],
        Annotation[#, 
         Style[0.000999800039992005, {
           GrayLevel[0]}], "Tooltip"]& ]}, {}, {}}, {{}, {}, 
      {RGBColor[1, 0, 0], Thickness[Large], LineBox[CompressedData["
1:eJwVlnc81u8XxvUge++ZvUdJQ+HcVFaRUURkZyRJshIZFSURmSEr+2uEZH1s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        "]]}}},
    AspectRatio->NCache[GoldenRatio^(-1), 0.6180339887498948],
    Axes->{True, True},
    AxesLabel->{
      FormBox["\"x1\"", TraditionalForm], 
      FormBox["\"freq\"", TraditionalForm]},
    AxesOrigin->{-1., 0},
    FrameTicks->{{Automatic, Automatic}, {Automatic, Automatic}},
    GridLines->{None, None},
    PlotRange->{{-1, 13}, {All, All}},
    PlotRangePadding->{{
       Scaled[0.02], 
       Scaled[0.02]}, {
       Scaled[0.02], 
       Scaled[0.1]}},
    Ticks->{Automatic, Automatic}], ",", 
   GraphicsBox[{
     {RGBColor[0.798413061722744, 0.824719615472648, 0.968322270542458], 
      EdgeForm[Opacity[0.7]], {}, 
      {RGBColor[0.798413061722744, 0.824719615472648, 0.968322270542458], 
       EdgeForm[Opacity[0.7]], 
       TagBox[
        TooltipBox[
         TagBox[
          DynamicBox[{
            FEPrivate`If[
             CurrentValue["MouseOver"], 
             EdgeForm[{
               GrayLevel[0.5], 
               AbsoluteThickness[1.5], 
               Opacity[0.66]}], {}, {}], 
            
            RectangleBox[{-1., 0.}, {-0.8, 0.18996200759848034`}, 
             "RoundingRadius" -> 0]},
           
           ImageSizeCache->{{15.5, 18.588835534213686`}, {12.597155703346225`,
             44.114928368734745`}}],
          StatusArea[#, 0.18996200759848034`]& ,
          TagBoxNote->"0.18996200759848034"],
         StyleBox["0.18996200759848034`", {
           GrayLevel[0]}, StripOnInput -> False]],
        Annotation[#, 
         Style[0.18996200759848034`, {
           GrayLevel[0]}], "Tooltip"]& ], 
       TagBox[
        TooltipBox[
         TagBox[
          DynamicBox[{
            FEPrivate`If[
             CurrentValue["MouseOver"], 
             EdgeForm[{
               GrayLevel[0.5], 
               AbsoluteThickness[1.5], 
               Opacity[0.66]}], {}, {}], 
            
            RectangleBox[{-0.8, 0.}, {-0.6, 0.2619476104779043}, 
             "RoundingRadius" -> 0]},
           
           ImageSizeCache->{{17.588835534213686`, 20.677671068427372`}, {
            1.032526061725335, 44.114928368734745`}}],
          StatusArea[#, 0.2619476104779043]& ,
          TagBoxNote->"0.2619476104779043"],
         StyleBox["0.2619476104779043`", {
           GrayLevel[0]}, StripOnInput -> False]],
        Annotation[#, 
         Style[0.2619476104779043, {
           GrayLevel[0]}], "Tooltip"]& ], 
       TagBox[
        TooltipBox[
         TagBox[
          DynamicBox[{
            FEPrivate`If[
             CurrentValue["MouseOver"], 
             EdgeForm[{
               GrayLevel[0.5], 
               AbsoluteThickness[1.5], 
               Opacity[0.66]}], {}, {}], 
            
            RectangleBox[{-0.6, 0.}, {-0.4, 0.4349130173965208}, 
             "RoundingRadius" -> 0]},
           
           ImageSizeCache->{{19.677671068427372`, 
            22.766506602641055`}, {-26.7547090493916, 44.114928368734745`}}],
          StatusArea[#, 0.4349130173965208]& ,
          TagBoxNote->"0.4349130173965208"],
         StyleBox["0.4349130173965208`", {
           GrayLevel[0]}, StripOnInput -> False]],
        Annotation[#, 
         Style[0.4349130173965208, {
           GrayLevel[0]}], "Tooltip"]& ], 
       TagBox[
        TooltipBox[
         TagBox[
          DynamicBox[{
            FEPrivate`If[
             CurrentValue["MouseOver"], 
             EdgeForm[{
               GrayLevel[0.5], 
               AbsoluteThickness[1.5], 
               Opacity[0.66]}], {}, {}], 
            
            RectangleBox[{-0.4, 0.}, {-0.2, 0.3709258148370326}, 
             "RoundingRadius" -> 0]},
           
           ImageSizeCache->{{21.766506602641055`, 
            24.85534213685474}, {-16.475038256839667`, 44.114928368734745`}}],
          StatusArea[#, 0.3709258148370326]& ,
          TagBoxNote->"0.3709258148370326"],
         StyleBox["0.3709258148370326`", {
           GrayLevel[0]}, StripOnInput -> False]],
        Annotation[#, 
         Style[0.3709258148370326, {
           GrayLevel[0]}], "Tooltip"]& ], 
       TagBox[
        TooltipBox[
         TagBox[
          DynamicBox[{
            FEPrivate`If[
             CurrentValue["MouseOver"], 
             EdgeForm[{
               GrayLevel[0.5], 
               AbsoluteThickness[1.5], 
               Opacity[0.66]}], {}, {}], 
            
            RectangleBox[{-0.2, 0.}, {0., 0.5008998200359928}, 
             "RoundingRadius" -> 0]},
           
           ImageSizeCache->{{23.85534213685474, 
            26.944177671068427`}, {-37.355619554210755`, 
            44.114928368734745`}}],
          StatusArea[#, 0.5008998200359928]& ,
          TagBoxNote->"0.5008998200359928"],
         StyleBox["0.5008998200359928`", {
           GrayLevel[0]}, StripOnInput -> False]],
        Annotation[#, 
         Style[0.5008998200359928, {
           GrayLevel[0]}], "Tooltip"]& ], 
       TagBox[
        TooltipBox[
         TagBox[
          DynamicBox[{
            FEPrivate`If[
             CurrentValue["MouseOver"], 
             EdgeForm[{
               GrayLevel[0.5], 
               AbsoluteThickness[1.5], 
               Opacity[0.66]}], {}, {}], 
            
            RectangleBox[{0., 0.}, {0.2, 0.45490901819636076`}, 
             "RoundingRadius" -> 0]},
           
           ImageSizeCache->{{25.944177671068427`, 
            29.033013205282113`}, {-29.96710617206407, 44.114928368734745`}}],
          StatusArea[#, 0.45490901819636076`]& ,
          TagBoxNote->"0.45490901819636076"],
         StyleBox["0.45490901819636076`", {
           GrayLevel[0]}, StripOnInput -> False]],
        Annotation[#, 
         Style[0.45490901819636076`, {
           GrayLevel[0]}], "Tooltip"]& ], 
       TagBox[
        TooltipBox[
         TagBox[
          DynamicBox[{
            FEPrivate`If[
             CurrentValue["MouseOver"], 
             EdgeForm[{
               GrayLevel[0.5], 
               AbsoluteThickness[1.5], 
               Opacity[0.66]}], {}, {}], 
            RectangleBox[{0.2, 0.}, {0.4, 0.4669066186762647}, 
             "RoundingRadius" -> 0]},
           
           ImageSizeCache->{{28.033013205282113`, 
            31.1218487394958}, {-31.89454444566754, 44.114928368734745`}}],
          StatusArea[#, 0.4669066186762647]& ,
          TagBoxNote->"0.4669066186762647"],
         StyleBox["0.4669066186762647`", {
           GrayLevel[0]}, StripOnInput -> False]],
        Annotation[#, 
         Style[0.4669066186762647, {
           GrayLevel[0]}], "Tooltip"]& ], 
       TagBox[
        TooltipBox[
         TagBox[
          DynamicBox[{
            FEPrivate`If[
             CurrentValue["MouseOver"], 
             EdgeForm[{
               GrayLevel[0.5], 
               AbsoluteThickness[1.5], 
               Opacity[0.66]}], {}, {}], 
            
            RectangleBox[{0.4, 0.}, {0.6, 0.3569286142771447}, 
             "RoundingRadius" -> 0]},
           
           ImageSizeCache->{{30.1218487394958, 
            33.21068427370948}, {-14.226360270968954`, 44.114928368734745`}}],
          StatusArea[#, 0.3569286142771447]& ,
          TagBoxNote->"0.3569286142771447"],
         StyleBox["0.3569286142771447`", {
           GrayLevel[0]}, StripOnInput -> False]],
        Annotation[#, 
         Style[0.3569286142771447, {
           GrayLevel[0]}], "Tooltip"]& ], 
       TagBox[
        TooltipBox[
         TagBox[
          DynamicBox[{
            FEPrivate`If[
             CurrentValue["MouseOver"], 
             EdgeForm[{
               GrayLevel[0.5], 
               AbsoluteThickness[1.5], 
               Opacity[0.66]}], {}, {}], 
            
            RectangleBox[{0.6, 0.}, {0.8, 0.3269346130773844}, 
             "RoundingRadius" -> 0]},
           
           ImageSizeCache->{{32.21068427370948, 
            35.29951980792317}, {-9.407764586960205, 44.114928368734745`}}],
          StatusArea[#, 0.3269346130773844]& ,
          TagBoxNote->"0.3269346130773844"],
         StyleBox["0.3269346130773844`", {
           GrayLevel[0]}, StripOnInput -> False]],
        Annotation[#, 
         Style[0.3269346130773844, {
           GrayLevel[0]}], "Tooltip"]& ], 
       TagBox[
        TooltipBox[
         TagBox[
          DynamicBox[{
            FEPrivate`If[
             CurrentValue["MouseOver"], 
             EdgeForm[{
               GrayLevel[0.5], 
               AbsoluteThickness[1.5], 
               Opacity[0.66]}], {}, {}], 
            
            RectangleBox[{0.8, 0.}, {1., 0.2879424115176965}, 
             "RoundingRadius" -> 0]},
           
           ImageSizeCache->{{34.29951980792317, 
            37.388355342136855`}, {-3.1435901977489067`, 
            44.114928368734745`}}],
          StatusArea[#, 0.2879424115176965]& ,
          TagBoxNote->"0.2879424115176965"],
         StyleBox["0.2879424115176965`", {
           GrayLevel[0]}, StripOnInput -> False]],
        Annotation[#, 
         Style[0.2879424115176965, {
           GrayLevel[0]}], "Tooltip"]& ], 
       TagBox[
        TooltipBox[
         TagBox[
          DynamicBox[{
            FEPrivate`If[
             CurrentValue["MouseOver"], 
             EdgeForm[{
               GrayLevel[0.5], 
               AbsoluteThickness[1.5], 
               Opacity[0.66]}], {}, {}], 
            
            RectangleBox[{1., 0.}, {1.2, 0.1719656068786243}, 
             "RoundingRadius" -> 0]},
           
           ImageSizeCache->{{36.388355342136855`, 39.47719087635054}, {
            15.488313113751452`, 44.114928368734745`}}],
          StatusArea[#, 0.1719656068786243]& ,
          TagBoxNote->"0.1719656068786243"],
         StyleBox["0.1719656068786243`", {
           GrayLevel[0]}, StripOnInput -> False]],
        Annotation[#, 
         Style[0.1719656068786243, {
           GrayLevel[0]}], "Tooltip"]& ], 
       TagBox[
        TooltipBox[
         TagBox[
          DynamicBox[{
            FEPrivate`If[
             CurrentValue["MouseOver"], 
             EdgeForm[{
               GrayLevel[0.5], 
               AbsoluteThickness[1.5], 
               Opacity[0.66]}], {}, {}], 
            
            RectangleBox[{1.2, 0.}, {1.4, 0.24395120975804843`}, 
             "RoundingRadius" -> 0]},
           
           ImageSizeCache->{{38.47719087635054, 41.56602641056423}, {
            3.9236834721305414`, 44.114928368734745`}}],
          StatusArea[#, 0.24395120975804843`]& ,
          TagBoxNote->"0.24395120975804843"],
         StyleBox["0.24395120975804843`", {
           GrayLevel[0]}, StripOnInput -> False]],
        Annotation[#, 
         Style[0.24395120975804843`, {
           GrayLevel[0]}], "Tooltip"]& ], 
       TagBox[
        TooltipBox[
         TagBox[
          DynamicBox[{
            FEPrivate`If[
             CurrentValue["MouseOver"], 
             EdgeForm[{
               GrayLevel[0.5], 
               AbsoluteThickness[1.5], 
               Opacity[0.66]}], {}, {}], 
            
            RectangleBox[{1.4, 0.}, {1.6, 0.1719656068786241}, 
             "RoundingRadius" -> 0]},
           
           ImageSizeCache->{{40.56602641056423, 43.65486194477791}, {
            15.488313113751484`, 44.114928368734745`}}],
          StatusArea[#, 0.1719656068786241]& ,
          TagBoxNote->"0.1719656068786241"],
         StyleBox["0.1719656068786241`", {
           GrayLevel[0]}, StripOnInput -> False]],
        Annotation[#, 
         Style[0.1719656068786241, {
           GrayLevel[0]}], "Tooltip"]& ], 
       TagBox[
        TooltipBox[
         TagBox[
          DynamicBox[{
            FEPrivate`If[
             CurrentValue["MouseOver"], 
             EdgeForm[{
               GrayLevel[0.5], 
               AbsoluteThickness[1.5], 
               Opacity[0.66]}], {}, {}], 
            
            RectangleBox[{1.6, 0.}, {1.8, 0.1119776044791042}, 
             "RoundingRadius" -> 0]},
           
           ImageSizeCache->{{42.65486194477791, 45.74369747899159}, {
            25.125504481768882`, 44.114928368734745`}}],
          StatusArea[#, 0.1119776044791042]& ,
          TagBoxNote->"0.1119776044791042"],
         StyleBox["0.1119776044791042`", {
           GrayLevel[0]}, StripOnInput -> False]],
        Annotation[#, 
         Style[0.1119776044791042, {
           GrayLevel[0]}], "Tooltip"]& ], 
       TagBox[
        TooltipBox[
         TagBox[
          DynamicBox[{
            FEPrivate`If[
             CurrentValue["MouseOver"], 
             EdgeForm[{
               GrayLevel[0.5], 
               AbsoluteThickness[1.5], 
               Opacity[0.66]}], {}, {}], 
            
            RectangleBox[{1.8, 0.}, {2., 0.11797640471905621`}, 
             "RoundingRadius" -> 0]},
           
           ImageSizeCache->{{44.74369747899159, 47.83253301320528}, {
            24.16178534496714, 44.114928368734745`}}],
          StatusArea[#, 0.11797640471905621`]& ,
          TagBoxNote->"0.11797640471905621"],
         StyleBox["0.11797640471905621`", {
           GrayLevel[0]}, StripOnInput -> False]],
        Annotation[#, 
         Style[0.11797640471905621`, {
           GrayLevel[0]}], "Tooltip"]& ], 
       TagBox[
        TooltipBox[
         TagBox[
          DynamicBox[{
            FEPrivate`If[
             CurrentValue["MouseOver"], 
             EdgeForm[{
               GrayLevel[0.5], 
               AbsoluteThickness[1.5], 
               Opacity[0.66]}], {}, {}], 
            
            RectangleBox[{2., 0.}, {2.2, 0.09398120375924807}, 
             "RoundingRadius" -> 0]},
           
           ImageSizeCache->{{46.83253301320528, 49.921368547418965`}, {
            28.016661892174128`, 44.114928368734745`}}],
          StatusArea[#, 0.09398120375924807]& ,
          TagBoxNote->"0.09398120375924807"],
         StyleBox["0.09398120375924807`", {
           GrayLevel[0]}, StripOnInput -> False]],
        Annotation[#, 
         Style[0.09398120375924807, {
           GrayLevel[0]}], "Tooltip"]& ], 
       TagBox[
        TooltipBox[
         TagBox[
          DynamicBox[{
            FEPrivate`If[
             CurrentValue["MouseOver"], 
             EdgeForm[{
               GrayLevel[0.5], 
               AbsoluteThickness[1.5], 
               Opacity[0.66]}], {}, {}], 
            
            RectangleBox[{2.2, 0.}, {2.4, 0.07098580283943222}, 
             "RoundingRadius" -> 0]},
           
           ImageSizeCache->{{48.921368547418965`, 52.01020408163265}, {
            31.71091858324744, 44.114928368734745`}}],
          StatusArea[#, 0.07098580283943222]& ,
          TagBoxNote->"0.07098580283943222"],
         StyleBox["0.07098580283943222`", {
           GrayLevel[0]}, StripOnInput -> False]],
        Annotation[#, 
         Style[0.07098580283943222, {
           GrayLevel[0]}], "Tooltip"]& ], 
       TagBox[
        TooltipBox[
         TagBox[
          DynamicBox[{
            FEPrivate`If[
             CurrentValue["MouseOver"], 
             EdgeForm[{
               GrayLevel[0.5], 
               AbsoluteThickness[1.5], 
               Opacity[0.66]}], {}, {}], 
            
            RectangleBox[{2.4, 0.}, {2.6, 0.04499100179964003}, 
             "RoundingRadius" -> 0]},
           
           ImageSizeCache->{{51.01020408163265, 54.09903961584634}, {
            35.88703484272168, 44.114928368734745`}}],
          StatusArea[#, 0.04499100179964003]& ,
          TagBoxNote->"0.04499100179964003"],
         StyleBox["0.04499100179964003`", {
           GrayLevel[0]}, StripOnInput -> False]],
        Annotation[#, 
         Style[0.04499100179964003, {
           GrayLevel[0]}], "Tooltip"]& ], 
       TagBox[
        TooltipBox[
         TagBox[
          DynamicBox[{
            FEPrivate`If[
             CurrentValue["MouseOver"], 
             EdgeForm[{
               GrayLevel[0.5], 
               AbsoluteThickness[1.5], 
               Opacity[0.66]}], {}, {}], 
            
            RectangleBox[{2.6, 0.}, {2.8, 0.04099180163967212}, 
             "RoundingRadius" -> 0]},
           
           ImageSizeCache->{{53.09903961584634, 56.18787515006002}, {
            36.52951426725616, 44.114928368734745`}}],
          StatusArea[#, 0.04099180163967212]& ,
          TagBoxNote->"0.04099180163967212"],
         StyleBox["0.04099180163967212`", {
           GrayLevel[0]}, StripOnInput -> False]],
        Annotation[#, 
         Style[0.04099180163967212, {
           GrayLevel[0]}], "Tooltip"]& ], 
       TagBox[
        TooltipBox[
         TagBox[
          DynamicBox[{
            FEPrivate`If[
             CurrentValue["MouseOver"], 
             EdgeForm[{
               GrayLevel[0.5], 
               AbsoluteThickness[1.5], 
               Opacity[0.66]}], {}, {}], 
            
            RectangleBox[{2.8, 0.}, {3., 0.03099380123975202}, 
             "RoundingRadius" -> 0]},
           
           ImageSizeCache->{{55.18787515006002, 58.27671068427371}, {
            38.13571282859241, 44.114928368734745`}}],
          StatusArea[#, 0.03099380123975202]& ,
          TagBoxNote->"0.03099380123975202"],
         StyleBox["0.03099380123975202`", {
           GrayLevel[0]}, StripOnInput -> False]],
        Annotation[#, 
         Style[0.03099380123975202, {
           GrayLevel[0]}], "Tooltip"]& ], 
       TagBox[
        TooltipBox[
         TagBox[
          DynamicBox[{
            FEPrivate`If[
             CurrentValue["MouseOver"], 
             EdgeForm[{
               GrayLevel[0.5], 
               AbsoluteThickness[1.5], 
               Opacity[0.66]}], {}, {}], 
            
            RectangleBox[{3., 0.}, {3.2, 0.03599280143971203}, 
             "RoundingRadius" -> 0]},
           
           ImageSizeCache->{{57.27671068427371, 60.365546218487395`}, {
            37.332613547924296`, 44.114928368734745`}}],
          StatusArea[#, 0.03599280143971203]& ,
          TagBoxNote->"0.03599280143971203"],
         StyleBox["0.03599280143971203`", {
           GrayLevel[0]}, StripOnInput -> False]],
        Annotation[#, 
         Style[0.03599280143971203, {
           GrayLevel[0]}], "Tooltip"]& ], 
       TagBox[
        TooltipBox[
         TagBox[
          DynamicBox[{
            FEPrivate`If[
             CurrentValue["MouseOver"], 
             EdgeForm[{
               GrayLevel[0.5], 
               AbsoluteThickness[1.5], 
               Opacity[0.66]}], {}, {}], 
            
            RectangleBox[{3.2, 0.}, {3.4, 0.0339932013597281}, 
             "RoundingRadius" -> 0]},
           
           ImageSizeCache->{{59.365546218487395`, 62.45438175270108}, {
            37.65385326019153, 44.114928368734745`}}],
          StatusArea[#, 0.0339932013597281]& ,
          TagBoxNote->"0.0339932013597281"],
         StyleBox["0.0339932013597281`", {
           GrayLevel[0]}, StripOnInput -> False]],
        Annotation[#, 
         Style[0.0339932013597281, {
           GrayLevel[0]}], "Tooltip"]& ], 
       TagBox[
        TooltipBox[
         TagBox[
          DynamicBox[{
            FEPrivate`If[
             CurrentValue["MouseOver"], 
             EdgeForm[{
               GrayLevel[0.5], 
               AbsoluteThickness[1.5], 
               Opacity[0.66]}], {}, {}], 
            
            RectangleBox[{3.4, 0.}, {3.6, 0.01399720055988801}, 
             "RoundingRadius" -> 0]},
           
           ImageSizeCache->{{61.45438175270108, 64.54321728691477}, {
            40.86625038286402, 44.114928368734745`}}],
          StatusArea[#, 0.01399720055988801]& ,
          TagBoxNote->"0.01399720055988801"],
         StyleBox["0.01399720055988801`", {
           GrayLevel[0]}, StripOnInput -> False]],
        Annotation[#, 
         Style[0.01399720055988801, {
           GrayLevel[0]}], "Tooltip"]& ], 
       TagBox[
        TooltipBox[
         TagBox[
          DynamicBox[{
            FEPrivate`If[
             CurrentValue["MouseOver"], 
             EdgeForm[{
               GrayLevel[0.5], 
               AbsoluteThickness[1.5], 
               Opacity[0.66]}], {}, {}], 
            
            RectangleBox[{3.6, 0.}, {3.8, 0.01999600079984006}, 
             "RoundingRadius" -> 0]},
           
           ImageSizeCache->{{63.54321728691477, 66.63205282112845}, {
            39.90253124606227, 44.114928368734745`}}],
          StatusArea[#, 0.01999600079984006]& ,
          TagBoxNote->"0.01999600079984006"],
         StyleBox["0.01999600079984006`", {
           GrayLevel[0]}, StripOnInput -> False]],
        Annotation[#, 
         Style[0.01999600079984006, {
           GrayLevel[0]}], "Tooltip"]& ], 
       TagBox[
        TooltipBox[
         TagBox[
          DynamicBox[{
            FEPrivate`If[
             CurrentValue["MouseOver"], 
             EdgeForm[{
               GrayLevel[0.5], 
               AbsoluteThickness[1.5], 
               Opacity[0.66]}], {}, {}], 
            
            RectangleBox[{3.8, 0.}, {4., 0.01299740051989601}, 
             "RoundingRadius" -> 0]},
           
           ImageSizeCache->{{65.63205282112845, 68.72088835534214}, {
            41.02687023899764, 44.114928368734745`}}],
          StatusArea[#, 0.01299740051989601]& ,
          TagBoxNote->"0.01299740051989601"],
         StyleBox["0.01299740051989601`", {
           GrayLevel[0]}, StripOnInput -> False]],
        Annotation[#, 
         Style[0.01299740051989601, {
           GrayLevel[0]}], "Tooltip"]& ], 
       TagBox[
        TooltipBox[
         TagBox[
          DynamicBox[{
            FEPrivate`If[
             CurrentValue["MouseOver"], 
             EdgeForm[{
               GrayLevel[0.5], 
               AbsoluteThickness[1.5], 
               Opacity[0.66]}], {}, {}], 
            
            RectangleBox[{4., 0.}, {4.2, 0.009998000399920007}, 
             "RoundingRadius" -> 0]},
           
           ImageSizeCache->{{67.72088835534214, 70.80972388955583}, {
            41.50872980739851, 44.114928368734745`}}],
          StatusArea[#, 0.009998000399920007]& ,
          TagBoxNote->"0.009998000399920007"],
         StyleBox["0.009998000399920007`", {
           GrayLevel[0]}, StripOnInput -> False]],
        Annotation[#, 
         Style[0.009998000399920007, {
           GrayLevel[0]}], "Tooltip"]& ], 
       TagBox[
        TooltipBox[
         TagBox[
          DynamicBox[{
            FEPrivate`If[
             CurrentValue["MouseOver"], 
             EdgeForm[{
               GrayLevel[0.5], 
               AbsoluteThickness[1.5], 
               Opacity[0.66]}], {}, {}], 
            
            RectangleBox[{4.2, 0.}, {4.4, 0.010997800439912008`}, 
             "RoundingRadius" -> 0]},
           
           ImageSizeCache->{{69.80972388955583, 72.89855942376951}, {
            41.34810995126489, 44.114928368734745`}}],
          StatusArea[#, 0.010997800439912008`]& ,
          TagBoxNote->"0.010997800439912008"],
         StyleBox["0.010997800439912008`", {
           GrayLevel[0]}, StripOnInput -> False]],
        Annotation[#, 
         Style[0.010997800439912008`, {
           GrayLevel[0]}], "Tooltip"]& ], 
       TagBox[
        TooltipBox[
         TagBox[
          DynamicBox[{
            FEPrivate`If[
             CurrentValue["MouseOver"], 
             EdgeForm[{
               GrayLevel[0.5], 
               AbsoluteThickness[1.5], 
               Opacity[0.66]}], {}, {}], 
            
            RectangleBox[{4.4, 0.}, {4.6, 0.008998200359928047}, 
             "RoundingRadius" -> 0]},
           
           ImageSizeCache->{{71.89855942376951, 74.98739495798318}, {
            41.669349663532124`, 44.114928368734745`}}],
          StatusArea[#, 0.008998200359928047]& ,
          TagBoxNote->"0.008998200359928047"],
         StyleBox["0.008998200359928047`", {
           GrayLevel[0]}, StripOnInput -> False]],
        Annotation[#, 
         Style[0.008998200359928047, {
           GrayLevel[0]}], "Tooltip"]& ], 
       TagBox[
        TooltipBox[
         TagBox[
          DynamicBox[{
            FEPrivate`If[
             CurrentValue["MouseOver"], 
             EdgeForm[{
               GrayLevel[0.5], 
               AbsoluteThickness[1.5], 
               Opacity[0.66]}], {}, {}], 
            
            RectangleBox[{4.6, 0.}, {4.8, 0.002999400119976002}, 
             "RoundingRadius" -> 0]},
           
           ImageSizeCache->{{73.98739495798318, 77.07623049219687}, {
            42.633068800333874`, 44.114928368734745`}}],
          StatusArea[#, 0.002999400119976002]& ,
          TagBoxNote->"0.002999400119976002"],
         StyleBox["0.002999400119976002`", {
           GrayLevel[0]}, StripOnInput -> False]],
        Annotation[#, 
         Style[0.002999400119976002, {
           GrayLevel[0]}], "Tooltip"]& ], 
       TagBox[
        TooltipBox[
         TagBox[
          DynamicBox[{
            FEPrivate`If[
             CurrentValue["MouseOver"], 
             EdgeForm[{
               GrayLevel[0.5], 
               AbsoluteThickness[1.5], 
               Opacity[0.66]}], {}, {}], 
            
            RectangleBox[{4.8, 0.}, {5., 0.011997600479904008`}, 
             "RoundingRadius" -> 0]},
           
           ImageSizeCache->{{76.07623049219687, 79.16506602641057}, {
            41.18749009513126, 44.114928368734745`}}],
          StatusArea[#, 0.011997600479904008`]& ,
          TagBoxNote->"0.011997600479904008"],
         StyleBox["0.011997600479904008`", {
           GrayLevel[0]}, StripOnInput -> False]],
        Annotation[#, 
         Style[0.011997600479904008`, {
           GrayLevel[0]}], "Tooltip"]& ], 
       TagBox[
        TooltipBox[
         TagBox[
          DynamicBox[{
            FEPrivate`If[
             CurrentValue["MouseOver"], 
             EdgeForm[{
               GrayLevel[0.5], 
               AbsoluteThickness[1.5], 
               Opacity[0.66]}], {}, {}], 
            
            RectangleBox[{5., 0.}, {5.2, 0.01299740051989601}, 
             "RoundingRadius" -> 0]},
           
           ImageSizeCache->{{78.16506602641057, 81.25390156062426}, {
            41.02687023899764, 44.114928368734745`}}],
          StatusArea[#, 0.01299740051989601]& ,
          TagBoxNote->"0.01299740051989601"],
         StyleBox["0.01299740051989601`", {
           GrayLevel[0]}, StripOnInput -> False]],
        Annotation[#, 
         Style[0.01299740051989601, {
           GrayLevel[0]}], "Tooltip"]& ], 
       TagBox[
        TooltipBox[
         TagBox[
          DynamicBox[{
            FEPrivate`If[
             CurrentValue["MouseOver"], 
             EdgeForm[{
               GrayLevel[0.5], 
               AbsoluteThickness[1.5], 
               Opacity[0.66]}], {}, {}], 
            
            RectangleBox[{5.2, 0.}, {5.4, 0.003999200159968003}, 
             "RoundingRadius" -> 0]},
           
           ImageSizeCache->{{80.25390156062426, 83.34273709483794}, {
            42.47244894420025, 44.114928368734745`}}],
          StatusArea[#, 0.003999200159968003]& ,
          TagBoxNote->"0.003999200159968003"],
         StyleBox["0.003999200159968003`", {
           GrayLevel[0]}, StripOnInput -> False]],
        Annotation[#, 
         Style[0.003999200159968003, {
           GrayLevel[0]}], "Tooltip"]& ], 
       TagBox[
        TooltipBox[
         TagBox[
          DynamicBox[{
            FEPrivate`If[
             CurrentValue["MouseOver"], 
             EdgeForm[{
               GrayLevel[0.5], 
               AbsoluteThickness[1.5], 
               Opacity[0.66]}], {}, {}], 
            
            RectangleBox[{5.4, 0.}, {5.6, 0.0029994001199760154`}, 
             "RoundingRadius" -> 0]},
           
           ImageSizeCache->{{82.34273709483794, 85.43157262905162}, {
            42.633068800333874`, 44.114928368734745`}}],
          StatusArea[#, 0.0029994001199760154`]& ,
          TagBoxNote->"0.0029994001199760154"],
         StyleBox["0.0029994001199760154`", {
           GrayLevel[0]}, StripOnInput -> False]],
        Annotation[#, 
         Style[0.0029994001199760154`, {
           GrayLevel[0]}], "Tooltip"]& ], 
       TagBox[
        TooltipBox[
         TagBox[
          DynamicBox[{
            FEPrivate`If[
             CurrentValue["MouseOver"], 
             EdgeForm[{
               GrayLevel[0.5], 
               AbsoluteThickness[1.5], 
               Opacity[0.66]}], {}, {}], 
            
            RectangleBox[{5.6, 0.}, {5.8, 0.004999000199960003}, 
             "RoundingRadius" -> 0]},
           
           ImageSizeCache->{{84.43157262905162, 87.5204081632653}, {
            42.311829088066624`, 44.114928368734745`}}],
          StatusArea[#, 0.004999000199960003]& ,
          TagBoxNote->"0.004999000199960003"],
         StyleBox["0.004999000199960003`", {
           GrayLevel[0]}, StripOnInput -> False]],
        Annotation[#, 
         Style[0.004999000199960003, {
           GrayLevel[0]}], "Tooltip"]& ], 
       TagBox[
        TooltipBox[
         TagBox[
          DynamicBox[{
            FEPrivate`If[
             CurrentValue["MouseOver"], 
             EdgeForm[{
               GrayLevel[0.5], 
               AbsoluteThickness[1.5], 
               Opacity[0.66]}], {}, {}], 
            
            RectangleBox[{5.8, 0.}, {6., 0.005998800239952004}, 
             "RoundingRadius" -> 0]},
           
           ImageSizeCache->{{86.5204081632653, 89.60924369747899}, {
            42.151209231933, 44.114928368734745`}}],
          StatusArea[#, 0.005998800239952004]& ,
          TagBoxNote->"0.005998800239952004"],
         StyleBox["0.005998800239952004`", {
           GrayLevel[0]}, StripOnInput -> False]],
        Annotation[#, 
         Style[0.005998800239952004, {
           GrayLevel[0]}], "Tooltip"]& ], 
       TagBox[
        TooltipBox[
         TagBox[
          DynamicBox[{
            FEPrivate`If[
             CurrentValue["MouseOver"], 
             EdgeForm[{
               GrayLevel[0.5], 
               AbsoluteThickness[1.5], 
               Opacity[0.66]}], {}, {}], 
            
            RectangleBox[{6., 0.}, {6.2, 0.0019996000799840014`}, 
             "RoundingRadius" -> 0]},
           
           ImageSizeCache->{{88.60924369747899, 91.69807923169269}, {
            42.793688656467495`, 44.114928368734745`}}],
          StatusArea[#, 0.0019996000799840014`]& ,
          TagBoxNote->"0.0019996000799840014"],
         StyleBox["0.0019996000799840014`", {
           GrayLevel[0]}, StripOnInput -> False]],
        Annotation[#, 
         Style[0.0019996000799840014`, {
           GrayLevel[0]}], "Tooltip"]& ], 
       TagBox[
        TooltipBox[
         TagBox[
          DynamicBox[{
            FEPrivate`If[
             CurrentValue["MouseOver"], 
             EdgeForm[{
               GrayLevel[0.5], 
               AbsoluteThickness[1.5], 
               Opacity[0.66]}], {}, {}], 
            
            RectangleBox[{6.2, 0.}, {6.4, 0.002999400119976002}, 
             "RoundingRadius" -> 0]},
           
           ImageSizeCache->{{90.69807923169269, 93.78691476590637}, {
            42.633068800333874`, 44.114928368734745`}}],
          StatusArea[#, 0.002999400119976002]& ,
          TagBoxNote->"0.002999400119976002"],
         StyleBox["0.002999400119976002`", {
           GrayLevel[0]}, StripOnInput -> False]],
        Annotation[#, 
         Style[0.002999400119976002, {
           GrayLevel[0]}], "Tooltip"]& ], 
       TagBox[
        TooltipBox[
         TagBox[
          DynamicBox[{
            FEPrivate`If[
             CurrentValue["MouseOver"], 
             EdgeForm[{
               GrayLevel[0.5], 
               AbsoluteThickness[1.5], 
               Opacity[0.66]}], {}, {}], 
            
            RectangleBox[{6.4, 0.}, {6.6, 0.00399920015996802}, 
             "RoundingRadius" -> 0]},
           
           ImageSizeCache->{{92.78691476590637, 95.87575030012005}, {
            42.472448944200245`, 44.114928368734745`}}],
          StatusArea[#, 0.00399920015996802]& ,
          TagBoxNote->"0.00399920015996802"],
         StyleBox["0.00399920015996802`", {
           GrayLevel[0]}, StripOnInput -> False]],
        Annotation[#, 
         Style[0.00399920015996802, {
           GrayLevel[0]}], "Tooltip"]& ], 
       TagBox[
        TooltipBox[
         TagBox[
          DynamicBox[{
            FEPrivate`If[
             CurrentValue["MouseOver"], 
             EdgeForm[{
               GrayLevel[0.5], 
               AbsoluteThickness[1.5], 
               Opacity[0.66]}], {}, {}], 
            
            RectangleBox[{6.6, 0.}, {6.8, 0.0019996000799840014`}, 
             "RoundingRadius" -> 0]},
           
           ImageSizeCache->{{94.87575030012005, 97.96458583433373}, {
            42.793688656467495`, 44.114928368734745`}}],
          StatusArea[#, 0.0019996000799840014`]& ,
          TagBoxNote->"0.0019996000799840014"],
         StyleBox["0.0019996000799840014`", {
           GrayLevel[0]}, StripOnInput -> False]],
        Annotation[#, 
         Style[0.0019996000799840014`, {
           GrayLevel[0]}], "Tooltip"]& ], 
       TagBox[
        TooltipBox[
         TagBox[
          DynamicBox[{
            FEPrivate`If[
             CurrentValue["MouseOver"], 
             EdgeForm[{
               GrayLevel[0.5], 
               AbsoluteThickness[1.5], 
               Opacity[0.66]}], {}, {}], 
            
            RectangleBox[{6.8, 0.}, {7., 0.002999400119976002}, 
             "RoundingRadius" -> 0]},
           
           ImageSizeCache->{{96.96458583433373, 100.05342136854742`}, {
            42.633068800333874`, 44.114928368734745`}}],
          StatusArea[#, 0.002999400119976002]& ,
          TagBoxNote->"0.002999400119976002"],
         StyleBox["0.002999400119976002`", {
           GrayLevel[0]}, StripOnInput -> False]],
        Annotation[#, 
         Style[0.002999400119976002, {
           GrayLevel[0]}], "Tooltip"]& ], 
       TagBox[
        TooltipBox[
         TagBox[
          DynamicBox[{
            FEPrivate`If[
             CurrentValue["MouseOver"], 
             EdgeForm[{
               GrayLevel[0.5], 
               AbsoluteThickness[1.5], 
               Opacity[0.66]}], {}, {}], 
            
            RectangleBox[{7., 0.}, {7.2, 0.003999200159968003}, 
             "RoundingRadius" -> 0]},
           
           ImageSizeCache->{{99.05342136854742, 102.1422569027611}, {
            42.47244894420025, 44.114928368734745`}}],
          StatusArea[#, 0.003999200159968003]& ,
          TagBoxNote->"0.003999200159968003"],
         StyleBox["0.003999200159968003`", {
           GrayLevel[0]}, StripOnInput -> False]],
        Annotation[#, 
         Style[0.003999200159968003, {
           GrayLevel[0]}], "Tooltip"]& ], 
       TagBox[
        TooltipBox[
         TagBox[
          DynamicBox[{
            FEPrivate`If[
             CurrentValue["MouseOver"], 
             EdgeForm[{
               GrayLevel[0.5], 
               AbsoluteThickness[1.5], 
               Opacity[0.66]}], {}, {}], 
            
            RectangleBox[{7.2, 0.}, {7.4, 0.005998800239952004}, 
             "RoundingRadius" -> 0]},
           
           ImageSizeCache->{{101.1422569027611, 104.23109243697479`}, {
            42.151209231933, 44.114928368734745`}}],
          StatusArea[#, 0.005998800239952004]& ,
          TagBoxNote->"0.005998800239952004"],
         StyleBox["0.005998800239952004`", {
           GrayLevel[0]}, StripOnInput -> False]],
        Annotation[#, 
         Style[0.005998800239952004, {
           GrayLevel[0]}], "Tooltip"]& ], 
       TagBox[
        TooltipBox[
         TagBox[
          DynamicBox[{
            FEPrivate`If[
             CurrentValue["MouseOver"], 
             EdgeForm[{
               GrayLevel[0.5], 
               AbsoluteThickness[1.5], 
               Opacity[0.66]}], {}, {}], 
            
            RectangleBox[{7.4, 0.}, {7.6, 0.0029994001199760154`}, 
             "RoundingRadius" -> 0]},
           
           ImageSizeCache->{{103.23109243697479`, 106.31992797118848`}, {
            42.633068800333874`, 44.114928368734745`}}],
          StatusArea[#, 0.0029994001199760154`]& ,
          TagBoxNote->"0.0029994001199760154"],
         StyleBox["0.0029994001199760154`", {
           GrayLevel[0]}, StripOnInput -> False]],
        Annotation[#, 
         Style[0.0029994001199760154`, {
           GrayLevel[0]}], "Tooltip"]& ], 
       TagBox[
        TooltipBox[
         TagBox[
          DynamicBox[{
            FEPrivate`If[
             CurrentValue["MouseOver"], 
             EdgeForm[{
               GrayLevel[0.5], 
               AbsoluteThickness[1.5], 
               Opacity[0.66]}], {}, {}], 
            
            RectangleBox[{7.6, 0.}, {7.8, 0.002999400119976002}, 
             "RoundingRadius" -> 0]},
           
           ImageSizeCache->{{105.31992797118848`, 108.40876350540216`}, {
            42.633068800333874`, 44.114928368734745`}}],
          StatusArea[#, 0.002999400119976002]& ,
          TagBoxNote->"0.002999400119976002"],
         StyleBox["0.002999400119976002`", {
           GrayLevel[0]}, StripOnInput -> False]],
        Annotation[#, 
         Style[0.002999400119976002, {
           GrayLevel[0]}], "Tooltip"]& ], 
       TagBox[
        TooltipBox[
         TagBox[
          DynamicBox[{
            FEPrivate`If[
             CurrentValue["MouseOver"], 
             EdgeForm[{
               GrayLevel[0.5], 
               AbsoluteThickness[1.5], 
               Opacity[0.66]}], {}, {}], 
            
            RectangleBox[{7.8, 0.}, {8., 0.0009998000399920007}, 
             "RoundingRadius" -> 0]},
           
           ImageSizeCache->{{107.40876350540216`, 110.49759903961585`}, {
            42.95430851260112, 44.114928368734745`}}],
          StatusArea[#, 0.0009998000399920007]& ,
          TagBoxNote->"0.0009998000399920007"],
         StyleBox["0.0009998000399920007`", {
           GrayLevel[0]}, StripOnInput -> False]],
        Annotation[#, 
         Style[0.0009998000399920007, {
           GrayLevel[0]}], "Tooltip"]& ], {}, 
       TagBox[
        TooltipBox[
         TagBox[
          DynamicBox[{
            FEPrivate`If[
             CurrentValue["MouseOver"], 
             EdgeForm[{
               GrayLevel[0.5], 
               AbsoluteThickness[1.5], 
               Opacity[0.66]}], {}, {}], 
            
            RectangleBox[{8.2, 0.}, {8.4, 0.0029994001199759885`}, 
             "RoundingRadius" -> 0]},
           
           ImageSizeCache->{{111.58643457382952`, 114.67527010804322`}, {
            42.633068800333874`, 44.114928368734745`}}],
          StatusArea[#, 0.0029994001199759885`]& ,
          TagBoxNote->"0.0029994001199759885"],
         StyleBox["0.0029994001199759885`", {
           GrayLevel[0]}, StripOnInput -> False]],
        Annotation[#, 
         Style[0.0029994001199759885`, {
           GrayLevel[0]}], "Tooltip"]& ], 
       TagBox[
        TooltipBox[
         TagBox[
          DynamicBox[{
            FEPrivate`If[
             CurrentValue["MouseOver"], 
             EdgeForm[{
               GrayLevel[0.5], 
               AbsoluteThickness[1.5], 
               Opacity[0.66]}], {}, {}], 
            
            RectangleBox[{8.4, 0.}, {8.6, 0.000999800039992005}, 
             "RoundingRadius" -> 0]},
           
           ImageSizeCache->{{113.67527010804322`, 116.76410564225691`}, {
            42.95430851260112, 44.114928368734745`}}],
          StatusArea[#, 0.000999800039992005]& ,
          TagBoxNote->"0.000999800039992005"],
         StyleBox["0.000999800039992005`", {
           GrayLevel[0]}, StripOnInput -> False]],
        Annotation[#, 
         Style[0.000999800039992005, {
           GrayLevel[0]}], "Tooltip"]& ], {}, 
       TagBox[
        TooltipBox[
         TagBox[
          DynamicBox[{
            FEPrivate`If[
             CurrentValue["MouseOver"], 
             EdgeForm[{
               GrayLevel[0.5], 
               AbsoluteThickness[1.5], 
               Opacity[0.66]}], {}, {}], 
            
            RectangleBox[{8.8, 0.}, {9., 0.000999800039992005}, 
             "RoundingRadius" -> 0]},
           
           ImageSizeCache->{{117.8529411764706, 120.94177671068428`}, {
            42.95430851260112, 44.114928368734745`}}],
          StatusArea[#, 0.000999800039992005]& ,
          TagBoxNote->"0.000999800039992005"],
         StyleBox["0.000999800039992005`", {
           GrayLevel[0]}, StripOnInput -> False]],
        Annotation[#, 
         Style[0.000999800039992005, {
           GrayLevel[0]}], "Tooltip"]& ], 
       TagBox[
        TooltipBox[
         TagBox[
          DynamicBox[{
            FEPrivate`If[
             CurrentValue["MouseOver"], 
             EdgeForm[{
               GrayLevel[0.5], 
               AbsoluteThickness[1.5], 
               Opacity[0.66]}], {}, {}], 
            
            RectangleBox[{9., 0.}, {9.2, 0.00199960007998401}, 
             "RoundingRadius" -> 0]},
           
           ImageSizeCache->{{119.94177671068428`, 123.03061224489795`}, {
            42.793688656467495`, 44.114928368734745`}}],
          StatusArea[#, 0.00199960007998401]& ,
          TagBoxNote->"0.00199960007998401"],
         StyleBox["0.00199960007998401`", {
           GrayLevel[0]}, StripOnInput -> False]],
        Annotation[#, 
         Style[0.00199960007998401, {
           GrayLevel[0]}], "Tooltip"]& ], 
       TagBox[
        TooltipBox[
         TagBox[
          DynamicBox[{
            FEPrivate`If[
             CurrentValue["MouseOver"], 
             EdgeForm[{
               GrayLevel[0.5], 
               AbsoluteThickness[1.5], 
               Opacity[0.66]}], {}, {}], 
            
            RectangleBox[{9.2, 0.}, {9.4, 0.005998800239951977}, 
             "RoundingRadius" -> 0]},
           
           ImageSizeCache->{{122.03061224489795`, 125.11944777911165`}, {
            42.15120923193301, 44.114928368734745`}}],
          StatusArea[#, 0.005998800239951977]& ,
          TagBoxNote->"0.005998800239951977"],
         StyleBox["0.005998800239951977`", {
           GrayLevel[0]}, StripOnInput -> False]],
        Annotation[#, 
         Style[0.005998800239951977, {
           GrayLevel[0]}], "Tooltip"]& ], {}, 
       TagBox[
        TooltipBox[
         TagBox[
          DynamicBox[{
            FEPrivate`If[
             CurrentValue["MouseOver"], 
             EdgeForm[{
               GrayLevel[0.5], 
               AbsoluteThickness[1.5], 
               Opacity[0.66]}], {}, {}], 
            
            RectangleBox[{9.6, 0.}, {9.8, 0.003999200159967985}, 
             "RoundingRadius" -> 0]},
           
           ImageSizeCache->{{126.20828331332532`, 129.29711884753902`}, {
            42.47244894420025, 44.114928368734745`}}],
          StatusArea[#, 0.003999200159967985]& ,
          TagBoxNote->"0.003999200159967985"],
         StyleBox["0.003999200159967985`", {
           GrayLevel[0]}, StripOnInput -> False]],
        Annotation[#, 
         Style[0.003999200159967985, {
           GrayLevel[0]}], "Tooltip"]& ], 
       TagBox[
        TooltipBox[
         TagBox[
          DynamicBox[{
            FEPrivate`If[
             CurrentValue["MouseOver"], 
             EdgeForm[{
               GrayLevel[0.5], 
               AbsoluteThickness[1.5], 
               Opacity[0.66]}], {}, {}], 
            
            RectangleBox[{9.8, 0.}, {10., 0.000999800039992005}, 
             "RoundingRadius" -> 0]},
           
           ImageSizeCache->{{128.29711884753902`, 131.3859543817527}, {
            42.95430851260112, 44.114928368734745`}}],
          StatusArea[#, 0.000999800039992005]& ,
          TagBoxNote->"0.000999800039992005"],
         StyleBox["0.000999800039992005`", {
           GrayLevel[0]}, StripOnInput -> False]],
        Annotation[#, 
         Style[0.000999800039992005, {
           GrayLevel[0]}], "Tooltip"]& ], {}, 
       TagBox[
        TooltipBox[
         TagBox[
          DynamicBox[{
            FEPrivate`If[
             CurrentValue["MouseOver"], 
             EdgeForm[{
               GrayLevel[0.5], 
               AbsoluteThickness[1.5], 
               Opacity[0.66]}], {}, {}], 
            
            RectangleBox[{10.2, 0.}, {10.4, 0.0019996000799839923`}, 
             "RoundingRadius" -> 0]},
           
           ImageSizeCache->{{132.47478991596637`, 135.56362545018007`}, {
            42.7936886564675, 44.114928368734745`}}],
          StatusArea[#, 0.0019996000799839923`]& ,
          TagBoxNote->"0.0019996000799839923"],
         StyleBox["0.0019996000799839923`", {
           GrayLevel[0]}, StripOnInput -> False]],
        Annotation[#, 
         Style[0.0019996000799839923`, {
           GrayLevel[0]}], "Tooltip"]& ], {}, 
       TagBox[
        TooltipBox[
         TagBox[
          DynamicBox[{
            FEPrivate`If[
             CurrentValue["MouseOver"], 
             EdgeForm[{
               GrayLevel[0.5], 
               AbsoluteThickness[1.5], 
               Opacity[0.66]}], {}, {}], 
            
            RectangleBox[{10.6, 0.}, {10.8, 0.0009998000399919962}, 
             "RoundingRadius" -> 0]},
           
           ImageSizeCache->{{136.65246098439374`, 139.74129651860744`}, {
            42.95430851260112, 44.114928368734745`}}],
          StatusArea[#, 0.0009998000399919962]& ,
          TagBoxNote->"0.0009998000399919962"],
         StyleBox["0.0009998000399919962`", {
           GrayLevel[0]}, StripOnInput -> False]],
        Annotation[#, 
         Style[0.0009998000399919962, {
           GrayLevel[0]}], "Tooltip"]& ], 
       TagBox[
        TooltipBox[
         TagBox[
          DynamicBox[{
            FEPrivate`If[
             CurrentValue["MouseOver"], 
             EdgeForm[{
               GrayLevel[0.5], 
               AbsoluteThickness[1.5], 
               Opacity[0.66]}], {}, {}], 
            
            RectangleBox[{10.8, 0.}, {11., 0.000999800039992005}, 
             "RoundingRadius" -> 0]},
           
           ImageSizeCache->{{138.74129651860744`, 141.8301320528211}, {
            42.95430851260112, 44.114928368734745`}}],
          StatusArea[#, 0.000999800039992005]& ,
          TagBoxNote->"0.000999800039992005"],
         StyleBox["0.000999800039992005`", {
           GrayLevel[0]}, StripOnInput -> False]],
        Annotation[#, 
         Style[0.000999800039992005, {
           GrayLevel[0]}], "Tooltip"]& ], 
       TagBox[
        TooltipBox[
         TagBox[
          DynamicBox[{
            FEPrivate`If[
             CurrentValue["MouseOver"], 
             EdgeForm[{
               GrayLevel[0.5], 
               AbsoluteThickness[1.5], 
               Opacity[0.66]}], {}, {}], 
            
            RectangleBox[{11., 0.}, {11.2, 0.000999800039992005}, 
             "RoundingRadius" -> 0]},
           
           ImageSizeCache->{{140.8301320528211, 143.9189675870348}, {
            42.95430851260112, 44.114928368734745`}}],
          StatusArea[#, 0.000999800039992005]& ,
          TagBoxNote->"0.000999800039992005"],
         StyleBox["0.000999800039992005`", {
           GrayLevel[0]}, StripOnInput -> False]],
        Annotation[#, 
         Style[0.000999800039992005, {
           GrayLevel[0]}], "Tooltip"]& ], {}, {}, {}, {}, {}, {}, 
       TagBox[
        TooltipBox[
         TagBox[
          DynamicBox[{
            FEPrivate`If[
             CurrentValue["MouseOver"], 
             EdgeForm[{
               GrayLevel[0.5], 
               AbsoluteThickness[1.5], 
               Opacity[0.66]}], {}, {}], 
            
            RectangleBox[{12.4, 0.}, {12.6, 0.000999800039992005}, 
             "RoundingRadius" -> 0]},
           
           ImageSizeCache->{{155.45198079231693`, 158.5408163265306}, {
            42.95430851260112, 44.114928368734745`}}],
          StatusArea[#, 0.000999800039992005]& ,
          TagBoxNote->"0.000999800039992005"],
         StyleBox["0.000999800039992005`", {
           GrayLevel[0]}, StripOnInput -> False]],
        Annotation[#, 
         Style[0.000999800039992005, {
           GrayLevel[0]}], "Tooltip"]& ]}, {}, {}}, {{}, {}, 
      {RGBColor[1, 0, 0], Thickness[Large], LineBox[CompressedData["
1:eJwVlnc01/8Xx+2s7L1CxodI+RJS3VeRyMguJBQZIatCJFtIZO9khOyd6G0m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        "]]}}},
    AspectRatio->NCache[GoldenRatio^(-1), 0.6180339887498948],
    Axes->{True, True},
    AxesLabel->{
      FormBox["\"x2\"", TraditionalForm], 
      FormBox["\"freq\"", TraditionalForm]},
    AxesOrigin->{-1., 0},
    FrameTicks->{{Automatic, Automatic}, {Automatic, Automatic}},
    GridLines->{None, None},
    PlotRange->NCache[{{-1, 
        Rational[63, 5]}, {All, All}}, {{-1, 12.6}, {All, All}}],
    PlotRangePadding->{{
       Scaled[0.02], 
       Scaled[0.02]}, {
       Scaled[0.02], 
       Scaled[0.1]}},
    Ticks->{Automatic, Automatic}]}], "}"}]], "Output",
 ImageSize->{389, 126},
 ImageMargins->{{0, 0}, {0, 0}},
 ImageRegion->{{0, 1}, {0, 1}},
 CellLabel->"Out[5]=",
 CellID->138974571]
}, Open  ]]
}, Closed]]
}, Open  ]],

Cell[CellGroupData[{

Cell[TextData[ButtonBox["MORE ABOUT",
 BaseStyle->None,
 Appearance->{Automatic, None},
 Evaluator->None,
 Method->"Preemptive",
 ButtonFunction:>(FrontEndExecute[{
    FrontEnd`SelectionMove[
     FrontEnd`SelectedNotebook[], All, ButtonCell], 
    FrontEndToken["OpenCloseGroup"], 
    FrontEnd`SelectionMove[
     FrontEnd`SelectedNotebook[], After, 
     CellContents]}]& )]], "MoreAboutSection",
 CellID->38303248],

Cell[TextData[ButtonBox["Metropolis",
 BaseStyle->"Link",
 ButtonData->"paclet:Metropolis/guide/Metropolis"]], "MoreAbout",
 CellID->131455931]
}, Open  ]],

Cell[" ", "FooterCell"]
},
Saveable->False,
ScreenStyleEnvironment->"Working",
WindowSize->{725, 730},
WindowMargins->{{0, Automatic}, {Automatic, 0}},
WindowTitle->"Metropolis - Wolfram Mathematica",
TaggingRules->{
 "ModificationHighlight" -> False, 
  "Metadata" -> {
   "built" -> "{2013, 2, 10, 17, 35, 45.6021086}", "context" -> "Metropolis`",
     "keywords" -> {}, "index" -> True, "label" -> "Metropolis Paclet Symbol",
     "language" -> "en", "paclet" -> "Metropolis", "status" -> "None", 
    "summary" -> 
    "Metropolis[logpdf, logpdfjump, randomjump, x0, n] generates n Metropolis \
updates starting from x0. The target distribution is p[x] with the proposal \
distribution J[xt, x] randomized by x = randomjump[xt] given current xt. Here \
logpdf[x] = Log[p[x]] and logpdfjump[xt,x] = Log[J[xt,x]]", "synonyms" -> {}, 
    "title" -> "Metropolis", "type" -> "Symbol", "uri" -> 
    "Metropolis/ref/Metropolis"}, "LinkTrails" -> "", "SearchTextTranslated" -> 
  ""},
CellContext->"Global`",
FrontEndVersion->"8.0 for Microsoft Windows (64-bit) (February 23, 2011)",
StyleDefinitions->Notebook[{
   Cell[
    StyleData[
    StyleDefinitions -> FrontEnd`FileName[{"Wolfram"}, "Reference.nb"]]], 
   Cell[
    StyleData["Input"], CellContext -> "Global`"], 
   Cell[
    StyleData["Output"], CellContext -> "Global`"]}, Visible -> False, 
  FrontEndVersion -> "8.0 for Microsoft Windows (64-bit) (February 23, 2011)",
   StyleDefinitions -> "Default.nb"]
]
(* End of Notebook Content *)

(* Internal cache information *)
(*CellTagsOutline
CellTagsIndex->{
 "PrimaryExamplesSection"->{
  Cell[6372, 184, 460, 13, 70, "PrimaryExamplesSection",
   CellTags->"PrimaryExamplesSection",
   CellID->42576746]}
 }
*)
(*CellTagsIndex
CellTagsIndex->{
 {"PrimaryExamplesSection", 194671, 4986}
 }
*)
(*NotebookFileOutline
Notebook[{
Cell[603, 21, 76, 1, 70, "SymbolColorBar"],
Cell[682, 24, 185, 5, 70, "LinkTrail"],
Cell[870, 31, 2319, 53, 70, "AnchorBarGrid",
 CellID->1],
Cell[CellGroupData[{
Cell[3214, 88, 53, 1, 70, "ObjectName",
 CellID->1224892054],
Cell[3270, 91, 675, 17, 70, "Usage",
 CellID->982511436]
}, Open  ]],
Cell[CellGroupData[{
Cell[3982, 113, 580, 17, 70, "NotesSection",
 CellGroupingRules->{"SectionGrouping", 50},
 CellID->234101828],
Cell[4565, 132, 142, 4, 70, "Notes",
 CellID->85410339],
Cell[4710, 138, 73, 1, 70, "Notes",
 CellID->182314388],
Cell[4786, 141, 1407, 32, 70, "3ColumnTableMod",
 CellID->497869101],
Cell[6196, 175, 139, 4, 70, "Notes",
 CellID->588891594]
}, Closed]],
Cell[CellGroupData[{
Cell[6372, 184, 460, 13, 70, "PrimaryExamplesSection",
 CellTags->"PrimaryExamplesSection",
 CellID->42576746],
Cell[CellGroupData[{
Cell[6857, 201, 148, 5, 70, "ExampleSection",
 CellID->121590637],
Cell[7008, 208, 119, 3, 70, "Input",
 CellID->223020744],
Cell[7130, 213, 87, 1, 70, "ExampleText",
 CellID->34393979],
Cell[CellGroupData[{
Cell[7242, 218, 1775, 52, 70, "Input",
 CellID->294782483],
Cell[9020, 272, 222, 7, 36, "Output",
 CellID->2875901]
}, Open  ]],
Cell[9257, 282, 60, 1, 70, "ExampleText",
 CellID->12821957],
Cell[CellGroupData[{
Cell[9342, 287, 279, 10, 70, "Input",
 CellID->891160780],
Cell[9624, 299, 243, 8, 36, "Output",
 CellID->184204190]
}, Open  ]],
Cell[9882, 310, 54, 1, 70, "ExampleText",
 CellID->60397930],
Cell[CellGroupData[{
Cell[9961, 315, 398, 12, 28, "Input",
 CellID->111189838],
Cell[10362, 329, 7336, 128, 137, "Output",
 CellID->697109720]
}, Open  ]],
Cell[17713, 460, 101, 2, 70, "ExampleText",
 CellID->77108731],
Cell[CellGroupData[{
Cell[17839, 466, 973, 29, 70, "Input",
 CellID->66917885],
Cell[18815, 497, 35001, 885, 70, "Output",
 CellID->261084222]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[53865, 1388, 228, 7, 70, "ExampleSection",
 CellID->589267740],
Cell[54096, 1397, 96, 1, 70, "ExampleText",
 CellID->73418448],
Cell[CellGroupData[{
Cell[54217, 1402, 3382, 100, 70, "Input",
 CellID->251956668],
Cell[57602, 1504, 211, 7, 36, "Output",
 CellID->439528226]
}, Open  ]],
Cell[57828, 1514, 61, 1, 70, "ExampleText",
 CellID->183647242],
Cell[CellGroupData[{
Cell[57914, 1519, 255, 9, 70, "Input",
 CellID->265777506],
Cell[58172, 1530, 177, 5, 36, "Output",
 CellID->738619732]
}, Open  ]],
Cell[58364, 1538, 54, 1, 70, "ExampleText",
 CellID->193398859],
Cell[CellGroupData[{
Cell[58443, 1543, 573, 17, 70, "Input",
 CellID->309231733],
Cell[59019, 1562, 11125, 190, 70, "Output",
 CellID->66405875]
}, Open  ]],
Cell[CellGroupData[{
Cell[70181, 1757, 389, 12, 70, "Input",
 CellID->357276416],
Cell[70573, 1771, 3645, 67, 70, "Output",
 CellID->2079456]
}, Open  ]],
Cell[74233, 1841, 126, 3, 70, "ExampleText",
 CellID->395721816],
Cell[CellGroupData[{
Cell[74384, 1848, 2267, 68, 70, "Input",
 CellID->10673344],
Cell[76654, 1918, 222, 7, 36, "Output",
 CellID->32820248],
Cell[76879, 1927, 115555, 2993, 70, "Output",
 CellID->138974571]
}, Open  ]]
}, Closed]]
}, Open  ]],
Cell[CellGroupData[{
Cell[192495, 4927, 419, 12, 70, "MoreAboutSection",
 CellID->38303248],
Cell[192917, 4941, 143, 3, 70, "MoreAbout",
 CellID->131455931]
}, Open  ]],
Cell[193075, 4947, 23, 0, 70, "FooterCell"]
}
]
*)

(* End of internal cache information *)

