(* Content-type: application/vnd.wolfram.mathematica *)

(*** Wolfram Notebook File ***)
(* http://www.wolfram.com/nb *)

(* CreatedBy='Mathematica 8.0' *)

(*CacheID: 234*)
(* Internal cache information:
NotebookFileLineBreakTest
NotebookFileLineBreakTest
NotebookDataPosition[       157,          7]
NotebookDataLength[     16132,        473]
NotebookOptionsPosition[     13205,        382]
NotebookOutlinePosition[     15000,        431]
CellTagsIndexPosition[     14915,        426]
WindowTitle->RandomNegativeBinomial - Wolfram Mathematica
WindowFrame->Normal*)

(* Beginning of Notebook Content *)
Notebook[{
Cell[" ", "SymbolColorBar",
 CellMargins->{{Inherited, Inherited}, {-5, 0}}],

Cell[TextData[{
 ButtonBox["RandomGenerator",
  BaseStyle->{"Link", "LinkTrail"},
  ButtonData->"paclet:RandomGenerator/guide/RandomGenerator"],
 StyleBox[" > ", "LinkTrailSeparator"]
}], "LinkTrail"],

Cell[BoxData[GridBox[{
   {Cell["RANDOMGENERATOR PACKAGE PACLET SYMBOL", "PacletNameCell"], Cell[
    TextData[{
     Cell[BoxData[
      ActionMenuBox[
       FrameBox["\<\"More About \[RightGuillemet]\"\>",
        StripOnInput->False], {"\<\"RandomGenerator\"\>":>
       Documentation`HelpLookup[
       "paclet:RandomGenerator/guide/RandomGenerator"]},
       Appearance->None,
       MenuAppearance->Automatic]],
      LineSpacing->{1.4, 0}],
     "\[ThickSpace]\[ThickSpace]\[ThickSpace]\[ThickSpace]\[ThickSpace]\
\[ThickSpace]",
     Cell[BoxData[
      ActionMenuBox[
       FrameBox["\<\"URL \[RightGuillemet]\"\>",
        StripOnInput->
         False], {"\<\"RandomGenerator/ref/RandomNegativeBinomial\"\>":>
       None, "\<\"Copy Mathematica url\"\>":>
       Module[{DocumentationSearch`Private`nb$}, 
        DocumentationSearch`Private`nb$ = NotebookPut[
           Notebook[{
             Cell["RandomGenerator/ref/RandomNegativeBinomial"]}, Visible -> 
            False]]; SelectionMove[
         DocumentationSearch`Private`nb$, All, Notebook]; 
        FrontEndTokenExecute[DocumentationSearch`Private`nb$, "Copy"]; 
        NotebookClose[DocumentationSearch`Private`nb$]; Null], 
       Delimiter, "\<\"Copy web url\"\>":>
       Module[{DocumentationSearch`Private`nb$}, 
        DocumentationSearch`Private`nb$ = NotebookPut[
           Notebook[{
             Cell[
              BoxData[
               MakeBoxes[
                Hyperlink[
                "http://reference.wolfram.com/mathematica/RandomGenerator/ref/\
RandomNegativeBinomial.html"], StandardForm]], "Input", TextClipboardType -> 
              "PlainText"]}, Visible -> False]]; 
        SelectionMove[DocumentationSearch`Private`nb$, All, Notebook]; 
        FrontEndTokenExecute[DocumentationSearch`Private`nb$, "Copy"]; 
        NotebookClose[DocumentationSearch`Private`nb$]; 
        Null], "\<\"Go to web url\"\>":>FrontEndExecute[{
         NotebookLocate[{
           URL[
            StringJoin[
            "http://reference.wolfram.com/mathematica/", 
             "RandomGenerator/ref/RandomNegativeBinomial", ".html"]], 
           None}]}]},
       Appearance->None,
       MenuAppearance->Automatic]],
      LineSpacing->{1.4, 0}]
    }], "AnchorBar"]}
  }]], "AnchorBarGrid",
 GridBoxOptions->{GridBoxItemSize->{"Columns" -> {
     Scaled[0.65], {
      Scaled[0.34]}}, "ColumnsIndexed" -> {}, "Rows" -> {{1.}}, 
   "RowsIndexed" -> {}}},
 CellID->1],

Cell[CellGroupData[{

Cell["RandomNegativeBinomial", "ObjectName",
 CellID->1224892054],

Cell[BoxData[GridBox[{
   {"", Cell[TextData[{
     Cell[BoxData[
      RowBox[{"RandomNegativeBinomial", "[", 
       RowBox[{
        StyleBox["n", "TI"], ",", " ", 
        StyleBox["p", "TI"], ",", " ", 
        StyleBox["lb", "TI"], ",", " ", 
        StyleBox["s", "TI"]}], "]"}]], "InlineFormula"],
     "\[LineSeparator]gives s random draws from the negative binomial \
distribution with parameters n and p, truncted to be not less than lb. The \
defaults lb=0 and s=1."
    }]]}
  }]], "Usage",
 GridBoxOptions->{
 GridBoxBackground->{
  "Columns" -> {{None}}, "ColumnsIndexed" -> {}, "Rows" -> {{None}}, 
   "RowsIndexed" -> {}}}]
}, Open  ]],

Cell[CellGroupData[{

Cell[TextData[ButtonBox["EXAMPLES",
 BaseStyle->None,
 Appearance->{Automatic, None},
 Evaluator->None,
 Method->"Preemptive",
 ButtonFunction:>(FrontEndExecute[{
    FrontEnd`SelectionMove[
     FrontEnd`SelectedNotebook[], All, ButtonCell], 
    FrontEndToken["OpenCloseGroup"], 
    FrontEnd`SelectionMove[
     FrontEnd`SelectedNotebook[], After, 
     CellContents]}]& )]], "PrimaryExamplesSection",
 CellTags->"PrimaryExamplesSection",
 CellID->85057143],

Cell[CellGroupData[{

Cell[TextData[{
 "Basic Examples",
 "\[NonBreakingSpace]\[NonBreakingSpace]",
 Cell["(1)", "ExampleCount"]
}], "ExampleSection",
 CellID->144596749],

Cell[BoxData[
 RowBox[{"Needs", "[", "\"\<RandomGenerator`\>\"", "]"}]], "Input",
 CellLabel->"In[1]:=",
 CellID->168232019],

Cell[CellGroupData[{

Cell[BoxData[{
 RowBox[{"Clear", "[", 
  RowBox[{"n", ",", "p", ",", "lb", ",", "f", ",", "c", ",", "res"}], 
  "]"}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{"n", "=", "7"}], ";"}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{"p", "=", "0.5"}], ";"}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{"lb", "=", "5"}], ";"}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{"size", "=", "10000"}], ";"}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{
   RowBox[{"f", "[", "x_", "]"}], ":=", 
   RowBox[{
    RowBox[{"Binomial", "[", 
     RowBox[{
      RowBox[{"n", "+", "x", "-", "1"}], ",", 
      RowBox[{"n", "-", "1"}]}], "]"}], "  ", 
    RowBox[{"p", "^", "n"}], " ", 
    RowBox[{
     RowBox[{"(", 
      RowBox[{"1", "-", "p"}], ")"}], "^", "x"}]}]}], 
  ";"}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{"c", "=", 
   RowBox[{"N", "[", 
    RowBox[{"Sum", "[", 
     RowBox[{
      RowBox[{"f", "[", "x", "]"}], ",", 
      RowBox[{"{", 
       RowBox[{"x", ",", "lb", ",", "Infinity"}], "}"}]}], "]"}], "]"}]}], 
  ";"}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{"res", "=", 
   RowBox[{"RandomNegativeBinomial", "[", 
    RowBox[{"n", ",", "p", ",", "lb", ",", "size"}], "]"}]}], 
  ";"}], "\[IndentingNewLine]", 
 RowBox[{"Show", "[", 
  RowBox[{
   RowBox[{"ListPlot", "[", 
    RowBox[{
     RowBox[{"Table", "[", 
      RowBox[{
       RowBox[{"{", 
        RowBox[{"i", ",", 
         RowBox[{
          RowBox[{"Count", "[", 
           RowBox[{"res", ",", "i"}], "]"}], "/", "size"}]}], "}"}], ",", 
       RowBox[{"{", 
        RowBox[{"i", ",", "lb", ",", 
         RowBox[{"Max", "[", "res", "]"}]}], "}"}]}], "]"}], ",", 
     RowBox[{"PlotRange", "\[Rule]", "All"}], ",", 
     RowBox[{"PlotMarkers", "\[Rule]", "Automatic"}]}], "]"}], ",", 
   RowBox[{"Plot", "[", 
    RowBox[{
     RowBox[{
      RowBox[{"f", "[", "x", "]"}], "/", "c"}], ",", 
     RowBox[{"{", 
      RowBox[{"x", ",", "lb", ",", 
       RowBox[{"Max", "[", "res", "]"}]}], "}"}], ",", 
     RowBox[{"PlotStyle", "\[Rule]", 
      RowBox[{"{", 
       RowBox[{"Red", ",", "Thick"}], "}"}]}], ",", 
     RowBox[{"PlotRange", "\[Rule]", "All"}]}], "]"}]}], "]"}]}], "Input",
 CellLabel->"In[2]:=",
 CellID->104442095],

Cell[BoxData[
 GraphicsBox[
  TagBox[RasterBox[CompressedData["
1:eJztnW1oFEcYxw9tbEirUYJtis2LQULqG9FKfatYCMq1ppCLEFKCGBIJhkJM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    "], {{0, 113}, {180, 0}}, {0, 255},
    ColorFunction->RGBColor],
   BoxForm`ImageTag["Byte", ColorSpace -> "RGB", Interleaving -> True],
   Selectable->False],
  BaseStyle->"ImageGraphics",
  ImageSize->Automatic,
  ImageSizeRaw->{180, 113},
  PlotRange->{{0, 180}, {0, 113}}]], "Output",
 ImageSize->{184, 113},
 ImageMargins->{{0, 0}, {0, 0}},
 ImageRegion->{{0, 1}, {0, 1}},
 CellLabel->"Out[2]=",
 CellID->258469057]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[{
 RowBox[{
  RowBox[{
   RowBox[{"RandomNegativeBinomial", "[", 
    RowBox[{"n", ",", "p", ",", "lb", ",", "size"}], "]"}], ";"}], "//", 
  "Timing"}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{
   RowBox[{"Table", "[", 
    RowBox[{
     RowBox[{"RandomNegativeBinomial", "[", 
      RowBox[{"n", ",", "p", ",", "lb"}], "]"}], ",", 
     RowBox[{"{", "size", "}"}]}], "]"}], ";"}], "//", 
  "Timing"}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{
   RowBox[{"RandomInteger", "[", 
    RowBox[{
     RowBox[{"NegativeBinomialDistribution", "[", 
      RowBox[{"n", ",", "p"}], "]"}], ",", "size"}], "]"}], ";"}], "//", 
  "Timing"}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{
   RowBox[{"Table", "[", 
    RowBox[{
     RowBox[{"RandomInteger", "[", 
      RowBox[{"NegativeBinomialDistribution", "[", 
       RowBox[{"n", ",", "p"}], "]"}], "]"}], ",", 
     RowBox[{"{", "size", "}"}]}], "]"}], ";"}], "//", "Timing"}]}], "Input",
 CellLabel->"In[3]:=",
 CellID->322293134],

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{"0.640000000000002`", ",", "Null"}], "}"}]], "Output",
 ImageSize->{85, 15},
 ImageMargins->{{0, 0}, {0, 0}},
 ImageRegion->{{0, 1}, {0, 1}},
 CellLabel->"Out[3]=",
 CellID->9416940],

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{"1.641000000000002`", ",", "Null"}], "}"}]], "Output",
 ImageSize->{92, 15},
 ImageMargins->{{0, 0}, {0, 0}},
 ImageRegion->{{0, 1}, {0, 1}},
 CellLabel->"Out[3]=",
 CellID->92424901],

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{"2.609024107869118`*^-15", ",", "Null"}], "}"}]], "Output",
 ImageSize->{147, 18},
 ImageMargins->{{0, 0}, {0, 0}},
 ImageRegion->{{0, 1}, {0, 1}},
 CellLabel->"Out[3]=",
 CellID->345194208],

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{"8.203000000000001`", ",", "Null"}], "}"}]], "Output",
 ImageSize->{92, 15},
 ImageMargins->{{0, 0}, {0, 0}},
 ImageRegion->{{0, 1}, {0, 1}},
 CellLabel->"Out[3]=",
 CellID->57632146]
}, Open  ]]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell[TextData[ButtonBox["MORE ABOUT",
 BaseStyle->None,
 Appearance->{Automatic, None},
 Evaluator->None,
 Method->"Preemptive",
 ButtonFunction:>(FrontEndExecute[{
    FrontEnd`SelectionMove[
     FrontEnd`SelectedNotebook[], All, ButtonCell], 
    FrontEndToken["OpenCloseGroup"], 
    FrontEnd`SelectionMove[
     FrontEnd`SelectedNotebook[], After, 
     CellContents]}]& )]], "MoreAboutSection",
 CellID->38303248],

Cell[TextData[ButtonBox["RandomGenerator",
 BaseStyle->"Link",
 ButtonData->"paclet:RandomGenerator/guide/RandomGenerator"]], "MoreAbout",
 CellID->283104158]
}, Open  ]],

Cell[" ", "FooterCell"]
},
Saveable->False,
ScreenStyleEnvironment->"Working",
WindowSize->{725, 730},
WindowMargins->{{0, Automatic}, {Automatic, 0}},
WindowTitle->"RandomNegativeBinomial - Wolfram Mathematica",
TaggingRules->{
 "ModificationHighlight" -> False, 
  "Metadata" -> {
   "built" -> "{2013, 2, 10, 20, 28, 31.9801773}", "context" -> 
    "RandomGenerator`RandomDiscreteDistribution`", "keywords" -> {}, "index" -> 
    True, "label" -> "RandomGenerator Package Paclet Symbol", "language" -> 
    "en", "paclet" -> "RandomGenerator Package", "status" -> "None", 
    "summary" -> 
    "RandomNegativeBinomial[n, p, lb, s] gives s random draws from the \
negative binomial distribution with parameters n and p, truncted to be not \
less than lb. The defaults lb=0 and s=1.", "synonyms" -> {}, "title" -> 
    "RandomNegativeBinomial", "type" -> "Symbol", "uri" -> 
    "RandomGenerator/ref/RandomNegativeBinomial"}, "LinkTrails" -> "", 
  "SearchTextTranslated" -> ""},
CellContext->"Global`",
FrontEndVersion->"8.0 for Microsoft Windows (64-bit) (February 23, 2011)",
StyleDefinitions->Notebook[{
   Cell[
    StyleData[
    StyleDefinitions -> FrontEnd`FileName[{"Wolfram"}, "Reference.nb"]]], 
   Cell[
    StyleData["Input"], CellContext -> "Global`"], 
   Cell[
    StyleData["Output"], CellContext -> "Global`"]}, Visible -> False, 
  FrontEndVersion -> "8.0 for Microsoft Windows (64-bit) (February 23, 2011)",
   StyleDefinitions -> "Default.nb"]
]
(* End of Notebook Content *)

(* Internal cache information *)
(*CellTagsOutline
CellTagsIndex->{
 "PrimaryExamplesSection"->{
  Cell[4131, 118, 460, 13, 70, "PrimaryExamplesSection",
   CellTags->"PrimaryExamplesSection",
   CellID->85057143]}
 }
*)
(*CellTagsIndex
CellTagsIndex->{
 {"PrimaryExamplesSection", 14773, 419}
 }
*)
(*NotebookFileOutline
Notebook[{
Cell[615, 21, 76, 1, 70, "SymbolColorBar"],
Cell[694, 24, 200, 5, 70, "LinkTrail"],
Cell[897, 31, 2464, 58, 70, "AnchorBarGrid",
 CellID->1],
Cell[CellGroupData[{
Cell[3386, 93, 65, 1, 70, "ObjectName",
 CellID->1224892054],
Cell[3454, 96, 640, 17, 70, "Usage"]
}, Open  ]],
Cell[CellGroupData[{
Cell[4131, 118, 460, 13, 70, "PrimaryExamplesSection",
 CellTags->"PrimaryExamplesSection",
 CellID->85057143],
Cell[CellGroupData[{
Cell[4616, 135, 148, 5, 70, "ExampleSection",
 CellID->144596749],
Cell[4767, 142, 124, 3, 70, "Input",
 CellID->168232019],
Cell[CellGroupData[{
Cell[4916, 149, 2212, 67, 209, "Input",
 CellID->104442095],
Cell[7131, 218, 3452, 64, 134, "Output",
 CellID->258469057]
}, Open  ]],
Cell[CellGroupData[{
Cell[10620, 287, 996, 30, 70, "Input",
 CellID->322293134],
Cell[11619, 319, 221, 7, 36, "Output",
 CellID->9416940],
Cell[11843, 328, 222, 7, 36, "Output",
 CellID->92424901],
Cell[12068, 337, 229, 7, 39, "Output",
 CellID->345194208],
Cell[12300, 346, 222, 7, 36, "Output",
 CellID->57632146]
}, Open  ]]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[12583, 360, 419, 12, 70, "MoreAboutSection",
 CellID->38303248],
Cell[13005, 374, 158, 3, 70, "MoreAbout",
 CellID->283104158]
}, Open  ]],
Cell[13178, 380, 23, 0, 70, "FooterCell"]
}
]
*)

(* End of internal cache information *)

