(* Content-type: application/vnd.wolfram.mathematica *)

(*** Wolfram Notebook File ***)
(* http://www.wolfram.com/nb *)

(* CreatedBy='Mathematica 9.0' *)

(*CacheID: 234*)
(* Internal cache information:
NotebookFileLineBreakTest
NotebookFileLineBreakTest
NotebookDataPosition[       157,          7]
NotebookDataLength[     13448,        429]
NotebookOptionsPosition[      9800,        327]
NotebookOutlinePosition[     12223,        383]
CellTagsIndexPosition[     12138,        378]
WindowTitle->TreeTransitionProb - Wolfram Mathematica
WindowFrame->Normal*)

(* Beginning of Notebook Content *)
Notebook[{
Cell[" ", "SymbolColorBar",
 CellMargins->{{Inherited, Inherited}, {-5, 0}}],

Cell[TextData[{
 ButtonBox["SMARTree",
  BaseStyle->{"Link", "LinkTrail"},
  ButtonData->"paclet:SMARTree/guide/SMARTree"],
 StyleBox[" > ", "LinkTrailSeparator"]
}], "LinkTrail"],

Cell[BoxData[GridBox[{
   {Cell["SMARTREE PACLET SYMBOL", "PacletNameCell"], Cell[TextData[{
     Cell[BoxData[
      ActionMenuBox[
       FrameBox["\<\"More About \[RightGuillemet]\"\>",
        StripOnInput->False], {"\<\"SMARTree\"\>":>Documentation`HelpLookup[
       "paclet:SMARTree/guide/SMARTree"]},
       Appearance->None,
       MenuAppearance->Automatic]],
      LineSpacing->{1.4, 0}],
     "\[ThickSpace]\[ThickSpace]\[ThickSpace]\[ThickSpace]\[ThickSpace]\
\[ThickSpace]",
     Cell[BoxData[
      ActionMenuBox[
       FrameBox["\<\"URL \[RightGuillemet]\"\>",
        StripOnInput->False], {"\<\"SMARTree/ref/TreeTransitionProb\"\>":>
       None, "\<\"Copy Mathematica url\"\>":>
       Module[{DocumentationSearch`Private`nb$}, 
        DocumentationSearch`Private`nb$ = NotebookPut[
           Notebook[{
             Cell["SMARTree/ref/TreeTransitionProb"]}, Visible -> False]]; 
        SelectionMove[DocumentationSearch`Private`nb$, All, Notebook]; 
        FrontEndTokenExecute[DocumentationSearch`Private`nb$, "Copy"]; 
        NotebookClose[DocumentationSearch`Private`nb$]; Null], 
       Delimiter, "\<\"Copy web url\"\>":>
       Module[{DocumentationSearch`Private`nb$}, 
        DocumentationSearch`Private`nb$ = NotebookPut[
           Notebook[{
             Cell[
              BoxData[
               MakeBoxes[
                Hyperlink[
                "http://reference.wolfram.com/mathematica/SMARTree/ref/\
TreeTransitionProb.html"], StandardForm]], "Input", TextClipboardType -> 
              "PlainText"]}, Visible -> False]]; 
        SelectionMove[DocumentationSearch`Private`nb$, All, Notebook]; 
        FrontEndTokenExecute[DocumentationSearch`Private`nb$, "Copy"]; 
        NotebookClose[DocumentationSearch`Private`nb$]; 
        Null], "\<\"Go to web url\"\>":>FrontEndExecute[{
         NotebookLocate[{
           URL[
            StringJoin[
            "http://reference.wolfram.com/mathematica/", 
             "SMARTree/ref/TreeTransitionProb", ".html"]], None}]}]},
       Appearance->None,
       MenuAppearance->Automatic]],
      LineSpacing->{1.4, 0}]
    }], "AnchorBar"]}
  }]], "AnchorBarGrid",
 GridBoxOptions->{GridBoxItemSize->{"Columns" -> {
     Scaled[0.65], {
      Scaled[0.34]}}, "ColumnsIndexed" -> {}, "Rows" -> {{1.}}, 
   "RowsIndexed" -> {}}},
 CellID->1],

Cell["TreeTransitionProb", "ObjectName",
 CellID->1224892054],

Cell[CellGroupData[{

Cell[BoxData[GridBox[{
   {"", Cell[TextData[{
     Cell[BoxData[
      RowBox[{"TransitionProb", "[", "tree", "]"}]], "InlineFormula"],
     "\[LineSeparator]returns the invisible transtion probabiliy from tree."
    }]]},
   {"", Cell[TextData[{
     Cell[BoxData[
      RowBox[{"TransitionProb", "[", 
       RowBox[{"tree1", ",", "tree2"}], "]"}]], "InlineFormula"],
     "\[LineSeparator]returns the transtion probability between tree1 and \
tree2, based on the mSMC model, for the two trees of distance no greater than \
2. If ",
     StyleBox["tree1",
      FontSlant->"Italic"],
     "=",
     StyleBox["tree2",
      FontSlant->"Italic"],
     ", ",
     StyleBox["TransitionProb[tree1,tree2]",
      FontSlant->"Italic"],
     " = ",
     StyleBox["TransitionProb[tree1]",
      FontSlant->"Italic"],
     "."
    }]]},
   {"", Cell[TextData[{
     Cell[BoxData[
      RowBox[{"TreeTransitionProb", "[", 
       RowBox[{"tree1", ",", 
        RowBox[{"{", 
         RowBox[{"e1", ",", "tc"}], "}"}]}], "]"}]], "InlineFormula"],
     "\[LineSeparator]returns the transtion probability between ",
     StyleBox["tree1",
      FontSlant->"Italic"],
     " and ",
     StyleBox["tree2",
      FontSlant->"Italic"],
     ", where ",
     StyleBox["tree2",
      FontSlant->"Italic"],
     " is produced from ",
     StyleBox["tree1",
      FontSlant->"Italic"],
     " by the mSMC model, under the constraints that the recombination event \
occurs on branch ",
     StyleBox["e1",
      FontSlant->"Italic"],
     " of ",
     StyleBox["tree1",
      FontSlant->"Italic"],
     " and the emerging branch coaleces on a particular branch at time ",
     StyleBox["tc",
      FontSlant->"Italic"],
     "."
    }]]},
   {"", Cell[TextData[{
     Cell[BoxData[
      RowBox[{"TreeTransitionProb", "[", 
       RowBox[{"tree1", ",", 
        RowBox[{"{", 
         RowBox[{"e1", ",", 
          RowBox[{"-", "1"}], ",", "e2"}], "}"}]}], "]"}]], "InlineFormula"],
     "\[LineSeparator]returns the transtion probability between ",
     StyleBox["tree1",
      FontSlant->"Italic"],
     " and ",
     StyleBox["tree2",
      FontSlant->"Italic"],
     ", where ",
     StyleBox["tree2",
      FontSlant->"Italic"],
     " is produced from ",
     StyleBox["tree1",
      FontSlant->"Italic"],
     " by the mSMC model, under the constraints that the recombination event \
occurs on branch ",
     StyleBox["e1",
      FontSlant->"Italic"],
     " of ",
     StyleBox["tree1",
      FontSlant->"Italic"],
     " and the emerging branch coaleces with branch ",
     StyleBox["e2",
      FontSlant->"Italic"],
     " of ",
     StyleBox["tree1",
      FontSlant->"Italic"],
     "."
    }]]}
  }]], "Usage",
 GridBoxOptions->{
 GridBoxBackground->{
  "Columns" -> {{None}}, "ColumnsIndexed" -> {}, 
   "Rows" -> {None, None, None, {None}}, "RowsIndexed" -> {}}},
 CellID->982511436],

Cell[CellGroupData[{

Cell[TextData[Cell[BoxData[
 ButtonBox[
  FrameBox[
   StyleBox[
    RowBox[{"MORE", " ", "INFORMATION"}], "NotesFrameText"],
   StripOnInput->False],
  Appearance->{Automatic, None},
  BaseStyle->None,
  ButtonFunction:>(FrontEndExecute[{
     FrontEnd`SelectionMove[
      FrontEnd`SelectedNotebook[], All, ButtonCell], 
     FrontEndToken["OpenCloseGroup"], 
     FrontEnd`SelectionMove[
      FrontEnd`SelectedNotebook[], After, CellContents]}]& ),
  Evaluator->None,
  Method->"Preemptive"]]]], "NotesSection",
 CellGroupingRules->{"SectionGrouping", 50},
 CellID->671647458],

Cell["\<\
A branch is given by a pair of nodes of the tree. Nodes of a tree is named by \
integers starting from 1.\
\>", "Notes",
 CellID->1067943069]
}, Closed]]
}, Open  ]],

Cell[CellGroupData[{

Cell[TextData[ButtonBox["EXAMPLES",
 BaseStyle->None,
 Appearance->{Automatic, None},
 Evaluator->None,
 Method->"Preemptive",
 ButtonFunction:>(FrontEndExecute[{
    FrontEnd`SelectionMove[
     FrontEnd`SelectedNotebook[], All, ButtonCell], 
    FrontEndToken["OpenCloseGroup"], 
    FrontEnd`SelectionMove[
     FrontEnd`SelectedNotebook[], After, 
     CellContents]}]& )]], "PrimaryExamplesSection",
 CellTags->"PrimaryExamplesSection",
 CellID->7648669],

Cell[CellGroupData[{

Cell[TextData[{
 "Basic Examples",
 "\[NonBreakingSpace]\[NonBreakingSpace]",
 Cell["(1)", "ExampleCount"]
}], "ExampleSection",
 CellID->220797628],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"<<", "SMARTree`"}]], "Input",
 CellLabel->"In[1]:=",
 CellID->623793352],

Cell[BoxData[
 RowBox[{
  StyleBox[
   RowBox[{"General", "::", "compat"}], "MessageName"], ":", 
  " ", "\<\"Combinatorica Graph and Permutations functionality has been \
superseded by preloaded functionaliy. The package now being loaded may \
conflict with this. Please see the Compatibility Guide for details.\"\>"}]], \
"Message", "MSG",
 GeneratedCell->False,
 CellAutoOverwrite->False,
 CellLabel->"During evaluation of In[1]:=",
 CellID->351691725]
}, Open  ]],

Cell[BoxData[{
 RowBox[{
  RowBox[{"tree1", "=", 
   RowBox[{"RandomCoalescentTree", "[", "5", "]"}]}], 
  ";"}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{"tree2", "=", 
   RowBox[{"Last", "[", 
    RowBox[{"RandomNextLocalTree", "[", "tree1", "]"}], "]"}]}], 
  ";"}]}], "Input",
 CellLabel->"In[2]:=",
 CellID->3416801],

Cell[CellGroupData[{

Cell[BoxData[{
 RowBox[{"TreeTransitionProb", "[", "tree1", "]"}], "\[IndentingNewLine]", 
 RowBox[{"TreeTransitionProb", "[", 
  RowBox[{"tree1", ",", "tree2"}], "]"}]}], "Input",
 CellLabel->"In[3]:=",
 CellID->327428633],

Cell[BoxData["0.3789311397914762`"], "Output",
 ImageSize->{60, 15},
 ImageMargins->{{0, 0}, {0, 0}},
 ImageRegion->{{0, 1}, {0, 1}},
 CellLabel->"Out[3]=",
 CellID->336882876],

Cell[BoxData["0.14037476686187428`"], "Output",
 ImageSize->{60, 15},
 ImageMargins->{{0, 0}, {0, 0}},
 ImageRegion->{{0, 1}, {0, 1}},
 CellLabel->"Out[3]=",
 CellID->256885342]
}, Open  ]]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell[TextData[ButtonBox["MORE ABOUT",
 BaseStyle->None,
 Appearance->{Automatic, None},
 Evaluator->None,
 Method->"Preemptive",
 ButtonFunction:>(FrontEndExecute[{
    FrontEnd`SelectionMove[
     FrontEnd`SelectedNotebook[], All, ButtonCell], 
    FrontEndToken["OpenCloseGroup"], 
    FrontEnd`SelectionMove[
     FrontEnd`SelectedNotebook[], After, 
     CellContents]}]& )]], "MoreAboutSection",
 CellID->38303248],

Cell[TextData[ButtonBox["SMARTree",
 BaseStyle->"Link",
 ButtonData->"paclet:SMARTree/guide/SMARTree"]], "MoreAbout",
 CellID->42880380]
}, Open  ]],

Cell[" ", "FooterCell"]
},
Saveable->False,
ScreenStyleEnvironment->"Working",
WindowSize->{725, 730},
WindowMargins->{{0, Automatic}, {Automatic, 0}},
WindowTitle->"TreeTransitionProb - Wolfram Mathematica",
TaggingRules->{
 "ModificationHighlight" -> False, 
  "Metadata" -> {
   "built" -> "{2013, 7, 10, 15, 42, 10.0361750}", "context" -> "SMARTree`", 
    "keywords" -> {}, "index" -> True, "label" -> "SMARTree Paclet Symbol", 
    "language" -> "en", "paclet" -> "SMARTree", "status" -> "None", "summary" -> 
    "TransitionProb[tree] returns the invisible transtion probabiliy from \
tree.TransitionProb[tree1, tree2] returns the transtion probability between \
tree1 and tree2, based on the mSMC model, for the two trees of distance no \
greater than 2. If tree1=tree2, TransitionProb[tree1,tree2] = \
TransitionProb[tree1].TreeTransitionProb[tree1, {e1, tc}] returns the \
transtion probability between tree1 and tree2, where tree2 is produced from \
tree1 by the mSMC model, under the constraints that the recombination event \
occurs on branch e1 of tree1 and the emerging branch coaleces on a particular \
branch at time tc.TreeTransitionProb[tree1, {e1, -1, e2}] returns the \
transtion probability between tree1 and tree2, where tree2 is produced from \
tree1 by the mSMC model, under the constraints that the recombination event \
occurs on branch e1 of tree1 and the emerging branch coaleces with branch e2 \
of tree1.", "synonyms" -> {}, "title" -> "TreeTransitionProb", "type" -> 
    "Symbol", "uri" -> "SMARTree/ref/TreeTransitionProb"}, "LinkTrails" -> "",
   "SearchTextTranslated" -> ""},
CellContext->"Global`",
FrontEndVersion->"9.0 for Microsoft Windows (64-bit) (January 25, 2013)",
StyleDefinitions->Notebook[{
   Cell[
    StyleData[
    StyleDefinitions -> FrontEnd`FileName[{"Wolfram"}, "Reference.nb"]]], 
   Cell[
    StyleData["Input"], CellContext -> "Global`"], 
   Cell[
    StyleData["Output"], CellContext -> "Global`"]}, Visible -> False, 
  FrontEndVersion -> "9.0 for Microsoft Windows (64-bit) (January 25, 2013)", 
  StyleDefinitions -> "Default.nb"]
]
(* End of Notebook Content *)

(* Internal cache information *)
(*CellTagsOutline
CellTagsIndex->{
 "PrimaryExamplesSection"->{
  Cell[6982, 219, 459, 13, 70, "PrimaryExamplesSection",
   CellTags->"PrimaryExamplesSection",
   CellID->7648669]}
 }
*)
(*CellTagsIndex
CellTagsIndex->{
 {"PrimaryExamplesSection", 11997, 371}
 }
*)
(*NotebookFileOutline
Notebook[{
Cell[611, 21, 76, 1, 70, "SymbolColorBar"],
Cell[690, 24, 179, 5, 70, "LinkTrail"],
Cell[872, 31, 2335, 53, 70, "AnchorBarGrid",
 CellID->1],
Cell[3210, 86, 61, 1, 70, "ObjectName",
 CellID->1224892054],
Cell[CellGroupData[{
Cell[3296, 91, 2878, 95, 70, "Usage",
 CellID->982511436],
Cell[CellGroupData[{
Cell[6199, 190, 580, 17, 70, "NotesSection",
 CellGroupingRules->{"SectionGrouping", 50},
 CellID->671647458],
Cell[6782, 209, 151, 4, 70, "Notes",
 CellID->1067943069]
}, Closed]]
}, Open  ]],
Cell[CellGroupData[{
Cell[6982, 219, 459, 13, 70, "PrimaryExamplesSection",
 CellTags->"PrimaryExamplesSection",
 CellID->7648669],
Cell[CellGroupData[{
Cell[7466, 236, 148, 5, 70, "ExampleSection",
 CellID->220797628],
Cell[CellGroupData[{
Cell[7639, 245, 96, 3, 70, "Input",
 CellID->623793352],
Cell[7738, 250, 455, 11, 70, "Message",
 CellID->351691725]
}, Open  ]],
Cell[8208, 264, 324, 11, 70, "Input",
 CellID->3416801],
Cell[CellGroupData[{
Cell[8557, 279, 223, 5, 70, "Input",
 CellID->327428633],
Cell[8783, 286, 176, 5, 36, "Output",
 CellID->336882876],
Cell[8962, 293, 177, 5, 36, "Output",
 CellID->256885342]
}, Open  ]]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[9200, 305, 419, 12, 70, "MoreAboutSection",
 CellID->38303248],
Cell[9622, 319, 136, 3, 70, "MoreAbout",
 CellID->42880380]
}, Open  ]],
Cell[9773, 325, 23, 0, 70, "FooterCell"]
}
]
*)

(* End of internal cache information *)

